Package: BayesTools
Title: Tools for Bayesian Analyses
Version: 0.3.0
Description: Provides tools for conducting Bayesian analyses and Bayesian model averaging 
    (Kass and Raftery, 1995, <doi:10.1080/01621459.1995.10476572>, 
    Hoeting et al., 1999, <doi:10.1214/ss/1009212519>). The package contains 
    functions for creating a wide range of prior distribution objects, mixing posterior 
    samples from 'JAGS' and 'Stan' models, plotting posterior distributions, and etc...
    The tools for working with prior distribution span from visualization, generating 'JAGS' 
    and 'bridgesampling' syntax to basic functions such as rng, quantile, and distribution functions. 
Maintainer: František Bartoš <f.bartos96@gmail.com>
Authors@R: c( 
    person("František", "Bartoš",     role = c("aut", "cre"),
    email   = "f.bartos96@gmail.com", comment = c(ORCID = "0000-0002-0018-5573"))
    )
URL: https://fbartos.github.io/BayesTools/
BugReports: https://github.com/FBartos/BayesTools/issues
License: GPL-3
Encoding: UTF-8
LazyData: true
RoxygenNote: 7.3.3
SystemRequirements: JAGS >= 4.3.0 (https://mcmc-jags.sourceforge.io/)
Depends: R (>= 4.1.0), stats
Imports: graphics, grid, extraDistr, mvtnorm, coda, bridgesampling,
        parallel, ggplot2, Rdpack, rlang
Suggests: scales, testthat (>= 3.3.0), vdiffr, covr, knitr, rstan,
        rjags, runjags, lme4, BayesFactor, RoBMA, rmarkdown
RdMacros: Rdpack
VignetteBuilder: knitr
Config/testthat/edition: 3
NeedsCompilation: no
Packaged: 2026-05-05 14:46:53 UTC; fbart
Author: František Bartoš [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-0018-5573>)
Repository: CRAN
Date/Publication: 2026-05-06 05:20:02 UTC
Built: R 4.5.2; ; 2026-05-06 06:40:29 UTC; unix

Package: BayesTools
Title: Tools for Bayesian Analyses
Version: 0.3.0
Description: Provides tools for conducting Bayesian analyses and Bayesian model averaging 
    (Kass and Raftery, 1995, <doi:10.1080/01621459.1995.10476572>, 
    Hoeting et al., 1999, <doi:10.1214/ss/1009212519>). The package contains 
    functions for creating a wide range of prior distribution objects, mixing posterior 
    samples from 'JAGS' and 'Stan' models, plotting posterior distributions, and etc...
    The tools for working with prior distribution span from visualization, generating 'JAGS' 
    and 'bridgesampling' syntax to basic functions such as rng, quantile, and distribution functions. 
Maintainer: František Bartoš <f.bartos96@gmail.com>
Authors@R: c( 
    person("František", "Bartoš",     role = c("aut", "cre"),
    email   = "f.bartos96@gmail.com", comment = c(ORCID = "0000-0002-0018-5573"))
    )
URL: https://fbartos.github.io/BayesTools/
BugReports: https://github.com/FBartos/BayesTools/issues
License: GPL-3
Encoding: UTF-8
LazyData: true
RoxygenNote: 7.3.3
SystemRequirements: JAGS >= 4.3.0 (https://mcmc-jags.sourceforge.io/)
Depends: R (>= 4.1.0), stats
Imports: graphics, grid, extraDistr, mvtnorm, coda, bridgesampling,
        parallel, ggplot2, Rdpack, rlang
Suggests: scales, testthat (>= 3.3.0), vdiffr, covr, knitr, rstan,
        rjags, runjags, lme4, BayesFactor, RoBMA, rmarkdown
RdMacros: Rdpack
VignetteBuilder: knitr
Config/testthat/edition: 3
NeedsCompilation: no
Packaged: 2026-05-05 14:46:53 UTC; fbart
Author: František Bartoš [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-0018-5573>)
Repository: CRAN
Date/Publication: 2026-05-06 05:20:02 UTC
Built: R 4.6.0; ; 2026-05-06 07:01:08 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(ccdf,prior)
S3method(cdf,prior)
S3method(density,prior)
S3method(lines,prior)
S3method(lpdf,prior)
S3method(mccdf,prior)
S3method(mcdf,prior)
S3method(mean,prior)
S3method(mlpdf,prior)
S3method(mpdf,prior)
S3method(mquant,prior)
S3method(pdf,default)
S3method(pdf,prior)
S3method(plot,prior)
S3method(print,BayesTools_table)
S3method(print,prior)
S3method(quant,prior)
S3method(range,prior)
S3method(rng,prior)
S3method("[",BayesTools_BF)
S3method("[",BayesTools_table)
S3method(sd,default)
S3method(sd,prior)
S3method(update,BayesTools_table)
S3method(var,default)
S3method(var,prior)
export(JAGS_add_priors)
export(JAGS_bridgesampling)
export(JAGS_bridgesampling_posterior)
export(JAGS_check_and_list_autofit_settings)
export(JAGS_check_and_list_fit_settings)
export(JAGS_check_convergence)
export(JAGS_diagnostics)
export(JAGS_diagnostics_autocorrelation)
export(JAGS_diagnostics_density)
export(JAGS_diagnostics_trace)
export(JAGS_estimates_empty_table)
export(JAGS_estimates_table)
export(JAGS_evaluate_formula)
export(JAGS_extend)
export(JAGS_fit)
export(JAGS_formula)
export(JAGS_formula_design)
export(JAGS_get_inits)
export(JAGS_inference_empty_table)
export(JAGS_inference_table)
export(JAGS_marglik_parameters)
export(JAGS_marglik_parameters_formula)
export(JAGS_marglik_priors)
export(JAGS_marglik_priors_formula)
export(JAGS_parameter_names)
export(JAGS_summary_table)
export(JAGS_to_monitor)
export(Savage_Dickey_BF)
export(add_column)
export(as_marginal_inference)
export(as_mixed_posteriors)
export(bridgesampling_object)
export(ccdf)
export(cdf)
export(check_bool)
export(check_char)
export(check_int)
export(check_list)
export(check_real)
export(compute_inference)
export(contr.independent)
export(contr.meandif)
export(contr.orthonormal)
export(dmpoint)
export(dpoint)
export(ensemble_diagnostics_empty_table)
export(ensemble_diagnostics_table)
export(ensemble_estimates_empty_table)
export(ensemble_estimates_table)
export(ensemble_inference)
export(ensemble_inference_empty_table)
export(ensemble_inference_table)
export(ensemble_summary_empty_table)
export(ensemble_summary_table)
export(format_BF)
export(format_parameter_names)
export(formula_add_intercept)
export(geom_prior)
export(geom_prior_list)
export(inclusion_BF)
export(interpret)
export(interpret2)
export(interpret_records)
export(interpret_tables)
export(is.prior)
export(is.prior.PEESE)
export(is.prior.PET)
export(is.prior.discrete)
export(is.prior.factor)
export(is.prior.independent)
export(is.prior.meandif)
export(is.prior.mixture)
export(is.prior.none)
export(is.prior.orthonormal)
export(is.prior.point)
export(is.prior.simple)
export(is.prior.spike_and_slab)
export(is.prior.treatment)
export(is.prior.vector)
export(is.prior.weightfunction)
export(is_prior_bias)
export(is_prior_phacking)
export(lines_prior_list)
export(lpdf)
export(marginal_estimates_table)
export(marginal_inference)
export(marginal_posterior)
export(mccdf)
export(mcdf)
export(mdone.sided)
export(mdone.sided_fixed)
export(mdtwo.sided)
export(mdtwo.sided_fixed)
export(mix_posteriors)
export(mlpdf)
export(model_summary_empty_table)
export(model_summary_table)
export(models_inference)
export(mpdf)
export(mpone.sided)
export(mpone.sided_fixed)
export(mptwo.sided)
export(mptwo.sided_fixed)
export(mqone.sided)
export(mqone.sided_fixed)
export(mqtwo.sided)
export(mqtwo.sided_fixed)
export(mquant)
export(pdf)
export(phack_alpha_from_pi_null)
export(phack_backend_constants)
export(phack_pi_null)
export(plot_marginal)
export(plot_models)
export(plot_posterior)
export(plot_prior_list)
export(plot_transformed_prior)
export(pmpoint)
export(ppoint)
export(prior)
export(prior_PEESE)
export(prior_PET)
export(prior_bias)
export(prior_factor)
export(prior_informed)
export(prior_informed_medicine_names)
export(prior_mixture)
export(prior_none)
export(prior_phacking)
export(prior_spike_and_slab)
export(prior_weightfunction)
export(qmpoint)
export(qpoint)
export(quant)
export(remove_column)
export(rmpoint)
export(rng)
export(rone.sided)
export(rone.sided_fixed)
export(rpoint)
export(rtwo.sided)
export(rtwo.sided_fixed)
export(runjags_estimates_empty_table)
export(runjags_estimates_table)
export(runjags_inference_empty_table)
export(runjags_inference_table)
export(sd)
export(selection_backend_spec)
export(stan_estimates_table)
export(transform_factor_samples)
export(transform_meandif_samples)
export(transform_orthonormal_samples)
export(transform_prior_samples)
export(transform_scale_samples)
export(var)
export(weightfunctions_mapping)
export(wf_cumulative)
export(wf_fixed)
export(wf_independent)
importFrom(Rdpack,reprompt)
importFrom(grDevices,pdf)
importFrom(rlang,.data)
importFrom(stats,density)
importFrom(stats,fft)
importFrom(stats,median)
importFrom(stats,sd)
importFrom(stats,setNames)
importFrom(stats,var)
importFrom(utils,head)
importFrom(utils,tail)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# BayesTools

<!-- badges: start -->

[![R-CMD-check](https://github.com/FBartos/BayesTools/workflows/R-CMD-check/badge.svg)](https://github.com/FBartos/BayesTools/actions)
[![Codecov test
coverage](https://codecov.io/gh/FBartos/BayesTools/branch/master/graph/badge.svg)](https://app.codecov.io/gh/FBartos/BayesTools?branch=master)
[![CRAN_Status_Badge](https://www.r-pkg.org/badges/version/BayesTools)](https://cran.r-project.org/package=BayesTools)
<!-- badges: end -->

The goal of BayesTools is to provide functions that simplify building R
packages focused on Bayesian inference and Bayesian model-averaging.

Currently, the package provides several tools:

-   prior distribution (with S3 methods for plot/print/pdf/cdf/rng/…)
-   JAGS models automation (generating JAGS model syntax and
    `bridgesampling` marginal likelihood functions for prior
    distributions, various wrappers, …)
-   model-averaging (mixing posterior distributions, computing Bayes
    factors, generating summary tables, …)

## Installation

You can install the released version of BayesTools from
[CRAN](https://CRAN.R-project.org) with:

``` r
install.packages("BayesTools")
```

And the development version from [GitHub](https://github.com/) with:

``` r
# install.packages("devtools")
devtools::install_github("FBartos/BayesTools")
```

## Prior Distributions

Prior distribution can be created with the `prior` function.

``` r
library(BayesTools)
#> Loading required namespace: runjags
#> 
#> Attaching package: 'BayesTools'
#> The following objects are masked from 'package:stats':
#> 
#>     sd, var
#> The following object is masked from 'package:grDevices':
#> 
#>     pdf

p0 <- prior(distribution = "point",  parameters = list(location = 0))
p1 <- prior(distribution = "normal", parameters = list(mean = 0, sd = 1))
p2 <- prior(distribution = "normal", parameters = list(mean = 0, sd = 1), truncation = list(0, Inf))
```

The priors can be easily visualized with many possible arguments

``` r
plot(p0)
```

<img src="man/figures/README-unnamed-chunk-3-1.png" width="50%" />

``` r
plot(p1, lwd = 2, col = "blue", par_name = bquote(mu))
```

<img src="man/figures/README-unnamed-chunk-3-2.png" width="50%" />

``` r
plot(p2, plot_type = "ggplot")
```

<img src="man/figures/README-unnamed-chunk-3-3.png" width="50%" />

``` r
plot(p2, plot_type = "ggplot", xlim = c(-2, 2)) + geom_prior(p1, col = "red", lty = 2)
```

<img src="man/figures/README-unnamed-chunk-3-4.png" width="50%" />

``` r
plot(p1, transformation = "exp")
```

<img src="man/figures/README-unnamed-chunk-3-5.png" width="50%" />

All priors also contain some basic S3 methods.

``` r
# S3 methods
set.seed(1)
rng(p0, 10)
#>  [1] 0 0 0 0 0 0 0 0 0 0
rng(p1, 10)
#>  [1] -0.6264538  0.1836433 -0.8356286  1.5952808  0.3295078 -0.8204684
#>  [7]  0.4874291  0.7383247  0.5757814 -0.3053884
rng(p2, 10)
#>  [1] 0.38984324 1.12493092 0.94383621 0.82122120 0.59390132 0.91897737
#>  [7] 0.78213630 0.07456498 0.61982575 0.41794156

pdf(p0, c(-1, 0, 1))
#> [1]   0 Inf   0
pdf(p1, c(-1, 0, 1))
#> [1] 0.2419707 0.3989423 0.2419707
pdf(p2, c(-1, 0, 1))
#> [1] 0.0000000 0.7978846 0.4839414

cdf(p0, c(-1, 0, 1))
#> [1] 0 1 1
cdf(p1, c(-1, 0, 1))
#> [1] 0.1586553 0.5000000 0.8413447
cdf(p2, c(-1, 0, 1))
#> [1] 0.0000000 0.0000000 0.6826895

mean(p0)
#> [1] 0
mean(p1)
#> [1] 0
mean(p2)
#> [1] 0.7978846

sd(p0)
#> [1] 0
sd(p1)
#> [1] 1
sd(p2)
#> [1] 0.6028103

print(p0)
#> Spike(0)
print(p1, short_name = TRUE)
#> N(0, 1)
```

## JAGS Automation

The packages simplifies development of JAGS models by automatically
taking care of the prior distributions relevant portion of the code.

First, we generate few samples from a normal distribution and use the
previously specified prior distributions as priors for the mean (passed
with a list).

``` r
# get some data
set.seed(1)
data <- list(
  x = rnorm(10),
  N = 10
)
data$x
#>  [1] -0.6264538  0.1836433 -0.8356286  1.5952808  0.3295078 -0.8204684
#>  [7]  0.4874291  0.7383247  0.5757814 -0.3053884

## create and fit models
# define priors
priors_list0 <- list(mu = p0)
priors_list1 <- list(mu = p1)
priors_list2 <- list(mu = p2)
```

We create a `model_syntax` that defines likelihood of the data for the
JAGs model and fit the models with the `JAGS_fit` wrapper that
automatically adds prior distributions to the syntax, generates starting
values, creates list of monitored variables, and contains additional
control options (most of the functionality is build upon `runjags`
package).

``` r
# define likelihood for the data
model_syntax <-
  "model{
    for(i in 1:N){
      x[i] ~ dnorm(mu, 1)
    }
  }"

# fit the models
fit0 <- JAGS_fit(model_syntax, data, priors_list0, seed = 0)
#> Loading required namespace: rjags
fit1 <- JAGS_fit(model_syntax, data, priors_list1, seed = 1)
fit2 <- JAGS_fit(model_syntax, data, priors_list2, seed = 2)
```

The `runjags_estimates_table` function then provides a nicely formated
summary for the fitted model.

``` r
# formatted summary tables
runjags_estimates_table(fit1, priors_list1)
#>     Mean    SD 0.025   0.5 0.975 error(MCMC) error(MCMC)/SD   ESS R-hat
#> mu 0.116 0.304 -0.469 0.117 0.715     0.00242          0.008 15748 1.000
```

We create a `log_posterior` function that defines the log likelihood of
the data for marginal likelihood estimation via `bridgesampling` (while
creating a dummy bridge sampling object for the model without any
posterior samples).

``` r
# define log posterior for bridge sampling
log_posterior <- function(parameters, data){
  sum(dnorm(data$x, parameters$mu, 1, log = TRUE))
}

# get marginal likelihoods
marglik0 <- list(
  logml = sum(dnorm(data$x, mean(p0), 1, log = TRUE))
)
class(marglik0) <- "bridge"
marglik1 <- JAGS_bridgesampling(fit1, data, priors_list1, log_posterior)
marglik2 <- JAGS_bridgesampling(fit2, data, priors_list2, log_posterior)

marglik1
#> Bridge sampling estimate of the log marginal likelihood: -13.1383
#> Estimate obtained in 4 iteration(s) via method "normal".
```

## Model-Averaging

The package also simplifies implementation of Bayesian model-averaging
(see e.g., `RoBMA` package).

First, we create a list of model objects, each containing the JAGS fit,
marginal likelihood, list of prior distribution, prior weights, and
generated model summaries. Then we apply the `models_inference`
automatically calculating basic model-averaging information. Finally, we
can use `model_summary_table` to summarize the individual models.

``` r
## create an object with the models
models <- list(
  list(fit = fit0, marglik = marglik0, priors = priors_list0, prior_weights = 1, fit_summary = runjags_estimates_table(fit0, priors_list0)),
  list(fit = fit1, marglik = marglik1, priors = priors_list1, prior_weights = 1, fit_summary = runjags_estimates_table(fit1, priors_list1)),
  list(fit = fit2, marglik = marglik2, priors = priors_list2, prior_weights = 1, fit_summary = runjags_estimates_table(fit2, priors_list2))
)
# compare and summarize the models
models <- models_inference(models)

# create model-summaries
model_summary_table(models[[1]])
#>                                                                 
#>  Model               1             Parameter prior distributions
#>  Prior prob.     0.333                                          
#>  log(marglik)   -12.02                                          
#>  Post. prob.     0.570                                          
#>  Inclusion BF    2.655
model_summary_table(models[[2]])
#>                                                                 
#>  Model               2             Parameter prior distributions
#>  Prior prob.     0.333                         mu ~ Normal(0, 1)
#>  log(marglik)   -13.14                                          
#>  Post. prob.     0.186                                          
#>  Inclusion BF    0.458
model_summary_table(models[[3]])
#>                                                                 
#>  Model               3             Parameter prior distributions
#>  Prior prob.     0.333                 mu ~ Normal(0, 1)[0, Inf]
#>  log(marglik)   -12.87                                          
#>  Post. prob.     0.243                                          
#>  Inclusion BF    0.644
```

Moreover, we can draw inference based on the whole ensemble for the
common parameters with the `ensemble_inference` function, or mixed the
posterior distributions based on marginal likelihoods with the
`mix_posteriors` functions. The various summary functions then create
tables for the inference, estimates, model summary, and MCMC
diagnostics.

``` r
## draw model based inference
inference          <- ensemble_inference(model_list = models, parameters = "mu", is_null_list = list("mu" = 1))

# automatically mix posteriors
mixed_posteriors   <- mix_posteriors(model_list = models, parameters = "mu", is_null_list = list("mu" = 1), seed = 1)

# summarizes the model-averaging summary
ensemble_inference_table(inference, parameters = "mu")
#>    Models Prior prob. Post. prob. Inclusion BF
#> mu    2/3       0.667       0.430        0.377
ensemble_estimates_table(mixed_posteriors,  parameters = "mu")
#>     Mean Median  0.025 0.975
#> mu 0.091  0.000 -0.218 0.523
ensemble_summary_table(models, parameters = "mu")
#>  Model       Prior mu       Prior prob. log(marglik) Post. prob. Inclusion BF
#>      1                            0.333       -12.02       0.570        2.655
#>      2         Normal(0, 1)       0.333       -13.14       0.186        0.458
#>      3 Normal(0, 1)[0, Inf]       0.333       -12.87       0.243        0.644
ensemble_diagnostics_table(models, parameters = "mu", remove_spike_0 = FALSE)
#>  Model       Prior mu       max[error(MCMC)] max[error(MCMC)/SD] min(ESS)
#>      1             Spike(0)               NA                  NA       NA
#>      2         Normal(0, 1)          0.00242               0.008    15748
#>      3 Normal(0, 1)[0, Inf]          0.00162               0.008    16078
#>  max(R-hat)
#>          NA
#>       1.000
#>       1.000
```

The packages also provides functions for plotting model-averaged
posterior distributions.

``` r
### plotting
oldpar <- graphics::par(no.readonly = TRUE)
on.exit(graphics::par(mar = oldpar[["mar"]]))

# plot model-average posteriors
par(mar = c(4, 4, 1, 4))
plot_posterior(mixed_posteriors, parameter = "mu")
```

<img src="man/figures/README-unnamed-chunk-11-1.png" width="50%" />

``` r
plot_posterior(mixed_posteriors, parameter = "mu", lwd = 2, col = "black", prior = TRUE, dots_prior = list(col = "grey", lwd = 2), xlim = c(-2, 2))
```

<img src="man/figures/README-unnamed-chunk-11-2.png" width="50%" />

``` r
plot_posterior(mixed_posteriors, parameter = "mu", transformation = "exp", lwd = 2, col = "red", prior = TRUE, dots_prior = list(col = "blue", lty = 2))
```

<img src="man/figures/README-unnamed-chunk-11-3.png" width="50%" />

Or comparing estimates from the different models.

``` r
plot_models(model_list = models, samples = mixed_posteriors, inference = inference, parameter = "mu", col = "blue")
```

<img src="man/figures/README-unnamed-chunk-12-1.png" width="50%" />

``` r
plot_models(model_list = models, samples = mixed_posteriors, inference = inference, parameter = "mu", prior = TRUE, plot_type = "ggplot")
```

<img src="man/figures/README-unnamed-chunk-12-2.png" width="50%" />

## ----setup, include = FALSE---------------------------------------------------
is_check <- ("CheckExEnv" %in% search()) ||
             any(c("_R_CHECK_TIMINGS_", "_R_CHECK_LICENSE_") %in% names(Sys.getenv()))
knitr::opts_chunk$set(
  collapse = TRUE,
  comment  = "#>",
  eval     = !is_check,
  dev      = "png"
)
if(.Platform$OS.type == "windows"){
  knitr::opts_chunk$set(dev.args = list(type = "cairo"))
}

## -----------------------------------------------------------------------------
library(BayesTools)

## -----------------------------------------------------------------------------
set.seed(-68) # set seed for reproducibility

N     <- 100      # number of observations
x     <- rnorm(N) # continuous predictor
alpha <- -0.5     # intercept
beta  <- 0.15     # small effect

# compute the mean parameter for each predictor value
mu <- alpha + beta * x

# generate the response for each observation 
y  <- rnorm(N, mean = mu, sd = 1) 

## -----------------------------------------------------------------------------
summary(lm(y ~ x))

## -----------------------------------------------------------------------------
model_likelihood <- 
"model{
  for(i in 1:N){
    y[i] ~ dnorm(mu[i], pow(sigma, -2))
  }
}
"

## -----------------------------------------------------------------------------
formula_M0 <- list("mu" = ~ 1)
formula_M1 <- list("mu" = ~ 1 + x)

## -----------------------------------------------------------------------------
# prior on the test parameter
prior_beta  <- prior(distribution = "normal", parameters = list(mean = 0, sd = 0.5))

# priors on the nuisance parameters
prior_int   <- prior(distribution = "normal", parameters = list(mean = 0, sd = 5))
prior_sigma <- prior(distribution = "normal", parameters = list(mean = 0, sd = 5), truncation = list(0, Inf))

# the data list
data_list <- list(
  y = y,
  N = N
)
data_formula <- data.frame(
  x = x
)

## -----------------------------------------------------------------------------
M0 <- JAGS_fit(
  # specification for the `model_likelihood` part
  model_syntax = model_likelihood,
  data         = list(y = y, N = N),
  prior_list   = list("sigma" = prior_sigma),

  # specification for the `formula_M0` part 
  formula_list       = formula_M0,
  formula_prior_list = list("mu" = list("intercept" = prior_int)),
  formula_data_list  = list("mu" = data_formula),
  
  # seed for reproducibility
  seed         = 0
)

M1 <- JAGS_fit(
  model_syntax = model_likelihood,
  data         = list(y = y, N = N),
  prior_list   = list("sigma" = prior_sigma),
  formula_list       = formula_M1,
  formula_prior_list = list("mu" = list("intercept" = prior_int, "x" = prior_beta)),
  formula_data_list  = list("mu" = data_formula),
  seed         = 1
)

## -----------------------------------------------------------------------------
JAGS_estimates_table(M1)

## -----------------------------------------------------------------------------
log_posterior <- function(parameters, data){
  sum(dnorm(
    x    = data[["y"]],
    mean = parameters[["mu"]],
    sd   = parameters[["sigma"]],
    log  = TRUE
  ))
}

## -----------------------------------------------------------------------------
marglik_model_H0 <- JAGS_bridgesampling(
  # specification for the model part
  fit           = M0,
  log_posterior = log_posterior,
  data          = list(y = y, N = N),
  prior_list    = list("sigma" = prior_sigma),

  # specification for the formula` part 
  formula_list       = formula_M0,
  formula_prior_list = list("mu" = list("intercept" = prior_int)),
  formula_data_list  = list("mu" = data_formula)
)

marglik_model_H1 <- JAGS_bridgesampling(
  fit           = M1,
  log_posterior = log_posterior,
  data          = list(y = y, N = N),
  prior_list    = list("sigma" = prior_sigma),
  formula_list       = formula_M1,
  formula_prior_list = list("mu" = list("intercept" = prior_int, "x" = prior_beta)),
  formula_data_list  = list("mu" = data_formula),
)

## -----------------------------------------------------------------------------
models_list <- models_inference(list(
  list(model = M0, marglik = marglik_model_H0, prior_weights = 1/2),
  list(model = M1, marglik = marglik_model_H1, prior_weights = 1/2)
))
ensemble_info <- ensemble_inference(models_list, parameters = "x", is_null_list = list("x" = c(TRUE, FALSE)))

ensemble_inference_table(ensemble_info, parameters = "x")

## -----------------------------------------------------------------------------
prior_beta_spike_and_slab <- prior_spike_and_slab(
  prior_parameter = prior(distribution = "normal", parameters = list(mean = 0, sd = 0.5)),
  prior_inclusion = prior(distribution = "spike", parameters = list(location = 0.5)) 
)


## -----------------------------------------------------------------------------
MS <- JAGS_fit(
  model_syntax = model_likelihood,
  data         = list(y = y, N = N),
  prior_list   = list("sigma" = prior_sigma),
  formula_list       = formula_M1,
  formula_prior_list = list("mu" = list("intercept" = prior_int, "x" = prior_beta_spike_and_slab)),
  formula_data_list  = list("mu" = data_formula),
  seed         = 1
)

## -----------------------------------------------------------------------------
JAGS_estimates_table(MS, conditional = TRUE)

## -----------------------------------------------------------------------------
JAGS_inference_table(MS, BF_diagnostics = TRUE)

## ----setup, include = FALSE---------------------------------------------------
is_check <- ("CheckExEnv" %in% search()) ||
             any(c("_R_CHECK_TIMINGS_", "_R_CHECK_LICENSE_") %in% names(Sys.getenv()))
knitr::opts_chunk$set(
  collapse = TRUE,
  comment  = "#>",
  eval     = !is_check,
  dev      = "png"
)
if(.Platform$OS.type == "windows"){
  knitr::opts_chunk$set(dev.args = list(type = "cairo"))
}

## -----------------------------------------------------------------------------
library(BayesTools)
data("kitchen_rolls")
x <- kitchen_rolls$mean_NEO[kitchen_rolls$rotation == "counter"]
y <- kitchen_rolls$mean_NEO[kitchen_rolls$rotation == "clock"]

## ----fig_dist, fig.width = 4, fig.height = 4, dpi = 300, out.width = "60%", fig.align = "center"----
h1 <- hist(x, breaks = 15, plot = FALSE)
h2 <- hist(y, breaks = 15, plot = FALSE)
par(mar = c(4, 4, 0, 1))
plot(h1, col = rgb(0,0,1,1/4), xlim = c(-1, 2), ylim = c(0, 16), las = 1, main = "", xlab = "mean NEO PI-R")
plot(h2, col = rgb(1,0,0,1/4), add = TRUE)
legend("topright", legend = c("Counter", "Clock"), fill = c(rgb(0,0,1,1/4), rgb(1,0,0,1/4)), bty = "n")

## -----------------------------------------------------------------------------
ttest_model <- 
"model{
  for(i in 1:Nx){
    x[i] ~ dnorm(mu - delta*sigma/2, pow(sigma, -2))
  }
  for(i in 1:Ny){
    y[i] ~ dnorm(mu + delta*sigma/2, pow(sigma, -2))
  }
}
"

## -----------------------------------------------------------------------------
ttest_priors_H0 <- list(
  mu    = prior("cauchy",      parameters = list(location = 0, scale = 10)),
  sigma = prior("exponential", parameters = list(rate = 2)),
  delta = prior("spike",       parameters = list(location = 0))
)
ttest_priors_Hp <- list(
  mu    = prior("cauchy",      parameters = list(location = 0, scale = 10)),
  sigma = prior("exponential", parameters = list(rate = 2)),
  delta = prior("cauchy",      parameters = list(location = 0, scale = 1),
                truncation = list(lower = 0, upper = Inf))
)

## ----fig_priors, fig.width = 8, fig.height = 8, dpi = 300, out.width = "100%", fig.align = "center"----
par(mfrow = c(2, 2), mar = c(4, 4, 2, 1))
plot(ttest_priors_H0$mu,    par_name = bquote(mu), xlim = c(-50, 50))
plot(ttest_priors_H0$sigma, par_name = bquote(sigma), xlim = c(0, 50))
plot(ttest_priors_H0$delta, par_name = bquote(delta), 
     xlim = c(-1, 1), main = bquote(H[0]))
plot(ttest_priors_Hp$delta, par_name = bquote(delta), 
     xlim = c(0, 5), main = bquote(H[1]))

## -----------------------------------------------------------------------------
ttest_data <- list(
  x  = x,
  y  = y,
  Nx = length(x),
  Ny = length(y)
)

## -----------------------------------------------------------------------------
ttest_model_H0 <- JAGS_fit(
  model_syntax = ttest_model,
  data         = ttest_data,
  prior_list   = ttest_priors_H0,
  seed         = 0
) 

ttest_model_Hp <- JAGS_fit(
  model_syntax = ttest_model,
  data         = ttest_data,
  prior_list   = ttest_priors_Hp,
  seed         = 1
) 

## -----------------------------------------------------------------------------
runjags_estimates_table(ttest_model_Hp)

## -----------------------------------------------------------------------------
log_posterior <- function(parameters, data){
  loglik_x <- sum(dnorm(
    x    = data[["x"]],
    mean = parameters[["mu"]] - parameters[["delta"]] * parameters[["sigma"]] / 2,
    sd   = parameters[["sigma"]],
    log  = TRUE
  ))
  loglik_y <- sum(dnorm(
    x    = data[["y"]],
    mean = parameters[["mu"]] + parameters[["delta"]] * parameters[["sigma"]] / 2,
    sd   = parameters[["sigma"]], 
    log  = TRUE
  ))
  return(loglik_x + loglik_y)
}

## -----------------------------------------------------------------------------
marglik_model_H0 <- JAGS_bridgesampling(
  fit           = ttest_model_H0,
  log_posterior = log_posterior,
  data          = ttest_data,
  prior_list    = ttest_priors_H0
)

marglik_model_Hp <- JAGS_bridgesampling(
  fit           = ttest_model_Hp,
  log_posterior = log_posterior,
  data          = ttest_data,
  prior_list    = ttest_priors_Hp
)

## -----------------------------------------------------------------------------
bridgesampling::bf(marglik_model_H0, marglik_model_Hp)

## -----------------------------------------------------------------------------
models_list <- models_inference(list(
  list(model = ttest_model_H0, marglik = marglik_model_H0, prior_weights = 1/2),
  list(model = ttest_model_Hp, marglik = marglik_model_Hp, prior_weights = 1/2)
))
ensemble_info <- ensemble_inference(models_list, parameters = "delta", is_null_list = list("delta" = c(TRUE, FALSE)))

ensemble_inference_table(ensemble_info, "delta", BF01 = TRUE)

## -----------------------------------------------------------------------------
BayesFactor_ttest <- BayesFactor::ttestBF(x = x, y = y, rscale = "wide", nullInterval = c(0, Inf))
BayesFactor_ttest
1/exp(BayesFactor_ttest@bayesFactor$bf[2])

#' @title Creates a prior distribution for a p-hacking selection kernel
#'
#' @description \code{prior_phacking()} creates a prior distribution for a
#' mass-preserving p-hacking redistribution kernel. The initial backend supports
#' linear and quadratic power depletion between \code{source} and
#' \code{target}, with depleted probability mass redistributed between
#' \code{target} and \code{destination}.
#'
#' @param side side geometry. Currently only \code{"one-sided"} is supported.
#' @param target target p-value cut point.
#' @param source source p-value cut point. Must be larger than \code{target}.
#' @param destination destination p-value cut point. Must be smaller than
#' \code{target}.
#' @param form power depletion form, either \code{"linear"} or
#' \code{"quadratic"}.
#' @param alpha prior distribution for the p-hacking severity parameter.
#' @param report_scale reporting scale for deterministic summaries.
#' @param prior_weights prior odds associated with a given distribution.
#'
#' @return \code{prior_phacking()} returns an object of class \code{"prior"}.
#'
#' @export
prior_phacking <- function(side = "one-sided",
                           target = .025,
                           source = .25,
                           destination = .005,
                           form = c("linear", "quadratic"),
                           alpha = prior("beta", list(1, 1)),
                           report_scale = "pi_null",
                           prior_weights = 1){

  check_char(side, "side")
  check_real(target, "target", lower = 0, upper = 1, allow_bound = FALSE, allow_NA = FALSE)
  check_real(source, "source", lower = 0, upper = 1, allow_bound = FALSE, allow_NA = FALSE)
  check_real(destination, "destination", lower = 0, upper = 1, allow_bound = FALSE, allow_NA = FALSE)
  form <- match.arg(form)
  check_char(report_scale, "report_scale", allow_values = c("pi_null", "alpha"))
  check_real(prior_weights, "prior_weights", lower = 0, allow_bound = FALSE)

  side <- .weightfunction_normalize_side(side)
  if(side != "one-sided"){
    stop("P-hacking priors currently support only one-sided p-value geometry.", call. = FALSE)
  }

  if(!(destination < target && target < source)){
    stop("P-hacking p-value cut points must satisfy destination < target < source.", call. = FALSE)
  }

  .phack_validate_alpha_prior(alpha)
  constants <- phack_backend_constants(form = form, source = source, destination = destination, target = target)

  output <- list(
    distribution  = "phacking",
    side          = side,
    target        = target,
    source        = source,
    destination   = destination,
    form          = form,
    q             = constants$q,
    alpha         = alpha,
    report_scale  = report_scale,
    parameters    = list(
      target      = target,
      source      = source,
      destination = destination,
      form        = form
    ),
    truncation    = list(lower = 0, upper = 1),
    prior_weights = prior_weights
  )

  class(output) <- c("prior", "prior.phacking")

  return(output)
}

#' @title Creates a composed publication-bias prior
#'
#' @description \code{prior_bias()} composes a step-selection
#' \code{prior_weightfunction()} and/or a \code{prior_phacking()} object. It does
#' not introduce new math; the backend compiler treats the composition as the
#' product of both kernels.
#'
#' @param selection optional \code{prior_weightfunction()} object.
#' @param phacking optional \code{prior_phacking()} object.
#' @inheritParams prior_phacking
#'
#' @return \code{prior_bias()} returns an object of class \code{"prior"}.
#'
#' @export
prior_bias <- function(selection = NULL, phacking = NULL, prior_weights = 1){

  check_real(prior_weights, "prior_weights", lower = 0, allow_bound = FALSE)

  if(is.null(selection) && is.null(phacking)){
    stop("At least one of 'selection' or 'phacking' must be specified.", call. = FALSE)
  }
  if(!is.null(selection) && !is.prior.weightfunction(selection)){
    stop("'selection' must be a weightfunction prior created by prior_weightfunction().", call. = FALSE)
  }
  if(!is.null(phacking) && !is_prior_phacking(phacking)){
    stop("'phacking' must be a p-hacking prior created by prior_phacking().", call. = FALSE)
  }
  if(!is.null(selection) && !is.null(phacking) && selection$side != phacking$side){
    stop("'selection' and 'phacking' must use the same side geometry.", call. = FALSE)
  }

  output <- list(
    distribution  = "bias",
    selection     = selection,
    phacking      = phacking,
    parameters    = list(),
    truncation    = list(lower = 0, upper = Inf),
    prior_weights = prior_weights
  )

  class(output) <- c("prior", "prior.bias")

  return(output)
}

#' @title Reports whether x is a p-hacking prior
#'
#' @param x object to test.
#'
#' @return A logical value.
#'
#' @export
is_prior_phacking <- function(x){
  inherits(x, "prior.phacking")
}

#' @title Reports whether x is a composed publication-bias prior
#'
#' @inheritParams is_prior_phacking
#'
#' @return A logical value.
#'
#' @export
is_prior_bias <- function(x){
  inherits(x, "prior.bias")
}

is.prior.phacking <- function(x){
  is_prior_phacking(x)
}

is.prior.bias <- function(x){
  is_prior_bias(x)
}

.phacking_report_parameter <- function(prior){

  if(!is_prior_phacking(prior)){
    stop("'prior' must be a p-hacking prior.", call. = FALSE)
  }

  if(is.null(prior$report_scale)){
    return("pi_null")
  }

  prior$report_scale
}

.phacking_unreported_parameters <- function(prior){

  setdiff(c("alpha", "pi_null"), .phacking_report_parameter(prior))
}

.selection_phacking_report_parameters <- function(phacking_priors){

  if(length(phacking_priors) == 0L){
    return(character())
  }

  unique(vapply(phacking_priors, .phacking_report_parameter, character(1)))
}

.selection_phacking_unreported_parameters <- function(phacking_priors){

  setdiff(c("alpha", "pi_null"), .selection_phacking_report_parameters(phacking_priors))
}

.selection_prior_phacking_report_parameters <- function(prior){

  .selection_phacking_report_parameters(.selection_prior_phacking_priors(prior))
}

#' @title P-hacking calibration helpers
#'
#' @description \code{phack_pi_null()} converts the sampled severity
#' \code{alpha} to the amount of null probability mass depleted from the source
#' interval. \code{phack_alpha_from_pi_null()} applies the inverse calibration.
#' \code{phack_backend_constants()} returns the z-scale constants used by the
#' backend.
#'
#' @param alpha p-hacking severity parameter.
#' @param pi_null null probability mass depleted from the source interval.
#' @param form power depletion form, either \code{"linear"} or
#' \code{"quadratic"}.
#' @param source source p-value cut point.
#' @param destination destination p-value cut point.
#' @param target target p-value cut point.
#'
#' @return Numeric vector for the calibration helpers and a named list for
#' \code{phack_backend_constants()}.
#'
#' @export
phack_pi_null <- function(alpha, form, source, destination, target = .025){

  check_real(alpha, "alpha", lower = 0, upper = 1, check_length = 0, allow_NA = FALSE)
  if(any(alpha >= 1)){
    stop("'alpha' must be lower than 1.", call. = FALSE)
  }
  constants <- phack_backend_constants(form = form, source = source, destination = destination, target = target)

  return(alpha * constants$pi_null_per_alpha)
}

#' @rdname phack_pi_null
#' @export
phack_alpha_from_pi_null <- function(pi_null, form, source, destination, target = .025){

  check_real(pi_null, "pi_null", lower = 0, check_length = 0, allow_NA = FALSE)
  constants <- phack_backend_constants(form = form, source = source, destination = destination, target = target)

  if(any(pi_null >= constants$pi_null_per_alpha)){
    stop("'pi_null' is too large for alpha <= 1 under the specified p-hacking geometry.", call. = FALSE)
  }

  return(pi_null / constants$pi_null_per_alpha)
}

#' @rdname phack_pi_null
#' @export
phack_backend_constants <- function(form, source, destination, target = .025){

  form <- match.arg(form, choices = c("linear", "quadratic"))
  check_real(target, "target", lower = 0, upper = 1, allow_bound = FALSE, allow_NA = FALSE)
  check_real(source, "source", lower = 0, upper = 1, allow_bound = FALSE, allow_NA = FALSE)
  check_real(destination, "destination", lower = 0, upper = 1, allow_bound = FALSE, allow_NA = FALSE)

  if(!(destination < target && target < source)){
    stop("P-hacking p-value cut points must satisfy destination < target < source.", call. = FALSE)
  }

  q <- switch(
    form,
    "linear"    = 1L,
    "quadratic" = 2L
  )

  z_source_lower <- stats::qnorm(1 - source)
  z_target       <- stats::qnorm(1 - target)
  z_dest_upper   <- stats::qnorm(1 - destination)

  source_mass <- .phack_power_null_moment(z_source_lower, z_target, q, anchor = z_source_lower, reverse = FALSE)
  dest_mass   <- .phack_power_null_moment(z_target, z_dest_upper, q, anchor = z_dest_upper, reverse = TRUE)

  return(list(
    form                    = form,
    q                       = q,
    phack_kind              = .phack_kind(form),
    target                  = target,
    source                  = source,
    destination             = destination,
    z_source                = c(z_source_lower, z_target),
    z_destination           = c(z_target, z_dest_upper),
    source_null_mass        = source_mass,
    destination_null_mass   = dest_mass,
    pi_null_per_alpha       = source_mass,
    beta_null_per_alpha     = source_mass / dest_mass
  ))
}

#' @title Compile selection priors for backend consumers
#'
#' @description \code{selection_backend_spec()} compiles step-selection and
#' p-hacking prior objects into active backend parameters. The returned object
#' contains stable p/z geometry, JAGS prior/transform code, monitor names,
#' initial values, and data constants.
#'
#' @param priors a selection prior, p-hacking prior, composed bias prior,
#' \code{prior_none()}, \code{prior_mixture()}, or a list of those priors.
#' @param backend backend target. Currently only \code{"jags"} is supported.
#' @param names list of backend parameter names.
#' @param global_breaks optional global p-value break grid.
#'
#' @return A list describing the compiled backend specification.
#'
#' @export
selection_backend_spec <- function(priors,
                                   backend = "jags",
                                   names = list(omega = "omega", alpha = "alpha"),
                                   global_breaks = NULL){

  check_char(backend, "backend", allow_values = "jags")
  check_list(names, "names", check_names = c("omega", "alpha", "pi_null", "beta_null", "phack_kind", "phack_z_source", "phack_z_dest", "phack_z_destination"), allow_other = FALSE)
  names <- .selection_backend_names(names)

  branches <- .selection_normalize_priors(priors)
  branch_info <- lapply(branches, .selection_branch_info)

  has_selection <- vapply(branch_info, function(x) !is.null(x$selection), logical(1))
  has_phacking  <- vapply(branch_info, function(x) !is.null(x$phacking),  logical(1))
  branch_type   <- vapply(branch_info, function(x) x$type, character(1))

  mode <- .selection_backend_mode(any(has_selection), any(has_phacking))

  step_priors <- lapply(branch_info[has_selection], function(x) x$selection)
  if(is.null(global_breaks)){
    breaks <- if(length(step_priors) > 0L){
      weightfunctions_mapping(step_priors, cuts_only = TRUE, one_sided = TRUE)
    }else{
      c(0, 1)
    }
  }else{
    breaks <- .selection_validate_global_breaks(global_breaks)
    if(length(step_priors) > 0L){
      required_breaks <- weightfunctions_mapping(step_priors, cuts_only = TRUE, one_sided = TRUE)
      if(!all(vapply(required_breaks, function(x) any(abs(x - breaks) < sqrt(.Machine$double.eps)), logical(1)))){
        stop("'global_breaks' must contain all step-selection p-value breaks.", call. = FALSE)
      }
    }
  }
  n_bins <- length(breaks) - 1L

  prior_weights <- vapply(branches, function(x) x$prior_weights, numeric(1))
  uses_indicator <- length(branches) > 1L
  indicator_terms <- if(uses_indicator){
    paste0("equals(bias_indicator, ", seq_along(branches), ")")
  }else{
    rep("1", length(branches))
  }

  prior_code <- character()
  transform_code <- character()

  if(uses_indicator){
    prior_code <- c(prior_code, paste0("bias_indicator ~ dcat(c(", paste0(prior_weights, collapse = ", "), "))"))
  }

  for(i in seq_along(branches)){
    component_id <- if(uses_indicator) i else NULL
    step_code <- if(uses_indicator || !is.null(branch_info[[i]]$selection) || !is.null(branch_info[[i]]$phacking)){
      .selection_jags_step_component_code(branch_info[[i]]$selection, component_id = component_id, n_bins = n_bins, global_cuts = breaks)
    }else{
      character()
    }
    phacking_code <- if(uses_indicator || !is.null(branch_info[[i]]$phacking)){
      .JAGS_phacking_component_syntax(branch_info[[i]]$phacking, component_id = component_id)
    }else{
      character()
    }
    prior_code <- c(
      prior_code,
      step_code,
      phacking_code
    )
  }

  if(uses_indicator){
    for(j in seq_len(n_bins)){
      transform_code <- c(
        transform_code,
        paste0(
          names$omega, "[", j, "] <- ",
          paste0("omega_component_", seq_along(branches), "[", j, "] * ", indicator_terms, collapse = " + ")
        )
      )
    }

    transform_code <- c(
      transform_code,
      .selection_jags_active_scalar("alpha_component_", names$alpha, indicator_terms, seq_along(branches)),
      .selection_jags_active_scalar("phack_kind_component_", names$phack_kind, indicator_terms, seq_along(branches)),
      .selection_jags_active_scalar("pi_null_component_", names$pi_null, indicator_terms, seq_along(branches)),
      .selection_jags_active_scalar("beta_null_component_", names$beta_null, indicator_terms, seq_along(branches))
    )

    if(any(has_phacking)){
      for(k in 1:2){
        transform_code <- c(
          transform_code,
          .selection_jags_active_vector("phack_z_source_component_", names$phack_z_source, k, indicator_terms, seq_along(branches)),
          .selection_jags_active_vector("phack_z_dest_component_", names$phack_z_dest, k, indicator_terms, seq_along(branches))
        )
      }
    }
  }

  monitor <- character()
  if(uses_indicator){
    monitor <- c(monitor, "bias_indicator")
  }
  if(any(has_selection) || any(has_phacking)){
    monitor <- c(monitor, names$omega)
  }
  if(!uses_indicator && any(has_selection)){
    monitor <- c(monitor, .JAGS_monitor_private.weightfunction(branch_info[[which(has_selection)[1L]]]$selection))
  }
  if(any(has_phacking)){
    monitor <- c(monitor, names$alpha, names$phack_kind, names$pi_null)
  }

  phacking_priors <- lapply(branch_info[has_phacking], function(x) x$phacking)
  phacking <- .selection_backend_phacking_info(phacking_priors, names)

  init <- .selection_backend_init(branch_info, breaks, prior_weights, names, uses_indicator)

  return(list(
    mode           = mode,
    branch_type    = branch_type,
    prior_weights  = prior_weights,
    step           = list(
      breaks          = breaks,
      n_bins          = as.integer(n_bins),
      coefficient     = names$omega,
      coefficient_ids = paste0(names$omega, "[", seq_len(n_bins), "]"),
      z_lower         = stats::qnorm(1 - breaks[-1]),
      z_upper         = stats::qnorm(1 - breaks[-length(breaks)])
    ),
    phacking      = phacking,
    prior_code    = paste0(prior_code[nzchar(prior_code)], collapse = "\n"),
    transform_code = paste0(transform_code[nzchar(transform_code)], collapse = "\n"),
    monitor       = unique(monitor),
    init          = init,
    data          = list(
      sel_p_cuts       = breaks,
      sel_z_lower      = stats::qnorm(1 - breaks[-1]),
      sel_z_upper      = stats::qnorm(1 - breaks[-length(breaks)]),
      sel_n_bins       = as.integer(n_bins),
      phack_component_z_source = phacking$branch_z_source,
      phack_component_z_dest   = phacking$branch_z_destination,
      phack_component_q        = phacking$branch_q,
      phack_component_beta_null_per_alpha = phacking$branch_beta_null_per_alpha,
      kernel_mode      = .selection_mode_code(mode)
    )
  ))
}

.phack_validate_alpha_prior <- function(alpha){

  .check_prior(alpha)

  if(!is.prior.simple(alpha)){
    stop("'alpha' must be a simple prior distribution.", call. = FALSE)
  }
  if(is.prior.discrete(alpha) && !is.prior.point(alpha)){
    stop("'alpha' must be a continuous prior distribution or a point prior.", call. = FALSE)
  }

  if(is.prior.point(alpha)){
    if(alpha$parameters[["location"]] < 0 || alpha$parameters[["location"]] >= 1){
      stop("Point p-hacking alpha priors must be in the interval [0, 1).", call. = FALSE)
    }
  }else{
    if(alpha$truncation[["lower"]] < 0 || alpha$truncation[["upper"]] > 1){
      stop("P-hacking alpha priors must have support within [0, 1].", call. = FALSE)
    }
  }

  return()
}

.phack_kind <- function(form){
  switch(
    form,
    "linear"    = 1L,
    "quadratic" = 2L
  )
}

.phack_power_null_moment <- function(lower, upper, q, anchor, reverse){

  width <- upper - lower
  m0 <- stats::pnorm(upper) - stats::pnorm(lower)
  m1 <- stats::dnorm(lower) - stats::dnorm(upper)

  if(q == 1L){
    if(reverse){
      return((anchor * m0 - m1) / width)
    }else{
      return((m1 - anchor * m0) / width)
    }
  }

  m2 <- m0 + lower * stats::dnorm(lower) - upper * stats::dnorm(upper)
  if(reverse){
    return((m2 - 2 * anchor * m1 + anchor^2 * m0) / width^2)
  }

  return((m2 - 2 * anchor * m1 + anchor^2 * m0) / width^2)
}

.selection_backend_names <- function(names){

  backend_names <- names
  if(!is.null(backend_names$phack_z_destination) && is.null(backend_names$phack_z_dest)){
    backend_names$phack_z_dest <- backend_names$phack_z_destination
  }
  backend_names$phack_z_destination <- NULL

  defaults <- list(
    omega               = "omega",
    alpha               = "alpha",
    pi_null             = "pi_null",
    beta_null           = "beta_null",
    phack_kind          = "phack_kind",
    phack_z_source      = "phack_z_source",
    phack_z_dest        = "phack_z_dest"
  )
  defaults[base::names(backend_names)] <- backend_names

  for(i in seq_along(defaults)){
    check_char(defaults[[i]], paste0("names$", names(defaults)[i]))
  }

  return(defaults)
}

.selection_normalize_priors <- function(priors){

  if(is.null(priors)){
    priors <- list(prior_none())
  }else if(is.prior(priors) && is.prior.mixture(priors)){
    priors <- as.list(priors)
  }else if(is.prior(priors)){
    priors <- list(priors)
  }else if(is.list(priors)){
    if(length(priors) == 0L){
      priors <- list(prior_none())
    }
  }else{
    stop("'priors' must be a prior, prior mixture, list of priors, or NULL.", call. = FALSE)
  }

  if(!all(vapply(priors, is.prior, logical(1)))){
    stop("'priors' must contain only prior objects.", call. = FALSE)
  }

  allowed <- vapply(priors, function(x){
    is.prior.none(x) || is.prior.PET(x) || is.prior.PEESE(x) ||
      is.prior.weightfunction(x) || is_prior_phacking(x) || is_prior_bias(x)
  }, logical(1))

  if(!all(allowed)){
    stop("'priors' contains unsupported selection prior objects.", call. = FALSE)
  }

  return(priors)
}

.selection_branch_info <- function(prior){

  if(is.prior.none(prior)){
    return(list(type = "none", selection = NULL, phacking = NULL))
  }
  if(is.prior.PET(prior)){
    return(list(type = "PET", selection = NULL, phacking = NULL))
  }
  if(is.prior.PEESE(prior)){
    return(list(type = "PEESE", selection = NULL, phacking = NULL))
  }
  if(is.prior.weightfunction(prior)){
    return(list(type = "weightfunction", selection = prior, phacking = NULL))
  }
  if(is_prior_phacking(prior)){
    return(list(type = "phack", selection = NULL, phacking = prior))
  }
  if(is_prior_bias(prior)){
    has_selection <- !is.null(prior$selection)
    has_phacking  <- !is.null(prior$phacking)
    type <- if(has_selection && has_phacking){
      "combined"
    }else if(has_selection){
      "weightfunction"
    }else{
      "phack"
    }
    return(list(type = type, selection = prior$selection, phacking = prior$phacking))
  }

  stop("Unsupported selection prior object.", call. = FALSE)
}

.selection_backend_mode <- function(has_selection, has_phacking){

  if(has_selection && has_phacking){
    return("step_phack_power")
  }
  if(has_selection){
    return("step")
  }
  if(has_phacking){
    return("phack_power")
  }

  return("none")
}

.selection_mode_code <- function(mode){
  switch(
    mode,
    "none"             = 0L,
    "step"             = 1L,
    "phack_power"      = 2L,
    "step_phack_power" = 3L
  )
}

.selection_validate_global_breaks <- function(global_breaks){

  check_real(global_breaks, "global_breaks", check_length = 0, allow_NA = FALSE)
  if(length(global_breaks) < 2L){
    stop("'global_breaks' must contain at least 0 and 1.", call. = FALSE)
  }
  if(any(global_breaks < 0 | global_breaks > 1)){
    stop("'global_breaks' must be within [0, 1].", call. = FALSE)
  }
  if(anyDuplicated(global_breaks)){
    stop("'global_breaks' must not contain duplicate values.", call. = FALSE)
  }
  if(!all(global_breaks == cummax(global_breaks))){
    stop("'global_breaks' must be monotonically increasing.", call. = FALSE)
  }
  if(!isTRUE(all.equal(global_breaks[1], 0)) || !isTRUE(all.equal(global_breaks[length(global_breaks)], 1))){
    stop("'global_breaks' must start at 0 and end at 1.", call. = FALSE)
  }

  return(global_breaks)
}

.selection_jags_step_component_code <- function(selection, component_id, n_bins, global_cuts){

  if(is.null(selection)){
    return(.JAGS_weightfunction_none_component_syntax(component_id = component_id, n_bins = n_bins))
  }

  return(.JAGS_weightfunction_component_syntax(
    prior           = selection,
    component_id    = component_id,
    global_cuts     = global_cuts,
    force_one_sided = TRUE
  ))
}

.selection_jags_active_scalar <- function(prefix, target, indicator_terms, component_ids){

  paste0(
    target, " <- ",
    paste0(prefix, component_ids, " * ", indicator_terms, collapse = " + ")
  )
}

.selection_jags_active_vector <- function(prefix, target, index, indicator_terms, component_ids){

  paste0(
    target, "[", index, "] <- ",
    paste0(prefix, component_ids, "[", index, "] * ", indicator_terms, collapse = " + ")
  )
}

.selection_backend_phacking_info <- function(phacking_priors, names){

  if(length(phacking_priors) == 0L){
    return(list(
      form                 = "none",
      q                    = 0L,
      z_source             = c(0, 0),
      z_destination        = c(0, 0),
      coefficient          = names$alpha,
      coefficient_ids      = names$alpha,
      branch_form          = character(),
      branch_q             = integer(),
      branch_phack_kind    = integer(),
      branch_beta_null_per_alpha = numeric(),
      branch_z_source      = matrix(nrow = 0L, ncol = 2L),
      branch_z_destination = matrix(nrow = 0L, ncol = 2L)
    ))
  }

  constants <- lapply(phacking_priors, function(x){
    phack_backend_constants(x$form, x$source, x$destination, target = x$target)
  })

  forms <- vapply(constants, function(x) x$form, character(1))
  q <- vapply(constants, function(x) x$q, integer(1))
  z_source <- do.call(rbind, lapply(constants, function(x) x$z_source))
  z_destination <- do.call(rbind, lapply(constants, function(x) x$z_destination))

  return(list(
    form                 = if(length(unique(forms)) == 1L) forms[1] else unique(forms),
    q                    = if(length(unique(q)) == 1L) q[1] else unique(q),
    z_source             = if(nrow(z_source) == 1L) as.numeric(z_source[1,]) else z_source,
    z_destination        = if(nrow(z_destination) == 1L) as.numeric(z_destination[1,]) else z_destination,
    coefficient          = names$alpha,
    coefficient_ids      = names$alpha,
    branch_form          = forms,
    branch_q             = q,
    branch_phack_kind    = vapply(constants, function(x) x$phack_kind, integer(1)),
    branch_beta_null_per_alpha = vapply(constants, function(x) x$beta_null_per_alpha, numeric(1)),
    branch_z_source      = z_source,
    branch_z_destination = z_destination
  ))
}

.selection_backend_init <- function(branch_info, breaks, prior_weights, names, uses_indicator){

  active_branch <- which.max(prior_weights)

  init <- list()
  for(i in seq_along(branch_info)){
    component_id <- if(uses_indicator) i else NULL
    if(!is.null(branch_info[[i]]$selection)){
      init <- c(init, .selection_JAGS_init_weightfunction_component(branch_info[[i]]$selection, component_id = component_id))
    }
    if(!is.null(branch_info[[i]]$phacking)){
      init <- c(init, .selection_JAGS_init_phacking_component(branch_info[[i]]$phacking, component_id = component_id))
    }
  }
  if(uses_indicator){
    init[["bias_indicator"]] <- active_branch
  }

  return(init)
}

.selection_JAGS_init_weightfunction_component <- function(prior, component_id = NULL){

  init <- list()
  if(prior$weights$type == "fixed"){
    return()
  }else if(prior$weights$type == "cumulative"){
    eta_name <- if(is.null(component_id)) "eta" else paste0("eta_component_", component_id)
    init[[eta_name]] <- stats::rgamma(length(prior$weights[["alpha"]]), shape = prior$weights[["alpha"]], rate = 1)
  }

  return(init)
}

.selection_JAGS_init_phacking_component <- function(prior, component_id = NULL){

  alpha_name <- if(is.null(component_id)) "alpha" else paste0("alpha_component_", component_id)
  .JAGS_init.simple(prior$alpha, alpha_name)
}

.selection_prior_branch_info <- function(prior){

  if(is.prior.mixture(prior)){
    return(lapply(prior, .selection_branch_info))
  }

  list(.selection_branch_info(prior))
}

.selection_prior_has_selection <- function(prior){

  branch_info <- .selection_prior_branch_info(prior)
  any(vapply(branch_info, function(x) !is.null(x$selection), logical(1)))
}

.selection_prior_has_phacking <- function(prior){

  branch_info <- .selection_prior_branch_info(prior)
  any(vapply(branch_info, function(x) !is.null(x$phacking), logical(1)))
}

.selection_prior_selection_priors <- function(prior){

  branch_info <- .selection_prior_branch_info(prior)
  selection_priors <- lapply(branch_info, function(x) x$selection)
  selection_priors[!vapply(selection_priors, is.null, logical(1))]
}

.selection_prior_phacking_priors <- function(prior){

  branch_info <- .selection_prior_branch_info(prior)
  phacking_priors <- lapply(branch_info, function(x) x$phacking)
  phacking_priors[!vapply(phacking_priors, is.null, logical(1))]
}

.selection_prior_stop_unsupported_generic <- function(generic, prior){

  if(is_prior_bias(prior)){
    selection_backend_spec(prior)
    stop(
      sprintf(
        "No %s is implemented for composed bias priors; use the selection or p-hacking component explicitly.",
        generic
      ),
      call. = FALSE
    )
  }

  if(is_prior_phacking(prior)){
    selection_backend_spec(prior)
    stop(
      sprintf(
        "No %s is implemented for p-hacking priors; use the alpha prior component explicitly.",
        generic
      ),
      call. = FALSE
    )
  }

  stop(sprintf("No %s is implemented for this prior.", generic), call. = FALSE)
}

.selection_prior_has_PET <- function(prior){

  if(is.prior.mixture(prior)){
    return(any(sapply(prior, is.prior.PET)))
  }

  is.prior.PET(prior)
}

.selection_prior_has_PEESE <- function(prior){

  if(is.prior.mixture(prior)){
    return(any(sapply(prior, is.prior.PEESE)))
  }

  is.prior.PEESE(prior)
}

.selection_bias_parameter_names <- function(prior, include_kind = TRUE){

  c(
    if(.selection_prior_has_PET(prior)) "PET",
    if(.selection_prior_has_PEESE(prior)) "PEESE",
    if(.selection_prior_has_selection(prior)) "omega",
    if(.selection_prior_has_phacking(prior)) c(
      .selection_prior_phacking_report_parameters(prior),
      if(include_kind) c("beta_null", "phack_kind")
    )
  )
}

.selection_initial_omega <- function(selection, breaks){

  local <- .selection_weightfunction_mean(selection)
  expansion <- .weightfunction_mapping_expansion(selection, force_one_sided = TRUE)
  global_bin_indices <- .weightfunction_global_bin_indices(breaks, expansion)

  vapply(seq_len(length(breaks) - 1L), function(i){
    ind <- global_bin_indices[i]
    local[expansion$index[ind]]
  }, numeric(1))
}

.selection_weightfunction_mean <- function(selection){

  components <- .weightfunction_marginal_components(selection)
  vapply(components, .prior_weightfunction_component_mean, numeric(1))
}

.selection_prior_mean <- function(prior){

  if(is.prior.point(prior)){
    return(prior$parameters[["location"]])
  }

  return(mean(prior))
}

.selection_format_number <- function(x){
  format(x, scientific = FALSE, digits = 16, trim = TRUE)
}

.JAGS_phacking_component_syntax <- function(prior, component_id = NULL){

  alpha_name <- if(is.null(component_id)) "alpha" else paste0("alpha_component_", component_id)
  kind_name <- if(is.null(component_id)) "phack_kind" else paste0("phack_kind_component_", component_id)
  pi_null_name <- if(is.null(component_id)) "pi_null" else paste0("pi_null_component_", component_id)
  beta_null_name <- if(is.null(component_id)) "beta_null" else paste0("beta_null_component_", component_id)
  z_source_name <- if(is.null(component_id)) "phack_z_source" else paste0("phack_z_source_component_", component_id)
  z_destination_name <- if(is.null(component_id)) "phack_z_dest" else paste0("phack_z_dest_component_", component_id)

  if(is.null(prior)){
    return(paste0(
      alpha_name, " <- 0\n",
      kind_name, " <- 0\n",
      pi_null_name, " <- 0\n",
      beta_null_name, " <- 0\n",
      z_source_name, "[1] <- 0\n",
      z_source_name, "[2] <- 0\n",
      z_destination_name, "[1] <- 0\n",
      z_destination_name, "[2] <- 0\n"
    ))
  }

  if(!is_prior_phacking(prior)){
    stop("'prior' must be a p-hacking prior.", call. = FALSE)
  }

  constants <- phack_backend_constants(prior$form, prior$source, prior$destination, target = prior$target)

  paste0(
    .JAGS_prior.simple(prior$alpha, alpha_name),
    kind_name, " <- ", constants$phack_kind, "\n",
    pi_null_name, " <- ", alpha_name, " * ", .selection_format_number(constants$pi_null_per_alpha), "\n",
    beta_null_name, " <- ", alpha_name, " * ", .selection_format_number(constants$beta_null_per_alpha), "\n",
    z_source_name, "[1] <- ", .selection_format_number(constants$z_source[1]), "\n",
    z_source_name, "[2] <- ", .selection_format_number(constants$z_source[2]), "\n",
    z_destination_name, "[1] <- ", .selection_format_number(constants$z_destination[1]), "\n",
    z_destination_name, "[2] <- ", .selection_format_number(constants$z_destination[2]), "\n"
  )
}

#' @title Check input
#'
#' @description A set of convenience functions for checking
#' objects/arguments to a function passed by a user.
#'
#' @param x object to be checked
#' @param name name of the object that will be print in the error
#' message.
#' @param check_length length of the object to be checked. Defaults to
#' \code{1}. Set to \code{0} in order to not check object length.
#' @param allow_NULL whether the object can be \code{NULL}.
#' If so, no checks are executed.
#' @param allow_NA whether the object can contain \code{NA} or \code{NaN}
#' values.
#' @param allow_values names of values allowed in a character vector.
#' Defaults to \code{NULL} (do not check).
#' @param check_names names of entries allowed in a list. Defaults to
#' \code{NULL} (do not check).
#' @param all_objects whether all entries in \code{check_names} must be
#' present. Defaults to \code{FALSE}.
#' @param allow_other whether additional entries then the specified in
#' \code{check_names} might be present
#' @param lower lower bound of allowed values.
#' Defaults to \code{-Inf} (do not check).
#' @param upper upper bound of allowed values.
#' Defaults to \code{Inf} (do not check).
#' @param allow_bound whether the values at the boundary are allowed.
#' Defaults to \code{TRUE}.
#' @param call string to be placed as a prefix to the error call.
#'
#' @examples
#' # check whether the object is logical
#' check_bool(TRUE, name = "input")
#'
#' # will throw an error on any other type
#' \dontrun{
#'   check_bool("TRUE", name = "input")
#' }
#'
#' @return returns \code{NULL}, called for the input check.
#'
#' @name check_input
#' @export check_bool
#' @export check_char
#' @export check_int
#' @export check_real
#' @export check_list

#' @rdname check_input
check_bool   <- function(x, name = deparse(substitute(x)), check_length = 1, allow_NULL = FALSE, allow_NA = TRUE, call = ""){

  if(is.null(x) || length(x) == 0){
    if(allow_NULL){
      return()
    }else{
      stop(paste0(call, "The '", name, "' argument cannot be NULL."), call. = FALSE)
    }
  }

  if(!is.logical(x) | !is.vector(x))
    stop(paste0(call, "The '", name, "' argument must be a logical vector."), call. = FALSE)

  if(check_length != 0  && length(x) != check_length)
    stop(paste0(call, "The '", name, "' argument must have length '", check_length, "'."), call. = FALSE)

  if(!allow_NA && anyNA(x))
    stop(paste0(call, "The '", name, "' argument cannot contain NA/NaN values."), call. = FALSE)

  return()
}

#' @rdname check_input
check_char   <- function(x, name = deparse(substitute(x)), check_length = 1, allow_values = NULL, allow_NULL = FALSE, allow_NA = TRUE, call = ""){

  if(is.null(x) || length(x) == 0){
    if(allow_NULL){
      return()
    }else{
      stop(paste0(call, "The '", name, "' argument cannot be NULL."), call. = FALSE)
    }
  }

  if(!is.character(x) | !is.vector(x))
    stop(paste0(call, "The '", name, "' argument must be a character vector."), call. = FALSE)

  if(check_length != 0 && length(x) != check_length)
    stop(paste0(call, "The '", name, "' argument must have length '", check_length, "'."), call. = FALSE)

  if(!allow_NA && anyNA(x))
    stop(paste0(call, "The '", name, "' argument cannot contain NA/NaN values."), call. = FALSE)

  if(!is.null(allow_values)){
    unrecognized <- !is.na(x) & !x %in% allow_values
    if(any(unrecognized))
      stop(paste0(call, "The '", paste0(x[unrecognized], collapse = "', '") ,"' values are not recognized by the '", name, "' argument."), call. = FALSE)
  }

  return()
}

#' @rdname check_input
check_real   <- function(x, name = deparse(substitute(x)), lower = -Inf, upper = Inf, allow_bound = TRUE, check_length = 1, allow_NULL = FALSE, allow_NA = TRUE, call = ""){

  if(is.null(x) || length(x) == 0){
    if(allow_NULL){
      return()
    }else{
      stop(paste0(call, "The '", name, "' argument cannot be NULL."), call. = FALSE)
    }
  }

  if(!allow_NA && anyNA(x))
    stop(paste0(call, "The '", name, "' argument cannot contain NA/NaN values."), call. = FALSE)

  if(check_length != 0 && length(x) != check_length)
    stop(paste0(call, "The '", name, "' argument must have length '", check_length, "'."), call. = FALSE)

  if(!is.numeric(x) | !is.vector(x))
    stop(paste0(call, "The '", name, "' argument must be a numeric vector."), call. = FALSE)

  x_no_na <- x[!is.na(x)]

  if(!is.infinite(lower)){
    if(!allow_bound){
      if(any(x_no_na <= lower))
        stop(paste0(call, "The '", name ,"' must be higher than ", lower,"."), call. = FALSE)
    }else{
      if(any(x_no_na < lower))
        stop(paste0(call, "The '", name ,"' must be equal or higher than ", lower,"."), call. = FALSE)
    }
  }

  if(!is.infinite(upper)){
    if(!allow_bound){
      if(any(x_no_na >= upper))
        stop(paste0(call, "The '", name ,"' must be lower than ", upper,"."), call. = FALSE)
    }else{
      if(any(x_no_na > upper))
        stop(paste0(call, "The '", name ,"' must be equal or lower than ", upper,"."), call. = FALSE)
    }
  }

  return()
}

#' @rdname check_input
check_int    <- function(x, name = deparse(substitute(x)), lower = -Inf, upper = Inf, allow_bound = TRUE, check_length = 1, allow_NULL = FALSE, allow_NA = TRUE, call = ""){

  if(is.null(x) || length(x) == 0){
    if(allow_NULL){
      return()
    }else{
      stop(paste0(call, "The '", name, "' argument cannot be NULL."), call. = FALSE)
    }
  }

  check_real(x, name = name, lower = lower, upper = upper, allow_bound = allow_bound, check_length = check_length, allow_NULL = allow_NULL, allow_NA = allow_NA, call = call)

  if(!all(.is.wholenumber(x, na.rm = TRUE)))
    stop(paste0(call, "The '", name ,"' argument must be an integer vector."), call. = FALSE)

  return()
}

#' @rdname check_input
check_list   <- function(x, name = deparse(substitute(x)), check_length = 0, check_names = NULL, all_objects = FALSE, allow_other = FALSE, allow_NULL = FALSE, call = ""){

  if(is.null(x) || length(x) == 0){
    if(allow_NULL){
      return()
    }else{
      stop(paste0(call, "The '", name, "' argument cannot be NULL."), call. = FALSE)
    }
  }

  if(!is.list(x))
    stop(paste0(call, "The '", name, "' argument must be a list."), call. = FALSE)

  if(check_length != 0 && length(x) != check_length)
    stop(paste0(call, "The '", name, "' argument must have length '", check_length, "'."), call. = FALSE)

  if(!is.null(check_names)){

    if(all_objects && any(!check_names %in% names(x)))
      stop(paste0(call, "The '", paste0(check_names[!check_names %in% names(x)], collapse = "', '") ,"' objects are missing in the '", name, "' argument."), call. = FALSE)

    if(!allow_other && any(!names(x) %in% check_names))
      stop(paste0(call, "The '", paste0(names(x)[!names(x) %in% check_names], collapse = "', '") ,"' objects are not recognized by the '", name, "' argument."), call. = FALSE)
  }

  return()
}

# helper functions
.is.wholenumber  <- function(x, na.rm = FALSE, tol = .Machine$double.eps^0.5){
  if(na.rm){
    return(stats::na.omit(abs(x - round(x))) < tol)
  }else{
    return(abs(x - round(x)) < tol)
  }
}

# check transformation argument
.check_transformation_input <- function(transformation, transformation_arguments, transformation_settings){

  if(!is.null(transformation)){
    if(is.character(transformation)){
      check_char(transformation, "transformation")
    }else if(is.list(transformation)){
      check_list(transformation, "transformation", check_length = 3, check_names = c("fun", "inv", "jac"), all_objects = TRUE)
    }else{
      stop("Uknown format of the 'transformation' argument.")
    }
  }
  check_list(transformation_arguments, "transformation_arguments", allow_NULL = TRUE)
  check_bool(transformation_settings, "transformation_settings")

  return()
}

# check whether package is installed
.check_rstan   <- function(){
  if(!requireNamespace("rstan", quietly = TRUE))
    stop("rstan package needs to be installed. Run 'install.packages('rstan')'")
  else
    return(invisible(TRUE))
}
.check_runjags <- function(){
  if(!requireNamespace("runjags", quietly = TRUE))
    stop("runjags package needs to be installed. Run 'install.packages('runjags')'")
  else
    return(invisible(TRUE))
}

# extract stan samples into a named matrix
.extract_stan <- function(fit, drop = TRUE){

  .check_rstan()
  if(!inherits(fit, what = "stanfit"))
    stop("'fit' must be an rstan fit")

  # order permutations to correspond to the other
  # (otherwise, models with same seed produce different posterior draws, like wtf stan???)
  for(i in seq_along(fit@sim$permutation)){
    fit@sim$permutation[[i]] <- seq_along(fit@sim$permutation[[i]])
  }

  model_samples <- rstan::extract(fit)
  par_names     <- names(model_samples)
  par_dims      <- sapply(model_samples, function(s)if(is.matrix(s)) ncol(s) else if(drop) 1 else 0)
  par_names     <- unlist(sapply(seq_along(par_names), function(p){
    if(par_dims[p] == {if(drop) 1 else 0}){
      return(par_names[p])
    }else{
      return(paste0(par_names[p], "[", 1:par_dims[p],"]"))
    }
  }))
  model_samples <- do.call(cbind, model_samples)
  colnames(model_samples) <- par_names

  return(model_samples)
}

# JASP progress bar functions
.JASP_progress_bar_active <- function(package_source = "jaspBase"){
  if(!requireNamespace(package_source, quietly = TRUE)){
    return(FALSE)
  }

  namespace <- asNamespace(package_source)
  if(!exists("jaspResultsCalledFromJasp", envir = namespace, inherits = FALSE)){
    return(FALSE)
  }

  isTRUE(get("jaspResultsCalledFromJasp", envir = namespace, inherits = FALSE)())
}
.JASP_progress_bar_call <- function(function_name, package_source = "jaspBase", ...){
  if(!.JASP_progress_bar_active(package_source)){
    return(invisible(FALSE))
  }

  namespace <- asNamespace(package_source)
  if(!exists(function_name, envir = namespace, inherits = FALSE)){
    return(invisible(FALSE))
  }

  do.call(get(function_name, envir = namespace, inherits = FALSE), list(...))
  return(invisible(TRUE))
}
.JASP_progress_bar_start <- function(n, label = "", package_source = "jaspBase"){
  .JASP_progress_bar_call("startProgressbar", package_source, expectedTicks = n, label = label)
  return(invisible())
}
.JASP_progress_bar_tick  <- function(package_source = "jaspBase"){
  .JASP_progress_bar_call("progressbarTick", package_source)
  return(invisible())
}

# depreciate orthonormal
.depreciate.transform_orthonormal <- function(transform_orthonormal, transform_factors){

  if(isTRUE(transform_orthonormal)){
    warning("'transform_orthonormal' argument will be depreciated in favor of 'transform_factors' argument.", immediate. = TRUE)

    if(isFALSE(transform_factors)){
      transform_factors <- transform_orthonormal
    }
  }

  return(transform_factors)
}

#' @title Creates an informed prior distribution based on research
#'
#' @description \code{prior_informed} creates an informed prior distribution based on past
#' research. The prior can be visualized by the \code{plot} function.
#'
#' @param name name of the prior distribution. There are many options based on prior psychological
#' or medical research.
#' For psychology, the possible options are
#' \describe{
#'   \item{\code{"van Erp"}}{for an informed prior distribution for the heterogeneity parameter tau
#'   of meta-analytic effect size estimates based on standardized mean differences
#'   \insertCite{erp2017estimates}{BayesTools},}
#'   \item{\code{"Oosterwijk"}}{for an informed prior distribution for the effect sizes expected in
#'   social psychology based on prior elicitation with dr. Oosterwijk
#'   \insertCite{gronau2017bayesian}{BayesTools}.}
#' }
#' For medicine, the possible options are based on \insertCite{bartos2021bayesian;textual}{BayesTools}
#' and  \insertCite{bartos2023empirical;textual}{BayesTools}
#' who developed empirical prior distributions for the effect size and heterogeneity parameters of the
#' continuous outcomes (standardized mean differences), dichotomous outcomes (logOR, logRR, and risk differences),
#' and time to event outcomes (logHR) based on the Cochrane database of systematic reviews.
#' Use \code{"Cochrane"} for a prior distribution based on the whole database or call
#' \code{print(prior_informed_medicine_names)} to inspect the names of
#' available subfields and set the appropriate \code{parameter} and \code{type};
#' availability of subfield-specific priors depends on the selected \code{type}.
#' @param parameter parameter name describing what prior distribution is supposed to be produced in cases
#' where the \code{name} corresponds to multiple prior distributions. Relevant only for the empirical medical
#' prior distributions.
#' @param type prior type describing what prior distribution is supposed to be produced in cases
#' where the \code{name} and \code{parameter} correspond to multiple prior distributions. Relevant only for
#' the empirical medical prior distributions with the following options
#' \describe{
#'   \item{\code{"smd"}}{for standardized mean differences}
#'   \item{\code{"logOR"}}{for log odds ratios}
#'   \item{\code{"logRR"}}{for log risk ratios}
#'   \item{\code{"RD"}}{for risk differences}
#'   \item{\code{"logHR"}}{for log hazard ratios}
#' }
#'
#' @examples
#' # prior distribution representing expected effect sizes in social psychology
#' # based on prior elicitation with dr. Oosterwijk
#' p1 <- prior_informed("Oosterwijk")
#'
#' # the prior distribution can be visualized using the plot function
#' # (see ?plot.prior for all options)
#' plot(p1)
#'
#'
#' # empirical prior distribution for the standardized mean differences from the oral health
#' # medical subfield based on meta-analytic effect size estimates from the
#' # Cochrane database of systematic reviews
#' p2 <- prior_informed("Oral Health", parameter ="effect", type ="smd")
#' print(p2)
#'
#' @return \code{prior_informed} returns an object of class 'prior'.
#'
#' @references
#' \insertAllCited{}
#' @seealso [prior()], [prior_informed_medicine_names]
#' @export
prior_informed <- function(name, parameter = NULL, type ="smd"){

  check_char(name,"name")
  check_char(parameter,"parameter", allow_NULL = TRUE)
  check_char(type,"type",      allow_NULL = TRUE)

  name <- .prior_clean_input_name(name)

  if(name %in% .prior_clean_input_name(prior_informed_medicine_names)){
    # check for implemented metrics
    if(is.null(type)){
      type <- "smd"
    }
    type <- .prior_clean_input_name(type)
    check_char(type,"type", allow_values = c("smd","logor","logrr","loghr","rd"))

    if(!name %in% .prior_clean_input_name(.prior_informed_medicine_names_by_type(type))){
      stop(paste0("The requested medicine prior 'name' is not available for type '", type, "'."), call. = FALSE)
    }
    if(is.null(parameter)){
      stop("The 'parameter' argument must be specified for informed prior distributions from medicine.", call. = FALSE)
    }
    parameter <- .prior_clean_input_name(parameter)
    check_char(parameter, "parameter", allow_values = c("effect", "heterogeneity"))

    p <- switch(
      parameter,
      "effect"        = .prior_informed.medicine_effect(name, type),
      "heterogeneity" = .prior_informed.medicine_heterogeneity(name, type),
      stop("unknown 'parameter' argument for an informed prior distribution from medicine.")
    )

    return(p)

  }else if(name %in% .prior_clean_input_name(.prior_informed_psychology_names)){

    p <- .prior_informed.psychology(name)

    return(p)

  }else{
    stop("unknown prior 'name'. See '?prior_informed' for help.")
  }
}

.prior_informed_medicine_names_by_type <- function(type = NULL){

  names_by_type <- list(
    smd = setdiff(prior_informed_medicine_names, c(
      "Breast Cancer",
      "Childhood Cancer",
      "Epilepsy",
      "Fertility Regulation",
      "Gut",
      "Haematology",
      "Heart; Vascular",
      "Lung Cancer",
      "Multiple Sclerosis and Rare Diseases of the CNS",
      "Neuromuscular"
    )),
    logor = setdiff(prior_informed_medicine_names, c(
      "Inflammatory Bowel Disease",
      "Public Health",
      "Upper GI and Pancreatic Diseases"
    )),
    logrr = setdiff(prior_informed_medicine_names, c(
      "Inflammatory Bowel Disease",
      "Public Health",
      "Upper GI and Pancreatic Diseases"
    )),
    rd = setdiff(prior_informed_medicine_names, c(
      "Inflammatory Bowel Disease",
      "Public Health",
      "Upper GI and Pancreatic Diseases"
    )),
    loghr = "Cochrane"
  )

  if(is.null(type)){
    return(names_by_type)
  }

  names_by_type[[type]]
}


.prior_informed.medicine_effect        <- function(name, type){

  if(type =="smd"){

    # medical priors for continuous outcomes based on Bartoš et al. 2021
    # (can't use switch if the argument is to be modified)
    if(name == .prior_clean_input_name("Acute Respiratory Infections")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.38, df = 5))
    }else if(name == .prior_clean_input_name("Airways")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.38, df = 6))
    }else if(name == .prior_clean_input_name("Anaesthesia")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.55, df = 4))
    }else if(name == .prior_clean_input_name("Back and Neck")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.37, df = 5))
    }else if(name == .prior_clean_input_name("Bone, Joint and Muscle Trauma")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.40, df = 5))
    }else if(name == .prior_clean_input_name("Colorectal")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.51, df = 5))
    }else if(name == .prior_clean_input_name("Common Mental Disorders")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.55, df = 5))
    }else if(name == .prior_clean_input_name("Consumers and Communication")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.40, df = 5))
    }else if(name == .prior_clean_input_name("Cystic Fibrosis and Genetic Disorders")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.47, df = 5))
    }else if(name == .prior_clean_input_name("Dementia and Cognitive Improvement")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.45, df = 5))
    }else if(name == .prior_clean_input_name("Developmental, Psychosocial and Learning Problems")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.18, df = 5))
    }else if(name == .prior_clean_input_name("Drugs and Alcohol")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.33, df = 5))
    }else if(name == .prior_clean_input_name("Effective Practice and Organisation of Care")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.39, df = 5))
    }else if(name == .prior_clean_input_name("Emergency and Critical Care")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.39, df = 5))
    }else if(name == .prior_clean_input_name("ENT")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.43, df = 5))
    }else if(name == .prior_clean_input_name("Eyes and Vision")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.40, df = 6))
    }else if(name == .prior_clean_input_name("Gynaecological, Neuro-oncology and Orphan Cancer")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.45, df = 5))
    }else if(name == .prior_clean_input_name("Gynaecology and Fertility")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.38, df = 5))
    }else if(name == .prior_clean_input_name("Heart")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.42, df = 5))
    }else if(name == .prior_clean_input_name("Hepato-Biliary")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.60, df = 4))
    }else if(name == .prior_clean_input_name("HIV/AIDS")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.43, df = 5))
    }else if(name == .prior_clean_input_name("Hypertension")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.48, df = 3))
    }else if(name == .prior_clean_input_name("Incontinence")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.33, df = 6))
    }else if(name == .prior_clean_input_name("Infectious Diseases")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.59, df = 2))
    }else if(name == .prior_clean_input_name("Inflammatory Bowel Disease")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.40, df = 5))
    }else if(name == .prior_clean_input_name("Injuries")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.35, df = 5))
    }else if(name == .prior_clean_input_name("Kidney and Transplant")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.54, df = 5))
    }else if(name == .prior_clean_input_name("Metabolic and Endocrine Disorders")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.43, df = 5))
    }else if(name == .prior_clean_input_name("Methodology")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.49, df = 5))
    }else if(name == .prior_clean_input_name("Movement Disorders")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.42, df = 5))
    }else if(name == .prior_clean_input_name("Musculoskeletal")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.45, df = 6))
    }else if(name == .prior_clean_input_name("Neonatal")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.42, df = 5))
    }else if(name == .prior_clean_input_name("Oral Health")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.51, df = 5))
    }else if(name == .prior_clean_input_name("Pain, Palliative and Supportive Care")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.43, df = 5))
    }else if(name == .prior_clean_input_name("Pregnancy and Childbirth")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.33, df = 5))
    }else if(name == .prior_clean_input_name("Public Health")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.33, df = 5))
    }else if(name == .prior_clean_input_name("Schizophrenia")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.29, df = 4))
    }else if(name == .prior_clean_input_name("Sexually Transmitted Infections")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.42, df = 5))
    }else if(name == .prior_clean_input_name("Skin")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.48, df = 5))
    }else if(name == .prior_clean_input_name("Stroke")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.48, df = 5))
    }else if(name == .prior_clean_input_name("Tobacco Addiction")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.44, df = 4))
    }else if(name == .prior_clean_input_name("Upper GI and Pancreatic Diseases")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.45, df = 5))
    }else if(name == .prior_clean_input_name("Urology")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.44, df = 5))
    }else if(name == .prior_clean_input_name("Vascular")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.46, df = 5))
    }else if(name == .prior_clean_input_name("Work")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.42, df = 5))
    }else if(name == .prior_clean_input_name("Wounds")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.56, df = 5))
    }else if(name == .prior_clean_input_name("Cochrane")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.43, df = 5))
    }else{
      stop("unknown subfield 'name' argument for an informed prior distribution from medicine.")
    }

    return(p)

  }else if(type =="logor"){

    # medical priors for dichotomous outcomes based on Bartoš et al. 2023
    if(name == .prior_clean_input_name("Acute Respiratory Infections")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.48, df = 3))
    }else if(name == .prior_clean_input_name("Airways")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.37, df = 2))
    }else if(name == .prior_clean_input_name("Anaesthesia")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.79, df = 6))
    }else if(name == .prior_clean_input_name("Back and Neck")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.62, df = 4))
    }else if(name == .prior_clean_input_name("Bone, Joint and Muscle Trauma")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.44, df = 2))
    }else if(name == .prior_clean_input_name("Breast Cancer")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.39, df = 3))
    }else if(name == .prior_clean_input_name("Childhood Cancer")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.47, df = 4))
    }else if(name == .prior_clean_input_name("Colorectal")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.65, df = 4))
    }else if(name == .prior_clean_input_name("Common Mental Disorders")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.54, df = 4))
    }else if(name == .prior_clean_input_name("Consumers and Communication")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.48, df = 4))
    }else if(name == .prior_clean_input_name("Cystic Fibrosis and Genetic Disorders")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.40, df = 3))
    }else if(name == .prior_clean_input_name("Dementia and Cognitive Improvement")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.49, df = 4))
    }else if(name == .prior_clean_input_name("Developmental, Psychosocial and Learning Problems")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.83, df = 5))
    }else if(name == .prior_clean_input_name("Drugs and Alcohol")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.44, df = 4))
    }else if(name == .prior_clean_input_name("Effective Practice and Organisation of Care")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.51, df = 4))
    }else if(name == .prior_clean_input_name("Emergency and Critical Care")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.35, df = 3))
    }else if(name == .prior_clean_input_name("ENT")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.81, df = 4))
    }else if(name == .prior_clean_input_name("Epilepsy")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.88, df = 6))
    }else if(name == .prior_clean_input_name("Eyes and Vision")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.77, df = 5))
    }else if(name == .prior_clean_input_name("Fertility Regulation")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.46, df = 5))
    }else if(name == .prior_clean_input_name("Gut")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.63, df = 5))
    }else if(name == .prior_clean_input_name("Gynaecological, Neuro-oncology and Orphan Cancer")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.53, df = 4))
    }else if(name == .prior_clean_input_name("Gynaecology and Fertility")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.40, df = 2))
    }else if(name == .prior_clean_input_name("Haematology")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.57, df = 4))
    }else if(name == .prior_clean_input_name("Heart")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.20, df = 2))
    }else if(name == .prior_clean_input_name("Heart; Vascular")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.95, df = 4))
    }else if(name == .prior_clean_input_name("Hepato-Biliary")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.43, df = 3))
    }else if(name == .prior_clean_input_name("HIV/AIDS")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.32, df = 4))
    }else if(name == .prior_clean_input_name("Hypertension")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.28, df = 5))
    }else if(name == .prior_clean_input_name("Incontinence")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.75, df = 3))
    }else if(name == .prior_clean_input_name("Infectious Diseases")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.66, df = 3))
    }else if(name == .prior_clean_input_name("Injuries")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.60, df = 4))
    }else if(name == .prior_clean_input_name("Kidney and Transplant")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.53, df = 4))
    }else if(name == .prior_clean_input_name("Lung Cancer")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.61, df = 5))
    }else if(name == .prior_clean_input_name("Metabolic and Endocrine Disorders")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.29, df = 2))
    }else if(name == .prior_clean_input_name("Methodology")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.60, df = 5))
    }else if(name == .prior_clean_input_name("Movement Disorders")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.73, df = 5))
    }else if(name == .prior_clean_input_name("Multiple Sclerosis and Rare Diseases of the CNS")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.76, df = 4))
    }else if(name == .prior_clean_input_name("Musculoskeletal")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.59, df = 4))
    }else if(name == .prior_clean_input_name("Neonatal")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.29, df = 3))
    }else if(name == .prior_clean_input_name("Neuromuscular")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.70, df = 5))
    }else if(name == .prior_clean_input_name("Oral Health")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 1.13, df = 4))
    }else if(name == .prior_clean_input_name("Pain, Palliative and Supportive Care")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 1.00, df = 7))
    }else if(name == .prior_clean_input_name("Pregnancy and Childbirth")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.38, df = 2))
    }else if(name == .prior_clean_input_name("Schizophrenia")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.58, df = 4))
    }else if(name == .prior_clean_input_name("Sexually Transmitted Infections")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.61, df = 2))
    }else if(name == .prior_clean_input_name("Skin")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.81, df = 2))
    }else if(name == .prior_clean_input_name("Stroke")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.22, df = 2))
    }else if(name == .prior_clean_input_name("Tobacco Addiction")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.49, df = 5))
    }else if(name == .prior_clean_input_name("Urology")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.82, df = 5))
    }else if(name == .prior_clean_input_name("Vascular")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.68, df = 5))
    }else if(name == .prior_clean_input_name("Work")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.59, df = 4))
    }else if(name == .prior_clean_input_name("Wounds")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.60, df = 4))
    }else if(name == .prior_clean_input_name("Cochrane")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.58, df = 4))
    }else{
      stop("unknown subfield 'name' argument for an informed prior distribution from medicine.")
    }

    return(p)

  }else if(type =="logrr"){

    # medical priors for dichotomous outcomes based on Bartoš et al. 2023
    if(name == .prior_clean_input_name("Acute Respiratory Infections")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.27, df = 3))
    }else if(name == .prior_clean_input_name("Airways")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.25, df = 3))
    }else if(name == .prior_clean_input_name("Anaesthesia")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.45, df = 5))
    }else if(name == .prior_clean_input_name("Back and Neck")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.34, df = 3))
    }else if(name == .prior_clean_input_name("Bone, Joint and Muscle Trauma")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.26, df = 2))
    }else if(name == .prior_clean_input_name("Breast Cancer")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.21, df = 3))
    }else if(name == .prior_clean_input_name("Childhood Cancer")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.28, df = 3))
    }else if(name == .prior_clean_input_name("Colorectal")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.38, df = 4))
    }else if(name == .prior_clean_input_name("Common Mental Disorders")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.29, df = 3))
    }else if(name == .prior_clean_input_name("Consumers and Communication")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.13, df = 2))
    }else if(name == .prior_clean_input_name("Cystic Fibrosis and Genetic Disorders")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.21, df = 2))
    }else if(name == .prior_clean_input_name("Dementia and Cognitive Improvement")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.29, df = 3))
    }else if(name == .prior_clean_input_name("Developmental, Psychosocial and Learning Problems")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.39, df = 3))
    }else if(name == .prior_clean_input_name("Drugs and Alcohol")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.22, df = 4))
    }else if(name == .prior_clean_input_name("Effective Practice and Organisation of Care")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.30, df = 4))
    }else if(name == .prior_clean_input_name("Emergency and Critical Care")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.19, df = 2))
    }else if(name == .prior_clean_input_name("ENT")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.40, df = 4))
    }else if(name == .prior_clean_input_name("Epilepsy")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.58, df = 5))
    }else if(name == .prior_clean_input_name("Eyes and Vision")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.34, df = 4))
    }else if(name == .prior_clean_input_name("Fertility Regulation")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.34, df = 4))
    }else if(name == .prior_clean_input_name("Gut")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.40, df = 4))
    }else if(name == .prior_clean_input_name("Gynaecological, Neuro-oncology and Orphan Cancer")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.35, df = 4))
    }else if(name == .prior_clean_input_name("Gynaecology and Fertility")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.25, df = 2))
    }else if(name == .prior_clean_input_name("Haematology")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.31, df = 3))
    }else if(name == .prior_clean_input_name("Heart")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.17, df = 3))
    }else if(name == .prior_clean_input_name("Heart; Vascular")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.39, df = 3))
    }else if(name == .prior_clean_input_name("Hepato-Biliary")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.23, df = 3))
    }else if(name == .prior_clean_input_name("HIV/AIDS")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.21, df = 3))
    }else if(name == .prior_clean_input_name("Hypertension")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.25, df = 4))
    }else if(name == .prior_clean_input_name("Incontinence")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.39, df = 3))
    }else if(name == .prior_clean_input_name("Infectious Diseases")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.41, df = 3))
    }else if(name == .prior_clean_input_name("Injuries")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.27, df = 4))
    }else if(name == .prior_clean_input_name("Kidney and Transplant")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.35, df = 4))
    }else if(name == .prior_clean_input_name("Lung Cancer")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.38, df = 4))
    }else if(name == .prior_clean_input_name("Metabolic and Endocrine Disorders")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.13, df = 1))
    }else if(name == .prior_clean_input_name("Methodology")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.27, df = 5))
    }else if(name == .prior_clean_input_name("Movement Disorders")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.46, df = 4))
    }else if(name == .prior_clean_input_name("Multiple Sclerosis and Rare Diseases of the CNS")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.44, df = 3))
    }else if(name == .prior_clean_input_name("Musculoskeletal")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.33, df = 3))
    }else if(name == .prior_clean_input_name("Neonatal")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.18, df = 3))
    }else if(name == .prior_clean_input_name("Neuromuscular")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.41, df = 4))
    }else if(name == .prior_clean_input_name("Oral Health")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.55, df = 4))
    }else if(name == .prior_clean_input_name("Pain, Palliative and Supportive Care")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.59, df = 5))
    }else if(name == .prior_clean_input_name("Pregnancy and Childbirth")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.24, df = 2))
    }else if(name == .prior_clean_input_name("Schizophrenia")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.33, df = 3))
    }else if(name == .prior_clean_input_name("Sexually Transmitted Infections")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.39, df = 3))
    }else if(name == .prior_clean_input_name("Skin")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.41, df = 1))
    }else if(name == .prior_clean_input_name("Stroke")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.13, df = 2))
    }else if(name == .prior_clean_input_name("Tobacco Addiction")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.36, df = 5))
    }else if(name == .prior_clean_input_name("Urology")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.44, df = 3))
    }else if(name == .prior_clean_input_name("Vascular")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.43, df = 5))
    }else if(name == .prior_clean_input_name("Work")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.23, df = 3))
    }else if(name == .prior_clean_input_name("Wounds")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.37, df = 4))
    }else if(name == .prior_clean_input_name("Cochrane")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.32, df = 3))
    }else{
      stop("unknown subfield 'name' argument for an informed prior distribution from medicine.")
    }

  }else if(type =="rd"){

    # medical priors for dichotomous outcomes based on Bartoš et al. 2023
    if(name == .prior_clean_input_name("Acute Respiratory Infections")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.012, df = 1))
    }else if(name == .prior_clean_input_name("Airways")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.005, df = 1))
    }else if(name == .prior_clean_input_name("Anaesthesia")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.020, df = 1))
    }else if(name == .prior_clean_input_name("Back and Neck")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.056, df = 2))
    }else if(name == .prior_clean_input_name("Bone, Joint and Muscle Trauma")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.014, df = 1))
    }else if(name == .prior_clean_input_name("Breast Cancer")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.049, df = 2))
    }else if(name == .prior_clean_input_name("Childhood Cancer")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.021, df = 1))
    }else if(name == .prior_clean_input_name("Colorectal")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.010, df = 1))
    }else if(name == .prior_clean_input_name("Common Mental Disorders")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.051, df = 1))
    }else if(name == .prior_clean_input_name("Consumers and Communication")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.063, df = 2))
    }else if(name == .prior_clean_input_name("Cystic Fibrosis and Genetic Disorders")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.015, df = 1))
    }else if(name == .prior_clean_input_name("Dementia and Cognitive Improvement")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.017, df = 1))
    }else if(name == .prior_clean_input_name("Developmental, Psychosocial and Learning Problems")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.082, df = 2))
    }else if(name == .prior_clean_input_name("Drugs and Alcohol")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.025, df = 1))
    }else if(name == .prior_clean_input_name("Effective Practice and Organisation of Care")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.056, df = 2))
    }else if(name == .prior_clean_input_name("Emergency and Critical Care")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.019, df = 1))
    }else if(name == .prior_clean_input_name("ENT")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.042, df = 1))
    }else if(name == .prior_clean_input_name("Epilepsy")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.047, df = 2))
    }else if(name == .prior_clean_input_name("Eyes and Vision")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.041, df = 1))
    }else if(name == .prior_clean_input_name("Fertility Regulation")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.014, df = 1))
    }else if(name == .prior_clean_input_name("Gut")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.047, df = 2))
    }else if(name == .prior_clean_input_name("Gynaecological, Neuro-oncology and Orphan Cancer")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.024, df = 1))
    }else if(name == .prior_clean_input_name("Gynaecology and Fertility")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.021, df = 1))
    }else if(name == .prior_clean_input_name("Haematology")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.047, df = 1))
    }else if(name == .prior_clean_input_name("Heart")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.005, df = 1))
    }else if(name == .prior_clean_input_name("Heart; Vascular")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.083, df = 1))
    }else if(name == .prior_clean_input_name("Hepato-Biliary")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.014, df = 1))
    }else if(name == .prior_clean_input_name("HIV/AIDS")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.018, df = 2))
    }else if(name == .prior_clean_input_name("Hypertension")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.006, df = 2))
    }else if(name == .prior_clean_input_name("Incontinence")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.060, df = 2))
    }else if(name == .prior_clean_input_name("Infectious Diseases")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.019, df = 1))
    }else if(name == .prior_clean_input_name("Injuries")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.010, df = 1))
    }else if(name == .prior_clean_input_name("Kidney and Transplant")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.028, df = 1))
    }else if(name == .prior_clean_input_name("Lung Cancer")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.057, df = 2))
    }else if(name == .prior_clean_input_name("Metabolic and Endocrine Disorders")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.003, df = 1))
    }else if(name == .prior_clean_input_name("Methodology")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.097, df = 2))
    }else if(name == .prior_clean_input_name("Movement Disorders")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.055, df = 2))
    }else if(name == .prior_clean_input_name("Multiple Sclerosis and Rare Diseases of the CNS")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.067, df = 2))
    }else if(name == .prior_clean_input_name("Musculoskeletal")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.010, df = 1))
    }else if(name == .prior_clean_input_name("Neonatal")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.022, df = 1))
    }else if(name == .prior_clean_input_name("Neuromuscular")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.034, df = 1))
    }else if(name == .prior_clean_input_name("Oral Health")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.030, df = 1))
    }else if(name == .prior_clean_input_name("Pain, Palliative and Supportive Care")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.111, df = 2))
    }else if(name == .prior_clean_input_name("Pregnancy and Childbirth")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.012, df = 1))
    }else if(name == .prior_clean_input_name("Schizophrenia")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.028, df = 1))
    }else if(name == .prior_clean_input_name("Sexually Transmitted Infections")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.035, df = 1))
    }else if(name == .prior_clean_input_name("Skin")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.033, df = 1))
    }else if(name == .prior_clean_input_name("Stroke")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.014, df = 1))
    }else if(name == .prior_clean_input_name("Tobacco Addiction")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.042, df = 2))
    }else if(name == .prior_clean_input_name("Urology")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.041, df = 1))
    }else if(name == .prior_clean_input_name("Vascular")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.012, df = 1))
    }else if(name == .prior_clean_input_name("Work")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.062, df = 2))
    }else if(name == .prior_clean_input_name("Wounds")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.012, df = 1))
    }else if(name == .prior_clean_input_name("Cochrane")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.034, df = 1))
    }else{
      stop("unknown subfield 'name' argument for an informed prior distribution from medicine.")
    }

    return(p)

  }else if(type =="loghr"){

    # medical priors for dichotomous outcomes based on Bartoš et al. 2023
    if(name == .prior_clean_input_name("Cochrane")){
      p <- prior(distribution ="t", parameters = list(location = 0, scale = 0.13, df = 2))
    }else{
      stop("There are no subfield specific prior distributions for logHR.")
    }

    return(p)
  }
}
.prior_informed.medicine_heterogeneity <- function(name, type){

  if(type =="smd"){

    # medical priors for continuous outcomes based on Bartoš et al. 2021
    if(name == .prior_clean_input_name("Acute Respiratory Infections")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.73, scale = 0.46))
    }else if(name == .prior_clean_input_name("Airways")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 2.02, scale = 0.28))
    }else if(name == .prior_clean_input_name("Anaesthesia")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.62, scale = 0.64))
    }else if(name == .prior_clean_input_name("Back and Neck")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.75, scale = 0.57))
    }else if(name == .prior_clean_input_name("Bone, Joint and Muscle Trauma")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.52, scale = 0.28))
    }else if(name == .prior_clean_input_name("Colorectal")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.64, scale = 0.56))
    }else if(name == .prior_clean_input_name("Common Mental Disorders")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.62, scale = 0.45))
    }else if(name == .prior_clean_input_name("Consumers and Communication")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.56, scale = 0.14))
    }else if(name == .prior_clean_input_name("Cystic Fibrosis and Genetic Disorders")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.70, scale = 0.45))
    }else if(name == .prior_clean_input_name("Dementia and Cognitive Improvement")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.71, scale = 0.44))
    }else if(name == .prior_clean_input_name("Developmental, Psychosocial and Learning Problems")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.43, scale = 0.12))
    }else if(name == .prior_clean_input_name("Drugs and Alcohol")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.89, scale = 0.28))
    }else if(name == .prior_clean_input_name("Effective Practice and Organisation of Care")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.71, scale = 0.35))
    }else if(name == .prior_clean_input_name("Emergency and Critical Care")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.62, scale = 0.29))
    }else if(name == .prior_clean_input_name("ENT")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.85, scale = 0.48))
    }else if(name == .prior_clean_input_name("Eyes and Vision")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.86, scale = 0.41))
    }else if(name == .prior_clean_input_name("Gynaecological, Neuro-oncology and Orphan Cancer")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.67, scale = 0.46))
    }else if(name == .prior_clean_input_name("Gynaecology and Fertility")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.78, scale = 0.46))
    }else if(name == .prior_clean_input_name("Heart")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.83, scale = 0.47))
    }else if(name == .prior_clean_input_name("Hepato-Biliary")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.56, scale = 0.58))
    }else if(name == .prior_clean_input_name("HIV/AIDS")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.73, scale = 0.44))
    }else if(name == .prior_clean_input_name("Hypertension")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 2.01, scale = 0.38))
    }else if(name == .prior_clean_input_name("Incontinence")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.64, scale = 0.36))
    }else if(name == .prior_clean_input_name("Infectious Diseases")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.28, scale = 0.44))
    }else if(name == .prior_clean_input_name("Inflammatory Bowel Disease")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.76, scale = 0.39))
    }else if(name == .prior_clean_input_name("Injuries")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.80, scale = 0.34))
    }else if(name == .prior_clean_input_name("Kidney and Transplant")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.72, scale = 0.53))
    }else if(name == .prior_clean_input_name("Metabolic and Endocrine Disorders")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.71, scale = 0.37))
    }else if(name == .prior_clean_input_name("Methodology")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.72, scale = 0.51))
    }else if(name == .prior_clean_input_name("Movement Disorders")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.88, scale = 0.33))
    }else if(name == .prior_clean_input_name("Musculoskeletal")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.87, scale = 0.38))
    }else if(name == .prior_clean_input_name("Neonatal")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.68, scale = 0.38))
    }else if(name == .prior_clean_input_name("Oral Health")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.79, scale = 0.28))
    }else if(name == .prior_clean_input_name("Pain, Palliative and Supportive Care")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.69, scale = 0.42))
    }else if(name == .prior_clean_input_name("Pregnancy and Childbirth")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.86, scale = 0.32))
    }else if(name == .prior_clean_input_name("Public Health")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.76, scale = 0.23))
    }else if(name == .prior_clean_input_name("Schizophrenia")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.60, scale = 0.27))
    }else if(name == .prior_clean_input_name("Sexually Transmitted Infections")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.70, scale = 0.59))
    }else if(name == .prior_clean_input_name("Skin")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.64, scale = 0.51))
    }else if(name == .prior_clean_input_name("Stroke")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.71, scale = 0.40))
    }else if(name == .prior_clean_input_name("Tobacco Addiction")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.73, scale = 0.42))
    }else if(name == .prior_clean_input_name("Upper GI and Pancreatic Diseases")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.76, scale = 0.38))
    }else if(name == .prior_clean_input_name("Urology")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.73, scale = 0.45))
    }else if(name == .prior_clean_input_name("Vascular")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.66, scale = 0.50))
    }else if(name == .prior_clean_input_name("Work")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.76, scale = 0.39))
    }else if(name == .prior_clean_input_name("Wounds")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.54, scale = 0.41))
    }else if(name == .prior_clean_input_name("Cochrane")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.71, scale = 0.40))
    }else{
      stop("unknown subfield 'name' argument for an informed prior distribution from medicine.")
    }

    return(p)

  }else if(type =="logor"){

    # medical priors for dichotomous outcomes based on Bartoš et al. 2023
    if(name == .prior_clean_input_name("Acute Respiratory Infections")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.67, scale = 0.45))
    }else if(name == .prior_clean_input_name("Airways")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.35, scale = 0.27))
    }else if(name == .prior_clean_input_name("Anaesthesia")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 2.12, scale = 0.86))
    }else if(name == .prior_clean_input_name("Back and Neck")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.84, scale = 0.68))
    }else if(name == .prior_clean_input_name("Bone, Joint and Muscle Trauma")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.01, scale = 0.36))
    }else if(name == .prior_clean_input_name("Breast Cancer")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.59, scale = 0.48))
    }else if(name == .prior_clean_input_name("Childhood Cancer")){
      stop("no informed subfield prior distribution for 'Childhood Cancer' category.")
    }else if(name == .prior_clean_input_name("Colorectal")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.71, scale = 0.60))
    }else if(name == .prior_clean_input_name("Common Mental Disorders")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.63, scale = 0.45))
    }else if(name == .prior_clean_input_name("Consumers and Communication")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 2.37, scale = 0.86))
    }else if(name == .prior_clean_input_name("Cystic Fibrosis and Genetic Disorders")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.82, scale = 0.58))
    }else if(name == .prior_clean_input_name("Dementia and Cognitive Improvement")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.28, scale = 0.25))
    }else if(name == .prior_clean_input_name("Developmental, Psychosocial and Learning Problems")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.83, scale = 0.82))
    }else if(name == .prior_clean_input_name("Drugs and Alcohol")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.59, scale = 0.42))
    }else if(name == .prior_clean_input_name("Effective Practice and Organisation of Care")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.90, scale = 0.68))
    }else if(name == .prior_clean_input_name("Emergency and Critical Care")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.46, scale = 0.34))
    }else if(name == .prior_clean_input_name("ENT")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.73, scale = 0.71))
    }else if(name == .prior_clean_input_name("Epilepsy")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.71, scale = 0.43))
    }else if(name == .prior_clean_input_name("Eyes and Vision")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 2.09, scale = 0.94))
    }else if(name == .prior_clean_input_name("Fertility Regulation")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 2.21, scale = 0.71))
    }else if(name == .prior_clean_input_name("Gut")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.94, scale = 0.62))
    }else if(name == .prior_clean_input_name("Gynaecological, Neuro-oncology and Orphan Cancer")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.80, scale = 0.56))
    }else if(name == .prior_clean_input_name("Gynaecology and Fertility")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.24, scale = 0.28))
    }else if(name == .prior_clean_input_name("Haematology")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 2.91, scale = 0.66))
    }else if(name == .prior_clean_input_name("Heart")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.64, scale = 0.29))
    }else if(name == .prior_clean_input_name("Heart; Vascular")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.64, scale = 0.83))
    }else if(name == .prior_clean_input_name("Hepato-Biliary")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.58, scale = 0.40))
    }else if(name == .prior_clean_input_name("HIV/AIDS")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.76, scale = 0.36))
    }else if(name == .prior_clean_input_name("Hypertension")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.25, scale = 0.10))
    }else if(name == .prior_clean_input_name("Incontinence")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 2.07, scale = 1.09))
    }else if(name == .prior_clean_input_name("Infectious Diseases")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 2.08, scale = 0.86))
    }else if(name == .prior_clean_input_name("Injuries")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.52, scale = 0.49))
    }else if(name == .prior_clean_input_name("Kidney and Transplant")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.68, scale = 0.44))
    }else if(name == .prior_clean_input_name("Lung Cancer")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 2.04, scale = 0.68))
    }else if(name == .prior_clean_input_name("Metabolic and Endocrine Disorders")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 0.92, scale = 0.11))
    }else if(name == .prior_clean_input_name("Methodology")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 2.04, scale = 0.50))
    }else if(name == .prior_clean_input_name("Movement Disorders")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 2.14, scale = 0.64))
    }else if(name == .prior_clean_input_name("Multiple Sclerosis and Rare Diseases of the CNS")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 2.09, scale = 0.71))
    }else if(name == .prior_clean_input_name("Musculoskeletal")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.76, scale = 0.59))
    }else if(name == .prior_clean_input_name("Neonatal")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.80, scale = 0.42))
    }else if(name == .prior_clean_input_name("Neuromuscular")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.74, scale = 0.49))
    }else if(name == .prior_clean_input_name("Oral Health")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.85, scale = 0.70))
    }else if(name == .prior_clean_input_name("Pain, Palliative and Supportive Care")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.50, scale = 0.40))
    }else if(name == .prior_clean_input_name("Pregnancy and Childbirth")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.73, scale = 0.47))
    }else if(name == .prior_clean_input_name("Schizophrenia")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.92, scale = 0.69))
    }else if(name == .prior_clean_input_name("Sexually Transmitted Infections")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.72, scale = 0.47))
    }else if(name == .prior_clean_input_name("Skin")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.64, scale = 0.49))
    }else if(name == .prior_clean_input_name("Stroke")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.36, scale = 0.21))
    }else if(name == .prior_clean_input_name("Tobacco Addiction")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 2.51, scale = 0.63))
    }else if(name == .prior_clean_input_name("Urology")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.72, scale = 0.50))
    }else if(name == .prior_clean_input_name("Vascular")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.70, scale = 0.45))
    }else if(name == .prior_clean_input_name("Work")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.79, scale = 0.73))
    }else if(name == .prior_clean_input_name("Wounds")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 2.16, scale = 0.86))
    }else if(name == .prior_clean_input_name("Cochrane")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.77, scale = 0.55))
    }else{
      stop("unknown subfield 'name' argument for an informed prior distribution from medicine.")
    }

    return(p)

  }else if(type =="logrr"){

    # medical priors for dichotomous outcomes based on Bartoš et al. 2023
    if(name == .prior_clean_input_name("Acute Respiratory Infections")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.58, scale = 0.25))
    }else if(name == .prior_clean_input_name("Airways")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.36, scale = 0.15))
    }else if(name == .prior_clean_input_name("Anaesthesia")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.21, scale = 0.21))
    }else if(name == .prior_clean_input_name("Back and Neck")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.52, scale = 0.22))
    }else if(name == .prior_clean_input_name("Bone, Joint and Muscle Trauma")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 0.97, scale = 0.14))
    }else if(name == .prior_clean_input_name("Breast Cancer")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.49, scale = 0.24))
    }else if(name == .prior_clean_input_name("Childhood Cancer")){
      stop("no informed subfield prior distribution for 'Childhood Cancer' category.")
    }else if(name == .prior_clean_input_name("Colorectal")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.42, scale = 0.27))
    }else if(name == .prior_clean_input_name("Common Mental Disorders")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.43, scale = 0.20))
    }else if(name == .prior_clean_input_name("Consumers and Communication")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.48, scale = 0.14))
    }else if(name == .prior_clean_input_name("Cystic Fibrosis and Genetic Disorders")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.71, scale = 0.24))
    }else if(name == .prior_clean_input_name("Dementia and Cognitive Improvement")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.53, scale = 0.19))
    }else if(name == .prior_clean_input_name("Developmental, Psychosocial and Learning Problems")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 0.97, scale = 0.11))
    }else if(name == .prior_clean_input_name("Drugs and Alcohol")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.58, scale = 0.20))
    }else if(name == .prior_clean_input_name("Effective Practice and Organisation of Care")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.40, scale = 0.24))
    }else if(name == .prior_clean_input_name("Emergency and Critical Care")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.65, scale = 0.25))
    }else if(name == .prior_clean_input_name("ENT")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.50, scale = 0.24))
    }else if(name == .prior_clean_input_name("Epilepsy")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.74, scale = 0.27))
    }else if(name == .prior_clean_input_name("Eyes and Vision")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.60, scale = 0.26))
    }else if(name == .prior_clean_input_name("Fertility Regulation")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.59, scale = 0.31))
    }else if(name == .prior_clean_input_name("Gut")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.69, scale = 0.29))
    }else if(name == .prior_clean_input_name("Gynaecological, Neuro-oncology and Orphan Cancer")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.43, scale = 0.25))
    }else if(name == .prior_clean_input_name("Gynaecology and Fertility")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.29, scale = 0.16))
    }else if(name == .prior_clean_input_name("Haematology")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.22, scale = 0.11))
    }else if(name == .prior_clean_input_name("Heart")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.96, scale = 0.24))
    }else if(name == .prior_clean_input_name("Heart; Vascular")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.62, scale = 0.25))
    }else if(name == .prior_clean_input_name("Hepato-Biliary")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.44, scale = 0.19))
    }else if(name == .prior_clean_input_name("HIV/AIDS")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.66, scale = 0.23))
    }else if(name == .prior_clean_input_name("Hypertension")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.48, scale = 0.11))
    }else if(name == .prior_clean_input_name("Incontinence")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.53, scale = 0.33))
    }else if(name == .prior_clean_input_name("Infectious Diseases")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.30, scale = 0.27))
    }else if(name == .prior_clean_input_name("Injuries")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.46, scale = 0.19))
    }else if(name == .prior_clean_input_name("Kidney and Transplant")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.51, scale = 0.23))
    }else if(name == .prior_clean_input_name("Lung Cancer")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.49, scale = 0.26))
    }else if(name == .prior_clean_input_name("Metabolic and Endocrine Disorders")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.55, scale = 0.19))
    }else if(name == .prior_clean_input_name("Methodology")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.74, scale = 0.23))
    }else if(name == .prior_clean_input_name("Movement Disorders")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.54, scale = 0.24))
    }else if(name == .prior_clean_input_name("Multiple Sclerosis and Rare Diseases of the CNS")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.58, scale = 0.35))
    }else if(name == .prior_clean_input_name("Musculoskeletal")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.54, scale = 0.31))
    }else if(name == .prior_clean_input_name("Neonatal")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.89, scale = 0.30))
    }else if(name == .prior_clean_input_name("Neuromuscular")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.44, scale = 0.15))
    }else if(name == .prior_clean_input_name("Oral Health")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.50, scale = 0.37))
    }else if(name == .prior_clean_input_name("Pain, Palliative and Supportive Care")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.98, scale = 0.39))
    }else if(name == .prior_clean_input_name("Pregnancy and Childbirth")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.40, scale = 0.21))
    }else if(name == .prior_clean_input_name("Schizophrenia")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.47, scale = 0.26))
    }else if(name == .prior_clean_input_name("Sexually Transmitted Infections")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.43, scale = 0.21))
    }else if(name == .prior_clean_input_name("Skin")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.49, scale = 0.23))
    }else if(name == .prior_clean_input_name("Stroke")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.52, scale = 0.15))
    }else if(name == .prior_clean_input_name("Tobacco Addiction")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 2.02, scale = 0.36))
    }else if(name == .prior_clean_input_name("Urology")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.37, scale = 0.18))
    }else if(name == .prior_clean_input_name("Vascular")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.26, scale = 0.19))
    }else if(name == .prior_clean_input_name("Work")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.50, scale = 0.26))
    }else if(name == .prior_clean_input_name("Wounds")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.63, scale = 0.38))
    }else if(name == .prior_clean_input_name("Cochrane")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 1.51, scale = 0.23))
    }else{
      stop("unknown subfield 'name' argument for an informed prior distribution from medicine.")
    }

  }else if(type =="rd"){

    # medical priors for dichotomous outcomes based on Bartoš et al. 2023
    if(name == .prior_clean_input_name("Acute Respiratory Infections")){
      p <- prior(distribution ="normal", parameters = list(mean = 0, sd = 0.102), truncation = list(lower = 0))
    }else if(name == .prior_clean_input_name("Airways")){
      p <- prior(distribution ="normal", parameters = list(mean = 0, sd = 0.099), truncation = list(lower = 0))
    }else if(name == .prior_clean_input_name("Anaesthesia")){
      p <- prior(distribution ="normal", parameters = list(mean = 0, sd = 0.108), truncation = list(lower = 0))
    }else if(name == .prior_clean_input_name("Back and Neck")){
      p <- prior(distribution ="normal", parameters = list(mean = 0, sd = 0.102), truncation = list(lower = 0))
    }else if(name == .prior_clean_input_name("Bone, Joint and Muscle Trauma")){
      p <- prior(distribution ="normal", parameters = list(mean = 0, sd = 0.119), truncation = list(lower = 0))
    }else if(name == .prior_clean_input_name("Breast Cancer")){
      p <- prior(distribution ="normal", parameters = list(mean = 0, sd = 0.118), truncation = list(lower = 0))
    }else if(name == .prior_clean_input_name("Childhood Cancer")){
      stop("no informed subfield prior distribution for 'Childhood Cancer' category.")
    }else if(name == .prior_clean_input_name("Colorectal")){
      p <- prior(distribution ="normal", parameters = list(mean = 0, sd = 0.096), truncation = list(lower = 0))
    }else if(name == .prior_clean_input_name("Common Mental Disorders")){
      p <- prior(distribution ="normal", parameters = list(mean = 0, sd = 0.108), truncation = list(lower = 0))
    }else if(name == .prior_clean_input_name("Consumers and Communication")){
      p <- prior(distribution ="normal", parameters = list(mean = 0, sd = 0.103), truncation = list(lower = 0))
    }else if(name == .prior_clean_input_name("Cystic Fibrosis and Genetic Disorders")){
      p <- prior(distribution ="normal", parameters = list(mean = 0, sd = 0.092), truncation = list(lower = 0))
    }else if(name == .prior_clean_input_name("Dementia and Cognitive Improvement")){
      p <- prior(distribution ="normal", parameters = list(mean = 0, sd = 0.092), truncation = list(lower = 0))
    }else if(name == .prior_clean_input_name("Developmental, Psychosocial and Learning Problems")){
      p <- prior(distribution ="normal", parameters = list(mean = 0, sd = 0.132), truncation = list(lower = 0))
    }else if(name == .prior_clean_input_name("Drugs and Alcohol")){
      p <- prior(distribution ="normal", parameters = list(mean = 0, sd = 0.101), truncation = list(lower = 0))
    }else if(name == .prior_clean_input_name("Effective Practice and Organisation of Care")){
      p <- prior(distribution ="normal", parameters = list(mean = 0, sd = 0.106), truncation = list(lower = 0))
    }else if(name == .prior_clean_input_name("Emergency and Critical Care")){
      p <- prior(distribution ="normal", parameters = list(mean = 0, sd = 0.068), truncation = list(lower = 0))
    }else if(name == .prior_clean_input_name("ENT")){
      p <- prior(distribution ="normal", parameters = list(mean = 0, sd = 0.123), truncation = list(lower = 0))
    }else if(name == .prior_clean_input_name("Epilepsy")){
      p <- prior(distribution ="normal", parameters = list(mean = 0, sd = 0.086), truncation = list(lower = 0))
    }else if(name == .prior_clean_input_name("Eyes and Vision")){
      p <- prior(distribution ="normal", parameters = list(mean = 0, sd = 0.112), truncation = list(lower = 0))
    }else if(name == .prior_clean_input_name("Fertility Regulation")){
      p <- prior(distribution ="normal", parameters = list(mean = 0, sd = 0.082), truncation = list(lower = 0))
    }else if(name == .prior_clean_input_name("Gut")){
      p <- prior(distribution ="normal", parameters = list(mean = 0, sd = 0.090), truncation = list(lower = 0))
    }else if(name == .prior_clean_input_name("Gynaecological, Neuro-oncology and Orphan Cancer")){
      p <- prior(distribution ="normal", parameters = list(mean = 0, sd = 0.103), truncation = list(lower = 0))
    }else if(name == .prior_clean_input_name("Gynaecology and Fertility")){
      p <- prior(distribution ="normal", parameters = list(mean = 0, sd = 0.086), truncation = list(lower = 0))
    }else if(name == .prior_clean_input_name("Haematology")){
      p <- prior(distribution ="normal", parameters = list(mean = 0, sd = 0.089), truncation = list(lower = 0))
    }else if(name == .prior_clean_input_name("Heart")){
      p <- prior(distribution ="normal", parameters = list(mean = 0, sd = 0.070), truncation = list(lower = 0))
    }else if(name == .prior_clean_input_name("Heart; Vascular")){
      p <- prior(distribution ="normal", parameters = list(mean = 0, sd = 0.116), truncation = list(lower = 0))
    }else if(name == .prior_clean_input_name("Hepato-Biliary")){
      p <- prior(distribution ="normal", parameters = list(mean = 0, sd = 0.093), truncation = list(lower = 0))
    }else if(name == .prior_clean_input_name("HIV/AIDS")){
      p <- prior(distribution ="normal", parameters = list(mean = 0, sd = 0.097), truncation = list(lower = 0))
    }else if(name == .prior_clean_input_name("Hypertension")){
      p <- prior(distribution ="normal", parameters = list(mean = 0, sd = 0.059), truncation = list(lower = 0))
    }else if(name == .prior_clean_input_name("Incontinence")){
      p <- prior(distribution ="normal", parameters = list(mean = 0, sd = 0.125), truncation = list(lower = 0))
    }else if(name == .prior_clean_input_name("Infectious Diseases")){
      p <- prior(distribution ="normal", parameters = list(mean = 0, sd = 0.098), truncation = list(lower = 0))
    }else if(name == .prior_clean_input_name("Injuries")){
      p <- prior(distribution ="normal", parameters = list(mean = 0, sd = 0.137), truncation = list(lower = 0))
    }else if(name == .prior_clean_input_name("Kidney and Transplant")){
      p <- prior(distribution ="normal", parameters = list(mean = 0, sd = 0.089), truncation = list(lower = 0))
    }else if(name == .prior_clean_input_name("Lung Cancer")){
      p <- prior(distribution ="normal", parameters = list(mean = 0, sd = 0.116), truncation = list(lower = 0))
    }else if(name == .prior_clean_input_name("Metabolic and Endocrine Disorders")){
      p <- prior(distribution ="normal", parameters = list(mean = 0, sd = 0.091), truncation = list(lower = 0))
    }else if(name == .prior_clean_input_name("Methodology")){
      p <- prior(distribution ="normal", parameters = list(mean = 0, sd = 0.102), truncation = list(lower = 0))
    }else if(name == .prior_clean_input_name("Movement Disorders")){
      p <- prior(distribution ="normal", parameters = list(mean = 0, sd = 0.071), truncation = list(lower = 0))
    }else if(name == .prior_clean_input_name("Multiple Sclerosis and Rare Diseases of the CNS")){
      p <- prior(distribution ="normal", parameters = list(mean = 0, sd = 0.101), truncation = list(lower = 0))
    }else if(name == .prior_clean_input_name("Musculoskeletal")){
      p <- prior(distribution ="normal", parameters = list(mean = 0, sd = 0.109), truncation = list(lower = 0))
    }else if(name == .prior_clean_input_name("Neonatal")){
      p <- prior(distribution ="normal", parameters = list(mean = 0, sd = 0.088), truncation = list(lower = 0))
    }else if(name == .prior_clean_input_name("Neuromuscular")){
      p <- prior(distribution ="normal", parameters = list(mean = 0, sd = 0.075), truncation = list(lower = 0))
    }else if(name == .prior_clean_input_name("Oral Health")){
      p <- prior(distribution ="normal", parameters = list(mean = 0, sd = 0.121), truncation = list(lower = 0))
    }else if(name == .prior_clean_input_name("Pain, Palliative and Supportive Care")){
      p <- prior(distribution ="normal", parameters = list(mean = 0, sd = 0.106), truncation = list(lower = 0))
    }else if(name == .prior_clean_input_name("Pregnancy and Childbirth")){
      p <- prior(distribution ="normal", parameters = list(mean = 0, sd = 0.082), truncation = list(lower = 0))
    }else if(name == .prior_clean_input_name("Schizophrenia")){
      p <- prior(distribution ="normal", parameters = list(mean = 0, sd = 0.102), truncation = list(lower = 0))
    }else if(name == .prior_clean_input_name("Sexually Transmitted Infections")){
      p <- prior(distribution ="normal", parameters = list(mean = 0, sd = 0.117), truncation = list(lower = 0))
    }else if(name == .prior_clean_input_name("Skin")){
      p <- prior(distribution ="normal", parameters = list(mean = 0, sd = 0.123), truncation = list(lower = 0))
    }else if(name == .prior_clean_input_name("Stroke")){
      p <- prior(distribution ="normal", parameters = list(mean = 0, sd = 0.060), truncation = list(lower = 0))
    }else if(name == .prior_clean_input_name("Tobacco Addiction")){
      p <- prior(distribution ="normal", parameters = list(mean = 0, sd = 0.065), truncation = list(lower = 0))
    }else if(name == .prior_clean_input_name("Urology")){
      p <- prior(distribution ="normal", parameters = list(mean = 0, sd = 0.102), truncation = list(lower = 0))
    }else if(name == .prior_clean_input_name("Vascular")){
      p <- prior(distribution ="normal", parameters = list(mean = 0, sd = 0.119), truncation = list(lower = 0))
    }else if(name == .prior_clean_input_name("Work")){
      p <- prior(distribution ="normal", parameters = list(mean = 0, sd = 0.123), truncation = list(lower = 0))
    }else if(name == .prior_clean_input_name("Wounds")){
      p <- prior(distribution ="normal", parameters = list(mean = 0, sd = 0.088), truncation = list(lower = 0))
    }else if(name == .prior_clean_input_name("Cochrane")){
      p <- prior(distribution ="normal", parameters = list(mean = 0, sd = 0.099), truncation = list(lower = 0))
    }else{
      stop("unknown subfield 'name' argument for an informed prior distribution from medicine.")
    }

    return(p)

  }else if(type =="loghr"){

    # medical priors for dichotomous outcomes based on Bartoš et al. 2023
    if(name == .prior_clean_input_name("Cochrane")){
      p <- prior(distribution ="invgamma", parameters = list(shape = 2.42, scale = 0.30))
    }else{
      stop("There are no subfield specific prior distributions for logHR.")
    }

    return(p)
  }

}
.prior_informed.psychology             <- function(name){

  if(name == .prior_clean_input_name("van Erp")){
    p <- prior(distribution ="invgamma", parameters = list(shape = 1, scale = 0.15))
  }else if(name == .prior_clean_input_name("Oosterwijk")){
    p <- prior(distribution ="t", parameters = list(location = 0.35, scale = 0.102, df = 3), truncation = list(lower = 0))
  }else{
    stop("unknown prior 'name' for an informed prior distribution from psychology.")
  }

  return(p)
}

#' @title Names of medical subfields from the Cochrane database of systematic reviews
#'
#' @description Contains the union of names identifying the individual subfields from
#' the Cochrane database of systematic reviews. Type-specific availability depends on
#' the \code{type} argument passed to [prior_informed()].
#'
#' @examples
#' print(prior_informed_medicine_names)
#'
#' @return returns a character vector with names of medical subfields from Cochrane database of
#' systematic reviews.
#'
#' @seealso [prior_informed()]
#' @export
prior_informed_medicine_names <- c(
  "Acute Respiratory Infections",
  "Airways",
  "Anaesthesia",
  "Back and Neck",
  "Bone, Joint and Muscle Trauma",
  "Breast Cancer",
  "Childhood Cancer",
  "Colorectal",
  "Common Mental Disorders",
  "Consumers and Communication",
  "Cystic Fibrosis and Genetic Disorders",
  "Dementia and Cognitive Improvement",
  "Developmental, Psychosocial and Learning Problems",
  "Drugs and Alcohol",
  "Effective Practice and Organisation of Care",
  "Emergency and Critical Care",
  "ENT",
  "Epilepsy",
  "Eyes and Vision",
  "Fertility Regulation",
  "Gut",
  "Gynaecological, Neuro-oncology and Orphan Cancer",
  "Gynaecology and Fertility",
  "Haematology",
  "Heart",
  "Heart; Vascular",
  "Hepato-Biliary",
  "HIV/AIDS",
  "Hypertension",
  "Incontinence",
  "Infectious Diseases",
  "Inflammatory Bowel Disease",
  "Injuries",
  "Kidney and Transplant",
  "Lung Cancer",
  "Metabolic and Endocrine Disorders",
  "Methodology",
  "Movement Disorders",
  "Multiple Sclerosis and Rare Diseases of the CNS",
  "Musculoskeletal",
  "Neonatal",
  "Neuromuscular",
  "Oral Health",
  "Pain, Palliative and Supportive Care",
  "Pregnancy and Childbirth",
  "Public Health",
  "Schizophrenia",
  "Sexually Transmitted Infections",
  "Skin",
  "Stroke",
  "Tobacco Addiction",
  "Upper GI and Pancreatic Diseases",
  "Urology",
  "Vascular",
  "Work",
  "Wounds",
  "Cochrane"
)

.prior_informed_psychology_names <- c(
  "van Erp",
  "Oosterwijk"
)

#' @title Compute marginal likelihood of a 'JAGS' model
#'
#' @description A wrapper around
#' \link[bridgesampling]{bridge_sampler} that automatically
#' computes likelihood part dependent on the prior distribution
#' and prepares parameter samples. \code{log_posterior} must
#' specify a function that takes two arguments - a named list
#' of samples from the prior distributions and the data, and returns
#' log likelihood of the model part.
#'
#' @param fit model fitted with either \link[runjags]{runjags} posterior
#' samples obtained with \link[rjags]{rjags-package}
#' @param log_posterior function that takes a named list of samples, the data,
#' and additional list of parameters passed as \code{...} as input and
#' returns the log of the unnormalized posterior density of the model part
#' @param data list containing data to fit the model (not including data for the formulas)
#' @param prior_list named list of prior distribution
#' (names correspond to the parameter names) of parameters not specified within the
#' \code{formula_list}
#' @param formula_list named list of formulas to be added to the model
#' (names correspond to the parameter name created by each of the formula)
#' @param formula_data_list named list of data frames containing data for each formula
#' (names of the lists correspond to the parameter name created by each of the formula)
#' @param formula_prior_list named list of named lists of prior distributions
#' (names of the lists correspond to the parameter name created by each of the formula and
#' the names of the prior distribution correspond to the parameter names) of parameters specified
#' within the \code{formula}
#' @param formula_scale_list named list of named lists for standardizing continuous predictors
#' (names of the lists correspond to the parameter name created by each of the formula).
#' Each entry should be a named list where continuous predictors with \code{TRUE} values will
#' be standardized. Defaults to \code{NULL} (no standardization).
#' @param add_parameters vector of additional parameter names that should be used
#' in bridgesampling but were not specified in the \code{prior_list}
#' @param add_bounds list with two name vectors (\code{"lb"} and \code{"up"})
#' containing lower and upper bounds of the additional parameters that were not
#' specified in the \code{prior_list}
#' @param maxiter maximum number of iterations for the
#' \link[bridgesampling]{bridge_sampler}
#' @param silent whether the progress should be printed, defaults to \code{TRUE}
#' @param ... additional argument to the \link[bridgesampling]{bridge_sampler}
#' and \code{log_posterior} function
#'
#' @examples \dontrun{
#' # simulate data
#' set.seed(1)
#' data <- list(
#'   x = rnorm(10),
#'   N = 10
#' )
#' data$x
#'
#' # define priors
#' priors_list <- list(mu = prior("normal", list(0, 1)))
#'
#' # define likelihood for the data
#' model_syntax <-
#'   "model{
#'     for(i in 1:N){
#'       x[i] ~ dnorm(mu, 1)
#'     }
#'   }"
#'
#' # fit the models
#' fit <- JAGS_fit(model_syntax, data, priors_list)
#'
#' # define log posterior for bridge sampling
#' log_posterior <- function(parameters, data){
#'   sum(dnorm(data$x, parameters$mu, 1, log = TRUE))
#' }
#'
#' # get marginal likelihoods
#' marglik <- JAGS_bridgesampling(fit, log_posterior, data, priors_list)
#' }
#' @return \code{JAGS_bridgesampling} returns an object of class 'bridge'.
#'
#' @export
JAGS_bridgesampling <- function(fit, log_posterior, data = NULL, prior_list = NULL, formula_list = NULL, formula_data_list = NULL, formula_prior_list = NULL, formula_scale_list = NULL,
                                add_parameters = NULL, add_bounds = NULL,
                                maxiter = 10000, silent = TRUE, ...){

  ### check input
  check_bool(silent, "silent")
  check_int(maxiter, "maxiter", lower = 1)
  check_list(formula_list, "formula_list", allow_NULL = TRUE)
  check_list(formula_data_list, "formula_data_list", check_names = names(formula_list), allow_other = FALSE, all_objects = TRUE, allow_NULL = is.null(formula_list))
  check_list(formula_prior_list, "formula_prior_list", check_names = names(formula_list), allow_other = FALSE, all_objects = TRUE, allow_NULL = is.null(formula_list))
  check_list(formula_scale_list, "formula_scale_list", allow_NULL = TRUE)

  if(is.null(formula_scale_list) && !is.null(formula_list)){
    formula_scale_list <- .JAGS_formula_scale_list_from_fit(fit, names(formula_list))
  }

  # extract the posterior distribution
  posterior <- .fit_to_posterior(fit)

  ### prepare formula objects summary
  if(!is.null(formula_list)){

    # obtain settings for each formula
    formula_output <- list()
    for(parameter in names(formula_list)){
      formula_output[[parameter]] <- JAGS_formula(
        formula       = formula_list[[parameter]],
        parameter     = parameter,
        data          = formula_data_list[[parameter]],
        prior_list    = formula_prior_list[[parameter]],
        formula_scale = if(!is.null(formula_scale_list)) formula_scale_list[[parameter]] else NULL)
    }

    # merge with the rest of the input
    formula_list       <- lapply(formula_output, function(output) output[["formula"]])
    formula_prior_list <- lapply(formula_output, function(output) output[["prior_list"]])
    formula_data_list  <- lapply(formula_output, function(output) output[["data"]])

    all_prior_list <- c(prior_list, do.call(c, unname(formula_prior_list)))

  }else{
    all_prior_list <- prior_list
  }

  if(any(sapply(all_prior_list, is.prior.discrete)))
    stop("Discrete or spike and slab priors are not supported with bridgesampling.")

  ### extract relevant variables and upper and lower bound
  bridgesampling_posterior <- JAGS_bridgesampling_posterior(posterior = posterior, prior_list = all_prior_list, add_parameters = add_parameters, add_bounds = add_bounds)
  if(ncol(bridgesampling_posterior) == 0)
    stop("Bridge sampling cannot proceed without any estimated parameter")


  ### define the marglik function
  full_log_posterior <- function(samples.row, data, prior_list, formula_list, formula_data_list, formula_prior_list, add_parameters, ...){

    # prepare object for holding the parameters, later accessible to the user specified 'log_posterior'
    parameters <- list()
    if(length(prior_list) > 0){
      parameters <- c(parameters, JAGS_marglik_parameters(samples.row, prior_list))
    }
    if(length(formula_prior_list) > 0){
      parameters <- c(parameters, JAGS_marglik_parameters_formula(samples.row, formula_list, formula_data_list, formula_prior_list, parameters))
    }
    if(length(add_parameters) > 0){
      parameters <- c(parameters, samples.row[add_parameters])
    }

    # compute the marginal likelihoods
    marglik <- 0
    if(length(prior_list) > 0){
      marglik <- marglik + JAGS_marglik_priors(samples.row, prior_list)
    }
    if(length(formula_prior_list) > 0){
      marglik <- marglik + JAGS_marglik_priors_formula(samples.row, formula_prior_list)
    }
    marglik   <- marglik + log_posterior(parameters = parameters, data = data, ...)

    return(marglik)
  }


  ### perform bridgesampling
  marglik <- tryCatch(suppressWarnings(bridgesampling::bridge_sampler(
      samples            = bridgesampling_posterior,
      data               = data,
      log_posterior      = full_log_posterior,
      prior_list         = prior_list,
      formula_list       = formula_list,
      formula_data_list  = formula_data_list,
      formula_prior_list = formula_prior_list,
      lb                 = attr(bridgesampling_posterior, "lb"),
      ub                 = attr(bridgesampling_posterior, "ub"),
      silent             = silent,
      maxiter            = maxiter,
      add_parameters     = add_parameters,
      ...
    )), error = function(e)e)

  # add a warning attribute and call the warning if not silent
  if(!inherits(marglik, "error") && marglik[["niter"]] > maxiter){
    attr(marglik, "warning") <- "Marginal likelihood could not be estimated within the maximum number of itetations and might be more variable than usual."
    if(!silent)
      warning(attr(marglik, "warning"), immediate. = TRUE)
  }

  return(marglik)
}

.JAGS_formula_scale_list_from_fit <- function(fit, formula_parameters){

  formula_scale <- attr(fit, "formula_scale")
  if(is.null(formula_scale) || length(formula_scale) == 0L){
    return(NULL)
  }

  scale_list <- vector("list", length(formula_parameters))
  names(scale_list) <- formula_parameters

  for(parameter in intersect(formula_parameters, names(formula_scale))){
    parameter_scale <- formula_scale[[parameter]]
    if(is.null(parameter_scale) || length(parameter_scale) == 0L){
      next
    }

    scaled_terms <- names(parameter_scale)
    if(is.null(scaled_terms) || length(scaled_terms) == 0L){
      next
    }

    parameter_prefix <- paste0(parameter, "_")
    predictor_terms  <- ifelse(
      startsWith(scaled_terms, parameter_prefix),
      substring(scaled_terms, nchar(parameter_prefix) + 1L),
      scaled_terms
    )

    scale_list[[parameter]] <- as.list(stats::setNames(rep(TRUE, length(predictor_terms)), predictor_terms))
  }

  scale_list <- scale_list[!vapply(scale_list, is.null, logical(1))]
  if(length(scale_list) == 0L){
    return(NULL)
  }

  return(scale_list)
}

.fit_to_posterior <- function(fit){

  ### check the input and split it on posterior and data
  if(inherits(fit, "runjags")){

    # get posterior and merge chains
    posterior <- .extract_posterior_samples(fit, as_list = FALSE)

  }else if(is.list(fit) & all(sapply(fit, inherits, what = "mcarray"))){

    # rjags model with rjags::jags.samples
    # merge chains
    posterior <- do.call(cbind, lapply(names(fit), function(par){
      if(dim(fit[[par]])[1] > 1){
        samples <- do.call(rbind, lapply(1:(dim(fit[[par]]))[3],  function(chain)t(fit[[par]][,,chain])))
        colnames(samples) <- paste0(attr(fit[[par]], "varname"), "[",1:ncol(samples),"]")
      }else{
        samples <- matrix(do.call(c, lapply(1:(dim(fit[[par]]))[3],  function(chain)fit[[par]][,,chain])), ncol = 1)
        colnames(samples) <- attr(fit[[par]], "varname")
      }
      return(samples)
    }))

  }else if(inherits(fit, "mcmc.list")){

    # rjags model with rjags::coda.samples or samples extracted via coda::as.mcmc.list
    # merge chains
    posterior <- do.call(rbind, fit)

  }else if (inherits(fit, "mcmc") && length(dim(fit)) == 2) {

    # rjags model with samples extracted via coda::as.mcmc
    return(fit)

  } else {

    stop("the method is not implemented for this output")

  }

  return(posterior)
}


#' @title Create a 'bridgesampling' object
#'
#' @description prepares a 'bridgesampling' object with a given
#' log marginal likelihood.
#'
#' @param logml log marginal likelihood. Defaults to \code{-Inf}.
#'
#'
#' @return \code{JAGS_bridgesampling} returns an object of class 'bridge'.
#'
#' @export
bridgesampling_object <- function(logml = -Inf){

  marglik        <- list()
  marglik$logml  <- logml
  class(marglik) <- "bridge"

  return(marglik)
}


#' @title Prepare 'JAGS' posterior for 'bridgesampling'
#'
#' @description prepares posterior distribution for 'bridgesampling'
#' by removing unnecessary parameters and attaching lower and upper
#' bounds of parameters based on a list of prior distributions.
#'
#' @param posterior matrix of mcmc samples from the posterior
#' distribution
#'
#' @inheritParams JAGS_bridgesampling
#'
#' @return \code{JAGS_bridgesampling_posterior} returns a matrix of
#' posterior samples with 'lb' and 'ub' attributes carrying the
#' lower and upper boundaries.
#'
#' @export
JAGS_bridgesampling_posterior <- function(posterior, prior_list, add_parameters = NULL, add_bounds = NULL){

  # check the input
  if(!is.matrix(posterior))
    stop("'posterior' must be a matrix")
  if(!is.null(prior_list)){
    if(!is.list(prior_list))
      stop("'prior_list' must be a list.")
    if(is.prior(prior_list) | !all(sapply(prior_list, is.prior)))
      stop("'prior_list' must be a list of priors.")
  }
  if(!is.null(add_parameters)){
    if(!is.character(add_parameters))
      stop("'add_parameters' must be a character vector.")
    if(!is.list(add_bounds))
      stop("'add_bounds' must be a list.")
    if(length(add_bounds) != 2 | !all(names(add_bounds) %in% c("lb", "ub")))
      stop("'add_bounds' must contain lower and upper bounds ('lb' and 'ub').")
    if(length(add_bounds[["lb"]]) != length(add_parameters) | length(add_bounds[["ub"]]) != length(add_parameters))
      stop("lb' and 'ub' must have the same lenght as the 'add_parameters'.")
    if(!is.numeric(add_bounds[["lb"]]) | !is.numeric(add_bounds[["ub"]]))
      stop("lb' and 'ub' must be numeric vectors.")
  }

  # these are not generally possible because the component indicators are discrete and bridgesampling
  # package cannot currently deal with them
  if(any(sapply(prior_list, is.prior.spike_and_slab)))
    stop("Marginal likelihood computation for spike and slab priors is not implemented.")
  if(any(sapply(prior_list, is.prior.mixture))){
    .JAGS_marglik_stop_unsupported_mixture(prior_list[[which(sapply(prior_list, is.prior.mixture))[1L]]])
  }

  # get information about the specified parameters
  parameters_names <- .JAGS_bridgesampling_posterior_info(prior_list)

  # add the user defined parameters
  if(!is.null(add_parameters)){
    parameters_names_lb <- c(attr(parameters_names, "lb"), add_bounds[["lb"]])
    parameters_names_ub <- c(attr(parameters_names, "ub"), add_bounds[["ub"]])
    parameters_names <- c(parameters_names, add_parameters)
    attr(parameters_names, "lb") <- parameters_names_lb
    attr(parameters_names, "ub") <- parameters_names_ub
  }


  # check that all parameter names exist in the posterior
  if(!all(parameters_names %in% colnames(posterior)))
    stop("'posterior' does not contain all of the parameters corresponding to the 'prior_list' and the 'add_parameter' argument.")

  posterior <- posterior[,parameters_names, drop = FALSE]
  attr(posterior, "lb") <- attr(parameters_names, "lb")
  attr(posterior, "ub") <- attr(parameters_names, "ub")

  return(posterior)
}

.JAGS_marglik_stop_unsupported_mixture <- function(prior){

  if(inherits(prior, "prior.bias_mixture")){
    selection_backend_spec(prior)
    stop(
      "Marginal likelihood computation for bias mixture priors is not implemented because bridge sampling does not support discrete bias indicators.",
      call. = FALSE
    )
  }

  stop("Marginal likelihood computation for prior mixture priors is not implemented.", call. = FALSE)
}

.JAGS_bridgesampling_posterior_info                <- function(prior_list){

  # return empty string in case that no prior was specified
  if(length(prior_list) == 0){
    return(list())
  }

  if(!is.list(prior_list))
    stop("'prior_list' must be a list.")
  if(is.prior(prior_list) | !all(sapply(prior_list, is.prior)))
    stop("'prior_list' must be a list of priors.")
  .check_prior_list_unique_names(prior_list)


  # add the resulting parameters
  parameters    <- NULL
  parameters_lb <- NULL
  parameters_ub <- NULL
  for(i in seq_along(prior_list)){

    add_parameter <- NULL

    if(is.prior.weightfunction(prior_list[[i]])){

      add_parameter <- .JAGS_bridgesampling_posterior_info.weightfunction(prior_list[[i]])

    }else if(is_prior_phacking(prior_list[[i]])){

      add_parameter <- .JAGS_bridgesampling_posterior_info.phacking(prior_list[[i]])

    }else if(is_prior_bias(prior_list[[i]])){

      add_parameter <- .JAGS_bridgesampling_posterior_info.bias(prior_list[[i]])

    }else if(is.prior.mixture(prior_list[[i]])){

      .JAGS_marglik_stop_unsupported_mixture(prior_list[[i]])

    }else if(is.prior.PET(prior_list[[i]]) | is.prior.PEESE(prior_list[[i]])){

      add_parameter <- .JAGS_bridgesampling_posterior_info.PP(prior_list[[i]])

    }else if(is.prior.factor(prior_list[[i]])){

      add_parameter <- .JAGS_bridgesampling_posterior_info.factor(prior_list[[i]], names(prior_list)[i])

    }else if(is.prior.vector(prior_list[[i]])){

      add_parameter <- .JAGS_bridgesampling_posterior_info.vector(prior_list[[i]], names(prior_list)[i])

    }else if(is.prior.simple(prior_list[[i]])){

      add_parameter <- .JAGS_bridgesampling_posterior_info.simple(prior_list[[i]], names(prior_list)[i])

    }

    if(!is.null(add_parameter)){
      parameters    <- c(parameters,    add_parameter)
      parameters_lb <- c(parameters_lb, attr(add_parameter, "lb"))
      parameters_ub <- c(parameters_ub, attr(add_parameter, "ub"))
    }
  }

  attr(parameters, "lb") <- parameters_lb
  attr(parameters, "ub") <- parameters_ub

  return(parameters)
}
.JAGS_bridgesampling_posterior_info.simple         <- function(prior, parameter_name){

  .check_prior(prior)
  if(!is.prior.simple(prior))
    stop("improper prior provided")
  check_char(parameter_name, "parameter_name")


  if(prior[["distribution"]] == "invgamma"){
    parameter <- paste0("inv_", parameter_name)
    attr(parameter, "lb") <- prior$truncation[["upper"]]^-1
    attr(parameter, "ub") <- prior$truncation[["lower"]]^-1
  }else if(prior[["distribution"]] == "point"){
    parameter <- NULL
  }else{
    parameter <- parameter_name
    attr(parameter, "lb") <- prior$truncation[["lower"]]
    attr(parameter, "ub") <- prior$truncation[["upper"]]
  }

  names(attr(parameter, "lb")) <- parameter
  names(attr(parameter, "ub")) <- parameter

  return(parameter)
}
.JAGS_bridgesampling_posterior_info.vector         <- function(prior, parameter_name){

  .check_prior(prior)
  if(!is.prior.vector(prior))
    stop("improper prior provided")
  check_char(parameter_name, "parameter_name")
  check_int(prior$parameters[["K"]], "K", lower = 1)
  if(prior[["distribution"]] != "mpoint")
    .check_vector_truncation_unsupported(prior$truncation)

  if(prior[["distribution"]] == "mpoint"){
    parameter <- NULL
  }else{
    if(prior$parameters[["K"]] == 1){
      parameter <- parameter_name
    }else{
      parameter <- paste0(parameter_name, "[", 1:prior$parameters[["K"]], "]")
    }

    attr(parameter, "lb") <- rep(prior$truncation[["lower"]], prior$parameters[["K"]])
    attr(parameter, "ub") <- rep(prior$truncation[["upper"]], prior$parameters[["K"]])

    names(attr(parameter, "lb")) <- parameter
    names(attr(parameter, "ub")) <- parameter
  }

  return(parameter)
}
.JAGS_bridgesampling_posterior_info.factor         <- function(prior, parameter_name){

  .check_prior(prior)
  if(!is.prior.factor(prior))
    stop("improper prior provided")
  check_char(parameter_name, "parameter_name")

  if(is.prior.treatment(prior) | is.prior.independent(prior)){

    if(.get_prior_factor_levels(prior) == 1){

      parameter <- .JAGS_bridgesampling_posterior_info.simple(prior, parameter_name)

    }else{

      parameter    <- NULL
      parameter_lb <- NULL
      parameter_ub <- NULL

      for(i in 1:.get_prior_factor_levels(prior)){

        add_parameter <- .JAGS_bridgesampling_posterior_info.simple(prior, paste0(parameter_name, "[", i, "]"))

        parameter    <- c(parameter,    add_parameter)
        parameter_lb <- c(parameter_lb, attr(add_parameter, "lb"))
        parameter_ub <- c(parameter_ub, attr(add_parameter, "ub"))
      }

      attr(parameter, "lb") <- parameter_lb
      attr(parameter, "ub") <- parameter_ub

    }

  }else if(is.prior.orthonormal(prior) | is.prior.meandif(prior)){

    prior$parameters[["K"]] <- .get_prior_factor_levels(prior)

    parameter <- .JAGS_bridgesampling_posterior_info.vector(prior, parameter_name)

  }

  return(parameter)
}
.JAGS_bridgesampling_posterior_info.PP             <- function(prior){

  .check_prior(prior)
  if(!is.prior.PET(prior) & !is.prior.PEESE(prior))
    stop("improper prior provided")

  if(is.prior.PET(prior)){
    parameter <- .JAGS_bridgesampling_posterior_info.simple(prior, "PET")
  }else if(is.prior.PEESE(prior)){
    parameter <- .JAGS_bridgesampling_posterior_info.simple(prior, "PEESE")
  }

  return(parameter)
}
.JAGS_bridgesampling_posterior_info.weightfunction <- function(prior){

  .check_prior(prior)
  if(!is.prior.weightfunction(prior))
    stop("improper prior provided")

  J <- .weightfunction_n_bins(prior)

  if(prior$weights$type == "cumulative"){

    parameter <- paste0("eta[", seq_len(J), "]")
    attr(parameter, "lb") <- rep(0,   length(parameter))
    attr(parameter, "ub") <- rep(Inf, length(parameter))

  }else if(prior$weights$type == "independent" && prior$weights$scale == "omega"){

    parameter <- if(J > 1L) paste0("omega[", 2:J, "]") else NULL
    attr(parameter, "lb") <- rep(prior$weights$prior$truncation[["lower"]], length(parameter))
    attr(parameter, "ub") <- rep(prior$weights$prior$truncation[["upper"]], length(parameter))

  }else if(prior$weights$type == "independent" && prior$weights$scale == "log_omega"){

    parameter <- if(J > 1L) paste0("log_omega[", 2:J, "]") else NULL
    attr(parameter, "lb") <- rep(prior$weights$prior$truncation[["lower"]], length(parameter))
    attr(parameter, "ub") <- rep(prior$weights$prior$truncation[["upper"]], length(parameter))

  }else if(prior$weights$type == "fixed"){

    parameter <- NULL

  }

  if(!is.null(parameter)){
    names(attr(parameter, "lb")) <- parameter
    names(attr(parameter, "ub")) <- parameter
  }

  return(parameter)
}
.JAGS_bridgesampling_posterior_info.phacking <- function(prior){

  .check_prior(prior)
  if(!is_prior_phacking(prior))
    stop("improper prior provided")

  parameter <- .JAGS_bridgesampling_posterior_info.simple(prior$alpha, "alpha")

  return(parameter)
}
.JAGS_bridgesampling_posterior_info.bias <- function(prior){

  .check_prior(prior)
  if(!is_prior_bias(prior))
    stop("improper prior provided")

  selection_backend_spec(prior)

  parameter <- NULL
  parameter_lb <- NULL
  parameter_ub <- NULL

  if(!is.null(prior$selection)){
    selection_parameter <- .JAGS_bridgesampling_posterior_info.weightfunction(prior$selection)
    parameter <- c(parameter, selection_parameter)
    parameter_lb <- c(parameter_lb, attr(selection_parameter, "lb"))
    parameter_ub <- c(parameter_ub, attr(selection_parameter, "ub"))
  }
  if(!is.null(prior$phacking)){
    phacking_parameter <- .JAGS_bridgesampling_posterior_info.phacking(prior$phacking)
    parameter <- c(parameter, phacking_parameter)
    parameter_lb <- c(parameter_lb, attr(phacking_parameter, "lb"))
    parameter_ub <- c(parameter_ub, attr(phacking_parameter, "ub"))
  }

  attr(parameter, "lb") <- parameter_lb
  attr(parameter, "ub") <- parameter_ub
  return(parameter)
}
# .JAGS_bridgesampling_posterior_info.spike_and_slab <- function(prior, parameter_name){
#
#   .check_prior(prior)
#   if(!is.prior.spike_and_slab(prior))
#     stop("improper prior provided")
#   check_char(parameter_name, "parameter_name")
#
#   if(!is.prior.point(prior[["inclusion"]])){
#
#     parameter_variable  <- .JAGS_bridgesampling_posterior_info.simple(prior[["variable"]],  paste0(parameter_name, "_variable"))
#     parameter_inclusion <- .JAGS_bridgesampling_posterior_info.simple(prior[["inclusion"]], paste0(parameter_name, "_inclusion"))
#
#     parameter <- c(parameter_variable, parameter_inclusion)
#
#     attr(parameter, "lb") <- c(attr(parameter_variable, "lb"), attr(parameter_inclusion, "lb"))
#     attr(parameter, "ub") <- c(attr(parameter_variable, "ub"), attr(parameter_inclusion, "ub"))
#
#     names(attr(parameter, "lb")) <- c(names(attr(parameter_variable, "lb")), names(attr(parameter_inclusion, "lb")))
#     names(attr(parameter, "ub")) <- c(names(attr(parameter_variable, "ub")), names(attr(parameter_inclusion, "ub")))
#
#   }else{
#     parameter  <- .JAGS_bridgesampling_posterior_info.simple(prior[["variable"]],  paste0(parameter_name, "_variable"))
#   }
#
#
#   return(parameter)
# }

#' @title Compute marginal likelihood for 'JAGS' priors
#'
#' @description Computes marginal likelihood for the
#' prior part of a 'JAGS' model within 'bridgesampling'
#' function
#'
#' @param samples samples provided by bridgesampling
#' function
#'
#' @inheritParams JAGS_bridgesampling
#'
#' @return \code{JAGS_marglik_priors} returns a numeric value
#' of likelihood evaluated at the current posterior sample.
#'
#' @export JAGS_marglik_priors
#' @export JAGS_marglik_priors_formula
#' @name JAGS_marglik_priors
NULL

#' @rdname JAGS_marglik_priors
JAGS_marglik_priors                <- function(samples, prior_list){

  # return empty string in case that no prior was specified
  if(length(prior_list) == 0){
    return(list())
  }

  if(!is.list(prior_list))
    stop("'prior_list' must be a list.")
  if(is.prior(prior_list) | !all(sapply(prior_list, is.prior)))
    stop("'prior_list' must be a list of priors.")
  .check_prior_list_unique_names(prior_list)


  # add the resulting parameters
  marglik <- 0
  for(i in seq_along(prior_list)){

    if(is.prior.weightfunction(prior_list[[i]])){

      marglik <- marglik + .JAGS_marglik_priors.weightfunction(samples, prior_list[[i]])

    }else if(is_prior_phacking(prior_list[[i]])){

      marglik <- marglik + .JAGS_marglik_priors.phacking(samples, prior_list[[i]])

    }else if(is_prior_bias(prior_list[[i]])){

      marglik <- marglik + .JAGS_marglik_priors.bias(samples, prior_list[[i]])

    }else if(is.prior.mixture(prior_list[[i]])){

      .JAGS_marglik_stop_unsupported_mixture(prior_list[[i]])

    }else if(is.prior.PET(prior_list[[i]]) | is.prior.PEESE(prior_list[[i]])){

      marglik <- marglik + .JAGS_marglik_priors.PP(samples, prior_list[[i]])

    }else if(is.prior.factor(prior_list[[i]])){

      marglik <- marglik + .JAGS_marglik_priors.factor(samples, prior_list[[i]], names(prior_list)[i])

    }else if(is.prior.vector(prior_list[[i]])){

      marglik <- marglik + .JAGS_marglik_priors.vector(samples, prior_list[[i]], names(prior_list)[i])

    }else if(is.prior.simple(prior_list[[i]])){

      marglik <- marglik + .JAGS_marglik_priors.simple(samples, prior_list[[i]], names(prior_list)[i])

    }
  }

  return(marglik)
}


.JAGS_marglik_priors.simple         <- function(samples, prior, parameter_name){

  .check_prior(prior)
  if(!is.prior.simple(prior))
    stop("improper prior provided")
  check_char(parameter_name, "parameter_name")

  if(prior[["distribution"]] == "invgamma"){

    sampling_prior <- prior(
      "distribution" = "gamma",
      "parameters"   = list("shape" = prior$parameters[["shape"]], "rate" = prior$parameters[["scale"]]),
      "truncation"   = list("lower" = prior$truncation[["upper"]]^-1, "upper" = prior$truncation[["lower"]]^-1))
    marglik <- lpdf(sampling_prior, samples[[ paste0("inv_", parameter_name) ]])

  }else if(prior[["distribution"]] == "point"){

    marglik <- 0

  }else{

    marglik <- lpdf(prior, samples[[ parameter_name ]])

  }

  return(marglik)
}
.JAGS_marglik_priors.vector         <- function(samples, prior, parameter_name){

  .check_prior(prior)
  if(!is.prior.vector(prior))
    stop("improper prior provided")
  check_char(parameter_name, "parameter_name")
  check_int(prior$parameters[["K"]], "K", lower = 1)
  if(prior[["distribution"]] != "mpoint")
    .check_vector_truncation_unsupported(prior$truncation)

  if(prior[["distribution"]] == "mpoint"){
    marglik <- 0
  }else if(prior$parameters[["K"]] == 1){
    marglik <- lpdf(prior, samples[[ parameter_name ]])
  }else{
    marglik <- lpdf(prior, samples[ paste0(parameter_name, "[", 1:prior$parameters[["K"]], "]") ])
  }

  return(marglik)
}
.JAGS_marglik_priors.factor         <- function(samples, prior, parameter_name){

  .check_prior(prior)
  if(!is.prior.factor(prior))
    stop("improper prior provided")
  check_char(parameter_name, "parameter_name")

  if(is.prior.treatment(prior) | is.prior.independent(prior)){

    if(.get_prior_factor_levels(prior) == 1){

      marglik <- .JAGS_marglik_priors.simple(samples, prior, parameter_name)

    }else{

      marglik <- sum(sapply(1:.get_prior_factor_levels(prior), function(i) .JAGS_marglik_priors.simple(samples, prior, paste0(parameter_name, "[", i, "]"))))

    }

  }else if(is.prior.orthonormal(prior) | is.prior.meandif(prior)){

    prior$parameters[["K"]] <- .get_prior_factor_levels(prior)

    marglik <- .JAGS_marglik_priors.vector(samples, prior, parameter_name)

  }

  return(marglik)
}
.JAGS_marglik_priors.PP             <- function(samples, prior){

  .check_prior(prior)
  if(!is.prior.PET(prior) & !is.prior.PEESE(prior))
    stop("improper prior provided")

  if(is.prior.PET(prior)){
    marglik <- .JAGS_marglik_priors.simple(samples, prior, "PET")
  }else if(is.prior.PEESE(prior)){
    marglik <- .JAGS_marglik_priors.simple(samples, prior, "PEESE")
  }

  return(marglik)
}
.JAGS_marglik_priors.weightfunction <- function(samples, prior){

  .check_prior(prior)
  if(!is.prior.weightfunction(prior))
    stop("improper prior provided")

  J <- .weightfunction_n_bins(prior)

  if(prior$weights$type == "fixed"){

    marglik <- 0

  }else if(prior$weights$type == "cumulative"){

    marglik <- sum(stats::dgamma(samples[paste0("eta[", seq_len(J), "]")], shape = prior$weights$alpha, rate = 1, log = TRUE))

  }else if(prior$weights$type == "independent"){

    if(J == 1L){
      marglik <- 0
    }else if(prior$weights$scale == "omega"){
      marglik <- sum(mlpdf(prior$weights$prior, samples[paste0("omega[", 2:J, "]")]))
    }else if(prior$weights$scale == "log_omega"){
      marglik <- sum(mlpdf(prior$weights$prior, samples[paste0("log_omega[", 2:J, "]")]))
    }

  }

  return(marglik)
}
.JAGS_marglik_priors.phacking <- function(samples, prior){

  .check_prior(prior)
  if(!is_prior_phacking(prior))
    stop("improper prior provided")

  .JAGS_marglik_priors.simple(samples, prior$alpha, "alpha")
}
.JAGS_marglik_priors.bias <- function(samples, prior){

  .check_prior(prior)
  if(!is_prior_bias(prior))
    stop("improper prior provided")

  selection_backend_spec(prior)

  marglik <- 0
  if(!is.null(prior$selection)){
    marglik <- marglik + .JAGS_marglik_priors.weightfunction(samples, prior$selection)
  }
  if(!is.null(prior$phacking)){
    marglik <- marglik + .JAGS_marglik_priors.phacking(samples, prior$phacking)
  }

  return(marglik)
}
# .JAGS_marglik_priors.spike_and_slab <- function(samples, prior, parameter_name){
#
#   .check_prior(prior)
#   if(!is.prior.spike_and_slab(prior))
#     stop("improper prior provided")
#   check_char(parameter_name, "parameter_name")
#
#   marglik <- 0
#   if(!is.prior.point(prior[["inclusion"]])){
#     marglik <- marglik + .JAGS_marglik_priors.simple(samples, prior[["inclusion"]], paste0(parameter_name, "_inclusion"))
#   }
#
#   if(samples[[ paste0(if(prior[["variable"]][["distribution"]] == "invgamma") "inv_" else "", parameter_name, "_variable") ]] != 0){
#     marglik <- marglik + .JAGS_marglik_priors.simple(samples, prior[["variable"]], paste0(parameter_name, "_variable"))
#   }
#
#   return(marglik)
# }

#' @rdname JAGS_marglik_priors
JAGS_marglik_priors_formula <- function(samples, formula_prior_list){

  marglik <- 0

  for(parameter in names(formula_prior_list)){
    marglik <- marglik + JAGS_marglik_priors(samples, formula_prior_list[[parameter]])
  }

  return(marglik)
}

#' @title Extract parameters for 'JAGS' priors
#'
#' @description Extracts transformed parameters from the
#' prior part of a 'JAGS' model inside of a 'bridgesampling'
#' function (returns them as a named list)
#'
#' @param samples samples provided by bridgesampling
#' function
#' @param prior_list_parameters named list of prior distributions on model parameters
#' (not specified within the formula but that might scale the formula parameters)
#'
#' @return \code{JAGS_marglik_parameters} returns a named list
#' of (transformed) posterior samples.
#'
#' @inheritParams JAGS_bridgesampling
#' @export JAGS_marglik_parameters
#' @export JAGS_marglik_parameters_formula
#' @name JAGS_marglik_parameters
NULL

#' @rdname JAGS_marglik_parameters
JAGS_marglik_parameters                <- function(samples, prior_list){

  # return empty list in case that no prior was specified
  if(length(prior_list) == 0){
    return(list())
  }

  if(!is.list(prior_list))
    stop("'prior_list' must be a list.")
  if(is.prior(prior_list) | !all(sapply(prior_list, is.prior)))
    stop("'prior_list' must be a list of priors.")


  # add the resulting parameters
  parameters <- list()
  for(i in seq_along(prior_list)){

    if(is.prior.weightfunction(prior_list[[i]])){

      parameters <- c(parameters, .JAGS_marglik_parameters.weightfunction(samples, prior_list[[i]]))

    }else if(is_prior_phacking(prior_list[[i]])){

      parameters <- c(parameters, .JAGS_marglik_parameters.phacking(samples, prior_list[[i]]))

    }else if(is_prior_bias(prior_list[[i]])){

      parameters <- c(parameters, .JAGS_marglik_parameters.bias(samples, prior_list[[i]]))

    }else if(is.prior.mixture(prior_list[[i]])){

      .JAGS_marglik_stop_unsupported_mixture(prior_list[[i]])

    }else if(is.prior.PET(prior_list[[i]]) | is.prior.PEESE(prior_list[[i]])){

      parameters <- c(parameters, .JAGS_marglik_parameters.PP(samples, prior_list[[i]]))

    }else if(is.prior.factor(prior_list[[i]])){

      parameters <- c(parameters, .JAGS_marglik_parameters.factor(samples, prior_list[[i]], names(prior_list)[i]))

    }else if(is.prior.vector(prior_list[[i]])){

      parameters <- c(parameters, .JAGS_marglik_parameters.vector(samples, prior_list[[i]], names(prior_list)[i]))

    }else if(is.prior.simple(prior_list[[i]])){

      parameters <- c(parameters, .JAGS_marglik_parameters.simple(samples, prior_list[[i]], names(prior_list)[i]))

    }
  }

  return(parameters)
}


.JAGS_marglik_parameters.simple         <- function(samples, prior, parameter_name){

  .check_prior(prior)
  if(!is.prior.simple(prior))
    stop("improper prior provided")
  check_char(parameter_name, "parameter_name")


  parameter <- list()
  if(prior[["distribution"]] == "invgamma"){
    parameter[[parameter_name]] <- samples[[ paste0("inv_", parameter_name) ]]^-1
  }else if(prior[["distribution"]] == "point"){
    parameter[[parameter_name]] <- prior$parameters[["location"]]
  }else{
    parameter[[parameter_name]] <- samples[[ parameter_name ]]
  }

  return(parameter)
}
.JAGS_marglik_parameter_values          <- function(samples, prior, parameter_names){

  if(is.prior.point(prior)){
    return(rep(prior$parameters[["location"]], length(parameter_names)))
  }

  sample_names <- parameter_names
  if(prior[["distribution"]] == "invgamma"){
    sample_names <- paste0("inv_", parameter_names)
  }

  if(!all(sample_names %in% names(samples))){
    stop("'samples' does not contain all monitored formula prior parameters.", call. = FALSE)
  }

  values <- unname(unlist(samples[sample_names], use.names = FALSE))
  if(prior[["distribution"]] == "invgamma"){
    values <- values^-1
  }

  return(values)
}
.JAGS_marglik_parameters.vector         <- function(samples, prior, parameter_name){

  .check_prior(prior)
  if(!is.prior.vector(prior))
    stop("improper prior provided")
  check_char(parameter_name, "parameter_name")


  parameter <- list()
  if(prior$parameters[["K"]] == 1){
    parameter_monitor_name <- parameter_name
  }else{
    parameter_monitor_name <- paste0(parameter_name, "[", 1:prior$parameters[["K"]], "]")
  }

  if(prior[["distribution"]] == "mpoint"){
    parameter[[parameter_name]] <- rep(prior$parameters[["location"]], length(parameter_monitor_name))
  }else{
    parameter[[parameter_name]] <- samples[ parameter_monitor_name ]
  }

  return(parameter)
}
.JAGS_marglik_parameters.factor         <- function(samples, prior, parameter_name){

  .check_prior(prior)
  if(!is.prior.factor(prior))
    stop("improper prior provided")
  check_char(parameter_name, "parameter_name")


  if(is.prior.treatment(prior) | is.prior.independent(prior)){

    parameter <- list()
    if(.get_prior_factor_levels(prior) == 1){
      parameter_names <- parameter_name
    }else{
      parameter_names <- paste0(parameter_name, "[", 1:.get_prior_factor_levels(prior), "]")
    }
    parameter[[parameter_name]] <- .JAGS_marglik_parameter_values(samples, prior, parameter_names)

  }else if(is.prior.orthonormal(prior) | is.prior.meandif(prior)){

    prior$parameters[["K"]] <- .get_prior_factor_levels(prior)
    parameter <- .JAGS_marglik_parameters.vector(samples, prior, parameter_name)

  }


  return(parameter)
}
.JAGS_marglik_parameters.PP             <- function(samples, prior){

  .check_prior(prior)
  if(!is.prior.PET(prior) & !is.prior.PEESE(prior))
    stop("improper prior provided")

  if(is.prior.PET(prior)){
    parameter <- .JAGS_marglik_parameters.simple(samples, prior, "PET")
  }else if(is.prior.PEESE(prior)){
    parameter <- .JAGS_marglik_parameters.simple(samples, prior, "PEESE")
  }

  return(parameter)
}
.JAGS_marglik_parameters.weightfunction <- function(samples, prior){

  .check_prior(prior)
  if(!is.prior.weightfunction(prior))
    stop("improper prior provided")

  parameter <- list()
  J <- .weightfunction_n_bins(prior)

  if(prior$weights$type == "cumulative"){

    eta     <- samples[paste0("eta[", seq_len(J), "]")]
    std_eta <- eta / sum(eta)
    omega <- unname(rev(cumsum(rev(std_eta))))

  }else if(prior$weights$type == "independent"){

    omega <- rep(1, J)
    if(J > 1L){
      if(prior$weights$scale == "omega"){
        omega[2:J] <- samples[paste0("omega[", 2:J, "]")]
      }else if(prior$weights$scale == "log_omega"){
        omega[2:J] <- exp(samples[paste0("log_omega[", 2:J, "]")])
      }
    }
  }else if(prior$weights$type == "fixed"){

    omega <- unname(prior$weights$omega)

  }

  expansion <- .weightfunction_mapping_expansion(prior, force_one_sided = TRUE)
  parameter[["omega"]] <- unname(omega[expansion$index])

  return(parameter)
}
.JAGS_marglik_parameters.phacking <- function(samples, prior){

  .check_prior(prior)
  if(!is_prior_phacking(prior))
    stop("improper prior provided")

  alpha <- samples[["alpha"]]
  constants <- phack_backend_constants(prior$form, prior$source, prior$destination, target = prior$target)
  list(
    alpha     = alpha,
    pi_null   = alpha * constants$pi_null_per_alpha,
    beta_null = alpha * constants$beta_null_per_alpha
  )
}
.JAGS_marglik_parameters.bias <- function(samples, prior){

  .check_prior(prior)
  if(!is_prior_bias(prior))
    stop("improper prior provided")

  selection_backend_spec(prior)

  parameter <- list()
  if(!is.null(prior$selection)){
    parameter <- c(parameter, .JAGS_marglik_parameters.weightfunction(samples, prior$selection))
  }
  if(!is.null(prior$phacking)){
    parameter <- c(parameter, .JAGS_marglik_parameters.phacking(samples, prior$phacking))
  }

  return(parameter)
}
# .JAGS_marglik_parameters.spike_and_slab <- function(samples, prior, parameter_name){
#
#   .check_prior(prior)
#   if(!is.prior.spike_and_slab(prior))
#     stop("improper prior provided")
#   check_char(parameter_name, "parameter_name")
#
#   parameter <- list()
#   parameter[paste0(parameter_name, "_variable")]  <- .JAGS_marglik_parameters.simple(samples, prior[["variable"]],  paste0(parameter_name, "_variable"))
#   if(!is.prior.point(prior[[parameter_name]][["inclusion"]])){
#     parameter[paste0(parameter_name, "_inclusion")] <- .JAGS_marglik_parameters.simple(samples, prior[["inclusion"]], paste0(parameter_name, "_inclusion"))
#   }
#
#   return(parameter)
# }

#' @rdname JAGS_marglik_parameters
JAGS_marglik_parameters_formula      <- function(samples, formula_list, formula_data_list, formula_prior_list, prior_list_parameters){

  # return empty list in case that no prior was specified
  if(length(formula_prior_list) == 0){
    return(list())
  }

  parameters <- list()

  for(parameter in names(formula_prior_list)){
    # check for log(intercept) attribute on the formula
    log_intercept <- if(!is.null(formula_list[[parameter]])) isTRUE(attr(formula_list[[parameter]], "log(intercept)")) else FALSE
    parameters[[parameter]] <- .JAGS_marglik_parameters_formula_get(samples, parameter, formula_data_list[[parameter]], formula_prior_list[[parameter]], prior_list_parameters, log_intercept)
  }

  return(parameters)
}

.JAGS_marglik_parameters_formula_get <- function(samples, parameter, formula_data_list, formula_prior_list, prior_list_parameters, log_intercept = FALSE){

  formula_terms            <- names(formula_prior_list)
  names(formula_data_list) <- sub(paste0("^", parameter, "_data_"), paste0(parameter, "_"), names(formula_data_list))

  # start with intercept
  if(sum(formula_terms == paste0(parameter, "_intercept")) == 1){

    # check for scaling factors
    if(!is.null(attr(formula_prior_list[[paste0(parameter, "_intercept")]], "multiply_by"))){
      if(is.numeric(attr(formula_prior_list[[paste0(parameter, "_intercept")]], "multiply_by"))){
        multiply_by <- attr(formula_prior_list[[paste0(parameter, "_intercept")]], "multiply_by")
      }else{
        multiply_by <- prior_list_parameters[[attr(formula_prior_list[[paste0(parameter, "_intercept")]], "multiply_by")]]
      }
    }else{
      multiply_by <- 1
    }

    intercept_prior <- formula_prior_list[[paste0(parameter, "_intercept")]]
    intercept_value <- .JAGS_marglik_parameter_values(samples, intercept_prior, paste0(parameter, "_intercept"))
    # apply log transformation if log(intercept) attribute is set
    if(log_intercept){
      intercept_value <- log(intercept_value)
    }
    output <- multiply_by * rep(intercept_value, formula_data_list[[paste0("N_", parameter)]])

  }else{
    output <- rep(0, formula_data_list[[paste0("N_", parameter)]])
  }

  # add the remaining terms
  remaining_terms <- formula_terms[formula_terms != paste0(parameter, "_intercept")]
  if(length(remaining_terms) > 0){
    for(term in remaining_terms){

      # check for scaling factors
      if(!is.null(attr(formula_prior_list[[term]], "multiply_by"))){
        if(is.numeric(attr(formula_prior_list[[term]], "multiply_by"))){
          multiply_by <- attr(formula_prior_list[[term]], "multiply_by")
        }else{
          multiply_by <- prior_list_parameters[[attr(formula_prior_list[[term]], "multiply_by")]]
        }
      }else{
        multiply_by <- 1
      }


      if(is.prior.point(formula_prior_list[[term]]) && !is.prior.factor(formula_prior_list[[term]])){

        output <- output + multiply_by * formula_prior_list[[term]][["parameters"]][["location"]] * formula_data_list[[term]]

      }else if(is.prior.point(formula_prior_list[[term]]) && is.prior.factor(formula_prior_list[[term]])){

        if(.get_prior_factor_levels(formula_prior_list[[term]]) == 1){
          output <- output + multiply_by * formula_prior_list[[term]][["parameters"]][["location"]] * formula_data_list[[term]]
        }else{
          output <- output + multiply_by * formula_data_list[[term]] %*% rep(formula_prior_list[[term]][["parameters"]][["location"]], .get_prior_factor_levels(formula_prior_list[[term]]))
        }

      }else if(is.prior.factor(formula_prior_list[[term]])){

        if(.get_prior_factor_levels(formula_prior_list[[term]]) == 1){
          term_value <- .JAGS_marglik_parameter_values(samples, formula_prior_list[[term]], term)
          output     <- output + multiply_by * term_value * formula_data_list[[term]]
        }else{
          term_names  <- paste0(term,"[", 1:.get_prior_factor_levels(formula_prior_list[[term]]), "]")
          term_values <- .JAGS_marglik_parameter_values(samples, formula_prior_list[[term]], term_names)
          output      <- output + multiply_by * formula_data_list[[term]] %*% term_values
        }


      }else if(is.prior.simple(formula_prior_list[[term]])){

        term_value <- .JAGS_marglik_parameter_values(samples, formula_prior_list[[term]], term)
        output     <- output + multiply_by * term_value * formula_data_list[[term]]

      }

    }
  }


  return(as.vector(output))
}

#' @title Prints a prior object
#'
#' @param x a prior
#' @param short_name whether prior distribution names should be
#' shorted
#' @param parameter_names whether parameter names should be
#' printed
#' @param plot to return \link[base]{bquote} formatted
#'   prior name for plotting.
#' @param digits_estimates number of decimals to be displayed
#'   for printed parameters.
#' @param silent to silently return the print message.
#' @param ... additional arguments
#'
#' @examples
#' # create some prior distributions
#' p0 <- prior(distribution = "point",  parameters = list(location = 0))
#' p1 <- prior(distribution = "normal", parameters = list(mean = 0, sd = 1))
#'
#' # print them
#' p0
#' p1
#'
#' # use short names
#' print(p1, short_name = TRUE)
#'
#' # print parameter names
#' print(p1, parameter_names = TRUE)
#'
#' # generate bquote plotting syntax
#' plot(0, main = print(p1, plot = TRUE))
#'
#' @return \code{print.prior} invisibly returns the print statement.
#'
#' @seealso [prior()]
#' @rdname print.prior
#' @export
print.prior <- function(x, short_name = FALSE, parameter_names = FALSE, plot = FALSE,
                        digits_estimates = 2, silent = FALSE, ...){

  .check_prior(x, "x", allow_expressions = TRUE)
  check_bool(short_name, "short_name")
  check_bool(parameter_names, "parameter_names")
  check_int(digits_estimates, "digits_estimates", lower = 0)
  check_bool(plot, "plot")
  check_bool(silent, "silent")

  if(plot){
    silent <- TRUE
  }

  dots <- list(...)
  if(is.null(dots[["inline"]])){
    inline <- FALSE
  }else{
    inline <- dots[["inline"]]
    check_bool(inline, "inline")
  }

  if(.is_prior_expression(x)){
    x <- .prior_expression_to_character(x)
  }

  if(is.prior.none(x)){
    output <- .print.prior.none(x, short_name, parameter_names, plot, digits_estimates, silent)
  }else if(is.prior.simple(x) || is.prior.vector(x)){
    output <- .print.prior.simple(x, short_name, parameter_names, plot, digits_estimates, silent)
  }else if(is.prior.weightfunction(x)){
    output <- .print.prior.weightfunction(x, short_name, parameter_names, plot, digits_estimates, silent)
  }else if(is_prior_phacking(x)){
    output <- .print.prior.phacking(x, short_name, parameter_names, plot, digits_estimates, silent)
  }else if(is_prior_bias(x)){
    output <- .print.prior.bias(x, short_name, parameter_names, plot, digits_estimates, silent)
  }else if(is.prior.spike_and_slab(x)){
    output <- .print.prior.spike_and_slab(x, short_name, parameter_names, plot, digits_estimates, silent)
  }else if(is.prior.mixture(x)){
    output <- .print.prior.mixture(x, short_name, parameter_names, plot, digits_estimates, silent, inline)
  }


  if(!silent){
    cat(output)
  }
  return(invisible(output))
}

.print.prior.simple         <- function(x, short_name, parameter_names, plot, digits_estimates, silent){

  # check whether the range was truncated (before the object is modified)
  if(is.prior.vector(x)){
    needs_truncation <- FALSE
  }else{
    needs_truncation <- !.is_prior_default_range(x)
  }

  # deal with exceptions - Cauchy is passed as T
  if(x[["distribution"]] == "t" && x[["parameters"]][["df"]] == 1){
      x[["distribution"]] <- "Cauchy"
      x[["parameters"]]   <- list(
        location = x[["parameters"]][["location"]],
        scale    = x[["parameters"]][["scale"]])
  }else if(x[["distribution"]] == "mt" && x[["parameters"]][["df"]] == 1){
    x[["distribution"]] <- "mCauchy"
    x[["parameters"]]   <- list(
      location = x[["parameters"]][["location"]],
      scale    = x[["parameters"]][["scale"]])
  }

  ### prepare prior name
  if(short_name){
    out_name <- switch(
      x[["distribution"]],
      "normal"       = "N",
      "lognormal"    = "Ln",
      "t"            = "T",
      "Cauchy"       = "C",
      "gamma"        = "G",
      "invgamma"     = "Ig",
      "point"        = "S",
      "beta"         = "B",
      "bernoulli"    = "Br",
      "exp"          = "E",
      "uniform"      = "U",
      "mnormal"      = "mN",
      "mt"           = "mT",
      "mCauchy"      = "mC",
      "mpoint"       = "mS"
    )
  }else{
    out_name <- switch(
      x[["distribution"]],
      "normal"       = "Normal",
      "lognormal"    = "Lognormal",
      "t"            = "Student-t",
      "Cauchy"       = "Cauchy",
      "gamma"        = "Gamma",
      "invgamma"     = "InvGamma",
      "point"        = "Spike",
      "beta"         = "Beta",
      "bernoulli"    = "Bernoulli",
      "exp"          = "Exponential",
      "uniform"      = "Uniform",
      "mnormal"      = "mNormal",
      "mt"           = "mStudent-t",
      "mCauchy"      = "mCauchy",
      "mpoint"       = "mSpike"
    )
  }

  # add prefix
  if(is.prior.PET(x)){
    out_prefix <- "PET ~ "
  }else if(is.prior.PEESE(x)){
    out_prefix <- "PEESE ~ "
  }else if(is.prior.treatment(x)){
    out_prefix <- "treatment contrast: "
  }else if(is.prior.orthonormal(x)){
    out_prefix <- "orthonormal contrast: "
  }else if(is.prior.meandif(x)){
    out_prefix <- "mean difference contrast: "
  }else if(is.prior.independent(x)){
    out_prefix <- "independent contrast: "
  }else{
    out_prefix <- NULL
  }

  # remove the dimensions parameter from multivariate prior distributions
  if(is.prior.vector(x)){
    x[["parameters"]] <- x[["parameters"]][names(x[["parameters"]]) != "K"]
  }

  ### prepare prior parameters
  # round the parameters and truncation for printing
  for(i in seq_along(x[["parameters"]])){
    if(!is.character(x[["parameters"]][[i]])){
      x[["parameters"]][[i]] <- round(x[["parameters"]][[i]], digits_estimates)
    }
  }
  for(i in seq_along(x[["truncation"]])){
    x[["truncation"]][[i]] <- round(x[["truncation"]][[i]], digits_estimates)
  }

  if(parameter_names){
    out_parameters <- paste(sapply(seq_along(x[["parameters"]]), function(i){
      paste0(names(x[["parameters"]])[i], " = ", x[["parameters"]][[i]])
    }), collapse = ", ")
  }else{
    out_parameters <- paste(sapply(seq_along(x[["parameters"]]), function(i){
      x[["parameters"]][[i]]
    }), collapse = ", ")
  }


  ### prepare prior truncation
  if(needs_truncation){
    out_truncation <- paste(x[["truncation"]], collapse = ", ")
  }else{
    out_truncation <- NULL
  }


  ### combine the results together
  if(!plot){
    output <- paste0(
      if(!is.null(out_prefix)) out_prefix,
      out_name,
      if(!is.null(out_parameters)) paste0("(", out_parameters, ")"),
      if(!is.null(out_truncation)) paste0("[", out_truncation, "]"))
  }else{
    output <- bquote(
      .(if(!is.null(out_prefix)){ out_prefix }else{ bquote() } )*
        italic(.(out_name))*
        .(if(!is.null(out_parameters)){ paste0("(", out_parameters, ")") }else{ bquote() } )*
        .(if(!is.null(out_truncation)){
          if(is.infinite(x[["truncation"]][["lower"]])){
            bquote(""["["*-infinity*", "*.(x[["truncation"]][["upper"]])*"]"])
          }else if(is.infinite(x[["truncation"]][["upper"]])){
            bquote(""["["*.(x[["truncation"]][["lower"]])*", "*infinity*"]"])
          }else{
            bquote(""[.(paste0("[", out_truncation, "]"))])
          }
        }else{ bquote() })
    )
  }

  return(output)
}
.print.prior.weightfunction <- function(x, short_name, parameter_names, plot, digits_estimates, silent){

  ### prepare prior name
  # add prefix
  if(plot){
    out_prefix <- bquote(omega)
  }else{
    out_prefix <- "omega"
  }

  # type of steps
  if(short_name){
    steps_name <- switch(
      x[["side"]],
      "two-sided" = "2s: ",
      "one-sided" = "1s: "
    )
  }else{
    steps_name <- switch(
      x[["side"]],
      "two-sided" = "two-sided: ",
      "one-sided" = "one-sided: "
    )
  }

  # add steps
  out_steps  <- paste(trimws(x$steps, which = "left", whitespace = "0"), collapse = ", ")

  # distribution
  if(x$weights$type == "cumulative"){

    out_parameters <- paste(round(x$weights[["alpha"]], digits_estimates), collapse = ", ")
    if(parameter_names){
      out_parameters <- paste0("alpha = ", out_parameters)
    }
    if(short_name){
      out_distribution <- paste0("CumD")
    }else{
      out_distribution <- paste0("CumDirichlet")
    }

    if(!plot){
      output <- paste0(out_prefix, "[", steps_name, out_steps, "]", " ~ ", out_distribution, "(", out_parameters, ")")
    }else{
      output <- bquote(italic(.(out_prefix))[.(steps_name)*.(out_steps)]~"~"~italic(.(out_distribution))*(.(out_parameters)))
    }

  }else if(x$weights$type == "fixed"){

    out_parameters <- paste0(round(x$weights[["omega"]], digits_estimates), collapse = ", ")

    if(!plot){
      output <- paste0(out_prefix, "[", steps_name, out_steps, "]", " = ", "(", out_parameters, ")")
    }else{
      output <- bquote(italic(.(out_prefix))[.(steps_name)*.(out_steps)]~"="~(.(out_parameters)))
    }

  }else if(x$weights$type == "independent"){

    out_distribution <- if(x$weights$scale == "omega") "Independent" else "IndependentLog"
    out_parameters <- print(x$weights$prior, short_name = short_name, parameter_names = parameter_names,
                            plot = FALSE, digits_estimates = digits_estimates, silent = TRUE)

    if(!plot){
      output <- paste0(out_prefix, "[", steps_name, out_steps, "]", " ~ ", out_distribution, "(", out_parameters, ")")
    }else{
      output <- bquote(italic(.(out_prefix))[.(steps_name)*.(out_steps)]~"~"~italic(.(out_distribution))*(.(out_parameters)))
    }
  }

  return(output)
}
.print.prior.none           <- function(x, short_name, parameter_names, plot, digits_estimates, silent){

  ### prepare prior name
  out_name <- "None"


  ### combine the results together
  if(!plot){
    output <- out_name
  }else{
    output <- bquote(italic(.(out_name)))
  }

  return(output)
}
.print.prior.phacking       <- function(x, short_name, parameter_names, plot, digits_estimates, silent){

  report_parameter <- .phacking_report_parameter(x)
  out_prefix <- if(plot) as.name(report_parameter) else report_parameter
  form_name <- if(short_name) substr(x$form, 1, 1) else x$form
  alpha_prior <- print(x$alpha, short_name = short_name, parameter_names = parameter_names,
                       plot = FALSE, digits_estimates = digits_estimates, silent = TRUE)
  format_p <- function(value) format(value, scientific = FALSE, digits = max(digits_estimates, 2), trim = TRUE)
  out_parameters <- paste0(
    "target = ", format_p(x$target), ", ",
    "source = ", format_p(x$source), ", ",
    "destination = ", format_p(x$destination), ", ",
    "form = ", form_name
  )

  if(!plot){
    if(report_parameter == "alpha"){
      output <- paste0(out_prefix, "[phacking: ", out_parameters, "] ~ ", alpha_prior)
    }else{
      output <- paste0(out_prefix, "[phacking: ", out_parameters, "] derived from alpha ~ ", alpha_prior)
    }
  }else{
    if(report_parameter == "alpha"){
      output <- bquote(italic(.(out_prefix))[.(out_parameters)]~"~"~.(alpha_prior))
    }else{
      output <- bquote(italic(.(out_prefix))[.(out_parameters)]~" derived from "~alpha~"~"~.(alpha_prior))
    }
  }

  return(output)
}
.print.prior.bias           <- function(x, short_name, parameter_names, plot, digits_estimates, silent){

  parts <- character()
  if(!is.null(x$selection)){
    parts <- c(parts, print(x$selection, short_name = short_name, parameter_names = parameter_names,
                            plot = FALSE, digits_estimates = digits_estimates, silent = TRUE))
  }
  if(!is.null(x$phacking)){
    parts <- c(parts, print(x$phacking, short_name = short_name, parameter_names = parameter_names,
                            plot = FALSE, digits_estimates = digits_estimates, silent = TRUE))
  }

  if(!plot){
    output <- paste(parts, collapse = " * ")
  }else{
    output <- bquote(.(paste(parts, collapse = " * ")))
  }

  return(output)
}
.print.prior.spike_and_slab <- function(x, short_name, parameter_names, plot, digits_estimates, silent){

  variable  <- print(.get_spike_and_slab_variable(x),  short_name, parameter_names, plot, digits_estimates, silent = TRUE)
  inclusion <- print(.get_spike_and_slab_inclusion(x), short_name, parameter_names, plot, digits_estimates, silent = TRUE)

  ### combine the results together
  if(!plot){
    output <- paste0(variable, " * ", inclusion)
  }else{
    output <- bquote(.(variable) ~ "*" ~ .(inclusion))
  }

  return(output)
}
.print.prior.mixture        <- function(x, short_name, parameter_names, plot, digits_estimates, silent, inline){

  prior_names <- lapply(x, function(p){
    print(p,  short_name, parameter_names, plot, digits_estimates, silent = TRUE)
  })

  prior_weights <- attr(x, "prior_weights")
  prior_weights <- paste0("(", round(prior_weights, digits_estimates), "/", round(sum(prior_weights), digits_estimates), ")")

  prior_components <- attr(x, "components")
  if(all(prior_components %in% c("null", "alternative"))){
    prior_order      <- order(prior_components)
    prior_names      <- prior_names[prior_order]
    prior_weights    <- prior_weights[prior_order]
    prior_components <- prior_components[prior_order]
  }

  if(!plot){

    output <- NULL

    # inline printing for summary tables
    if(inline){
      output <- paste0(paste0(prior_weights, " * ", prior_names), collapse = " + ")
    }else{
      for(component in unique(prior_components)){
        output <- paste0(output, component, ":\n")
        for(i in seq_along(prior_components)[prior_components == component]){
          output <- paste0(output, "  ", prior_weights[i], " * ", prior_names[[i]], "\n")
        }
      }
    }

  }else{
    output <- Map(function(weight, prior) bquote(.(as.name(weight))~"*"~.(prior)), prior_weights, prior_names)
    output <- Reduce(function(x, y) bquote(.(x)~+~.(y)), output)
  }

  return(output)
}

#' @title Create BayesTools ensemble summary tables
#'
#' @description Creates estimate summaries based on posterior
#' distributions created by [mix_posteriors], inference summaries
#' based on inference created by [ensemble_inference], or ensemble
#' summary/diagnostics based on a list of [models_inference] models
#' (or [marginal_inference] in case of [marginal_estimates_table]).
#'
#' @param samples posterior samples created by [mix_posteriors]
#' @param inference model inference created by [ensemble_inference]
#' @param parameters character vector of parameters (or a
#' named list with of character vectors for summary and
#' diagnostics tables) specifying the parameters
#' (and their grouping) for the summary table
#' @param models list of [models_inference] model objects,
#' each of which containing a list of \code{priors}
#' and \code{inference} object, The \code{inference} must be a
#' named list with information about the model: model number
#' \code{m_number}, marginal likelihood \code{marglik}, prior and
#' posterior probability \code{prior_prob} and \code{post_prob},
#' inclusion Bayes factor \code{inclusion_BF}, and fit summary
#' generated by [runjags_estimates_table] for the diagnostics
#' table
#' @param probs quantiles for parameter estimates
#' @param logBF whether the Bayes factor should be on log scale
#' @param BF01 whether the Bayes factor should be inverted
#' @param short_name whether the prior distribution names should be
#' shortened. Defaults to \code{FALSE}.
#' @param remove_spike_0 whether prior distributions equal to spike
#' at 0 should be removed from the \code{prior_list}
#' @param transform_factors whether factors with orthonormal/meandif
#' prior distribution should be transformed to differences from the
#' grand mean
#' @param transform_orthonormal (to be depreciated) whether factors
#' with orthonormal prior distributions should be transformed to
#' differences from the grand mean
#' @param transform_scaled whether coefficients from standardized
#' continuous predictors should be transformed back to the original
#' scale. Defaults to \code{FALSE}.
#' @param formula_scale named list containing standardization information
#' (mean and sd) for each standardized predictor. Required when
#' \code{transform_scaled = TRUE} for ensemble/marginal tables. For
#' \code{runjags_estimates_table}, this is automatically extracted from
#' the fit object's \code{formula_scale} attribute.
#' @param title title to be added to the table
#' @param footnotes footnotes to be added to the table
#' @param warnings warnings to be added to the table
#' @param formula_prefix whether the parameter prefix from formula should
#' be printed. Defaults to \code{TRUE}.
#'
#'
#' @return \code{ensemble_estimates_table} returns a table with the
#' model-averaged estimates, \code{ensemble_inference_table} returns
#' a table with the prior and posterior probabilities and inclusion
#' Bayes factors, \code{ensemble_summary_table} returns a table with
#' overview of the models included in the ensemble, and
#' \code{ensemble_diagnostics_table} returns an overview of the MCMC
#' diagnostics for the models included in the ensemble. All of the
#' tables are objects of class 'BayesTools_table'.
#'
#' @export ensemble_estimates_table
#' @export ensemble_inference_table
#' @export ensemble_summary_table
#' @export ensemble_diagnostics_table
#' @export ensemble_estimates_empty_table
#' @export ensemble_inference_empty_table
#' @export ensemble_summary_empty_table
#' @export ensemble_diagnostics_empty_table
#' @export marginal_estimates_table
#' @name BayesTools_ensemble_tables
#'
#' @seealso [ensemble_inference] [mix_posteriors] [BayesTools_model_tables]
NULL

#' @rdname BayesTools_ensemble_tables
ensemble_estimates_table <- function(samples, parameters, probs = c(0.025, 0.975), title = NULL, footnotes = NULL, warnings = NULL, transform_factors = FALSE, transform_orthonormal = FALSE, formula_prefix = TRUE, transform_scaled = FALSE, formula_scale = NULL){

  # check input
  check_char(parameters, "parameters", check_length = 0)
  check_list(samples, "samples", check_names = parameters, all_objects = TRUE, allow_other = TRUE)
  check_real(probs, "probs", lower = 0, upper = 1, check_length = 0, allow_NULL = TRUE)
  check_char(title, "title", allow_NULL = TRUE)
  check_char(footnotes, "footnotes", check_length = 0, allow_NULL = TRUE)
  check_char(warnings, "warnings", check_length = 0, allow_NULL = TRUE)
  check_bool(transform_factors, "transform_factors")
  check_bool(transform_orthonormal, "transform_orthonormal")
  check_bool(formula_prefix, "formula_prefix")
  check_bool(transform_scaled, "transform_scaled")
  check_list(formula_scale, "formula_scale", allow_NULL = TRUE)

  # depreciate
  transform_factors <- .depreciate.transform_orthonormal(transform_orthonormal, transform_factors)


  # transform scaled coefficients back to original scale
  if(transform_scaled && !is.null(formula_scale) && length(formula_scale) > 0){
    samples <- .transform_scale_samples_list(samples, formula_scale)
  }

  # transform meandif/orthonormal posterior
  if(transform_factors){
    samples <- transform_factor_samples(samples)
  }


  # extract values
  estimates_table <- NULL
  for(parameter in parameters){

    if(is.matrix(samples[[parameter]])){

      par_summary <- cbind(
        "Mean"   = apply(samples[[parameter]], 2, mean),
        "Median" = apply(samples[[parameter]], 2, stats::median)
      )
      for(i in seq_along(probs)){
        par_summary <- cbind(par_summary, apply(samples[[parameter]], 2, stats::quantile, probs = probs[i]))
        colnames(par_summary)[ncol(par_summary)] <- probs[i]
      }

      if(inherits(samples[[parameter]], "mixed_posteriors.formula")){
        parameter_name <- format_parameter_names(colnames(samples[[parameter]]), formula_parameters = attr(samples[[parameter]], "formula_parameter"), formula_prefix = formula_prefix, formula_scale = formula_scale)
      }else{
        parameter_name <- colnames(samples[[parameter]])
      }

      rownames(par_summary) <- parameter_name
      estimates_table       <- rbind(estimates_table, par_summary)

    }else if(is.numeric(samples[[parameter]])){

      par_summary <- c(
        "Mean"   = mean(samples[[parameter]]),
        "Median" = stats::median((samples[[parameter]]))
      )
      for(i in seq_along(probs)){
        par_summary <- c(par_summary, stats::quantile(samples[[parameter]], probs = probs[i]))
        names(par_summary)[length(par_summary)] <- probs[i]
      }
      estimates_table <- rbind(estimates_table, par_summary)

      if(inherits(samples[[parameter]], "mixed_posteriors.formula")){
        parameter_name <- gsub(
          paste0(attr(samples[[parameter]], "formula_parameter"), "_"),
          if(formula_prefix) paste0("(", attr(samples[[parameter]], "formula_parameter"), ") ") else "",
          parameter)
        parameter_name <- gsub("__xXx__", ":", parameter_name)
      }else{
        parameter_name <- parameter
      }

      rownames(estimates_table)[nrow(estimates_table)] <- parameter_name


    }else{
      stop("Uknown parameter type.")
    }
  }

  # prepare output
  estimates_table                    <- data.frame(estimates_table)
  colnames(estimates_table)          <- gsub("X", "", colnames(estimates_table))
  class(estimates_table)             <- c("BayesTools_table", "BayesTools_ensemble_summary", class(estimates_table))
  attr(estimates_table, "type")      <- rep("estimate", ncol(estimates_table))
  attr(estimates_table, "rownames")  <- TRUE
  attr(estimates_table, "title")     <- title
  attr(estimates_table, "footnotes") <- footnotes
  attr(estimates_table, "warnings")  <- warnings

  return(estimates_table)
}

#' @rdname BayesTools_ensemble_tables
ensemble_inference_table <- function(inference, parameters, logBF = FALSE, BF01 = FALSE, title = NULL, footnotes = NULL, warnings = NULL){

  # check input
  check_char(parameters, "parameters", check_length = 0)
  check_list(inference, "inference", check_names = parameters, all_objects = TRUE, allow_other = TRUE)
  check_bool(logBF, "logBF")
  check_bool(BF01,  "BF01")
  check_char(title, "title", allow_NULL = TRUE)
  check_char(footnotes, "footnotes", check_length = 0, allow_NULL = TRUE)
  check_char(warnings, "warnings", check_length = 0, allow_NULL = TRUE)
  if(attr(inference,"conditional"))
    stop("The inference object cannot be 'conditional'.")

  # extract values
  inference_table <- NULL
  n_models        <- NULL
  for(parameter in parameters){
    inference_table <- rbind(inference_table, c(
      "models"       = sum(!attr(inference[[parameter]], "is_null")),
      "prior_prob"   = sum(inference[[parameter]][["prior_probs"]][!attr(inference[[parameter]], "is_null")]),
      "post_prob"    = sum(inference[[parameter]][["post_probs"]][!attr(inference[[parameter]], "is_null")] ),
      "inclusion_BF" = inference[[parameter]][["BF"]]
    ))
    rownames(inference_table)[nrow(inference_table)] <- attr(inference[[parameter]], "parameter_name")
    n_models <- c(n_models, length(attr(inference[[parameter]], "is_null")))
  }
  inference_table <- data.frame(inference_table)

  # format BF
  inference_table[,"inclusion_BF"]  <- format_BF(inference_table[,"inclusion_BF"], logBF = logBF, BF01 = BF01, inclusion = TRUE)

  # prepare output
  class(inference_table)             <- c("BayesTools_table", "BayesTools_ensemble_summary", class(inference_table))
  attr(inference_table, "type")      <- c("n_models", "prior_prob", "post_prob", "inclusion_BF")
  attr(inference_table, "rownames")  <- TRUE
  attr(inference_table, "n_models")  <- n_models
  attr(inference_table, "title")     <- title
  attr(inference_table, "footnotes") <- footnotes
  attr(inference_table, "warnings")  <- warnings

  return(inference_table)
}

#' @rdname BayesTools_ensemble_tables
ensemble_summary_table <- function(models, parameters, logBF = FALSE, BF01 = FALSE, title = NULL, footnotes = NULL, warnings = NULL,
                                   remove_spike_0 = TRUE, short_name = FALSE){

  # check input
  check_list(models, "models")
  for(i in seq_along(models)){
    model <- models[[i]]
    check_list(model, "model", check_names = "inference", allow_other = TRUE, all_objects = TRUE)
    prior_list <- attr(model[["fit"]], "prior_list")
    check_list(prior_list, "model:priors")
    if(!all(sapply(prior_list, is.prior)))
      stop("'model:priors' must be a list of priors.")
    model_inference <- model[["inference"]]
    check_list(model_inference, "model:inference", check_names = c("m_number", "marglik", "prior_prob", "post_prob", "inclusion_BF"), allow_other = TRUE, all_objects = TRUE)
    check_int(model_inference[["m_number"]],      "model_inference:model_number")
    check_real(model_inference[["marglik"]],      "model_inference:marglik")
    check_real(model_inference[["prior_prob"]],   "model_inference:prior_prob",   lower = 0, upper = 1)
    check_real(model_inference[["post_prob"]],    "model_inference:post_prob",   lower = 0, upper = 1)
    check_real(model_inference[["inclusion_BF"]], "model_inference:inclusion_BF", lower = 0)
  }
  if(is.list(parameters)){
    check_list(parameters, "parameters")
    check_char(names(parameters), "names(parameters)", check_length = FALSE, allow_NULL = length(parameters) == 0)
    sapply(parameters, check_char, name = "parameters", check_length = FALSE)
  }else{
    check_char(parameters, "parameters", check_length = FALSE)
    if(is.null(names(parameters))){
      names(parameters) <- parameters
    }
  }
  check_bool(logBF, "logBF")
  check_bool(BF01,  "BF01")
  check_char(title, "title", allow_NULL = TRUE)
  check_char(footnotes, "footnotes", check_length = 0, allow_NULL = TRUE)
  check_char(warnings, "warnings", check_length = 0, allow_NULL = TRUE)
  check_bool(short_name, "short_name")


  # create the output
  ensemble_table <- .ensemble_table_foundation(models, parameters, remove_spike_0, short_name)

  ensemble_table <- cbind(
    ensemble_table,
    "prior_prob"   = sapply(models, function(model)model[["inference"]][["prior_prob"]]),
    "marglik"      = sapply(models, function(model)model[["inference"]][["marglik"]]),
    "post_prob"    = sapply(models, function(model)model[["inference"]][["post_prob"]]),
    "inclusion_BF" = sapply(models, function(model)model[["inference"]][["inclusion_BF"]])
  )

  ensemble_table[,"inclusion_BF"]  <- format_BF(ensemble_table[,"inclusion_BF"], logBF = logBF, BF01 = BF01, inclusion = TRUE)

  # prepare output
  class(ensemble_table)             <- c("BayesTools_table", "BayesTools_ensemble_summary", class(ensemble_table))
  attr(ensemble_table, "type")      <- c("integer", rep("prior", length(parameters)), "prior_prob", "marglik", "post_prob", "inclusion_BF")
  attr(ensemble_table, "rownames")  <- FALSE
  attr(ensemble_table, "title")     <- title
  attr(ensemble_table, "footnotes") <- footnotes
  attr(ensemble_table, "warnings")  <- warnings

  return(ensemble_table)
}

#' @rdname BayesTools_ensemble_tables
ensemble_diagnostics_table <- function(models, parameters, title = NULL, footnotes = NULL, warnings = NULL, remove_spike_0 = TRUE, short_name = FALSE){

  # check input
  check_list(models, "models")
  for(i in seq_along(models)){
    model <- models[[i]]
    check_list(model, "model", check_names = c("fit_summary", "inference"), allow_other = TRUE, all_objects = TRUE)
    prior_list <- attr(model[["fit"]], "prior_list")
    check_list(prior_list, "model:priors")
    if(!all(sapply(prior_list, is.prior)))
      stop("'model:priors' must be a list of priors.")
    if(!is.null(model[["fit_summary"]]) && !(inherits(model[["fit_summary"]], "BayesTools_runjags_summary") || inherits(model[["fit_summary"]], "BayesTools_stan_summary")))
      stop("'fit_summary' must be a runjags or rstan summary generated by 'JAGS_estimates_table()' / 'stan_estimates_table()'.")
    model_inference <- model[["inference"]]
    check_list(model_inference, "model:inference", check_names = "m_number", allow_other = TRUE, all_objects = TRUE)
    check_int(model_inference[["m_number"]], "model_inference:model_number")
  }
  if(is.list(parameters)){
    check_list(parameters, "parameters")
    check_char(names(parameters), "names(parameters)", check_length = FALSE, allow_NULL = length(parameters) == 0)
    sapply(parameters, check_char, name = "parameters", check_length = FALSE)
  }else{
    check_char(parameters, "parameters", check_length = FALSE)
    if(is.null(names(parameters))){
      names(parameters) <- parameters
    }
  }
  check_char(title, "title", allow_NULL = TRUE)
  check_char(footnotes, "footnotes", check_length = 0, allow_NULL = TRUE)
  check_char(warnings, "warnings", check_length = 0, allow_NULL = TRUE)
  check_bool(short_name, "short_name")

  # create the output
  ensemble_table <- .ensemble_table_foundation(models, parameters, remove_spike_0, short_name)

  ensemble_table <- cbind(
    ensemble_table,
    "max_MCMC_error"    = sapply(models, function(model){
      .max_or_na(.fit_summary_column(model[["fit_summary"]], "MCMC_error"))
    }),
    "max_MCMC_SD_error" = sapply(models, function(model){
      .max_or_na(.fit_summary_column(model[["fit_summary"]], "MCMC_SD_error"))
    }),
    "min_ESS"     = sapply(models, function(model){
      .min_or_na(.fit_summary_column(model[["fit_summary"]], "ESS"))
    }),
    "max_R_hat"   = sapply(models, function(model){
      .max_or_na(.fit_summary_column(model[["fit_summary"]], "R_hat"))
    })
  )

  # prepare output
  class(ensemble_table)             <- c("BayesTools_table", "BayesTools_ensemble_summary", class(ensemble_table))
  attr(ensemble_table, "type")      <- c("integer", rep("prior", length(parameters)), "max_MCMC_error", "max_MCMC_SD_error", "min_ESS", "max_R_hat")
  attr(ensemble_table, "rownames")  <- FALSE
  attr(ensemble_table, "title")     <- title
  attr(ensemble_table, "footnotes") <- footnotes
  attr(ensemble_table, "warnings")  <- warnings

  return(ensemble_table)
}

#' @rdname BayesTools_ensemble_tables
ensemble_estimates_empty_table   <- function(probs = c(0.025, 0.975), title = NULL, footnotes = NULL, warnings = NULL){

  empty_table <- data.frame(matrix(nrow = 0, ncol = 2 + length(probs)))
  colnames(empty_table) <- c("Mean", "Median", probs)

  class(empty_table)             <- c("BayesTools_table", "BayesTools_ensemble_summary", class(empty_table))
  attr(empty_table, "type")      <- rep("estimate", ncol(empty_table))
  attr(empty_table, "rownames")  <- TRUE
  attr(empty_table, "title")     <- title
  attr(empty_table, "footnotes") <- footnotes
  attr(empty_table, "warnings")  <- warnings

  return(empty_table)
}

#' @rdname BayesTools_ensemble_tables
ensemble_inference_empty_table   <- function(title = NULL, footnotes = NULL, warnings = NULL){

  empty_table <- data.frame(matrix(nrow = 0, ncol = 4))
  colnames(empty_table) <- c("models", "prior_prob", "post_prob", "inclusion_BF")

  class(empty_table)             <- c("BayesTools_table", "BayesTools_ensemble_summary", class(empty_table))
  attr(empty_table, "type")      <- c("n_models", "prior_prob", "post_prob", "inclusion_BF")
  attr(empty_table, "rownames")  <- TRUE
  attr(empty_table, "title")     <- title
  attr(empty_table, "footnotes") <- footnotes
  attr(empty_table, "warnings")  <- warnings

  return(empty_table)
}

#' @rdname BayesTools_ensemble_tables
ensemble_summary_empty_table     <- function(title = NULL, footnotes = NULL, warnings = NULL){

  empty_table <- data.frame(matrix(nrow = 0, ncol = 5))
  colnames(empty_table) <- c("Model", "prior_prob", "marglik", "post_prob", "inclusion_BF")

  # prepare output
  class(empty_table)             <- c("BayesTools_table", "BayesTools_ensemble_summary", class(empty_table))
  attr(empty_table, "type")      <- c("integer", "prior_prob", "marglik", "post_prob", "inclusion_BF")
  attr(empty_table, "rownames")  <- FALSE
  attr(empty_table, "title")     <- title
  attr(empty_table, "footnotes") <- footnotes
  attr(empty_table, "warnings")  <- warnings

  return(empty_table)
}

#' @rdname BayesTools_ensemble_tables
ensemble_diagnostics_empty_table <- function(title = NULL, footnotes = NULL, warnings = NULL){

  empty_table <- data.frame(matrix(nrow = 0, ncol = 5))
  colnames(empty_table) <- c("Model", "max_MCMC_error", "max_MCMC_SD_error", "min_ESS", "max_R_hat")

  # prepare output
  class(empty_table)             <- c("BayesTools_table", "BayesTools_ensemble_summary", class(empty_table))
  attr(empty_table, "type")      <- c("integer", "max_MCMC_error", "max_MCMC_SD_error", "min_ESS", "max_R_hat")
  attr(empty_table, "rownames")  <- FALSE
  attr(empty_table, "title")     <- title
  attr(empty_table, "footnotes") <- footnotes
  attr(empty_table, "warnings")  <- warnings

  return(empty_table)
}

#' @rdname BayesTools_ensemble_tables
marginal_estimates_table <- function(samples, inference, parameters, probs = c(0.025, 0.5, 0.975), logBF = FALSE, BF01 = FALSE, title = NULL, footnotes = NULL, warnings = NULL, formula_prefix = TRUE, transform_scaled = FALSE, formula_scale = NULL){

  # check input
  check_char(parameters, "parameters", check_length = 0)
  check_list(samples, "samples", check_names = parameters, all_objects = TRUE, allow_other = TRUE)
  check_list(inference, "inference", check_names = parameters, all_objects = TRUE, allow_other = TRUE)
  check_real(probs, "probs", lower = 0, upper = 1, check_length = 0, allow_NULL = TRUE)
  check_bool(logBF, "logBF")
  check_bool(BF01,  "BF01")
  check_char(title, "title", allow_NULL = TRUE)
  check_char(footnotes, "footnotes", check_length = 0, allow_NULL = TRUE)
  check_char(warnings, "warnings", check_length = 0, allow_NULL = TRUE)
  check_bool(formula_prefix, "formula_prefix")
  check_bool(transform_scaled, "transform_scaled")
  check_list(formula_scale, "formula_scale", allow_NULL = TRUE)

  # transform scaled coefficients back to original scale
  if(transform_scaled && !is.null(formula_scale) && length(formula_scale) > 0){
    samples <- .transform_scale_samples_list(samples, formula_scale)
  }


  # extract values
  estimates_table <- NULL
  for(parameter in parameters){

    # extract the relevant information
    if(is.list(samples[[parameter]]) && length(samples[[parameter]]) > 1){
      temp_samples  <- .marginal_posterior_parameter_samples(samples, parameter)
      temp_BF       <- do.call(c, inference[[parameter]])
      temp_warnings <- do.call(c, lapply(names(inference[[parameter]]), function(lvl) {
        if(is.null(attr(inference[[parameter]][[lvl]], "warnings"))){
          return()
        }else{
          paste0("[", lvl, "]: ", attr(inference[[parameter]][[lvl]], "warnings"))
        }
      }))
    }else{
      temp_samples  <- .marginal_posterior_parameter_samples(samples, parameter)
      temp_BF       <- inference[[parameter]][[1]]
      if(is.null(attr(inference[[parameter]][[1]], "warnings"))){
        temp_warnings <- NULL
      }else{
        temp_warnings <- paste0(if(names(inference[[parameter]]) != "intercept") paste0("[", names(inference[[parameter]]), "]: ") else ": ", attr(inference[[parameter]][[1]], "warnings"))
      }
    }

    # add estimates
    par_summary <- cbind(
      "Mean" = vapply(temp_samples, mean, numeric(1)),
      "SD"   = vapply(temp_samples, stats::sd, numeric(1))
    )
    for(i in seq_along(probs)){
      par_summary <- cbind(par_summary, vapply(temp_samples, stats::quantile, numeric(1), probs = probs[i]))
      colnames(par_summary)[ncol(par_summary)] <- probs[i]
    }

    # add BF
    par_summary <- cbind(par_summary, "inclusion_BF" = temp_BF)


    # format parameter names
    if(inherits(samples[[parameter]], "marginal_posterior.formula")){
      if(length(names(samples[[parameter]])) == 1 && names(samples[[parameter]]) == "intercept"){
        parameter_name <- parameter
      }else{
        parameter_name <- paste0(parameter, "[", names(samples[[parameter]]), "]")
      }
      parameter_name <- format_parameter_names(parameter_name, formula_parameters = attr(samples[[parameter]], "formula_parameter"), formula_prefix = formula_prefix, formula_scale = formula_scale)
    }else{
      parameter_name <- paste0(parameter, "[", names(samples[[parameter]]), "]")
    }

    rownames(par_summary) <- parameter_name
    estimates_table       <- rbind(estimates_table, par_summary)

    # add warnings
    if(!is.null(temp_warnings)){
      warnings <- c(warnings, paste0(parameter, temp_warnings))
    }
  }

  # prepare output
  estimates_table                    <- data.frame(estimates_table)
  colnames(estimates_table)          <- gsub("X", "", colnames(estimates_table))
  estimates_table[,"inclusion_BF"]   <- format_BF(estimates_table[,"inclusion_BF"], logBF = logBF, BF01 = BF01, inclusion = TRUE)
  class(estimates_table)             <- c("BayesTools_table", "BayesTools_marginal_estimates", class(estimates_table))
  attr(estimates_table, "type")      <- c(rep("estimate", ncol(estimates_table) - 1), "inclusion_BF")
  attr(estimates_table, "rownames")  <- TRUE
  attr(estimates_table, "title")     <- title
  attr(estimates_table, "footnotes") <- footnotes
  attr(estimates_table, "warnings")  <- warnings

  return(estimates_table)
}



.ensemble_table_foundation <- function(models, parameters, remove_spike_0, short_name){

  model_rows <- list()
  for(i in seq_along(models)){

    model_row <- list()
    model_row[["Model"]] <- models[[i]][["inference"]][["m_number"]]

    temp_prior_list <- attr(models[[i]][["fit"]], "prior_list")

    for(p in seq_along(parameters)){
      if(is.list(parameters)){
        if(sum(names(temp_prior_list) %in% parameters[[p]]) == 1){
          temp_prior <- temp_prior_list[[parameters[[p]][parameters[[p]] %in% names(temp_prior_list)]]]
        }else if(sum(names(temp_prior_list) %in% parameters[[p]]) == 0){
          temp_prior <- prior_none()
        }else{
          stop("More than one prior matching the specified grouping.")
        }
      }else{
        if(any(names(temp_prior_list) == parameters[p])){
          temp_prior <- temp_prior_list[[parameters[p]]]
        }else{
          temp_prior <- prior_none()
        }
      }
      if(remove_spike_0 && is.prior.point(temp_prior) && temp_prior[["parameters"]][["location"]] == 0){
        model_row[[names(parameters)[p]]] <- ""
      }else if(is.prior.none(temp_prior)){
        model_row[[names(parameters)[p]]] <- ""
      }else{
        model_row[[names(parameters)[p]]] <- print(temp_prior, silent = TRUE, short_name = short_name)
      }
    }
    model_rows[[i]] <- model_row
  }

  summary_table <- data.frame(do.call(rbind, model_rows))
  for(i in 1:ncol(summary_table)){
    summary_table[,i] <- unlist(summary_table[,i])
  }

  colnames(summary_table) <- c("Model", names(parameters))

  parameters <- unlist(parameters) # deal with potential matching of multiple parameters withing a name
  for(p in seq_along(parameters)){

    parameter_name    <- parameters[p]
    formula_parameter <- unique(unlist(lapply(models, function(m) attr(attr(m[["fit"]], "prior_list")[[parameter_name]], "parameter"))))

    if(!is.null(unlist(formula_parameter))){
      parameter_name <- gsub(paste0(formula_parameter, "_"), paste0("(", formula_parameter, ") "), parameter_name)
      parameter_name <- gsub("__xXx__", ":", parameter_name)

      colnames(summary_table)[colnames(summary_table) == parameters[p]] <- parameter_name
    }
  }

  return(summary_table)
}
.fit_summary_column <- function(fit_summary, column){

  if(is.null(fit_summary) || !column %in% colnames(fit_summary)){
    return(NA_real_)
  }

  fit_summary[, column]
}
.max_or_na <- function(x){

  if(length(x) == 0 || all(is.na(x))){
    return(NA_real_)
  }

  max(x, na.rm = TRUE)
}
.min_or_na <- function(x){

  if(length(x) == 0 || all(is.na(x))){
    return(NA_real_)
  }

  min(x, na.rm = TRUE)
}


#' @title Create BayesTools model tables
#'
#' @description Creates model summary based on a model objects or
#' provides estimates table for a runjags fit.
#'
#' @param model model object containing a list of \code{priors}
#' and \code{inference} object, The \code{inference} must be a
#' named list with information about the model: model number
#' \code{m_number}, marginal likelihood \code{marglik}, prior and
#' posterior probability \code{prior_prob} and \code{post_prob},
#' and model inclusion Bayes factor \code{inclusion_BF}
#' @param fit runjags model fit
#' @param conditional summarizes estimates conditional on being included
#' in the model for spike and slab priors. Defaults to \code{FALSE}.
#' @param transformations named list of transformations to be applied
#' to specific parameters
#' @param model_description named list with additional description
#' to be added to the table
#' @param remove_inclusion whether estimates of the inclusion probabilities
#' should be excluded from the summary table. Defaults to \code{FALSE}.
#' @param remove_parameters parameters to be removed from the summary.
#' Can be \code{NULL} (default, no removal), a character vector of parameter
#' names to remove, or \code{TRUE} to remove all parameters that are not
#' part of any formula.
#' @param remove_formulas character vector of formula names whose parameters
#' should be removed from the summary. Defaults to \code{NULL}.
#' @param keep_parameters character vector of parameter names to keep.
#' All other parameters will be removed unless they belong to formulas
#' specified in \code{keep_formulas}. Defaults to \code{NULL}.
#' @param keep_formulas character vector of formula names whose parameters
#' should be kept. All other parameters will be removed unless they are
#' specified in \code{keep_parameters}. Defaults to \code{NULL}.
#' @param return_samples whether to return the transoformed and formated samples
#' instead of the table. Defaults to \code{FALSE}.
#' @param remove_diagnostics whether to exclude MCMC diagnostics (MCMC error,
#' ESS, R-hat) from the output table. Defaults to \code{FALSE}. Setting to
#' \code{TRUE} will exclude diagnostics columns regardless of the
#' \code{conditional} setting.
#' @param diagnostic_columns MCMC diagnostic columns to display in JAGS
#' estimates tables. Can be \code{"all"}, \code{"none"}, \code{TRUE},
#' \code{FALSE}, or a character vector containing any subset of
#' \code{"MCMC_error"}, \code{"MCMC_SD_error"}, \code{"ESS"}, and
#' \code{"R_hat"}. Defaults to the
#' \code{BayesTools.JAGS_estimates_diagnostic_columns} option, or all
#' diagnostics unless \code{remove_diagnostics = TRUE}.
#' @param BF_diagnostics whether to add MCMC diagnostics for Bayes factors
#' computed from model indicator frequencies. The Bayes factor error is
#' reported as a relative Monte Carlo standard error percentage. Defaults to
#' \code{FALSE}.
#' @param BF_diagnostic_columns MCMC diagnostic columns to display in JAGS
#' inclusion Bayes factor tables. Can be \code{"all"}, \code{"none"},
#' \code{TRUE}, \code{FALSE}, or a character vector containing any subset of
#' \code{"ESS"}, \code{"MCMC_error"}, and \code{"BF_error_percent"}.
#' Defaults to the \code{BayesTools.JAGS_BF_diagnostic_columns} option, or all
#' diagnostics when \code{BF_diagnostics = TRUE} and none otherwise.
#' @inheritParams BayesTools_ensemble_tables
#'
#'
#' @return \code{model_summary_table} returns a table with
#' overview of the fitted model, \code{runjags_estimates_table} returns
#' a table with MCMC estimates, and \code{runjags_estimates_empty_table}
#' returns an empty estimates table. All of the tables are objects of
#' class 'BayesTools_table'.
#'
#' @details For product-space JAGS inclusion Bayes factors, posterior
#' inclusion probabilities of exactly 0 or 1 cannot produce a finite
#' point estimate of the model odds ratio. In that case, inclusion BFs
#' marked with \code{"<"} or \code{">"} are finite-sample bounds: posterior
#' inclusion probabilities of 0 or 1 were replaced by \code{1/S} or
#' \code{(S - 1)/S}, where \code{S} is the number of posterior samples.
#' This is a reporting convention, not an unbiased finite Bayes-factor
#' estimate. If the prior inclusion probability is exactly 0 or 1, the
#' inclusion Bayes factor is undefined and reported as \code{NA}, because
#' the corresponding inclusion/exclusion comparison was not tested.
#'
#' @export JAGS_summary_table
#' @export JAGS_estimates_table
#' @export JAGS_inference_table
#' @export model_summary_table
#' @export runjags_estimates_table
#' @export runjags_inference_table
#' @export model_summary_empty_table
#' @export JAGS_inference_empty_table
#' @export JAGS_estimates_empty_table
#' @export runjags_estimates_empty_table
#' @export runjags_inference_empty_table
#' @export stan_estimates_table
#' @name BayesTools_model_tables
#'
#' @seealso [BayesTools_ensemble_tables]
NULL

#' @rdname BayesTools_model_tables
model_summary_table <- function(model, model_description = NULL, title = NULL, footnotes = NULL, warnings = NULL,
                                remove_spike_0 = TRUE, short_name = FALSE, formula_prefix = TRUE, remove_parameters = NULL){

  # check input
  check_list(model, "model", check_names = "inference", allow_other = TRUE, all_objects = TRUE)
  prior_list <- attr(model[["fit"]], "prior_list")
  check_list(prior_list, "model:priors")
  if(!all(sapply(prior_list, is.prior)))
    stop("'model:priors' must be a list of priors.")
  model_inference <- model[["inference"]]
  check_list(model_inference, "model:inference", check_names = c("m_number", "marglik", "prior_prob", "post_prob", "inclusion_BF"), allow_other = TRUE, all_objects = TRUE)
  check_int(model_inference[["m_number"]],      "model_inference:model_number")
  check_real(model_inference[["marglik"]],      "model_inference:marglik")
  check_real(model_inference[["prior_prob"]],   "model_inference:prior_prob",   lower = 0, upper = 1)
  check_real(model_inference[["post_prob"]],    "model_inference:post_prob",   lower = 0, upper = 1)
  check_real(model_inference[["inclusion_BF"]], "model_inference:inclusion_BF", lower = 0)
  check_list(model_description, "model_description", allow_NULL = TRUE)
  check_bool(short_name, "short_name")
  check_char(title, "title", allow_NULL = TRUE)
  check_char(footnotes, "footnotes", check_length = 0, allow_NULL = TRUE)
  check_char(warnings, "warnings", check_length = 0, allow_NULL = TRUE)
  check_bool(formula_prefix, "formula_prefix")
  check_char(remove_parameters, "remove_parameters", allow_NULL = TRUE, check_length = 0)

  # prepare the columns
  summary_names  <- c(
    "Model",
    if(!is.null(model_description)) names(model_description),
    "Prior prob.",
    "log(marglik)",
    "Post. prob.",
    "Inclusion BF")
  summary_values <- c(
    model_inference[["m_number"]],
    if(!is.null(model_description)) unlist(model_description),
    .format_column(model_inference[["prior_prob"]],   "probability"),
    .format_column(model_inference[["marglik"]],      "marglik"),
    .format_column(model_inference[["post_prob"]],    "probability"),
    .format_column(model_inference[["inclusion_BF"]], "BF"))

  summary_priors  <- "Parameter prior distributions"
  for(i in seq_along(prior_list)){
    # get the prior name
    if(is.prior.none(prior_list[[i]])){
      next
    }else if(remove_spike_0 && is.prior.point(prior_list[[i]]) && prior_list[[i]][["parameters"]][["location"]] == 0 || (names(prior_list)[[i]] %in% remove_parameters)){
      next
    }else if(is.prior.weightfunction(prior_list[[i]]) | is.prior.PET(prior_list[[i]]) | is.prior.PEESE(prior_list[[i]]) |
             is_prior_phacking(prior_list[[i]]) | is_prior_bias(prior_list[[i]])){
      temp_prior <- print(prior_list[[i]], silent = TRUE, short_name = short_name)
    }else if(is.prior.simple(prior_list[[i]]) | is.prior.vector(prior_list[[i]]) | is.prior.factor(prior_list[[i]]) | is.prior.spike_and_slab(prior_list[[i]]) | is.prior.mixture(prior_list[[i]])){
      temp_prior <- paste0(names(prior_list)[i], " ~ " , print(prior_list[[i]], silent = TRUE, short_name = short_name, inline = TRUE))
    }else if(is.prior.point(prior_list[[i]])){
      temp_prior <- paste0(names(prior_list)[i], " = " , print(prior_list[[i]], silent = TRUE, short_name = short_name))
    }
    # change the formula formatting
    if(!is.null(attr(prior_list[[i]], "parameter"))){
      temp_prior <- gsub(
        paste0(attr(prior_list[[i]], "parameter"), "_"),
        if(formula_prefix) paste0("(", attr(prior_list[[i]], "parameter"), ") ") else "",
        temp_prior)
      temp_prior <- gsub("__xXx__", ":", temp_prior)
    }
    summary_priors <- c(summary_priors, temp_prior)
  }

  if(length(summary_names) > length(summary_priors)){
    summary_priors <- c(summary_priors, rep("", length(summary_names) - length(summary_priors)))
  }else if(length(summary_names) < length(summary_priors)){
    summary_names  <- c(summary_names,  rep("", length(summary_priors) - length(summary_names)))
    summary_values <- c(summary_values, rep("", length(summary_priors) - length(summary_values)))
  }
  summary_names <- paste0(summary_names, "  ")

  summary_table <- data.frame(cbind(
    summary_names,
    summary_values,
    rep("           ", length(summary_names)),
    summary_priors
  ))
  names(summary_table) <- NULL

  # prepare output
  class(summary_table)             <- c("BayesTools_table", class(summary_table))
  attr(summary_table, "type")      <- c("string_left", "string", "string", "prior")
  attr(summary_table, "rownames")  <- FALSE
  attr(summary_table, "as.matrix") <- TRUE
  attr(summary_table, "title")     <- title
  attr(summary_table, "footnotes") <- footnotes
  attr(summary_table, "warnings")  <- warnings

  return(summary_table)
}

#' @rdname BayesTools_model_tables
runjags_estimates_table  <- function(fit, transformations = NULL, title = NULL, footnotes = NULL, warnings = NULL, conditional = FALSE,
                                     probs = c(0.025, 0.5, 0.975), remove_spike_0 = TRUE, transform_factors = FALSE, transform_orthonormal = FALSE,
                                     formula_prefix = TRUE, remove_inclusion = FALSE, remove_parameters = NULL, remove_formulas = NULL,
                                     keep_parameters = NULL, keep_formulas = NULL, return_samples = FALSE, transform_scaled = FALSE,
                                     remove_diagnostics = FALSE,
                                     diagnostic_columns = getOption("BayesTools.JAGS_estimates_diagnostic_columns", if(remove_diagnostics) "none" else "all")){

  .check_runjags()
  # most of the code is shared with .diagnostics_plot_data function (keep them in sync on update)

  # check fits
  if(!inherits(fit, "runjags"))
    stop("'fit' must be a runjags fit")
  if(!inherits(fit, "BayesTools_fit"))
    stop("'fit' must be a BayesTools fit")
  prior_list <- attr(fit, "prior_list")
  check_list(prior_list, "prior_list")
  if(!all(sapply(prior_list, is.prior)))
    stop("'prior_list' must be a list of priors.")
  check_list(transformations, "transformations", allow_NULL = TRUE)
  if(!is.null(transformations) && any(!sapply(transformations, function(trans)is.function(trans[["fun"]]))))
    stop("'transformations' must be list of functions in the 'fun' element.")
  check_char(title, "title", allow_NULL = TRUE)
  check_char(footnotes, "footnotes", check_length = 0, allow_NULL = TRUE)
  check_char(warnings, "warnings", check_length = 0, allow_NULL = TRUE)
  check_real(probs, "probs", lower = 0, upper = 1, check_length = 0)
  check_bool(remove_spike_0, "remove_spike_0")
  check_bool(conditional, "conditional")
  check_bool(transform_factors, "transform_factors")
  check_bool(transform_orthonormal, "transform_orthonormal")
  check_bool(formula_prefix, "formula_prefix")
  check_bool(transform_scaled, "transform_scaled")
  check_bool(remove_diagnostics, "remove_diagnostics")
  diagnostic_columns <- .normalize_diagnostic_columns(diagnostic_columns, .JAGS_estimates_diagnostic_columns(), "diagnostic_columns")
  if(remove_diagnostics){
    diagnostic_columns <- character()
  }
  summary_diagnostic_columns <- if(conditional) character() else diagnostic_columns
  if(!is.null(remove_parameters) && !is.logical(remove_parameters))
    check_char(remove_parameters, "remove_parameters", allow_NULL = TRUE, check_length = 0)
  if(is.logical(remove_parameters))
    check_bool(remove_parameters, "remove_parameters")
  check_char(remove_formulas, "remove_formulas", allow_NULL = TRUE, check_length = 0)
  check_char(keep_parameters, "keep_parameters", allow_NULL = TRUE, check_length = 0)
  check_char(keep_formulas, "keep_formulas", allow_NULL = TRUE, check_length = 0)

  # depreciate
  transform_factors <- .depreciate.transform_orthonormal(transform_orthonormal, transform_factors)

  # get model samples
  model_samples <- .extract_posterior_samples(fit, as_list = FALSE)

  # transform scaled coefficients back to original scale before filtering. Lower
  # order coefficients can depend on higher-order interaction columns.
  formula_scale <- NULL
  if(transform_scaled){
    formula_scale <- attr(fit, "formula_scale")
    if(!is.null(formula_scale) && length(formula_scale) > 0){
      model_samples <- transform_scale_samples(model_samples, formula_scale)
    }
  }

  ### remove un-wanted estimates (or support values) - spike and slab priors already dealt with later (also remove the item from prior list)
  # compute filtered parameters using the helper function
  remove_params_vec <- .filter_parameters(
    prior_list        = prior_list,
    remove_parameters = remove_parameters,
    remove_formulas   = remove_formulas,
    keep_parameters   = keep_parameters,
    keep_formulas     = keep_formulas,
    remove_spike_0    = remove_spike_0
  )

  cleaned       <- .remove_auxiliary_parameters(model_samples, prior_list, remove_params_vec)
  model_samples <- cleaned$model_samples
  prior_list    <- cleaned$prior_list

  # simplify mixture and spike and slab priors to simple priors
  # the samples and summary can be dealt with as any other prior (i.e., transformations later)
  for(par in names(prior_list)){
    if(is.prior.spike_and_slab(prior_list[[par]])){

      # process spike and slab using helper function
      processed     <- .process_spike_and_slab(model_samples, prior_list, par, conditional, remove_inclusion, warnings)
      model_samples <- processed$model_samples
      prior_list    <- processed$prior_list
      warnings      <- processed$warnings

    }else if(is.prior.mixture(prior_list[[par]])){

      # check for publication bias component
      is_bias_mixture <- inherits(prior_list[[par]], "prior.bias_mixture")
      is_PET          <- sapply(prior_list[[par]], is.prior.PET)
      is_PEESE        <- sapply(prior_list[[par]], is.prior.PEESE)
      if(is_bias_mixture){
        branch_info   <- .selection_prior_branch_info(prior_list[[par]])
        has_selection <- vapply(branch_info, function(x) !is.null(x$selection), logical(1))
        has_phacking  <- vapply(branch_info, function(x) !is.null(x$phacking), logical(1))
      }else{
        branch_info   <- vector("list", length(prior_list[[par]]))
        has_selection <- rep(FALSE, length(prior_list[[par]]))
        has_phacking  <- rep(FALSE, length(prior_list[[par]]))
      }

      # distinguish between null/alternative and component type notations
      components <- attr(prior_list[[par]], "components")

      if(any(is_PET | is_PEESE | has_selection | has_phacking)){

        # change the samples between conditional/averaged based on the preferences
        if(conditional){

          if(any(is_PET)){
            # compute the number of conditional samples
            n_conditional_samples <- sum(model_samples[,colnames(model_samples) == paste0(par, "_indicator")] %in% which(is_PET))

            # replace null samples with NAs (important for later transformations)
            model_samples[!model_samples[,colnames(model_samples) == paste0(par, "_indicator")] %in% which(is_PET), "PET"] <- NA

            # add warnings about conditional summary
            warnings <- c(warnings, .runjags_conditional_warning("PET", n_conditional_samples))
          }

          if(any(is_PEESE)){
            # compute the number of conditional samples
            n_conditional_samples <- sum(model_samples[,colnames(model_samples) == paste0(par, "_indicator")] %in% which(is_PEESE))

            # replace null samples with NAs (important for later transformations)
            model_samples[!model_samples[,colnames(model_samples) == paste0(par, "_indicator")] %in% which(is_PEESE), "PEESE"] <- NA

            # add warnings about conditional summary
            warnings <- c(warnings, .runjags_conditional_warning("PEESE", n_conditional_samples))
          }

          if(any(has_selection)){
            # compute the number of conditional samples
            n_conditional_samples <- sum(model_samples[,colnames(model_samples) == paste0(par, "_indicator")] %in% which(has_selection))

            # replace null samples with NAs (important for later transformations)
            model_samples[!model_samples[,colnames(model_samples) == paste0(par, "_indicator")] %in% which(has_selection), grepl("omega", colnames(model_samples))] <- NA

            # add warnings about conditional summary
            warnings <- c(warnings, .runjags_conditional_warning("omega", n_conditional_samples))
          }

          if(any(has_phacking)){
            # compute the number of conditional samples
            n_conditional_samples <- sum(model_samples[,colnames(model_samples) == paste0(par, "_indicator")] %in% which(has_phacking))

            # replace non-p-hacking samples with NAs
            phacking_report_columns <- .selection_phacking_report_parameters(lapply(branch_info[has_phacking], function(x) x$phacking))
            phacking_columns <- intersect(c(phacking_report_columns, "phack_kind"), colnames(model_samples))
            if(length(phacking_columns) > 0){
              model_samples[!model_samples[,colnames(model_samples) == paste0(par, "_indicator")] %in% which(has_phacking), phacking_columns] <- NA
            }

            # add warnings about conditional summary
            warning_columns <- intersect(phacking_report_columns, phacking_columns)
            if(length(warning_columns) > 0){
              warnings <- c(warnings, .runjags_conditional_warning(warning_columns, n_conditional_samples))
            }
          }

        }

        # re-format the weightfunctions
        if(any(has_selection) || any(has_phacking)){

          # rename
          omega_cuts      <- if(any(has_selection)){
            selection_priors <- lapply(branch_info[has_selection], function(x) x$selection)
            weightfunctions_mapping(selection_priors, cuts_only = TRUE, one_sided = TRUE)
          }else{
            c(0, 1)
          }
          omega_names_old <- paste0("omega[", 1:(length(omega_cuts)-1),"]")
          omega_names     <- sapply(1:(length(omega_cuts)-1), function(i)paste0("omega[",omega_cuts[i],",",omega_cuts[i+1],"]"))
          colnames(model_samples)[which(colnames(model_samples) %in% omega_names_old)] <- omega_names

          # remove if requested
          if("omega" %in% remove_parameters){
            model_samples <- model_samples[,!colnames(model_samples) %in% omega_names,drop=FALSE]
            if(any(has_selection)){
              prior_list[[par]][has_selection] <- NULL
            }
          }
        }

        # add the simpler priors to the prior list
        if(any(is_PET)){
          prior_list[["PET"]] <- prior_list[[par]][is_PET][1]
        }
        if(any(is_PEESE)){
          prior_list[["PEESE"]] <- prior_list[[par]][is_PEESE][1]
        }
        if(any(has_selection)){
          prior_list[["omega"]] <- branch_info[has_selection][[1]]$selection
        }
        if(any(has_phacking)){
          phacking_priors <- lapply(branch_info[has_phacking], function(x) x$phacking)
          drop_phacking <- .selection_phacking_unreported_parameters(phacking_priors)
          model_samples <- model_samples[, !colnames(model_samples) %in% drop_phacking, drop = FALSE]

          phacking_report_columns <- .selection_phacking_report_parameters(phacking_priors)
          for(phacking_report_column in phacking_report_columns){
            prior_list[[phacking_report_column]] <- phacking_priors[[1]]$alpha
          }
        }

      }else{

        # prepare parameter names
        if(inherits(prior_list[[par]], "prior.factor_mixture")){
          par_names <- .JAGS_prior_factor_names(par, prior_list[[par]])
        }else{
          par_names <- par
        }

        # change the samples between conditional/averaged based on the preferences
        if(conditional){

          if(all(components %in% c("null", "alternative"))){

            # select the corresponding indicators
            this_component_indicator <- which(components == "alternative")

            # compute the number of conditional samples
            n_conditional_samples <- sum(model_samples[,colnames(model_samples) == paste0(par, "_indicator")] %in% this_component_indicator)

            # replace null samples with NAs (important for later transformations)
            model_samples[!model_samples[,colnames(model_samples) == paste0(par, "_indicator")] %in% this_component_indicator, par_names] <- NA

            # add warnings about conditional summary
            warnings <- c(warnings, .runjags_conditional_warning(par_names, n_conditional_samples))

          }else{

            # remove the join samples and replace with individual conditional samples
            temp_position    <- min(which(colnames(model_samples) %in% par))
            temp_all_samples <- model_samples[, colnames(model_samples) %in% par,drop=FALSE]
            temp_new_samples <- list()
            model_samples    <- model_samples[,!colnames(model_samples) %in% par,drop=FALSE]

             # component-by-component replacement
            for(component in unique(components[components != "null"])){

              # create component specific samples
              temp_par_names <- paste0(par_names, "[", component, "]")
              temp_new_samples[[component]]           <- temp_all_samples
              colnames(temp_new_samples[[component]]) <- temp_par_names

              # select the corresponding indicators
              this_component_indicator <- which(components == component)

              # compute the number of conditional samples
              n_conditional_samples <- sum(model_samples[,colnames(model_samples) == paste0(par, "_indicator")] %in% this_component_indicator)

              # replace null samples with NAs (important for later transformations)
              temp_new_samples[[component]][!model_samples[,colnames(model_samples) == paste0(par, "_indicator")] %in% this_component_indicator,] <- NA

              # add warnings about conditional summary
              warnings <- c(warnings, .runjags_conditional_warning(temp_par_names, n_conditional_samples))

              # forward transformations to the conditional estimates
              if(!is.null(transformations[[par]])){
                transformations[[temp_par_names]] <- transformations[[par]]
                attr(prior_list[[par]][which(components == component)][1], "parameter") <- attr(prior_list[[par]], "parameter")
                prior_list[[temp_par_names]]      <- prior_list[[par]][which(components == component)][1]
              }
            }

            # place the transformed samples back
            model_samples <- cbind(
              if(temp_position > 1) model_samples[,1:(temp_position-1),drop=FALSE],
              do.call(cbind, temp_new_samples),
              if(temp_position <= ncol(model_samples)) model_samples[,temp_position:ncol(model_samples),drop=FALSE]
            )

            # remove the original parameter transformations
            if(!is.null(transformations[[par]])){
              transformations[[par]] <- NULL
              prior_list[[par]]      <- NULL
            }
          }
        }
      }

      # remove/rename the inclusions probabilities
      if(remove_inclusion){
        model_samples   <- model_samples[,colnames(model_samples) != paste0(par, "_indicator"),drop=FALSE]
      }else{
        if(all(components %in% c("null", "alternative"))){
          # replace and rename in the samples
          model_samples[,colnames(model_samples) == paste0(par, "_indicator")] <- ifelse(
            model_samples[,colnames(model_samples) == paste0(par, "_indicator")] %in% which(components == "alternative"), 1, 0)
          colnames(model_samples)[colnames(model_samples) == paste0(par, "_indicator")] <- paste0(par, " (inclusion)")
        }else{
          # extract
          temp_position <- min(which(colnames(model_samples) %in% paste0(par, "_indicator")))
          temp_samples  <- model_samples[,colnames(model_samples) == paste0(par, "_indicator")]
          model_samples <- model_samples[,colnames(model_samples) != paste0(par, "_indicator"),drop=FALSE]

          # compute component specific indicators
          temp_new_samples <- lapply(unique(components), function(component) ifelse(temp_samples %in% which(components == component), 1, 0))
          temp_new_samples <- do.call(cbind, temp_new_samples)
          colnames(temp_new_samples) <- paste0(par, " (inclusion: ", unique(components),")")

          # place the transformed samples back
          model_samples <- cbind(
            if(temp_position > 1) model_samples[,1:(temp_position-1),drop=FALSE],
            temp_new_samples,
            if(temp_position <= ncol(model_samples)) model_samples[,temp_position:ncol(model_samples),drop=FALSE]
          )
        }
      }
    }
  }

  # remove transformations for removed variables
  if(!is.null(transformations)){
    transformations <-  transformations[names(transformations) %in% names(prior_list)]
  }

  # apply transformations (not orthornormal if they are to be returned transformed to diffs)
  model_samples <- .apply_parameter_transformations(model_samples, transformations, prior_list, transform_factors)

  # transform orthonormal factors to differences from mean
  model_samples <- .transform_factor_contrasts(model_samples, prior_list, transform_factors, transformations)

  # rename factor levels
  model_samples <- .rename_factor_levels(model_samples, prior_list)

  # store parameter names before removing formula attachments
  parameter_names <- colnames(model_samples)

  # rename formula parameters
  if(any(!sapply(lapply(prior_list, attr, which = "parameter"), is.null))){
    colnames(model_samples) <- format_parameter_names(
      parameters         = colnames(model_samples),
      formula_parameters = unique(unlist(lapply(prior_list, attr, which = "parameter"))),
      formula_random     = unique(unlist(lapply(prior_list, attr, which = "random_factor"))),
      formula_prefix     = formula_prefix,
      formula_scale      = if(transform_scaled) formula_scale else NULL)
  }

  # return samples if requested
  if(return_samples){
    attr(model_samples, "prior_list") <- prior_list
    return(model_samples)
  }

  # compute the summary
  if(ncol(model_samples) == 0){
    return(runjags_estimates_empty_table(
      probs              = probs,
      title              = title,
      footnotes          = footnotes,
      warnings           = warnings,
      remove_diagnostics = remove_diagnostics,
      diagnostic_columns = summary_diagnostic_columns
    ))
  }else{
    runjags_summary <- .runjags_summary_fast(
      model_samples       = model_samples,
      n_samples           = fit$sample,
      n_chains            = length(fit$mcmc),
      conditional         = conditional,
      probs               = probs,
      remove_diagnostics  = remove_diagnostics,
      diagnostic_columns  = diagnostic_columns
    )
  }

  # prepare output
  n_estimate_cols <- 2 + length(probs)  # Mean, SD, quantiles
  class(runjags_summary)              <- c("BayesTools_table", "BayesTools_runjags_summary", class(runjags_summary))
  attr(runjags_summary, "type")       <- c(rep("estimate", n_estimate_cols), summary_diagnostic_columns)
  attr(runjags_summary, "parameters") <- parameter_names
  attr(runjags_summary, "rownames")   <- TRUE
  attr(runjags_summary, "title")      <- title
  attr(runjags_summary, "footnotes")  <- footnotes
  attr(runjags_summary, "warnings")   <- warnings

  return(runjags_summary)
}

#' @rdname BayesTools_model_tables
runjags_inference_table  <- function(fit, title = NULL, footnotes = NULL, warnings = NULL, formula_prefix = TRUE,
                                     logBF = FALSE, BF01 = FALSE, BF_diagnostics = FALSE,
                                     BF_diagnostic_columns = getOption("BayesTools.JAGS_BF_diagnostic_columns", if(BF_diagnostics) "all" else "none")){

  # check fits
  if(!inherits(fit, "runjags"))
    stop("'fit' must be a runjags fit")
  if(!inherits(fit, "BayesTools_fit"))
    stop("'fit' must be a BayesTools fit")
  prior_list <- attr(fit, "prior_list")
  check_list(prior_list, "prior_list")
  if(!all(sapply(prior_list, is.prior)))
    stop("'prior_list' must be a list of priors.")
  check_char(title, "title", allow_NULL = TRUE)
  check_char(footnotes, "footnotes", check_length = 0, allow_NULL = TRUE)
  check_char(warnings, "warnings", check_length = 0, allow_NULL = TRUE)
  check_bool(formula_prefix, "formula_prefix")
  check_bool(logBF, "logBF")
  check_bool(BF01,  "BF01")
  check_bool(BF_diagnostics, "BF_diagnostics")
  BF_diagnostic_columns <- .normalize_diagnostic_columns(BF_diagnostic_columns, .JAGS_BF_diagnostic_columns(), "BF_diagnostic_columns")
  BF_diagnostics        <- length(BF_diagnostic_columns) > 0
  BF_error_diagnostics  <- "BF_error_percent" %in% BF_diagnostic_columns

  # return empty table if none of the priors is spike and slab
  if(!any(sapply(prior_list, function(p) is.prior.spike_and_slab(p) | is.prior.mixture(p)))){
    runjags_summary <- runjags_inference_empty_table(
      title          = title,
      footnotes      = footnotes,
      warnings       = warnings,
      logBF          = logBF,
      BF01           = BF01,
      BF_diagnostics = BF_diagnostics,
      BF_diagnostic_columns = BF_diagnostic_columns
    )
    return(runjags_summary)
  }

  # extract samples
  model_samples   <- .extract_posterior_samples(fit, as_list = FALSE)
  if(BF_diagnostics){
    model_samples_list <- .extract_posterior_samples(fit, as_list = TRUE)
  }else{
    model_samples_list <- NULL
  }
  runjags_summary <- data.frame(matrix(nrow = 0, ncol = 4 + length(BF_diagnostic_columns)))
  colnames(runjags_summary) <- c("Parameter", "prior_prob", "post_prob", "inclusion_BF", BF_diagnostic_columns)
  BF_bound_operators <- character()

  for(par in names(prior_list)){
    if(is.prior.spike_and_slab(prior_list[[par]])){

      temp_prior_prob <- mean(.get_spike_and_slab_inclusion(prior_list[[par]]))
      temp_post_prob  <- mean(model_samples[,paste0(par, "_indicator")])
      temp_BF         <- inclusion_BF(
        prior_probs = c(null = 1 - temp_prior_prob, alternative = temp_prior_prob),
        post_probs  = c(null = 1 - temp_post_prob,  alternative = temp_post_prob),
        is_null     = c(TRUE, FALSE)
      )
      temp_BF_reporting <- .indicator_BF_reporting_value(temp_BF, temp_post_prob, temp_prior_prob, nrow(model_samples))
      temp_row        <- data.frame(
        Parameter    = par,
        prior_prob   = temp_prior_prob,
        post_prob    = temp_post_prob,
        inclusion_BF = temp_BF_reporting$value
      )
      if(BF_diagnostics){
        temp_indicator_list <- .runjags_indicator_list(model_samples_list, paste0(par, "_indicator"), 1)
        temp_diagnostics    <- .indicator_BF_diagnostics(temp_indicator_list, temp_prior_prob, temp_BF)
        temp_row            <- cbind(temp_row, .indicator_BF_diagnostic_row(temp_diagnostics)[, BF_diagnostic_columns, drop = FALSE])
        if(BF_error_diagnostics){
          warnings <- c(warnings, .indicator_BF_warnings(par, temp_diagnostics))
        }
      }

      runjags_summary <- rbind(runjags_summary, temp_row)
      BF_bound_operators <- c(BF_bound_operators, temp_BF_reporting$operator)
    }else if(is.prior.mixture(prior_list[[par]])){

      # extract the components and prior probabilities
      components      <- attr(prior_list[[par]], "components")
      temp_prior_prob <- attr(prior_list[[par]], "prior_weights")
      temp_prior_prob <- sapply(unique(components), function(component) sum(temp_prior_prob[which(components == component)])) / sum(temp_prior_prob)
      temp_post_prob  <- sapply(unique(components), function(component) mean(model_samples[,paste0(par, "_indicator")] %in% which(components == component)))

      # if only null and alternative are specified, removed the null component
      if(all(components %in% c("null", "alternative"))){

        if(all(components == "null")){
          temp_prior_prob <- c(temp_prior_prob, "alternative" = 0)
          temp_post_prob  <- c(temp_post_prob,  "alternative" = 0)
        }
        if(all(components == "alternative")){
          temp_prior_prob <- c("null" = 0, temp_prior_prob)
          temp_post_prob  <- c("null" = 0,  temp_post_prob)
        }

        temp_BF  <- inclusion_BF(prior_probs = temp_prior_prob, post_probs = temp_post_prob, is_null = names(temp_post_prob) != "alternative")
        temp_BF_reporting <- .indicator_BF_reporting_value(temp_BF, temp_post_prob[["alternative"]], temp_prior_prob[["alternative"]], nrow(model_samples))
        temp_row <- data.frame(
          Parameter    = par,
          prior_prob   = temp_prior_prob[["alternative"]],
          post_prob    = temp_post_prob[["alternative"]],
          inclusion_BF = temp_BF_reporting$value
        )
        if(BF_diagnostics){
          temp_indicator_list <- .runjags_indicator_list(model_samples_list, paste0(par, "_indicator"), which(components == "alternative"))
          temp_diagnostics    <- .indicator_BF_diagnostics(temp_indicator_list, temp_prior_prob[["alternative"]], temp_BF)
          temp_row            <- cbind(temp_row, .indicator_BF_diagnostic_row(temp_diagnostics)[, BF_diagnostic_columns, drop = FALSE])
          if(BF_error_diagnostics){
            warnings <- c(warnings, .indicator_BF_warnings(par, temp_diagnostics))
          }
        }

        runjags_summary <- rbind(runjags_summary, temp_row)
        BF_bound_operators <- c(BF_bound_operators, temp_BF_reporting$operator)

      }else{

        # compute summary for each component
        for(component in unique(components)){
          temp_parameter <- paste0(par, " [", component, "]")
          temp_BF        <- inclusion_BF(prior_probs = temp_prior_prob, post_probs = temp_post_prob, is_null = names(temp_post_prob) != component)
          temp_BF_reporting <- .indicator_BF_reporting_value(temp_BF, temp_post_prob[[component]], temp_prior_prob[[component]], nrow(model_samples))
          temp_row       <- data.frame(
            Parameter    = paste0(par, " [", component, "]"),
            prior_prob   = temp_prior_prob[[component]],
            post_prob    = temp_post_prob[[component]],
            inclusion_BF = temp_BF_reporting$value
          )
          if(BF_diagnostics){
            temp_indicator_list <- .runjags_indicator_list(model_samples_list, paste0(par, "_indicator"), which(components == component))
            temp_diagnostics    <- .indicator_BF_diagnostics(temp_indicator_list, temp_prior_prob[[component]], temp_BF)
            temp_row            <- cbind(temp_row, .indicator_BF_diagnostic_row(temp_diagnostics)[, BF_diagnostic_columns, drop = FALSE])
            if(BF_error_diagnostics){
              warnings <- c(warnings, .indicator_BF_warnings(temp_parameter, temp_diagnostics))
            }
          }

          runjags_summary <- rbind(runjags_summary, temp_row)
          BF_bound_operators <- c(BF_bound_operators, temp_BF_reporting$operator)
        }

      }
    }
  }

  # store parameter names before removing formula attachments
  parameter_names           <- runjags_summary$Parameter
  rownames(runjags_summary) <- parameter_names
  runjags_summary           <- runjags_summary[,-1]

  # format BF and BF MC error on the requested scale
  temp_inclusion_BF <- runjags_summary[,"inclusion_BF"]
  attr(temp_inclusion_BF, "bound_operator") <- BF_bound_operators
  runjags_summary[["inclusion_BF"]] <- format_BF(temp_inclusion_BF, logBF = logBF, BF01 = BF01, inclusion = TRUE)
  if("MCMC_error" %in% BF_diagnostic_columns){
    attr(runjags_summary[["MCMC_error"]], "name") <- "error(Post. prob.)"
  }
  if("BF_error_percent" %in% BF_diagnostic_columns){
    attr(runjags_summary[["BF_error_percent"]], "name") <- .BF_error_column_name(BF01)
  }

  # rename formula parameters
  if(any(!sapply(lapply(prior_list, attr, which = "parameter"), is.null))){
    rownames(runjags_summary) <- format_parameter_names(
      parameters         = rownames(runjags_summary),
      formula_parameters = unique(unlist(lapply(prior_list, attr, which = "parameter"))),
      formula_random     = unique(unlist(lapply(prior_list, attr, which = "random_factor"))),
      formula_prefix     = formula_prefix,
      formula_scale      = NULL)
  }

  class(runjags_summary)               <- c("BayesTools_table", "BayesTools_runjags_inference", class(runjags_summary))
  attr(runjags_summary, "type")        <- c("prior_prob", "post_prob", "inclusion_BF", .JAGS_BF_diagnostic_column_types(BF_diagnostic_columns))
  attr(runjags_summary, "parameters")  <- parameter_names
  attr(runjags_summary, "rownames")    <- TRUE
  attr(runjags_summary, "title")       <- title
  attr(runjags_summary, "footnotes")   <- footnotes
  attr(runjags_summary, "warnings")    <- warnings

  return(runjags_summary)
}

#' @rdname BayesTools_model_tables
JAGS_estimates_table <- runjags_estimates_table

#' @rdname BayesTools_model_tables
JAGS_inference_table <- runjags_inference_table

#' @rdname BayesTools_model_tables
JAGS_summary_table   <- model_summary_table

#' @rdname BayesTools_model_tables
model_summary_empty_table <- function(model_description = NULL, title = NULL, footnotes = NULL, warnings = NULL){

  check_list(model_description, "model_description", allow_NULL = TRUE)

  summary_names  <- c(
    "Model",
    if(!is.null(model_description)) names(model_description),
    "Prior prob.",
    "log(marglik)",
    "Post. prob.",
    "Inclusion BF")


  summary_names <- paste0(summary_names, "  ")

  empty_table <- data.frame(cbind(
    summary_names,
    rep("",            length(summary_names)),
    rep("           ", length(summary_names)),
    c("Parameter prior distributions", rep("", length(summary_names) - 1))
  ))
  names(empty_table) <- NULL

  # prepare output
  class(empty_table)             <- c("BayesTools_table", class(empty_table))
  attr(empty_table, "type")      <- c("string_left", "string", "string", "prior")
  attr(empty_table, "rownames")  <- FALSE
  attr(empty_table, "as.matrix") <- TRUE
  attr(empty_table, "title")     <- title
  attr(empty_table, "footnotes") <- footnotes
  attr(empty_table, "warnings")  <- warnings

  return(empty_table)
}

#' @rdname BayesTools_model_tables
runjags_estimates_empty_table <- function(probs = c(0.025, 0.5, 0.975), title = NULL, footnotes = NULL, warnings = NULL,
                                          remove_diagnostics = FALSE,
                                          diagnostic_columns = getOption("BayesTools.JAGS_estimates_diagnostic_columns", if(remove_diagnostics) "none" else "all")){

  check_bool(remove_diagnostics, "remove_diagnostics")
  diagnostic_columns <- .normalize_diagnostic_columns(diagnostic_columns, .JAGS_estimates_diagnostic_columns(), "diagnostic_columns")
  if(remove_diagnostics){
    diagnostic_columns <- character()
  }
  n_estimate_cols <- 2 + length(probs)  # Mean, SD, quantiles
  empty_table <- data.frame(matrix(nrow = 0, ncol = n_estimate_cols + length(diagnostic_columns)), check.names = FALSE)
  colnames(empty_table) <- c("Mean", "SD", as.character(probs), diagnostic_columns)

  class(empty_table)             <- c("BayesTools_table", "BayesTools_runjags_summary", class(empty_table))
  attr(empty_table, "type")      <- c(rep("estimate", n_estimate_cols), diagnostic_columns)
  attr(empty_table, "rownames")  <- FALSE
  attr(empty_table, "title")     <- title
  attr(empty_table, "footnotes") <- footnotes
  attr(empty_table, "warnings")  <- warnings

  return(empty_table)
}

#' @rdname BayesTools_model_tables
runjags_inference_empty_table <- function(title = NULL, footnotes = NULL, warnings = NULL,
                                          logBF = FALSE, BF01 = FALSE, BF_diagnostics = FALSE,
                                          BF_diagnostic_columns = getOption("BayesTools.JAGS_BF_diagnostic_columns", if(BF_diagnostics) "all" else "none")){

  check_char(title, "title", allow_NULL = TRUE)
  check_char(footnotes, "footnotes", check_length = 0, allow_NULL = TRUE)
  check_char(warnings, "warnings", check_length = 0, allow_NULL = TRUE)
  check_bool(logBF, "logBF")
  check_bool(BF01,  "BF01")
  check_bool(BF_diagnostics, "BF_diagnostics")
  BF_diagnostic_columns <- .normalize_diagnostic_columns(BF_diagnostic_columns, .JAGS_BF_diagnostic_columns(), "BF_diagnostic_columns")
  BF_diagnostics        <- length(BF_diagnostic_columns) > 0

  empty_table <- data.frame(matrix(nrow = 0, ncol = 3 + length(BF_diagnostic_columns)))
  colnames(empty_table) <- c("prior_prob", "post_prob", "inclusion_BF", BF_diagnostic_columns)
  attr(empty_table[["inclusion_BF"]], "name") <- .BF_column_name(logBF = logBF, BF01 = BF01, inclusion = TRUE)
  if("MCMC_error" %in% BF_diagnostic_columns){
    attr(empty_table[["MCMC_error"]], "name") <- "error(Post. prob.)"
  }
  if("BF_error_percent" %in% BF_diagnostic_columns){
    attr(empty_table[["BF_error_percent"]], "name") <- .BF_error_column_name(BF01)
  }

  class(empty_table)             <- c("BayesTools_table", "BayesTools_runjags_inference", class(empty_table))
  attr(empty_table, "type")      <- c("prior_prob", "post_prob", "inclusion_BF", .JAGS_BF_diagnostic_column_types(BF_diagnostic_columns))
  attr(empty_table, "rownames")  <- FALSE
  attr(empty_table, "title")     <- title
  attr(empty_table, "footnotes") <- footnotes
  attr(empty_table, "warnings")  <- warnings

  return(empty_table)
}

#' @rdname BayesTools_model_tables
JAGS_estimates_empty_table <- runjags_estimates_empty_table

#' @rdname BayesTools_model_tables
JAGS_inference_empty_table <- runjags_inference_empty_table

#' @rdname BayesTools_model_tables
stan_estimates_table  <- function(fit, transformations = NULL, title = NULL, footnotes = NULL, warnings = NULL){

  # this is a simplification of the runjags_estimates_table function for stan
  .check_rstan()

  # check fits
  if(!inherits(fit, "stanfit"))
    stop("'fit' must be a rstan fit")
  prior_list <- attr(fit, "prior_list")
  check_list(prior_list, "prior_list")
  if(!all(sapply(prior_list, is.prior)))
    stop("'prior_list' must be a list of priors.")
  check_list(transformations, "transformations", allow_NULL = TRUE)
  if(!is.null(transformations) && any(!sapply(transformations, function(trans)is.function(trans[["fun"]]))))
    stop("'transformations' must be list of functions in the 'fun' element.")
  check_char(title, "title", allow_NULL = TRUE)
  check_char(footnotes, "footnotes", check_length = 0, allow_NULL = TRUE)
  check_char(warnings, "warnings", check_length = 0, allow_NULL = TRUE)

  # obtain model information
  stan_summary  <- data.frame(rstan::summary(fit)$summary)
  model_samples <- .extract_stan(fit, drop = FALSE)

  # remove un-wanted columns
  stan_summary <- stan_summary[,!colnames(stan_summary) %in% c("X25.", "X75."), drop = FALSE]

  # remove un-wanted rows
  stan_summary <- stan_summary[-nrow(stan_summary),, drop = FALSE]

  # rename columns to match runjags output
  colnames(stan_summary) <- c("Mean", "MCerr", "SD", "Lower95", "Median", "Upper95", "SSeff", "psrf")

  # add MC.ofSD estimates
  stan_summary[, "MC.ofSD"] <- stan_summary[, "MCerr"] / stan_summary[, "SD"]


  # apply transformations
  if(!is.null(transformations)){
    for(par in names(transformations)){
      model_samples[,par] <- do.call(transformations[[par]][["fun"]], c(list(model_samples[,par]), transformations[[par]][["arg"]]))
      transformed_summary <- .stan_transformed_summary(model_samples[,par], stan_summary[par, "SSeff"])
      stan_summary[par, names(transformed_summary)] <- transformed_summary
      stan_summary[par, "MC.ofSD"] <- stan_summary[par, "MCerr"] / stan_summary[par, "SD"]
    }
  }


  # rename the rest
  colnames(stan_summary)[colnames(stan_summary) == "Lower95"] <- "0.025"
  colnames(stan_summary)[colnames(stan_summary) == "Median"]  <- "0.5"
  colnames(stan_summary)[colnames(stan_summary) == "Upper95"] <- "0.975"
  colnames(stan_summary)[colnames(stan_summary) == "MCerr"]   <- "MCMC_error"
  colnames(stan_summary)[colnames(stan_summary) == "MC.ofSD"] <- "MCMC_SD_error"
  colnames(stan_summary)[colnames(stan_summary) == "SSeff"]   <- "ESS"
  colnames(stan_summary)[colnames(stan_summary) == "psrf"]    <- "R_hat"

  # reorder the columns
  stan_summary <- stan_summary[,c("Mean", "SD", "0.025", "0.5", "0.975", "MCMC_error", "MCMC_SD_error", "ESS", "R_hat"), drop = FALSE]

  # store parameter names
  parameter_names <- rownames(stan_summary)

  # prepare output
  class(stan_summary)              <- c("BayesTools_table", "BayesTools_stan_summary", class(stan_summary))
  attr(stan_summary, "type")       <- c(rep("estimate", 5), "MCMC_error", "MCMC_SD_error", "ESS", "R_hat")
  attr(stan_summary, "parameters") <- parameter_names
  attr(stan_summary, "rownames")   <- TRUE
  attr(stan_summary, "title")      <- title
  attr(stan_summary, "footnotes")  <- footnotes
  attr(stan_summary, "warnings")   <- warnings

  return(stan_summary)
}

.stan_transformed_summary <- function(samples, SSeff){

  samples <- as.numeric(samples)
  qs      <- stats::quantile(samples, probs = c(0.025, 0.5, 0.975), na.rm = TRUE, names = FALSE)
  sd      <- stats::sd(samples, na.rm = TRUE)
  MCerr   <- if(is.finite(SSeff) && SSeff > 0) sd / sqrt(SSeff) else NA_real_

  c(
    "Mean"    = mean(samples, na.rm = TRUE),
    "MCerr"   = MCerr,
    "SD"      = sd,
    "Lower95" = qs[1],
    "Median"  = qs[2],
    "Upper95" = qs[3]
  )
}

#' @title Print a BayesTools table
#'
#' @param x a BayesTools_values_tables
#' @param ... additional arguments.
#'
#' @return \code{print.BayesTools_table} returns \code{NULL}.
#'
#' @exportS3Method
print.BayesTools_table <- function(x, ...){

  # print formatting
  for(i in seq_along(attr(x, "type"))){
    values         <- x[[i]]
    colnames(x)[i] <- .format_column_names(colnames(x)[i], attr(x, "type")[i], values)
    x[[i]]         <- .format_column(values, attr(x, "type")[i], attr(x, "n_models")[i])
  }

  # print title
  if(!is.null(attr(x, "title"))){
    cat(paste0(attr(x, "title"), "\n"))
  }

  # print the table
  print(as.data.frame(x), quote = FALSE, right = TRUE, row.names = attr(x, "rownames"))

  # print footnotes
  for(i in seq_along(attr(x, "footnotes"))){
    cat(paste0(attr(x, "footnotes")[i], "\n"))
  }

  # print warnings in red
  for(i in seq_along(attr(x, "warnings"))){
    cat(paste0("\033[0;31m", attr(x, "warnings")[i], "\033[0m\n"))
  }

  return(invisible())
}

#' @title Format Bayes factor
#'
#' @description Formats Bayes factor
#'
#' @param BF Bayes factor(s)
#' @param logBF log(BF)
#' @param BF01 1/BF
#' @param inclusion whether the Bayes factor is an inclusion BF (for naming purposes)
#'
#' @return \code{format_BF} returns a formatted Bayes factor.
#'
#' @export
format_BF <- function(BF, logBF = FALSE, BF01 = FALSE, inclusion = FALSE){

  if(!is.numeric(BF)){
    check_real(BF, "BF", lower = 0, check_length = FALSE, allow_NA = TRUE)
  }
  BF_names <- names(BF)
  bound_operator <- .standardize_BF_bound_operator(attr(BF, "bound_operator"), length(BF))
  BF <- as.numeric(BF)
  names(BF) <- BF_names
  check_real(BF, "BF", lower = 0, check_length = FALSE, allow_NA = TRUE)
  check_bool(logBF, "logBF")
  check_bool(BF01,  "BF01")

  if(BF01){
    BF   <- 1/BF
    bound_operator <- .invert_BF_bound_operator(bound_operator)
    name <- ifelse(inclusion, "Exclusion BF", "1/BF")
  }else{
    name <- ifelse(inclusion, "Inclusion BF", "BF")
  }

  if(logBF){
    BF   <- log(BF)
    name <- paste0("log(", name, ")")
  }

  attr(BF, "name")  <- name
  attr(BF, "logBF") <- logBF
  attr(BF, "BF01")  <- BF01
  attr(BF, "bound_operator") <- bound_operator
  if(any(!is.na(bound_operator))){
    class(BF) <- unique(c("BayesTools_BF", class(BF)))
  }

  return(BF)
}

#' @exportS3Method
`[.BayesTools_BF` <- function(x, i, ...){

  bound_operator <- attr(x, "bound_operator")
  out <- NextMethod("[")

  attr(out, "name")  <- attr(x, "name")
  attr(out, "logBF") <- attr(x, "logBF")
  attr(out, "BF01")  <- attr(x, "BF01")
  if(!is.null(bound_operator)){
    attr(out, "bound_operator") <- if(missing(i)) bound_operator else bound_operator[i]
  }
  if(any(!is.na(attr(out, "bound_operator")))){
    class(out) <- unique(c("BayesTools_BF", class(out)))
  }

  return(out)
}

#' @exportS3Method
`[.BayesTools_table` <- function(x, i, j, ..., drop = TRUE){

  original_names <- names(x)
  original_col_attributes <- lapply(x, attributes)

  out <- NextMethod("[")
  if(!is.data.frame(out)){
    return(out)
  }

  out_names <- names(out)
  source_cols <- .match_table_columns(out_names, original_names)
  valid_cols  <- !is.na(source_cols)

  if(length(source_cols) > 0 && any(valid_cols)){
    for(k in which(valid_cols)){
      out[[k]] <- .restore_table_column_attributes(out[[k]], original_col_attributes[[source_cols[k]]])
    }
  }

  type <- attr(x, "type")
  if(length(type) == length(original_names)){
    attr(out, "type") <- type[source_cols]
  }

  n_models <- attr(x, "n_models")
  if(length(n_models) == length(original_names)){
    attr(out, "n_models") <- n_models[source_cols]
  }

  selected_parameters <- .subset_table_parameters(x, out)
  if(!is.null(selected_parameters)){
    attr(out, "parameters") <- selected_parameters
  }
  attr(out, "warnings") <- .subset_table_warnings(attr(x, "warnings"), selected_parameters, rownames(out))

  out
}

.BF_column_name <- function(logBF = FALSE, BF01 = FALSE, inclusion = FALSE){

  if(BF01){
    name <- ifelse(inclusion, "Exclusion BF", "1/BF")
  }else{
    name <- ifelse(inclusion, "Inclusion BF", "BF")
  }

  if(logBF){
    name <- paste0("log(", name, ")")
  }

  return(name)
}

.runjags_indicator_list <- function(model_samples_list, indicator_name, indicator_values){

  lapply(model_samples_list, function(samples){
    samples <- as.matrix(samples)
    if(!indicator_name %in% colnames(samples)){
      stop(paste0("The '", indicator_name, "' indicator was not found in the posterior samples."), call. = FALSE)
    }
    as.numeric(samples[, indicator_name] %in% indicator_values)
  })
}

.indicator_MCMC_summary <- function(indicator_list){

  indicator_list <- lapply(indicator_list, as.numeric)
  indicator      <- unlist(indicator_list, use.names = FALSE)

  indicator_mcmc <- coda::as.mcmc.list(lapply(indicator_list, function(x){
    coda::as.mcmc(matrix(x, ncol = 1, dimnames = list(NULL, "indicator")))
  }))

  mcmc_summary <- summary(indicator_mcmc, quantiles = NULL)$statistics
  if(is.null(dim(mcmc_summary))){
    mcmc_summary <- t(mcmc_summary)
  }

  ESS <- coda::effectiveSize(indicator_mcmc)[1]
  if(is.nan(ESS)){
    ESS <- 0
  }

  return(list(
    post_prob       = mean(indicator),
    MCMC_error      = mcmc_summary[1,"Time-series SE"],
    MCMC_SD_error   = mcmc_summary[1,"Time-series SE"] / mcmc_summary[1,"SD"],
    ESS             = ESS,
    visits          = sum(indicator),
    n_samples       = length(indicator)
  ))
}

.indicator_BF_diagnostics <- function(indicator_list, prior_prob, BF){

  diagnostics <- .indicator_MCMC_summary(indicator_list)

  if(diagnostics$post_prob <= 0 || diagnostics$post_prob >= 1 ||
     prior_prob <= 0 || prior_prob >= 1 ||
     !is.finite(BF) || !is.finite(diagnostics$MCMC_error)){
    BF_error_percent <- NA_real_
  }else{
    log_BF_error <- diagnostics$MCMC_error / (diagnostics$post_prob * (1 - diagnostics$post_prob))
    BF_error_percent <- 100 * log_BF_error
  }

  diagnostics$BF_error_percent <- BF_error_percent

  return(diagnostics)
}

.indicator_BF_diagnostic_row <- function(diagnostics){

  data.frame(
    ESS              = diagnostics$ESS,
    MCMC_error       = diagnostics$MCMC_error,
    BF_error_percent = diagnostics$BF_error_percent
  )
}

.indicator_BF_reporting_value <- function(BF, post_prob, prior_prob, n_samples){

  out <- list(value = BF, operator = NA_character_)
  if(length(post_prob) != 1L || length(prior_prob) != 1L || length(n_samples) != 1L){
    return(out)
  }
  if(!is.finite(post_prob) || !is.finite(prior_prob) || !is.finite(n_samples) ||
     prior_prob <= 0 || prior_prob >= 1 || n_samples <= 1){
    return(out)
  }

  prior_odds <- prior_prob / (1 - prior_prob)
  if(post_prob >= 1){
    out$value    <- (n_samples - 1) / prior_odds
    out$operator <- ">"
  }else if(post_prob <= 0){
    out$value    <- 1 / ((n_samples - 1) * prior_odds)
    out$operator <- "<"
  }

  return(out)
}

.indicator_BF_warnings <- function(parameter, diagnostics, severe_limit = 20, warning_limit = 100){

  minority_visits <- min(diagnostics$visits, diagnostics$n_samples - diagnostics$visits)

  if(minority_visits == 0){
    return()
  }else if(minority_visits < severe_limit){
    return(stats::setNames(
      sprintf("Bayes factor MC error for %s is based on only %i posterior samples from the less frequent model.", parameter, minority_visits),
      parameter
    ))
  }else if(minority_visits < warning_limit){
    return(stats::setNames(
      sprintf("Bayes factor MC error for %s is based on %i posterior samples from the less frequent model.", parameter, minority_visits),
      parameter
    ))
  }else{
    return()
  }
}

.BF_error_column_name <- function(BF01 = FALSE){

  "error%(Inclusion BF)"
}

.standardize_BF_bound_operator <- function(operator, n){

  if(is.null(operator)){
    return(rep(NA_character_, n))
  }
  operator <- as.character(operator)
  if(length(operator) != n){
    operator <- rep_len(operator, n)
  }
  operator[!operator %in% c("<", ">")] <- NA_character_

  return(operator)
}

.invert_BF_bound_operator <- function(operator){

  operator <- as.character(operator)
  operator[operator == "<"] <- "TEMP_GT"
  operator[operator == ">"] <- "<"
  operator[operator == "TEMP_GT"] <- ">"

  return(operator)
}

.match_table_columns <- function(output_names, input_names){

  if(length(output_names) == 0){
    return(integer())
  }

  used <- rep(FALSE, length(input_names))
  vapply(output_names, function(output_name){
    matches <- which(input_names == output_name & !used)
    if(length(matches) == 0){
      return(NA_integer_)
    }
    used[matches[1]] <<- TRUE
    matches[1]
  }, integer(1))
}

.restore_table_column_attributes <- function(column, source_attributes){

  source_attributes <- source_attributes[!names(source_attributes) %in% c("names", "dim", "dimnames")]
  for(attribute in names(source_attributes)){
    attr(column, attribute) <- source_attributes[[attribute]]
  }

  column
}

.subset_table_parameters <- function(table, output){

  parameters <- attr(table, "parameters")
  if(is.null(parameters) || length(parameters) != nrow(table)){
    return(NULL)
  }

  row_indices <- match(rownames(output), rownames(table))
  if(any(is.na(row_indices))){
    return(NULL)
  }

  parameters[row_indices]
}

.subset_table_warnings <- function(warnings, selected_parameters = NULL, selected_rows = NULL){

  if(is.null(warnings) || length(warnings) == 0){
    return(warnings)
  }

  warning_names <- names(warnings)
  if(is.null(warning_names)){
    return(warnings)
  }
  warning_names[is.na(warning_names)] <- ""
  if(!any(nzchar(warning_names))){
    return(warnings)
  }

  selected <- unique(c(selected_parameters, selected_rows))
  if(length(selected) == 0){
    return(warnings[!nzchar(warning_names)])
  }

  keep <- !nzchar(warning_names) | warning_names %in% selected
  warnings[keep]
}

.format_BF_column <- function(x){

  out <- format(round(x, digits = 3), nsmall = 3)
  bound_operator <- .standardize_BF_bound_operator(attr(x, "bound_operator"), length(x))
  has_bound <- !is.na(bound_operator) & !is.na(x)
  if(any(has_bound)){
    out[has_bound] <- paste0(bound_operator[has_bound], trimws(out[has_bound]))
  }

  return(out)
}

.copy_BayesTools_table_attributes <- function(new_table, table, type){

  class(new_table)        <- class(table)
  attr(new_table, "type") <- type

  copied_attributes <- setdiff(
    names(attributes(table)),
    c("class", "type", "names", "dim", "dimnames")
  )
  for(a in copied_attributes){
    attr(new_table, a) <- attr(table, a)
  }

  new_table
}

#' @title Adds column to BayesTools table
#'
#' @description Adds column to a BayesTools table while not
#' breaking formatting, attributes, etc...
#'
#' @param table BayesTools table
#' @param column_title title of the new column
#' @param column_values values of the new column
#' @param column_position position of the new column (defaults to \code{NULL} which
#' appends the column to the end)
#' @param column_type type of values of the new column table (important for formatting,
#' defaults to \code{NULL} = the function tries to guess numeric / character based on the
#' \code{column_values} but many more specific types are available)
#'
#' @return returns an object of 'BayesTools_table' class.
#'
#' @export
add_column <- function(table, column_title, column_values, column_position = NULL, column_type = NULL){

  if(!inherits(table, "BayesTools_table"))
    stop("The 'table' must be of class 'BayesTools_table'.")
  check_char(column_title, "column_title")
  if(!(is.vector(column_values) | is.numeric(column_values)) || length(column_values) != nrow(table))
    stop("The 'column_values' must be a vector of the same length as has the table rows.")
  check_int(column_position, "column_position", allow_NULL = TRUE, lower = 0, upper = ncol(table) + 1)
  .check_table_types(column_type, "column_type", allow_NULL = TRUE)

  # fill defaults
  if(is.null(column_position)){
    column_position <- ncol(table) + 1
  }
  if(is.null(column_type)){
    if(all(.is.wholenumber(column_values))){
      column_type <- "integer"
    }else if(is.numeric(column_values)){
      column_type <- "estimate"
    }else if(is.character(column_values)){
      column_type <- "string"
    }else{
      stop("The 'column_type' could not be guessed. Please, supply it manually.")
    }
  }

  before <- if(column_position > 1) seq_len(column_position - 1) else integer(0)
  after  <- if(column_position <= ncol(table)) seq.int(max(column_position, 1L), ncol(table)) else integer(0)

  new_table <- table
  new_table[[ncol(table) + 1L]] <- column_values
  new_table <- new_table[, c(before, ncol(table) + 1L, after), drop = FALSE]

  names(new_table) <- c(
    if(length(before) > 0) colnames(table)[before],
    column_title,
    if(length(after) > 0) colnames(table)[after]
  )

  .copy_BayesTools_table_attributes(new_table, table, c(
    if(length(before) > 0) attr(table, "type")[before],
    column_type,
    if(length(after) > 0) attr(table, "type")[after]
  ))
}

#' @title Removes column to BayesTools table
#'
#' @description Removes column to a BayesTools table while not
#' breaking formatting, attributes, etc...
#'
#' @param table BayesTools table
#' @param column_position position of the to be removed column (defaults to \code{NULL} which
#' removes the last column)
#'
#' @return returns an object of 'BayesTools_table' class.
#'
#' @export
remove_column <- function(table, column_position = NULL){

  if(!inherits(table, "BayesTools_table"))
    stop("The 'table' must be of class 'BayesTools_table'.")
  check_int(column_position, "column_position", allow_NULL = TRUE, lower = 1, upper = ncol(table))

  # fill defaults
  if(is.null(column_position)){
    column_position <- ncol(table)
  }

  keep <- setdiff(seq_len(ncol(table)), column_position)
  new_table <- table[, keep, drop = FALSE]

  # transfer column names
  colnames(new_table) <- colnames(table)[keep]

  .copy_BayesTools_table_attributes(new_table, table, attr(table, "type")[keep])
}

#' @title Updates BayesTools table
#'
#' @description Updates BayesTools table while not breaking formatting, attributes, etc...
#'
#' @param object a BayesTools table
#' @param title title of the table
#' @param footnotes add footnotes to the table
#' @param warnings add warnings of the table
#' @param remove_parameters remove parameters from the table
#' @param logBF whether to format Bayes factors as log(BF)
#' @param BF01 whether to format Bayes factors as 1/BF
#' @param ... additional arguments.
#'
#' @return returns an object of 'BayesTools_table' class.
#' @export
update.BayesTools_table <- function(object, title = NULL, footnotes = NULL, warnings = NULL, remove_parameters = NULL, logBF = FALSE, BF01 = FALSE, ...){

  check_char(title, "title", allow_NULL = TRUE)
  check_char(footnotes, "footnotes", check_length = 0, allow_NULL = TRUE)
  check_char(warnings, "warnings", check_length = 0, allow_NULL = TRUE)
  check_char(remove_parameters, "remove_parameters", check_length = 0, allow_NULL = TRUE)
  check_bool(logBF, "logBF")
  check_bool(BF01,  "BF01")

  if(!is.null(footnotes)){
    attr(object, "footnotes") <- c(attr(object, "footnotes"), footnotes)
  }

  if(!is.null(warnings)){
    attr(object, "warnings")  <- c(attr(object, "warnings"), warnings)
  }

  if(!is.null(remove_parameters)){
    object <- object[!rownames(object) %in% remove_parameters,,drop=FALSE]
  }

  if(!is.null(title)){
    attr(object, "title") <- title
  }

  if(any(attr(object, "type") == "inclusion_BF")){
    for(BF_col in which(attr(object, "type") == "inclusion_BF")){
      BF_values <- object[[BF_col]]
      raw_BF    <- as.numeric(BF_values)
      bound_operator <- .standardize_BF_bound_operator(attr(BF_values, "bound_operator"), length(BF_values))
      if(isTRUE(attr(BF_values, "logBF"))){
        raw_BF <- exp(raw_BF)
      }
      if(isTRUE(attr(BF_values, "BF01"))){
        raw_BF <- 1 / raw_BF
        bound_operator <- .invert_BF_bound_operator(bound_operator)
      }
      attr(raw_BF, "bound_operator") <- bound_operator
      object[[BF_col]] <- format_BF(raw_BF, logBF = logBF, BF01 = BF01, inclusion = TRUE)
    }
  }
  if(any(attr(object, "type") == "BF_error")){
    BF_error_cols <- which(attr(object, "type") == "BF_error")
    if(length(BF_error_cols) == 1){
      attr(object[[BF_error_cols]], "name") <- .BF_error_column_name(BF01)
    }
  }

  return(object)
}

.format_column       <- function(x, type, n_models){
  if(is.null(x) || length(x) == 0){
    return(x)
  }else{
    return(switch(
      type,
      "integer"         = round(x),
      "prior"           = .center_priors(x),
      "string_left"     = .string_left(x),
      "string"          = x,
      "estimate"        = format(round(x, digits = 3), nsmall = 3),
      "prior_prob"      = format(round(x, digits = 3), nsmall = 3),
      "post_prob"       = format(round(x, digits = 3), nsmall = 3),
      "probability"     = format(round(x, digits = 3), nsmall = 3),
      "marglik"         = format(round(x, digits = 2), nsmall = 2),
      "BF"              = .format_BF_column(x),
      "inclusion_BF"    = .format_BF_column(x),
      "BF_error"        = format(round(x, digits = 3), nsmall = 3),
      "n_models"        = paste0(round(x), "/", n_models),
      "ESS"             = round(x),
      "R_hat"           = format(round(x, digits = 3), nsmall = 3),
      "MCMC_error"      = format(round(x, digits = 5), nsmall = 5),
      "MCMC_SD_error"   = format(round(x, digits = 3), nsmall = 3),
      "min_ESS"             = round(x),
      "max_R_hat"           = format(round(x, digits = 3), nsmall = 3),
      "max_MCMC_error"      = format(round(x, digits = 5), nsmall = 5),
      "max_MCMC_SD_error"   = format(round(x, digits = 3), nsmall = 3)
    ))
  }
}
.format_column_names <- function(x, type, values){
  if(is.null(x) || length(x) == 0){
    return(x)
  }else{
    return(switch(
      type,
      "integer"         = x,
      "prior"           = .string_center(paste0("Prior ", x), values),
      "string_left"     = .string_left(x, values),
      "string"          = x,
      "estimate"        = x,
      "probability"     = x,
      "prior_prob"      = "Prior prob.",
      "post_prob"       = "Post. prob.",
      "marglik"         = "log(marglik)",
      "BF"              = if(is.null(attr(values, "name"))) "BF"           else attr(values, "name"),
      "inclusion_BF"    = if(is.null(attr(values, "name"))) "Inclusion BF" else attr(values, "name"),
      "BF_error"        = if(is.null(attr(values, "name"))) x              else attr(values, "name"),
      "n_models"        = "Models",
      "ESS"             = "ESS",
      "R_hat"           = "R-hat",
      "MCMC_error"      = if(is.null(attr(values, "name"))) "error(MCMC)" else attr(values, "name"),
      "MCMC_SD_error"   = "error(MCMC)/SD",
      "min_ESS"             = "min(ESS)",
      "max_R_hat"           = "max(R-hat)",
      "max_MCMC_error"      = "max[error(MCMC)]",
      "max_MCMC_SD_error"   = "max[error(MCMC)/SD]",
    ))
  }
}
.center_priors <- function(x){

  if(any(grepl("~", x) | grepl("=", x))){

    position_tilda  <- regexpr("~", x)
    position_equal  <- regexpr("=", x)
    from_right      <- sapply(seq_along(x), function(i){
      if(position_tilda[i] != -1){
        return(nchar(x[i]) - position_tilda[i])
      }else if(position_equal[i] != -1){
        return(nchar(x[i]) - position_equal[i])
      }else{
        return(0)
      }
    })
    add_to_right  <- ifelse(from_right == 0, 0, max(from_right) - from_right)
    x <- paste0(x, sapply(seq_along(x), function(i)paste0(rep(" ", add_to_right[i]), collapse = "")))

  }

  return(x)
}
.string_left   <- function(x, reference = x){

  if(length(x) > 0){

    add_to_right <- max(nchar(reference)) - nchar(x)
    x <- paste0(x, sapply(seq_along(x), function(i)paste0(rep(" ", add_to_right[i]), collapse = "")))

  }

  return(x)
}
.string_center <- function(x, reference = x){

  if(length(x) > 0){

    add_to_sides <- max(nchar(reference)) - nchar(x)
    add_to_sides <- ifelse(add_to_sides < 0, 0, add_to_sides)
    x <- paste0( sapply(seq_along(x), function(i)paste0(rep(" ", round(add_to_sides[i]/2)), collapse = "")), x, sapply(seq_along(x), function(i)paste0(rep(" ", round(add_to_sides[i]/2)), collapse = "")))

  }

  return(x)
}

.check_table_types      <- function(x, name, allow_NULL = FALSE){
  check_char(x, name, allow_values = c(
    "integer", "prior", "string_left", "string", "estimate",
    "probability", "prior_prob", "post_prob",
    "marglik", "BF", "inclusion_BF", "BF_error", "n_models",
    "ESS", "R_hat", "MCMC_error", "MCMC_SD_error", "min_ESS", "max_R_hat", "max_MCMC_error", "max_MCMC_SD_error"),
    allow_NULL = allow_NULL)
}
.JAGS_estimates_diagnostic_columns <- function(){
  c("MCMC_error", "MCMC_SD_error", "ESS", "R_hat")
}
.JAGS_BF_diagnostic_columns <- function(){
  c("ESS", "MCMC_error", "BF_error_percent")
}
.JAGS_BF_diagnostic_column_types <- function(columns){
  unname(c(
    ESS              = "ESS",
    MCMC_error       = "MCMC_error",
    BF_error_percent = "BF_error"
  )[columns])
}
.normalize_diagnostic_columns <- function(columns, allowed_columns, name){

  if(is.null(columns) || length(columns) == 0){
    return(character())
  }

  if(is.logical(columns)){
    check_bool(columns, name, allow_NA = FALSE)
    if(columns){
      return(allowed_columns)
    }else{
      return(character())
    }
  }

  check_char(columns, name, check_length = 0, allow_NA = FALSE)

  shortcut_columns <- c("all", "none")
  if(any(columns %in% shortcut_columns)){
    if(length(columns) > 1){
      stop(paste0("The '", name, "' argument can use 'all' or 'none' only by itself."), call. = FALSE)
    }
    if(columns == "all"){
      return(allowed_columns)
    }else{
      return(character())
    }
  }

  check_char(columns, name, check_length = 0, allow_values = allowed_columns, allow_NA = FALSE)

  return(unique(columns))
}
.runjags_summary_fast   <- function(model_samples, n_samples, n_chains, conditional, probs = c(0.025, 0.975), remove_diagnostics = FALSE,
                                    diagnostic_columns = .JAGS_estimates_diagnostic_columns()){

  diagnostic_columns <- .normalize_diagnostic_columns(diagnostic_columns, .JAGS_estimates_diagnostic_columns(), "diagnostic_columns")
  if(remove_diagnostics){
    diagnostic_columns <- character()
  }

  # compute quantiles dynamically
  quantile_cols <- lapply(probs, function(p) apply(model_samples, 2, stats::quantile, probs = p, na.rm = TRUE))
  names(quantile_cols) <- as.character(probs)

  # the chains needs to be kept merged for conditional summary (due to NAs in the chains)
  runjags_summary <- cbind.data.frame(
    "Mean"   = apply(model_samples, 2, mean,          na.rm = TRUE),
    "SD"     = apply(model_samples, 2, stats::sd,    na.rm = TRUE),
    as.data.frame(quantile_cols, check.names = FALSE)
  )

  # remove all but Mean for inclusions
  quantile_col_names <- as.character(probs)
  runjags_summary[grepl("(inclusion)", rownames(runjags_summary)), c("SD", quantile_col_names)] <- NA

  # don't produce fit diagnostics for conditional samples (different chain lengths etc...) or if remove_diagnostics is TRUE
  if(conditional || length(diagnostic_columns) == 0){
    return(runjags_summary)
  }

  # split back the chains (allows for diagnostics calculation)
  model_samples_list <- split(as.data.frame(model_samples), rep(1:n_chains, each = n_samples), drop = FALSE)
  model_samples_list <- coda::as.mcmc.list(lapply(model_samples_list, coda::as.mcmc))
  mcmc_summary       <- summary(model_samples_list, quantiles = NULL)$statistics

  # fix single parameter summaries
  if(is.null(dim(mcmc_summary))){
    mcmc_summary <- t(mcmc_summary)
  }

  # add diagnostics to non-conditional samples
  runjags_diagnostics <- cbind.data.frame(
    "MCMC_error"    = mcmc_summary[,"Time-series SE"],
    "MCMC_SD_error" = mcmc_summary[,"Time-series SE"] / mcmc_summary[,"SD"],
    "ESS"           = coda::effectiveSize(model_samples_list),
    "R_hat"         = if(n_chains > 1) coda::gelman.diag(model_samples_list, multivariate = FALSE, autoburnin = FALSE)$psrf[,1] else NA
  )

  # remove incorrect NANs and NAs from the diagnostics
  runjags_diagnostics[is.nan(runjags_diagnostics[,"MCMC_SD_error"]),"MCMC_SD_error"] <- NA
  runjags_diagnostics[runjags_diagnostics[,"ESS"] == 0, "ESS"]                       <- 0
  runjags_diagnostics[is.nan(runjags_diagnostics[,"R_hat"]),"R_hat"]                 <- NA

  # first omega parameter is always constant
  runjags_diagnostics[grepl("omega[0,", rownames(runjags_diagnostics), fixed = TRUE), ] <- NA
  runjags_summary <- cbind.data.frame(runjags_summary, runjags_diagnostics[, diagnostic_columns, drop = FALSE])

  return(runjags_summary)
}
.runjags_conditional_warning <- function(parameters, n_samples, warning_limit = 500){
  if(n_samples == 0){
    return(sprintf("Conditional summary for %1$s parameter could not be computed due to no posterior samples.", paste0(parameters, collapse = ", ")))
  }else if(n_samples <= warning_limit){
    return(sprintf("Conditional summary for %1$s is based on %2$i samples.", paste0(parameters, collapse = ", "), n_samples))
  }else{
    return()
  }
}

# Helper function to transform scaled samples in list format (for ensemble/marginal tables)
# Uses the combinatorial unscaling algorithm via the helper in JAGS-formula.R
.transform_scale_samples_list <- function(samples, formula_scale){
  
  if(is.null(formula_scale) || length(formula_scale) == 0){
    return(samples)
  }
  
  sample_names <- names(samples)
  transformable <- sapply(samples, function(x) is.numeric(x) || is.matrix(x))
  transformable_names <- sample_names[transformable]
  
  if(length(transformable_names) == 0){
    return(samples)
  }

  n_samples <- if(is.matrix(samples[[transformable_names[1]]])){
    nrow(samples[[transformable_names[1]]])
  }else{
    length(samples[[transformable_names[1]]])
  }

  sample_columns <- lapply(transformable_names, function(name){
    if(is.matrix(samples[[name]])){
      temp_samples <- samples[[name]]
      if(is.null(colnames(temp_samples))){
        colnames(temp_samples) <- if(ncol(temp_samples) == 1) name else paste0(name, "[", seq_len(ncol(temp_samples)), "]")
      }
      return(temp_samples)
    }else{
      temp_samples <- matrix(samples[[name]], ncol = 1)
      colnames(temp_samples) <- name
      return(temp_samples)
    }
  })
  names(sample_columns) <- transformable_names

  posterior_matrix <- do.call(cbind, sample_columns)
  if(nrow(posterior_matrix) != n_samples){
    stop("All posterior sample elements must contain the same number of samples.", call. = FALSE)
  }

  posterior_matrix <- .apply_unscale_transform(posterior_matrix, formula_scale)

  for(name in transformable_names){
    old_attrs <- attributes(samples[[name]])
    column_names <- colnames(sample_columns[[name]])

    if(is.matrix(samples[[name]])){
      samples[[name]][, seq_along(column_names)] <- posterior_matrix[, column_names, drop = FALSE]
    }else{
      samples[[name]] <- posterior_matrix[, column_names]
      for(attr_name in setdiff(names(old_attrs), "names")){
        attr(samples[[name]], attr_name) <- old_attrs[[attr_name]]
      }
    }
  }
  
  return(samples)
}

args <- commandArgs(trailingOnly = TRUE)
profile <- if (length(args) > 0L && nzchar(args[[1L]])) args[[1L]] else Sys.getenv("BAYESTOOLS_TEST_PROFILE", "unit")

if (Sys.getenv("AGENT") == "") {
  Sys.setenv(AGENT = "1")
}

source(file.path("tests", "testthat", "helper-test-profiles.R"))

Sys.setenv(BAYESTOOLS_TEST_PROFILE = profile)

requested_profile_parts <- unique(unlist(strsplit(tolower(profile), "[,;[:space:]]+")))
selected_profiles <- bayestools_normalize_test_profiles(profile)
run_all_profiles <- "all" %in% requested_profile_parts

resolve_test_files_dir <- function() {
  test_files_dir <- Sys.getenv("BAYESTOOLS_TEST_FILES_DIR")
  if (test_files_dir == "") {
    test_files_dir <- file.path(tempdir(), "BayesTools_test_files")
  }
  if (!dir.exists(test_files_dir)) {
    dir.create(test_files_dir, showWarnings = FALSE, recursive = TRUE)
  }
  normalizePath(test_files_dir, winslash = "/", mustWork = TRUE)
}

Sys.setenv(BAYESTOOLS_TEST_FILES_DIR = resolve_test_files_dir())

if (any(c("fit", "visual-fixture") %in% selected_profiles)) {
  Sys.setenv(NOT_CRAN = "true")
}
if (any(c("visual", "visual-fixture") %in% selected_profiles)) {
  Sys.setenv(VDIFFR_RUN_TESTS = "true")
}
if ("fit" %in% selected_profiles && Sys.getenv("BAYESTOOLS_TEST_SKIP_REFIT") == "") {
  Sys.setenv(BAYESTOOLS_TEST_SKIP_REFIT = "false")
}

message("Running BayesTools test profile: ", Sys.getenv("BAYESTOOLS_TEST_PROFILE"))
message("Selected profile lane(s): ", paste(selected_profiles, collapse = ", "))
message("BAYESTOOLS_TEST_FILES_DIR: ", Sys.getenv("BAYESTOOLS_TEST_FILES_DIR"))
message("BAYESTOOLS_TEST_SKIP_REFIT: ", Sys.getenv("BAYESTOOLS_TEST_SKIP_REFIT", unset = "<unset>"))
message("NOT_CRAN: ", Sys.getenv("NOT_CRAN", unset = "<unset>"))
message("VDIFFR_RUN_TESTS: ", Sys.getenv("VDIFFR_RUN_TESTS", unset = "<unset>"))
message("AGENT: ", Sys.getenv("AGENT", unset = "<unset>"))

started <- Sys.time()
on.exit({
  elapsed <- difftime(Sys.time(), started, units = "secs")
  message("Elapsed seconds: ", round(as.numeric(elapsed), 1))
}, add = TRUE)

if (!requireNamespace("devtools", quietly = TRUE)) {
  stop("The devtools package is required to run tools/test-profile.R.", call. = FALSE)
}
if (!requireNamespace("testthat", quietly = TRUE)) {
  stop("The testthat package is required to run tools/test-profile.R.", call. = FALSE)
}
if (!exists("LlmReporter", envir = asNamespace("testthat"), inherits = FALSE)) {
  stop("testthat::LlmReporter is required; install testthat >= 3.3.0.", call. = FALSE)
}

test_filter <- if (run_all_profiles) NULL else bayestools_test_profile_filter(selected_profiles)

if (!is.null(test_filter)) {
  message("testthat filter selects ", length(bayestools_test_profile_selected_contexts(selected_profiles)), " context(s).")
}

message("testthat reporter: LlmReporter")
devtools::test(
  filter = test_filter,
  stop_on_failure = TRUE,
  reporter = testthat::LlmReporter$new()
)

#' @title Compute posterior probabilities and inclusion Bayes factors
#'
#' @description Computes prior probabilities, posterior probabilities,
#' and inclusion Bayes factors based either on (1) a list of models,
#' vector of parameters, and a list of indicators the models represent the
#' null or alternative hypothesis for each parameter, (2) on prior
#' model odds, marginal likelihoods, and indicator whether
#' the models represent the null or alternative hypothesis, or (3) list of
#' models for each model.
#'
#' @param model_list list of models, each of which contains marginal
#' likelihood estimated with bridge sampling \code{marglik} and prior model
#' odds \code{prior_weights}
#' @param parameters vector of parameters names for which inference should
#' be drawn
#' @param is_null_list list with entries for each parameter carrying either
#' logical vector of indicators specifying whether the model corresponds
#' to the null or alternative hypothesis (or an integer vector indexing models
#' corresponding to the null hypothesis)
#' @param prior_weights vector of prior model odds
#' @param margliks vector of marginal likelihoods
#' @param is_null logical vector of indicators specifying whether the model corresponds
#' to the null or alternative hypothesis (or an integer vector indexing models
#' corresponding to the null hypothesis)
#' @param conditional whether prior and posterior model probabilities should
#' be returned only for the conditional model. Defaults to \code{FALSE}
#'
#'
#' @return \code{compute_inference} returns a named list of prior probabilities,
#' posterior probabilities, and Bayes factors, \code{ppoint} gives the
#' distribution function, \code{ensemble_inference} gives a list of named lists of
#' inferences for each parameter, and \code{models_inference} returns a list of
#' models, each expanded by the inference list.
#'
#' @seealso [mix_posteriors] [BayesTools_ensemble_tables]
#'
#' @name ensemble_inference
#' @export compute_inference
#' @export ensemble_inference
#' @export models_inference
NULL

#' @rdname ensemble_inference
compute_inference <- function(prior_weights, margliks, is_null = NULL, conditional = FALSE){

  check_real(prior_weights, "prior_weights", lower = 0, check_length = 0)
  check_real(margliks,   "margliks", check_length = length(prior_weights))
  check_bool(conditional, "conditional")
  if(!is.null(is_null)){
    if(is.numeric(is_null)){
      check_int(is_null, "is_null", lower = 0, upper = length(prior_weights), check_length = 0)
      is_null <- c(1:length(prior_weights)) %in% is_null
    }else if(is.logical(is_null)){
      check_bool(is_null, "is_null", check_length = length(prior_weights))
    }else{
      stop("'is_null' argument must be either logical vector, integer vector, or NULL.")
    }
  }else{
    is_null <- rep(FALSE, length(prior_weights))
  }

  prior_probs <- .model_averaging_prior_probs(prior_weights)
  margliks    <- .model_averaging_margliks(margliks, prior_probs)
  post_probs  <- .model_averaging_post_probs(margliks, prior_probs)
  BF          <- inclusion_BF(prior_probs = prior_probs, margliks = margliks, is_null = is_null)

  if(conditional){
    if(all(is_null))
      stop("Conditional inference requires at least one non-null model.", call. = FALSE)
    prior_probs <- .model_averaging_prior_probs(ifelse(is_null, 0, prior_weights))
    margliks    <- .model_averaging_margliks(margliks, prior_probs)
    post_probs  <- .model_averaging_post_probs(margliks, prior_probs)
  }

  output <- list(
    prior_probs = prior_probs,
    post_probs  = post_probs,
    BF          = BF
  )

  attr(output, "is_null")     <- is_null
  attr(output, "conditional") <- conditional
  class(output) <- c(class(output), "inference")

  return(output)
}

.model_averaging_prior_probs <- function(prior_weights){

  if(any(!is.finite(prior_weights))){
    stop("'prior_weights' must be finite.", call. = FALSE)
  }
  if(sum(prior_weights) <= 0){
    stop("At least one prior model weight must be positive.", call. = FALSE)
  }

  prior_weights / sum(prior_weights)
}

.model_averaging_margliks <- function(margliks, prior_probs){

  if(any(is.infinite(margliks) & margliks > 0 & prior_probs > 0, na.rm = TRUE)){
    stop("Infinite positive marginal likelihoods are not supported.", call. = FALSE)
  }

  margliks[is.na(margliks)] <- -Inf

  if(!any(is.finite(margliks) & prior_probs > 0)){
    stop("No finite marginal likelihoods are available for models with positive prior probability.", call. = FALSE)
  }

  margliks
}

.model_averaging_post_probs <- function(margliks, prior_probs){

  unname(bridgesampling::post_prob(margliks, prior_prob = prior_probs))
}

.posterior_mixture_sample_counts <- function(post_probs, n_samples){

  .mixture_sample_counts(post_probs, n_samples)
}

.mixture_sample_counts <- function(probs, n_samples, preserve_positive = TRUE){

  probs <- as.numeric(probs)
  if(length(probs) == 0L){
    return(integer())
  }

  if(length(n_samples) != 1L || !is.finite(n_samples) || n_samples < 0 || !.is.wholenumber(n_samples)){
    stop("'n_samples' must be a non-negative integer.", call. = FALSE)
  }
  n_samples <- as.integer(n_samples)

  if(any(!is.finite(probs)) || any(probs < 0)){
    stop("'probs' must contain non-negative finite values.", call. = FALSE)
  }
  if(sum(probs) <= 0){
    stop("At least one mixture probability must be positive.", call. = FALSE)
  }

  probs <- probs / sum(probs)
  raw_counts <- probs * n_samples
  counts <- floor(raw_counts)

  remaining <- n_samples - sum(counts)
  if(remaining > 0L){
    add_order <- order(-(raw_counts - counts), -probs, seq_along(probs))
    counts[add_order[seq_len(remaining)]] <- counts[add_order[seq_len(remaining)]] + 1L
  }

  if(preserve_positive && n_samples > 0L){
    positive <- probs > 0
    missing <- which(positive & counts == 0L)
    if(length(missing) > 0L && sum(positive) <= n_samples){
      for(m in missing){
        donors <- which(counts > 1L)
        if(length(donors) == 0L){
          break
        }

        donor_cost <- abs((counts[donors] - 1L) - raw_counts[donors]) - abs(counts[donors] - raw_counts[donors])
        donor <- donors[order(donor_cost, -counts[donors], -probs[donors], donors)[1L]]
        counts[donor] <- counts[donor] - 1L
        counts[m] <- 1L
      }
    }
  }

  as.integer(counts)
}

#' @rdname ensemble_inference
ensemble_inference <- function(model_list, parameters, is_null_list, conditional = FALSE){

  # check input
  check_list(model_list, "model_list")
  check_char(parameters, "parameters", check_length = FALSE)
  check_list(is_null_list, "is_null_list", check_length = length(parameters))
  sapply(model_list, function(m)check_list(m, "model_list:model", check_names = c("marglik", "prior_weights"), all_objects = TRUE, allow_other = TRUE))
  if(!all(sapply(model_list, function(m)inherits(m[["marglik"]], what = "bridge"))))
    stop("model_list:marglik must contain 'bridgesampling' marginal likelihoods")
  sapply(model_list, function(m)check_real(m[["prior_weights"]], "model_list:prior_weights", lower = 0))


  # extract the object
  margliks      <- sapply(model_list, function(m) m[["marglik"]][["logml"]])
  prior_weights <- sapply(model_list, function(m) m[["prior_weights"]])

  out <- list()

  for(p in seq_along(parameters)){

    # prepare parameter specific values
    out[[parameters[p]]] <- compute_inference(prior_weights = prior_weights, margliks = margliks, is_null = is_null_list[[p]], conditional = conditional)

    # add parameter names
    parameter_name    <- parameters[p]
    formula_parameter <- unique(unlist(lapply(model_list, function(m) attr(attr(m[["fit"]], "prior_list")[[parameters[p]]], "parameter"))))

    if(!is.null(unlist(formula_parameter))){
      parameter_name <- format_parameter_names(parameter_name, formula_parameters = formula_parameter, formula_prefix = TRUE)
      class(out[[parameters[p]]]) <- c(class(out[[parameters[p]]]), "inference.formula")
      attr(out[[parameters[p]]], "formula_parameter")  <- formula_parameter
    }
    attr(out[[parameters[p]]], "parameter_name")  <- parameter_name

  }

  attr(out, "conditional") <- conditional
  return(out)
}

#' @rdname ensemble_inference
models_inference   <- function(model_list){

  sapply(model_list, function(m)check_list(m, "model_list:model", check_names = c("marglik", "prior_weights"), all_objects = TRUE, allow_other = TRUE))
  if(!all(sapply(model_list, function(m)inherits(m[["marglik"]], what = "bridge"))))
    stop("model_list:marglik must contain 'bridgesampling' marginal likelihoods")
  sapply(model_list, function(m)check_real(m[["prior_weights"]], "model_list:prior_weights", lower = 0))

  margliks    <- sapply(model_list, function(model)model[["marglik"]][["logml"]])
  prior_weights  <- sapply(model_list, function(model)model[["prior_weights"]])
  prior_probs <- .model_averaging_prior_probs(prior_weights)
  margliks    <- .model_averaging_margliks(margliks, prior_probs)
  post_probs  <- .model_averaging_post_probs(margliks, prior_probs)
  incl_BF     <- sapply(seq_along(model_list), function(i){
    is_null <- rep(TRUE, length(model_list))
    is_null[i] <- FALSE
    return(inclusion_BF(prior_probs = prior_probs, margliks = margliks, is_null = is_null))
  })

  for(i in seq_along(model_list)){
    model_list[[i]][["inference"]] <- list(
      "m_number"     = i,
      "marglik"      = margliks[i],
      "prior_prob"   = prior_probs[i],
      "post_prob"    = post_probs[i],
      "inclusion_BF" = incl_BF[i]
    )
  }

  return(model_list)
}

#' @title Model-average posterior distributions
#'
#' @description Model-averages posterior distributions based on
#' a list of models, vector of parameters, and a list of
#' indicators of the null or alternative hypothesis models
#' for each parameter.
#'
#' @param seed integer specifying seed for sampling posteriors for
#' model averaging. Defaults to \code{NULL}.
#' @param n_samples number of samples to be drawn for the model-averaged
#' posterior distribution
#' @inheritParams ensemble_inference
#'
#' @return \code{mix_posteriors} returns a named list of mixed posterior
#' distributions (either a vector of matrix).
#'
#' @seealso [ensemble_inference] [BayesTools_ensemble_tables] [as_mixed_posteriors]
#'
#' @name mix_posteriors
#' @export
mix_posteriors <- function(model_list, parameters, is_null_list, conditional = FALSE, seed = NULL, n_samples = 10000){

  # check input
  check_list(model_list, "model_list")
  check_char(parameters, "parameters", check_length = FALSE)
  check_list(is_null_list, "is_null_list", check_length = length(parameters))
  check_real(seed, "seed", allow_NULL = TRUE)
  check_int(n_samples, "n_samples")
  sapply(model_list, function(m)check_list(m, "model_list:model", check_names = c("fit", "marglik", "prior_weights"), all_objects = TRUE, allow_other = TRUE))
  if(!all(sapply(model_list, function(m) inherits(m[["fit"]], what = "runjags")) | sapply(model_list, function(m)inherits(m[["fit"]], what = "stanfit")) | sapply(model_list, function(m)inherits(m[["fit"]], what = "null_model"))))
    stop("model_list:fit must contain 'runjags' or 'rstan' models")
  if(!all(sapply(model_list, function(m) inherits(m[["marglik"]], what = "bridge"))))
    stop("model_list:marglik must contain 'bridgesampling' marginal likelihoods")
  if(!all(unlist(sapply(model_list, function(m) sapply(attr(m[["fit"]], "prior_list"), function(p) is.prior(p))))))
    stop("model_list:priors must contain 'BayesTools' priors")
  sapply(model_list, function(m) check_real(m[["prior_weights"]], "model_list:prior_weights", lower = 0))


  # extract the object
  fits           <- lapply(model_list, function(m) m[["fit"]])
  margliks       <- sapply(model_list, function(m) m[["marglik"]][["logml"]])
  priors         <- lapply(model_list, function(m) attr(m[["fit"]], "prior_list"))
  formula_priors <- lapply(model_list, function(m) m[["formula_priors"]])
  prior_weights  <- sapply(model_list, function(m) m[["prior_weights"]])

  inference  <- ensemble_inference(model_list, parameters, is_null_list, conditional)

  # Use one shared sampling seed so formula terms are drawn from aligned
  # posterior rows across parameters.
  if(!is.null(seed)){
    set.seed(seed)
    common_sample_seed <- seed
  }else{
    common_sample_seed <- sample(.Machine$integer.max, 1)
  }

  out <- list()

  for(p in seq_along(parameters)){

    # prepare parameter specific values
    temp_parameter    <- parameters[p]
    temp_inference    <- inference[[temp_parameter]]
    temp_priors       <- lapply(priors, function(p) p[[temp_parameter]])

    if(any(sapply(temp_priors, is.prior.weightfunction)) && all(sapply(temp_priors, is.prior.weightfunction) | sapply(temp_priors, .is_prior_weightfunction_null) | sapply(temp_priors, is.null))){
      # weightfunctions:

      # replace missing priors with default prior: none
      for(i in 1:length(fits)){
        if(is.null(temp_priors[[i]])){
          temp_priors[[i]] <- prior_none()
        }
      }

      # replace prior odds with the corresponding prior model odds
      for(i in seq_along(temp_priors)){
        temp_priors[[i]] <- .set_prior_model_weight(temp_priors[[i]], temp_inference$prior_probs[i])
      }

      out[[temp_parameter]] <- .mix_posteriors.weightfunction(fits, temp_priors, temp_parameter, temp_inference$post_probs, common_sample_seed, n_samples)

    }else if(any(sapply(temp_priors, is.prior.factor)) && all(sapply(temp_priors, is.prior.factor) | sapply(temp_priors, is.prior.point) | sapply(temp_priors, is.null))){
      # factor priors

      # replace missing priors with default prior: spike(0)
      for(i in 1:length(fits)){
        if(is.null(temp_priors[[i]])){
          temp_priors[[i]] <- prior("spike", parameters = list("location" = 0))
        }
      }

      # replace prior odds with the corresponding prior model odds
      for(i in seq_along(temp_priors)){
        temp_priors[[i]] <- .set_prior_model_weight(temp_priors[[i]], temp_inference$prior_probs[i])
      }

      out[[temp_parameter]] <- .mix_posteriors.factor(fits, temp_priors, temp_parameter, temp_inference$post_probs, common_sample_seed, n_samples)

    }else if(any(sapply(temp_priors, is.prior.vector)) && all(sapply(temp_priors, is.prior.vector) | sapply(temp_priors, is.prior.point) | sapply(temp_priors, is.null))){
      # vector priors:

      # replace missing priors with default prior: spike(0)
      for(i in 1:length(fits)){
        if(is.null(temp_priors[[i]])){
          temp_priors[[i]] <- prior("spike", parameters = list("location" = 0))
        }
      }

      # replace prior odds with the corresponding prior model odds
      for(i in seq_along(temp_priors)){
        temp_priors[[i]] <- .set_prior_model_weight(temp_priors[[i]], temp_inference$prior_probs[i])
      }

      out[[temp_parameter]] <- .mix_posteriors.vector(fits, temp_priors, temp_parameter, temp_inference$post_probs, common_sample_seed, n_samples)

    }else if(all(sapply(temp_priors, is.prior.simple) | sapply(temp_priors, is.prior.point) | sapply(temp_priors, is.null))){
      # simple priors:

      # replace missing priors with default prior: spike(0)
      for(i in 1:length(fits)){
        if(is.null(temp_priors[[i]])){
          temp_priors[[i]] <- prior("spike", parameters = list("location" = 0))
        }
      }

      # replace prior odds with the corresponding prior model odds
      for(i in seq_along(temp_priors)){
        temp_priors[[i]] <- .set_prior_model_weight(temp_priors[[i]], temp_inference$prior_probs[i])
      }

      out[[temp_parameter]] <- .mix_posteriors.simple(fits, temp_priors, temp_parameter, temp_inference$post_probs, common_sample_seed, n_samples)

    }else{
      stop("The posterior samples cannot be mixed: unsupported mixture of prior distributions.")
    }

    # add formula relevant information
    if(!is.null(unique(unlist(lapply(temp_priors, attr, which = "parameter"))))){
      class(out[[temp_parameter]]) <- c(class(out[[temp_parameter]]), "mixed_posteriors.formula")
      attr(out[[temp_parameter]], "formula_parameter")  <- unique(unlist(lapply(temp_priors, attr, which = "parameter")))
    }

  }

  class(out) <- c(class(out), "mixed_posteriors")
  return(out)
}

.mix_posteriors.simple         <- function(fits, priors, parameter, post_probs, seed = NULL, n_samples = 10000){

  # check input
  check_list(fits, "fits")
  check_list(priors, "priors", check_length = length(fits))
  check_char(parameter, "parameter")
  check_real(post_probs, "post_probs", lower = 0, upper = 1, check_length = length(fits))
  check_real(seed, "seed", allow_NULL = TRUE)
  check_int(n_samples, "n_samples")
  if(!all(sapply(fits, inherits, what = "runjags") | sapply(fits, inherits, what = "stanfit") | sapply(fits, inherits, what = "null_model")))
    stop("'fits' must be a list of 'runjags' or 'rstan' models")
  if(!all(sapply(priors, is.prior.simple) | sapply(priors, is.prior.point)))
    stop("'priors' must be a list of simple priors")

  # gather and check compatibility of prior distributions
  priors_info <- lapply(priors, function(p){
    if(is.prior.point(p) | is.prior.none(p)){
      return(FALSE)
    }else{
      list(
        "interaction"       = .is_prior_interaction(p),
        "interaction_terms" = attr(p, "interaction_terms")
      )
    }
  })
  priors_info <- priors_info[!sapply(priors_info, isFALSE)]
  if(length(priors_info) >= 2 && any(!unlist(lapply(priors_info, function(i) all.equal(i, priors_info[[1]]))))){
    stop("non-matching prior factor type specifications")
  }else if(length(priors_info) != 0){
    priors_info <- priors_info[[1]]
  }



  # When seed is NULL, continue using the current RNG state so mix_posteriors()
  # can coordinate draws across parameters.
  if(!is.null(seed)){
    set.seed(seed)
  }

  # prepare output objects
  samples <- NULL
  sample_ind <- NULL
  models_ind <- NULL

  # mix samples
  sample_counts <- .posterior_mixture_sample_counts(post_probs, n_samples)
  for(i in seq_along(fits)[sample_counts > 0]){

    # obtain posterior samples
    if(inherits(fits[[i]], "null_model")){
      # deal with a possibility of completely null model
      model_samples <- matrix()
    }else if(inherits(fits[[i]], "runjags")){
      model_samples <- .extract_posterior_samples(fits[[i]], as_list = FALSE)
      if(!is.matrix(model_samples)){
        # deal with automatic coercion into a vector in case of a single predictor
        model_samples <- matrix(model_samples, ncol = 1)
        colnames(model_samples) <- fits[[i]]$monitor
      }
    }else if(inherits(fits[[i]], "stanfit")){
      .check_rstan()
      model_samples <- .extract_stan(fits[[i]])
    }

    # sample indexes
    temp_ind <- sample(nrow(model_samples), sample_counts[i], replace = TRUE)

    if(is.prior.point(priors[[i]])){
      # not sampling the priors as the samples would be already transformed
      samples <- c(samples, rep(priors[[i]]$parameters[["location"]], length(temp_ind)))
    }else{
      samples <- c(samples, model_samples[temp_ind, parameter])
    }

    sample_ind <- c(sample_ind, temp_ind)
    models_ind <- c(models_ind, rep(i, length(temp_ind)))
  }

  samples <- unname(samples)
  attr(samples, "sample_ind") <- sample_ind
  attr(samples, "models_ind") <- models_ind
  attr(samples, "parameter")  <- parameter
  attr(samples, "prior_list") <- priors
  attr(samples, "interaction")       <- if(length(priors_info) == 0) FALSE else priors_info[["interaction"]]
  attr(samples, "interaction_terms") <- priors_info[["interaction_terms"]]
  class(samples) <- c("mixed_posteriors", "mixed_posteriors.simple")

  return(samples)
}
.mix_posteriors.vector         <- function(fits, priors, parameter, post_probs, seed = NULL, n_samples = 10000){

  # check input
  check_list(fits, "fits")
  check_list(priors, "priors", check_length = length(fits))
  check_char(parameter, "parameter")
  check_real(post_probs, "post_probs", lower = 0, upper = 1, check_length = length(fits))
  check_real(seed, "seed", allow_NULL = TRUE)
  check_int(n_samples, "n_samples")
  if(!all(sapply(fits, inherits, what = "runjags") | sapply(fits, inherits, what = "stanfit") | sapply(fits, inherits, what = "null_model")))
    stop("'fits' must be a list of 'runjags' or 'rstan' models")
  if(!all(sapply(priors, is.prior.vector) | sapply(priors, is.prior.point)))
    stop("'priors' must be a list of vector priors")

  # When seed is NULL, continue using the current RNG state so mix_posteriors()
  # can coordinate draws across parameters.
  if(!is.null(seed)){
    set.seed(seed)
  }

  # prepare output objects
  K <- unique(sapply(priors[sapply(priors, is.prior.vector)], function(p) p$parameters[["K"]]))
  if(length(K) != 1)
    stop("all vector priors must be of the same length")

  samples    <- matrix(nrow = 0, ncol = K)
  sample_ind <- NULL
  models_ind <- NULL

  # mix samples
  sample_counts <- .posterior_mixture_sample_counts(post_probs, n_samples)
  for(i in seq_along(fits)[sample_counts > 0]){

    # obtain posterior samples
    if(inherits(fits[[i]], "null_model")){
      # deal with a possibility of completely null model
      model_samples <- matrix()
    }else if(inherits(fits[[i]], "runjags")){
      model_samples <- .extract_posterior_samples(fits[[i]], as_list = FALSE)
      if(!is.matrix(model_samples)){
        # deal with automatic coercion into a vector in case of a single predictor
        model_samples <- matrix(model_samples, ncol = 1)
        colnames(model_samples) <- fits[[i]]$monitor
      }
    }else if(inherits(fits[[i]], "stanfit")){
      .check_rstan()
      model_samples <- .extract_stan(fits[[i]])
    }

    # sample indexes
    temp_ind <- sample(nrow(model_samples), sample_counts[i], replace = TRUE)

    if(is.prior.point(priors[[i]])){
      # not sampling the priors as the samples would be already transformed
      samples <- rbind(samples, matrix(priors[[i]]$parameters[["location"]], nrow = length(temp_ind), ncol = K))
    }else if(K == 1){
      samples <- rbind(samples, matrix(model_samples[temp_ind, parameter], nrow = length(temp_ind), ncol = K))
    }else{
      samples <- rbind(samples, model_samples[temp_ind, paste0(parameter,"[",1:K,"]")])
    }

    sample_ind <- c(sample_ind, temp_ind)
    models_ind <- c(models_ind, rep(i, length(temp_ind)))
  }

  rownames(samples) <- NULL
  colnames(samples) <- paste0(parameter,"[",1:K,"]")
  attr(samples, "sample_ind") <- sample_ind
  attr(samples, "models_ind") <- models_ind
  attr(samples, "parameter")  <- parameter
  attr(samples, "prior_list") <- priors
  class(samples) <- c("mixed_posteriors", "mixed_posteriors.vector")

  return(samples)
}
.mix_posteriors.factor         <- function(fits, priors, parameter, post_probs, seed = NULL, n_samples = 10000){

  # check input
  check_list(fits, "fits")
  check_list(priors, "priors", check_length = length(fits))
  check_char(parameter, "parameter")
  check_real(post_probs, "post_probs", lower = 0, upper = 1, check_length = length(fits))
  check_real(seed, "seed", allow_NULL = TRUE)
  check_int(n_samples, "n_samples")
  if(!all(sapply(fits, inherits, what = "runjags") | sapply(fits, inherits, what = "stanfit") | sapply(fits, inherits, what = "null_model")))
    stop("'fits' must be a list of 'runjags' or 'rstan' models")
  if(!all(sapply(priors, is.prior.factor) | sapply(priors, is.prior.point)))
    stop("'priors' must be a list of factor priors")

  # check the prior levels
  levels <- unique(sapply(priors[sapply(priors, is.prior.factor)], .get_prior_factor_levels))
  if(length(levels) != 1)
    stop("all factor priors must be of the same number of levels")

  # gather and check compatibility of prior distributions
  priors_info <- lapply(priors, function(p){
    if(is.prior.point(p) | is.prior.none(p)){
      return(FALSE)
    }else if(is.prior.factor(p)){
      return(list(
        "levels"            = .get_prior_factor_levels(p),
        "level_names"       = .get_prior_factor_level_names(p),
        "interaction"       = .is_prior_interaction(p),
        "interaction_terms" = attr(p, "interaction_terms"),
        "term_components"   = attr(p, "term_components"),
        "factor_terms"      = attr(p, "factor_terms"),
        "factor_contrasts"  = attr(p, "factor_contrasts"),
        "factor_design"     = attr(p, "factor_design"),
        "factor_cell_names" = attr(p, "factor_cell_names"),
        "treatment"         = is.prior.treatment(p),
        "independent"       = is.prior.independent(p),
        "orthonormal"       = is.prior.orthonormal(p),
        "meandif"           = is.prior.meandif(p)
      ))
    }else{
      stop("unsupported prior type")
    }
  })
  priors_info <- priors_info[!sapply(priors_info, isFALSE)]
  if(length(priors_info) >= 2 && any(!unlist(lapply(priors_info, function(i) all.equal(i, priors_info[[1]]))))){
    stop("non-matching prior factor type specifications")
  }
  if(length(priors_info) != 0){
    priors_info <- priors_info[[1]]
  }


  if(priors_info[["treatment"]]){

    if(levels == 1){

      samples <- .mix_posteriors.simple(fits, priors, parameter, post_probs, seed, n_samples)

      sample_ind <- attr(samples, "sample_ind")
      models_ind <- attr(samples, "models_ind")

      samples <- matrix(samples, ncol = 1)

    }else{

      # keep the same seed across levels
      if(is.null(seed)){
        seed <- sample(666666, 1)
      }

      samples <- lapply(1:levels, function(i) .mix_posteriors.simple(fits, priors, paste0(parameter, "[", i, "]"), post_probs, seed, n_samples))

      sample_ind <- attr(samples[[1]], "sample_ind")
      models_ind <- attr(samples[[1]], "models_ind")

      samples <- do.call(cbind, samples)

    }

    rownames(samples) <- NULL
    colnames(samples) <- paste0(parameter,"[",priors_info$level_names[-1],"]")
    attr(samples, "sample_ind") <- sample_ind
    attr(samples, "models_ind") <- models_ind
    attr(samples, "parameter")  <- parameter
    attr(samples, "prior_list") <- priors
    class(samples) <- c("mixed_posteriors", "mixed_posteriors.factor", "mixed_posteriors.vector")

  }else if(priors_info[["independent"]]){

    if(levels == 1){

      samples <- .mix_posteriors.simple(fits, priors, parameter, post_probs, seed, n_samples)

      sample_ind <- attr(samples, "sample_ind")
      models_ind <- attr(samples, "models_ind")

      samples <- matrix(samples, ncol = 1)

    }else{

      # keep the same seed across levels
      if(is.null(seed)){
        seed <- sample(666666, 1)
      }

      samples <- lapply(1:levels, function(i) .mix_posteriors.simple(fits, priors, paste0(parameter, "[", i, "]"), post_probs, seed, n_samples))

      sample_ind <- attr(samples[[1]], "sample_ind")
      models_ind <- attr(samples[[1]], "models_ind")

      samples <- do.call(cbind, samples)

    }

    rownames(samples) <- NULL
    colnames(samples) <- paste0(parameter,"[",priors_info$level_names,"]")
    attr(samples, "sample_ind") <- sample_ind
    attr(samples, "models_ind") <- models_ind
    attr(samples, "parameter")  <- parameter
    attr(samples, "prior_list") <- priors
    class(samples) <- c("mixed_posteriors", "mixed_posteriors.factor", "mixed_posteriors.vector")

  }else if(priors_info[["orthonormal"]] | priors_info[["meandif"]]){

    for(i in seq_along(priors)){
      if(is.prior.factor(priors[[i]])){
        priors[[i]]$parameters[["K"]] <- levels
      }
    }

    samples <- .mix_posteriors.vector(fits, priors, parameter, post_probs, seed, n_samples)
    class(samples) <- c(class(samples), "mixed_posteriors.factor")

  }

  attr(samples, "levels")            <- priors_info[["levels"]]
  attr(samples, "level_names")       <- priors_info[["level_names"]]
  attr(samples, "interaction")       <- if(length(priors_info) == 0) FALSE else priors_info[["interaction"]]
  attr(samples, "interaction_terms") <- priors_info[["interaction_terms"]]
  attr(samples, "term_components")   <- priors_info[["term_components"]]
  attr(samples, "factor_terms")      <- priors_info[["factor_terms"]]
  attr(samples, "factor_contrasts")  <- priors_info[["factor_contrasts"]]
  attr(samples, "factor_design")     <- priors_info[["factor_design"]]
  attr(samples, "factor_cell_names") <- priors_info[["factor_cell_names"]]
  attr(samples, "treatment")         <- priors_info[["treatment"]]
  attr(samples, "independent")       <- priors_info[["independent"]]
  attr(samples, "orthonormal")       <- priors_info[["orthonormal"]]
  attr(samples, "meandif")           <- priors_info[["meandif"]]

  return(samples)
}
.mix_posteriors.weightfunction <- function(fits, priors, parameter, post_probs, seed = NULL, n_samples = 10000){

  # check input
  check_list(fits, "fits")
  check_list(priors, "priors", check_length = length(fits))
  check_char(parameter, "parameter")
  check_real(post_probs, "post_probs", lower = 0, upper = 1, check_length = length(fits))
  check_real(seed, "seed", allow_NULL = TRUE)
  check_int(n_samples, "n_samples")
  if(!all(sapply(fits, inherits, what = "runjags") | sapply(fits, inherits, what = "stanfit") | sapply(fits, inherits, what = "null_model")))
    stop("'fits' must be a list of 'runjags' or 'rstan' models")
  if(!all(sapply(priors, is.prior.weightfunction) | sapply(priors, .is_prior_weightfunction_null)))
    stop("'priors' must be a list of weightfunction priors or point(1)/none null priors")


  # When seed is NULL, continue using the current RNG state so mix_posteriors()
  # can coordinate draws across parameters.
  if(!is.null(seed)){
    set.seed(seed)
  }

  # obtain mapping for the weight coefficients
  omega_mapping <- weightfunctions_mapping(priors)
  omega_cuts    <- weightfunctions_mapping(priors, cuts_only = TRUE)
  omega_names   <- sapply(1:(length(omega_cuts)-1), function(i)paste0("omega[",omega_cuts[i],",",omega_cuts[i+1],"]"))

  # prepare output objects
  samples    <- matrix(nrow = 0, ncol = length(omega_cuts) - 1)
  sample_ind <- NULL
  models_ind <- NULL

  # mix samples
  sample_counts <- .posterior_mixture_sample_counts(post_probs, n_samples)
  for(i in seq_along(fits)[sample_counts > 0]){

    # obtain posterior samples
    if(inherits(fits[[i]], "null_model")){
      # deal with a possibility of completely null model
      model_samples <- matrix()
    }else if(inherits(fits[[i]], "runjags")){
      model_samples <- .extract_posterior_samples(fits[[i]], as_list = FALSE)
      if(!is.matrix(model_samples)){
        # deal with automatic coercion into a vector in case of a single predictor
        model_samples <- matrix(model_samples, ncol = 1)
        colnames(model_samples) <- fits[[i]]$monitor
      }
    }else if(inherits(fits[[i]], "stanfit")){
      .check_rstan()
      model_samples <- .extract_stan(fits[[i]])
    }

    # sample indexes
    temp_ind <- sample(nrow(model_samples), sample_counts[i], replace = TRUE)

    if(.is_prior_weightfunction_null(priors[[i]])){
      samples <- rbind(samples, matrix(1, ncol = length(omega_cuts) - 1, nrow = length(temp_ind)))
    }else{
      samples <- rbind(samples, model_samples[temp_ind, paste0("omega[",omega_mapping[[i]],"]")])
    }

    sample_ind <- c(sample_ind, temp_ind)
    models_ind <- c(models_ind, rep(i, length(temp_ind)))
  }

  rownames(samples) <- NULL
  colnames(samples) <- omega_names
  attr(samples, "sample_ind") <- sample_ind
  attr(samples, "models_ind") <- models_ind
  attr(samples, "parameter")  <- parameter
  attr(samples, "prior_list") <- priors
  class(samples) <- c("mixed_posteriors", "mixed_posteriors.weightfunction")

  return(samples)
}

#' @title Export BayesTools JAGS model posterior distribution as model-average posterior distributions via \code{mix_posteriors}
#'
#' @description Creates a model-averages posterior distributions on a single
#' model that allows mimicking the [mix_posteriors] functionality. This function
#' is useful when the model-averaged ensemble is based on [prior_spike_and_slab]
#' or [prior_mixture] priors - the model-averaging is done within the model.
#'
#' @param model model fit via the [JAGS_fit] function
#' @param conditional a character vector of parameters to be conditioned on
#' @param conditional_rule a character string specifying the rule for conditioning.
#' Either "AND" or "OR". Defaults to "AND".
#' @param force_plots temporal argument allowing to generate conditional posterior samples
#' suitable for prior and posterior plots. Only available when conditioning on a
#' single parameter.
#' @param transform_scaled whether to transform samples from standardized (scaled) to
#' original (unscaled) scale. When \code{TRUE}, posterior samples are
#' transformed, and the result can be directly passed to [plot_posterior] which will
#' automatically detect the transformation and use transformed deterministic prior densities.
#' Requires a model fitted with \code{formula_scale_list}. Defaults to \code{FALSE}.
#' @param n_prior_samples controls the numerical grid used for transformed
#' prior densities when \code{transform_scaled = TRUE}. Defaults to 10000.
#' @inheritParams ensemble_inference
#'
#' @return \code{as_mix_posteriors} returns a named list of mixed posterior
#' distributions (either a vector of matrix).
#'
#' @seealso [mix_posteriors]
#'
#' @name as_mixed_posteriors
#' @export
as_mixed_posteriors <- function(model, parameters, conditional = NULL, conditional_rule = "AND", force_plots = FALSE,
                                 transform_scaled = FALSE, n_prior_samples = 10000){

  # check input
  if(!inherits(model, "BayesTools_fit"))
    stop("'model' must be a 'BayesTools_fit'")
  check_char(parameters, "parameters", check_length = FALSE)
  check_char(conditional, "conditional", check_length = FALSE, allow_values = c(parameters, "PET", "PEESE", "PETPEESE", "omega", "phacking", "alpha", "pi_null"), allow_NULL = TRUE)
  check_char(conditional_rule, "conditional_rule", allow_values = c("AND", "OR"))
  check_bool(transform_scaled, "transform_scaled")
  check_int(n_prior_samples, "n_prior_samples", lower = 1)

  # extract the list of priors
  priors <- attr(model, "prior_list")
  prior_density_priors <- priors

  # extract the samples
  model_samples <- suppressWarnings(coda::as.mcmc(model))
  if(!is.matrix(model_samples)){
    # deal with automatic coercion into a vector in case of a single predictor
    model_samples <- matrix(model_samples, ncol = 1)
    colnames(model_samples) <- model$monitor
  }

  # apply conditioning
  if(length(conditional) > 0){

    # subset the posterior distribution
    conditioning_samples <- do.call(cbind, lapply(conditional, function(parameter){

      # special cases for PET / PEESE / PET-PEESE / selection / p-hacking within the bias parameter
      if (!is.null(priors[["bias"]]) && is.prior.mixture(priors[["bias"]])) {
        bias_branch_info <- .selection_prior_branch_info(priors[["bias"]])
        if(parameter == "PET"){
          is_PET <- sapply(priors[["bias"]], is.prior.PET)
          return(model_samples[, "bias_indicator"] %in% which(is_PET))
        }
        if(parameter == "PEESE"){
          is_PEESE <- sapply(priors[["bias"]], is.prior.PEESE)
          return(model_samples[, "bias_indicator"] %in% which(is_PEESE))
        }
        if(parameter == "PETPEESE"){
          is_PET   <- sapply(priors[["bias"]], is.prior.PET)
          is_PEESE <- sapply(priors[["bias"]], is.prior.PEESE)
          return(model_samples[, "bias_indicator"] %in% which(is_PET | is_PEESE))
        }
        if(parameter == "omega"){
          has_selection <- vapply(bias_branch_info, function(x) !is.null(x$selection), logical(1))
          return(model_samples[, "bias_indicator"] %in% which(has_selection))
        }
        if(parameter %in% c("phacking", "alpha", "pi_null")){
          has_phacking <- vapply(bias_branch_info, function(x) !is.null(x$phacking), logical(1))
          return(model_samples[, "bias_indicator"] %in% which(has_phacking))
        }
      }

      if(parameter == "omega" && any(vapply(priors, function(x){
        is.prior.weightfunction(x) || (is_prior_bias(x) && !is.null(x$selection))
      }, logical(1)))){
        return(rep(TRUE, nrow(model_samples)))
      }

      if(parameter %in% c("phacking", "alpha", "pi_null") && any(vapply(priors, function(x){
        is_prior_phacking(x) || (is_prior_bias(x) && !is.null(x$phacking))
      }, logical(1)))){
        return(rep(TRUE, nrow(model_samples)))
      }

      # normal cases
      temp_prior <- priors[[parameter]]

      if(is.prior.spike_and_slab(temp_prior)){

        return(model_samples[,paste0(parameter, "_indicator")] == 1)

      }else if(is.prior.mixture(temp_prior)){

        components <- attr(temp_prior, "components")
        if(!all(components %in% c("null", "alternative")))
          stop("conditional mixture posterior distributions are available only for 'null' and 'alternative' components")

        return(model_samples[,paste0(parameter, "_indicator")] %in% which(components == "alternative"))

      }else{

        warning(sprintf("The parameter '%s' is not a conditional parameter. All samples are assumed to compe from the conditional posterior distribution.", parameter), call. = FALSE, immediate. = TRUE)
        return(rep(TRUE, nrow(model_samples)))

      }
    }))
    conditioning_samples <- apply(conditioning_samples, 1, ifelse(conditional_rule == "AND", all, any))

    if(sum(conditioning_samples) == 0){
      warning("No samples left after conditioning.", call. = FALSE, immediate. = TRUE)
      return(list())
    }


    model_samples <- model_samples[conditioning_samples,,drop=FALSE]

    # set prior weights to 0 for null distributions
    if(length(conditional) == 1){

      if (conditional %in% c("bias", "PET", "PEESE", "PETPEESE", "omega", "phacking", "alpha", "pi_null") && !is.null(priors[["bias"]]) && is.prior.mixture(priors[["bias"]])) {

        # special cases for PET / PEESE / PET-PEESE / selection / p-hacking
        bias_branch_info <- .selection_prior_branch_info(priors[["bias"]])
        if(conditional == "bias"){
          components <- attr(priors[["bias"]], "components")
          if(is.null(components) || length(components) != length(priors[["bias"]])){
            is_null <- sapply(priors[["bias"]], is.prior.none)
          }else{
            is_null <- components == "null"
          }
          for(i in seq(along = is_null)){
            if(is_null[i]){
              priors[["bias"]][[i]][["prior_weights"]] <- 0
            }
          }
        }else if(conditional == "PET"){
          is_PET <- sapply(priors[["bias"]], is.prior.PET)
          for(i in seq(along = is_PET)){
            if(!is_PET[i]){
              priors[["bias"]][[i]][["prior_weights"]] <- 0
            }
          }
        }else if(conditional == "PEESE"){
          is_PEESE <- sapply(priors[["bias"]], is.prior.PEESE)
          for(i in seq(along = is_PEESE)){
            if(!is_PEESE[i]){
              priors[["bias"]][[i]][["prior_weights"]] <- 0
            }
          }
        }else if(conditional == "PETPEESE"){
          is_PET   <- sapply(priors[["bias"]], is.prior.PET)
          is_PEESE <- sapply(priors[["bias"]], is.prior.PEESE)
          for(i in seq(along = is_PET)){
            if(!(is_PET[i] || is_PEESE[i])){
              priors[["bias"]][[i]][["prior_weights"]] <- 0
            }
          }
        }else if(conditional == "omega"){
          has_selection <- vapply(bias_branch_info, function(x) !is.null(x$selection), logical(1))
          for(i in seq(along = has_selection)){
            if(!has_selection[i]){
              priors[["bias"]][[i]][["prior_weights"]] <- 0
            }
          }
        }else if(conditional %in% c("phacking", "alpha", "pi_null")){
          has_phacking <- vapply(bias_branch_info, function(x) !is.null(x$phacking), logical(1))
          for(i in seq(along = has_phacking)){
            if(!has_phacking[i]){
              priors[["bias"]][[i]][["prior_weights"]] <- 0
            }
          }
        }

        # propagate the prior weights to the mixture prior itself
        attr(priors[["bias"]], "prior_weights") <- sapply(priors[["bias"]], \(x) x[["prior_weights"]])

      }else if(is.prior.mixture(priors[[conditional]])){

        components <- attr(priors[[conditional]], "components")
        for(i in seq(along = components)){
          if(components[i] == "null"){
            priors[[conditional]][[i]][["prior_weights"]] <- 0
          }
        }

        # propagate the prior weights to the mixture prior itself
        attr(priors[[conditional]], "prior_weights") <- sapply(priors[[conditional]], \(x) x[["prior_weights"]])
      }

    }
  }

  # extract formula_scale early for transform_scaled support
  formula_scale <- attr(model, "formula_scale")

  # apply scale transformation to posterior samples if requested
  if(transform_scaled && !is.null(formula_scale) && length(formula_scale) > 0){
    model_samples <- transform_scale_samples(model_samples, formula_scale)
  }

  out    <- list()

  for(p in seq_along(parameters)){

    # prepare parameter specific values
    temp_parameter <- parameters[p]
    temp_prior     <- priors[[temp_parameter]]

    if(is.prior.spike_and_slab(temp_prior)){
      # spike and slab priors
      out[[temp_parameter]] <- .as_mixed_posteriors.spike_and_slab(model_samples, temp_prior, temp_parameter)

    }else if(is.prior.mixture(temp_prior)){
      # mixture priors
      out[[temp_parameter]] <- .as_mixed_posteriors.mixture(model_samples, temp_prior, temp_parameter, conditional)

    }else if(is_prior_phacking(temp_prior)){
      # p-hacking priors
      out[[temp_parameter]] <- .as_mixed_posteriors.phacking(model_samples, temp_prior, temp_parameter)

    }else if(is_prior_bias(temp_prior)){
      # composed publication-bias priors
      out[[temp_parameter]] <- .as_mixed_posteriors.bias(model_samples, temp_prior, temp_parameter, conditional)

    }else if(is.prior.weightfunction(temp_prior)){
      # weight functions
      out[[temp_parameter]] <- .as_mixed_posteriors.weightfunction(model_samples, temp_prior, temp_parameter)

    }else if(is.prior.factor(temp_prior)){
      # factor priors
      out[[temp_parameter]] <- .as_mixed_posteriors.factor(model_samples, temp_prior, temp_parameter)

    }else if(is.prior.vector(temp_prior)){
      # vector priors
      out[[temp_parameter]] <- .as_mixed_posteriors.vector(model_samples, temp_prior, temp_parameter)

    }else if(is.prior.simple(temp_prior)){
      # simple priors
      out[[temp_parameter]] <- .as_mixed_posteriors.simple(model_samples, temp_prior, temp_parameter)

    }else{
      stop("The posterior samples cannot be mixed: unsupported prior distributions.")
    }

    # add formula relevant information
    if(!is.null(attr(temp_prior, which = "parameter"))){
      class(out[[temp_parameter]]) <- c(class(out[[temp_parameter]]), "mixed_posteriors.formula")
      attr(out[[temp_parameter]], "formula_parameter")  <- attr(temp_prior, which = "parameter")
    }

    # add conditioning information
    attr(out[[temp_parameter]], "conditional")      <- conditional
    attr(out[[temp_parameter]], "conditional_rule") <- conditional_rule

  }

  attr(out, "prior_list")       <- priors
  attr(out, "conditional")      <- conditional
  attr(out, "conditional_rule") <- conditional_rule

  # propagate formula_scale attribute for transform_scaled support
  if(!is.null(formula_scale)){
    attr(out, "formula_scale") <- formula_scale
  }

  prior_context_priors      <- prior_density_priors
  prior_context_conditional <- conditional
  special_conditionals      <- c("PET", "PEESE", "PETPEESE", "omega", "phacking", "alpha", "pi_null")
  if(length(conditional) == 1 &&
     conditional %in% special_conditionals &&
     !conditional %in% names(prior_density_priors)){
    prior_context_priors      <- priors
    prior_context_conditional <- NULL
  }

  # generate and store transformed prior densities if requested
  if(transform_scaled && !is.null(formula_scale) && length(formula_scale) > 0){
    prior_densities <- .generate_transformed_prior_densities(
      prior_list       = prior_context_priors,
      column_names     = colnames(model_samples),
      n_grid           = n_prior_samples,
      formula_scale    = formula_scale,
      conditional      = prior_context_conditional,
      conditional_rule = conditional_rule
    )
    attr(out, "prior_densities")       <- prior_densities
    attr(out, "prior_density_context") <- attr(prior_densities, "context")
    attr(out, "transform_scaled")      <- TRUE
  }else{
    attr(out, "prior_density_context") <- .prior_density_build_context(
      prior_list       = prior_context_priors,
      column_names     = colnames(model_samples),
      n_grid           = n_prior_samples,
      conditional      = prior_context_conditional,
      conditional_rule = conditional_rule
    )
  }

  class(out) <- c(class(out), "as_mixed_posteriors", "mixed_posteriors")
  return(out)
}

.as_mixed_posteriors.simple         <- function(model_samples, prior, parameter){

  # check input
  check_char(parameter, "parameter", check_length = FALSE)

  # gather information about the prior distribution
  prior_info <- list(
    "interaction"       = .is_prior_interaction(prior),
    "interaction_terms" = attr(prior, "interaction_terms")
  )

  # prepare output objects
  samples <- model_samples[, parameter]

  # format the output
  samples <- unname(samples)
  attr(samples, "sample_ind") <- FALSE
  attr(samples, "models_ind") <- rep(1, length(samples))
  attr(samples, "parameter")  <- parameter
  attr(samples, "prior_list") <- prior
  attr(samples, "interaction")       <- if(length(prior_info) == 0) FALSE else prior_info[["interaction"]]
  attr(samples, "interaction_terms") <- prior_info[["interaction_terms"]]
  class(samples) <- c("mixed_posteriors", "mixed_posteriors.simple")

  return(samples)
}
.as_mixed_posteriors.vector         <- function(model_samples, prior, parameter){

  # check input
  check_char(parameter, "parameter", check_length = FALSE)

  # gather information about the prior distribution
  K <- prior$parameter[["K"]]
  if(length(K) != 1)
    stop("all vector prior must be of the same length")

  # prepare output objects
  if(K == 1){
    samples <- model_samples[, parameter, drop = FALSE]
  }else{
    samples <- model_samples[, paste0(parameter,"[",1:K,"]"), drop = FALSE]
  }

  rownames(samples) <- NULL
  colnames(samples) <- paste0(parameter,"[",1:K,"]")
  attr(samples, "sample_ind") <- FALSE
  attr(samples, "models_ind") <- rep(1, nrow(samples))
  attr(samples, "parameter")  <- parameter
  attr(samples, "prior_list") <- prior
  class(samples) <- c("mixed_posteriors", "mixed_posteriors.vector")

  return(samples)
}
.as_mixed_posteriors.factor         <- function(model_samples, prior, parameter){

  # check input
  check_char(parameter, "parameter", check_length = FALSE)

  # gather information about the prior distribution
  prior_info <- list(
    "levels"            = .get_prior_factor_levels(prior),
    "level_names"       = .get_prior_factor_level_names(prior),
    "interaction"       = .is_prior_interaction(prior),
    "interaction_terms" = attr(prior, "interaction_terms"),
    "term_components"   = attr(prior, "term_components"),
    "factor_terms"      = attr(prior, "factor_terms"),
    "factor_contrasts"  = attr(prior, "factor_contrasts"),
    "factor_design"     = attr(prior, "factor_design"),
    "factor_cell_names" = attr(prior, "factor_cell_names"),
    "treatment"         = is.prior.treatment(prior),
    "independent"       = is.prior.independent(prior),
    "orthonormal"       = is.prior.orthonormal(prior),
    "meandif"           = is.prior.meandif(prior)
  )


  if(prior_info[["treatment"]]){

    if(prior_info[["levels"]] == 1){

      samples <- .as_mixed_posteriors.simple(model_samples, prior, parameter)
      samples <- matrix(samples, ncol = 1)

    }else{

      samples <- lapply(1:prior_info[["levels"]], function(i) .as_mixed_posteriors.simple(model_samples, prior, paste0(parameter, "[", i, "]")))
      samples <- do.call(cbind, samples)

    }

    level_names <- prior_info[["level_names"]]
    if(is.list(level_names)){
      level_names <- lapply(level_names, function(x) x[-1])
    }else{
      level_names <- level_names[-1]
    }

    rownames(samples) <- NULL
    colnames(samples) <- .format_factor_level_parameter_names(parameter, level_names, ncol(samples))
    attr(samples, "sample_ind") <- FALSE
    attr(samples, "models_ind") <- rep(1, nrow(samples))
    attr(samples, "parameter")  <- parameter
    attr(samples, "prior_list") <- prior
    class(samples) <- c("mixed_posteriors", "mixed_posteriors.factor", "mixed_posteriors.vector")

  }else if(prior_info[["independent"]]){

    if(prior_info[["levels"]] == 1){

      samples <- .as_mixed_posteriors.simple(model_samples, prior, parameter)
      samples <- matrix(samples, ncol = 1)

    }else{

      samples <- lapply(1:prior_info[["levels"]], function(i) .as_mixed_posteriors.simple(model_samples, prior, paste0(parameter, "[", i, "]")))
      samples <- do.call(cbind, samples)

    }

    rownames(samples) <- NULL
    colnames(samples) <- .format_factor_level_parameter_names(parameter, prior_info[["level_names"]], ncol(samples))
    attr(samples, "sample_ind") <- FALSE
    attr(samples, "models_ind") <- rep(1, nrow(samples))
    attr(samples, "parameter")  <- parameter
    attr(samples, "prior_list") <- prior
    class(samples) <- c("mixed_posteriors", "mixed_posteriors.factor", "mixed_posteriors.vector")

  }else if(prior_info[["orthonormal"]] | prior_info[["meandif"]]){

    prior$parameter[["K"]] <- prior_info[["levels"]]
    samples <- .as_mixed_posteriors.vector(model_samples, prior, parameter)
    class(samples) <- c(class(samples), "mixed_posteriors.factor")

  }

  attr(samples, "levels")            <- prior_info[["levels"]]
  attr(samples, "level_names")       <- prior_info[["level_names"]]
  attr(samples, "interaction")       <- if(length(prior_info) == 0) FALSE else prior_info[["interaction"]]
  attr(samples, "interaction_terms") <- prior_info[["interaction_terms"]]
  attr(samples, "term_components")   <- prior_info[["term_components"]]
  attr(samples, "factor_terms")      <- prior_info[["factor_terms"]]
  attr(samples, "factor_contrasts")  <- prior_info[["factor_contrasts"]]
  attr(samples, "factor_design")     <- prior_info[["factor_design"]]
  attr(samples, "factor_cell_names") <- prior_info[["factor_cell_names"]]
  attr(samples, "treatment")         <- prior_info[["treatment"]]
  attr(samples, "independent")       <- prior_info[["independent"]]
  attr(samples, "orthonormal")       <- prior_info[["orthonormal"]]
  attr(samples, "meandif")           <- prior_info[["meandif"]]

  return(samples)
}
.as_mixed_posteriors.weightfunction <- function(model_samples, prior, parameter){

  # check input
  check_char(parameter, "parameter", check_length = FALSE)


  # obtain mapping for the weight coefficients
  omega_info    <- .weightfunction_mapping_info(list(prior))
  omega_mapping <- omega_info$mapping
  omega_names   <- omega_info$names
  omega_par     <- omega_info$pars

  # prepare output objects
  samples <- model_samples[, omega_par, drop = FALSE]

  rownames(samples) <- NULL
  colnames(samples) <- omega_names
  attr(samples, "sample_ind") <- FALSE
  attr(samples, "models_ind") <- rep(1, nrow(samples))
  attr(samples, "parameter")  <- parameter
  attr(samples, "prior_list") <- prior
  samples <- .weightfunction_set_omega_context(samples, omega_info)
  class(samples) <- c("mixed_posteriors", "mixed_posteriors.weightfunction")

  return(samples)
}
.as_mixed_posteriors.phacking      <- function(model_samples, prior, parameter){

  # check input
  check_char(parameter, "parameter", check_length = FALSE)

  par_names <- intersect(.phacking_report_parameter(prior), colnames(model_samples))
  samples   <- model_samples[, par_names, drop = FALSE]

  rownames(samples) <- NULL
  attr(samples, "sample_ind") <- FALSE
  attr(samples, "models_ind") <- rep(1, nrow(samples))
  attr(samples, "parameter")  <- parameter
  attr(samples, "prior_list") <- prior
  class(samples) <- c("mixed_posteriors", "mixed_posteriors.phacking")

  return(samples)
}
.as_mixed_posteriors.bias          <- function(model_samples, prior, parameter, conditional){

  # check input
  check_char(parameter, "parameter", check_length = FALSE)

  branch_info   <- .selection_prior_branch_info(prior)
  has_selection <- vapply(branch_info, function(x) !is.null(x$selection), logical(1))
  has_phacking  <- vapply(branch_info, function(x) !is.null(x$phacking),  logical(1))

  out_names <- NULL
  par_names <- NULL

  if(any(has_selection)){
    selection_priors <- lapply(branch_info[has_selection], function(x) x$selection)
    omega_info       <- .weightfunction_mapping_info(selection_priors, one_sided = TRUE)
    omega_names      <- omega_info$names
    omega_par        <- omega_info$pars
  }
  if(any(has_phacking)){
    phacking_priors <- lapply(branch_info[has_phacking], function(x) x$phacking)
    phacking_par    <- .selection_phacking_report_parameters(phacking_priors)
    phacking_names  <- phacking_par
  }

  if(length(conditional) > 0 && any(c("omega", "phacking", "alpha", "pi_null") %in% conditional)){
    if("omega" %in% conditional && any(has_selection)){
      out_names <- c(out_names, omega_names)
      par_names <- c(par_names, omega_par)
    }
    if("phacking" %in% conditional && any(has_phacking)){
      out_names <- c(out_names, phacking_names)
      par_names <- c(par_names, phacking_par)
    }
    if("alpha" %in% conditional && any(has_phacking)){
      out_names <- c(out_names, "alpha")
      par_names <- c(par_names, "alpha")
    }
    if("pi_null" %in% conditional && any(has_phacking)){
      out_names <- c(out_names, "pi_null")
      par_names <- c(par_names, "pi_null")
    }
  }else{
    if(any(has_selection)){
      out_names <- c(out_names, omega_names)
      par_names <- c(par_names, omega_par)
    }
    if(any(has_phacking)){
      out_names <- c(out_names, phacking_names)
      par_names <- c(par_names, phacking_par)
    }
  }

  if(is.null(par_names)){
    par_names <- character()
    out_names <- character()
  }
  keep_unique <- !duplicated(par_names)
  par_names <- par_names[keep_unique]
  out_names <- out_names[keep_unique]
  keep_par <- par_names %in% colnames(model_samples)
  par_names <- par_names[keep_par]
  out_names <- out_names[keep_par]
  samples   <- model_samples[, par_names, drop = FALSE]

  rownames(samples) <- NULL
  colnames(samples) <- out_names
  attr(samples, "sample_ind") <- FALSE
  attr(samples, "models_ind") <- rep(1, nrow(samples))
  attr(samples, "parameter")  <- parameter
  attr(samples, "prior_list") <- prior
  if(any(has_selection)){
    samples <- .weightfunction_set_omega_context(samples, omega_info)
  }
  class(samples) <- c("mixed_posteriors", "mixed_posteriors.bias")

  return(samples)
}
.as_mixed_posteriors.spike_and_slab <- function(model_samples, prior, parameter){

  # check input
  check_char(parameter, "parameter", check_length = FALSE)

  prior_variable <- .get_spike_and_slab_variable(prior)

  # prepare output objects
  if(is.prior.factor(prior_variable)){

    samples <- .as_mixed_posteriors.factor(model_samples, prior_variable, parameter)
    attr(samples, "models_ind") <- as.vector(model_samples[,paste0(parameter, "_indicator")])

  }else if(is.prior.simple(prior_variable)){

    samples <- .as_mixed_posteriors.simple(model_samples, prior_variable, parameter)
    attr(samples, "models_ind") <- as.vector(model_samples[,paste0(parameter, "_indicator")])

  }

  class(samples) <- c("mixed_posteriors.spike_and_slab", class(samples))
  attr(samples, "prior_list") <- prior

  return(samples)
}
.as_mixed_posteriors.mixture        <- function(model_samples, prior, parameter, conditional){

  # check input
  check_char(parameter, "parameter", check_length = FALSE)


  # prepare output objects
  if(inherits(prior, "prior.bias_mixture")){

    is_PET            <- sapply(prior, is.prior.PET)
    is_PEESE          <- sapply(prior, is.prior.PEESE)
    branch_info       <- .selection_prior_branch_info(prior)
    has_selection     <- vapply(branch_info, function(x) !is.null(x$selection), logical(1))
    has_phacking      <- vapply(branch_info, function(x) !is.null(x$phacking), logical(1))

    # prepare weightfunction parameter names
    if(any(has_selection)){
      selection_priors <- lapply(branch_info[has_selection], function(x) x$selection)
      omega_info    <- .weightfunction_mapping_info(selection_priors, one_sided = TRUE)
      omega_mapping <- omega_info$mapping
      omega_names   <- omega_info$names
      omega_par     <- omega_info$pars
    }
    if(any(has_phacking)){
      phacking_priors <- lapply(branch_info[has_phacking], function(x) x$phacking)
      phacking_par    <- .selection_phacking_report_parameters(phacking_priors)
      phacking_names  <- phacking_par
    }

    # deal with conditional parameters
    if(length(conditional) > 0 && any(c("PET", "PEESE", "PETPEESE", "omega", "phacking", "alpha", "pi_null") %in% conditional)){

      out_names <- NULL
      par_names <- NULL

      if("omega" %in% conditional && any(has_selection)){
        out_names <- c(out_names, omega_names)
        par_names <- c(par_names, omega_par)
      }
      if("phacking" %in% conditional && any(has_phacking)){
        out_names <- c(out_names, phacking_names)
        par_names <- c(par_names, phacking_par)
      }
      if("alpha" %in% conditional && any(has_phacking)){
        out_names <- c(out_names, "alpha")
        par_names <- c(par_names, "alpha")
      }
      if("pi_null" %in% conditional && any(has_phacking)){
        out_names <- c(out_names, "pi_null")
        par_names <- c(par_names, "pi_null")
      }
      if("PETPEESE" %in% conditional){
        # subset in case only PET/PEESE is supplied
        out_names <- c(out_names, colnames(model_samples)[colnames(model_samples) %in% c("PET", "PEESE")])
        par_names <- c(par_names, colnames(model_samples)[colnames(model_samples) %in% c("PET", "PEESE")])
      }
      if("PET" %in% conditional && any(is_PET)){
        out_names <- c(out_names, "PET")
        par_names <- c(par_names, "PET")
      }
      if("PEESE" %in% conditional && any(is_PEESE)){
        out_names <- c(out_names, "PEESE")
        par_names <- c(par_names, "PEESE")
      }

    }else{

      out_names <- NULL
      par_names <- NULL

      if(any(has_selection)){
        out_names <- c(out_names, omega_names)
        par_names <- c(par_names, omega_par)
      }
      if(any(has_phacking)){
        out_names <- c(out_names, phacking_names)
        par_names <- c(par_names, phacking_par)
      }
      if(any(is_PET)){
        out_names <- c(out_names, "PET")
        par_names <- c(par_names, "PET")
      }
      if(any(is_PEESE)){
        out_names <- c(out_names, "PEESE")
        par_names <- c(par_names, "PEESE")
      }
    }

    # select samples
    if(is.null(par_names)){
      par_names <- character()
      out_names <- character()
    }
    keep_unique <- !duplicated(par_names)
    par_names <- par_names[keep_unique]
    out_names <- out_names[keep_unique]
    keep_par <- par_names %in% colnames(model_samples)
    par_names <- par_names[keep_par]
    out_names <- out_names[keep_par]
    samples   <- model_samples[, par_names,drop=FALSE]
    indicator <- model_samples[,paste0(parameter, "_indicator")]

    rownames(samples) <- NULL
    colnames(samples) <- out_names
    attr(samples, "sample_ind") <- FALSE
    attr(samples, "models_ind") <- as.vector(indicator)
    attr(samples, "parameter")  <- parameter
    attr(samples, "prior_list") <- prior
    if(any(has_selection)){
      samples <- .weightfunction_set_omega_context(samples, omega_info)
    }
    class(samples) <- c("mixed_posteriors", "mixed_posteriors.bias")

  }else{

    if(inherits(prior, "prior.simple_mixture")){
      samples <- .as_mixed_posteriors.simple(model_samples, prior, parameter)
    }else if(inherits(prior, "prior.factor_mixture")){
      samples <- .as_mixed_posteriors.factor(model_samples, prior, parameter)
    }
    attr(samples, "models_ind") <- as.vector(model_samples[,paste0(parameter, "_indicator")])

  }

  class(samples) <- c("mixed_posteriors.mixture", class(samples))
  attr(samples, "prior_list") <- prior

  return(samples)
}

#' @title Compute inclusion Bayes factors
#'
#' @description Computes inclusion Bayes factors based on prior model probabilities,
#' posterior model probabilities (or marginal likelihoods), and indicator whether
#' the models represent the null or alternative hypothesis.
#'
#' @param prior_probs vector of prior model probabilities
#' @param post_probs vector of posterior model probabilities
#' @param margliks vector of marginal likelihoods.
#' @param is_null logical vector of indicators whether the model corresponds
#' to the null or alternative hypothesis (or an integer vector indexing models
#' corresponding to the null hypothesis)
#'
#' @details Supplying \code{margliks} as the input is preferred since it is better at dealing with
#' under/overflow (posterior probabilities are very close to either 0 or 1). In case that both the
#' \code{post_probs} and \code{margliks} are supplied, the results are based on \code{margliks}.
#' If the prior probability of either the null or alternative hypothesis is
#' zero, the Bayes factor is undefined and \code{NA} is returned.
#'
#' @return \code{inclusion_BF} returns a Bayes factor.
#'
#' @export
inclusion_BF         <- function(prior_probs, post_probs, margliks, is_null){


  if(is.numeric(is_null)){
    check_int(is_null, "is_null", lower = 1, upper = length(prior_probs), check_length = 0)
    is_null <- c(1:length(prior_probs)) %in% is_null
  }else if(is.logical(is_null)){
    check_bool(is_null, "is_null", check_length = length(prior_probs))
  }else{
    stop("'is_null' argument must be either logical vector, integer vector, or NULL.")
  }

  if(!missing(prior_probs) && !missing(margliks)){
    return(.inclusion_BF.margliks(prior_probs = prior_probs, margliks = margliks, is_null = is_null))
  }else if(!missing(prior_probs) && !missing(post_probs)){
    return(.inclusion_BF.probs(prior_probs = prior_probs, post_probs = post_probs, is_null = is_null))
  }else{
    stop("'prior_probs' and either 'post_probs' or 'marglik' must be specified.")
  }
}

.inclusion_BF.probs    <- function(prior_probs, post_probs, is_null){

  check_real(prior_probs, "prior_probs", lower = 0, upper = 1, check_length = 0)
  check_real(post_probs,  "post_probs", lower = 0, upper = 1, check_length = length(prior_probs))

  prior_alt  <- sum(prior_probs[!is_null])
  prior_null <- sum(prior_probs[is_null])
  post_alt   <- sum(post_probs[!is_null])
  post_null  <- sum(post_probs[is_null])

  if(isTRUE(all.equal(prior_alt, 0)) || isTRUE(all.equal(prior_null, 0))){
    return(NA_real_)
  }

  if(isTRUE(all.equal(post_alt, 1))){
    return(Inf)
  }else if(isTRUE(all.equal(post_null, 1))){
    return(0)
  }else{
    return(
      (post_alt / post_null) / (prior_alt / prior_null)
    )
  }
}
.inclusion_BF.margliks <- function(prior_probs, margliks, is_null){

  check_real(prior_probs, "prior_probs", lower = 0, upper = 1, check_length = 0)
  check_real(margliks,  "margliks", check_length = length(prior_probs))

  prior_alt  <- sum(prior_probs[!is_null])
  prior_null <- sum(prior_probs[is_null])

  if(isTRUE(all.equal(prior_alt, 0)) || isTRUE(all.equal(prior_null, 0))){
    return(NA_real_)
  }

  margliks <- .model_averaging_margliks(margliks, prior_probs)

  active <- prior_probs > 0 & is.finite(margliks)
  if(!any(active & !is_null)){
    return(0)
  }
  if(!any(active & is_null)){
    return(Inf)
  }

  # subtract the max among positive-prior finite models to avoid overflow.
  margliks <- margliks - max(margliks[active])

  alt_ind  <- active & !is_null
  null_ind <- active & is_null

  alt_marginal  <- sum(exp(margliks[alt_ind])  * prior_probs[alt_ind])
  null_marginal <- sum(exp(margliks[null_ind]) * prior_probs[null_ind])

  if(alt_marginal == 0 && null_marginal == 0){
    return(NaN)
  }
  if(alt_marginal == 0){
    return(0)
  }
  if(null_marginal == 0){
    return(Inf)
  }

  return(
    (alt_marginal / null_marginal) / (prior_alt / prior_null)
  )
}


#' @title Create coefficient mapping between multiple weightfunctions
#'
#' @description Creates coefficients mapping between multiple weightfunctions.
#'
#' @param prior_list list of prior distributions
#' @param cuts_only whether only p-value cuts should be returned
#' @param one_sided force one-sided output
#'
#' @return \code{weightfunctions_mapping} returns a list of indices
#' mapping the publication weights omega from the individual weightfunctions
#' into a joint weightfunction.
#'
#' @export
weightfunctions_mapping <- function(prior_list, cuts_only = FALSE, one_sided = FALSE){

  # check input
  if(!all(sapply(prior_list, is.prior.weightfunction) | sapply(prior_list, .is_prior_weightfunction_null)))
    stop("'priors' must be a list of weightfunction priors or point(1)/none null priors")
  check_bool(cuts_only, "cuts_only")
  check_bool(one_sided, "one_sided")

  force_one_sided <- one_sided || any(sapply(prior_list, function(prior){
    is.prior.weightfunction(prior) && prior$side == "one-sided"
  }))

  prior_expansions <- lapply(prior_list, function(prior){
    if(!is.prior.weightfunction(prior)){
      return(NULL)
    }
    .weightfunction_mapping_expansion(prior, force_one_sided)
  })

  all_cuts <- .weightfunction_unique_cuts(unlist(lapply(prior_expansions, function(expansion){
    if(is.null(expansion)) NULL else expansion$cuts
  })))
  if(length(all_cuts) == 0L){
    all_cuts <- c(0, 1)
  }

  # return the naming for summary function if only asked for labels
  if(cuts_only){
    return(all_cuts)
  }

  # create mapping to weights
  omega_mapping <- list()
  for(p in seq_along(prior_list)){
    if(is.prior.weightfunction(prior_list[[p]])){
      expansion <- prior_expansions[[p]]
      omega_mapping[[p]] <- expansion$index[.weightfunction_global_bin_indices(all_cuts, expansion)]
    }
  }


  return(omega_mapping)
}

.weightfunction_mapping_info <- function(prior_list, one_sided = FALSE){

  cuts <- weightfunctions_mapping(prior_list, cuts_only = TRUE, one_sided = one_sided)
  list(
    mapping = weightfunctions_mapping(prior_list, one_sided = one_sided),
    cuts    = cuts,
    names   = .weightfunction_omega_names(cuts),
    pars    = paste0("omega[", seq_len(length(cuts) - 1L), "]"),
    one_sided = one_sided
  )
}

.weightfunction_set_omega_context <- function(samples, omega_context){

  if(is.null(omega_context)){
    return(samples)
  }

  attr(samples, "omega_context") <- omega_context
  prior_list <- attr(samples, "prior_list")
  if(!is.null(prior_list)){
    attr(prior_list, "omega_context") <- omega_context
    attr(samples, "prior_list") <- prior_list
  }

  samples
}

.weightfunction_omega_names <- function(cuts){
  sapply(seq_len(length(cuts) - 1L), function(i){
    paste0("omega[", cuts[i], ",", cuts[i + 1L], "]")
  })
}

.weightfunction_unique_cuts <- function(cuts, tolerance = sqrt(.Machine$double.eps)){

  cuts <- sort(cuts)
  if(length(cuts) <= 1L){
    return(cuts)
  }

  out <- cuts[1L]
  for(cut in cuts[-1L]){
    if(abs(cut - out[length(out)]) <= tolerance){
      if(abs(cut) <= tolerance || abs(out[length(out)]) <= tolerance){
        out[length(out)] <- 0
      }else if(abs(cut - 1) <= tolerance || abs(out[length(out)] - 1) <= tolerance){
        out[length(out)] <- 1
      }else{
        out[length(out)] <- min(cut, out[length(out)])
      }
    }else{
      out <- c(out, cut)
    }
  }

  out
}

.weightfunction_global_bin_indices <- function(global_cuts, expansion, tolerance = sqrt(.Machine$double.eps)){

  vapply(seq_len(length(global_cuts) - 1L), function(i){
    ind <- which(
      global_cuts[i] >= expansion$lower - tolerance &
        global_cuts[i + 1L] <= expansion$upper + tolerance
    )
    if(length(ind) != 1L){
      stop("Could not map global weightfunction bin to a local bin.", call. = FALSE)
    }
    ind
  }, integer(1))
}

.weightfunction_mapping_expansion <- function(prior, force_one_sided = FALSE){

  if(prior$side == "two-sided" && force_one_sided){
    J <- .weightfunction_n_bins(prior)
    cuts <- c(0, prior$steps / 2, 1 - rev(prior$steps) / 2, 1)
    index <- c(seq_len(J), seq.int(J - 1L, 1L))
  }else{
    cuts <- .weightfunction_local_cuts(prior)
    index <- seq_len(.weightfunction_n_bins(prior))
  }

  list(
    cuts  = cuts,
    lower = cuts[-length(cuts)],
    upper = cuts[-1],
    index = index
  )
}

#' @title Multivariate point mass distribution
#'
#' @description Density, distribution function,
#' quantile function and random generation for multivariate point distribution.
#'
#' @param x,q vector or matrix of quantiles.
#' @param p vector of probabilities.
#' @param n number of observations.
#' @param location vector of locations corresponding to the location of individual points.
#' Alternatively, a matrix with rows corresponding to the location of individual samples
#' and columns correspond to the location of individual points.
#' @param log,log.p logical; if \code{TRUE}, probabilities
#' \code{p} are given as \code{log(p)}.
#' @param lower.tail logical; if \code{TRUE} (default), probabilities
#' are \eqn{P[X \le x]}, otherwise, \eqn{P[X > x]}.
#'
#' @examples
#' # draw samples from a multivariate point distribution
#' rmpoint(10, location = c(0, 1))
#'
#' @return \code{dpoint} gives the density, \code{ppoint} gives the
#' distribution function, \code{qpoint} gives the quantile function,
#' and \code{rpoint} generates random deviates.
#'
#' @export dmpoint
#' @export rmpoint
#' @export pmpoint
#' @export qmpoint
#' @name mpoint
NULL

#### wrappers ####
#' @rdname mpoint
dmpoint <- function(x, location, log = FALSE){

  # common input check
  .check_log(log)
  .check_x(as.vector(x))
  check_real(as.vector(location), "location", check_length = 0)

  # transform arguments to the matrix form
  if(!is.matrix(x)){
    x <- matrix(x, nrow = 1)
  }
  if(!is.matrix(location)){
    if(length(location) == 1){
      location <- rep(location, ncol(x))
    }
    location <- matrix(location, nrow = 1)
  }

  if((nrow(location) != nrow(x) & nrow(location) != 1) || ncol(location) != ncol(x))
    stop("Non matching dimensions of 'location' and 'x'.")

  if(nrow(location) != nrow(x) & nrow(location) == 1){
    location <- matrix(location, nrow = nrow(x), ncol = ncol(location), byrow = TRUE)
  }

  lik <- vapply(seq_len(nrow(x)), function(i){
    if(anyNA(location[i,]) || anyNA(x[i,])){
      return(NA_real_)
    }else if(all(location[i,] == x[i,])){
      return(Inf)
    }else{
      return(0)
    }
  }, numeric(1))

  if(log){
    lik <- log(lik)
  }

  return(lik)
}
#' @rdname mpoint
rmpoint <- function(n, location){

  # common input check
  .check_n(n)
  check_real(as.vector(location), "location", check_length = 0)

  # transform arguments to the matrix form
  if(!is.matrix(location)){
    location <- matrix(location, nrow = 1)
  }

  if(nrow(location) != n & nrow(location) != 1)
    stop("Incompatible dimensions of requested number of samples and 'location'.")

  # 'sample' distribution
  if(nrow(location) == n){
    return(location)
  }else{
    x <- matrix(location, byrow = TRUE, ncol = ncol(location), nrow = n)
  }

  return(x)
}
#' @rdname mpoint
pmpoint <- function(q, location, lower.tail = TRUE, log.p = FALSE){

  # common input check
  .check_log.p(log.p)
  .check_lower.tail(lower.tail)
  .check_q(as.vector(q))
  check_real(as.vector(location), "location", check_length = 0)

  # transform arguments to the matrix form
  if(!is.matrix(q)){
    q <- matrix(q, nrow = 1)
  }

  if(!is.matrix(location)){
    if(length(location) == 1){
      location <- rep(location, ncol(q))
    }
    location <- matrix(location, nrow = 1)
  }

  if(nrow(location) != nrow(q) & nrow(location) != 1 || ncol(location) != ncol(q))
    stop("Non matching dimensions of 'location' and 'q'.")

  if(nrow(location) != nrow(q) & nrow(location) == 1){
    location <- matrix(location, nrow = nrow(q), ncol = ncol(location), byrow = TRUE)
  }

  p <- as.numeric(apply(location <= q, 1, all))

  if(!lower.tail){
    p <- 1 - p
  }
  if(log.p){
    p <- log(p)
  }

  return(p)
}
#' @rdname mpoint
qmpoint <- function(p, location, lower.tail = TRUE, log.p = FALSE){

  # common input check
  .check_log.p(log.p)
  .check_lower.tail(lower.tail)
  .check_p(p, log.p)
  check_real(as.vector(location), "location", check_length = 0)

  # transform arguments to the matrix form
  if(!is.matrix(location)){
    location <- matrix(location, nrow = 1)
  }

  if(nrow(location) != length(p) & nrow(location) != 1)
    stop("Non matching dimensions of 'location' and 'p'.")

  if(nrow(location) != length(p) & nrow(location) == 1){
    location <- matrix(location, nrow = length(p), ncol = ncol(location), byrow = TRUE)
  }

  if(log.p){
    p <- exp(p)
  }

  q <- do.call(rbind, lapply(seq_along(p), function(i){
    if(is.na(p[i])){
      return(matrix(NA_real_, nrow = 1, ncol = ncol(location)))
    }else if(all(p[i] > 0)){
      return(location[i,])
    }else{
      return(matrix(ifelse(lower.tail, -Inf, Inf), nrow = 1, ncol = ncol(location)))
    }
  }))

  return(q)
}

#' @title Creates a prior distribution
#'
#' @description \code{prior} creates a prior distribution.
#' The prior can be visualized by the \code{plot} function.
#'
#' @param distribution name of the prior distribution. The
#' possible options are
#' \describe{
#'   \item{\code{"point"}}{for a point density characterized by a
#'   \code{location} parameter.}
#'   \item{\code{"normal"}}{for a normal distribution characterized
#'   by a \code{mean} and \code{sd} parameters.}
#'   \item{\code{"lognormal"}}{for a lognormal distribution characterized
#'   by a \code{meanlog} and \code{sdlog} parameters.}
#'   \item{\code{"cauchy"}}{for a Cauchy distribution characterized
#'   by a \code{location} and \code{scale} parameters. Internally
#'   converted into a generalized t-distribution with \code{df = 1}.}
#'   \item{\code{"t"}}{for a generalized t-distribution characterized
#'   by a \code{location}, \code{scale}, and \code{df} parameters.}
#'   \item{\code{"gamma"}}{for a gamma distribution characterized
#'   by either \code{shape} and \code{rate}, or \code{shape} and
#'   \code{scale} parameters. The later is internally converted to
#'   the \code{shape} and \code{rate} parametrization}
#'   \item{\code{"invgamma"}}{for an inverse-gamma distribution
#'   characterized by a \code{shape} and \code{scale} parameters. The
#'   JAGS part uses a 1/gamma distribution with a shape and rate
#'   parameter.}
#'   \item{\code{"beta"}}{for a beta distribution
#'   characterized by an \code{alpha} and \code{beta} parameters.}
#'   \item{\code{"exp"}}{for an exponential distribution
#'   characterized by either \code{rate} or \code{scale}
#'   parameter. The later is internally converted to
#'   \code{rate}.}
#'   \item{\code{"uniform"}}{for a uniform distribution defined on a
#'   range from \code{a} to \code{b}}
#' }
#' @param parameters list of appropriate parameters for a given
#' \code{distribution}.
#' @param truncation list with two elements, \code{lower} and
#' \code{upper}, that define the lower and upper truncation of the
#' distribution. Defaults to \code{list(lower = -Inf, upper = Inf)}.
#' The truncation is automatically set to the bounds of the support.
#' @param prior_weights prior odds associated with a given distribution.
#' The value is passed into the model fitting function, which creates models
#' corresponding to all combinations of prior distributions for each of
#' the model parameters and sets the model priors odds to the product
#' of its prior distributions.
#'
#' @examples
#' # create a standard normal prior distribution
#' p1 <- prior(distribution = "normal", parameters = list(mean = 1, sd = 1))
#'
#' # create a half-normal standard normal prior distribution
#' p2 <- prior(distribution = "normal", parameters = list(mean = 1, sd = 1),
#' truncation = list(lower = 0, upper = Inf))
#'
#' # the prior distribution can be visualized using the plot function
#' # (see ?plot.prior for all options)
#' plot(p1)
#'
#' @return \code{prior} and \code{prior_none} return an object of class 'prior'.
#' A named list containing the distribution name, parameters, and prior weights.
#'
#' @name prior
#' @export prior
#' @export prior_none
#' @seealso [plot.prior()], \link[stats]{Normal}, \link[stats]{Lognormal}, \link[stats]{Cauchy},
#' \link[stats]{Beta}, \link[stats]{Exponential},
#' \link[extraDistr]{LocationScaleT}, \link[extraDistr]{InvGamma}.

#' @rdname prior
prior <- function(distribution, parameters, truncation = list(lower = -Inf, upper = Inf), prior_weights = 1){

  # general input check (detailed checks are performed withing the constructors)
  check_char(distribution, "distribution")
  check_list(parameters, "parameters")
  #sapply(seq_along(parameters), function(i)check_real(parameters[[i]], names(parameters[[i]]), check_length = 0))
  check_list(truncation, "truncation")
  .check_prior_weight(prior_weights)

  # clean the input name
  distribution <- .prior_clean_input_name(distribution)

  if(distribution %in% c("norm", "normal")){
    distribution <- "normal"
  }else if(distribution %in% c("lnorm", "lognormal")){
    distribution <- "lognormal"
  }else if(distribution %in% c("t", "student", "studentt")){
    distribution <- "t"
  }else if(distribution %in% c("cauchy")){
    distribution <- "cauchy"
  }else if(distribution %in% c("invgamma", "inversegamma")){
    distribution <- "invgamma"
  }else if(distribution %in% c("gamma")){
    distribution <- "gamma"
  }else if(distribution %in% c("beta")){
    distribution <- "beta"
  }else if(distribution %in% c("bernoulli", "bernouli", "bern")){
    distribution <- "bernoulli"
  }else if(distribution %in% c("exp", "exponential")){
    distribution <- "exp"
  }else if(distribution %in% c("uniform", "unif")){
    distribution <- "uniform"
  }else if(distribution %in% c("point", "spike")){
    distribution <- "point"
  }else if(distribution %in% c("multivariatenorm", "multivariatenormal", "mnorm", "mnormal")){
    distribution <- "mnormal"
  }else if(distribution %in% c("multivariatet", "multivariatestudent", "mt", "mstudent")){
    distribution <- "mt"
  }else if(distribution %in% c("multivariatecauchy", "mcauchy")){
    distribution <- "mcauchy"
  }else if(distribution %in% c("mpoint", "mspike")){
    distribution <- "mpoint"
  }else{
    stop(paste0("The specified distribution name '", distribution,"' is not known. Please, see '?prior' for more information about supported prior distributions."))
  }

  # check the passed settings
  output <- do.call(paste0(".prior_", distribution), list(parameters = parameters, truncation = truncation))

  # add the prior odds
  output$prior_weights <- prior_weights

  return(output)
}

#' @rdname prior
prior_none <- function(prior_weights = 1){

  .check_prior_weight(prior_weights)

  out <- list()
  out$distribution <- "none"
  out$prior_weights   <- prior_weights
  class(out)       <- c("prior", "prior.none")

  return(out)
}


#' @title Creates a prior distribution for a weight function
#'
#' @description \code{prior_weightfunction} creates a prior distribution for
#' fitting a RoBMA selection model. The \code{side} and \code{steps} arguments
#' define the p-value bins, and the \code{weights} argument defines the prior on
#' the publication weights in those bins.
#'
#' @param side side geometry. Either \code{"one-sided"} or \code{"two-sided"}.
#' @param steps increasing p-value cut points between 0 and 1.
#' @param weights a weight-prior object created by \code{wf_cumulative()},
#' \code{wf_fixed()}, or \code{wf_independent()}.
#' @param reference reference bin. Currently only \code{"most_significant"} is
#' supported and fixes the most significant bin to \code{omega = 1}.
#' @param prior_weights prior odds associated with a given distribution.
#'
#' @examples
#' p1 <- prior_weightfunction(
#'   side = "one-sided",
#'   steps = c(.05, .10),
#'   weights = wf_cumulative(alpha = c(1, 1, 1))
#' )
#'
#' p2 <- prior_weightfunction(
#'   side = "one-sided",
#'   steps = c(.05),
#'   weights = wf_independent(prior("beta", list(1, 1)))
#' )
#'
#' @return \code{prior_weightfunction} returns an object of class 'prior'.
#'
#' @export prior_weightfunction
#' @seealso [plot.prior()]
prior_weightfunction <- function(side = "one-sided", steps = c(.025, .05),
                                 weights = wf_cumulative(),
                                 reference = "most_significant",
                                 prior_weights = 1){

  check_char(side, "side")
  check_real(steps, "steps", check_length = 0)
  check_char(reference, "reference", allow_values = "most_significant")
  .check_prior_weight(prior_weights)

  side <- .weightfunction_normalize_side(side)
  steps <- .weightfunction_validate_steps(steps)
  bins <- .weightfunction_bins(steps)
  weights <- .weightfunction_validate_weights(weights, n_bins = nrow(bins), reference = reference)
  truncation <- .weightfunction_weights_truncation(weights)

  output <- list(
    distribution  = "weightfunction",
    side          = side,
    steps         = steps,
    bins          = bins,
    reference     = reference,
    weights       = weights,
    parameters    = list(steps = steps),
    truncation    = truncation,
    prior_weights = prior_weights
  )

  class(output) <- c("prior", "prior.weightfunction")

  return(output)
}

#' @rdname prior_weightfunction
#' @param alpha positive cumulative-Dirichlet concentration parameters. If
#' omitted, a flat Dirichlet prior is used with one concentration parameter per
#' bin.
#' @export
wf_cumulative <- function(alpha = NULL){

  if(!is.null(alpha)){
    check_real(alpha, "alpha", lower = 0, allow_bound = FALSE, check_length = 0, allow_NA = FALSE)
  }

  out <- list(type = "cumulative", alpha = alpha)
  class(out) <- c("weightfunction_weights", "weightfunction_weights.cumulative")
  return(out)
}

#' @rdname prior_weightfunction
#' @param omega fixed non-negative relative publication weights, one per bin.
#' The reference-bin weight must be exactly 1.
#' @export
wf_fixed <- function(omega){

  check_real(omega, "omega", lower = 0, check_length = 0, allow_NA = FALSE)

  out <- list(type = "fixed", omega = omega)
  class(out) <- c("weightfunction_weights", "weightfunction_weights.fixed")
  return(out)
}

#' @rdname prior_weightfunction
#' @param prior prior distribution for each non-reference weight.
#' @param scale latent scale for independent weights. \code{"omega"} places the
#' prior directly on the non-negative publication weight. \code{"log_omega"}
#' places the prior on \code{log(omega)} and transforms with
#' \code{omega = exp(log_omega)}, allowing weights above one whenever the log
#' prior assigns mass above zero.
#' @export
wf_independent <- function(prior, scale = "omega"){

  .check_prior(prior)
  if(!is.prior.simple(prior) || is.prior.point(prior) || is.prior.discrete(prior)){
    stop("'prior' must be a continuous simple prior distribution.", call. = FALSE)
  }

  check_char(scale, "scale", allow_values = c("omega", "log_omega", "log"))
  if(scale == "log"){
    scale <- "log_omega"
  }

  if(scale == "omega"){
    if(prior$truncation[["lower"]] < 0){
      stop("Independent omega-scale weight priors must have non-negative support.", call. = FALSE)
    }
  }

  out <- list(type = "independent", scale = scale, prior = prior)
  class(out) <- c("weightfunction_weights", "weightfunction_weights.independent")
  return(out)
}

#' @title Creates a prior distribution for PET or PEESE models
#'
#' @description \code{prior} creates a prior distribution for fitting a PET or
#' PEESE style models in RoBMA. The prior distribution can be visualized
#' by the \code{plot} function.
#'
#' @examples
#' # create a half-Cauchy prior distribution
#' # (PET and PEESE specific functions automatically set lower truncation at 0)
#' p1 <- prior_PET(distribution = "Cauchy", parameters = list(location = 0, scale = 1))
#'
#' plot(p1)
#'
#' @return \code{prior_PET} and \code{prior_PEESE} return an object of class 'prior'.
#'
#' @inheritParams prior
#' @export prior_PET
#' @export prior_PEESE
#' @seealso [plot.prior()], [prior()]
#' @name prior_PP
NULL

#' @rdname prior_PP
prior_PET   <- function(distribution, parameters, truncation = list(lower = 0, upper = Inf), prior_weights = 1){

  output <- prior(distribution, parameters, truncation, prior_weights)

  class(output) <- c(class(output), "prior.PET")

  return(output)
}
#' @rdname prior_PP
prior_PEESE <- function(distribution, parameters, truncation = list(lower = 0, upper = Inf), prior_weights = 1){

  output <- prior(distribution, parameters, truncation, prior_weights)

  class(output) <- c(class(output), "prior.PEESE")

  return(output)
}

#' @title Creates a prior distribution for factors
#'
#' @description \code{prior_factor} creates a prior distribution for fitting
#' models with factor predictors. (Note that results across different operating
#' systems might vary due to differences in JAGS numerical precision.)
#'
#' @param contrast type of contrast for the prior distribution. The possible options are
#' \describe{
#'   \item{\code{"meandif"}}{for contrast centered around the grand mean
#'   with equal marginal distributions, making the prior distribution exchangeable
#'   across factor levels. In contrast to \code{"orthonormal"}, the marginal distributions
#'   are identical regardless of the number of factor levels and the specified prior
#'   distribution corresponds to the difference from grand mean for each factor level.
#'   Only supports \code{distribution = "mnormal"} and \code{distribution = "mt"}
#'   which generates the corresponding multivariate normal/t distributions.}
#'   \item{\code{"orthonormal"}}{for contrast centered around the grand mean
#'   with equal marginal distributions, making the prior distribution exchangeable
#'   across factor levels. Only supports \code{distribution = "mnormal"} and
#'   \code{distribution = "mt"} which generates the corresponding multivariate normal/t
#'   distributions.}
#'   \item{\code{"treatment"}}{for contrasts using the first level as a comparison
#'   group and setting equal prior distribution on differences between the individual
#'   factor levels and the comparison level.}
#'   \item{\code{"independent"}}{for contrasts specifying dependent prior distribution
#'   for each factor level (note that this leads to an overparameterized model if the
#'   intercept is included).}
#' }
#'
#'
#' @examples
#' # create an orthonormal prior distribution
#' p1 <- prior_factor(distribution = "mnormal", contrast = "orthonormal",
#'                    parameters = list(mean = 0, sd = 1))
#'
#' @return return an object of class 'prior'.
#'
#' @inheritParams prior
#' @export  prior_factor
#' @seealso [prior()]
prior_factor <- function(distribution, parameters, truncation = list(lower = -Inf, upper = Inf), prior_weights = 1, contrast = "meandif"){

  # general input check (detailed checks are performed withing the constructors)
  check_char(contrast, "contrast", allow_values = c("meandif", "orthonormal", "treatment", "dummy", "independent"))

  # check its compatibility with the contrasts
  if(contrast %in% c("meandif", "orthonormal")){

    # add the (yet unspecified) dimensions parameter
    if(is.null(names(parameters))){
      parameters <- c(parameters, NA)
    }else{
      parameters[["K"]] <- NA
    }

    # change spike/point into mpoint dispatch
    if(distribution %in% c("spike", "mspike", "point", "mpoint"))
      distribution <- "mpoint"

    if(!distribution %in% c("multivariatenorm", "multivariatenormal", "mnorm", "mnormal",
                            "multivariatet", "multivariatestudent", "mt", "mstudent",
                            "multivariatecauchy", "mcauchy",
                            "mpoint"))
      stop(paste0("'", contrast,"' contrasts require multivariate prior disribution."))

    # generate the prior object
    output <- prior(distribution = distribution, parameters = parameters, truncation = truncation, prior_weights = prior_weights)

    if(!is.prior.vector(output))
      stop(paste0("'", contrast,"' contrasts require vector prior distribution."))
    if(output[["distribution"]] != "mpoint" && !all(sapply(output[["truncation"]], is.infinite)))
      stop(paste0("'", contrast,"' contrasts do not support truncation."))

    class(output) <- c(class(output), "prior.factor", paste0("prior.", contrast))

  }else if(contrast %in% c("treatment", "dummy")){

    # generate the prior object
    output <- prior(distribution = distribution, parameters = parameters, truncation = truncation, prior_weights = prior_weights)

    if(!is.prior.simple(output))
      stop("'treatment' contrasts require univariate prior distribution.")

    output <- prior(distribution = distribution, parameters = parameters, truncation = truncation, prior_weights = prior_weights)

    class(output) <- c(class(output), "prior.factor", "prior.treatment")

  }else if(contrast  == "independent"){

    # generate the prior object
    output <- prior(distribution = distribution, parameters = parameters, truncation = truncation, prior_weights = prior_weights)

    if(!is.prior.simple(output))
      stop("'independent' contrasts require univariate prior distribution.")

    output <- prior(distribution = distribution, parameters = parameters, truncation = truncation, prior_weights = prior_weights)

    class(output) <- c(class(output), "prior.factor", "prior.independent")
  }

  return(output)
}


#' @title Creates a spike and slab prior distribution
#'
#' @description \code{prior_spike_and_slab} creates a spike and slab prior
#' distribution corresponding to the specification in
#' \insertCite{kuo1998variable;textual}{BayesTools} (see
#' \insertCite{ohara2009review;textual}{BayesTools} for further details). I.e.,
#' a prior distribution is multiplied by an independent indicator with values
#' either zero or one.
#'
#' @param prior_parameter a prior distribution for the parameter
#' @param prior_inclusion a prior distribution for the inclusion probability. The
#' inclusion probability must be bounded within 0 and 1 range. Defaults to
#' \code{prior("spike", parameters = list(location = 0.5))} which corresponds to 1/2
#' prior probability of including the slab prior distribution (but other prior
#' distributions, like beta etc can be also specified).
#'
#'
#' @examples
#' # create a spike and slab prior distribution
#' p1 <- prior_spike_and_slab(
#'    prior(distribution = "normal", parameters = list(mean = 0, sd = 1)),
#'    prior_inclusion = prior(distribution = "beta", parameters = list(alpha = 1, beta = 1))
#' )
#'
#' @return return an object of class 'prior'.
#'
#' @inheritParams prior
#' @seealso [prior()]
#' @export
prior_spike_and_slab <- function(prior_parameter,
                                 prior_inclusion = prior(distribution = "spike", parameters = list(location = 0.5)),
                                 prior_weights = 1){
  if(!is.prior(prior_parameter))
    stop("'prior_parameter' must be a prior distribution")
  if(!is.prior(prior_inclusion))
    stop("'prior_inclusion' must be a prior distribution")
  .check_prior_weight(prior_weights)
  if(is.prior.point(prior_inclusion) && (prior_inclusion$parameters[["location"]] < 0 | prior_inclusion$parameters[["location"]] > 1))
    stop("The probability parameter of 'prior_inclusion' must be within 0 and 1.")
  if(!is.prior.point(prior_inclusion) && (prior_inclusion$truncation[["lower"]] < 0 | prior_inclusion$truncation[["upper"]] > 1))
    stop("The range of the probability parameter (set via the 'truncation' argument) of 'prior_inclusion' must be within 0 and 1.")

  # Create the spike component (point at 0)
  if(is.prior.factor(prior_parameter)){
    # For factor priors, create a factor spike
    priors_type <- .get_prior_factor_list_type(list(prior_parameter))
    contrast_type <- gsub("prior.", "", priors_type[["class"]], fixed = TRUE)
    
    spike_component <- prior_factor(
      distribution = "point",
      parameters   = list(location = 0),
      contrast     = contrast_type
    )
  } else {
    # For simple priors, create a simple spike
    spike_component <- prior(
      distribution = "point", 
      parameters   = list(location = 0)
    )
  }
  
  # Create the mixture using the mixture backend
  mixture_output <- prior_mixture(
    prior_list = list(prior_parameter, spike_component),
    components = c("alternative", "null")
  )
  
  # Store inclusion prior as attribute so it can be retrieved by helper functions
  attr(mixture_output, "inclusion_prior") <- prior_inclusion
  attr(mixture_output, "model_prior_weights") <- prior_weights
  
  # Add spike_and_slab classes for specialized behavior while keeping mixture functionality
  if(is.prior.factor(prior_parameter)){
    # obtain and store the contrast type
    priors_type <- .get_prior_factor_list_type(list(prior_parameter))
    
    attr(prior_parameter, "K") <- priors_type[["K"]]
    class(mixture_output) <- c("prior", "prior.spike_and_slab", "prior.factor_spike_and_slab", 
                              class(mixture_output)[-1], priors_type[["class"]])
  }else if(is.prior.simple(prior_parameter)){
    class(mixture_output) <- c("prior", "prior.spike_and_slab", "prior.simple_spike_and_slab", 
                              class(mixture_output)[-1])
  }else{
    stop("The 'prior_parameter' must be either a simple or factor prior distribution.")
  }

  return(mixture_output)
}

# Helper functions to extract variable and inclusion from spike_and_slab mixture structure
.get_spike_and_slab_variable <- function(spike_and_slab_prior) {
  if (!is.prior.spike_and_slab(spike_and_slab_prior)) {
    stop("This function only works with spike_and_slab priors")
  }
 
  # Find the alternative component (this is the variable/slab part)
  components    <- attr(spike_and_slab_prior, "components") 
  alternative_idx <- which(components == "alternative")
  
  return(spike_and_slab_prior[[alternative_idx]])
}

.get_spike_and_slab_inclusion <- function(spike_and_slab_prior) {
  if (!is.prior.spike_and_slab(spike_and_slab_prior)) {
    stop("This function only works with spike_and_slab priors")
  }
  
  # For backward compatibility, use stored inclusion if available
  if (!is.null(spike_and_slab_prior[["inclusion"]])) {
    return(spike_and_slab_prior[["inclusion"]])
  }
  
  # Get inclusion prior from attribute
  inclusion_prior <- attr(spike_and_slab_prior, "inclusion_prior")
  return(inclusion_prior)
}

# Setter functions to allow modifying variable and inclusion components
.set_spike_and_slab_variable_attr <- function(spike_and_slab_prior, attr_name, value) {
  if (!is.prior.spike_and_slab(spike_and_slab_prior)) {
    stop("This function only works with spike_and_slab priors")
  }
  
  # Find the alternative component (this is the variable/slab part)
  components <- attr(spike_and_slab_prior, "components") 
  alternative_idx <- which(components == "alternative")

  # Set attribute on the variable component
  attr(spike_and_slab_prior[[alternative_idx]], attr_name) <- value
  
  return(spike_and_slab_prior)
}




#' @title Creates a mixture of prior distributions
#' @description \code{prior_mixture} creates a mixture of prior distributions.
#' This is a more generic version of the \code{prior_spike_and_slab} function.
#'
#' @param prior_list a list of prior distributions to be mixed.
#' @param is_null a logical vector indicating which of the prior distributions
#' should be considered as a null distribution. Defaults to \code{rep(FALSE, length(prior_list))}.
#' @param components a character vector indicating which of the prior distributions
#' belong to the same mixture component (this is an alternative specification to the \code{is_null} argument).
#' Defaults to \code{NULL} (i.e., \code{is_null} is used.
#'
#' @seealso [prior()]
#' @export
prior_mixture <- function(prior_list, is_null = rep(FALSE, length(prior_list)), components = NULL){

  .check_prior_list(prior_list, allow_expressions = TRUE)
  check_bool(is_null, "is_null",       check_length = length(prior_list), allow_NULL = TRUE)
  check_char(components, "components", check_length = length(prior_list), allow_NULL = TRUE)
  if(is.null(is_null) && is.null(components))
    stop("Either 'is_null' or 'components' must be specified.")

  if(is.null(components)){
    components <- ifelse(is_null, "null", "alternative")
  }

  for(i in seq_along(prior_list)){
    attr(prior_list[[i]], "component") <- components[i]
  }


  # distinguish normal, factor, and publication bias mixture priors
  if(any(sapply(prior_list, is.prior.factor))){

    # test that the prior is either a factor prior or a spike prior
    if(!all(sapply(prior_list, is.prior.factor) | sapply(prior_list, is.prior.point) | sapply(prior_list, is.prior.none)))
      stop("Factor prior mixture requires that all priors are either factor priors or spike prior distributions")

    # obtain and store the contrast type
    priors_type <- .get_prior_factor_list_type(prior_list)

    # change prior none/spikes into factor prior spikes
    for(i in seq_along(prior_list)){
      if(is.prior.point(prior_list[[i]])){
        # Save mixture metadata before recreating the factor-compatible prior.
        component_attr <- attr(prior_list[[i]], "component")
        prior_weight   <- .prior_model_weight(prior_list[[i]])
        prior_list[[i]] <- prior_factor(
          distribution = "point",
          parameters   = list(location = prior_list[[i]][["parameters"]][["location"]]),
          prior_weights = prior_weight,
          contrast     = gsub("prior.", "", priors_type[["class"]], fixed = TRUE)
        )
        attr(prior_list[[i]], "component") <- component_attr
      }else if(is.prior.none(prior_list[[i]])){
        # Save mixture metadata before recreating the factor-compatible prior.
        component_attr <- attr(prior_list[[i]], "component")
        prior_weight   <- .prior_model_weight(prior_list[[i]])
        prior_list[[i]] <- prior_factor(
          distribution = "point",
          parameters   = list(location = 0),
          prior_weights = prior_weight,
          contrast     = gsub("prior.", "", priors_type[["class"]], fixed = TRUE)
        )
        attr(prior_list[[i]], "component") <- component_attr
      }
    }

    attr(prior_list, "K")  <- priors_type[["K"]]
    class(prior_list)      <- c("prior", "prior.factor_mixture", "prior.mixture", priors_type[["class"]])


  }else if(any(sapply(prior_list, is.prior.PET)) || any(sapply(prior_list, is.prior.PEESE)) ||
           any(sapply(prior_list, is.prior.weightfunction)) || any(sapply(prior_list, is_prior_phacking)) ||
           any(sapply(prior_list, is_prior_bias))){

    # test that the prior is either a PET, PEESE, weightfunction, p-hacking,
    # composed bias, or none prior
    if(!all(sapply(prior_list, is.prior.PET) | sapply(prior_list, is.prior.PEESE) |
            sapply(prior_list, is.prior.weightfunction) | sapply(prior_list, is_prior_phacking) |
            sapply(prior_list, is_prior_bias) | sapply(prior_list, is.prior.none)))
      stop("Publication-bias prior mixtures require PET, PEESE, weightfunction, p-hacking, composed bias, or none prior distributions.")

    class(prior_list) <- c("prior", "prior.bias_mixture", "prior.mixture")


  }else if(any(sapply(prior_list, is.prior.simple))){

    # test that all priors are simple priors
    if(!all(sapply(prior_list, is.prior.simple) | sapply(prior_list, is.prior.none)))
      stop("Simple prior mixture requires that all priors are simple prior distributions")

    # change none into prior spikes
    for(i in seq_along(prior_list)){
      if(is.prior.none(prior_list[[i]])){
        prior_weight <- .prior_model_weight(prior_list[[i]])
        prior_list[[i]] <- prior(
          distribution = "point",
          parameters   = list(location = 0),
          prior_weights = prior_weight
        )
      }
    }

    class(prior_list) <- c("prior", "prior.simple_mixture", "prior.mixture")

  }else{
    stop("The prior mixture must contain either factors, publication bias components, or simple prior distributions.")
  }

  attr(prior_list, "components")    <- components
  attr(prior_list, "prior_weights") <- sapply(prior_list, .prior_model_weight)

  return(prior_list)
}

#### functions for constructing prior distributions ####
.prior_normal    <- function(parameters, truncation){

  output <- list()

  # check overall settings
  parameters <- .check_and_name_parameters(parameters, c("mean", "sd"), "normal")
  truncation <- .check_and_set_truncation(truncation)

  # check individual parameters
  .check_parameter(parameters$mean, "mean")
  .check_parameter(parameters$sd,   "sd")
  .check_parameter_positive(parameters$sd, "sd")

  # add the values to the output
  output$distribution <- "normal"
  output$parameters   <- parameters
  output$truncation   <- truncation

  class(output) <- c("prior", "prior.simple")

  return(output)
}
.prior_lognormal <- function(parameters, truncation){

  output <- list()

  # check overall settings
  parameters <- .check_and_name_parameters(parameters, c("meanlog", "sdlog"), "lognormal")
  truncation <- .check_and_set_truncation(truncation, lower = 0)

  # check individual parameters
  .check_parameter(parameters$meanlog, "meanlog")
  .check_parameter(parameters$sdlog,   "sdlog")
  .check_parameter_positive(parameters$sdlog, "sdlog")

  # add the values to the output
  output$distribution <- "lognormal"
  output$parameters   <- parameters
  output$truncation   <- truncation

  class(output) <- c("prior", "prior.simple")

  return(output)
}
.prior_cauchy    <- function(parameters, truncation){

  output <- list()

  # check overall settings
  parameters <- .check_and_name_parameters(parameters, c("location", "scale"), "Cauchy")
  truncation <- .check_and_set_truncation(truncation)

  # check individual parameters
  .check_parameter(parameters$location, "location")
  .check_parameter(parameters$scale,    "scale")
  .check_parameter_positive(parameters$scale, "scale")

  # deal with as with a t-distribution
  parameters$df <- 1

  output$distribution <- "t"
  output$parameters   <- parameters
  output$truncation   <- truncation

  class(output) <- c("prior", "prior.simple")

  return(output)
}
.prior_t         <- function(parameters, truncation){

  output <- list()

  # check overall settings
  parameters <- .check_and_name_parameters(parameters, c("location", "scale", "df"), "student-t")
  truncation <- .check_and_set_truncation(truncation)

  # check individual parameters
  .check_parameter(parameters$location, "location")
  .check_parameter(parameters$scale,    "scale")
  .check_parameter(parameters$df,       "df")
  .check_parameter_positive(parameters$scale, "scale")
  .check_parameter_positive(parameters$df,    "df")

  # add the values to the output
  output$distribution <- "t"
  output$parameters   <- parameters
  output$truncation   <- truncation

  class(output) <- c("prior", "prior.simple")

  return(output)
}
.prior_gamma     <- function(parameters, truncation){

  output <- list()

  # deal with possible scale parametrization
  if(!is.null(names(parameters))){
    if("scale" %in% names(parameters)){
      parameters <- .check_and_name_parameters(parameters, c("shape", "scale"), "gamma")
      parameters$rate  <- 1/parameters$scale
      parameters$scale <- NULL
    }
  }

  # check overall settings
  parameters <- .check_and_name_parameters(parameters, c("shape", "rate"), "gamma")
  truncation <- .check_and_set_truncation(truncation, lower = 0)

  # check individual parameters
  .check_parameter(parameters$shape, "shape")
  .check_parameter(parameters$rate,  "rate")
  .check_parameter_positive(parameters$shape, "shape")
  .check_parameter_positive(parameters$rate,  "rate")

  # add the values to the output
  output$distribution <- "gamma"
  output$parameters   <- parameters
  output$truncation   <- truncation

  class(output) <- c("prior", "prior.simple")

  return(output)
}
.prior_invgamma  <- function(parameters, truncation){

  output <- list()

  # check overall settings
  parameters <- .check_and_name_parameters(parameters, c("shape", "scale"), "invgamma")
  truncation <- .check_and_set_truncation(truncation, lower = 0)

  # check individual parameters
  .check_parameter(parameters$shape, "shape")
  .check_parameter(parameters$scale, "scale")
  .check_parameter_positive(parameters$shape, "shape")
  .check_parameter_positive(parameters$scale, "scale")

  # add the values to the output
  output$distribution <- "invgamma"
  output$parameters   <- parameters
  output$truncation   <- truncation

  class(output) <- c("prior", "prior.simple")

  return(output)
}
.prior_exp       <- function(parameters, truncation){

  output <- list()

  # deal with possible scale parametrization
  if(!is.null(names(parameters))){
    if("scale" %in% names(parameters)){
      parameters <- .check_and_name_parameters(parameters, c("scale"), "exp")
      parameters$rate  <- 1/parameters$scale
      parameters$scale <- NULL
    }
  }

  # check overall settings
  parameters <- .check_and_name_parameters(parameters, c("rate"), "exp")
  truncation <- .check_and_set_truncation(truncation, lower = 0)

  # check individual parameters
  .check_parameter(parameters$rate, "rate")
  .check_parameter_positive(parameters$rate, "rate")

  # add the values to the output
  output$distribution <- "exp"
  output$parameters   <- parameters
  output$truncation   <- truncation

  class(output) <- c("prior", "prior.simple")

  return(output)
}
.prior_beta      <- function(parameters, truncation){

  output <- list()

  # check overall settings
  parameters <- .check_and_name_parameters(parameters, c("alpha", "beta"), "beta")
  truncation <- .check_and_set_truncation(truncation, lower = 0, upper = 1)

  # check individual parameters
  .check_parameter(parameters$alpha, "alpha")
  .check_parameter(parameters$beta,  "beta")
  .check_parameter_positive(parameters$alpha, "alpha")
  .check_parameter_positive(parameters$beta,  "beta")

  # add the values to the output
  output$distribution <- "beta"
  output$parameters   <- parameters
  output$truncation   <- truncation

  class(output) <- c("prior", "prior.simple")

  return(output)
}
.prior_bernoulli <- function(parameters, truncation){

  output <- list()

  # check overall settings
  parameters <- .check_and_name_parameters(parameters, "probability", "bernoulli")
  truncation <- .check_and_set_truncation(truncation, lower = 0, upper = 1)

  # check individual parameters
  .check_parameter(parameters$probability, "probability")
  .check_parameter_range(parameters$probability, "probability", lower = 0, upper = 1, include_bounds = TRUE)
  bernoulli_support <- c(0, 1)
  bernoulli_keep    <- bernoulli_support >= truncation[["lower"]] & bernoulli_support <= truncation[["upper"]]
  if(!any(bernoulli_keep))
    stop("Bernoulli truncation must contain at least one support point: 0 or 1.", call. = FALSE)
  if(!is.expression(parameters$probability)){
    bernoulli_prob <- c(1 - parameters$probability, parameters$probability)
    if(sum(bernoulli_prob[bernoulli_keep]) <= 0)
      stop("Bernoulli truncation must retain positive probability mass.", call. = FALSE)
  }

  # add the values to the output
  output$distribution <- "bernoulli"
  output$parameters   <- parameters
  output$truncation   <- truncation

  class(output) <- c("prior", "prior.simple", "prior.discrete")

  return(output)
}
.prior_uniform   <- function(parameters, truncation){

  output <- list()

  # check overall settings
  parameters <- .check_and_name_parameters(parameters, c("a", "b"), "uniform")

  # check individual parameters
  .check_parameter(parameters$a, "a")
  .check_parameter(parameters$b, "b")

  if(parameters$a >= parameters$b)
    stop("Parameter 'a' must be lower than the parameter 'b'.")

  truncation <- .check_and_set_truncation(truncation, lower = parameters$a, upper = parameters$b)

  # add the values to the output
  output$distribution <- "uniform"
  output$parameters   <- parameters
  output$truncation   <- truncation

  class(output) <- c("prior", "prior.simple")

  return(output)
}
.prior_point     <- function(parameters, truncation){

  output <- list()

  # check overall settings
  parameters <- .check_and_name_parameters(parameters, c("location"), "point")

  # check individual parameters
  .check_parameter(parameters$location, "location")
  .check_point_truncation(truncation, parameters$location, "point")

  # add the values to the output
  output$distribution <- "point"
  output$parameters   <- parameters
  output$truncation   <- list(lower = parameters$location, upper = parameters$location)

  class(output) <- c("prior", "prior.simple", "prior.point")

  return(output)
}
.prior_mnormal   <- function(parameters, truncation){

  output <- list()

  # check overall settings
  parameters <- .check_and_name_parameters(parameters, c("mean", "sd", "K"), "multivariate normal")
  truncation <- .check_and_set_truncation(truncation)
  .check_vector_truncation_unsupported(truncation)

  # check individual parameters
  .check_parameter(parameters$mean, "mean")
  .check_parameter(parameters$sd,   "sd")
  .check_parameter_positive(parameters$sd, "sd")
  .check_parameter_dimensions(parameters$K, "K", allow_NA = TRUE)   # allow undetermined dimensions if called by prior_factor

  # add the values to the output
  output$distribution <- "mnormal"
  output$parameters   <- parameters
  output$truncation   <- truncation

  class(output) <- c("prior", "prior.vector")

  return(output)
}
.prior_mcauchy   <- function(parameters, truncation){

  output <- list()

  # check overall settings
  parameters <- .check_and_name_parameters(parameters, c("location", "scale", "K"), "multivariate Cauchy")
  truncation <- .check_and_set_truncation(truncation)
  .check_vector_truncation_unsupported(truncation)

  # check individual parameters
  .check_parameter(parameters$location, "location")
  .check_parameter(parameters$scale,    "scale")
  .check_parameter_positive(parameters$scale, "scale")
  .check_parameter_dimensions(parameters$K,   "K", allow_NA = TRUE)   # allow undetermined dimensions if called by prior_factor

  # deal with as with a t-distribution
  parameters$df <- 1

  output$distribution <- "mt"
  output$parameters   <- parameters
  output$truncation   <- truncation

  class(output) <- c("prior", "prior.vector")

  return(output)
}
.prior_mt        <- function(parameters, truncation){

  output <- list()

  # check overall settings
  parameters <- .check_and_name_parameters(parameters, c("location", "scale", "df", "K"), "multivariate student-t")
  truncation <- .check_and_set_truncation(truncation)
  .check_vector_truncation_unsupported(truncation)

  # check individual parameters
  .check_parameter(parameters$location, "location")
  .check_parameter(parameters$scale,    "scale")
  .check_parameter(parameters$df,       "df")
  .check_parameter_positive(parameters$scale, "scale")
  .check_parameter_positive(parameters$df,    "df")
  .check_parameter_dimensions(parameters$K,   "K", allow_NA = TRUE)   # allow undetermined dimensions if called by prior_factor


  # add the values to the output
  output$distribution <- "mt"
  output$parameters   <- parameters
  output$truncation   <- truncation

  class(output) <- c("prior", "prior.vector")

  return(output)
}
.prior_mpoint    <- function(parameters, truncation){

  output <- list()

  # check overall settings
  parameters <- .check_and_name_parameters(parameters, c("location", "K"), "multivariate point")

  # check individual parameters
  .check_parameter(parameters$location, "location")
  .check_parameter_dimensions(parameters$K, "K", allow_NA = TRUE)   # allow undetermined dimensions if called by prior_factor
  .check_point_truncation(truncation, parameters$location, "multivariate point")

  # add the values to the output
  output$distribution <- "mpoint"
  output$parameters   <- parameters
  output$truncation   <- list(lower = parameters$location, upper = parameters$location)

  class(output) <- c("prior", "prior.vector", "prior.point")

  return(output)
}
.check_point_truncation <- function(truncation, location, distribution){

  if(is.expression(location)){
    return(invisible(NULL))
  }

  if(length(truncation) > 2)
    stop("More than two truncation points were supplied.", call. = FALSE)

  if(!is.null(names(truncation))){
    if(!all(names(truncation) %in% c("lower", "upper")))
      stop("Truncation points must be named 'lower' and 'upper'.", call. = FALSE)
    if(length(truncation) == 1){
      if(names(truncation) == "lower"){
        truncation$upper <- Inf
      }else if(names(truncation) == "upper"){
        truncation$lower <- -Inf
      }
    }
  }else{
    if(length(truncation) == 2){
      names(truncation) <- c("lower", "upper")
    }else if(length(truncation) == 1){
      names(truncation) <- "lower"
      truncation$upper <- Inf
    }else{
      truncation <- list(lower = -Inf, upper = Inf)
    }
  }

  truncation <- truncation[c("lower", "upper")]

  if(!is.numeric(truncation[["lower"]]) || length(truncation[["lower"]]) != 1 ||
     !is.numeric(truncation[["upper"]]) || length(truncation[["upper"]]) != 1){
    stop("Truncation points must be numeric values.", call. = FALSE)
  }

  if(truncation[["lower"]] > truncation[["upper"]]){
    stop("The lower truncation point must be lower or equal to the upper truncation point.", call. = FALSE)
  }

  if(truncation[["lower"]] > location || truncation[["upper"]] < location){
    stop(paste0("The ", distribution, " prior truncation must contain the point location."), call. = FALSE)
  }

  invisible(NULL)
}
.check_vector_truncation_unsupported <- function(truncation){

  if(!is.infinite(truncation[["lower"]]) || !is.infinite(truncation[["upper"]])){
    stop("Vector priors do not support truncation.", call. = FALSE)
  }
}
.weightfunction_normalize_side <- function(side){

  side_clean <- .prior_clean_input_name(side)
  if(side_clean %in% c("onesided", "one")){
    return("one-sided")
  }
  if(side_clean %in% c("twosided", "two")){
    return("two-sided")
  }

  stop("'side' must be either 'one-sided' or 'two-sided'.", call. = FALSE)
}
.weightfunction_validate_steps <- function(steps){

  check_real(steps, "steps", check_length = 0, allow_NA = FALSE)

  if(length(steps) == 0){
    stop("'steps' must contain at least one p-value cut point.", call. = FALSE)
  }
  if(any(steps >= 1) || any(steps <= 0)){
    stop("'steps' must be higher than 0 and lower than 1.", call. = FALSE)
  }
  if(anyDuplicated(steps)){
    stop("'steps' must not contain duplicate cut points.", call. = FALSE)
  }
  if(!all(steps == cummax(steps))){
    stop("'steps' must be monotonically increasing.", call. = FALSE)
  }

  steps
}
.weightfunction_bins <- function(steps){

  cuts <- c(0, steps, 1)
  data.frame(
    lower     = cuts[-length(cuts)],
    upper     = cuts[-1],
    reference = c(TRUE, rep(FALSE, length(cuts) - 2L))
  )
}
.weightfunction_validate_weights <- function(weights, n_bins, reference){

  if(!inherits(weights, "weightfunction_weights")){
    stop("'weights' must be created by wf_cumulative(), wf_fixed(), or wf_independent().", call. = FALSE)
  }

  if(weights$type == "cumulative"){
    if(is.null(weights$alpha)){
      weights$alpha <- rep(1, n_bins)
    }
    check_real(weights$alpha, "alpha", lower = 0, allow_bound = FALSE, check_length = n_bins, allow_NA = FALSE)

  }else if(weights$type == "fixed"){
    check_real(weights$omega, "omega", lower = 0, check_length = n_bins, allow_NA = FALSE)
    if(reference == "most_significant" && !isTRUE(all.equal(weights$omega[1], 1))){
      stop("The reference-bin fixed weight must be exactly 1.", call. = FALSE)
    }

  }else if(weights$type == "independent"){
    .check_prior(weights$prior)
    if(!is.prior.simple(weights$prior) || is.prior.point(weights$prior) || is.prior.discrete(weights$prior)){
      stop("'weights$prior' must be a continuous simple prior distribution.", call. = FALSE)
    }
    if(weights$scale == "omega"){
      if(weights$prior$truncation[["lower"]] < 0){
        stop("Independent omega-scale weight priors must have non-negative support.", call. = FALSE)
      }
    }else if(weights$scale != "log_omega"){
      stop("Unsupported independent weight prior scale.", call. = FALSE)
    }

  }else{
    stop("Unsupported weightfunction weight prior type.", call. = FALSE)
  }

  weights
}
.weightfunction_weights_truncation <- function(weights){

  if(weights$type == "cumulative"){
    return(list(lower = 0, upper = 1))
  }

  if(weights$type == "fixed"){
    return(list(lower = 0, upper = max(1, weights$omega, na.rm = TRUE)))
  }

  if(weights$type == "independent"){
    if(weights$scale == "omega"){
      upper <- weights$prior$truncation[["upper"]]
    }else if(weights$scale == "log_omega"){
      upper <- weights$prior$truncation[["upper"]]
      upper <- if(is.infinite(upper)) Inf else exp(upper)
    }else{
      stop("Unsupported independent weight prior scale.", call. = FALSE)
    }

    return(list(lower = 0, upper = max(1, upper)))
  }

  stop("Unsupported weightfunction weight prior type.", call. = FALSE)
}
.weightfunction_n_bins <- function(prior){

  if(!is.prior.weightfunction(prior)){
    stop("'prior' must be a weightfunction prior.", call. = FALSE)
  }

  nrow(prior$bins)
}
.weightfunction_is_fixed <- function(prior){
  prior$weights$type == "fixed"
}
.weightfunction_local_cuts <- function(prior){

  c(prior$bins$lower[1], prior$bins$upper)
}
.weightfunction_reference_index <- function(prior){
  which(prior$bins$reference)
}
.weightfunction_alpha_marginal <- function(alpha, index){

  if(index <= 1L){
    return(list(type = "point", location = 1))
  }

  list(
    type  = "beta",
    alpha = sum(alpha[index:length(alpha)]),
    beta  = sum(alpha[seq_len(index - 1L)])
  )
}
.weightfunction_rng <- function(prior, n){

  J <- .weightfunction_n_bins(prior)

  if(prior$weights$type == "fixed"){
    out <- matrix(rep(prior$weights$omega, each = n), nrow = n)

  }else if(prior$weights$type == "cumulative"){
    theta <- extraDistr::rdirichlet(n, alpha = prior$weights$alpha)
    out   <- t(apply(theta[,J:1, drop = FALSE], 1, cumsum))[,J:1, drop = FALSE]
    # Keep the reference bin exact; row-wise cumulative sums can leave 1 +/- eps.
    out[,1L] <- 1

  }else if(prior$weights$type == "independent"){
    out <- matrix(1, nrow = n, ncol = J)
    if(J > 1L){
      draws <- matrix(rng(prior$weights$prior, n * (J - 1L)), nrow = n, ncol = J - 1L)
      if(prior$weights$scale == "log_omega"){
        draws <- exp(draws)
      }
      out[,2:J] <- draws
    }
  }

  colnames(out) <- paste0("omega[", seq_len(J), "]")
  out
}
.weightfunction_marginal_components <- function(prior){

  J <- .weightfunction_n_bins(prior)

  if(prior$weights$type == "fixed"){
    return(lapply(prior$weights$omega, function(x){
      list(type = "point", location = x)
    }))
  }

  if(prior$weights$type == "cumulative"){
    return(lapply(seq_len(J), function(j){
      .weightfunction_alpha_marginal(prior$weights$alpha, j)
    }))
  }

  if(prior$weights$type == "independent"){
    return(lapply(seq_len(J), function(j){
      if(j == 1L){
        list(type = "point", location = 1)
      }else{
        list(type = "prior", prior = prior$weights$prior, scale = prior$weights$scale)
      }
    }))
  }
}
.prior_weightfunction_component_range <- function(component, quantiles = .005){

  switch(
    component$type,
    "point" = c(component$location, component$location),
    "beta"  = c(0, 1),
    "prior" = {
      if(component$scale == "omega"){
        lower <- component$prior$truncation[["lower"]]
        upper <- component$prior$truncation[["upper"]]

        lower <- if(is.infinite(lower)) mquant(component$prior, quantiles) else lower
        upper <- if(is.infinite(upper)) mquant(component$prior, 1 - quantiles) else upper
        c(lower, upper)
      }else{
        lower <- component$prior$truncation[["lower"]]
        upper <- component$prior$truncation[["upper"]]

        lower <- if(is.infinite(lower)) 0 else exp(lower)
        upper <- if(is.infinite(upper)) exp(mquant(component$prior, 1 - quantiles)) else exp(upper)
        c(lower, upper)
      }
    }
  )
}
.weightfunction_range <- function(prior, quantiles = .005){

  ranges <- do.call(rbind, lapply(
    .weightfunction_marginal_components(prior),
    .prior_weightfunction_component_range,
    quantiles = quantiles
  ))

  x_range <- range(c(0, 1, as.vector(ranges)), finite = TRUE)
  if(x_range[1] == x_range[2]){
    x_range <- range(c(0, 1, x_range), finite = TRUE)
  }

  x_range
}
.prior_weightfunction_component_cdf <- function(component, q){

  switch(
    component$type,
    "point" = ppoint(q, location = component$location),
    "beta"  = stats::pbeta(q, shape1 = component$alpha, shape2 = component$beta),
    "prior" = {
      if(component$scale == "omega"){
        mcdf(component$prior, q)
      }else{
        p <- numeric(length(q))
        p[q <= 0] <- 0
        inside <- q > 0
        p[inside] <- mcdf(component$prior, log(q[inside]))
        p
      }
    }
  )
}
.prior_weightfunction_component_ccdf <- function(component, q){
  if(component$type == "point"){
    return(ppoint(q, location = component$location, lower.tail = FALSE))
  }
  1 - .prior_weightfunction_component_cdf(component, q)
}
.prior_weightfunction_component_lpdf <- function(component, x){

  switch(
    component$type,
    "point" = dpoint(x, location = component$location, log = TRUE),
    "beta"  = stats::dbeta(x, shape1 = component$alpha, shape2 = component$beta, log = TRUE),
    "prior" = {
      if(component$scale == "omega"){
        mlpdf(component$prior, x)
      }else{
        out <- rep(-Inf, length(x))
        inside <- x > 0
        out[inside] <- mlpdf(component$prior, log(x[inside])) - log(x[inside])
        out
      }
    }
  )
}
.prior_weightfunction_component_quant <- function(component, p){

  switch(
    component$type,
    "point" = qpoint(p, location = component$location),
    "beta"  = stats::qbeta(p, shape1 = component$alpha, shape2 = component$beta),
    "prior" = {
      if(component$scale == "omega"){
        mquant(component$prior, p)
      }else{
        exp(mquant(component$prior, p))
      }
    }
  )
}
.prior_weightfunction_component_mean <- function(component){

  switch(
    component$type,
    "point" = component$location,
    "beta"  = component$alpha / (component$alpha + component$beta),
    "prior" = {
      if(component$scale == "omega"){
        mean(component$prior)
      }else{
        stats::integrate(
          f     = function(x, prior) {
            y <- exp(x) * pdf(prior, x)
            y[!is.finite(y)] <- 0
            y
          },
          lower = component$prior$truncation[["lower"]],
          upper = component$prior$truncation[["upper"]],
          prior = component$prior
        )$value
      }
    }
  )
}
.prior_weightfunction_component_var <- function(component){

  switch(
    component$type,
    "point" = 0,
    "beta"  = (component$alpha * component$beta) /
      ((component$alpha + component$beta)^2 * (component$alpha + component$beta + 1)),
    "prior" = {
      if(component$scale == "omega"){
        var(component$prior)
      }else{
        m1 <- .prior_weightfunction_component_mean(component)
        m2 <- stats::integrate(
          f     = function(x, prior) {
            y <- exp(2 * x) * pdf(prior, x)
            y[!is.finite(y)] <- 0
            y
          },
          lower = component$prior$truncation[["lower"]],
          upper = component$prior$truncation[["upper"]],
          prior = component$prior
        )$value
        m2 - m1^2
      }
    }
  )
}

#### elementary prior related functions ####
#' @title Elementary prior related functions
#'
#' @description Density (pdf / lpdf), distribution
#' function (cdf / ccdf), quantile function (quant),
#' random generation (rng), mean, standard deviation (sd),
#' and marginal variants of the functions (mpdf, mlpf, mcdf,
#' mccdf, mquant) for prior distributions.
#'
#' @param x prior distribution
#' @param y vector of observations
#' @param q vector or matrix of quantiles
#' @param p vector of probabilities
#' @param n number of observations
#' @param ... unused arguments
#'
#' @examples
#' # create a standard normal prior distribution
#' p1 <- prior(distribution = "normal", parameters = list(mean = 1, sd = 1))
#'
#' # generate a random sample from the prior
#' rng(p1, 10)
#'
#' # compute cumulative density function
#' cdf(p1, 0)
#'
#' # obtain quantile
#' quant(p1, .5)
#'
#' # compute probability density
#' pdf(p1, c(0, 1, 2))
#'
#' @return \code{pdf} (\code{mpdf}) and \code{lpdf} (\code{mlpdf}) give
#' the (marginal) density and the log of (marginal) density,
#' \code{cdf} (\code{mcdf}) and \code{ccdf} (\code{mccdf}) give the
#' (marginal) distribution and the complement of (marginal) distribution function,
#' \code{quant} (\code{mquant}) give the (marginal) quantile function,
#' and \code{rng} generates random deviates for an object of class 'prior'.
#'
#' @exportS3Method rng prior
#' @exportS3Method cdf prior
#' @exportS3Method ccdf prior
#' @exportS3Method quant prior
#' @exportS3Method lpdf prior
#' @exportS3Method pdf prior
#' @exportS3Method mcdf prior
#' @exportS3Method mccdf prior
#' @exportS3Method mquant prior
#' @exportS3Method mlpdf prior
#' @exportS3Method mpdf prior
#' @name prior_functions
NULL


#### joint distribution functions ####
#' @rdname prior_functions
rng.prior   <- function(x, n, ...){

  prior <- x

  .check_n(n)
  .check_prior(prior)

  dots  <- list(...)
  if(!is.null(dots[["transform_factor_samples"]])){
    check_bool(dots[["transform_factor_samples"]], "transform_factor_samples")
    transform_factor_samples <- dots[["transform_factor_samples"]]
  }else{
    transform_factor_samples <- TRUE
  }
  if(!is.null(dots[["sample_components"]])){
    check_bool(dots[["sample_components"]], "sample_components")
    sample_components <- dots[["sample_components"]]
  }else{
    sample_components <- FALSE
  }

  if(is.prior.spike_and_slab(prior)){

    inclusion <- stats::rbinom(n, size = 1, prob = rng(.get_spike_and_slab_inclusion(prior), n))

    if(sample_components)
      return(inclusion)

    x         <- rng(.get_spike_and_slab_variable(prior), n) * inclusion
    attr(x, "inclusion") <- inclusion

  }else if(is.prior.mixture(prior)){

    component_probabilities <- attr(prior, "prior_weights")
    components              <- sample(seq_along(component_probabilities), size = n, replace = TRUE, prob = component_probabilities)

    if(sample_components)
      return(components)

    if(inherits(prior, "prior.bias_mixture")){

      branch_info <- lapply(prior, .selection_branch_info)
      spec        <- selection_backend_spec(prior)

      is_PET        <- sapply(prior, is.prior.PET)
      is_PEESE      <- sapply(prior, is.prior.PEESE)
      has_selection <- vapply(branch_info, function(x) !is.null(x$selection), logical(1))
      has_phacking  <- vapply(branch_info, function(x) !is.null(x$phacking),  logical(1))

      out_names <- character()
      if(any(has_selection)){
        omega_names <- paste0("omega[", seq_len(spec$step$n_bins), "]")
        out_names   <- c(out_names, omega_names)

        selection_priors <- lapply(branch_info[has_selection], function(x) x$selection)
        omega_mapping    <- weightfunctions_mapping(selection_priors, one_sided = TRUE)
        selection_index  <- which(has_selection)
      }
      if(any(has_phacking)){
        out_names <- c(out_names, "alpha", "pi_null")
      }
      if(any(is_PET)){
        out_names <- c(out_names, "PET")
      }
      if(any(is_PEESE)){
        out_names <- c(out_names, "PEESE")
      }

      x <- matrix(0, nrow = n, ncol = length(out_names))
      colnames(x) <- out_names
      if(any(has_selection)){
        x[, omega_names] <- 1
      }

      for(component in unique(components)){
        component_rows <- component == components
        n_component    <- sum(component_rows)

        if(has_selection[component]){
          selection_i <- match(component, selection_index)
          selection_samples <- rng(branch_info[[component]]$selection, n_component)
          x[component_rows, omega_names] <- selection_samples[, paste0("omega[", omega_mapping[[selection_i]], "]"), drop = FALSE]
        }

        if(has_phacking[component]){
          phacking_samples <- rng(branch_info[[component]]$phacking, n_component)
          x[component_rows, c("alpha", "pi_null")] <- phacking_samples[, c("alpha", "pi_null"), drop = FALSE]
        }

        if(is_PET[component]){
          x[component_rows, "PET"] <- rng(prior[[component]], n_component, transform_factor_samples = FALSE)
        }
        if(is_PEESE[component]){
          x[component_rows, "PEESE"] <- rng(prior[[component]], n_component, transform_factor_samples = FALSE)
        }
      }

    }else if(inherits(prior, "prior.factor_mixture")){

      prior_type <- .get_prior_factor_list_type(prior)

      if(transform_factor_samples){
        x <- matrix(NA, nrow = n, ncol = prior_type[["K"]] + 1)
      }else{
        x <- matrix(NA, nrow = n, ncol = prior_type[["K"]])
      }

      for(component in unique(components)){
        x[component == components,] <- rng(prior[[component]], sum(component == components), transform_factor_samples = transform_factor_samples)
      }

    }else if(inherits(prior, "prior.simple_mixture")){

      x <- rep(NA, n)
      for(component in unique(components)){
        x[component == components] <- rng(prior[[component]], sum(component == components))
      }

    }else{
      stop("unsupported prior mixture type")
    }


    attr(x, "components") <- components

  }else if(is.prior.simple(prior)){

    x <- .prior_simple_rng(prior, n)

  }else if(transform_factor_samples && (is.prior.orthonormal(prior) | is.prior.meandif(prior))){

    par1 <- switch(
      prior[["distribution"]],
      "mnormal" = prior$parameter[["mean"]],
      "mt"      = prior$parameter[["location"]],
      "mpoint"  = prior$parameter[["location"]]
    )

    if(length(par1) != 1){
      stop("unsported distribution specification in 'rng' -- non-symmetric")
    }

    if(par1 != 0){
      stop("the orthonormal/meandif prior distribution must be centered")
    }

    if(is.na(prior$parameters[["K"]]) && !is.null(attr(prior, "levels"))){
      prior$parameters[["K"]] <- .get_prior_factor_levels(prior)
    }else if(is.na(prior$parameters[["K"]])){
      prior$parameters[["K"]] <- 1
      warning("number of factor levels / dimensionality of the prior distribution was not specified -- assuming two factor levels")
    }

    par1 <- rep(0, prior$parameter[["K"]])

    if(prior[["distribution"]] %in% c("mnormal", "mt")){
      par2 <- diag(switch(
        prior[["distribution"]],
        "mnormal" = prior$parameter[["sd"]]^2,
        "mt"      = prior$parameter[["scale"]]^2
      ), ncol = prior$parameter[["K"]], nrow = prior$parameter[["K"]])
    }

    x <- switch(
      prior[["distribution"]],
      "mnormal"    = mvtnorm::rmvnorm(n, mean = par1, sigma = par2),
      "mt"         = mvtnorm::rmvt(n, delta = par1, sigma = par2, df = prior$parameter[["df"]], type = "shifted"),
      "mpoint"     = rmpoint(n, location = par1)
    )


    if(is.prior.orthonormal(prior)){
      x <- x %*% t(contr.orthonormal(1:(prior$parameters[["K"]] + 1)))
    }else if(is.prior.meandif(prior)){
      x <- x %*% t(contr.meandif(1:(prior$parameters[["K"]] + 1)))
    }

  }else if(is.prior.vector(prior)){

    if(prior[["distribution"]] != "mpoint")
      .check_vector_truncation_unsupported(prior$truncation)

    par1 <- switch(
      prior[["distribution"]],
      "mnormal" = prior$parameter[["mean"]],
      "mt"      = prior$parameter[["location"]],
      "mpoint"  = prior$parameter[["location"]]
    )

    if(length(par1) != 1){
      stop("unsported distribution specification in 'rng' -- non-symmetric")
    }

    # TODO: generalize this to priors with covariances
    par1 <- rep(par1, length = prior$parameter[["K"]])

    if(prior[["distribution"]] %in% c("mnormal", "mt")){
      par2 <- diag(switch(
        prior[["distribution"]],
        "mnormal" = prior$parameter[["sd"]]^2,
        "mt"      = prior$parameter[["scale"]]^2
      ), ncol = prior$parameter[["K"]], nrow = prior$parameter[["K"]])
    }

    x <- switch(
      prior[["distribution"]],
      "mnormal"    = mvtnorm::rmvnorm(n, mean = par1, sigma = par2),
      "mt"         = mvtnorm::rmvt(n, delta = par1, sigma = par2, df = prior$parameter[["df"]], type = "shifted"),
      "mpoint"     = rmpoint(n, location = par1)
    )

  }else if(is.prior.weightfunction(prior)){

    x <- .weightfunction_rng(prior, n)

  }else if(is_prior_phacking(prior)){

    alpha <- rng(prior$alpha, n)
    x <- cbind(
      alpha   = alpha,
      pi_null = phack_pi_null(alpha, prior$form, prior$source, prior$destination, target = prior$target)
    )

  }else if(is_prior_bias(prior)){

    selection_backend_spec(prior)

    x <- NULL
    if(!is.null(prior$selection)){
      x <- cbind(x, rng(prior$selection, n))
    }
    if(!is.null(prior$phacking)){
      x <- cbind(x, rng(prior$phacking, n))
    }

  }

  return(x)
}
#' @rdname prior_functions
cdf.prior   <- function(x, q, ...){

  prior <- x

  .check_q(as.vector(q))
  .check_prior(prior)

  if(is.prior.spike_and_slab(prior)){

    stop("No cdfs are implemented for spike and slab priors.")

  }else if(is.prior.mixture(prior)){

    stop("No cdfs are implemented for prior mixtures.")

  }else if(is.prior.simple(prior)){

    p <- .prior_simple_cdf(prior, q)

  }else if(is.prior.vector(prior)){

    stop("No cdfs are implemented for vector priors.")

  }else if(is.prior.weightfunction(prior)){

    stop("Only marginal cdfs are implemented for prior weightfunctions.")

  }else if(is_prior_phacking(prior) || is_prior_bias(prior)){

    .selection_prior_stop_unsupported_generic("cdf", prior)

  }

  return(p)
}
#' @rdname prior_functions
ccdf.prior  <- function(x, q, ...){

  prior <- x

  .check_q(as.vector(q))
  .check_prior(prior)

  if(is.prior.spike_and_slab(prior)){

    stop("No ccdf are implemented for spike and slab priors.")

  }else if(is.prior.mixture(prior)){

    stop("No ccdf are implemented for prior mixtures.")

  }else if(is.prior.simple(prior)){

    p <- .prior_simple_ccdf(prior, q)

  }else if(is.prior.vector(prior)){

    stop("No cdfs are implemented for vector priors.")

  }else if(is.prior.weightfunction(prior)){

    stop("Only marginal ccdf functions are implemented for prior weightfunctions.")

  }else if(is_prior_phacking(prior) || is_prior_bias(prior)){

    .selection_prior_stop_unsupported_generic("ccdf", prior)

  }

  return(p)
}
#' @rdname prior_functions
lpdf.prior  <- function(x, y, ...){

  prior <- x
  x     <- y

  .check_x(as.vector(x))
  .check_prior(prior)

  if(is.prior.spike_and_slab(prior)){

    stop("No lpdf are implemented for spike and slab priors.")

  }else if(is.prior.mixture(prior)){

    stop("No lpdf are implemented for prior mixtures.")

  }else if(is.prior.simple(prior)){

    log_lik <- .prior_simple_lpdf(prior, x)

  }else if(is.prior.vector(prior)){

    if(prior[["distribution"]] != "mpoint")
      .check_vector_truncation_unsupported(prior$truncation)

    par1 <- switch(
      prior[["distribution"]],
      "mnormal" = prior$parameter[["mean"]],
      "mt"      = prior$parameter[["location"]],
      "mpoint"  = prior$parameter[["location"]]
    )

    # TODO: generalize this to priors with covariances
    par1 <- rep(par1, length = prior$parameter[["K"]])

    if(prior[["distribution"]] %in% c("mnormal", "mt")){
      par2 <- diag(switch(
        prior[["distribution"]],
        "mnormal" = prior$parameter[["sd"]]^2,
        "mt"      = prior$parameter[["scale"]]^2
      ), ncol = prior$parameter[["K"]], nrow = prior$parameter[["K"]])
    }

    log_lik <- switch(
      prior[["distribution"]],
      "mnormal"    = mvtnorm::dmvnorm(x, mean = par1, sigma = par2, log = TRUE),
      "mt"         = mvtnorm::dmvt(x, delta = par1, sigma = par2, df = prior$parameter[["df"]], type = "shifted", log = TRUE),
      "mpoint"     = dmpoint(x, location = par1, log = TRUE)
    )

  }else if(is.prior.weightfunction(prior)){

    stop("Only marginal lpdf are implemented for prior weightfunctions.")

  }else if(is_prior_phacking(prior) || is_prior_bias(prior)){

    .selection_prior_stop_unsupported_generic("lpdf", prior)

  }

  return(log_lik)
}
#' @rdname prior_functions
pdf.prior   <- function(x, y, ...){

  prior <- x
  x     <- y

  .check_x(as.vector(x))
  .check_prior(prior)

  if(is.prior.simple(prior)){
    lik <- .prior_simple_pdf(prior, x)
  }else{
    log_lik <- lpdf.prior(prior, x)
    lik     <- exp(log_lik)
  }

  return(lik)
}
#' @rdname prior_functions
quant.prior <- function(x, p, ...){

  prior <- x

  .check_p(p, log.p = FALSE)
  .check_prior(prior)

  if(is.prior.spike_and_slab(prior)){

    stop("No quantile functions are implemented for spike and slab priors.")

  }else if(is.prior.mixture(prior)){

    stop("No quantile functions are implemented for prior mixtures.")

  }else if(is.prior.simple(prior)){

    q <- .prior_simple_quant(prior, p)

  }else if(is.prior.vector(prior) && !is.prior.factor(prior)){

    q <- mquant(prior, p)

  }else if(is.prior.weightfunction(prior)){

    stop("Only marginal quantile functions are implemented for prior weightfunctions.")

  }else if(is_prior_phacking(prior) || is_prior_bias(prior)){

    .selection_prior_stop_unsupported_generic("quantile functions", prior)

  }

  return(q)
}

.prior_simple_base_d <- function(prior, x, log = FALSE){

  switch(
    prior[["distribution"]],
    "normal"    = stats::dnorm(x, mean = prior$parameters[["mean"]], sd = prior$parameters[["sd"]], log = log),
    "lognormal" = stats::dlnorm(x, meanlog = prior$parameters[["meanlog"]], sdlog = prior$parameters[["sdlog"]], log = log),
    "t"         = extraDistr::dlst(x, df = prior$parameters[["df"]], mu = prior$parameters[["location"]], sigma = prior$parameters[["scale"]], log = log),
    "gamma"     = stats::dgamma(x, shape = prior$parameters[["shape"]], rate = prior$parameters[["rate"]], log = log),
    "invgamma"  = extraDistr::dinvgamma(x, alpha = prior$parameters[["shape"]], beta = prior$parameters[["scale"]], log = log),
    "beta"      = stats::dbeta(x, shape1 = prior$parameters[["alpha"]], shape2 = prior$parameters[["beta"]], log = log),
    "bernoulli" = stats::dbinom(x, size = 1, prob = prior$parameters[["probability"]], log = log),
    "exp"       = stats::dexp(x, rate = prior$parameters[["rate"]], log = log),
    "uniform"   = stats::dunif(x, min = prior$parameters[["a"]], max = prior$parameters[["b"]], log = log),
    "point"     = dpoint(x, location = prior$parameters[["location"]], log = log)
  )
}

.prior_simple_base_p <- function(prior, q, lower.tail = TRUE){

  switch(
    prior[["distribution"]],
    "normal"    = stats::pnorm(q, mean = prior$parameters[["mean"]], sd = prior$parameters[["sd"]], lower.tail = lower.tail, log.p = FALSE),
    "lognormal" = stats::plnorm(q, meanlog = prior$parameters[["meanlog"]], sdlog = prior$parameters[["sdlog"]], lower.tail = lower.tail, log.p = FALSE),
    "t"         = extraDistr::plst(q, df = prior$parameters[["df"]], mu = prior$parameters[["location"]], sigma = prior$parameters[["scale"]], lower.tail = lower.tail, log.p = FALSE),
    "gamma"     = stats::pgamma(q, shape = prior$parameters[["shape"]], rate = prior$parameters[["rate"]], lower.tail = lower.tail, log.p = FALSE),
    "invgamma"  = extraDistr::pinvgamma(q, alpha = prior$parameters[["shape"]], beta = prior$parameters[["scale"]], lower.tail = lower.tail, log.p = FALSE),
    "beta"      = stats::pbeta(q, shape1 = prior$parameters[["alpha"]], shape2 = prior$parameters[["beta"]], lower.tail = lower.tail, log.p = FALSE),
    "bernoulli" = stats::pbinom(q, size = 1, prob = prior$parameters[["probability"]], lower.tail = lower.tail, log.p = FALSE),
    "exp"       = stats::pexp(q, rate = prior$parameters[["rate"]], lower.tail = lower.tail, log.p = FALSE),
    "uniform"   = stats::punif(q, min = prior$parameters[["a"]], max = prior$parameters[["b"]], lower.tail = lower.tail, log.p = FALSE),
    "point"     = ppoint(q, location = prior$parameters[["location"]], lower.tail = lower.tail, log.p = FALSE)
  )
}

.prior_simple_base_q <- function(prior, p){

  switch(
    prior[["distribution"]],
    "normal"    = stats::qnorm(p, mean = prior$parameters[["mean"]], sd = prior$parameters[["sd"]], lower.tail = TRUE, log.p = FALSE),
    "lognormal" = stats::qlnorm(p, meanlog = prior$parameters[["meanlog"]], sdlog = prior$parameters[["sdlog"]], lower.tail = TRUE, log.p = FALSE),
    "t"         = extraDistr::qlst(p, df = prior$parameters[["df"]], mu = prior$parameters[["location"]], sigma = prior$parameters[["scale"]], lower.tail = TRUE, log.p = FALSE),
    "gamma"     = stats::qgamma(p, shape = prior$parameters[["shape"]], rate = prior$parameters[["rate"]], lower.tail = TRUE, log.p = FALSE),
    "invgamma"  = extraDistr::qinvgamma(p, alpha = prior$parameters[["shape"]], beta = prior$parameters[["scale"]], lower.tail = TRUE, log.p = FALSE),
    "beta"      = stats::qbeta(p, shape1 = prior$parameters[["alpha"]], shape2 = prior$parameters[["beta"]], lower.tail = TRUE, log.p = FALSE),
    "bernoulli" = stats::qbinom(p, size = 1, prob = prior$parameters[["probability"]], lower.tail = TRUE, log.p = FALSE),
    "exp"       = stats::qexp(p, rate = prior$parameters[["rate"]], lower.tail = TRUE, log.p = FALSE),
    "uniform"   = stats::qunif(p, min = prior$parameters[["a"]], max = prior$parameters[["b"]], lower.tail = TRUE, log.p = FALSE),
    "point"     = qpoint(p, location = prior$parameters[["location"]], lower.tail = TRUE, log.p = FALSE)
  )
}

.prior_simple_base_r <- function(prior, n){

  switch(
    prior[["distribution"]],
    "normal"    = stats::rnorm(n, mean = prior$parameters[["mean"]], sd = prior$parameters[["sd"]]),
    "lognormal" = stats::rlnorm(n, meanlog = prior$parameters[["meanlog"]], sdlog = prior$parameters[["sdlog"]]),
    "t"         = extraDistr::rlst(n, df = prior$parameters[["df"]], mu = prior$parameters[["location"]], sigma = prior$parameters[["scale"]]),
    "gamma"     = stats::rgamma(n, shape = prior$parameters[["shape"]], rate = prior$parameters[["rate"]]),
    "invgamma"  = extraDistr::rinvgamma(n, alpha = prior$parameters[["shape"]], beta = prior$parameters[["scale"]]),
    "beta"      = stats::rbeta(n, shape1 = prior$parameters[["alpha"]], shape2 = prior$parameters[["beta"]]),
    "bernoulli" = stats::rbinom(n, size = 1, prob = prior$parameters[["probability"]]),
    "exp"       = stats::rexp(n, rate = prior$parameters[["rate"]]),
    "uniform"   = stats::runif(n, min = prior$parameters[["a"]], max = prior$parameters[["b"]]),
    "point"     = rpoint(n, location = prior$parameters[["location"]])
  )
}

.prior_simple_discrete_support <- function(prior){

  switch(
    prior[["distribution"]],
    "bernoulli" = c(0, 1)
  )
}

.prior_simple_discrete_prob <- function(prior){

  switch(
    prior[["distribution"]],
    "bernoulli" = c(1 - prior$parameters[["probability"]], prior$parameters[["probability"]])
  )
}

.prior_simple_truncated_discrete <- function(prior){

  support <- .prior_simple_discrete_support(prior)
  prob    <- .prior_simple_discrete_prob(prior)
  keep    <- support >= prior$truncation[["lower"]] & support <= prior$truncation[["upper"]]

  support <- support[keep]
  prob    <- prob[keep]

  if(length(support) == 0 || sum(prob) <= 0){
    stop("The truncated discrete prior has no probability mass in its support.", call. = FALSE)
  }

  list(
    support = support,
    prob    = prob / sum(prob)
  )
}

.prior_simple_cdf <- function(prior, q){

  if(.is_prior_default_range(prior)){
    return(.prior_simple_base_p(prior, q, lower.tail = TRUE))
  }

  p        <- numeric(length(q))
  q_lower  <- q < prior$truncation[["lower"]]
  q_higher <- q > prior$truncation[["upper"]]
  q_inside <- !q_lower & !q_higher

  p[q_lower]  <- 0
  p[q_higher] <- 1

  if(any(q_inside)){
    p[q_inside] <- .prior_simple_base_p(prior, q[q_inside], lower.tail = TRUE)

    if(prior[["distribution"]] != "point"){
      C1          <- .prior_C1(prior)
      p[q_inside] <- (p[q_inside] - C1) / (.prior_C2(prior) - C1)
    }
  }

  return(p)
}

.prior_simple_ccdf <- function(prior, q){

  if(.is_prior_default_range(prior)){
    return(.prior_simple_base_p(prior, q, lower.tail = FALSE))
  }

  p        <- numeric(length(q))
  q_lower  <- q < prior$truncation[["lower"]]
  q_higher <- q > prior$truncation[["upper"]]
  q_inside <- !q_lower & !q_higher

  p[q_lower]  <- 1
  p[q_higher] <- 0

  if(any(q_inside)){
    p[q_inside] <- .prior_simple_base_p(prior, q[q_inside], lower.tail = FALSE)

    if(prior[["distribution"]] != "point"){
      C2          <- .prior_C2(prior)
      p[q_inside] <- (p[q_inside] - (1 - C2)) / (C2 - .prior_C1(prior))
    }
  }

  return(p)
}

.prior_simple_lpdf <- function(prior, x){

  if(.is_prior_default_range(prior)){
    return(.prior_simple_base_d(prior, x, log = TRUE))
  }

  log_lik <- .prior_simple_base_d(prior, x, log = TRUE)
  log_lik[x < prior$truncation[["lower"]] | x > prior$truncation[["upper"]]] <- -Inf

  if(prior[["distribution"]] != "point"){
    log_lik <- log_lik - log(.prior_C(prior))
  }

  return(log_lik)
}

.prior_simple_pdf <- function(prior, x){

  if(.is_prior_default_range(prior)){
    return(.prior_simple_base_d(prior, x, log = FALSE))
  }

  lik <- .prior_simple_base_d(prior, x, log = FALSE)
  lik[x < prior$truncation[["lower"]] | x > prior$truncation[["upper"]]] <- 0

  if(prior[["distribution"]] != "point"){
    lik <- lik / .prior_C(prior)
  }

  return(lik)
}

.prior_simple_quant <- function(prior, p){

  if(.is_prior_default_range(prior)){
    return(.prior_simple_base_q(prior, p))
  }

  if(is.prior.discrete(prior)){
    discrete <- .prior_simple_truncated_discrete(prior)
    cdf      <- cumsum(discrete[["prob"]])
    return(vapply(p, function(p_i) discrete[["support"]][which(cdf >= p_i)[1]], numeric(1)))
  }

  C1 <- .prior_C1(prior)
  .prior_simple_base_q(prior, C1 + p * (.prior_C2(prior) - C1))
}

.prior_simple_quant_optim <- function(prior, p){

  if(!is.infinite(prior$truncation[["lower"]]) & !is.infinite(prior$truncation[["upper"]])){
    start_value <- prior$truncation[["lower"]] + (prior$truncation[["upper"]]  - prior$truncation[["lower"]]) / 2
  }else if(!is.infinite(prior$truncation[["upper"]])){
    start_value <- prior$truncation[["upper"]] - 1
  }else if(!is.infinite(prior$truncation[["lower"]])){
    start_value <- prior$truncation[["lower"]] + 1
  }else{
    start_value <- 0
  }

  sapply(p, function(p_i){
    stats::optim(
      par     = start_value,
      fn      = function(x, prior, p_i)(.prior_simple_cdf(prior, x) - p_i)^2,
      lower   = prior$truncation[["lower"]],
      upper   = prior$truncation[["upper"]],
      prior   = prior,
      p_i     = p_i,
      method  = "L-BFGS-B",
      control = list(
        factr = 1e3
      )
    )$par
  })
}

.prior_simple_rng <- function(prior, n){

  if(.is_prior_default_range(prior)){
    return(.prior_simple_base_r(prior, n))
  }

  if(is.prior.discrete(prior)){
    discrete <- .prior_simple_truncated_discrete(prior)
    return(sample(discrete[["support"]], size = n, replace = TRUE, prob = discrete[["prob"]]))
  }

  C1 <- .prior_C1(prior)
  .prior_simple_base_q(prior, stats::runif(n, min = C1, max = .prior_C2(prior)))
}

.prior_simple_rng_rejection <- function(prior, n){

  x  <- NULL
  nn <- round(n * 1 / .prior_C(prior) * 1.10)

  while(length(x) < n){
    temp_x <- .prior_simple_base_r(prior, nn)
    x      <- c(x, temp_x[temp_x >= prior$truncation[["lower"]] & temp_x <= prior$truncation[["upper"]]])
  }

  x[1:n]
}

.prior_C1 <- function(prior){

  if(is.prior.simple(prior)){

    C1 <- switch(
      prior[["distribution"]],
      "normal"    = stats::pnorm(prior$truncation[["lower"]], mean = prior$parameters[["mean"]], sd = prior$parameters[["sd"]], lower.tail = TRUE, log.p = FALSE),
      "lognormal" = stats::plnorm(prior$truncation[["lower"]], meanlog = prior$parameters[["meanlog"]], sdlog = prior$parameters[["sdlog"]], lower.tail = TRUE, log.p = FALSE),
      "t"         = extraDistr::plst(prior$truncation[["lower"]], df = prior$parameters[["df"]], mu = prior$parameters[["location"]], sigma = prior$parameters[["scale"]], lower.tail = TRUE, log.p = FALSE),
      "gamma"     = stats::pgamma(prior$truncation[["lower"]], shape = prior$parameters[["shape"]], rate = prior$parameters[["rate"]], lower.tail = TRUE, log.p = FALSE),
      "invgamma"  = extraDistr::pinvgamma(prior$truncation[["lower"]], alpha = prior$parameters[["shape"]], beta = prior$parameters[["scale"]], lower.tail = TRUE, log.p = FALSE),
      "beta"      = stats::pbeta(prior$truncation[["lower"]], shape1 = prior$parameters[["alpha"]], shape2 = prior$parameters[["beta"]], lower.tail = TRUE, log.p = FALSE),
      "bernoulli" = stats::pbinom(ceiling(prior$truncation[["lower"]]) - 1, size = 1, prob = prior$parameters[["probability"]], lower.tail = TRUE, log.p = FALSE),
      "exp"       = stats::pexp(prior$truncation[["lower"]], rate = prior$parameters[["rate"]], lower.tail = TRUE, log.p = FALSE),
      "uniform"   = stats::punif(prior$truncation[["lower"]], min = prior$parameters[["a"]], max = prior$parameters[["b"]], lower.tail = TRUE, log.p = FALSE),
      "point"     = ppoint(prior$truncation[["lower"]], location = prior$parameters[["location"]], lower.tail = TRUE, log.p = FALSE)
    )

  }

  return(C1)
}
.prior_C2 <- function(prior){

  if(is.prior.simple(prior)){

    C2 <- switch(
      prior[["distribution"]],
      "normal"    = stats::pnorm(prior$truncation[["upper"]], mean = prior$parameters[["mean"]], sd = prior$parameters[["sd"]], lower.tail = TRUE, log.p = FALSE),
      "lognormal" = stats::plnorm(prior$truncation[["upper"]], meanlog = prior$parameters[["meanlog"]], sdlog = prior$parameters[["sdlog"]], lower.tail = TRUE, log.p = FALSE),
      "t"         = extraDistr::plst(prior$truncation[["upper"]], df = prior$parameters[["df"]], mu = prior$parameters[["location"]], sigma = prior$parameters[["scale"]], lower.tail = TRUE, log.p = FALSE),
      "gamma"     = stats::pgamma(prior$truncation[["upper"]], shape = prior$parameters[["shape"]], rate = prior$parameters[["rate"]], lower.tail = TRUE, log.p = FALSE),
      "invgamma"  = extraDistr::pinvgamma(prior$truncation[["upper"]], alpha = prior$parameters[["shape"]], beta = prior$parameters[["scale"]], lower.tail = TRUE, log.p = FALSE),
      "beta"      = stats::pbeta(prior$truncation[["upper"]], shape1 = prior$parameters[["alpha"]], shape2 = prior$parameters[["beta"]], lower.tail = TRUE, log.p = FALSE),
      "bernoulli" = stats::pbinom(prior$truncation[["upper"]], size = 1, prob = prior$parameters[["probability"]], lower.tail = TRUE, log.p = FALSE),
      "exp"       = stats::pexp(prior$truncation[["upper"]], rate = prior$parameters[["rate"]], lower.tail = TRUE, log.p = FALSE),
      "uniform"   = stats::punif(prior$truncation[["upper"]], min = prior$parameters[["a"]], max = prior$parameters[["b"]], lower.tail = TRUE, log.p = FALSE),
      "point"     = ppoint(prior$truncation[["upper"]], location = prior$parameters[["location"]], lower.tail = TRUE, log.p = FALSE)
    )

  }

  return(C2)
}
.prior_C  <- function(prior){

  if(is.prior.simple(prior)){

    C <- .prior_C2(prior) - .prior_C1(prior)

  }

  return(C)
}

.prior_vector_dimension <- function(prior){

  K <- prior$parameters[["K"]]
  .check_parameter_dimensions(K, "K", allow_NA = FALSE)
  as.integer(K)
}
.prior_vector_primary_location <- function(prior){

  location <- switch(
    prior[["distribution"]],
    "mnormal" = prior$parameters[["mean"]],
    "mt"      = prior$parameters[["location"]],
    "mpoint"  = prior$parameters[["location"]]
  )

  if(length(location) != 1){
    stop("Only exchangeable vector prior distributions with scalar location parameters are supported.", call. = FALSE)
  }

  location
}
.prior_vector_x_matrix <- function(x, K, name = "x"){

  if(is.matrix(x)){
    if(ncol(x) != K){
      stop(paste0("The '", name, "' argument must have ", K, " columns."), call. = FALSE)
    }
    return(x)
  }

  matrix(rep(x, times = K), ncol = K)
}
.prior_vector_marginal_cdf <- function(prior, q, lower.tail = TRUE){

  K     <- .prior_vector_dimension(prior)
  q_mat <- .prior_vector_x_matrix(q, K, "q")
  loc   <- .prior_vector_primary_location(prior)

  out <- switch(
    prior[["distribution"]],
    "mnormal" = apply(q_mat, 2, stats::pnorm, mean = loc, sd = prior$parameters[["sd"]], lower.tail = lower.tail),
    "mt"      = apply(q_mat, 2, extraDistr::plst, df = prior$parameters[["df"]], mu = loc, sigma = prior$parameters[["scale"]], lower.tail = lower.tail),
    "mpoint"  = apply(q_mat, 2, ppoint, location = loc, lower.tail = lower.tail)
  )

  matrix(out, nrow = nrow(q_mat), ncol = K)
}
.prior_vector_marginal_lpdf <- function(prior, x){

  K     <- .prior_vector_dimension(prior)
  x_mat <- .prior_vector_x_matrix(x, K, "x")
  loc   <- .prior_vector_primary_location(prior)

  out <- switch(
    prior[["distribution"]],
    "mnormal" = apply(x_mat, 2, stats::dnorm, mean = loc, sd = prior$parameters[["sd"]], log = TRUE),
    "mt"      = apply(x_mat, 2, extraDistr::dlst, df = prior$parameters[["df"]], mu = loc, sigma = prior$parameters[["scale"]], log = TRUE),
    "mpoint"  = apply(x_mat, 2, dpoint, location = loc, log = TRUE)
  )

  matrix(out, nrow = nrow(x_mat), ncol = K)
}
.prior_vector_marginal_quant <- function(prior, p){

  K   <- .prior_vector_dimension(prior)
  loc <- .prior_vector_primary_location(prior)

  q <- switch(
    prior[["distribution"]],
    "mnormal" = stats::qnorm(p, mean = loc, sd = prior$parameters[["sd"]]),
    "mt"      = extraDistr::qlst(p, df = prior$parameters[["df"]], mu = loc, sigma = prior$parameters[["scale"]]),
    "mpoint"  = qpoint(p, location = loc)
  )

  matrix(rep(q, times = K), ncol = K)
}
.prior_vector_mean <- function(prior){

  K   <- .prior_vector_dimension(prior)
  loc <- .prior_vector_primary_location(prior)

  m <- switch(
    prior[["distribution"]],
    "mnormal" = loc,
    "mt"      = ifelse(prior$parameters[["df"]] > 1, loc, NaN),
    "mpoint"  = loc
  )

  rep(m, K)
}
.prior_vector_var <- function(prior){

  K <- .prior_vector_dimension(prior)

  v <- switch(
    prior[["distribution"]],
    "mnormal" = prior$parameters[["sd"]]^2,
    "mt"      = ifelse(prior$parameters[["df"]] > 2, prior$parameters[["scale"]]^2 * prior$parameters[["df"]] / (prior$parameters[["df"]] - 2), NaN),
    "mpoint"  = 0
  )

  rep(v, K)
}

# tools
.is_prior_default_range   <- function(prior){
  default_range <- switch(
    prior[["distribution"]],
    "normal"    = is.infinite(prior$truncation[["lower"]])          & is.infinite(prior$truncation[["upper"]]),
    "lognormal" = isTRUE(all.equal(prior$truncation[["lower"]], 0)) & is.infinite(prior$truncation[["upper"]]),
    "t"         = is.infinite(prior$truncation[["lower"]])          & is.infinite(prior$truncation[["upper"]]),
    "gamma"     = isTRUE(all.equal(prior$truncation[["lower"]], 0)) & is.infinite(prior$truncation[["upper"]]),
    "invgamma"  = isTRUE(all.equal(prior$truncation[["lower"]], 0)) & is.infinite(prior$truncation[["upper"]]),
    "beta"      = isTRUE(all.equal(prior$truncation[["lower"]], 0)) & isTRUE(all.equal(prior$truncation[["upper"]], 1)),
    "bernoulli" = isTRUE(all.equal(prior$truncation[["lower"]], 0)) & isTRUE(all.equal(prior$truncation[["upper"]], 1)),
    "exp"       = isTRUE(all.equal(prior$truncation[["lower"]], 0)) & is.infinite(prior$truncation[["upper"]]),
    "uniform"   = isTRUE(all.equal(prior$truncation[["lower"]], prior$parameters[["a"]])) &
      isTRUE(all.equal(prior$truncation[["upper"]], prior$parameters[["b"]])),
    "point"     = TRUE,
    "mpoint"          = TRUE,
    "weightfunction"  = TRUE,
    "none"            = TRUE
  )
}

#### marginal distribution functions ####
# weightfunctions require just marginals for plotting and etc
# (also, the joint are not implemented in general, since they differ between JAGS/Stan implementations)
#' @rdname prior_functions
mcdf.prior   <- function(x, q, ...){

  prior <- x

  .check_q(as.vector(q))
  .check_prior(prior)

  if(is.prior.spike_and_slab(prior)){

    stop("No mcdf are implemented for spike and slab priors.")

  }else if(is.prior.mixture(prior)){

    stop("No mcdf are implemented for prior mixtures.")

  }else if(is.prior.simple(prior)){

    p <- .prior_simple_cdf(prior, q)

  }else if(is.prior.weightfunction(prior)){

    components <- .weightfunction_marginal_components(prior)
    p <- do.call(cbind, lapply(components, .prior_weightfunction_component_cdf, q = q))

  }else if(is.prior.vector(prior) && !is.prior.factor(prior)){

    p <- .prior_vector_marginal_cdf(prior, q, lower.tail = TRUE)

  }else if(is.prior.orthonormal(prior) | is.prior.meandif(prior)){

    par1 <- switch(
      prior[["distribution"]],
      "mnormal" = prior$parameter[["mean"]],
      "mt"      = prior$parameter[["location"]],
      "mpoint"  = prior$parameter[["location"]]
    )

    if(length(par1) != 1){
      stop("unsported distribution specification in 'rng' -- non-symmetric")
    }

    if(par1 != 0){
      stop("the orthonormal/meandif prior distribution must be centered")
    }

    if(is.na(prior$parameters[["K"]]) && !is.null(attr(prior, "levels"))){
      prior$parameters[["K"]] <- .get_prior_factor_levels(prior)
    }else if(is.na(prior$parameters[["K"]])){
      prior$parameters[["K"]] <- 1
      warning("number of factor levels / dimensionality of the prior distribution was not specified -- assuming two factor levels")
    }

    if(prior[["distribution"]] %in% c("mnormal", "mt")){
      if(is.prior.orthonormal(prior)){
        par2 <- sqrt(sum( (contr.orthonormal(1:(prior$parameters[["K"]] + 1))[1,] * switch(
          prior[["distribution"]],
          "mnormal" = prior$parameters[["sd"]],
          "mt"      = prior$parameters[["scale"]]) )^2 ))
      }else if(is.prior.meandif(prior)){
        par2 <- sqrt(sum( (contr.meandif(1:(prior$parameters[["K"]] + 1))[1,] * switch(
          prior[["distribution"]],
          "mnormal" = prior$parameters[["sd"]],
          "mt"      = prior$parameters[["scale"]]) )^2 ))
      }
    }

    p <- switch(
      prior[["distribution"]],
      "mnormal" = stats::pnorm(q, mean = 0, sd = par2),
      "mt"      = extraDistr::plst(q, df = prior$parameters[["df"]], mu = 0, sigma = par2),
      "mpoint"  = ppoint(q, location = 0)
    )

  }

  return(p)
}
#' @rdname prior_functions
mccdf.prior  <- function(x, q, ...){

  prior <- x

  .check_q(as.vector(q))
  .check_prior(prior)

  if(is.prior.spike_and_slab(prior)){

    stop("No mccdf are implemented for spike and slab priors.")

  }else if(is.prior.mixture(prior)){

    stop("No mccdf are implemented for prior mixtures.")

  }else if(is.prior.simple(prior)){

    p <- .prior_simple_ccdf(prior, q)

  }else if(is.prior.weightfunction(prior)){

    components <- .weightfunction_marginal_components(prior)
    p <- do.call(cbind, lapply(components, .prior_weightfunction_component_ccdf, q = q))

  }else if(is.prior.vector(prior) && !is.prior.factor(prior)){

    p <- .prior_vector_marginal_cdf(prior, q, lower.tail = FALSE)

  }else if(is.prior.orthonormal(prior) | is.prior.meandif(prior)){

    par1 <- switch(
      prior[["distribution"]],
      "mnormal" = prior$parameter[["mean"]],
      "mt"      = prior$parameter[["location"]],
      "mpoint"  = prior$parameter[["location"]]
    )

    if(length(par1) != 1){
      stop("unsported distribution specification in 'rng' -- non-symmetric")
    }

    if(par1 != 0){
      stop("the orthonormal/meandif prior distribution must be centered")
    }

    if(is.na(prior$parameters[["K"]]) && !is.null(attr(prior, "levels"))){
      prior$parameters[["K"]] <- .get_prior_factor_levels(prior)
    }else if(is.na(prior$parameters[["K"]])){
      prior$parameters[["K"]] <- 1
      warning("number of factor levels / dimensionality of the prior distribution was not specified -- assuming two factor levels")
    }

    if(prior[["distribution"]] %in% c("mnormal", "mt")){
      if(is.prior.orthonormal(prior)){
        par2 <- sqrt(sum( (contr.orthonormal(1:(prior$parameters[["K"]] + 1))[1,] * switch(
          prior[["distribution"]],
          "mnormal" = prior$parameters[["sd"]],
          "mt"      = prior$parameters[["scale"]]) )^2 ))
      }else if(is.prior.meandif(prior)){
        par2 <- sqrt(sum( (contr.meandif(1:(prior$parameters[["K"]] + 1))[1,] * switch(
          prior[["distribution"]],
          "mnormal" = prior$parameters[["sd"]],
          "mt"      = prior$parameters[["scale"]]) )^2 ))
      }
    }

    p <- switch(
      prior[["distribution"]],
      "mnormal" = stats::pnorm(q, mean = 0, sd = par2, lower.tail = FALSE),
      "mt"      = extraDistr::plst(q, df = prior$parameters[["df"]], mu = 0, sigma = par2, lower.tail = FALSE),
      "mpoint"  = ppoint(q, location = 0, lower.tail = FALSE),
    )

  }

  return(p)
}
#' @rdname prior_functions
mlpdf.prior  <- function(x, y, ...){

  prior <- x
  x     <- y

  .check_x(as.vector(x))
  .check_prior(prior)

  if(is.prior.spike_and_slab(prior)){

    stop("No mlpdf are implemented for spike and slab priors.")

  }else if(is.prior.mixture(prior)){

    stop("No mlpdf are implemented for prior mixtures.")

  }else if(is.prior.simple(prior)){

    log_lik <- .prior_simple_lpdf(prior, x)

  }else if(is.prior.weightfunction(prior)){

    components <- .weightfunction_marginal_components(prior)
    log_lik <- do.call(cbind, lapply(components, .prior_weightfunction_component_lpdf, x = x))

  }else if(is.prior.vector(prior) && !is.prior.factor(prior)){

    log_lik <- .prior_vector_marginal_lpdf(prior, x)

  }else if(is.prior.orthonormal(prior) | is.prior.meandif(prior)){

    par1 <- switch(
      prior[["distribution"]],
      "mnormal" = prior$parameter[["mean"]],
      "mt"      = prior$parameter[["location"]],
      "mpoint"  = prior$parameter[["location"]]
    )

    if(length(par1) != 1){
      stop("unsported distribution specification in 'rng' -- non-symmetric")
    }

    if(par1 != 0){
      stop("the orthonormal/meandif prior distribution must be centered")
    }

    if(is.na(prior$parameters[["K"]]) && !is.null(attr(prior, "levels"))){
      prior$parameters[["K"]] <- .get_prior_factor_levels(prior)
    }else if(is.na(prior$parameters[["K"]])){
      prior$parameters[["K"]] <- 1
      warning("number of factor levels / dimensionality of the prior distribution was not specified -- assuming two factor levels")
    }

    if(prior[["distribution"]] %in% c("mnormal", "mt")){
      if(is.prior.orthonormal(prior)){
        par2 <- sqrt(sum( (contr.orthonormal(1:(prior$parameters[["K"]] + 1))[1,] * switch(
          prior[["distribution"]],
          "mnormal" = prior$parameters[["sd"]],
          "mt"      = prior$parameters[["scale"]]) )^2 ))
      }else if(is.prior.meandif(prior)){
        par2 <- sqrt(sum( (contr.meandif(1:(prior$parameters[["K"]] + 1))[1,] * switch(
          prior[["distribution"]],
          "mnormal" = prior$parameters[["sd"]],
          "mt"      = prior$parameters[["scale"]]) )^2 ))
      }
    }

    log_lik <- switch(
      prior[["distribution"]],
      "mnormal" = stats::dnorm(x, mean = 0, sd = par2, log = TRUE),
      "mt"      = extraDistr::dlst(x, df = prior$parameters[["df"]], mu = 0, sigma = par2, log = TRUE),
      "mpoint"  = dpoint(x, location = 0, log = TRUE),
    )

  }

  return(log_lik)
}
#' @rdname prior_functions
mpdf.prior   <- function(x, y, ...){

  prior <- x
  x     <- y

  .check_x(as.vector(x))
  .check_prior(prior)

  if(is.prior.simple(prior)){
    lik <- .prior_simple_pdf(prior, x)
  }else{
    log_lik <- mlpdf.prior(prior, x)
    lik     <- exp(log_lik)
  }

  return(lik)
}
#' @rdname prior_functions
mquant.prior <- function(x, p, ...){

  prior <- x

  .check_p(p, log.p = FALSE)
  .check_prior(prior)

  if(is.prior.spike_and_slab(prior)){

    stop("No quantile functions are implemented for spike and slab priors.")

  }else if(is.prior.mixture(prior)){

    stop("No quantile functions are implemented for prior mixtures.")

  }else if(is.prior.simple(prior)){

    q <- .prior_simple_quant(prior, p)

  }else if(is.prior.weightfunction(prior)){

    components <- .weightfunction_marginal_components(prior)
    q <- do.call(cbind, lapply(components, .prior_weightfunction_component_quant, p = p))

  }else if(is.prior.vector(prior) && !is.prior.factor(prior)){

    q <- .prior_vector_marginal_quant(prior, p)

  }else if(is.prior.orthonormal(prior) | is.prior.meandif(prior)){

    par1 <- switch(
      prior[["distribution"]],
      "mnormal" = prior$parameter[["mean"]],
      "mt"      = prior$parameter[["location"]],
      "mpoint"  = prior$parameter[["location"]]
    )

    if(length(par1) != 1){
      stop("unsported distribution specification in 'rng' -- non-symmetric")
    }

    if(par1 != 0){
      stop("the orthonormal/meandif prior distribution must be centered")
    }

    if(is.na(prior$parameters[["K"]]) && !is.null(attr(prior, "levels"))){
      prior$parameters[["K"]] <- .get_prior_factor_levels(prior)
    }else if(is.na(prior$parameters[["K"]])){
      prior$parameters[["K"]] <- 1
      warning("number of factor levels / dimensionality of the prior distribution was not specified -- assuming two factor levels")
    }

    if(prior[["distribution"]] %in% c("mnormal", "mt")){
      if(is.prior.orthonormal(prior)){
        par2 <- sqrt(sum( (contr.orthonormal(1:(prior$parameters[["K"]] + 1))[1,] * switch(
          prior[["distribution"]],
          "mnormal" = prior$parameters[["sd"]],
          "mt"      = prior$parameters[["scale"]]) )^2 ))
      }else if(is.prior.meandif(prior)){
        par2 <- sqrt(sum( (contr.meandif(1:(prior$parameters[["K"]] + 1))[1,] * switch(
          prior[["distribution"]],
          "mnormal" = prior$parameters[["sd"]],
          "mt"      = prior$parameters[["scale"]]) )^2 ))
      }
    }

    q <- switch(
      prior[["distribution"]],
      "mnormal" = stats::qnorm(p, mean = 0, sd = par2),
      "mt"      = extraDistr::qlst(p, df = prior$parameters[["df"]], mu = 0, sigma = par2),
      "mpoint"  = qpoint(p, location = 0)
    )

  }

  return(q)
}

### create generic method
#' @title Creates generics for common statistical functions
#'
#' @description Density (pdf / lpdf), distribution
#' function (cdf / ccdf), quantile function (quant),
#' random generation (rng), mean, standard deviation (sd),
#' and marginal variants of the functions (mpdf, mlpf, mcdf,
#' mccdf, mquant).
#'
#' @param x main argument
#' @param ... unused arguments
#'
#' @return \code{pdf} (\code{mpdf}) and \code{lpdf} (\code{mlpdf}) give
#' the (marginal) density and the log of (marginal) density,
#' \code{cdf} (\code{mcdf}) and \code{ccdf} (\code{mccdf}) give the
#' (marginal) distribution and the complement of (marginal) distribution function,
#' \code{quant} (\code{mquant}) give the (marginal) quantile function,
#' and \code{rng} generates random deviates for an object of class 'prior'.
#'
#' The \code{pdf} function proceeds to PDF graphics device if \code{x} is a character.
#'
#' @export rng
#' @export cdf
#' @export ccdf
#' @export quant
#' @export lpdf
#' @export pdf
#' @export mcdf
#' @export mccdf
#' @export mquant
#' @export mlpdf
#' @export mpdf
#' @importFrom grDevices pdf
#' @name prior_functions_methods
NULL

#' @rdname prior_functions_methods
rng    <- function(x, ...){
  UseMethod("rng")
}
#' @rdname prior_functions_methods
cdf    <- function(x, ...){
  UseMethod("cdf")
}
#' @rdname prior_functions_methods
ccdf   <- function(x, ...){
  UseMethod("ccdf")
}
#' @rdname prior_functions_methods
quant  <- function(x, ...){
  UseMethod("quant")
}
#' @rdname prior_functions_methods
lpdf   <- function(x, ...){
  UseMethod("lpdf")
}
#' @rdname prior_functions_methods
pdf    <- function(x, ...){
  UseMethod("pdf")
}
#' @rdname prior_functions_methods
mcdf   <- function(x, ...){
  UseMethod("mcdf")
}
#' @rdname prior_functions_methods
mccdf  <- function(x, ...){
  UseMethod("mccdf")
}
#' @rdname prior_functions_methods
mquant <- function(x, ...){
  UseMethod("mquant")
}
#' @rdname prior_functions_methods
mlpdf  <- function(x, ...){
  UseMethod("mlpdf")
}
#' @rdname prior_functions_methods
mpdf   <- function(x, ...){
  UseMethod("mpdf")
}

#' @export
pdf.default  <- function(x, ...){
  grDevices::pdf(x, ...)
}


#### additional prior related function ####
#' @title Prior mean
#'
#' @description Computes mean of a prior
#' distribution. (In case of orthonormal prior distributions
#' for factors, the mean of for the deviations from intercept
#' is returned.)
#'
#' @param x a prior
#' @param ... unused
#'
#' @examples
#' # create a standard normal prior distribution
#' p1 <- prior(distribution = "normal", parameters = list(mean = 1, sd = 1))
#'
#' # compute mean of the prior distribution
#' mean(p1)
#'
#' @return a mean of an object of class 'prior'.
#'
#' @seealso [prior()]
#' @rdname mean.prior
#' @export
mean.prior   <- function(x, ...){

  .check_prior(x, "x")

  if(is.prior.spike_and_slab(x)){

    m <- mean(.get_spike_and_slab_variable(x)) * mean(.get_spike_and_slab_inclusion(x))

  }else if(is.prior.mixture(x)){

    stop("No mean is implemented for prior mixtures.")

  }else if(is.prior.simple(x)){

    if(.is_prior_default_range(x)){

      m <- switch(
        x[["distribution"]],
        "normal"    = x$parameters[["mean"]],
        "lognormal" = exp(x$parameters[["meanlog"]] + x$parameters[["sdlog"]]^2/2),
        "t"         = ifelse(x$parameters[["df"]] > 1, x$parameters[["location"]], NaN),
        "gamma"     = x$parameters[["shape"]] / x$parameters[["rate"]],
        "invgamma"  = ifelse(x$parameters[["shape"]] > 1, x$parameters[["scale"]]/(x$parameters[["shape"]] - 1), NaN),
        "beta"      = x$parameters[["alpha"]] / (x$parameters[["alpha"]] + x$parameters[["beta"]]),
        "bernoulli" = x$parameters[["probability"]],
        "exp"       = 1 / x$parameters[["rate"]],
        "uniform"   = (x$parameters[["a"]] + x$parameters[["b"]]) / 2,
        "point"     = x$parameters[["location"]]
      )

    }else{

      if(is.prior.discrete(x)){
        discrete <- .prior_simple_truncated_discrete(x)
        return(sum(discrete[["support"]] * discrete[["prob"]]))
      }

      # check for undefined values when the problematic tail is still unbounded
      if(x[["distribution"]] == "t"){
        if(x$parameters[["df"]] <= 1 &&
           (is.infinite(x$truncation[["lower"]]) || is.infinite(x$truncation[["upper"]]))){
          return(NaN)
        }
      }
      if(x[["distribution"]] == "invgamma"){
        if(x$parameters[["shape"]] <= 1 && is.infinite(x$truncation[["upper"]])){
          return(NaN)
        }
      }

      m <- stats::integrate(
        f       = function(x, prior) x * pdf(prior, x),
        lower   = x$truncation[["lower"]],
        upper   = x$truncation[["upper"]],
        prior   = x
      )$value

    }


  }else if(is.prior.weightfunction(x)){

    m <- .mean.weightfunction(x)

  }else if(is_prior_phacking(x) || is_prior_bias(x)){

    .selection_prior_stop_unsupported_generic("mean", x)

  }else if(is.prior.vector(x) && !is.prior.factor(x)){

    m <- .prior_vector_mean(x)

  }else if(is.prior.orthonormal(x) | is.prior.meandif(x)){

    par1 <- switch(
      x[["distribution"]],
      "mnormal" = x$parameter[["mean"]],
      "mt"      = x$parameter[["location"]],
      "mpoint"  = x$parameter[["location"]]
    )

    if(length(par1) != 1){
      stop("unsported distribution specification in 'rng' -- non-symmetric")
    }

    if(par1 != 0){
      stop("the orthonormal/meandif prior distribution must be centered")
    }

    if(x[["distribution"]] == "mt" && x$parameters[["df"]] <= 1){
      m <- NaN
    }else{
      # orthonormal prior distributions must be centered
      m <- 0
    }

  }

  return(m)
}


### create generic methods from sd and var
#' @title Creates generic for sd function
#'
#' @param x main argument
#' @param ... additional arguments
#'
#' @return \code{sd} returns a standard deviation
#' of the supplied object (if it is either a numeric vector
#' or an object of class 'prior').
#'
#' @seealso \link[stats]{sd}
#' @export
sd  <- function(x, ...){
  UseMethod("sd")
}

#' @title Creates generic for var function
#'
#' @param x main argument
#' @param ... additional arguments
#'
#' @return \code{var} returns a variance
#' of the supplied object (if it is either a numeric vector
#' or an object of class 'prior').
#'
#' @seealso \link[stats]{cor}
#' @export
var <- function(x, ...){
  UseMethod("var")
}

#' @export
sd.default  <- function(x, ...){
  stats::sd(x, ...)
}
#' @export
var.default <- function(x, ...){
  stats::var(x, ...)
}


#' @title Prior var
#'
#' @description Computes variance
#' of a prior distribution.
#'
#' @param x a prior
#' @param ... unused arguments
#'
#' @examples
#' # create a standard normal prior distribution
#' p1 <- prior(distribution = "normal", parameters = list(mean = 1, sd = 1))
#'
#' # compute variance of the prior distribution
#' var(p1)
#'
#' @return a variance of an object of class 'prior'.
#'
#' @seealso [prior()]
#' @importFrom stats var
#' @rdname var.prior
#' @export
var.prior   <- function(x, ...){

  .check_prior(x, "x")

  if(is.prior.spike_and_slab(x)){

    inclusion_mean <- mean(.get_spike_and_slab_inclusion(x))
    variable_mean  <- mean(.get_spike_and_slab_variable(x))
    variable_E2    <- var(.get_spike_and_slab_variable(x)) + variable_mean^2

    var <- inclusion_mean * variable_E2 - (inclusion_mean * variable_mean)^2

  }else if(is.prior.mixture(x)){

    stop("No var is implemented for prior mixtures.")

  }else if(is.prior.simple(x)){

    if(.is_prior_default_range(x)){

      var <- switch(
        x[["distribution"]],
        "normal"    = x$parameters[["sd"]]^2,
        "lognormal" = (exp(x$parameters[["sdlog"]]^2) - 1) * exp(2 * x$parameters[["meanlog"]] + x$parameters[["sdlog"]]^2),
        "t"         = ifelse(x$parameters[["df"]] > 2, x$parameters[["scale"]]^2 * x$parameters[["df"]] / (x$parameters[["df"]] - 2), NaN),
        "gamma"     = x$parameters[["shape"]] / x$parameters[["rate"]]^2,
        "invgamma"  = ifelse(x$parameters[["shape"]] > 2, x$parameters[["scale"]]^2 / ((x$parameters[["shape"]] - 1)^2 * (x$parameters[["shape"]] - 2)), NaN),
        "beta"      = (x$parameters[["alpha"]] * x$parameters[["beta"]]) / ((x$parameters[["alpha"]] + x$parameters[["beta"]])^2 * (x$parameters[["alpha"]] + x$parameters[["beta"]] + 1)),
        "bernoulli" = (x$parameters[["probability"]] * (1 - x$parameters[["probability"]]) ),
        "exp"       = 1 / x$parameters[["rate"]]^2,
        "uniform"   = (x$parameters[["b"]] - x$parameters[["a"]])^2 / 12,
        "point"     = 0
      )

    }else{

      if(is.prior.discrete(x)){
        discrete <- .prior_simple_truncated_discrete(x)
        E2       <- sum(discrete[["support"]]^2 * discrete[["prob"]])
        return(E2 - mean(x)^2)
      }

      # check for undefined values when the problematic tail is still unbounded
      if(x[["distribution"]] == "t"){
        if(x$parameters[["df"]] <= 2 &&
           (is.infinite(x$truncation[["lower"]]) || is.infinite(x$truncation[["upper"]]))){
          return(NaN)
        }
      }
      if(x[["distribution"]] == "invgamma"){
        if(x$parameters[["shape"]] <= 2 && is.infinite(x$truncation[["upper"]])){
          return(NaN)
        }
      }

      E2 <- stats::integrate(
        f       = function(x, prior) x^2 * pdf(prior, x),
        lower   = x$truncation[["lower"]],
        upper   = x$truncation[["upper"]],
        prior   = x
      )$value

      var <- E2 - mean(x)^2
    }


  }else if(is.prior.weightfunction(x)){

    var <- .var.weightfunction(x)

  }else if(is_prior_phacking(x) || is_prior_bias(x)){

    .selection_prior_stop_unsupported_generic("variance", x)

  }else if(is.prior.vector(x) && !is.prior.factor(x)){

    var <- .prior_vector_var(x)

  }else if(is.prior.orthonormal(x) | is.prior.meandif(x)){

    par1 <- switch(
      x[["distribution"]],
      "mnormal" = x$parameter[["mean"]],
      "mt"      = x$parameter[["location"]],
      "mpoint"  = x$parameter[["location"]]
    )

    if(length(par1) != 1){
      stop("unsported distribution specification in 'rng' -- non-symmetric")
    }

    if(par1 != 0){
      stop("the orthonormal/meandif prior distribution must be centered")
    }

    if(x[["distribution"]] == "mpoint"){
      var <- 0
    }else if(x[["distribution"]] == "mt" && x$parameters[["df"]] <= 2){
      var <- NaN
    }else{


      if(is.na(x$parameters[["K"]]) && !is.null(attr(x, "levels"))){
        x$parameters[["K"]] <- .get_prior_factor_levels(x)
      }else if(is.na(x$parameters[["K"]])){
        x$parameters[["K"]] <- 1
        warning("number of factor levels / dimensionality of the prior distribution was not specified -- assuming two factor levels")
      }


      if(is.prior.orthonormal(x)){
        par2 <- sqrt(sum( (contr.orthonormal(1:(x$parameters[["K"]] + 1))[1,] * switch(
          x[["distribution"]],
          "mnormal" = x$parameters[["sd"]],
          "mt"      = x$parameters[["scale"]]) )^2 ))
      }else if(is.prior.meandif(x)){
        par2 <- sqrt(sum( (contr.meandif(1:(x$parameters[["K"]] + 1))[1,] * switch(
          x[["distribution"]],
          "mnormal" = x$parameters[["sd"]],
          "mt"      = x$parameters[["scale"]]) )^2 ))
      }

      # use the univariate functions
      var <- switch(
        x[["distribution"]],
        "mnormal" = var.prior(prior("normal", parameters = list(mean = 0, sd = par2))),
        "mt"      = var.prior(prior("t",      parameters = list(location = 0, scale = par2, df = x$parameters[["df"]]))))
    }

  }

  return(var)
}

#' @title Prior sd
#'
#' @description Computes standard deviation
#' of a prior distribution.
#'
#' @param x a prior
#' @param ... unused arguments
#'
#' @examples
#' # create a standard normal prior distribution
#' p1 <- prior(distribution = "normal", parameters = list(mean = 1, sd = 1))
#'
#' # compute sd of the prior distribution
#' sd(p1)
#'
#' @return a standard deviation of an object of class 'prior'.
#'
#' @seealso [prior()]
#' @importFrom stats sd
#' @rdname sd.prior
#' @export
sd.prior     <- function(x, ...){

  .check_prior(x, "x")

  sd <- sqrt(var(x))

  return(sd)
}

.mean.weightfunction <- function(prior){

  components <- .weightfunction_marginal_components(prior)
  m <- vapply(components, .prior_weightfunction_component_mean, numeric(1))

  return(m)
}
.var.weightfunction  <- function(prior){

  components <- .weightfunction_marginal_components(prior)
  m <- vapply(components, .prior_weightfunction_component_var, numeric(1))

  return(m)
}

#' @title Point mass distribution
#'
#' @description Density, distribution function,
#' quantile function and random generation for point distribution.
#'
#' @param x,q vector or matrix of quantiles.
#' @param p vector of probabilities.
#' @param n number of observations.
#' @param location vector of locations.
#' @param log,log.p logical; if \code{TRUE}, probabilities
#' \code{p} are given as \code{log(p)}.
#' @param lower.tail logical; if \code{TRUE} (default), probabilities
#' are \eqn{P[X \le x]}, otherwise, \eqn{P[X > x]}.
#'
#' @examples
#' # draw samples from a point distribution
#' rpoint(10, location = 1)
#'
#' @return \code{dpoint} gives the density, \code{ppoint} gives the
#' distribution function, \code{qpoint} gives the quantile function,
#' and \code{rpoint} generates random deviates.
#'
#' @export dpoint
#' @export rpoint
#' @export ppoint
#' @export qpoint
#' @name point
NULL

#### wrappers ####
#' @rdname point
dpoint <- function(x, location, log = FALSE){

  # common input check
  .check_log(log)
  .check_x(x)
  check_real(location, "location", check_length = 0)

  if(length(location) != length(x) & length(location) != 1)
    stop("Non matching dimensions of 'location' and 'x'.")

  if(length(location) != length(x) & length(location) == 1){
    location <- rep(location, length(x))
  }

  lik <- ifelse(
    is.na(location) | is.na(x),
    NA_real_,
    ifelse(location == x, Inf, 0)
  )

  if(log){
    lik <- log(lik)
  }

  return(lik)
}
#' @rdname point
rpoint <- function(n, location){

  # common input check
  .check_n(n)
  check_real(location, "location", check_length = 0)

  if(length(location) != n & length(location) != 1)
    stop("Incompatible dimensions of requested number of samples and 'location'.")

  if(length(location) != n & length(location) == 1){
    x <- rep(location, n)
  }else{
    x <- location
  }

  return(x)
}
#' @rdname point
ppoint <- function(q, location, lower.tail = TRUE, log.p = FALSE){

  # common input check
  .check_log.p(log.p)
  .check_lower.tail(lower.tail)
  .check_q(q)
  check_real(location, "location", check_length = 0)

  if(length(location) != length(q) & length(location) != 1)
    stop("Non matching dimensions of 'location' and 'q'.")

  if(length(location) != length(q) & length(location) == 1){
    location <- rep(location, length(q))
  }

  p <- ifelse(location <= q, 1, 0)

  if(!lower.tail){
    p <- 1 - p
  }
  if(log.p){
    p <- log(p)
  }

  return(p)
}
#' @rdname point
qpoint <- function(p, location, lower.tail = TRUE, log.p = FALSE){

  # common input check
  .check_log.p(log.p)
  .check_lower.tail(lower.tail)
  .check_p(p, log.p)
  check_real(location, "location", check_length = 0)

  if(length(location) != length(p) & length(location) != 1)
    stop("Non matching dimensions of 'location' and 'p'.")

  if(length(location) != length(p) & length(location) == 1){
    location <- rep(location, length(p))
  }

  if(log.p){
    p <- exp(p)
  }

  q <- ifelse(p > 0, location, ifelse(lower.tail, -Inf, Inf))

  return(q)
}

#' @title Fits a 'JAGS' model
#'
#' @description A wrapper around
#' \link[runjags]{run.jags}  that simplifies fitting 'JAGS' models
#' with usage with pre-specified model part of the 'JAGS' syntax, data and list
#' of prior distributions.
#' @param model_syntax jags syntax for the model part
#' @param data list containing data to fit the model (not including data for the formulas)
#' @param prior_list named list of prior distribution
#' (names correspond to the parameter names) of parameters not specified within the
#' \code{formula_list}
#' @param formula_list named list of formulas to be added to the model
#' (names correspond to the parameter name created by each of the formula)
#' @param formula_data_list named list of data frames containing data for each formula
#' (names of the lists correspond to the parameter name created by each of the formula)
#' @param formula_prior_list named list of named lists of prior distributions
#' (names of the lists correspond to the parameter name created by each of the formula and
#' the names of the prior distribution correspond to the parameter names) of parameters specified
#' within the \code{formula}
#' @param formula_scale_list named list of named lists for standardizing continuous predictors
#' (names of the lists correspond to the parameter name created by each of the formula).
#' Each entry should be a named list where continuous predictors with \code{TRUE} values will
#' be standardized. Defaults to \code{NULL} (no standardization).
#' @param chains number of chains to be run, defaults to \code{4}
#' @param adapt number of samples used for adapting the MCMC chains, defaults to \code{500}
#' @param burnin number of burnin iterations of the MCMC chains, defaults to \code{1000}
#' @param sample number of sampling iterations of the MCMC chains, defaults to \code{4000}
#' @param thin thinning interval for the MCMC samples, defaults to \code{1}
#' @param autofit whether the models should be refitted until convergence criteria
#' specified in \code{autofit_control}. Defaults to \code{FALSE}.
#' @param autofit_control a list of arguments controlling the autofit function.
#' Possible options are:
#' \describe{
#'   \item{max_Rhat}{maximum R-hat error for the autofit function.
#'   Defaults to \code{1.05}.}
#'   \item{min_ESS}{minimum effective sample size. Defaults to \code{500}.}
#'   \item{max_error}{maximum MCMC error. Defaults to \code{0.01}.}
#'   \item{max_SD_error}{maximum MCMC error as the proportion of standard
#'   deviation of the parameters. Defaults to \code{0.05}.}
#'   \item{max_time}{list specifying the time \code{time} and \code{units}
#'   after which the automatic fitting function is stopped. The units arguments
#'   need to correspond to \code{units} passed to \link[base]{difftime} function.}
#'   \item{max_extend}{number of times after which the automatic fitting function is stopped.}
#'   \item{sample_extend}{number of samples between each convergence check. Defaults to
#'   \code{1000}.}
#'   \item{restarts}{number of times new initial values should be generated in case the model
#'   fails to initialize. Defaults to \code{10}.}
#'   \item{check_indicators}{whether model indicator variables should be included
#'   in convergence checks. Defaults to \code{FALSE}.}
#' }
#' @param parallel whether the chains should be run in parallel \code{FALSE}
#' @param cores number of cores used for multithreading if \code{parallel = TRUE},
#'  defaults to \code{chains}
#' @param silent whether the function should proceed silently, defaults to \code{TRUE}
#' @param seed seed for random number generation
#' @param add_parameters vector of additional parameter names that should be used
#' monitored but were not specified in the \code{prior_list}
#' @param required_packages character vector specifying list of packages containing
#' JAGS models required for sampling (in case that the function is run in parallel or in
#' detached R session). Defaults to \code{NULL}.
#' @param fit a 'BayesTools_fit' object (created by \code{JAGS_fit()} function) to be
#' extended
#' @param ... additional hidden arguments
#'
#' @examples \dontrun{
#' # simulate data
#' set.seed(1)
#' data <- list(
#'   x = rnorm(10),
#'   N = 10
#' )
#' data$x
#'
#' # define priors
#' priors_list <- list(mu = prior("normal", list(0, 1)))
#'
#' # define likelihood for the data
#' model_syntax <-
#'   "model{
#'     for(i in 1:N){
#'       x[i] ~ dnorm(mu, 1)
#'     }
#'   }"
#'
#' # fit the models
#' fit <- JAGS_fit(model_syntax, data, priors_list)
#' }
#'
#' @return \code{JAGS_fit} returns an object of class 'runjags' and 'BayesTools_fit'.
#'
#' @seealso [JAGS_check_convergence()]
#'
#' @export JAGS_fit
#' @export JAGS_extend
#' @name JAGS_fit
NULL

.JAGS_require_packages <- function(required_packages, cl = NULL){

  if(length(required_packages) == 0)
    return(invisible(logical(0)))

  required_packages <- unique(required_packages)

  if(is.null(cl)){
    package_loaded <- vapply(required_packages, requireNamespace, logical(1), quietly = TRUE)
  }else{
    package_loaded <- vapply(required_packages, function(package){
      all(unlist(parallel::clusterCall(
        cl,
        function(package) requireNamespace(package, quietly = TRUE),
        package
      ), use.names = FALSE))
    }, logical(1))
  }

  missing_packages <- names(package_loaded)[!package_loaded]
  if(length(missing_packages) > 0)
    stop(
      paste0(
        "Required packages are not available: '",
        paste0(missing_packages, collapse = "', '"),
        "'."
      ),
      call. = FALSE
    )

  invisible(package_loaded)
}

#' @rdname JAGS_fit
JAGS_fit <- function(model_syntax, data = NULL, prior_list = NULL, formula_list = NULL, formula_data_list = NULL, formula_prior_list = NULL, formula_scale_list = NULL,
                     chains = 4, adapt = 500, burnin = 1000, sample = 4000, thin = 1,
                     autofit = FALSE, autofit_control = list(max_Rhat = 1.05, min_ESS = 500, max_error = 0.01, max_SD_error = 0.05, max_time = list(time = 60, unit = "mins"), sample_extend = 1000, restarts = 10, max_extend = 10, check_indicators = FALSE),
                     parallel = FALSE, cores = chains, silent = TRUE, seed = NULL,
                     add_parameters = NULL, required_packages = NULL, ...){

  .check_runjags()
  dots <- list(...)

  ### check input
  .check_JAGS_syntax(model_syntax)
  JAGS_check_and_list_fit_settings(chains, adapt, burnin, sample, thin, autofit, parallel, cores, silent, seed)
  autofit_control <- JAGS_check_and_list_autofit_settings(autofit_control)
  check_char(add_parameters, "add_parameters", check_length = 0, allow_NULL = TRUE)
  check_char(required_packages, "required_packages", check_length = 0, allow_NULL = TRUE, allow_NA = FALSE)
  check_list(formula_list, "formula_list", allow_NULL = TRUE)
  check_list(formula_data_list, "formula_data_list", check_names = names(formula_list), allow_other = FALSE, all_objects = TRUE, allow_NULL = is.null(formula_list))
  check_list(formula_prior_list, "formula_prior_list", check_names = names(formula_list), allow_other = FALSE, all_objects = TRUE, allow_NULL = is.null(formula_list))
  check_list(formula_scale_list, "formula_scale_list", allow_NULL = TRUE)

  ### add formulas
  if(!is.null(formula_list)){

    # obtain settings for each formula
    formula_output <- list()
    for(parameter in names(formula_list)){
      formula_output[[parameter]] <- JAGS_formula(
        formula        = formula_list[[parameter]],
        parameter      = parameter,
        data           = formula_data_list[[parameter]],
        prior_list     = formula_prior_list[[parameter]],
        formula_scale  = if(!is.null(formula_scale_list)) formula_scale_list[[parameter]] else NULL)
    }

    # merge with the rest of the input
    prior_list     <- c(do.call(c, unname(lapply(formula_output, function(output) output[["prior_list"]]))), prior_list)
    data           <- c(do.call(c, unname(lapply(formula_output, function(output) output[["data"]]))),       data)
    formula_syntax <- paste0(lapply(formula_output, function(output) output[["formula_syntax"]]), collapse = "")
    
    # collect formula_scale information
    formula_scale_info <- lapply(formula_output, function(output) output[["formula_scale"]])
    formula_scale_info <- formula_scale_info[!sapply(formula_scale_info, is.null)]
    if(length(formula_scale_info) == 0) formula_scale_info <- NULL
    formula_design_info <- lapply(formula_output, function(output) output[["formula_design"]])

    # add the formula syntax to the model syntax
    opening_bracket <- regexpr("{", model_syntax, fixed = TRUE)[1]
    syntax_start    <- substr(model_syntax, 1, opening_bracket)
    syntax_end      <- substr(model_syntax, opening_bracket + 1, nchar(model_syntax))
    model_syntax    <- paste0(syntax_start, "\n", formula_syntax, "\n", syntax_end)
  }else{
    formula_scale_info <- NULL
    formula_design_info <- NULL
  }


  ### create the model call
  model_call <- list(
    model     = JAGS_add_priors(syntax = model_syntax, prior_list = prior_list),
    data      = data,
    inits     = JAGS_get_inits(prior_list, chains = chains, seed = seed),
    monitor   = c(JAGS_to_monitor(prior_list), add_parameters),
    n.chains  = chains,
    adapt     = adapt,
    burnin    = burnin,
    sample    = sample,
    thin      = thin,
    summarise = FALSE
  )

  # parallel vs. not
  if(parallel){
    cl <- parallel::makePSOCKcluster(cores)
    on.exit(try(parallel::stopCluster(cl), silent = TRUE), add = TRUE)
    .JAGS_require_packages(required_packages, cl)
    model_call <- c(
      model_call,
      method = "rjparallel",
      cl     = list(cl)
    )
  }else{
    .JAGS_require_packages(required_packages)
    model_call <- c(
      model_call,
      method = "rjags"
    )
  }


  if(!is.null(seed)){
    set.seed(seed)
  }

  # set silent mode
  if(silent){
    user_silent.jags    <- runjags::runjags.getOption("silent.jags")
    user_silent.runjags <- runjags::runjags.getOption("silent.runjags")
    on.exit(runjags::runjags.options(silent.jags = user_silent.jags, silent.runjags = user_silent.runjags), add = TRUE)
    runjags::runjags.options(silent.jags = TRUE, silent.runjags = TRUE)
  }

  start_time <- Sys.time()
  # special fitting procedure for JASP
  # singlcore interrupted fits allowing for bar progression
  if(isTRUE(dots[["is_JASP"]])){

    model_call_adapt  <- model_call
    model_call_adapt[["sample"]] <- 1 # at least one burnin & adapt need to be specified

    # adapt & burnin
    .JASP_progress_bar_start(n = 1, label = paste0(if(!is.null(dots[["is_JASP_prefix"]])) paste0(dots[["is_JASP_prefix"]], ": "), "Adapting and burnin the model"))
    fit <- tryCatch(do.call(runjags::run.jags, model_call_adapt), error = function(e) e)
    .JASP_progress_bar_tick()

    # sample
    .JASP_progress_bar_start(n = 5, label = paste0(if(!is.null(dots[["is_JASP_prefix"]])) paste0(dots[["is_JASP_prefix"]], ": "), "Sampling the model"))
    for(i in 1:5){
      if(!inherits(fit, "error")){
        fit <- tryCatch(runjags::extend.jags(fit, burnin = 0, sample = floor((model_call[["sample"]])/5)), error = function(e)e)
        .JASP_progress_bar_tick()
      }
    }

  }else{
    if(is.null(autofit_control[["restarts"]])){
      fit <- tryCatch(do.call(runjags::run.jags, model_call), error = function(e) e)
    }else{
      for(i in 1:autofit_control[["restarts"]]){
        fit <- tryCatch(do.call(runjags::run.jags, model_call), error = function(e) e)
        if(!inherits(fit, "error")){
          break
        }else{
          # restart with different inits
          model_call$inits <- JAGS_get_inits(prior_list, chains = chains, seed = if(!is.null(seed)) seed + i)
        }
      }
    }
  }



  if(inherits(fit, "error") & !silent)
    warning(paste0("The model estimation failed with the following error: ", fit$message), immediate. = TRUE)

  if(autofit && !inherits(fit, "error")){

    converged  <- JAGS_check_convergence(fit, prior_list, autofit_control[["max_Rhat"]], autofit_control[["min_ESS"]], autofit_control[["max_error"]], autofit_control[["max_SD_error"]], check_indicators = autofit_control[["check_indicators"]], fail_fast = TRUE)
    itteration <- 1

    if(!converged && isTRUE(dots[["is_JASP"]]))
      .JASP_progress_bar_start(n = if (!is.null(autofit_control[["max_extend"]])) autofit_control[["max_extend"]] else 10, label = paste0(if(!is.null(dots[["is_JASP_prefix"]])) paste0(dots[["is_JASP_prefix"]], ": "), "Extending the model (autofit)"))

    while(!converged){

      if(!is.null(autofit_control[["max_time"]]) && difftime(Sys.time(), start_time, units = autofit_control[["max_time"]][["unit"]]) > autofit_control[["max_time"]][["time"]]){
        if(!silent){
          attr(fit, "warning") <- "The automatic model fitting was terminated due to the 'max_time' constraint."
          warning(attr(fit, "warning"), immediate. = TRUE)
        }
        break
      }
      if(!is.null(autofit_control[["max_extend"]]) && itteration > autofit_control[["max_extend"]]){
        if(!silent){
          attr(fit, "warning") <- "The automatic model fitting was terminated due to the 'max_extend' constraint."
          warning(attr(fit, "warning"), immediate. = TRUE)
        }
        break
      }

      fit <- tryCatch(runjags::extend.jags(fit, sample = autofit_control[["sample_extend"]]), error = function(e)e)

      if(inherits(fit, "error")){
        if(!silent)
          warning(paste0("The model estimation failed with the following error: ", fit$message), immediate. = TRUE)
        break
      }

      fit <- runjags::add.summary(fit)

      converged  <- JAGS_check_convergence(fit, prior_list, autofit_control[["max_Rhat"]], autofit_control[["min_ESS"]], autofit_control[["max_error"]], autofit_control[["max_SD_error"]], check_indicators = autofit_control[["check_indicators"]], fail_fast = TRUE)
      itteration <- itteration + 1

      if(isTRUE(dots[["is_JASP"]]))
        .JASP_progress_bar_tick()
    }
  }

  # add information to the fitted object
  attr(fit, "prior_list")   <- prior_list
  attr(fit, "model_syntax") <- model_syntax
  attr(fit, "required_packages") <- required_packages
  if(!is.null(formula_scale_info)){
    # Keep formula_scale as a nested list keyed by parameter name
    # Each element contains the scaling info for that parameter's predictors
    attr(fit, "formula_scale") <- formula_scale_info
  }
  if(!is.null(formula_design_info)){
    attr(fit, "formula_design") <- formula_design_info
  }

  class(fit) <- c(class(fit), "BayesTools_fit")

  return(fit)
}

#' @rdname JAGS_fit
JAGS_extend <- function(fit, autofit_control = list(max_Rhat = 1.05, min_ESS = 500, max_error = 0.01, max_SD_error = 0.05, max_time = list(time = 60, unit = "mins"), sample_extend = 1000, restarts = 10, max_extend = 10, check_indicators = FALSE),
                        parallel = FALSE, cores = NULL, silent = TRUE, seed = NULL){

  if(!inherits(fit, "BayesTools_fit"))
    stop("'fit' must be a 'BayesTools_fit'")

  # extract fitting information
  prior_list        <- attr(fit, "prior_list")
  model_syntax      <- attr(fit, "model_syntax")
  required_packages <- attr(fit, "required_packages")
  formula_scale     <- attr(fit, "formula_scale")
  formula_design    <- attr(fit, "formula_design")
  autofit_control <- JAGS_check_and_list_autofit_settings(autofit_control)

  # parallel vs. not
  if(parallel){
    if(is.null(cores)){
      cores <- length(fit[["mcmc"]])
    }
    cl <- parallel::makePSOCKcluster(cores)
    on.exit(try(parallel::stopCluster(cl), silent = TRUE), add = TRUE)
    .JAGS_require_packages(required_packages, cl)
    refit_call <- list(
      runjags.object = fit,
      sample         = autofit_control[["sample_extend"]],
      method         = "rjparallel",
      cl             = cl,
      summarise      = FALSE
    )
  }else{
    .JAGS_require_packages(required_packages)
    refit_call <- list(
      runjags.object = fit,
      sample         = autofit_control[["sample_extend"]],
      method         = "rjags",
      summarise      = FALSE
    )
  }


  if(!is.null(seed)){
    set.seed(seed)
  }

  # set silent mode
  if(silent){
    user_silent.jags    <- runjags::runjags.getOption("silent.jags")
    user_silent.runjags <- runjags::runjags.getOption("silent.runjags")
    on.exit(runjags::runjags.options(silent.jags = user_silent.jags, silent.runjags = user_silent.runjags), add = TRUE)
    runjags::runjags.options(silent.jags = TRUE, silent.runjags = TRUE)
  }

  start_time <- Sys.time()
  itteration <- 0
  converged  <- FALSE

  while(!converged){

    if(!is.null(autofit_control[["max_time"]]) && difftime(Sys.time(), start_time, units = autofit_control[["max_time"]][["unit"]]) > autofit_control[["max_time"]][["time"]]){
      if(!silent){
        attr(fit, "warning") <- "The automatic model fitting was terminated due to the 'max_time' constraint."
        warning(attr(fit, "warning"), immediate. = TRUE)
      }
      break
    }
    if(!is.null(autofit_control[["max_extend"]]) && itteration >= autofit_control[["max_extend"]]){
      if(!silent){
        attr(fit, "warning") <- "The automatic model fitting was terminated due to the 'max_extend' constraint."
        warning(attr(fit, "warning"), immediate. = TRUE)
      }
      break
    }

    fit <- tryCatch(do.call(runjags::extend.jags, refit_call), error = function(e)e)

    if(inherits(fit, "error")){
      if(!silent)
        warning(paste0("The model estimation failed with the following error: ", fit$message), immediate. = TRUE)

      break
    }

    converged <- JAGS_check_convergence(fit, prior_list, autofit_control[["max_Rhat"]], autofit_control[["min_ESS"]], autofit_control[["max_error"]], autofit_control[["max_SD_error"]], check_indicators = autofit_control[["check_indicators"]], fail_fast = TRUE)

    # update the refit call
    if(!converged){
      itteration <- itteration + 1
      refit_call$runjags.object <- fit
    }
  }

  # add information to the fitted object
  attr(fit, "prior_list")   <- prior_list
  attr(fit, "model_syntax") <- model_syntax
  attr(fit, "required_packages") <- required_packages
  if(!is.null(formula_scale)){
    attr(fit, "formula_scale") <- formula_scale
  }
  if(!is.null(formula_design)){
    attr(fit, "formula_design") <- formula_design
  }

  class(fit) <- c(class(fit), "BayesTools_fit")

  return(fit)
}


#' @title Assess convergence of a runjags model
#'
#' @description Checks whether the supplied \link[runjags]{runjags-package} model
#' satisfied convergence criteria.
#' @param fit a runjags model
#' @param prior_list named list of prior distribution
#' (names correspond to the parameter names)
#' @param max_Rhat maximum R-hat error for the autofit function.
#'   Defaults to \code{1.05}.
#' @param min_ESS minimum effective sample size. Defaults to \code{500}.
#' @param max_error maximum MCMC error. Defaults to \code{0.01}.
#' @param max_SD_error maximum MCMC error as the proportion of standard
#'   deviation of the parameters. Defaults to \code{0.05}.
#' @param add_parameters vector of additional parameter names that should be used
#' (only allows removing last, fixed, omega element if omega is tracked manually).
#' @param fail_fast whether the function should stop after the first failed convergence check.
#' @param check_indicators whether model indicator variables should be included
#' in convergence checks. Defaults to \code{FALSE}.
#'
#' @examples \dontrun{
#' # simulate data
#' set.seed(1)
#' data <- list(
#'   x = rnorm(10),
#'   N = 10
#' )
#' data$x
#'
#' # define priors
#' priors_list <- list(mu = prior("normal", list(0, 1)))
#'
#' # define likelihood for the data
#' model_syntax <-
#'   "model{
#'     for(i in 1:N){
#'       x[i] ~ dnorm(mu, 1)
#'     }
#'   }"
#'
#' # fit the models
#' fit <- JAGS_fit(model_syntax, data, priors_list)
#' JAGS_check_convergence(fit, priors_list)
#' }
#' @return \code{JAGS_check_convergence} returns a boolean
#' indicating whether the model converged or not, with an
#' attribute 'errors' carrying the failed convergence checks (if any).
#'
#' @seealso [JAGS_fit()]
#' @export
JAGS_check_convergence <- function(fit, prior_list, max_Rhat = 1.05, min_ESS = 500, max_error = 0.01, max_SD_error = 0.05, add_parameters = NULL, fail_fast = FALSE, check_indicators = FALSE){

  # check input
  if(!inherits(fit, "runjags"))
    stop("'fit' must be a runjags fit")
  check_list(prior_list, "prior_list", allow_NULL = TRUE)
  if(!is.null(prior_list) && any(!sapply(prior_list, is.prior)))
    stop("'prior_list' must be a list of priors.")
  check_real(max_Rhat,     "max_Rhat",     lower = 1, allow_NULL = TRUE)
  check_real(min_ESS,      "min_ESS",      lower = 0, allow_NULL = TRUE)
  check_real(max_error,    "max_error",    lower = 0, allow_NULL = TRUE)
  check_real(max_SD_error, "max_SD_error", lower = 0, upper = 1, allow_NULL = TRUE)
  check_char(add_parameters, "add_parameters", check_length = 0, allow_NULL = TRUE)
  check_bool(check_indicators, "check_indicators")

  # extract samples and parameter information
  mcmc_samples_list <- .extract_posterior_samples(fit, as_list = TRUE)
  mcmc_samples      <- do.call(rbind, mcmc_samples_list)
  
  # build remove_parameters list: point priors, spike priors, indicators, inclusions
  remove_params <- c(
    # point priors
    names(prior_list)[sapply(prior_list, is.prior.point)],
    # mixture with single point prior
    names(prior_list)[sapply(prior_list, function(p) {
      is.prior.mixture(p) && length(p) == 1 && is.prior.point(p[[1]])
    })],
    # add_parameters that should be excluded
    add_parameters
  )
  
  # use helper to remove auxiliary parameters
  cleaned <- .remove_auxiliary_parameters(mcmc_samples, prior_list, remove_params)
  mcmc_samples <- cleaned$model_samples
  
  # remove auxiliary inclusion probabilities and, by default, model indicators
  indicator_cols <- grepl("_indicator", colnames(mcmc_samples))
  inclusion_cols <- grepl("_inclusion", colnames(mcmc_samples))
  mcmc_samples <- mcmc_samples[, !(inclusion_cols | (!check_indicators & indicator_cols)), drop = FALSE]
  
  if(ncol(mcmc_samples) == 0){
    return(TRUE)
  }
  
  # convert back to mcmc.list for convergence checks
  n_chains <- length(mcmc_samples_list)
  samples_per_chain <- nrow(mcmc_samples) / n_chains
  mcmc_samples_list_cleaned <- lapply(1:n_chains, function(i) {
    start_idx <- (i - 1) * samples_per_chain + 1
    end_idx <- i * samples_per_chain
    coda::as.mcmc(mcmc_samples[start_idx:end_idx, , drop = FALSE])
  })
  mcmc_samples <- coda::as.mcmc.list(mcmc_samples_list_cleaned)

  ### check the convergence
  fails <- NULL

  # assess R-hat
  if(!is.null(max_Rhat)){
    if(length(fit$mcmc) == 1){
      warning("Only one chain was run. R-hat cannot be computed.", immediate. = TRUE)
    }else{
      temp_Rhat <- coda::gelman.diag(mcmc_samples, multivariate = FALSE, autoburnin = FALSE)$psrf
      temp_Rhat[is.na(temp_Rhat)] <- 1
      temp_Rhat <- max(temp_Rhat)
      if(temp_Rhat > max_Rhat){
        fails <- c(fails, paste0("R-hat ", round(temp_Rhat, 3), " is larger than the set target (", max_Rhat, ")."))
        if(fail_fast){
          return(FALSE)
        }
      }
    }
  }

  if(!is.null(min_ESS)){
    temp_ESS <- coda::effectiveSize(mcmc_samples)
    temp_ESS[is.nan(temp_ESS) | temp_ESS == 0] <- Inf
    temp_ESS <- min(temp_ESS)
    if(temp_ESS < min_ESS){
      fails <- c(fails, paste0("ESS ", round(temp_ESS), " is lower than the set target (", min_ESS, ")."))
      if(fail_fast){
        return(FALSE)
      }
    }
  }

  # compute the MCMC error and & SD error
  if(!(is.null(max_error) && is.null(max_SD_error))){
    temp_summary <- summary(mcmc_samples, quantiles = NULL)$statistics
    if(is.null(dim(temp_summary))){
      temp_summary <- t(temp_summary)
    }
  }


  if(!is.null(max_error)){
    temp_error    <- temp_summary[,"Time-series SE"]
    temp_error[is.na(temp_error)] <- 0
    temp_error    <- max(temp_error)
    if(temp_error > max_error){
      fails <- c(fails, paste0("MCMC error ", round(temp_error, 5), " is larger than the set target (", max_error, ")."))
      if(fail_fast){
        return(FALSE)
      }
    }
  }

  if(!is.null(max_SD_error)){
    temp_error_SD <- temp_summary[,"Time-series SE"] / temp_summary[,"SD"]
    temp_error_SD[is.na(temp_error_SD)] <- 0
    temp_error_SD <- max(temp_error_SD)
    if(temp_error_SD > max_SD_error){
      fails <- c(fails, paste0("MCMC SD error ", round(temp_error_SD, 3), " is larger than the set target (", max_SD_error, ")."))
      if(fail_fast){
        return(FALSE)
      }
    }
  }

  converged <- length(fails) == 0
  attr(converged, "errors") <- fails
  return(converged)
}


#' @title Add 'JAGS' prior
#'
#' @description Adds priors to a 'JAGS' syntax.
#'
#' @param syntax JAGS model syntax
#' @param prior_list named list of prior distribution
#' (names correspond to the parameter names)
#'
#' @return \code{JAGS_add_priors} returns a JAGS syntax.
#'
#' @export
JAGS_add_priors           <- function(syntax, prior_list){

  # return the original syntax in case that no prior was specified
  if(length(prior_list) == 0){
    return(syntax)
  }

  check_list(prior_list, "prior_list")
  if(is.prior(prior_list) | !all(sapply(prior_list, is.prior)))
    stop("'prior_list' must be a list of priors.")
  .check_prior_list_unique_names(prior_list)
  .check_JAGS_syntax(syntax)

  # identify parts of the syntax
  opening_bracket <- regexpr("{", syntax, fixed = TRUE)[1]
  syntax_start    <- substr(syntax, 1, opening_bracket)
  syntax_end      <- substr(syntax, opening_bracket + 1, nchar(syntax))

  # create the priors relevant syntax
  syntax_priors <- .JAGS_add_priors.fun(prior_list)

  # merge everything back together
  syntax <- paste0(syntax_start, "\n", syntax_priors, "\n", syntax_end)

  return(syntax)
}

.JAGS_add_priors.fun       <- function(prior_list){

  syntax_priors <- ""
  syntax_attributes <- NULL

  for(i in seq_along(prior_list)){

    if(is.prior.weightfunction(prior_list[[i]])){

      syntax_priors <- paste(syntax_priors, .JAGS_prior.weightfunction(prior_list[[i]]))

    }else if(is_prior_phacking(prior_list[[i]])){

      syntax_priors <- paste(syntax_priors, .JAGS_prior.phacking(prior_list[[i]]))

    }else if(is_prior_bias(prior_list[[i]])){

      syntax_priors <- paste(syntax_priors, .JAGS_prior.bias(prior_list[[i]]))

    }else if(is.prior.PET(prior_list[[i]]) | is.prior.PEESE(prior_list[[i]])){

      syntax_priors <- paste(syntax_priors, .JAGS_prior.PP(prior_list[[i]]))

    }else if(is.prior.spike_and_slab(prior_list[[i]])){

      syntax_priors <- paste(syntax_priors, .JAGS_prior.spike_and_slab(prior_list[[i]], names(prior_list)[i]))

    }else if(is.prior.mixture(prior_list[[i]])){

      syntax_priors <- paste(syntax_priors, .JAGS_prior.mixture(prior_list[[i]], names(prior_list)[i]))

    }else if(is.prior.factor(prior_list[[i]])){

      syntax_priors <- paste(syntax_priors, .JAGS_prior.factor(prior_list[[i]], names(prior_list)[i]))

    }else if(is.prior.vector(prior_list[[i]])){

      syntax_priors <- paste(syntax_priors, .JAGS_prior.vector(prior_list[[i]], names(prior_list)[i]))

    }else if(is.prior.simple(prior_list[[i]])){

      syntax_priors <- paste(syntax_priors, .JAGS_prior.simple(prior_list[[i]], names(prior_list)[i]))

    }
  }

  return(syntax_priors)
}
.JAGS_prior.simple         <- function(prior, parameter_name){

  .check_prior(prior, allow_expressions = TRUE)
  if(!is.prior.simple(prior))
    stop("improper prior provided")
  check_char(parameter_name, "parameter_name")

  # parse expressions to test
  if(.is_prior_expression(prior)){
    prior <- .prior_expression_to_character(prior)
  }

  # distribution
  syntax <- switch(
    prior[["distribution"]],
    "point"     = paste0(parameter_name," = ",prior$parameter[["location"]]),
    "normal"    = paste0(parameter_name," ~ dnorm(",prior$parameter[["mean"]],",", .JAGS_parameter_to_precision(prior$parameter[["sd"]]),")"),
    "lognormal" = paste0(parameter_name," ~ dlnorm(",prior$parameter[["meanlog"]],",", .JAGS_parameter_to_precision(prior$parameter[["sdlog"]]),")"),
    "t"         = paste0(parameter_name," ~ dt(",prior$parameter[["location"]],",", .JAGS_parameter_to_precision(prior$parameter[["scale"]]),",", prior$parameter[["df"]],")"),
    "gamma"     = paste0(parameter_name," ~ dgamma(",prior$parameter[["shape"]],",",prior$parameter[["rate"]],")"),
    "invgamma"  = paste0("inv_",parameter_name," ~ dgamma(",prior$parameter[["shape"]],",",prior$parameter[["scale"]],")"),
    "exp"       = paste0(parameter_name," ~ dexp(",prior$parameter[["rate"]],")"),
    "beta"      = paste0(parameter_name," ~ dbeta(",prior$parameter[["alpha"]],",",prior$parameter[["beta"]],")"),
    "bernoulli" = paste0(parameter_name," ~ dbern(",prior$parameter[["probability"]],")"),
    "uniform"   = paste0(parameter_name," ~ dunif(",prior$parameter[["a"]],",",prior$parameter[["b"]],")")
  )

  # add truncation
  if(!.is_prior_default_range(prior)){
    # the truncation for invgamma needs to be done in reverse since we sample from gamma
    if(prior[["distribution"]] == "invgamma"){
      syntax <- paste0(syntax, "T(",
                       ifelse(is.infinite(prior$truncation[["upper"]]^-1),"",prior$truncation[["upper"]]^-1),
                       ",",
                       ifelse(is.infinite(prior$truncation[["lower"]]^-1),"",prior$truncation[["lower"]]^-1),
                       ")")
    }else{
      syntax <- paste0(syntax, "T(",
                       ifelse(is.infinite(prior$truncation[["lower"]]),"",prior$truncation[["lower"]]),
                       ",",
                       ifelse(is.infinite(prior$truncation[["upper"]]),"",prior$truncation[["upper"]]),
                       ")")
    }
  }

  # finish the line
  syntax <- paste0(syntax, "\n")

  # transform the parameter in case of inverse-gamma
  if(prior[["distribution"]] == "invgamma"){
    syntax <- paste0(syntax, "  ", parameter_name," = pow(inv_",parameter_name,", -1)\n")
  }

  return(syntax)
}
.JAGS_prior.vector         <- function(prior, parameter_name){

  .check_prior(prior, allow_expressions = TRUE)
  if(!is.prior.vector(prior))
    stop("improper prior provided")
  check_char(parameter_name, "parameter_name")
  check_int(prior$parameters[["K"]], "K", lower = 1)
  if(prior[["distribution"]] != "mpoint")
    .check_vector_truncation_unsupported(prior$truncation)

  # parse expressions to test
  if(.is_prior_expression(prior)){
    prior <- .prior_expression_to_character(prior)
  }

  if(prior[["distribution"]] %in% c("mnormal", "mt")){
    # create the location/means vector the sigma matrix

    par1 <- switch(
      prior[["distribution"]],
      "mnormal" = prior$parameter[["mean"]],
      "mt"      = prior$parameter[["location"]]
    )
    par2 <- switch(
      prior[["distribution"]],
      "mnormal" = prior$parameter[["sd"]],
      "mt"      = prior$parameter[["scale"]]
    )

    par2 <- .JAGS_parameter_to_precision(par2)

    # TODO: beautify this code by specific JAGS distributions?
    if(prior[["distribution"]] == "mt"){
      # using the chisq * covariance parametrization since the mt fails with 1 df
      # (using a common df parameter as in Rouder et al. 2012)
      syntax <- paste0("prior_par1_", parameter_name, " = rep(0,", prior$parameter[["K"]], ")\n")
      syntax <- paste0(syntax, "prior_par_s_", parameter_name, " ~ dgamma(", prior$parameter[["df"]]/2, ", ", prior$parameter[["df"]]/2,")\n")
      syntax <- paste0(
        syntax,
        "for(i in 1:", prior$parameters[["K"]], "){\n",
        "  prior_par2_", parameter_name, "[i,i] <- ", par2, "\n",
        "  for(j in 1:(i-1)){\n",
        "    prior_par2_", parameter_name, "[i,j] <- 0\n",
        "  }\n",
        "  for (j in (i+1):", prior$parameters[["K"]], "){\n",
        "    prior_par2_", parameter_name, "[i,j] <- 0\n",
        "  }\n",
        "}\n",
        "prior_par_z_", parameter_name, " ~ dmnorm(prior_par1_", parameter_name, ",prior_par2_", parameter_name, ")\n",
        "for(i in 1:", prior$parameters[["K"]], "){\n",
        "  ", parameter_name, "[i] <- prior_par_z_", parameter_name, "[i]/sqrt(prior_par_s_", parameter_name, ") + ", par1, " \n",
        "}\n")
    }else if(prior[["distribution"]] == "mnormal"){
      syntax <- paste0("prior_par1_", parameter_name, " = rep(", par1, ",", prior$parameter[["K"]], ")\n")
      syntax <- paste0(
        syntax,
        "for(i in 1:", prior$parameters[["K"]], "){\n",
        "  prior_par2_", parameter_name, "[i,i] <- ", par2, "\n",
        "  for(j in 1:(i-1)){\n",
        "    prior_par2_", parameter_name, "[i,j] <- 0\n",
        "  }\n",
        "  for (j in (i+1):", prior$parameters[["K"]], "){\n",
        "    prior_par2_", parameter_name, "[i,j] <- 0\n",
        "  }\n",
        "}\n")
      syntax <- paste0(syntax, parameter_name," ~ dmnorm(prior_par1_", parameter_name, ",prior_par2_", parameter_name, ")\n")
    }

  }else if(prior[["distribution"]] == "mpoint"){

    syntax <- paste0(
      "for(i in 1:", prior$parameters[["K"]], "){\n",
      "  ", parameter_name, "[i] = ", prior$parameter[["location"]], " \n",
      "}\n")

  }


  return(syntax)
}
.JAGS_prior.factor         <- function(prior, parameter_name){

  .check_prior(prior, allow_expressions = TRUE)
  if(!is.prior.factor(prior))
    stop("improper prior provided")
  check_char(parameter_name, "parameter_name")
  check_int(.get_prior_factor_levels(prior), "levels", lower = 1)

  if(is.prior.treatment(prior) | is.prior.independent(prior)){

    syntax <- paste0(
      "for(i in 1:", .get_prior_factor_levels(prior), "){\n",
      "  ", .JAGS_prior.simple(prior, paste0(parameter_name, "[i]")),
      "}\n")

  }else if(is.prior.orthonormal(prior) | is.prior.meandif(prior)){

    prior$parameters[["K"]] <- .get_prior_factor_levels(prior)

    syntax <- .JAGS_prior.vector(prior, parameter_name)

  }

  return(syntax)
}
.JAGS_prior.PP             <- function(prior){

  .check_prior(prior, allow_expressions = TRUE)
  if(!is.prior.PET(prior) & !is.prior.PEESE(prior))
    stop("improper prior provided")

  if(is.prior.PET(prior)){
    syntax <- .JAGS_prior.simple(prior, "PET")
  }else if(is.prior.PEESE(prior)){
    syntax <- .JAGS_prior.simple(prior, "PEESE")
  }

  return(syntax)
}
.JAGS_prior.weightfunction <- function(prior){

  .check_prior(prior)
  if(!is.prior.weightfunction(prior))
    stop("improper prior provided")

  spec <- selection_backend_spec(prior)
  return(.JAGS_selection_backend_syntax(spec))
}
.JAGS_prior.phacking      <- function(prior){

  .check_prior(prior)
  if(!is_prior_phacking(prior))
    stop("improper prior provided")

  spec <- selection_backend_spec(prior)
  return(.JAGS_selection_backend_syntax(spec))
}
.JAGS_prior.bias          <- function(prior){

  .check_prior(prior)
  if(!is_prior_bias(prior))
    stop("improper prior provided")

  spec <- selection_backend_spec(prior)
  return(.JAGS_selection_backend_syntax(spec))
}

.JAGS_selection_backend_syntax <- function(spec){

  code <- c(spec$prior_code, spec$transform_code)
  code <- code[nzchar(code)]
  if(length(code) == 0L){
    return("")
  }
  code <- sub("[\r\n]+$", "", code)

  paste0(paste0(code, collapse = "\n"), "\n")
}

.JAGS_weightfunction_component_syntax <- function(prior, component_id = NULL, global_cuts = NULL, force_one_sided = FALSE){

  J <- .weightfunction_n_bins(prior)
  syntax <- character()

  expansion     <- .weightfunction_mapping_expansion(prior, force_one_sided)
  all_cuts      <- if(is.null(global_cuts)) expansion$cuts else global_cuts
  needs_mapping <- !identical(all_cuts, .weightfunction_local_cuts(prior)) ||
    !identical(expansion$index, seq_len(J))

  omega_local <- if(is.null(component_id) && !needs_mapping) "omega" else if(is.null(component_id)) "omega_local" else paste0("omega_local_component_", component_id)
  omega_target <- if(is.null(component_id)) "omega" else paste0("omega_component_", component_id)

  if(prior$weights$type == "cumulative"){
    eta_name <- if(is.null(component_id)) "eta" else paste0("eta_component_", component_id)
    std_eta_name <- if(is.null(component_id)) "std_eta" else paste0("std_eta_component_", component_id)

    for(i in seq_len(J)){
      syntax <- paste0(syntax, eta_name, "[", i, "] ~ dgamma(", prior$weights$alpha[i], ", 1)\n")
    }
    syntax <- paste0(syntax,
      "for(j in 1:", J, "){\n",
      "  ", std_eta_name, "[j] <- ", eta_name, "[j] / sum(", eta_name, ")\n",
      "}\n",
      omega_local, "[1] <- 1\n"
    )
    if(J > 1L){
      syntax <- paste0(syntax,
        "for(j in 2:", J, "){\n",
        "  ", omega_local, "[j] <- sum(", std_eta_name, "[j:", J, "])\n",
        "}\n"
      )
    }

  }else if(prior$weights$type == "fixed"){
    for(i in seq_len(J)){
      syntax <- paste0(syntax, omega_local, "[", i, "] <- ", prior$weights$omega[i], "\n")
    }

  }else if(prior$weights$type == "independent"){
    syntax <- paste0(syntax, omega_local, "[1] <- 1\n")
    if(J > 1L){
      for(i in 2:J){
        if(prior$weights$scale == "omega"){
          syntax <- paste0(syntax, .JAGS_prior.simple(prior$weights$prior, paste0(omega_local, "[", i, "]")))
        }else if(prior$weights$scale == "log_omega"){
          log_omega_name <- if(is.null(component_id)) "log_omega" else paste0("log_omega_component_", component_id)
          syntax <- paste0(
            syntax,
            .JAGS_prior.simple(prior$weights$prior, paste0(log_omega_name, "[", i, "]")),
            omega_local, "[", i, "] <- exp(", log_omega_name, "[", i, "])\n"
          )
        }
      }
    }
  }

  if(!is.null(component_id) || needs_mapping){
    global_bin_indices <- .weightfunction_global_bin_indices(all_cuts, expansion)
    for(i in seq_len(length(all_cuts) - 1L)){
      ind <- global_bin_indices[i]
      syntax <- paste0(syntax, omega_target, "[", i, "] <- ", omega_local, "[", expansion$index[ind], "]\n")
    }
  }

  syntax
}
.JAGS_weightfunction_none_component_syntax <- function(component_id, n_bins){

  syntax <- character()
  omega_target <- if(is.null(component_id)) "omega" else paste0("omega_component_", component_id)
  for(i in seq_len(n_bins)){
    syntax <- paste0(syntax, omega_target, "[", i, "] <- 1\n")
  }
  syntax
}
.JAGS_prior.spike_and_slab <- function(prior, parameter_name){

  .check_prior(prior)
  if(!is.prior.spike_and_slab(prior))
    stop("improper prior provided")
  check_char(parameter_name, "parameter_name")

  prior_variable_list  <- list(.get_spike_and_slab_variable(prior))
  prior_inclusion_list <- list(.get_spike_and_slab_inclusion(prior))
  names(prior_variable_list)  <- paste0(parameter_name, "_variable")
  names(prior_inclusion_list) <- paste0(parameter_name, "_inclusion")

  syntax <- paste0(
    .JAGS_add_priors.fun(prior_variable_list),
    .JAGS_add_priors.fun(prior_inclusion_list),
    parameter_name, "_indicator ~ dbern(",   paste0(parameter_name, "_inclusion"), ")\n",
    parameter_name, " = ",  parameter_name, "_variable * ", parameter_name, "_indicator\n"
  )

  return(syntax)
}
.JAGS_prior.mixture        <- function(prior_list, parameter_name){

  .check_prior_list(prior_list, allow_expressions = TRUE)
  if(!is.prior.mixture(prior_list))
    stop("improper prior provided")
  check_char(parameter_name, "parameter_name")

  if(inherits(prior_list, "prior.bias_mixture")){

    # dispatch between publication bias prior mixture and a standard prior mixture
    is_PET            <- sapply(prior_list, is.prior.PET)
    is_PEESE          <- sapply(prior_list, is.prior.PEESE)
    is_weightfunction <- sapply(prior_list, is.prior.weightfunction)
    is_phacking       <- sapply(prior_list, is_prior_phacking)
    is_bias           <- sapply(prior_list, is_prior_bias)
    is_none           <- sapply(prior_list, is.prior.none)
    branch_info       <- lapply(prior_list, .selection_branch_info)
    has_selection     <- vapply(branch_info, function(x) !is.null(x$selection), logical(1))
    has_phacking      <- vapply(branch_info, function(x) !is.null(x$phacking),  logical(1))

    # if any prior is bias related, the whole component must be dispatching publication bias
    if(any(!(is_PET | is_PEESE | is_weightfunction | is_phacking | is_bias | is_none)))
      stop("Mixture of publication bias and standard priors is not supported.")

    prior_weights <- attr(prior_list, "prior_weights")
    if(any(has_selection) || any(has_phacking)){
      spec <- selection_backend_spec(prior_list)
      syntax <- .JAGS_selection_backend_syntax(spec)
    }else{
      syntax <- paste0(" bias_indicator ~ dcat(c(", paste0(prior_weights, collapse = ", "), "))\n")
    }

    if(any(is_PET)){
      if(sum(is_PET) > 1) stop("Only one PET style publication bias adjustment is allowed.")

      named_prior_PET <- prior_list[[which(is_PET)]]
      class(named_prior_PET) <- class(named_prior_PET)[!class(named_prior_PET) %in% "prior.PET"]
      named_prior_PET <- list("PET_1" = named_prior_PET)

      syntax <- paste0(
        syntax,
        .JAGS_add_priors.fun(named_prior_PET),
        " PET <- PET_1 * equals(bias_indicator, ", which(is_PET), ")\n"
      )
    }
    if(any(is_PEESE)){
      if(sum(is_PEESE) > 1) stop("Only one PEESE style publication bias adjustment is allowed.")

      named_prior_PEESE <- prior_list[[which(is_PEESE)]]
      class(named_prior_PEESE) <- class(named_prior_PEESE)[!class(named_prior_PEESE) %in% "prior.PEESE"]
      named_prior_PEESE <- list("PEESE_1" = named_prior_PEESE)

      syntax <- paste0(
        syntax,
        .JAGS_add_priors.fun(named_prior_PEESE),
        " PEESE <- PEESE_1 * equals(bias_indicator, ", which(is_PEESE), ")\n"
      )
    }

  }else{

    prior_weights    <- attr(prior_list, "prior_weights")
    prior_components <- as.list(prior_list)
    class(prior_components) <- "list"
    names(prior_components) <- paste0(parameter_name, "_component_", seq_along(prior_components))

    syntax <- paste0(
      " ", parameter_name, "_indicator ~ dcat(c(", paste0(prior_weights, collapse = ", "), "))\n",
      sapply(.JAGS_add_priors.fun(prior_components), paste, collapse = "\n"),
      " ", parameter_name, " = ",  paste0(names(prior_components), " * ", paste0("(", parameter_name, "_indicator == ", seq_along(prior_components), ")"), collapse = " + "), "\n"
    )
  }

  return(syntax)
}


.add_JAGS_vector   <- function(name, vector){

  if(!is.vector(vector))
    stop("vector must be a vector")
  check_char(name, "name")

  syntax <- paste0(" ", name, " = c(", paste0(vector, collapse = ", "), ")\n")

  return(syntax)
}
.add_JAGS_matrix   <- function(name, matrix){

  if(!is.matrix(matrix))
    stop("matrix must be a matrix")
  check_char(name, "name")

  syntax <- ""

  # this unfortunatelly cannot be defined on row/column basis
  # I tried simplifying this before but only possible initialization is elementwise
  for(i in 1:nrow(matrix)){
   syntax <- paste0(
     syntax, " ",
     paste0(name,"[", i, ",", seq_len(ncol(matrix)), "] = ", matrix[i,], collapse = "; "), "\n"
   )
  }

  return(syntax)
}
.check_JAGS_syntax <- function(syntax){

  check_char(syntax, "syntax", allow_NULL = TRUE)
  if(is.null(syntax)){
    syntax <- "model{}"
  }
  if(!grepl("model", syntax, fixed = TRUE))
    stop("syntax must be a JAGS model syntax")
  if(!grepl("{", syntax, fixed = TRUE))
    stop("syntax must be a JAGS model syntax")
  if(!grepl("}", syntax, fixed = TRUE))
    stop("syntax must be a JAGS model syntax")
}
.JAGS_parameter_to_precision <- function(parameter){

  if(is.character(parameter)){
    parameter <- paste0("1/pow(", parameter, ", 2)")
  }else{
    parameter <- 1/parameter^2
  }

  return(parameter)
}

#' @title Create initial values for 'JAGS' model
#'
#' @description Creates initial values for priors in
#' a 'JAGS' model.
#'
#' @param chains number of chains
#' @param seed seed for random number generation
#'
#' @inheritParams JAGS_add_priors
#'
#' @return \code{JAGS_add_priors} returns a list of JAGS
#' initial values.
#'
#' @export
JAGS_get_inits            <- function(prior_list, chains, seed){

  # return empty list in case that no prior was specified
  if(length(prior_list) == 0){
    return(list())
  }

  check_int(chains, "chains", lower = 1)
  check_real(seed, "seed", allow_NULL = TRUE)
  check_list(prior_list, "prior_list")
  if(is.prior(prior_list) | !all(sapply(prior_list, is.prior)))
    stop("'prior_list' must be a list of priors.")
  .check_prior_list_unique_names(prior_list)


  # select seed at random if none was specified
  if(is.null(seed)){
    seed <- sample(666666, 1)
  }
  set.seed(seed)


  # create the starting values
  inits <- vector("list", chains)
  for(j in 1:chains){

    temp_inits <- .JAGS_get_inits.fun(prior_list)

    temp_inits[[".RNG.seed"]] <- seed + j
    temp_inits[[".RNG.name"]] <- if(chains > 4) "lecuyer::RngStream" else "base::Super-Duper"

    inits[[j]] <- temp_inits
  }

  return(inits)
}

.JAGS_get_inits.fun        <- function(prior_list){

  temp_inits <- list()

  for(i in seq_along(prior_list)){

    if(is.prior.point(prior_list[[i]])){

      next

    }else if(is.prior.weightfunction(prior_list[[i]])){

      temp_inits <- c(temp_inits, .JAGS_init.weightfunction(prior_list[[i]]))

    }else if(is_prior_phacking(prior_list[[i]])){

      temp_inits <- c(temp_inits, .JAGS_init.phacking(prior_list[[i]]))

    }else if(is_prior_bias(prior_list[[i]])){

      temp_inits <- c(temp_inits, .JAGS_init.bias(prior_list[[i]]))

    }else if(is.prior.PET(prior_list[[i]]) | is.prior.PEESE(prior_list[[i]])){

      temp_inits <- c(temp_inits, .JAGS_init.PP(prior_list[[i]]))

    }else if(is.prior.spike_and_slab(prior_list[[i]])){

      temp_inits <- c(temp_inits, .JAGS_init.spike_and_slab(prior_list[[i]], names(prior_list)[i]))

    }else if(is.prior.mixture(prior_list[[i]])){

      temp_inits <- c(temp_inits, .JAGS_init.mixture(prior_list[[i]], names(prior_list)[i]))

    }else if(is.prior.factor(prior_list[[i]])){

      temp_inits <- c(temp_inits, .JAGS_init.factor(prior_list[[i]], names(prior_list)[i]))

    }else if(is.prior.vector(prior_list[[i]])){

      temp_inits <- c(temp_inits, .JAGS_init.vector(prior_list[[i]], names(prior_list)[i]))

    }else if(is.prior.simple(prior_list[[i]])){

      temp_inits <- c(temp_inits, .JAGS_init.simple(prior_list[[i]], names(prior_list)[i]))

    }
  }

  return(temp_inits)
}
.JAGS_init.simple          <- function(prior, parameter_name){

  .check_prior(prior, allow_expressions = TRUE)
  if(!is.prior.simple(prior))
    stop("improper prior provided")
  check_char(parameter_name, "parameter_name")

  # no initialization for expression priors: require higher-level input
  if(.is_prior_expression(prior)){
    return()
  }

  if(prior[["distribution"]] == "point"){

    return()

  }else{
    init <- list()

    if(prior[["distribution"]] == "invgamma"){

      sampling_prior <- prior(
        "distribution" = "gamma",
        "parameters"   = list("shape" = prior$parameters[["shape"]], "rate" = prior$parameters[["scale"]]),
        "truncation"   = list("lower" = prior$truncation[["upper"]]^-1, "upper" = prior$truncation[["lower"]]^-1))
      init[[paste0("inv_", parameter_name)]] <- rng(sampling_prior, 1)

    }else{

      init[[parameter_name]] <- rng(prior, 1)

    }
  }

  return(init)
}
.JAGS_init.vector          <- function(prior, parameter_name){

  .check_prior(prior, allow_expressions = TRUE)
  if(!is.prior.vector(prior))
    stop("improper prior provided")
  check_char(parameter_name, "parameter_name")

  # no initialization for expression priors: require higher-level input
  if(.is_prior_expression(prior)){
    return()
  }

  if(prior[["distribution"]] == "point"){

    return()

  }else{

    init <- list()


    if(prior[["distribution"]] == "mt"){
      init[[paste0("prior_par_s_", parameter_name)]] <- rng(prior("gamma", list(shape = prior$parameters[["df"]]/2, rate = prior$parameters[["df"]]/2)), 1)
      init[[paste0("prior_par_z_", parameter_name)]] <- rng(prior("mnormal", list(mean = 0, sd = prior$parameters[["scale"]], K = prior$parameters[["K"]])), 1)[1,]
    }else{
      init[[parameter_name]] <- rng(prior, 1)[1,]
    }

  }

  return(init)
}
.JAGS_init.factor          <- function(prior, parameter_name){

  .check_prior(prior, allow_expressions = TRUE)
  if(!is.prior.factor(prior))
    stop("improper prior provided")
  check_char(parameter_name, "parameter_name")
  check_int(.get_prior_factor_levels(prior), "levels", lower = 1)

  # no initialization for expression priors: require higher-level input
  if(.is_prior_expression(prior)){
    return()
  }

  if(is.prior.treatment(prior) | is.prior.independent(prior)){

    init <- list()
    init[[parameter_name]] <- rng(prior, .get_prior_factor_levels(prior))

  }else if(is.prior.orthonormal(prior) | is.prior.meandif(prior)){

    prior$parameters[["K"]] <- .get_prior_factor_levels(prior)

    # remove the orthonormal/meandif class, otherwise samples from the transformed distributions are generated
    class(prior) <- class(prior)[!class(prior) %in% c("prior.orthonormal", "prior.meandif")]

    init <- .JAGS_init.vector(prior, parameter_name)

  }

  return(init)
}
.JAGS_init.PP              <- function(prior){

  .check_prior(prior)
  if(!is.prior.PET(prior) & !is.prior.PEESE(prior))
    stop("improper prior provided")

  if(is.prior.PET(prior)){
    init <- .JAGS_init.simple(prior, "PET")
  }else if(is.prior.PEESE(prior)){
    init <- .JAGS_init.simple(prior, "PEESE")
  }

  return(init)
}
.JAGS_init.weightfunction  <- function(prior, component_id = NULL){

  .check_prior(prior)
  if(!is.prior.weightfunction(prior))
    stop("improper prior provided")

  if(is.null(component_id)){
    return(selection_backend_spec(prior)$init)
  }

  init <- list()
  if(prior$weights$type == "fixed"){
    return()
  }else if(prior$weights$type == "cumulative"){
    eta_name <- paste0("eta_component_", component_id)
    init[[eta_name]] <- stats::rgamma(length(prior$weights[["alpha"]]), shape = prior$weights[["alpha"]], rate = 1)
  }

  return(init)
}
.JAGS_init.phacking       <- function(prior, component_id = NULL){

  .check_prior(prior)
  if(!is_prior_phacking(prior))
    stop("improper prior provided")

  if(is.null(component_id)){
    return(selection_backend_spec(prior)$init)
  }

  alpha_name <- paste0("alpha_component_", component_id)
  init <- .JAGS_init.simple(prior$alpha, alpha_name)

  return(init)
}
.JAGS_init.bias           <- function(prior){

  .check_prior(prior)
  if(!is_prior_bias(prior))
    stop("improper prior provided")

  return(selection_backend_spec(prior)$init)
}
.JAGS_init.spike_and_slab  <- function(prior, parameter_name){

  .check_prior(prior)
  if(!is.prior.spike_and_slab(prior))
    stop("improper prior provided")

  prior_variable        <- list(.get_spike_and_slab_variable(prior))
  names(prior_variable) <- paste0(parameter_name, "_variable")
  init <- .JAGS_get_inits.fun(prior_variable)

  if(!is.prior.point(.get_spike_and_slab_inclusion(prior))){
    init[[paste0(parameter_name, "_inclusion")]] <- rng(.get_spike_and_slab_inclusion(prior), 1)
  }


  return(init)
}
.JAGS_init.mixture         <- function(prior_list, parameter_name){

  .check_prior_list(prior_list, allow_expressions = TRUE)
  if(!is.prior.mixture(prior_list))
    stop("improper prior provided")
  check_char(parameter_name, "parameter_name")

  if(inherits(prior_list, "prior.bias_mixture")){

    # dispatch between publication bias prior mixture and a standard prior mixture
    is_PET            <- sapply(prior_list, is.prior.PET)
    is_PEESE          <- sapply(prior_list, is.prior.PEESE)
    is_weightfunction <- sapply(prior_list, is.prior.weightfunction)
    is_phacking       <- sapply(prior_list, is_prior_phacking)
    is_bias           <- sapply(prior_list, is_prior_bias)
    is_none           <- sapply(prior_list, is.prior.none)
    branch_info       <- lapply(prior_list, .selection_branch_info)
    has_selection     <- vapply(branch_info, function(x) !is.null(x$selection), logical(1))
    has_phacking      <- vapply(branch_info, function(x) !is.null(x$phacking),  logical(1))

    init <- list()

    # if any prior is bias related, the whole component must be dispatching publication bias
    if(any(!(is_PET | is_PEESE | is_weightfunction | is_phacking | is_bias | is_none)))
      stop("Mixture of publication bias and standard priors is not supported.")

    if(any(is_PET)){
      if(sum(is_PET) > 1) stop("Only one PET style publication bias adjustment is allowed.")

      named_prior_PET <- prior_list[[which(is_PET)]]
      class(named_prior_PET) <- class(named_prior_PET)[!class(named_prior_PET) %in% "prior.PET"]
      named_prior_PET <- list("PET_1" = named_prior_PET)

      init <- c(init, .JAGS_get_inits.fun(named_prior_PET))
    }
    if(any(is_PEESE)){
      if(sum(is_PEESE) > 1) stop("Only one PEESE style publication bias adjustment is allowed.")

      named_prior_PEESE <- prior_list[[which(is_PEESE)]]
      class(named_prior_PEESE) <- class(named_prior_PEESE)[!class(named_prior_PEESE) %in% "prior.PEESE"]
      named_prior_PEESE <- list("PEESE_1" = named_prior_PEESE)

      init <- c(init, .JAGS_get_inits.fun(named_prior_PEESE))
    }
    if(any(has_selection) || any(has_phacking)){
      init <- c(init, selection_backend_spec(prior_list)$init)
    }else{
      init[["bias_indicator"]] <- rng(prior_list, 1, sample_components = TRUE)
    }

  }else{

    prior_components <- as.list(prior_list)
    class(prior_components) <- "list"
    names(prior_components) <- paste0(parameter_name, "_component_", seq_along(prior_components))

    init <- .JAGS_get_inits.fun(prior_components)
    init[[paste0(parameter_name, "_indicator")]] <- rng(prior_list, 1, sample_components = TRUE)

  }

  return(init)
}


#' @title Create list of monitored parameters for 'JAGS' model
#'
#' @description Creates a vector of parameter names to be
#' monitored in a 'JAGS' model.
#'
#' @inheritParams JAGS_add_priors
#'
#' @return \code{JAGS_to_monitor} returns a character vector of
#' parameter names.
#'
#' @export
JAGS_to_monitor             <- function(prior_list){

  # return empty string in case that no prior was specified
  if(length(prior_list) == 0){
    return("")
  }

  check_list(prior_list, "prior_list")
  if(is.prior(prior_list) | !all(sapply(prior_list, is.prior)))
    stop("'prior_list' must be a list of priors.")
  .check_prior_list_unique_names(prior_list)


  # add the monitored parameters
  monitor <- character()
  for(i in seq_along(prior_list)){

    if(is.prior.weightfunction(prior_list[[i]])){

      monitor <- c(monitor, .JAGS_monitor.weightfunction(prior_list[[i]]))

    }else if(is_prior_phacking(prior_list[[i]])){

      monitor <- c(monitor, .JAGS_monitor.phacking(prior_list[[i]]))

    }else if(is_prior_bias(prior_list[[i]])){

      monitor <- c(monitor, .JAGS_monitor.bias(prior_list[[i]]))

    }else if(is.prior.PET(prior_list[[i]]) | is.prior.PEESE(prior_list[[i]])){

      monitor <- c(monitor, .JAGS_monitor.PP(prior_list[[i]]))

    }else if(is.prior.spike_and_slab(prior_list[[i]])){

      monitor <- c(monitor, .JAGS_monitor.spike_and_slab(prior_list[[i]], names(prior_list)[i]))

    }else if(is.prior.mixture(prior_list[[i]])){

      monitor <- c(monitor, .JAGS_monitor.mixture(prior_list[[i]], names(prior_list)[i]))

    }else if(is.prior.factor(prior_list[[i]])){

      monitor <- c(monitor, .JAGS_monitor.factor(prior_list[[i]], names(prior_list)[i]))

    }else if(is.prior.vector(prior_list[[i]])){

      monitor <- c(monitor, .JAGS_monitor.vector(prior_list[[i]], names(prior_list)[i]))

    }else if(is.prior.simple(prior_list[[i]])){

      monitor <- c(monitor, .JAGS_monitor.simple(prior_list[[i]], names(prior_list)[i]))

    }
  }

  if(length(monitor) == 0L){
    return("")
  }

  return(monitor)
}


.JAGS_monitor.simple         <- function(prior, parameter_name){

  .check_prior(prior, allow_expressions = TRUE)
  if(!(is.prior.simple(prior) | is.prior.vector(prior) | is.prior.factor(prior)))
    stop("improper prior provided")
  check_char(parameter_name, "parameter_name")

  if(prior[["distribution"]] %in% c("point", "mpoint")){
    monitor <- character()
  }else if(prior[["distribution"]] == "invgamma"){
    monitor <- c(parameter_name, paste0("inv_", parameter_name))
  }else{
    monitor <- parameter_name
  }

  return(monitor)
}
.JAGS_monitor.vector         <- function(prior, parameter_name){

  monitor <- .JAGS_monitor.simple(prior, parameter_name)

  return(monitor)
}
.JAGS_monitor.factor         <- function(prior, parameter_name){

  monitor <- .JAGS_monitor.simple(prior, parameter_name)

  return(monitor)
}
.JAGS_monitor.PP             <- function(prior){

  .check_prior(prior)
  if(!is.prior.PET(prior) & !is.prior.PEESE(prior))
    stop("improper prior provided")

  if(is.prior.PET(prior)){
    monitor <- .JAGS_monitor.simple(prior, "PET")
  }else if(is.prior.PEESE(prior)){
    monitor <- .JAGS_monitor.simple(prior, "PEESE")
  }

  return(monitor)
}
.JAGS_monitor.weightfunction <- function(prior){

  .check_prior(prior)
  if(!is.prior.weightfunction(prior))
    stop("improper prior provided")

  return(selection_backend_spec(prior)$monitor)
}
.JAGS_monitor_private.weightfunction <- function(prior){

  if(prior$weights$type == "cumulative"){
    return("eta")
  }
  if(prior$weights$type == "independent" && prior$weights$scale == "log_omega"){
    return("log_omega")
  }

  character()
}
.JAGS_monitor.phacking      <- function(prior){

  .check_prior(prior)
  if(!is_prior_phacking(prior))
    stop("improper prior provided")

  selection_backend_spec(prior)$monitor
}
.JAGS_monitor.bias          <- function(prior){

  .check_prior(prior)
  if(!is_prior_bias(prior))
    stop("improper prior provided")

  selection_backend_spec(prior)$monitor
}
.JAGS_monitor.spike_and_slab <- function(prior, parameter_name){

  .check_prior(prior)
  if(!is.prior.spike_and_slab(prior))
    stop("improper prior provided")
  check_char(parameter_name, "parameter_name")

  prior_variable  <- list(.get_spike_and_slab_variable(prior))
  prior_inclusion <- list(.get_spike_and_slab_inclusion(prior))
  names(prior_variable)  <- paste0(parameter_name, "_variable")
  names(prior_inclusion) <- paste0(parameter_name, "_inclusion")

  monitor <- c(
    paste0(parameter_name, "_indicator"),
    JAGS_to_monitor(prior_inclusion),
    parameter_name,
    JAGS_to_monitor(prior_variable)
  )

  return(monitor)
}
.JAGS_monitor.mixture        <- function(prior_list, parameter_name){

  .check_prior_list(prior_list, allow_expressions = TRUE)
  if(!is.prior.mixture(prior_list))
    stop("improper prior provided")
  check_char(parameter_name, "parameter_name")

  if(inherits(prior_list, "prior.bias_mixture")){

    # dispatch between publication bias prior mixture and a standard prior mixture
    is_PET            <- sapply(prior_list, is.prior.PET)
    is_PEESE          <- sapply(prior_list, is.prior.PEESE)
    is_weightfunction <- sapply(prior_list, is.prior.weightfunction)
    is_phacking       <- sapply(prior_list, is_prior_phacking)
    is_bias           <- sapply(prior_list, is_prior_bias)
    is_none           <- sapply(prior_list, is.prior.none)
    branch_info       <- lapply(prior_list, .selection_branch_info)
    has_selection     <- vapply(branch_info, function(x) !is.null(x$selection), logical(1))
    has_phacking      <- vapply(branch_info, function(x) !is.null(x$phacking),  logical(1))

    # if any prior is bias related, the whole component must be dispatching publication bias
    if(any(!(is_PET | is_PEESE | is_weightfunction | is_phacking | is_bias | is_none)))
      stop("Mixture of publication bias and standard priors is not supported.")

    monitor <- if(any(has_selection) || any(has_phacking)){
      selection_backend_spec(prior_list)$monitor
    }else{
      "bias_indicator"
    }

    if(any(is_PET)){
      if(sum(is_PET) > 1) stop("Only one PET style publication bias adjustment is allowed.")

      monitor <- c(monitor, "PET")
    }
    if(any(is_PEESE)){
      if(sum(is_PEESE) > 1) stop("Only one PEESE style publication bias adjustment is allowed.")

      monitor <- c(monitor, "PEESE")
    }
  }else{

    monitor <- c(paste0(parameter_name, "_indicator"), parameter_name)
  }

  return(unique(monitor))
}

#' @title Check and list 'JAGS' fitting settings
#'
#' @description Checks and lists settings for the
#' [JAGS_fit] function.
#'
#' @param check_mins named list of minimal values for which
#' should some input be checked. Defaults to:
#' \describe{
#'   \item{chains}{\code{1}}
#'   \item{adapt}{\code{50}}
#'   \item{burnin}{\code{50}}
#'   \item{sample}{\code{100}}
#'   \item{thin}{\code{1}}
#' }
#' @param skip_sample_extend whether \code{sample_extend}
#' is allowed to be NULL and skipped in the check
#'
#' @inheritParams JAGS_fit
#' @inheritParams check_input
#'
#' @return \code{JAGS_check_and_list_fit_settings} invisibly returns a
#' list of checked fit settings. \code{JAGS_check_and_list_autofit_settings}
#' invisibly returns a list of checked autofit settings.
#' parameter names.
#'
#' @export JAGS_check_and_list_fit_settings
#' @export JAGS_check_and_list_autofit_settings
#' @name JAGS_check_and_list
NULL

#' @rdname JAGS_check_and_list
JAGS_check_and_list_fit_settings     <- function(chains, adapt, burnin, sample, thin, autofit, parallel, cores, silent, seed, check_mins = list(chains = 1, adapt = 50, burnin = 50, sample = 100, thin = 1), call = ""){

  check_int(chains, "chains", lower = check_mins[["chains"]], call = call)
  check_int(adapt,  "adapt",  lower = check_mins[["adapt"]],  call = call)
  check_int(burnin, "burnin", lower = check_mins[["burnin"]], call = call)
  check_int(sample, "sample", lower = check_mins[["sample"]], call = call)
  check_int(thin,   "thin",   lower = check_mins[["thin"]],   call = call)
  check_bool(parallel, "parallel",                call = call)
  check_int(cores,     "cores", lower = 1,        call = call)
  check_bool(autofit,  "autofit",                 call = call)
  check_bool(silent,   "silent",                  call = call)
  check_int(seed,      "seed", allow_NULL = TRUE, call = call)

  return(invisible(list(
    chains   = chains,
    adapt    = adapt,
    burnin   = burnin,
    sample   = sample,
    thin     = thin,
    autofit  = autofit,
    parallel = parallel,
    cores    = cores,
    silent   = silent,
    seed     = seed
  )))
}

#' @rdname JAGS_check_and_list
JAGS_check_and_list_autofit_settings <- function(autofit_control, skip_sample_extend = FALSE, call = ""){

  check_list(autofit_control, "autofit_control", check_names = c("max_Rhat", "min_ESS", "max_error", "max_SD_error",  "max_time", "sample_extend", "restarts", "max_extend", "check_indicators"), call = call)
  if(is.null(autofit_control[["check_indicators"]])){
    autofit_control[["check_indicators"]] <- FALSE
  }
  check_real(autofit_control[["max_Rhat"]],     "max_Rhat",     lower = 1, allow_NULL = TRUE, call = call)
  check_real(autofit_control[["min_ESS"]],      "min_ESS",      lower = 0, allow_NULL = TRUE, call = call)
  check_real(autofit_control[["max_error"]],    "max_error",    lower = 0, allow_NULL = TRUE, call = call)
  check_real(autofit_control[["max_SD_error"]], "max_SD_error", lower = 0, upper = 1, allow_NULL = TRUE, call = call)
  check_bool(autofit_control[["check_indicators"]], "check_indicators", call = call)
  check_list(autofit_control[["max_time"]],     "max_time", check_names = c("time", "unit"), check_length = 2, allow_NULL = TRUE, call = call)
  if(!is.null(autofit_control[["max_time"]])){
    if(is.null(names(autofit_control[["max_time"]]))){
      names(autofit_control[["max_time"]]) <- c("time", "unit")
    }
    check_real(autofit_control[["max_time"]][["time"]], "max_time:time", lower = 0, call = call)
    check_char(autofit_control[["max_time"]][["unit"]], "max_time:unit", allow_values = c("secs", "mins", "hours", "days", "weeks"), call = call)
  }
  check_int(autofit_control[["sample_extend"]], "sample_extend", lower = 1, allow_NULL = skip_sample_extend, call = call)
  check_int(autofit_control[["restarts"]], "restarts", lower = 1, allow_NULL = TRUE, call = call)
  check_int(autofit_control[["max_extend"]], "max_extend", lower = 1, allow_NULL = TRUE, call = call)

  return(invisible(autofit_control))
}

# functions for checking input into custom distributions related functions
.check_log        <- function(log){
  check_bool(log, "log", allow_NA = FALSE)
}
.check_log.p      <- function(log.p){
  check_bool(log.p, "log.p", allow_NA = FALSE)
}
.check_lower.tail <- function(lower.tail){
  check_bool(lower.tail, "lower.tail", allow_NA = FALSE)
}
.check_x <- function(x, lower = -Inf, upper = Inf){
  check_real(x, "x", lower = lower, upper = upper, check_length = 0)
}
.check_n <- function(n){
  check_int(n, "n", lower = 1, allow_NA = FALSE)
}
.check_q <- function(q, lower = -Inf, upper = Inf){
  check_real(q, "q", lower = lower, upper = upper, check_length = 0)
}
.check_p <- function(p, log.p){
  if(!log.p){
    check_real(p, "p", lower = 0, upper = 1, check_length = 0)
  }else{
    check_real(p, "log(p)", upper = 0, check_length = 0)
  }
}

#' @title Interpret ensemble inference and estimates
#'
#' @description Provides textual summary for posterior
#' distributions created by [mix_posteriors] and ensemble inference
#' created by [ensemble_inference].
#'
#' @param inference model inference created by [ensemble_inference]
#' @param samples posterior samples created by [mix_posteriors]
#' @param specification list of lists specifying the generated text.
#' Each inner list carries: (1) \code{inference} specifying the name of
#' in the \code{inference} entry and optionally \code{inference_name}
#' as a name to use in the text and \code{inference_BF_name} as a
#' symbol to be used instead of the default \code{"BF"}. Finite-sample
#' BF bounds can be supplied either as a \code{bound_operator} attribute
#' on the numeric BF or as \code{inference_BF_bound_operator} /
#' \code{BF_bound_operator} with value \code{"<"} or \code{">"}. (2) \code{samples}
#' specifying the name of in the \code{samples} entry and optionally
#' \code{samples_name} as a name to use in the text, \code{samples_units} as
#' a unit text to be appended after the estimate, and \code{samples_conditional}
#' specifying whether the estimate is conditional or model-averaged.
#' @param method character specifying name of the method to be
#' appended at the beginning of each sentence.
#'
#'
#' @return \code{interpret} returns character.
#'
#' @export interpret
#' @export interpret2
#' @name interpret
#'
#' @seealso [ensemble_inference] [mix_posteriors] [BayesTools_model_tables] [BayesTools_ensemble_tables]
NULL

#' @rdname interpret
interpret <- function(inference, samples, specification, method){

  # check input
  check_list(specification, "specification", check_length = 0)
  sapply(specification, function(s){
    check_char(s$inference, allow_NULL = TRUE)
    check_char(s$inference_name, allow_NULL = TRUE)
    check_char(s$inference_BF_name, allow_NULL = TRUE)
    check_char(s$inference_BF_bound_operator, allow_values = c("<", ">"), allow_NULL = TRUE)
    check_char(s$samples, allow_NULL = TRUE)
    check_char(s$samples_name, allow_NULL = TRUE)
    check_char(s$samples_units, allow_NULL = TRUE)
    check_bool(s$samples_conditional, allow_NULL = TRUE)
  })
  check_list(inference, "inference", check_names = sapply(specification, function(s) s$inference), all_objects = TRUE, allow_other = TRUE)
  check_list(samples, "samples", check_names = unlist(sapply(specification, function(s) s$samples)), all_objects = TRUE, allow_other = TRUE)
  check_char(method, allow_NULL = TRUE)


  output <- ""

  for(i in seq_along(specification)){
    output <- paste0(output, .interpret.specification(inference, samples, specification[[i]], method), if(i != length(specification)) " ")
  }

  return(output)
}

#' @rdname interpret
interpret2                <- function(specification, method = NULL){

  # check input
  check_list(specification, "specification", check_length = 0)
  sapply(specification, function(s){
    check_char(s$inference_name,       "inference_name",        allow_NULL = TRUE)
    check_char(s$inference_BF_name,    "inference_BF_name",     allow_NULL = TRUE)
    if(!is.null(s$inference_BF) && !is.numeric(s$inference_BF))
      stop("The 'inference_BF' argument must be a numeric vector.", call. = FALSE)
    check_real(if(is.null(s$inference_BF)) NULL else as.numeric(s$inference_BF), "inference_BF", allow_NULL = TRUE)
    check_char(s$inference_BF_bound_operator, "inference_BF_bound_operator", allow_values = c("<", ">"), allow_NULL = TRUE)
    check_char(s$estimate_name,        "estimate_name",         allow_NULL = TRUE)
    check_real(s$estimate_samples,     "estimate_samples",      allow_NULL = TRUE, check_length = 0)
    check_char(s$estimate_units,       "estimate_units",        allow_NULL = TRUE)
    check_bool(s$estimate_conditional, "estimate_conditional",  allow_NULL = TRUE)
  })
  check_char(method, allow_NULL = TRUE)


  output <- ""

  for(i in seq_along(specification)){
    output <- paste0(output, .interpret.specification2(specification[[i]], method), if(i != length(specification)) " ")
  }

  return(output)
}

.interpret.specification  <- function(inference, samples, specification, method){

  temp_inference <- inference[[specification[["inference"]]]]
  temp_BF <- temp_inference[["BF"]]
  temp_BF_bound_operator <- temp_inference[["BF_bound_operator"]]
  if(is.null(temp_BF_bound_operator)){
    temp_BF_bound_operator <- temp_inference[["inference_BF_bound_operator"]]
  }
  if(is.null(temp_BF_bound_operator)){
    temp_BF_bound_operator <- specification[["inference_BF_bound_operator"]]
  }
  text_BF <- .interpret.BF(temp_BF, if(!is.null(specification[["inference_name"]])) specification[["inference_name"]] else specification[["inference"]],
                           specification[["inference_BF_name"]], temp_BF_bound_operator)

  if(is.null(specification[["samples"]])){
    return(paste0(method, " found ", text_BF, "."))
  }

  temp_par <- samples[[specification[["samples"]]]]
  text_par <- .interpret.par(temp_par, if(!is.null(specification[["samples_name"]])) specification[["samples_name"]] else specification[["samples"]],
                             specification[["samples_units"]], specification[["samples_conditional"]])

  return(paste0(method, " found ", text_BF, ", ", text_par, "."))
}
.interpret.specification2 <- function(specification, method){

  text_BF <- .interpret.BF(specification[["inference_BF"]], specification[["inference_name"]], specification[["inference_BF_name"]],
                           specification[["inference_BF_bound_operator"]])

  if(is.null(specification[["estimate_samples"]])){
    return(paste0(method, " found ", text_BF, "."))
  }

  text_par <- .interpret.par(specification[["estimate_samples"]], specification[["estimate_name"]], specification[["estimate_units"]], specification[["estimate_conditional"]])

  return(paste0(method, " found ", text_BF, ", ", text_par, "."))
}
.interpret.BF             <- function(BF, name, BF_name, BF_bound_operator = NULL){

  if(!is.numeric(BF))
    stop("The 'inference_BF' argument must be a numeric vector.", call. = FALSE)
  bound_operator <- if(is.null(BF_bound_operator)) attr(BF, "bound_operator") else BF_bound_operator
  bound_operator <- .standardize_BF_bound_operator(bound_operator, length(BF))
  BF <- as.numeric(BF)
  check_real(BF, "inference_BF", lower = 0, allow_bound = FALSE, allow_NA = FALSE)
  if(!is.finite(BF))
    stop("The 'inference_BF' argument must be finite.", call. = FALSE)

  bound_operator <- bound_operator[1]
  has_bound <- !is.na(bound_operator)
  bound_identifies_side <- !has_bound ||
    (bound_operator == ">" && BF >= 1) ||
    (bound_operator == "<" && BF <= 1)

  if(!bound_identifies_side){
    text <- paste0("a finite-sample bound for the evidence concerning the ", name)
  }else if(isTRUE(all.equal(BF, 1)) && !has_bound){
    text <- paste0("no evidence for or against the ", name)
  }else if(abs(log(BF)) > log(10)){
    text <- "strong evidence"
  }else if(abs(log(BF)) > log(3)){
    text <- "moderate evidence"
  }else{
    text <- "weak evidence"
  }

  if(bound_identifies_side && !(isTRUE(all.equal(BF, 1)) && !has_bound)){
    if(BF > 1 || (has_bound && isTRUE(all.equal(BF, 1)) && bound_operator == ">")){
      text <- paste0(text, " in favor of the ", name)
    }else{
      text <- paste0(text, " against the ", name)
    }
    if(has_bound){
      text <- paste0("at least ", text)
    }
  }

  BF <- format(round(BF, if(BF < 1) 3 else 2), nsmall = if(BF < 1) 3 else 2)

  if(is.null(BF_name)){
    text <- paste0(text, if(has_bound) ", BF " else ", BF = ", if(has_bound) paste(bound_operator, BF) else BF)
  }else{
    text <- paste0(text, ", ", BF_name, if(has_bound) " " else " = ", if(has_bound) paste(bound_operator, BF) else BF)
  }

  return(text)
}
.interpret.par            <- function(samples, name, unit, conditional){

  check_real(samples, "estimate_samples", allow_NA = FALSE, check_length = 0)
  if(length(samples) == 0)
    stop("The 'estimate_samples' argument cannot be empty.", call. = FALSE)
  if(any(!is.finite(samples)))
    stop("The 'estimate_samples' argument must be finite.", call. = FALSE)
  check_char(name, "estimate_name")

  est <- mean(samples)
  CI  <- unname(stats::quantile(samples, probs = c(0.025, 0.975)))

  est <- format(round(est, 3), nsmall = 3)
  CI  <- format(round(CI, 3),  nsmall = 3)

  text <- "with mean"

  if(is.null(conditional) || !conditional){
    text <- paste0(text, " model-averaged")
  }else{
    text <- paste0(text, " conditional")
  }

  if(!is.null(unit)){
    text <- paste0(text, " estimate ", name, " = ", est, " ", unit, ", 95% CI [", CI[1], ", ", CI[2], "]")
  }else{
    text <- paste0(text, " estimate ", name, " = ", est, ", 95% CI [", CI[1], ", ", CI[2], "]")
  }

  return(text)
}


#' @title Normalize structured interpretation sources
#'
#' @description \code{interpret_records} normalizes interpretation inputs from
#' one or more structured sources into traceable evidence, estimate, header,
#' note, prior, and footnote records. It is intended for downstream packages
#' that already own the domain-specific ordering and wording of an
#' interpretation, but want BayesTools to consistently extract table rows,
#' preserve Bayes-factor metadata, and carry traceability information.
#'
#' @param sources named list of interpretation sources. Each source can be a
#' \code{data.frame}/\code{BayesTools_table}, or a list with entries
#' \code{data}, optional \code{type}, and optional per-source \code{schema}.
#' Supported source types are \code{"table"}, \code{"record"}, and
#' \code{"records"}. Table schemas can define columns such as \code{row},
#' \code{BF}, \code{central}, \code{lower}, \code{upper}, and metadata such as
#' \code{BF_orientation}, \code{BF_scale}, \code{BF_name},
#' \code{BF_bound_operator}, \code{lower_prob}, \code{upper_prob},
#' \code{interval_level}, \code{units}, and \code{conditioning}.
#' @param plan ordered list specifying which records to create. Plan items can
#' have \code{kind = "evidence"}, \code{"estimate"}, \code{"pair"},
#' \code{"test_estimate"}, \code{"for_each"}, \code{"header"}, \code{"note"},
#' \code{"prior"}, or \code{"footnote"}. Pair items contain \code{evidence}
#' and/or \code{estimate} references. Dynamic \code{"for_each"} items generate
#' records from source row order.
#' @param output whether to return normalized \code{"records"} or a simple
#' generic \code{"text"} rendering. Downstream packages should usually request
#' \code{"records"} and render domain-specific text themselves.
#' @param missing how missing optional sources or rows should be handled.
#' @param method optional method name used only by the generic text renderer.
#' @param digits number of digits used only by the generic text renderer.
#' @param spec alias for \code{plan} used by \code{interpret_tables}.
#' @param ... additional arguments passed from \code{interpret_tables} to
#' \code{interpret_records}.
#'
#' @return \code{interpret_records} returns a data frame with class
#' \code{"BayesTools_interpret_records"} when \code{output = "records"}, or a
#' character vector when \code{output = "text"}. \code{interpret_tables} is a
#' convenience wrapper around \code{interpret_records}.
#'
#' @export
interpret_records <- function(sources, plan, output = c("records", "text"),
                              missing = c("error", "skip", "warn"),
                              method = NULL, digits = 3){

  output  <- match.arg(output)
  missing <- match.arg(missing)

  check_list(sources, "sources", check_length = 0)
  check_list(plan, "plan", check_length = 0)
  check_char(method, "method", allow_NULL = TRUE)
  check_int(digits, "digits", lower = 0)

  if(is.null(names(sources)) || any(!nzchar(names(sources)))){
    stop("The 'sources' argument must be a named list.", call. = FALSE)
  }

  sources <- .interpret_prepare_sources(sources)
  plan    <- .interpret_expand_plan(plan, sources, missing)

  record_list <- list()
  for(i in seq_along(plan)){
    item_records <- .interpret_plan_item_records(
      item          = plan[[i]],
      sources       = sources,
      missing       = missing,
      default_order = i
    )
    if(length(item_records) > 0){
      record_list <- c(record_list, item_records)
    }
  }

  records <- .interpret_bind_records(record_list)
  if(nrow(records) > 0){
    records <- records[order(records$order, seq_len(nrow(records))), , drop = FALSE]
    rownames(records) <- NULL
  }

  if(output == "text"){
    return(.interpret_records_text(records, method = method, digits = digits))
  }

  return(records)
}

#' @rdname interpret_records
#' @export
interpret_tables <- function(sources, spec, ...){
  interpret_records(sources = sources, plan = spec, ...)
}

.interpret_prepare_sources <- function(sources){

  out <- vector("list", length(sources))
  names(out) <- names(sources)

  for(source_name in names(sources)){
    source <- sources[[source_name]]

    if(inherits(source, "data.frame")){
      source <- list(type = "table", data = source, schema = list())
    }else if(is.list(source) && is.null(source[["data"]]) && !is.null(source[["kind"]])){
      source <- list(type = "record", data = source, schema = list())
    }else if(!is.list(source) || is.null(source[["data"]])){
      stop(paste0("The 'sources:", source_name, "' entry must be a table or a list with a 'data' entry."), call. = FALSE)
    }

    type <- source[["type"]]
    if(is.null(type)){
      type <- .interpret_infer_source_type(source[["data"]])
    }
    check_char(type, paste0("sources:", source_name, ":type"), allow_values = c("table", "record", "records"))

    schema <- source[["schema"]]
    if(is.null(schema)){
      schema <- list()
    }
    if(length(schema) > 0){
      check_list(schema, paste0("sources:", source_name, ":schema"), check_length = 0)
    }

    out[[source_name]] <- list(
      name     = source_name,
      type     = type,
      data     = source[["data"]],
      schema   = schema,
      metadata = source[["metadata"]]
    )
  }

  return(out)
}

.interpret_infer_source_type <- function(data){

  if(inherits(data, "data.frame")){
    return("table")
  }
  if(is.list(data) && !is.null(data[["kind"]])){
    return("record")
  }
  if(is.list(data)){
    return("records")
  }

  stop("Could not infer an interpretation source type.", call. = FALSE)
}

.interpret_expand_plan <- function(plan, sources, missing){

  expanded <- list()
  for(i in seq_along(plan)){
    item <- plan[[i]]
    check_list(item, paste0("plan:", i), check_length = 0)
    kind <- .interpret_plan_kind(item)

    if(kind == "for_each"){
      expanded <- c(expanded, .interpret_expand_for_each(item, sources, missing, i))
    }else{
      item[["kind"]] <- kind
      if(is.null(item[["order"]])){
        item[["order"]] <- i
      }
      expanded[[length(expanded) + 1L]] <- item
    }
  }

  return(expanded)
}

.interpret_plan_kind <- function(item){

  kind <- item[["kind"]]
  if(is.null(kind)){
    kind <- item[["type"]]
  }
  if(is.null(kind)){
    if(!is.null(item[["evidence"]]) && !is.null(item[["estimate"]])){
      kind <- "pair"
    }else if(!is.null(item[["text"]])){
      kind <- "note"
    }else{
      kind <- "evidence"
    }
  }
  check_char(kind, "plan:kind", allow_values = c(
    "evidence", "test", "estimate", "pair", "test_estimate",
    "for_each", "header", "note", "prior", "footnote", "record"
  ))

  if(kind == "test"){
    kind <- "evidence"
  }
  if(kind == "test_estimate"){
    kind <- "pair"
  }

  return(kind)
}

.interpret_expand_for_each <- function(item, sources, missing, index){

  source_name <- item[["source"]]
  if(is.null(source_name) && !is.null(item[["evidence"]])){
    source_name <- item[["evidence"]][["source"]]
  }
  if(is.null(source_name)){
    stop("A 'for_each' plan item must define 'source' or 'evidence$source'.", call. = FALSE)
  }

  source <- .interpret_get_source(sources, source_name, item, missing)
  if(is.null(source)){
    return(list())
  }

  rows <- item[["rows"]]
  if(is.null(rows) || identical(rows, "source_order")){
    rows <- .interpret_source_row_labels(source)
  }
  if(!is.null(item[["include"]])){
    rows <- rows[rows %in% item[["include"]]]
  }
  if(!is.null(item[["exclude"]])){
    rows <- rows[!rows %in% item[["exclude"]]]
  }
  if(!is.null(item[["row_regex"]])){
    rows <- rows[grepl(item[["row_regex"]], rows)]
  }

  template <- item[["template"]]
  if(is.null(template)){
    template <- if(!is.null(item[["pair_with"]]) || !is.null(item[["estimate"]])) "pair" else "evidence"
  }
  if(template == "test_estimate"){
    template <- "pair"
  }

  expanded <- list()
  base_order <- item[["order"]]
  if(is.null(base_order)){
    base_order <- index
  }

  for(j in seq_along(rows)){
    row <- rows[[j]]
    child <- item
    child[["kind"]] <- template
    child[["rows"]] <- NULL
    child[["include"]] <- NULL
    child[["exclude"]] <- NULL
    child[["row_regex"]] <- NULL
    child[["template"]] <- NULL
    child[["pair_with"]] <- NULL
    child[["row"]] <- row
    child[["order"]] <- base_order + j
    child[["item_id"]] <- .interpret_item_id(item, row)

    if(template == "pair"){
      evidence <- item[["evidence"]]
      if(is.null(evidence)){
        evidence <- list()
      }
      evidence[["source"]] <- .interpret_or(evidence[["source"]], source_name)
      evidence[["row"]]    <- .interpret_or(evidence[["row"]], row)

      estimate <- item[["estimate"]]
      if(is.null(estimate)){
        estimate <- list()
      }
      estimate[["source"]] <- .interpret_or(estimate[["source"]], item[["pair_with"]])
      estimate[["row"]]    <- .interpret_or(estimate[["row"]], row)

      child[["evidence"]] <- evidence
      child[["estimate"]] <- estimate
    }else{
      child[["source"]] <- source_name
    }

    expanded[[length(expanded) + 1L]] <- child
  }

  return(expanded)
}

.interpret_item_id <- function(item, row){

  prefix <- .interpret_or(item[["item_id"]], item[["id"]])
  row_id <- gsub("[^A-Za-z0-9_]+", "_", as.character(row))
  row_id <- gsub("^_+|_+$", "", row_id)
  if(is.null(prefix)){
    return(row_id)
  }

  paste(prefix, row_id, sep = ".")
}

.interpret_plan_item_records <- function(item, sources, missing, default_order){

  kind <- .interpret_plan_kind(item)
  item[["kind"]] <- kind
  if(is.null(item[["order"]])){
    item[["order"]] <- default_order
  }

  if(kind %in% c("header", "note", "prior", "footnote")){
    return(list(.interpret_complete_record(
      record = list(kind = kind, text = item[["text"]]),
      item   = item
    )))
  }

  if(kind == "record"){
    return(list(.interpret_complete_record(item, item)))
  }

  if(kind == "pair"){
    records <- list()
    if(!is.null(item[["evidence"]])){
      evidence <- .interpret_ref_record(
        ref     = .interpret_merge_ref(item[["evidence"]], item, "evidence", 0),
        kind    = "evidence",
        sources = sources,
        missing = missing
      )
      if(!is.null(evidence)){
        records[[length(records) + 1L]] <- evidence
      }
    }
    if(!is.null(item[["estimate"]])){
      estimate <- .interpret_ref_record(
        ref     = .interpret_merge_ref(item[["estimate"]], item, "estimate", 0.1),
        kind    = "estimate",
        sources = sources,
        missing = missing
      )
      if(!is.null(estimate)){
        records[[length(records) + 1L]] <- estimate
      }
    }
    return(records)
  }

  record <- .interpret_ref_record(
    ref     = .interpret_merge_ref(item, item, kind, 0),
    kind    = kind,
    sources = sources,
    missing = missing
  )
  if(is.null(record)){
    return(list())
  }

  return(list(record))
}

.interpret_merge_ref <- function(ref, item, kind, order_offset){

  if(is.null(ref)){
    ref <- list()
  }
  inherited <- c(
    "section", "item_id", "id", "label", "source", "row",
    "optional", "record_id"
  )
  for(field in inherited){
    if(is.null(ref[[field]]) && !is.null(item[[field]])){
      ref[[field]] <- item[[field]]
    }
  }

  ref[["kind"]] <- kind
  ref[["order"]] <- .interpret_or(ref[["order"]], item[["order"]] + order_offset)
  ref[["item_id"]] <- .interpret_or(ref[["item_id"]], ref[["id"]])

  return(ref)
}

.interpret_ref_record <- function(ref, kind, sources, missing){

  source_name <- ref[["source"]]
  if(is.null(source_name)){
    return(.interpret_handle_missing("Interpretation plan item is missing a source.", ref, missing))
  }

  source <- .interpret_get_source(sources, source_name, ref, missing)
  if(is.null(source)){
    return(NULL)
  }

  if(source[["type"]] == "table"){
    return(switch(
      kind,
      evidence = .interpret_table_evidence_record(source, ref, missing),
      estimate = .interpret_table_estimate_record(source, ref, missing),
      stop(paste0("Table sources cannot produce records of kind '", kind, "'."), call. = FALSE)
    ))
  }

  record <- .interpret_direct_source_record(source, ref, kind, missing)
  if(is.null(record)){
    return(NULL)
  }

  return(.interpret_complete_record(record, ref, source_name, ref[["row"]]))
}

.interpret_get_source <- function(sources, source_name, item, missing){

  if(!source_name %in% names(sources)){
    return(.interpret_handle_missing(paste0("Interpretation source '", source_name, "' was not found."), item, missing))
  }

  sources[[source_name]]
}

.interpret_direct_source_record <- function(source, ref, kind, missing){

  data <- source[["data"]]

  if(source[["type"]] == "record"){
    record <- data
  }else if(source[["type"]] == "records"){
    record <- .interpret_select_direct_record(data, ref, missing)
  }else{
    stop("Internal error: unsupported direct interpretation source.", call. = FALSE)
  }
  if(is.null(record)){
    return(NULL)
  }

  record <- .interpret_merge_lists(source[["schema"]], record)
  record <- .interpret_merge_lists(record, ref)
  record[["kind"]] <- .interpret_or(record[["kind"]], kind)

  if(record[["kind"]] == "evidence"){
    record <- .interpret_complete_BF_record(record)
  }

  return(record)
}

.interpret_select_direct_record <- function(data, ref, missing){

  row <- ref[["row"]]
  if(inherits(data, "data.frame")){
    if(is.null(row)){
      if(nrow(data) == 1L){
        return(as.list(data[1, , drop = FALSE]))
      }
      return(.interpret_handle_missing("Direct records source has multiple rows; specify 'row'.", ref, missing))
    }
    row_index <- .interpret_match_record_row(data, row)
    if(is.na(row_index)){
      return(.interpret_handle_missing(paste0("Record row '", row, "' was not found."), ref, missing))
    }
    return(as.list(data[row_index, , drop = FALSE]))
  }

  if(is.null(row)){
    if(length(data) == 1L && is.list(data[[1L]])){
      return(data[[1L]])
    }
    return(.interpret_handle_missing("Direct records source has multiple records; specify 'row'.", ref, missing))
  }

  if(!is.null(names(data)) && row %in% names(data)){
    return(data[[row]])
  }

  record_ids <- vapply(data, function(x) as.character(.interpret_or(x[["record_id"]], "")), character(1))
  row_index <- match(row, record_ids)
  if(is.na(row_index)){
    return(.interpret_handle_missing(paste0("Record row '", row, "' was not found."), ref, missing))
  }

  data[[row_index]]
}

.interpret_match_record_row <- function(data, row){

  row <- as.character(row)
  candidates <- list(rownames(data))
  if("record_id" %in% colnames(data)){
    candidates[[length(candidates) + 1L]] <- as.character(data[["record_id"]])
  }
  if("row" %in% colnames(data)){
    candidates[[length(candidates) + 1L]] <- as.character(data[["row"]])
  }

  for(candidate in candidates){
    row_index <- match(row, candidate)
    if(!is.na(row_index)){
      return(row_index)
    }
  }

  NA_integer_
}

.interpret_table_evidence_record <- function(source, ref, missing){

  data <- source[["data"]]
  schema <- source[["schema"]]
  row_index <- .interpret_table_row_index(source, ref[["row"]], ref, missing)
  if(is.null(row_index)){
    return(NULL)
  }
  row_label <- .interpret_source_row_labels(source)[row_index]

  BF_column <- .interpret_table_column(
    data       = data,
    schema     = schema,
    ref        = ref,
    field      = "BF",
    candidates = .interpret_BF_candidate_columns(data)
  )
  if(is.null(BF_column)){
    return(.interpret_handle_missing("Could not identify a Bayes-factor column.", ref, missing))
  }

  BF_vector <- data[[BF_column]]
  BF_value  <- as.numeric(BF_vector[row_index])

  BF_scale <- .interpret_table_scalar("BF_scale", data, row_index, schema, ref)
  if(is.null(BF_scale)){
    BF_scale <- if(isTRUE(attr(BF_vector, "logBF"))) "log" else "linear"
  }

  BF_orientation <- .interpret_table_scalar("BF_orientation", data, row_index, schema, ref)
  if(is.null(BF_orientation)){
    BF_orientation <- .interpret_infer_BF_orientation(data, BF_column, BF_vector)
  }

  BF_name <- .interpret_table_scalar("BF_name", data, row_index, schema, ref)
  if(is.null(BF_name)){
    BF_name <- .interpret_or(attr(BF_vector, "name"), BF_column)
  }

  BF_bound_operator <- .interpret_table_scalar("BF_bound_operator", data, row_index, schema, ref)
  if(is.null(BF_bound_operator)){
    BF_bound_operator <- attr(BF_vector[row_index], "bound_operator")
  }
  if(is.null(BF_bound_operator)){
    BF_bound_operator <- attr(BF_vector, "bound_operator")
    if(!is.null(BF_bound_operator)){
      BF_bound_operator <- BF_bound_operator[row_index]
    }
  }
  BF_bound_operator <- .standardize_BF_bound_operator(BF_bound_operator, 1)

  record <- list(
    kind              = "evidence",
    source            = source[["name"]],
    row               = row_label,
    label             = .interpret_or(.interpret_table_scalar("label", data, row_index, schema, ref), row_label),
    BF_value          = BF_value,
    BF_name           = BF_name,
    BF_scale          = BF_scale,
    BF_orientation    = BF_orientation,
    BF_bound_operator = BF_bound_operator,
    BF_target         = .interpret_table_scalar("BF_target", data, row_index, schema, ref),
    BF_complement     = .interpret_table_scalar("BF_complement", data, row_index, schema, ref)
  )
  record <- .interpret_merge_lists(record, ref)

  return(.interpret_complete_record(.interpret_complete_BF_record(record), ref, source[["name"]], row_label))
}

.interpret_table_estimate_record <- function(source, ref, missing){

  data <- source[["data"]]
  schema <- source[["schema"]]
  row_index <- .interpret_table_row_index(source, ref[["row"]], ref, missing)
  if(is.null(row_index)){
    return(NULL)
  }
  row_label <- .interpret_source_row_labels(source)[row_index]

  central_column <- .interpret_table_column(
    data       = data,
    schema     = schema,
    ref        = ref,
    field      = "central",
    candidates = c("Mean", "Median", "Mode", "mean", "median", "mode")
  )
  if(is.null(central_column)){
    return(.interpret_handle_missing("Could not identify a central estimate column.", ref, missing))
  }

  interval_columns <- .interpret_interval_columns(data, schema, ref, central_column)

  central_name <- .interpret_table_scalar("central_name", data, row_index, schema, ref)
  if(is.null(central_name)){
    central_name <- tolower(central_column)
  }

  record <- list(
    kind           = "estimate",
    source         = source[["name"]],
    row            = row_label,
    label          = .interpret_or(.interpret_table_scalar("label", data, row_index, schema, ref), row_label),
    parameter      = .interpret_or(.interpret_table_scalar("parameter", data, row_index, schema, ref), row_label),
    central_name   = central_name,
    central_value  = as.numeric(data[[central_column]][row_index]),
    lower_value    = if(is.null(interval_columns[["lower"]])) NA_real_ else as.numeric(data[[interval_columns[["lower"]]]][row_index]),
    upper_value    = if(is.null(interval_columns[["upper"]])) NA_real_ else as.numeric(data[[interval_columns[["upper"]]]][row_index]),
    lower_prob     = interval_columns[["lower_prob"]],
    upper_prob     = interval_columns[["upper_prob"]],
    interval_level = interval_columns[["interval_level"]],
    units          = .interpret_table_scalar("units", data, row_index, schema, ref),
    conditioning   = .interpret_table_scalar("conditioning", data, row_index, schema, ref)
  )
  record <- .interpret_merge_lists(record, ref)

  return(.interpret_complete_record(record, ref, source[["name"]], row_label))
}

.interpret_table_row_index <- function(source, row, item, missing){

  data <- source[["data"]]
  if(is.null(row)){
    if(nrow(data) == 1L){
      return(1L)
    }
    return(.interpret_handle_missing("Table source has multiple rows; specify 'row'.", item, missing))
  }

  if(is.numeric(row)){
    if(length(row) != 1L || is.na(row) || row < 1 || row > nrow(data)){
      return(.interpret_handle_missing(paste0("Table row index '", row, "' is out of range."), item, missing))
    }
    return(as.integer(row))
  }

  row <- as.character(row)
  candidates <- list(
    .interpret_source_row_labels(source),
    attr(data, "parameters")
  )

  schema_row <- source[["schema"]][["row"]]
  if(!is.null(schema_row) && schema_row %in% colnames(data)){
    candidates[[length(candidates) + 1L]] <- as.character(data[[schema_row]])
  }

  for(candidate in candidates){
    if(is.null(candidate)){
      next
    }
    row_index <- match(row, as.character(candidate))
    if(!is.na(row_index)){
      return(row_index)
    }
  }

  .interpret_handle_missing(paste0("Table row '", row, "' was not found."), item, missing)
}

.interpret_source_row_labels <- function(source){

  data <- source[["data"]]
  if(!inherits(data, "data.frame")){
    if(is.list(data) && !is.null(names(data))){
      return(names(data))
    }
    return(as.character(seq_along(data)))
  }

  schema_row <- source[["schema"]][["row"]]
  if(!is.null(schema_row) && schema_row %in% colnames(data)){
    return(as.character(data[[schema_row]]))
  }

  rows <- rownames(data)
  if(is.null(rows)){
    rows <- as.character(seq_len(nrow(data)))
  }

  rows
}

.interpret_table_column <- function(data, schema, ref, field, candidates){

  value <- .interpret_or(ref[[field]], schema[[field]])
  value <- .interpret_or(value, ref[[paste0(field, "_column")]])
  value <- .interpret_or(value, schema[[paste0(field, "_column")]])

  if(!is.null(value) && length(value) == 1L && value %in% colnames(data)){
    return(value)
  }

  for(candidate in candidates){
    if(candidate %in% colnames(data)){
      return(candidate)
    }
  }

  NULL
}

.interpret_table_scalar <- function(field, data, row_index, schema, ref){

  value <- .interpret_or(ref[[field]], schema[[field]])
  if(is.null(value)){
    return(NULL)
  }
  if(length(value) == 1L && is.character(value) && value %in% colnames(data)){
    return(data[[value]][row_index])
  }
  if(length(value) > 1L && length(value) >= row_index){
    return(value[row_index])
  }

  value
}

.interpret_BF_candidate_columns <- function(data){

  type <- attr(data, "type")
  if(!is.null(type)){
    type_match <- which(type %in% c("inclusion_BF", "BF"))
    if(length(type_match) > 0){
      return(colnames(data)[type_match])
    }
  }

  c("inclusion_BF", "BF", "Bayes factor", "bayes_factor")
}

.interpret_infer_BF_orientation <- function(data, BF_column, BF_vector){

  type <- attr(data, "type")
  BF_type <- NULL
  if(!is.null(type) && BF_column %in% colnames(data)){
    BF_type <- type[match(BF_column, colnames(data))]
  }
  BF01 <- isTRUE(attr(BF_vector, "BF01"))

  if(identical(BF_type, "inclusion_BF") || identical(BF_column, "inclusion_BF")){
    return(if(BF01) "exclusion_over_inclusion" else "inclusion_over_exclusion")
  }

  if(BF01){
    "null_over_alternative"
  }else{
    "alternative_over_null"
  }
}

.interpret_interval_columns <- function(data, schema, ref, central_column){

  lower <- .interpret_or(ref[["lower"]], schema[["lower"]])
  upper <- .interpret_or(ref[["upper"]], schema[["upper"]])
  lower_prob <- .interpret_or(ref[["lower_prob"]], schema[["lower_prob"]])
  upper_prob <- .interpret_or(ref[["upper_prob"]], schema[["upper_prob"]])

  if(is.null(lower) && !is.null(lower_prob)){
    lower <- .interpret_probability_column(data, lower_prob)
  }
  if(is.null(upper) && !is.null(upper_prob)){
    upper <- .interpret_probability_column(data, upper_prob)
  }

  if(is.null(lower) || is.null(upper)){
    probability_columns <- .interpret_probability_columns(data)
    probability_columns <- setdiff(probability_columns, central_column)
    if(length(probability_columns) >= 2L){
      probs <- as.numeric(probability_columns)
      lower <- probability_columns[which.min(probs)]
      upper <- probability_columns[which.max(probs)]
      lower_prob <- .interpret_or(lower_prob, as.numeric(lower))
      upper_prob <- .interpret_or(upper_prob, as.numeric(upper))
    }
  }

  if(!is.null(lower) && !lower %in% colnames(data)){
    lower <- NULL
  }
  if(!is.null(upper) && !upper %in% colnames(data)){
    upper <- NULL
  }

  if(is.null(lower_prob) && !is.null(lower)){
    lower_prob <- suppressWarnings(as.numeric(lower))
  }
  if(is.null(upper_prob) && !is.null(upper)){
    upper_prob <- suppressWarnings(as.numeric(upper))
  }

  interval_level <- .interpret_or(ref[["interval_level"]], schema[["interval_level"]])
  if(is.null(interval_level) && is.finite(lower_prob) && is.finite(upper_prob)){
    interval_level <- upper_prob - lower_prob
  }

  list(
    lower          = lower,
    upper          = upper,
    lower_prob     = if(is.null(lower_prob)) NA_real_ else as.numeric(lower_prob),
    upper_prob     = if(is.null(upper_prob)) NA_real_ else as.numeric(upper_prob),
    interval_level = if(is.null(interval_level)) NA_real_ else as.numeric(interval_level)
  )
}

.interpret_probability_column <- function(data, prob){

  prob <- as.numeric(prob)
  candidates <- c(as.character(prob), format(prob, trim = TRUE, scientific = FALSE))
  candidates <- unique(candidates)
  match <- candidates[candidates %in% colnames(data)]
  if(length(match) > 0){
    return(match[1])
  }

  NULL
}

.interpret_probability_columns <- function(data){

  probs <- suppressWarnings(as.numeric(colnames(data)))
  colnames(data)[!is.na(probs) & probs > 0 & probs < 1]
}

.interpret_complete_BF_record <- function(record){

  BF_value <- as.numeric(record[["BF_value"]])
  BF_scale <- .interpret_or(record[["BF_scale"]], "linear")
  BF_orientation <- .interpret_or(record[["BF_orientation"]], "target_over_complement")
  BF_bound_operator <- .standardize_BF_bound_operator(record[["BF_bound_operator"]], 1)

  check_char(BF_scale, "BF_scale", allow_values = c("linear", "log"))
  check_char(BF_orientation, "BF_orientation", allow_values = c(
    "inclusion_over_exclusion", "exclusion_over_inclusion",
    "alternative_over_null", "null_over_alternative",
    "target_over_complement", "complement_over_target"
  ))

  BF_linear <- if(BF_scale == "log") exp(BF_value) else BF_value
  canonical_bound <- BF_bound_operator
  if(BF_orientation %in% c("exclusion_over_inclusion", "null_over_alternative", "complement_over_target")){
    BF_linear <- 1 / BF_linear
    canonical_bound <- .invert_BF_bound_operator(canonical_bound)
  }

  record[["BF_value"]] <- BF_value
  record[["BF_scale"]] <- BF_scale
  record[["BF_orientation"]] <- BF_orientation
  record[["BF_bound_operator"]] <- BF_bound_operator
  record[["BF_canonical_value"]] <- BF_linear
  record[["BF_canonical_bound_operator"]] <- canonical_bound
  record[["BF_canonical_orientation"]] <- "target_over_complement"

  record
}

.interpret_complete_record <- function(record, item, source_name = NULL, row = NULL){

  empty <- .interpret_empty_record()
  record <- .interpret_merge_lists(empty, record)

  record[["section"]] <- .interpret_fill(record[["section"]], item[["section"]])
  record[["item_id"]] <- .interpret_fill(.interpret_fill(record[["item_id"]], item[["item_id"]]), item[["id"]])
  record[["source"]]  <- .interpret_fill(record[["source"]], source_name)
  record[["row"]]     <- .interpret_fill(record[["row"]], row)
  record[["order"]]   <- .interpret_fill(record[["order"]], item[["order"]])
  record[["record_id"]] <- .interpret_fill(record[["record_id"]], .interpret_record_id(record))

  record
}

.interpret_empty_record <- function(){

  list(
    record_id = NA_character_,
    kind = NA_character_,
    section = NA_character_,
    item_id = NA_character_,
    source = NA_character_,
    row = NA_character_,
    order = NA_real_,
    label = NA_character_,
    text = NA_character_,
    parameter = NA_character_,
    units = NA_character_,
    conditioning = NA_character_,
    central_name = NA_character_,
    central_value = NA_real_,
    lower_value = NA_real_,
    upper_value = NA_real_,
    lower_prob = NA_real_,
    upper_prob = NA_real_,
    interval_level = NA_real_,
    BF_value = NA_real_,
    BF_name = NA_character_,
    BF_scale = NA_character_,
    BF_orientation = NA_character_,
    BF_bound_operator = NA_character_,
    BF_canonical_value = NA_real_,
    BF_canonical_bound_operator = NA_character_,
    BF_canonical_orientation = NA_character_,
    BF_target = NA_character_,
    BF_complement = NA_character_
  )
}

.interpret_record_id <- function(record){

  pieces <- c(record[["section"]], record[["item_id"]], record[["kind"]])
  pieces <- pieces[!is.na(pieces) & nzchar(pieces)]
  if(length(pieces) == 0){
    return(NA_character_)
  }

  paste(pieces, collapse = ".")
}

.interpret_bind_records <- function(records){

  columns <- names(.interpret_empty_record())
  if(length(records) == 0){
    out <- as.data.frame(.interpret_empty_record(), stringsAsFactors = FALSE)[0, , drop = FALSE]
    class(out) <- c("BayesTools_interpret_records", class(out))
    return(out)
  }

  rows <- lapply(records, function(record){
    record <- .interpret_merge_lists(.interpret_empty_record(), record)
    record <- record[columns]
    as.data.frame(record, stringsAsFactors = FALSE)
  })

  out <- do.call(rbind, rows)
  rownames(out) <- NULL
  class(out) <- c("BayesTools_interpret_records", class(out))

  out
}

.interpret_records_text <- function(records, method = NULL, digits = 3){

  if(nrow(records) == 0){
    return(character())
  }

  out <- character()
  for(i in seq_len(nrow(records))){
    record <- records[i, , drop = FALSE]
    kind <- record[["kind"]]
    text <- switch(
      kind,
      header = record[["text"]],
      note = record[["text"]],
      prior = record[["text"]],
      footnote = record[["text"]],
      evidence = .interpret_record_evidence_text(record),
      estimate = .interpret_record_estimate_text(record, digits),
      NA_character_
    )
    if(!is.na(text) && nzchar(text)){
      out <- c(out, text)
    }
  }

  if(!is.null(method) && length(out) > 0 && !any(records[["kind"]] == "header")){
    out[1] <- paste(method, out[1])
  }

  class(out) <- c("BayesTools_interpret_text", "character")
  out
}

.interpret_record_evidence_text <- function(record){

  BF <- record[["BF_canonical_value"]]
  if(!is.finite(BF)){
    return(paste0("evidence for ", record[["label"]], " unavailable"))
  }
  BF_name <- .interpret_canonical_BF_name(record)
  bound <- record[["BF_canonical_bound_operator"]]
  if(is.na(bound)){
    bound <- NULL
  }

  .interpret.BF(BF, record[["label"]], BF_name, bound)
}

.interpret_canonical_BF_name <- function(record){

  if(record[["BF_orientation"]] %in% c("inclusion_over_exclusion", "exclusion_over_inclusion")){
    return("Inclusion BF")
  }
  if(record[["BF_orientation"]] %in% c("alternative_over_null", "null_over_alternative")){
    return("BF")
  }

  BF_name <- record[["BF_name"]]
  BF_scale <- record[["BF_scale"]]
  if(is.na(BF_name) || !nzchar(BF_name) || is.na(BF_scale) || BF_scale != "linear"){
    return("BF")
  }

  BF_name
}

.interpret_record_estimate_text <- function(record, digits){

  central <- format(round(record[["central_value"]], digits), nsmall = digits)
  lower   <- format(round(record[["lower_value"]], digits), nsmall = digits)
  upper   <- format(round(record[["upper_value"]], digits), nsmall = digits)
  level   <- if(is.finite(record[["interval_level"]])) paste0(round(100 * record[["interval_level"]]), "%") else "uncertainty"

  text <- paste0("with ", record[["central_name"]], " estimate ", record[["parameter"]], " = ", central)
  if(!is.na(record[["units"]]) && nzchar(record[["units"]])){
    text <- paste(text, record[["units"]])
  }
  if(is.finite(record[["lower_value"]]) && is.finite(record[["upper_value"]])){
    text <- paste0(text, ", ", level, " interval [", lower, ", ", upper, "]")
  }
  if(!is.na(record[["conditioning"]]) && nzchar(record[["conditioning"]])){
    text <- paste0(text, " (", record[["conditioning"]], ")")
  }

  text
}

.interpret_merge_lists <- function(x, y){

  if(is.null(y)){
    return(x)
  }
  for(name in names(y)){
    x[[name]] <- y[[name]]
  }
  x
}

.interpret_or <- function(x, y){
  if(is.null(x)) y else x
}

.interpret_fill <- function(x, y){
  if(is.null(x) || length(x) == 0 || (length(x) == 1L && is.na(x))) y else x
}

.interpret_handle_missing <- function(message, item, missing){

  optional <- isTRUE(item[["optional"]])
  if(optional || missing == "skip"){
    return(NULL)
  }
  if(missing == "warn"){
    warning(message, call. = FALSE)
    return(NULL)
  }

  stop(message, call. = FALSE)
}

##' BayesTools
##'
##' BayesTools: Provides tools for conducting Bayesian analyses. The package contains
##' functions for creating a wide range of prior distribution objects, mixing posterior
##' samples from JAGS and Stan models, plotting posterior distributions, and etc...
##' The tools for working with prior distribution span from visualization, generating JAGS
##' and bridgesampling syntax to basic functions such as rng, quantile, and distribution functions.
##'
##'
##' @name BayesTools
##' @aliases BayesTools-package BayesTools
##' @docType package
##' @author František Bartoš \email{f.bartos96@@gmail.com}
##' @keywords package
##' @importFrom rlang .data
##' @importFrom Rdpack reprompt
##' @importFrom stats fft median setNames
##' @importFrom utils head tail
"_PACKAGE"

#' @title Weight functions
#'
#' @description Marginal density, marginal distribution function,
#' marginal quantile function and random generation for weight functions.
#'
#' @param x,q vector or matrix of quantiles.
#' @param p vector of probabilities.
#' @param n number of observations.
#' @param alpha vector or matrix with concentration parameters for
#' the Dirichlet distribution for a monotonic one.sided or a two.sided
#' weight function.
#' @param alpha1 vector or matrix with concentration parameters for
#' the Dirichlet distribution for the expected direction of non-monotonic
#' one.sided of weight function.
#' @param alpha2 vector or matrix with concentration parameters for
#' the Dirichlet distribution for the unexpected direction of non-monotonic
#' one.sided of weight function.
#' @param omega vector or matrix of fixed non-negative relative weights for a
#' one.sided or a two.sided weight function. The first/reference weight must be
#' exactly 1.
#' @param log,log.p logical; if \code{TRUE}, probabilities
#' \code{p} are given as \code{log(p)}.
#' @param lower.tail logical; if \code{TRUE} (default), probabilities
#' are \eqn{P[X \le x]}, otherwise, \eqn{P[X > x]}.
#'
#'
#' @examples
#' # draw samples from a two-sided weight function
#' rtwo.sided(10, alpha = c(1, 1))
#'
#' # draw samples from a monotone one-sided weight function
#' rone.sided(10, alpha = c(1, 1, 1))
#'
#' # draw samples from a non-monotone one-sided weight function
#' rone.sided(10, alpha1 = c(1, 1), alpha2 = c(1, 1))
#'
#' @return \code{mdone.sided}, \code{mdtwo.sided}, \code{mdone.sided_fixed},
#' and \code{mdtwo.sided_fixed} give the marginal density,
#' \code{mpone.sided}, \code{mptwo.sided}, \code{mpone.sided_fixed},
#' and \code{mptwo.sided_fixed} give the marginal distribution function,
#' \code{mqone.sided}, \code{mqtwo.sided}, \code{mqone.sided_fixed},
#' and \code{mqtwo.sided_fixed} give the marginal quantile function,
#' and \code{rone.sided}, \code{rtwo.sided}, \code{rone.sided_fixed},
#' and \code{rtwo.sided_fixed} generate random deviates.
#'
#' @export mdone.sided
#' @export mdtwo.sided
#' @export mdone.sided_fixed
#' @export mdtwo.sided_fixed
#' @export rone.sided
#' @export rtwo.sided
#' @export rone.sided_fixed
#' @export rtwo.sided_fixed
#' @export mpone.sided
#' @export mptwo.sided
#' @export mpone.sided_fixed
#' @export mptwo.sided_fixed
#' @export mqone.sided
#' @export mqtwo.sided
#' @export mqone.sided_fixed
#' @export mqtwo.sided_fixed
#' @name weightfunctions
NULL

#### wrappers ####
#' @rdname weightfunctions
mdone.sided <- function(x, alpha = NULL, alpha1 = NULL, alpha2 = NULL, log = FALSE){

  parameterization <- .weightfunctions_one_sided_parameterization(alpha, alpha1, alpha2)

  # common input check
  .check_log(log)
  .check_x(x, lower = 0, upper = 1)

  if(parameterization == "monotonic"){
    lik <- .mdone.sided_monotonic(x, alpha, log)
  }else{
    lik <- .mdone.sided_general(x, alpha1, alpha2, log)
  }

  return(lik)
}
#' @rdname weightfunctions
mdtwo.sided <- function(x, alpha, log = FALSE){

  # common input check
  .check_log(log)
  .check_x(x, lower = 0, upper = 1)

  lik <- mdone.sided(x, alpha = alpha, log = log)

  return(lik)
}
#' @rdname weightfunctions
mdone.sided_fixed <- function(x, omega, log = FALSE){

  # common input check
  .check_log(log)
  .check_x(x, lower = 0)

  lik <- .mdone.sided_fixed(x, omega, log)

  return(lik)
}
#' @rdname weightfunctions
mdtwo.sided_fixed <- function(x, omega, log = FALSE){

  # common input check
  .check_log(log)
  .check_x(x, lower = 0)

  lik <- mdone.sided_fixed(x, omega = omega, log = log)

  return(lik)
}

#' @rdname weightfunctions
rone.sided <- function(n, alpha = NULL, alpha1 = NULL, alpha2 = NULL){

  parameterization <- .weightfunctions_one_sided_parameterization(alpha, alpha1, alpha2)

  # common input check
  .check_n(n)

  if(parameterization == "monotonic"){
    x <- .rone.sided_monotonic(n, alpha)
  }else{
    x <- .rone.sided_general(n, alpha1, alpha2)
  }

  return(x)
}
#' @rdname weightfunctions
rtwo.sided <- function(n, alpha){

  # common input check
  .check_n(n)

  x <- rone.sided(n, alpha = alpha)

  return(x)
}
#' @rdname weightfunctions
rone.sided_fixed <- function(n, omega){

  # common input check
  .check_n(n)

  x <- .rone.sided_fixed(n, omega)

  return(x)
}
#' @rdname weightfunctions
rtwo.sided_fixed <- function(n, omega){

  # common input check
  .check_n(n)

  x <- rone.sided_fixed(n, omega = omega)

  return(x)
}

#' @rdname weightfunctions
mpone.sided <- function(q, alpha = NULL, alpha1 = NULL, alpha2 = NULL, lower.tail = TRUE, log.p = FALSE){

  parameterization <- .weightfunctions_one_sided_parameterization(alpha, alpha1, alpha2)

  # common input check
  .check_log.p(log.p)
  .check_lower.tail(lower.tail)
  .check_q(q, lower = 0, upper = 1)

  if(parameterization == "monotonic"){
    p <- .mpone.sided_monotonic(q, alpha, lower.tail, log.p)
  }else{
    p <- .mpone.sided_general(q, alpha1, alpha2, lower.tail, log.p)
  }

  return(p)
}
#' @rdname weightfunctions
mptwo.sided <- function(q, alpha, lower.tail = TRUE, log.p = FALSE){

  # common input check
  .check_log.p(log.p)
  .check_lower.tail(lower.tail)
  .check_q(q, lower = 0, upper = 1)

  p <- mpone.sided(q, alpha = alpha, lower.tail = lower.tail, log.p = log.p)

  return(p)
}
#' @rdname weightfunctions
mpone.sided_fixed <- function(q, omega, lower.tail = TRUE, log.p = FALSE){

  # common input check
  .check_log.p(log.p)
  .check_lower.tail(lower.tail)
  .check_q(q, lower = 0)

  p <- .mpone.sided_fixed(q, omega, lower.tail, log.p)

  return(p)
}
#' @rdname weightfunctions
mptwo.sided_fixed <- function(q, omega, lower.tail = TRUE, log.p = FALSE){

  # common input check
  .check_log.p(log.p)
  .check_lower.tail(lower.tail)
  .check_q(q, lower = 0)

  p <- mpone.sided_fixed(q, omega = omega, lower.tail = lower.tail, log.p = log.p)

  return(p)
}

#' @rdname weightfunctions
mqone.sided <- function(p, alpha = NULL, alpha1 = NULL, alpha2 = NULL, lower.tail = TRUE, log.p = FALSE){

  parameterization <- .weightfunctions_one_sided_parameterization(alpha, alpha1, alpha2)

  # common input check
  .check_log.p(log.p)
  .check_lower.tail(lower.tail)
  .check_p(p, log.p)

  if(parameterization == "monotonic"){
    q <- .mqone.sided_monotonic(p, alpha, lower.tail, log.p)
  }else{
    q <- .mqone.sided_general(p, alpha1, alpha2, lower.tail, log.p)
  }

  return(q)
}
#' @rdname weightfunctions
mqtwo.sided <- function(p, alpha, lower.tail = TRUE, log.p = FALSE){

  # common input check
  .check_log.p(log.p)
  .check_lower.tail(lower.tail)
  .check_p(p, log.p)

  q <- mqone.sided(p, alpha = alpha, lower.tail = lower.tail, log.p = log.p)

  return(q)
}
#' @rdname weightfunctions
mqone.sided_fixed <- function(p, omega, lower.tail = TRUE, log.p = FALSE){

  # common input check
  .check_log.p(log.p)
  .check_lower.tail(lower.tail)
  .check_p(p, log.p)

  q <- .mqone.sided_fixed(p, omega, lower.tail, log.p)

  return(q)
}
#' @rdname weightfunctions
mqtwo.sided_fixed <- function(p, omega, lower.tail = TRUE, log.p = FALSE){

  # common input check
  .check_log.p(log.p)
  .check_lower.tail(lower.tail)
  .check_p(p, log.p)

  q <- mqone.sided_fixed(p, omega = omega, lower.tail = lower.tail, log.p = log.p)

  return(q)
}


###### helper functions
#### density functions ####
.mdone.sided_general   <- function(x, alpha1, alpha2, log){

  stop("Not implemented")
  # would require product of two beta-distributed variables, since the weights in the opposite direction
  # start at the first cutoff modeled by the first alpha parameter
  # https://math.stackexchange.com/questions/1073364/product-of-two-beta-distributed-random-variables
  # https://www.dm.fct.unl.pt/sites/www.dm.fct.unl.pt/files/preprints/2012/7_12.pdf
  # probably solvable with Meijer g-function

  # # input check
  # .weightfunctions_check_alpha(alpha1, "alpha1")
  # .weightfunctions_check_alpha(alpha2, "alpha2")
  #
  # # transform to matrices for easier manipulation and checks
  # if(length(x) == 1){
  #   x <- rep(x, nrow(alpha1))
  # }
  # if(!is.matrix(alpha1)){
  #   alpha1 <- matrix(alpha1, nrow = 1)
  # }
  # if(!is.matrix(alpha2)){
  #   alpha2 <- matrix(alpha2, nrow = 1)
  # }
  #
  # if(nrow(alpha1) != nrow(alpha2))
  #   stop("Non matching dimensions of 'alpha1' and 'alpha2'.")
  # if(nrow(alpha1) != length(x) & nrow(alpha1) != 1)
  #   stop("Non matching dimensions of 'alpha' and 'x'.")
  #
  # if(nrow(alpha1) != length(x) & nrow(alpha1) == 1){
  #   alpha1 <- do.call(rbind, lapply(1:length(x), function(i)alpha1))
  #   alpha2 <- do.call(rbind, lapply(1:length(x), function(i)alpha2))
  # }
  #
  #
  # lik1 <- .mdone.sided_monotonic(x, alpha = alpha1, log = log)
  # the side in the unexpected direction starts with the first step of alpha1, e.g., something like this:
  # lik2 <- .dmone.sided_monotonic(x, alpha = alpha2, log = log) * beta(alpha[,1], t(apply(alpha[,-1], 1, sum)))
  #
  # return(cbind(lik2, lik1))
}
.mdone.sided_monotonic <- function(x, alpha, log){

  # input check
  .weightfunctions_check_alpha(alpha, "alpha")

  # transform to matrices for easier manipulation and checks
  if(!is.matrix(alpha)){
    alpha <- matrix(alpha, nrow = 1)
  }
  if(length(x) == 1){
    x <- rep(x, nrow(alpha))
  }

  if(nrow(alpha) != length(x) & nrow(alpha) != 1)
    stop("Non matching dimensions of 'alpha' and 'x'.")

  if(nrow(alpha) != length(x) & nrow(alpha) == 1){
    alpha <- do.call(rbind, lapply(1:length(x), function(i)alpha))
  }

  # marginals of reversed cumulative sums of Dirichlet variables are beta
  # distributed; the first/reference publication weight is fixed to one.
  alpha_alpha <- t(apply(alpha[,ncol(alpha):1, drop = FALSE], 1, cumsum))[, ncol(alpha):1, drop = FALSE]
  alpha_beta  <- cbind(0, t(apply(alpha, 1, cumsum))[, -ncol(alpha), drop = FALSE])

  lik <- cbind(
    dpoint(x, location = 1, log = log),
    do.call(cbind, lapply(2:ncol(alpha), function(i){
      stats::dbeta(x, shape1 = alpha_alpha[,i], shape2 = alpha_beta[,i], log = log)
    }))
  )

  return(lik)
}
.mdone.sided_fixed     <- function(x, omega, log){

  .weightfunctions_check_omega(omega, "omega")

  # transform to matrices for easier manipulation and checks
  if(!is.matrix(omega)){
    omega <- matrix(omega, nrow = 1)
  }
  if(length(x) == 1){
    x <- rep(x, nrow(omega))
  }

  if(nrow(omega) != length(x) & nrow(omega) != 1)
    stop("Non matching dimensions of 'omega' and 'x'.")

  if(nrow(omega) != length(x) & nrow(omega) == 1){
    omega <- do.call(rbind, lapply(1:length(x), function(i)omega))
  }

  lik <- do.call(cbind, lapply(1:ncol(omega), function(i)dpoint(x, location = omega[,i], log = log)))
}

#### random number generators ####
.rone.sided_general   <- function(n, alpha1, alpha2){

  # input check
  .weightfunctions_check_alpha(alpha1, "alpha1")
  .weightfunctions_check_alpha(alpha2, "alpha2")

  # transform to matrices for easier manipulation and checks
  if(!is.matrix(alpha1)){
    alpha1 <- matrix(alpha1, nrow = 1)
  }
  if(!is.matrix(alpha2)){
    alpha2 <- matrix(alpha2, nrow = 1)
  }

  if(nrow(alpha1) != nrow(alpha2))
    stop("Non matching dimensions of 'alpha1' and 'alpha2'.")
  if(nrow(alpha1) != n & nrow(alpha1) != 1)
    stop("Incompatible dimensions of requested number of samples and 'alpha'.")

  if(nrow(alpha1) != n & nrow(alpha1) == 1){
    alpha1 <- do.call(rbind, lapply(1:n, function(i)alpha1))
    alpha2 <- do.call(rbind, lapply(1:n, function(i)alpha2))
  }

  x1 <- extraDistr::rdirichlet(n, alpha = alpha1)
  x2 <- extraDistr::rdirichlet(n, alpha = alpha2) * (1 - x1[,1])

  x <- matrix(ncol = ncol(alpha1) + ncol(alpha2) - 1, nrow = n)

  x[,ncol(alpha2):ncol(x)] <- t(apply(x1, 1, cumsum))
  for(i in 2:ncol(alpha2)){
    x[,i-1] = apply(x2[,i:ncol(alpha2), drop = FALSE], 1, sum) + x1[,1]
  }

  return(x[,ncol(x):1,drop = FALSE])
}
.rone.sided_monotonic <- function(n, alpha){

  # input check
  .weightfunctions_check_alpha(alpha, "alpha")

  # transform to matrices for easier manipulation and checks
  if(!is.matrix(alpha)){
    alpha <- matrix(alpha, nrow = 1)
  }

  if(nrow(alpha) != n & nrow(alpha) != 1)
    stop("Incompatible dimensions of requested number of samples and 'alpha'.")

  if(nrow(alpha) != n & nrow(alpha) == 1){
    alpha <- do.call(rbind, lapply(1:n, function(i)alpha))
  }

  x <- extraDistr::rdirichlet(n, alpha = alpha)
  x <- t(apply(x[,ncol(x):1, drop = FALSE], 1, cumsum))[,ncol(x):1, drop = FALSE]
  x[,1] <- 1

  return(x)
}
.rone.sided_fixed     <- function(n, omega){

  .weightfunctions_check_omega(omega, "omega")

  # transform to matrices for easier manipulation and checks
  if(!is.matrix(omega)){
    omega <- matrix(omega, nrow = 1)
  }

  if(nrow(omega) != n & nrow(omega) != 1)
    stop("Incompatible dimensions of requested number of samples and 'omega'.")

  if(nrow(omega) != n & nrow(omega) == 1){
    omega <- do.call(rbind, lapply(1:n, function(i)omega))
  }

  x <- omega

  return(x)
}

#### marginal distribution functions ####
.mpone.sided_general   <- function(q, alpha1, alpha2, lower.tail, log.p){

  stop("Not implemented")

  # # input check
  # .weightfunctions_check_alpha(alpha1, "alpha1")
  # .weightfunctions_check_alpha(alpha2, "alpha2")
  #
  #
  # # transform to matrices for easier manipulation and checks
  # if(!is.matrix(alpha1)){
  #   alpha1 <- matrix(alpha1, nrow = 1)
  # }
  # if(!is.matrix(alpha2)){
  #   alpha2 <- matrix(alpha2, nrow = 1)
  # }
  # if(length(q) == 1){
  #   q <- rep(q, nrow(alpha1))
  # }
  #
  # if(nrow(alpha1) != nrow(alpha2))
  #   stop("Non matching dimensions of 'alpha1' and 'alpha2'.")
  # if(nrow(alpha1) != length(q) & nrow(alpha1) != 1)
  #   stop("Non matching dimensions of 'alpha' and 'q'.")

}
.mpone.sided_monotonic <- function(q, alpha, lower.tail, log.p){

  # input check
  .weightfunctions_check_alpha(alpha, "alpha")

  # transform to matrices for easier manipulation and checks
  if(!is.matrix(alpha)){
    alpha <- matrix(alpha, nrow = 1)
  }
  if(length(q) == 1){
    q <- rep(q, nrow(alpha))
  }

  if(nrow(alpha) != length(q) & nrow(alpha) != 1)
    stop("Non matching dimensions of 'alpha' and 'q'.")

  if(nrow(alpha) != length(q) & nrow(alpha) == 1){
    alpha <- do.call(rbind, lapply(1:length(q), function(i)alpha))
  }

  # marginals of reversed cumulative sums of Dirichlet variables are beta
  # distributed; the first/reference publication weight is fixed to one.
  alpha_alpha <- t(apply(alpha[,ncol(alpha):1, drop = FALSE], 1, cumsum))[, ncol(alpha):1, drop = FALSE]
  alpha_beta  <- cbind(0, t(apply(alpha, 1, cumsum))[, -ncol(alpha), drop = FALSE])

  p <- cbind(
    ppoint(q, location = 1, lower.tail = lower.tail, log.p = log.p),
    do.call(cbind, lapply(2:ncol(alpha), function(i){
      stats::pbeta(q, shape1 = alpha_alpha[,i], shape2 = alpha_beta[,i], log.p = log.p, lower.tail = lower.tail)
    }))
  )

  return(p)
}
.mpone.sided_fixed     <- function(q, omega, lower.tail, log.p){

  .weightfunctions_check_omega(omega, "omega")

  # transform to matrices for easier manipulation and checks
  if(!is.matrix(omega)){
    omega <- matrix(omega, nrow = 1)
  }
  if(length(q) == 1){
    q <- rep(q, nrow(omega))
  }

  if(nrow(omega) != length(q) & nrow(omega) != 1)
    stop("Non matching dimensions of 'omega' and 'q'.")

  if(nrow(omega) != length(q) & nrow(omega) == 1){
    omega <- do.call(rbind, lapply(1:length(q), function(i)omega))
  }

  p <- do.call(cbind, lapply(1:ncol(omega), function(i)ppoint(q, location = omega[,i], lower.tail = lower.tail, log.p = log.p)))

  return(p)
}

#### marginal quantile functions ####
.mqone.sided_general   <- function(p, alpha1, alpha2, lower.tail, log.p){

  stop("Not implemented")

  # # input check
  # .weightfunctions_check_alpha(alpha1, "alpha1")
  # .weightfunctions_check_alpha(alpha2, "alpha2")
  #
  #
  # # transform to matrices for easier manipulation and checks
  # if(!is.matrix(alpha1)){
  #   alpha1 <- matrix(alpha1, nrow = 1)
  # }
  # if(!is.matrix(alpha2)){
  #   alpha2 <- matrix(alpha2, nrow = 1)
  # }
  # if(length(p) == 1){
  #   p <- rep(p, nrow(alpha1))
  # }
  #
  # if(nrow(alpha1) != nrow(alpha2))
  #   stop("Non matching dimensions of 'alpha1' and 'alpha2'.")
  # if(nrow(alpha1) != length(p) & nrow(alpha1) != 1)
  #   stop("Non matching dimensions of 'alpha' and 'p'.")
  #
  # if(nrow(alpha1) != length(p) & nrow(alpha1) == 1){
  #   alpha1 <- do.call(rbind, lapply(1:length(p), function(i)alpha1))
  #   alpha2 <- do.call(rbind, lapply(1:length(p), function(i)alpha2))
  # }

}
.mqone.sided_monotonic <- function(p, alpha, lower.tail, log.p){

  # input check
  .weightfunctions_check_alpha(alpha, "alpha")


  # transform to matrices for easier manipulation and checks
  if(!is.matrix(alpha)){
    alpha <- matrix(alpha, nrow = 1)
  }
  if(length(p) == 1){
    p <- rep(p, nrow(alpha))
  }

  if(nrow(alpha) != length(p) & nrow(alpha) != 1)
    stop("Non matching dimensions of 'alpha' and 'p'.")

  if(nrow(alpha) != length(p) & nrow(alpha) == 1){
    alpha <- do.call(rbind, lapply(1:length(p), function(i)alpha))
  }

  # marginals of reversed cumulative sums of Dirichlet variables are beta
  # distributed; the first/reference publication weight is fixed to one.
  alpha_alpha <- t(apply(alpha[,ncol(alpha):1, drop = FALSE], 1, cumsum))[, ncol(alpha):1, drop = FALSE]
  alpha_beta  <- cbind(0, t(apply(alpha, 1, cumsum))[, -ncol(alpha), drop = FALSE])

  q <- cbind(
    qpoint(p, location = 1, lower.tail = lower.tail, log.p = log.p),
    do.call(cbind, lapply(2:ncol(alpha), function(i){
      stats::qbeta(p, shape1 = alpha_alpha[,i], shape2 = alpha_beta[,i], lower.tail = lower.tail, log.p = log.p)
    }))
  )

  # make sure that the range is correct qpoint operates on unrestricted range
  q <- ifelse(q == -Inf, 0, q)
  q <- ifelse(q ==  Inf, 1, q)
  return(q)
}
.mqone.sided_fixed     <- function(p, omega, lower.tail, log.p){

  .weightfunctions_check_omega(omega, "omega")

  # transform to matrices for easier manipulation and checks
  if(!is.matrix(omega)){
    omega <- matrix(omega, nrow = 1)
  }
  if(length(p) == 1){
    p <- rep(p, nrow(omega))
  }

  if(nrow(omega) != length(p) & nrow(omega) != 1)
    stop("Non matching dimensions of 'omega' and 'p'.")

  if(nrow(omega) != length(p) & nrow(omega) == 1){
    omega <- do.call(rbind, lapply(1:length(p), function(i)omega))
  }

  q <- do.call(cbind, lapply(1:ncol(omega), function(i)qpoint(p, location = omega[,i], lower.tail = lower.tail, log.p = log.p)))

  # make sure that the range is correct
  q <- ifelse(q == -Inf, 0, q)
  q <- ifelse(q ==  Inf, omega, q)
  return(q)
}

### helper functions
.weightfunctions_one_sided_parameterization <- function(alpha, alpha1, alpha2){
  has_alpha  <- !is.null(alpha)
  has_alpha1 <- !is.null(alpha1)
  has_alpha2 <- !is.null(alpha2)

  if(has_alpha && (has_alpha1 || has_alpha2))
    stop("Use exactly one one-sided weightfunction parameterization: either 'alpha' or both 'alpha1' and 'alpha2'.")
  if(has_alpha)
    return("monotonic")
  if(has_alpha1 && has_alpha2)
    return("general")
  if(has_alpha1 || has_alpha2)
    stop("The general one-sided weightfunction parameterization requires both 'alpha1' and 'alpha2'.")

  stop("A one-sided weightfunction parameterization is required: use either 'alpha' or both 'alpha1' and 'alpha2'.")
}
.weightfunctions_check_alpha <- function(alpha, name = "alpha"){
  if(!is.numeric(alpha) | !(is.vector(alpha) | is.matrix(alpha)))
    stop(paste0("'", name, "' must be a numeric vector or a matrix."))
  if(is.vector(alpha))if(length(alpha) < 2)
    stop(paste0("'", name, "' must be a vector of length at least 2."))
  if(is.matrix(alpha))if(ncol(alpha) < 2)
    stop(paste0("'", name, "' must be a matrix with at least 2 columns."))
  if(!all(alpha > 0))
    stop(paste0("'", name, "' must be positive."))
}
.weightfunctions_check_omega <- function(omega, name = "omega"){
  if(!is.numeric(omega) | !(is.vector(omega) | is.matrix(omega)))
    stop(paste0("'", name, "' must be a numeric vector or a matrix."))
  if(is.vector(omega))if(length(omega) < 2)
    stop(paste0("'", name, "' must be a vector of length at least 2."))
  if(is.matrix(omega))if(ncol(omega) < 2)
    stop(paste0("'", name, "' must be a matrix with at least 2 columns."))
  if(!all(omega >= 0))
    stop(paste0("'", name, "' must be non-negative."))
  if(is.vector(omega) && !isTRUE(all.equal(omega[1], 1)))
    stop(paste0("The reference-bin '", name, "' weight must be exactly 1."))
  if(is.matrix(omega) && !all(vapply(seq_len(nrow(omega)), function(i) isTRUE(all.equal(omega[i,1], 1)), logical(1))))
    stop(paste0("The reference-bin '", name, "' weight must be exactly 1."))
}

# functions for checking input into prior distributions related functions
.check_and_name_parameters <- function(parameters, names, distribution){

  if(length(parameters) != length(names))
    stop(paste0(distribution, " prior distribution requires ", length(names), " parameters."), call. = FALSE)
  if(!is.null(names(parameters))){
    if(!all(names(parameters) %in% names))
      stop(paste0("Parameters ", paste(paste0("'", names(parameters)[!names(parameters) %in% c(names, "")], "'"), sep = ", ", collapse = ""), " are not supported for a ", distribution," distribution."), call. = FALSE)
  }else{
    names(parameters) <- names
  }

  # order
  parameters <- parameters[names]

  return(parameters)
}
.check_and_set_truncation  <- function(truncation, lower = -Inf, upper = Inf){

  if(length(truncation) > 2)
    stop("More than two truncation points were supplied.")

  if(!is.null(names(truncation))){
    if(!all(names(truncation) %in% c("lower", "upper")))
      stop("Truncation points must be named 'lower' and 'upper'.")
    if(length(truncation) == 1){
      if(names(truncation) == "lower"){
        truncation$upper  <- upper
      }else if(names(truncation) == "upper"){
        truncation$lower  <- lower
      }
    }
  }else{
    if(length(truncation) == 2){
      names(truncation) <- c("lower", "upper")
    }else if(length(truncation) == 1){
      names(truncation) <- "lower"
      truncation$upper  <- upper
    }else{
      truncation <- list(
        lower = lower,
        upper = upper
      )
    }
  }


  if(lower != -Inf){
    # change the default value to a distribution specific value or throw an error of misspecified by user
    if(truncation$lower == -Inf){
      truncation$lower <- lower
    }else if(truncation$lower < lower)
      stop(paste0("Lower truncation point must be larger or equal to ", lower, "."))
  }

  if(upper != Inf){
    # change the default value to a distribution specific value or throw an error of misspecified by user
    if(truncation$upper == Inf){
      truncation$upper <- upper
    }else if(truncation$upper > upper)
      stop(paste0("Upper truncation point must be smaller or equal to ", upper, "."))
  }

  if(truncation$lower >= truncation$upper)
    stop("The lower truncation point must be lower than the upper truncation points.")

  # order
  truncation <- truncation[c("lower", "upper")]

  return(truncation)
}
.check_parameter           <- function(parameter, name, length = 1){

  # allow expressions to be forwarded through the construction functions
  if(is.expression(parameter))
    return()

  if(length == -1){
    if(!is.numeric(parameter) | !is.vector(parameter) | length(parameter) > 1)
      stop(paste0("The '", name, "' must be a numeric vector of length at least 2."), call. = FALSE)
  }else if(length == 0){
    if(!is.numeric(parameter) | !is.vector(parameter))
      stop(paste0("The '", name, "' must be a numeric vector."), call. = FALSE)
  }else{
    if(!is.numeric(parameter) | !is.vector(parameter) | length(parameter) != length)
      stop(paste0("The '", name, "' must be a numeric vector of length ", length, "."), call. = FALSE)
  }
}
.check_parameter_positive  <- function(parameter, name, include_zero = FALSE){

  # allow expressions to be forwarded through the construction functions
  if(is.expression(parameter))
    return()

  if(include_zero){
    if(any(parameter < 0))
      stop(paste0("The '", name, "' must be non-negative."), call. = FALSE)
  }else{
    if(any(parameter <= 0))
      stop(paste0("The '", name, "' must be positive."), call. = FALSE)
  }
}
.check_parameter_negative  <- function(parameter, name, include_zero = FALSE){

  # allow expressions to be forwarded through the construction functions
  if(is.expression(parameter))
    return()

  if(include_zero){
    if(any(parameter > 0))
      stop(paste0("The '", name, "' must be non-positive."), call. = FALSE)
  }else{
    if(any(parameter >= 0))
      stop(paste0("The '", name, "' must be negative."), call. = FALSE)
  }
}
.check_parameter_dimensions<- function(parameter, name, allow_NA = FALSE){

  # allow expressions to be forwarded through the construction functions
  if(is.expression(parameter))
    return()

  if(length(parameter) != 1)
    stop(paste0("The '", name, "' must be a numeric vector of length 1."), call. = FALSE)

  if(is.na(parameter)){
    if(allow_NA)
      return()
    stop(paste0("The '", name, "' must be defined."), call. = FALSE)
  }

  if(!is.numeric(parameter))
    stop(paste0("The '", name, "' must be a numeric vector of length 1."), call. = FALSE)

  if(!.is.wholenumber(parameter))
    stop(paste0("The '", name ,"' must be an integer"), call. = FALSE)

  if(parameter < 1)
    stop(paste0("The '", name ,"' must be a positive integer"), call. = FALSE)
}
.check_parameter_range          <- function(parameter, name, lower, upper, include_bounds = FALSE){

  # allow expressions to be forwarded through the construction functions
  if(is.expression(parameter))
    return()

  if(include_bounds){
    if(any(parameter < lower) | any(parameter > upper))
      stop(paste0("The '", name, "' must be higher than ", lower, " and lower than ", upper, "."), call. = FALSE)
  }else{
    if(any(parameter <= lower) | any(parameter >= upper))
      stop(paste0("The '", name, "' must be higher or equal to than ", lower, " and lower or equal to than ", upper, "."), call. = FALSE)
  }
}

.get_prior_factor_levels       <- function(prior){
  if(is.prior.independent(prior)){
    return(attr(prior, "levels"))
  }else if(is.prior.treatment(prior)){
    return(attr(prior, "levels") - 1)
  }else if(is.prior.orthonormal(prior)){
    return(attr(prior, "levels") - 1)
  }else if(is.prior.meandif(prior)){
    return(attr(prior, "levels") - 1)
  }else if(is.prior.point(prior)){
    # allow to deal with spike priors assuming there is an intercept & the prior is not independent
    return(attr(prior, "levels") - 1)
  }
}
.get_prior_factor_level_names  <- function(prior){
  if(is.null(attr(prior, "level_names"))){
    if(is.prior.independent(prior)){
      return(1:.get_prior_factor_levels(prior))
    }else{
      return(1:(.get_prior_factor_levels(prior)+1))
    }
  }else{
    return(attr(prior, "level_names"))
  }
}
.get_prior_factor_list_type    <- function(prior_list){

  priors_type <- do.call(rbind, lapply(prior_list, function(p) {
    if(is.prior.point(p) | is.prior.none(p)){
      return(data.frame(
        "treatment"         = NA,
        "independent"       = NA,
        "orthonormal"       = NA,
        "meandif"           = NA,
        "K"                 = NA
      ))
    }else{
      return(data.frame(
        "treatment"         = is.prior.treatment(p),
        "independent"       = is.prior.independent(p),
        "orthonormal"       = is.prior.orthonormal(p),
        "meandif"           = is.prior.meandif(p),
        "K"                 = if(!is.null(p[["parameters"]][["K"]])) p[["parameters"]][["K"]] else NA
      ))
    }
  }))
  priors_type <- priors_type[!apply(priors_type, 1, function(x) all(is.na(x))),]
  priors_type <- unique(priors_type)

  if(nrow(priors_type) > 1)
    stop("The prior_list must contain only one type of factor priors.")

  class_type <- unlist(priors_type[-ncol(priors_type)])
  class_type <- names(class_type)[which(class_type)]

  return(list(
    "class" = paste0("prior.", class_type),
    "K"     = priors_type[["K"]]
  ))
}
.is_prior_interaction          <- function(prior){
  if(is.null(attr(prior, "interaction"))){
    return(FALSE)
  }else{
    return(attr(prior, "interaction"))
  }
}
.is_prior_expression           <- function(prior){

  parameters <- prior[["parameters"]]
  if(is.null(parameters)){
    return(FALSE)
  }

  for(i in seq_along(parameters)){
    if(is.expression(parameters[[i]])){
      return(TRUE)
    }
  }

  return(FALSE)
}
.prior_expression_to_character <- function(prior){
  prior[["parameters"]] <- lapply(prior[["parameters"]], function(x){
    if(is.expression(x)){
      return(.clean_from_expression(deparse(x)))
    }else{
      return(x)
    }
  })
  return(prior)
}

#' @title Reports whether x is a a prior object
#'
#' @description Reports whether x is a a prior object. Note that
#' point priors inherit the prior.simple property
#'
#' @param x an object of test
#'
#' @examples
#' # create some prior distributions
#' p0 <- prior(distribution = "point",  parameters = list(location = 0))
#' p1 <- prior_PET(distribution = "normal", parameters = list(mean = 0, sd = 1))
#'
#' is.prior(p0)
#' is.prior.simple(p0)
#' is.prior.point(p0)
#' is.prior.PET(p0)
#'
#' is.prior(p1)
#' is.prior.simple(p1)
#' is.prior.point(p1)
#' is.prior.PET(p1)
#'
#' @return returns a boolean indicating whether the test object
#' is a prior (of specific type).
#'
#' @export is.prior
#' @export is.prior.simple
#' @export is.prior.discrete
#' @export is.prior.vector
#' @export is.prior.point
#' @export is.prior.none
#' @export is.prior.PET
#' @export is.prior.PEESE
#' @export is.prior.weightfunction
#' @export is.prior.factor
#' @export is.prior.orthonormal
#' @export is.prior.meandif
#' @export is.prior.treatment
#' @export is.prior.independent
#' @export is.prior.spike_and_slab
#' @export is.prior.mixture
#' @name is.prior
NULL

#' @rdname is.prior
is.prior                 <- function(x){
  inherits(x, "prior")
}
#' @rdname is.prior
is.prior.point           <- function(x){
  inherits(x, "prior.point")
}
#' @rdname is.prior
is.prior.none            <- function(x){
  inherits(x, "prior.none")
}
#' @rdname is.prior
is.prior.simple          <- function(x){
  inherits(x, "prior.simple")
}
#' @rdname is.prior
is.prior.discrete        <- function(x){
  inherits(x, "prior.discrete")
}
#' @rdname is.prior
is.prior.vector          <- function(x){
  inherits(x, "prior.vector")
}
#' @rdname is.prior
is.prior.PET             <- function(x){
  inherits(x, "prior.PET")
}
#' @rdname is.prior
is.prior.PEESE           <- function(x){
  inherits(x, "prior.PEESE")
}
#' @rdname is.prior
is.prior.weightfunction  <- function(x){
  inherits(x, "prior.weightfunction")
}
.is_prior_weightfunction_null <- function(x){
  is.prior.none(x) ||
    (is.prior.point(x) &&
       length(x$parameters[["location"]]) == 1L &&
       isTRUE(all.equal(x$parameters[["location"]], 1)))
}
#' @rdname is.prior
is.prior.factor          <- function(x){
  inherits(x, "prior.factor")
}
#' @rdname is.prior
is.prior.orthonormal     <- function(x){
  inherits(x, "prior.orthonormal")
}
#' @rdname is.prior
is.prior.treatment       <- function(x){
  inherits(x, "prior.treatment")
}
#' @rdname is.prior
is.prior.independent     <- function(x){
  inherits(x, "prior.independent")
}
#' @rdname is.prior
is.prior.spike_and_slab  <- function(x){
  inherits(x, "prior.spike_and_slab")
}
#' @rdname is.prior
is.prior.meandif         <- function(x){
  inherits(x, "prior.meandif")
}
#' @rdname is.prior
is.prior.mixture         <- function(x){
  inherits(x, "prior.mixture")
}

.check_prior <- function(prior, name = "prior", allow_expressions = FALSE){

  if(!is.prior(prior))
    stop(paste0("The '", name, "' argument must be a valid prior object."), call. = FALSE)

  if(!allow_expressions && .is_prior_expression(prior))
    stop(paste0("The '", name, "' argument must not contain parameter expressions."), call. = FALSE)

  return()
}
.prior_model_weight <- function(prior){

  prior_weight <- attr(prior, "model_prior_weights", exact = TRUE)
  if(!is.null(prior_weight)){
    return(prior_weight)
  }

  prior[["prior_weights"]]
}
.set_prior_model_weight <- function(prior, prior_weight){

  if(is.prior.mixture(prior)){
    attr(prior, "model_prior_weights") <- prior_weight
  }else{
    prior[["prior_weights"]] <- prior_weight
  }

  prior
}
.check_prior_weight <- function(prior_weights, name = "prior_weights"){

  check_real(prior_weights, name, lower = 0, allow_bound = FALSE, allow_NA = FALSE)

  if(any(!is.finite(prior_weights)))
    stop(paste0("The '", name, "' must be finite."), call. = FALSE)

  return()
}
.check_prior_list_unique_names <- function(prior_list, name = "prior_list"){

  prior_names <- names(prior_list)
  if(!is.null(prior_names) && anyDuplicated(prior_names)){
    duplicate_names <- unique(prior_names[duplicated(prior_names)])
    stop(
      paste0(
        "The '", name, "' argument must not contain duplicate names ('",
        paste0(duplicate_names, collapse = "', '"), "')."
      ),
      call. = FALSE
    )
  }

  return()
}
.check_prior_list <- function(prior_list, name = "prior_list", check_length = 0, allow_NULL = FALSE,
                              allow_prior.point = TRUE, allow_prior.simple = TRUE, allow_prior.discrete = TRUE, allow_prior.vector = TRUE,
                              allow_prior.PET = TRUE, allow_prior.PEESE = TRUE, allow_prior.weightfunction = TRUE, allow_prior.factor = TRUE,
                              allow_expressions = FALSE){

  check_list(prior_list, name, check_length = check_length, allow_other = FALSE, allow_NULL = allow_NULL)

  if(allow_NULL && is.null(prior_list) || length(prior_list) == 0)
    return()

  for(i in seq_along(prior_list)){
    .check_prior(prior_list[[i]], paste0(name, "[[", i, "]] argument must be a prior distribution."), allow_expressions = allow_expressions)

    if(!allow_prior.point && is.prior.point(prior_list[[i]]))
      stop(paste0("The '", name, "' argument must not contain point priors (element [[", i, "]])."), call. = FALSE)

    if(!allow_prior.simple && is.prior.simple(prior_list[[i]]))
      stop(paste0("The '", name, "' argument must not contain simple priors (element [[", i, "]])."), call. = FALSE)

    if(!allow_prior.discrete && is.prior.discrete(prior_list[[i]]))
      stop(paste0("The '", name, "' argument must not contain discrete priors (element [[", i, "]])."), call. = FALSE)

    if(!allow_prior.vector && is.prior.vector(prior_list[[i]]))
      stop(paste0("The '", name, "' argument must not contain vector priors (element [[", i, "]])."), call. = FALSE)

    if(!allow_prior.PET && is.prior.PET(prior_list[[i]]))
      stop(paste0("The '", name, "' argument must not contain PET priors (element [[", i, "]])."), call. = FALSE)

    if(!allow_prior.PEESE && is.prior.PEESE(prior_list[[i]]))
      stop(paste0("The '", name, "' argument must not contain PEESE priors (element [[", i, "]])."), call. = FALSE)

    if(!allow_prior.weightfunction && is.prior.weightfunction(prior_list[[i]]))
      stop(paste0("The '", name, "' argument must not contain weightfunction priors (element [[", i, "]])."), call. = FALSE)

    if(!allow_prior.factor && is.prior.factor(prior_list[[i]]))
      stop(paste0("The '", name, "' argument must not contain factor priors (element [[", i, "]])."), call. = FALSE)

    if(is.prior.mixture(prior_list[[i]])){
      stop(paste0("The '", name, "' argument must not contain mixture priors (element [[", i, "]])."), call. = FALSE)
    }
  }

  return()
}


# clean user input
.prior_clean_input_name <- function(name){

  name <- tolower(name)
  name <- gsub(".", "", name, fixed = TRUE)
  name <- gsub(",", "", name, fixed = TRUE)
  name <- gsub(";", "", name, fixed = TRUE)
  name <- gsub("/", "", name, fixed = TRUE)
  name <- gsub("-", "", name, fixed = TRUE)
  name <- gsub("_", "", name, fixed = TRUE)
  name <- gsub(" ", "", name, fixed = TRUE)

  return(name)
}

#' @title Create JAGS formula syntax and data object
#'
#' @description Creates a JAGS formula syntax, prepares data input, and
#' returns modified prior list for further processing in the \code{JAGS_fit}
#' function.
#'
#' @param formula formula specifying the right hand side of the assignment (the
#' left hand side is ignored). If the formula contains \code{-1}, it will be
#' automatically converted to include an intercept with a spike(0) prior.
#' The formula can also have a \code{"log(intercept)"} attribute set to \code{TRUE}
#' to generate syntax of the form \code{log(intercept) + sum(beta_i * x_i)}, which
#' is useful for parameters that must be positive (e.g., standard deviation).
#' @param parameter name of the parameter to be created with the formula
#' @param data data.frame containing predictors included in the formula
#' @param prior_list named list of prior distribution of parameters specified within
#' the \code{formula}. When using \code{-1} in the formula, an "intercept" prior
#' can be explicitly specified; otherwise, \code{prior("spike", list(0))} is
#' automatically added. The list can also include two special entries:
#' @param formula_scale named list specifying whether to standardize continuous predictors.
#' If \code{NULL} (default), no standardization is applied. If a named list is provided,
#' continuous predictors with \code{TRUE} values will be standardized (mean-centered and
#' scaled by standard deviation). The intercept is never standardized.
#' \describe{
#'   \item{\code{"__default_continuous"}}{A prior to use for any continuous predictors
#'     (including the intercept) that are not explicitly specified in the prior list.}
#'   \item{\code{"__default_factor"}}{A prior to use for any factor predictors
#'     (including interactions involving factors) that are not explicitly specified
#'     in the prior list.}
#' }
#' These default priors allow for more concise specification when many predictors
#' share the same prior distribution.
#'
#' @details When a formula with \code{-1} (no intercept) is specified, the
#' function automatically removes the \code{-1}, adds an intercept back to the
#' formula, and includes a spike(0) prior for the intercept to ensure equivalent
#' model behavior while maintaining consistent formula parsing.
#'
#' When using default priors (\code{"__default_continuous"} or \code{"__default_factor"}),
#' explicitly specified priors for individual terms take precedence over the defaults.
#' The defaults are only applied to terms that are not already in the prior list.
#'
#' @examples
#' # simulate data
#' set.seed(1)
#' df <- data.frame(
#'   y      = rnorm(60),
#'   x_cont = rnorm(60),
#'   x_bin  = rbinom(60, 1, .5),
#'   x_fac3 = factor(rep(c("A", "B", "C"), 20), levels = c("A", "B", "C")),
#'   x_fac4 = factor(rep(c("A", "B", "C", "D"), 15), levels = c("A", "B", "C", "D"))
#' )
#'
#' # specify priors with intercept
#' prior_list <- list(
#' "intercept"     = prior("normal", list(0, 1)),
#' "x_cont"        = prior("normal", list(0, .5)),
#' "x_fac3"        = prior_factor("normal",  list(0, 1),  contrast = "treatment"),
#' "x_fac4"        = prior_factor("mnormal", list(0, 1),  contrast = "orthonormal"),
#' "x_fac3:x_fac4" = prior_factor("mnormal", list(0, .5), contrast = "orthonormal")
#' )
#'
#' # create the formula object
#' formula_obj <- JAGS_formula(
#'   formula = ~ x_cont + x_fac3 * x_fac4,
#'   parameter = "mu", data = df, prior_list = prior_list)
#'
#' # using -1 notation (automatically adds spike(0) intercept)
#' prior_list_no_intercept <- list(
#'   "x_fac3" = prior_factor("normal", list(0, 1), contrast = "treatment")
#' )
#' formula_no_intercept <- JAGS_formula(
#'   formula = ~ x_fac3 - 1,
#'   parameter = "mu", data = df, prior_list = prior_list_no_intercept)
#' # Equivalent to specifying intercept = prior("spike", list(0))
#'
#' # using default priors for simpler specification
#' prior_list_defaults <- list(
#'   "__default_continuous" = prior("normal", list(0, 1)),
#'   "__default_factor"     = prior_factor("normal", list(0, 0.5), contrast = "treatment")
#' )
#' formula_defaults <- JAGS_formula(
#'   formula = ~ x_cont + x_fac3,
#'   parameter = "mu", data = df, prior_list = prior_list_defaults)
#' # intercept and x_cont get the default continuous prior
#' # x_fac3 gets the default factor prior
#'
#' @return \code{JAGS_formula} returns a list containing the formula JAGS syntax,
#' JAGS data object, modified prior_list, and (if standardization was applied) a
#' \code{formula_scale} list with standardization information for back-transformation.
#'
#' @seealso [JAGS_fit()]
#' @export
JAGS_formula <- function(formula, parameter, data, prior_list, formula_scale = NULL){

  if(!is.language(formula))
    stop("'formula' must be a formula")
  check_char(parameter, "parameter")
  if(!is.data.frame(data))
    stop("'data' must be a data.frame")
  check_list(prior_list, "prior_list")
  if(any(!sapply(prior_list, is.prior)))
    stop("'prior_list' must be a list of priors.")
  # formula_scale can be TRUE/FALSE (apply to all) or a named list
  if(!is.null(formula_scale) && !is.logical(formula_scale) && !is.list(formula_scale)){
    stop("'formula_scale' must be NULL, TRUE, FALSE, or a named list")
  }


  # remove the specified response
  formula <- .remove_response(formula)
  # store log(intercept) attribute (for models relying on mu = log(intercept) + sum(beta_i * x_i) trick
  # exp(mu) = intercept * exp(sum(beta_i * x_i)) (e.g., Poisson regression / regression with log link etc...)
  log_intercept  <- isTRUE(attr(formula, "log(intercept)"))
  # store expressions (included later as the literal character input)
  expressions    <- .extract_expressions(formula)
  # store random effects (included later via a formula interface)
  random_effects <- .extract_random_effects(formula)
  # remove expressions and random effects from the formula
  formula <- .remove_expressions(formula)
  formula <- .remove_random_effects(formula)

  # handle -1 (no intercept) formulas: always add intercept back with spike(0) prior
  no_intercept_specified <- attr(stats::terms(formula), "intercept") == 0
  if(no_intercept_specified){
    # remove -1 from formula and add intercept back
    formula <- formula_add_intercept(formula)
    # add spike(0) prior for intercept if not already specified
    if(!"intercept" %in% names(prior_list)){
      prior_list[["intercept"]] <- prior("spike", list(0))
    }
  }

  # obtain predictors characteristics factors
  formula_terms    <- stats::terms(formula)
  has_intercept    <- attr(formula_terms, "intercept") == 1
  predictors       <- as.character(attr(formula_terms, "variables"))[-1]
  if(any(!predictors %in% colnames(data)))
    stop(paste0("The ", paste0("'", predictors[!predictors %in% colnames(data)], "'", collapse = ", ")," predictor variable is missing in the data set."))
  predictors_type  <- sapply(predictors, function(predictor){
    if(is.factor(data[,predictor]) | is.character(data[,predictor])){
      return("factor")
    }else{
      return("continuous")
    }
  })
  model_terms      <- c(if(has_intercept) "intercept", attr(formula_terms, "term.labels"))
  model_terms_type <- sapply(model_terms, function(model_term){
    model_term <- strsplit(model_term, ":")[[1]]
    if(length(model_term) == 1 && model_term == "intercept"){
      return("continuous")
    }else if(any(predictors_type[model_term] == "factor")){
      return("factor")
    }else{
      return("continuous")
    }
  })

  # separate prior lists: extract random effects priors
  prior_list_random_effects <- list()
  if(length(random_effects) > 0){
    # store the random effects specific priors in the corresponding entry
    for(i in seq_along(random_effects)){
      prior_list_random_effects[[i]] <- prior_list[which(.get_grouping_factor(names(prior_list)) == attr(random_effects[[i]], "grouping_factor"))]
    }
    # remove the random effects specific priors from the prior list
    prior_list <- prior_list[.get_grouping_factor(names(prior_list)) == ""]
  }

  # handle default priors: __default_factor and __default_continuous
  default_factor_prior     <- prior_list[["__default_factor"]]
  default_continuous_prior <- prior_list[["__default_continuous"]]
  has_defaults <- !is.null(default_factor_prior) || !is.null(default_continuous_prior)

  # remove default priors from prior_list before validation
  prior_list[["__default_factor"]]     <- NULL
  prior_list[["__default_continuous"]] <- NULL

  # fill in missing priors with defaults based on term type
  if(has_defaults){
    for(term in model_terms){
      if(!term %in% names(prior_list)){
        term_type <- model_terms_type[[term]]
        if(term_type == "factor" && !is.null(default_factor_prior)){
          prior_list[[term]] <- default_factor_prior
        }else if(term_type == "continuous" && !is.null(default_continuous_prior)){
          prior_list[[term]] <- default_continuous_prior
        }
      }
    }
  }

  # check that all predictors have a prior distribution
  check_list(prior_list, "prior_list", check_names = model_terms, allow_other = FALSE, all_objects = TRUE)

  # check the prior distribution for each predictor
  # assign factor contrasts to the data based on prior distributions
  if(any(predictors_type == "factor")){

    for(factor in names(predictors_type[predictors_type == "factor"])){

      # select the corresponding prior for the variable
      this_prior <- prior_list[[factor]]

      if(is.prior.treatment(this_prior)){
        stats::contrasts(data[,factor]) <- "contr.treatment"
      }else if(is.prior.independent(this_prior)){
        stats::contrasts(data[,factor]) <- "contr.independent"
      }else if(is.prior.orthonormal(this_prior)){
        stats::contrasts(data[,factor]) <- "contr.orthonormal"
      }else if(is.prior.meandif(this_prior)){
        stats::contrasts(data[,factor]) <- "contr.meandif"
      }else{
        stop(paste0("Unsupported prior distribution defined for '", factor, "' factor variable. See '?prior_factor' for details."))
      }
    }
  }
  if(any(predictors_type == "continuous")){

    for(continuous in names(predictors_type[predictors_type == "continuous"])){

      # select the corresponding prior for the variable
      this_prior <- prior_list[[continuous]]

      if(is.prior.factor(this_prior)|| is.prior.discrete(this_prior) || is.prior.PET(this_prior) || is.prior.PEESE(this_prior) || is.prior.weightfunction(this_prior)){
        stop(paste0("Unsupported prior distribution defined for '", continuous, "' continuous variable. See '?prior' for details."))
      }
    }

    # standardize continuous predictors if requested
    scale_info <- list()
    if(!is.null(formula_scale)){
      for(continuous in names(predictors_type[predictors_type == "continuous"])){
        # determine if this predictor should be scaled
        should_scale <- FALSE
        if(is.logical(formula_scale) && length(formula_scale) == 1){
          # formula_scale = TRUE/FALSE applies to all continuous predictors
          should_scale <- isTRUE(formula_scale)
        }else if(is.list(formula_scale) && !is.null(formula_scale[[continuous]])){
          # named list: check specific predictor
          should_scale <- isTRUE(formula_scale[[continuous]])
        }

        if(should_scale){
          # store original mean and sd
          scale_info[[continuous]] <- list(
            mean = mean(data[, continuous], na.rm = TRUE),
            sd   = stats::sd(data[, continuous], na.rm = TRUE)
          )
          if(is.na(scale_info[[continuous]]$sd) || !is.finite(scale_info[[continuous]]$sd) || scale_info[[continuous]]$sd <= 0){
            stop(paste0("Cannot standardize predictor '", continuous, "' because its standard deviation must be positive and finite."), call. = FALSE)
          }
          # standardize the predictor
          data[, continuous] <- (data[, continuous] - scale_info[[continuous]]$mean) / scale_info[[continuous]]$sd
        }
      }
    }
  }

  # get the default design matrix
  model_frame  <- stats::model.frame(formula, data = data)
  model_matrix <- stats::model.matrix(model_frame, formula = formula, data = data)
  raw_column_names <- colnames(model_matrix)

  # check whether intercept is unique parameter
  if(sum(grepl("intercept", names(prior_list))) > 1)
    stop("only the intercept parameter can contain 'intercept' in its name.")
  # check whether any reserved term is in usage (note: __default_factor/__default_continuous are reserved but already removed from prior_list)
  reserved_terms <- c("__xXx__", "__xREx__", "xRE_PRECx", "xRE_CORx", "xRE_Zx", "xRE_STDx", "xRE_COEFx", "xRE_MAPx", "xRE_COEFx", "xRE_DATAx", "__default_factor", "__default_continuous")
  for(reserved_term in reserved_terms){
    if(any(grepl(reserved_term, colnames(data))))
      stop(paste0("'", reserved_term, "' string is internally used by the BayesTools package and can't be used for naming variables."))
  }


  # replace interaction signs (due to JAGS incompatibility)
  colnames(model_matrix)  <- gsub(":", "__xXx__", colnames(model_matrix))
  column_names            <- colnames(model_matrix)
  names(prior_list)       <- gsub(":", "__xXx__", names(prior_list))
  names(model_terms_type) <- gsub(":", "__xXx__", names(model_terms_type))
  model_terms             <- gsub(":", "__xXx__", model_terms)

  # prepare syntax & data based on the formula
  formula_syntax <- NULL
  random_syntax  <- NULL
  JAGS_data      <- list()
  jags_data_names <- list()
  JAGS_data[[paste0("N_", parameter)]] <- nrow(data)

  # add intercept and prepare the indexing vector
  if(has_intercept){
    terms_indexes    <- attr(model_matrix, "assign") + 1
    terms_indexes[1] <- 0

    # use log(intercept) if the formula has the log(intercept) attribute
    if(log_intercept){
      formula_syntax <- c(formula_syntax, paste0("log(", parameter, "_intercept)"))
    }else{
      formula_syntax <- c(formula_syntax, paste0(parameter, "_intercept"))
    }
  }else{
    terms_indexes    <- attr(model_matrix, "assign")
  }

  # add remaining terms (omitting the intercept indexed as NA)
  for(i in unique(terms_indexes[terms_indexes > 0])){

    # extract the corresponding prior distribution for a given coefficient
    this_prior <- prior_list[[model_terms[i]]]

    # check whether the term is an interaction or not and save the corresponding attributes
    attr(this_prior, "interaction") <- grepl("__xXx__", model_terms[i])
    if(.is_prior_interaction(this_prior)){
      attr(this_prior, "interaction_terms") <- strsplit(model_terms[i], "__xXx__")[[1]]
    }


    if(model_terms_type[i] == "continuous"){

      # continuous variables or interactions of continuous variables are simple predictors
      data_name <- paste0(parameter, "_data_", model_terms[i])
      JAGS_data[[data_name]] <- model_matrix[,terms_indexes == i]
      jags_data_names[[model_terms[i]]] <- data_name

      formula_syntax <- c(formula_syntax, paste0(
        if(!is.null(attr(this_prior, "multiply_by"))) paste0(attr(this_prior, "multiply_by"), " * "),
        parameter, "_", model_terms[i],
        " * ",
        parameter, "_data_", model_terms[i], "[i]"
      ))

    }else if(model_terms_type[i] == "factor"){

      # factor variables or interactions with a factor requires factor style prior

      # add levels information attributes to factors
      if(is.prior.independent(this_prior)){
        attr(this_prior, "levels") <- sum(terms_indexes == i)
      }else{
        attr(this_prior, "levels") <- sum(terms_indexes == i) + 1
      }
      if(.is_prior_interaction(this_prior)){
        level_names <- list()
        for(sub_term in strsplit(model_terms[i], "__xXx__")[[1]]){
          if(predictors_type[sub_term] == "factor"){
            level_names[[sub_term]] <- levels(data[,sub_term])
          }
        }
        attr(this_prior, "level_names") <- level_names
      }else{
        attr(this_prior, "level_names") <- levels(data[,model_terms[i]])
      }
      attr(this_prior, "term_components") <- strsplit(model_terms[i], "__xXx__", fixed = TRUE)[[1]]
      attr(this_prior, "factor_terms") <- if(is.list(attr(this_prior, "level_names"))) {
        names(attr(this_prior, "level_names"))
      } else {
        model_terms[i]
      }
      attr(this_prior, "factor_contrasts") <- vapply(attr(this_prior, "factor_terms"), function(factor_term) {
        factor_contrast <- attr(data[[factor_term]], "contrasts")
        if(is.null(factor_contrast)){
          "contr.treatment"
        }else if(is.character(factor_contrast)){
          factor_contrast[1]
        }else{
          stop("Unsupported matrix-valued factor contrast metadata.", call. = FALSE)
        }
      }, character(1))
      factor_design_info <- .factor_term_design_from_formula(
        formula         = formula,
        data            = data,
        predictors      = predictors,
        predictors_type = predictors_type,
        term_index      = i,
        term_components = attr(this_prior, "term_components"),
        factor_terms    = attr(this_prior, "factor_terms"),
        has_intercept   = has_intercept
      )
      attr(this_prior, "factor_design")     <- factor_design_info[["design"]]
      attr(this_prior, "factor_cell_names") <- factor_design_info[["cell_names"]]

      data_name <- paste0(parameter, "_data_", model_terms[i])
      JAGS_data[[data_name]] <- model_matrix[,terms_indexes == i, drop = FALSE]
      jags_data_names[[model_terms[i]]] <- data_name
      formula_syntax <- c(formula_syntax, paste0(
        if(!is.null(attr(this_prior, "multiply_by"))) paste0(attr(this_prior, "multiply_by"), " * "),
        "inprod(",
        parameter, "_", model_terms[i],
        ", ",
        parameter, "_data_", model_terms[i], "[i,])"
      ))

    }else{
      stop("Unrecognized model term.")
    }

    # update the corresponding prior distribution back into the prior list
    # (and forward attributes to lower level components in the case of spike and slab and mixture priors)
    if(is.prior.spike_and_slab(this_prior) || is.prior.mixture(this_prior)){
      for(p in seq_along(this_prior)){
        attr(this_prior, "levels")            -> attr(this_prior[[p]], "levels")
        attr(this_prior, "level_names")       -> attr(this_prior[[p]], "level_names")
        attr(this_prior, "interaction")       -> attr(this_prior[[p]], "interaction")
        attr(this_prior, "interaction_terms") -> attr(this_prior[[p]], "interaction_terms")
        attr(this_prior, "term_components")   -> attr(this_prior[[p]], "term_components")
        attr(this_prior, "factor_terms")      -> attr(this_prior[[p]], "factor_terms")
        attr(this_prior, "factor_contrasts")  -> attr(this_prior[[p]], "factor_contrasts")
        attr(this_prior, "factor_design")     -> attr(this_prior[[p]], "factor_design")
        attr(this_prior, "factor_cell_names") -> attr(this_prior[[p]], "factor_cell_names")
      }
      this_prior -> prior_list[[model_terms[i]]]
    }else{
      this_prior -> prior_list[[model_terms[i]]]
    }

  }

  # add expressions input back to the formula
  for(i in seq_along(expressions)){
    formula_syntax <- c(formula_syntax, .clean_from_expression(expressions[[i]]))
  }

  # add random effects back to the formula
  random_scale_terms <- character()
  for(random_i in seq_along(random_effects)){
    temp_random   <- .JAGS_random_effect_formula(random_effects[[random_i]], parameter, data, prior_list_random_effects[[random_i]])

    for(data_i in seq_along(temp_random[["data"]])){
      JAGS_data[[names(temp_random[["data"]])[data_i]]] <- temp_random[["data"]][[data_i]]
    }
    random_key <- paste0("__xREx__", attr(random_effects[[random_i]], "grouping_factor"))
    jags_data_names[[random_key]] <- names(temp_random[["data"]])

    random_syntax  <- c(random_syntax,  temp_random[["random_syntax"]])
    formula_syntax <- c(formula_syntax, temp_random[["formula_term"]])
    prior_list     <- c(prior_list, temp_random[["prior_list"]])
    random_scale_terms <- c(random_scale_terms, temp_random[["random_scale_terms"]])
  }

  # finish the syntax
  formula_syntax <- paste0(
    "for(i in 1:N_", parameter, "){\n",
    "  ", parameter, "[i] = ", paste0(formula_syntax, collapse = " + "), "\n",
    "}\n")
  formula_syntax <- paste0(formula_syntax, paste0(random_syntax, collapse = "\n"), collapse = "\n")

  # add the parameter name as a prefix and attribute to each prior in the list
  names(prior_list) <- paste0(parameter, "_", names(prior_list))
  for(i in seq_along(prior_list)){
    attr(prior_list[[i]], "parameter") <- parameter
  }

  # preserve log(intercept) attribute on output formula
  if(log_intercept){
    attr(formula, "log(intercept)") <- TRUE
  }

  output <- list(
    formula_syntax = formula_syntax,
    data           = JAGS_data,
    prior_list     = prior_list,
    formula        = formula
  )

  # add scale information if standardization was applied
  if(exists("scale_info") && length(scale_info) > 0){
    # add parameter prefix to scale_info names for consistency
    names(scale_info) <- paste0(parameter, "_", names(scale_info))
    # store the parameter prefix as an attribute for later retrieval
    attr(scale_info, "parameter") <- parameter
    # store log_intercept attribute for proper unscaling transformation
    attr(scale_info, "log_intercept") <- log_intercept
    if(length(random_scale_terms) > 0){
      names(random_scale_terms) <- paste0(parameter, "_", names(random_scale_terms))
      attr(scale_info, "random_effect_terms") <- random_scale_terms
    }
    output$formula_scale <- scale_info
  }

  design_formula <- formula
  attr(design_formula, "log(intercept)") <- NULL

  output$formula_design <- .JAGS_formula_design_object(
    parameter         = parameter,
    formula           = design_formula,
    model_frame       = model_frame,
    model_matrix      = model_matrix,
    raw_column_names  = raw_column_names,
    column_names      = column_names,
    predictors        = predictors,
    predictors_type   = predictors_type,
    model_terms       = model_terms,
    model_terms_type  = model_terms_type,
    prior_list        = prior_list,
    formula_scale     = output$formula_scale,
    expressions       = expressions,
    random_effects    = random_effects,
    jags_data_names   = jags_data_names
  )

  return(output)
}

.JAGS_formula_design_object <- function(parameter, formula, model_frame, model_matrix,
                                        raw_column_names, column_names,
                                        predictors, predictors_type,
                                        model_terms, model_terms_type,
                                        prior_list, formula_scale,
                                        expressions, random_effects,
                                        jags_data_names){

  formula_terms <- stats::terms(formula)
  attr(formula_terms, ".Environment") <- emptyenv()
  formula_output <- formula
  environment(formula_output) <- emptyenv()
  model_frame_output <- model_frame
  attr(model_frame_output, "terms") <- formula_terms

  qr_info <- qr(model_matrix)
  aliased <- rep(FALSE, ncol(model_matrix))
  if(qr_info$rank < ncol(model_matrix)){
    aliased[qr_info$pivot[(qr_info$rank + 1L):ncol(model_matrix)]] <- TRUE
  }
  names(aliased) <- colnames(model_matrix)

  factor_predictors <- names(predictors_type)[predictors_type == "factor"]
  xlevels <- lapply(factor_predictors, function(predictor){
    if(predictor %in% names(model_frame) && is.factor(model_frame[[predictor]])){
      levels(model_frame[[predictor]])
    }else{
      NULL
    }
  })
  names(xlevels) <- factor_predictors
  xlevels <- xlevels[!vapply(xlevels, is.null, logical(1))]

  out <- list(
    parameter          = parameter,
    formula            = formula_output,
    model_frame        = model_frame_output,
    model_matrix       = model_matrix,
    column_names       = column_names,
    raw_column_names   = raw_column_names,
    assign             = attr(model_matrix, "assign"),
    terms              = formula_terms,
    contrasts          = attr(model_matrix, "contrasts"),
    xlevels            = xlevels,
    predictors         = predictors,
    predictor_types    = predictors_type,
    model_terms        = model_terms,
    model_terms_type   = model_terms_type,
    prior_list         = prior_list,
    formula_scale      = formula_scale,
    rank               = qr_info$rank,
    qr_pivot           = qr_info$pivot,
    aliased            = aliased,
    transformed_terms  = expressions,
    random_effects     = random_effects,
    jags_data_names    = jags_data_names
  )
  class(out) <- c("BayesTools_formula_design", "list")

  return(out)
}

#' @title Extract Fitted JAGS Formula Design Metadata
#'
#' @description Returns the fitted formula design metadata stored by
#' [JAGS_fit()]. The design contains the processed formula, fitted model frame,
#' exact model matrix used for JAGS data construction, JAGS-safe coefficient
#' names, contrast and factor-level metadata, rank diagnostics, prior metadata,
#' and formula-scale information.
#'
#' @param fit a fitted object returned by [JAGS_fit()].
#' @param parameter optional formula parameter name. If \code{NULL}, all stored
#' formula designs are returned.
#'
#' @return A named list of formula designs, or one formula design when
#' \code{parameter} is supplied. Returns \code{NULL} when no formula design
#' metadata is stored on \code{fit}.
#'
#' @seealso [JAGS_fit()] [JAGS_formula()]
#' @export
JAGS_formula_design <- function(fit, parameter = NULL){

  check_char(parameter, "parameter", allow_NULL = TRUE)

  formula_design <- attr(fit, "formula_design")
  if(is.null(formula_design)){
    return(NULL)
  }

  if(is.null(parameter)){
    return(formula_design)
  }

  if(!parameter %in% names(formula_design)){
    stop("Formula design for parameter '", parameter, "' was not found.", call. = FALSE)
  }

  return(formula_design[[parameter]])
}

.JAGS_random_effect_formula <- function(formula, parameter, data, prior_list){

  # extract the grouping factor information
  grouping_factor        <- attr(formula, "grouping_factor")
  grouping_independent   <- attr(formula, "independent")
  grouping_factor_levels <- levels(as.factor(data[[grouping_factor]]))
  grouping_mapping       <- as.numeric(as.factor(data[[grouping_factor]]))

  # TODO: expand to factor random effects
  # needs to implement LKJ correlation matrix
  if(!grouping_independent){
    stop("Only independent random effects are supported yet.")
  }

  # remove the grouping factor from the formula
  formula <- .remove_grouping_factor(formula)
  formula <- stats::as.formula(paste("~", formula))

  # obtain predictors characteristics factors (copy from formula)
  formula_terms    <- stats::terms(formula)
  has_intercept    <- attr(formula_terms, "intercept") == 1
  predictors       <- as.character(attr(formula_terms, "variables"))[-1]
  if(any(!predictors %in% colnames(data)))
    stop(paste0("The ", paste0("'", predictors[!predictors %in% colnames(data)], "'", collapse = ", ")," predictor variable is missing in the data set."))
  predictors_type  <- sapply(predictors, function(predictor){
    if(is.factor(data[,predictor]) | is.character(data[,predictor])){
      return("factor")
    }else{
      return("continuous")
    }
  })
  model_terms      <- c(if(has_intercept) "intercept", attr(formula_terms, "term.labels"))
  model_terms_type <- sapply(model_terms, function(model_term){
    model_term <- strsplit(model_term, ":")[[1]]
    if(length(model_term) == 1 && model_term == "intercept"){
      return("continuous")
    }else if(any(predictors_type[model_term] == "factor")){
      return("factor")
    }else{
      return("continuous")
    }
  })

  # check that all priors have a lower bound on 0 or their range is > 0, if not, throw a warning and correct
  for(i in seq_along(prior_list)){
    if(is.prior.spike_and_slab(prior_list[[i]]) || is.prior.mixture(prior_list[[i]])){
      for(j in seq_along(prior_list[[i]])){
        if(range(prior_list[[i]][[j]])[1] < 0){
          warning(paste0("The lower bound of the ", j ,"-th component in '", names(prior_list)[i], "' prior distribution is below 0. Correcting to 0."), immediate. = TRUE)
          prior_list[[i]][[j]]$truncation$lower <- 0
        }
      }
    }else{
      if(range(prior_list[[i]])[1] < 0){
        warning(paste0("The lower bound of the '", names(prior_list)[i], "' prior distribution is below 0. Correcting to 0."), immediate. = TRUE)
        prior_list[[i]]$truncation$lower <- 0
      }
    }
  }

  # drop the grouping factor from the prior names
  names(prior_list) <- .remove_grouping_factor(names(prior_list))
  # check that all terms have a prior distribution
  check_list(prior_list, "prior_list", check_names = model_terms, allow_other = TRUE, all_objects = TRUE)

  # get the default design matrix
  model_frame  <- stats::model.frame(formula, data = data)
  model_matrix <- stats::model.matrix(model_frame, formula = formula, data = data)

  # check whether intercept is unique parameter
  if(sum(grepl("intercept", names(prior_list))) > 1)
    stop("only the intercept parameter can contain 'intercept' in its name.")
  # check whether any reserved term is in usage
  reserved_terms <- c("__xXx__", "__xREx__", "xRE_PRECx", "xRE_CORx", "xRE_Zx", "xRE_STDx", "xRE_COEFx", "xRE_MAPx", "xRE_COEFx", "xRE_DATAx")
  for(reserved_term in reserved_terms){
    if(any(grepl(reserved_term, names(prior_list))))
      stop(paste0("'", reserved_term, "' string is internally used by the BayesTools package and can't be used for naming variables or prior distributions."))
  }

  # replace interaction signs (due to JAGS incompatibility)
  colnames(model_matrix)  <- gsub(":", "__xXx__", colnames(model_matrix))
  names(prior_list)       <- gsub(":", "__xXx__", names(prior_list))
  names(model_terms_type) <- gsub(":", "__xXx__", names(model_terms_type))
  model_terms             <- gsub(":", "__xXx__", model_terms)

  # prepare syntax & data based on the formula
  parameter_suffix <- paste0("_xREx__", grouping_factor)       # priors should not be named with parameter name (done on exit from formula)
  parameter        <- paste0(parameter, "_", parameter_suffix) # variables should be named already here
  random_syntax    <- NULL
  JAGS_data        <- list()
  new_prior_list   <- list()
  random_scale_terms <- character()

  ### in essence, the following prepares constructors that:
  # 1) samples standardized random effects xRE_Zx[ids, predictors] from a multivariate normal distribution
  # 2) create a vector of by-parameter standard deviation of the random effects xRE_STDx[predictors]
  # 3) multiplies the standardized random effects by parameter-specific standard deviations to create xRE_COEFx[ids, predictors] matrix
  # 4) computes the per observation formula output based on indexing the by-id COEF and selecting the observation variables
  # 5) appends the per-observation output to the higher order formula (done in the formula call itself)

  n_id  <- length(grouping_factor_levels)
  n_par <- ncol(model_matrix)

  # step 1:
  # TODO: get the identity matrix to sample the standardized random effects (update once correlated random effects are available with cholesky sampled correlation matrix)
  random_syntax <- c(random_syntax, .add_JAGS_matrix(name = paste0(parameter, "_xRE_PRECx"), diag(1, n_par)))
  random_syntax <- c(random_syntax, paste0(
    " for(i in 1:",n_id,"){\n",
    "   ",paste0(parameter, "_xRE_Zx"),"[i,1:", n_par ,"] ~ dmnorm(rep(0, ", n_par,"), ", paste0(parameter, "_xRE_PRECx"), ")\n",
    " }\n"
  ))

  # step 2
  if(has_intercept){
    terms_indexes    <- attr(model_matrix, "assign") + 1
    terms_indexes[1] <- 0

    new_prior_list[[paste0(parameter_suffix, "_intercept")]] <- prior_list[["intercept"]]
    random_scale_terms[[paste0(parameter_suffix, "_intercept")]] <- "intercept"
    attr(new_prior_list[[paste0(parameter_suffix, "_intercept")]], "random_factor") <- grouping_factor
    random_syntax   <- c(random_syntax, paste0(
      parameter, "_xRE_STDx[1] = ", parameter, "_", "intercept"
    ))
  }else{
    terms_indexes    <- attr(model_matrix, "assign")
  }

  # add remaining terms (omitting the intercept indexed as NA)
  for(i in unique(terms_indexes[terms_indexes > 0])){

    # extract the corresponding prior distribution for a given coefficient
    this_prior <- prior_list[[model_terms[i]]]

    # check whether the term is an interaction or not and save the corresponding attributes
    attr(this_prior, "interaction") <- grepl("__xXx__", model_terms[i])
    if(.is_prior_interaction(this_prior)){
      attr(this_prior, "interaction_terms") <- strsplit(model_terms[i], "__xXx__")[[1]]
    }

    if(!is.null(attr(this_prior, "multiply_by")))
      stop("'multiply_by' attribute is inadmissible for random effects")

    if(model_terms_type[i] == "continuous"){

      random_syntax <- c(random_syntax, paste0(
        parameter, "_xRE_STDx[", i, "] = ", parameter, "_", model_terms[i]
      ))

    }else if(model_terms_type[i] == "factor"){

      # factor random effects use the same contrasts as set in the upstream formula
      # (in the rare case that no factor_prior on the upstream formula was set, this might default to a treatment contrast)

      ## parameterization
      # treatment contrasts: independent variances for the comparison factor levels
      # independent contrasts: independent variances for each factor level
      # meandif/orthonormal contrasts: one total variance for the factor
      if(is.null(attr(data[[model_terms[i]]], "contrasts")) || attr(data[[model_terms[i]]], "contrasts") %in% c("contr.treatment", "contr.independent")){

        # determine the prior type
        if(is.null(attr(data[[model_terms[i]]], "contrasts")) || attr(data[[model_terms[i]]], "contrasts") == "contr.treatment"){
          temp_prior_type <- "prior.treatment"
          attr(this_prior, "levels") <- sum(terms_indexes == i) + 1
        }else{
          temp_prior_type <- "prior.independent"
          attr(this_prior, "levels") <- sum(terms_indexes == i)
        }

        # store level information
        if(.is_prior_interaction(this_prior)){
          level_names <- list()
          for(sub_term in strsplit(model_terms[i], "__xXx__")[[1]]){
            if(model_terms_type[sub_term] == "factor"){
              level_names[[sub_term]] <- levels(data[,sub_term])
            }
          }
          attr(this_prior, "level_names") <- level_names
        }else{
          attr(this_prior, "level_names") <- levels(data[,model_terms[i]])
        }

        # distribute the individual coefficients to the STD
        for(j in 1:sum(terms_indexes == i)){
          random_syntax <- c(random_syntax, paste0(
            parameter, "_xRE_STDx[", i + j - 1, "] = ", parameter, "_", model_terms[i], "[", j, "]"
          ))
        }

        # transform the simple prior into treatment / independent factor prior (for each of the coefficients)
        if(is.prior.simple(this_prior)){
          class(this_prior) <- c(class(this_prior), "prior.factor", temp_prior_type)
        }else if(is.prior.spike_and_slab(this_prior) || is.prior.mixture(this_prior)){
          for(p in seq_along(this_prior)){
            class(this_prior[[p]]) <- c(class(this_prior[[p]]), "prior.factor", temp_prior_type)
          }
          if(is.prior.spike_and_slab(this_prior)){
            class(this_prior) <- c(class(this_prior)[!class(this_prior) %in% c("prior.simple_spike_and_slab")], "prior.factor_spike_and_slab", temp_prior_type)
          } else {
            class(this_prior) <- c(class(this_prior)[!class(this_prior) %in% c("prior.simple_mixture")],  "prior.factor_mixture", temp_prior_type)
          }
        }

      }else if(attr(data[[model_terms[i]]], "contrasts") %in% c("contr.orthonormal", "contr.meandif")){

        # do not change the prior type of create multiple levels
        # distribute the same coefficients to the STD
        for(j in 1:sum(terms_indexes == i)){
          random_syntax <- c(random_syntax, paste0(
            parameter, "_xRE_STDx[", i + j - 1, "] = ", parameter, "_", model_terms[i]
          ))
        }
        # no prior transformation needed

      }else{
        stop("Unsupported factor contrasts for the random effects.")
      }


    }else{
      stop("Unrecognized model term.")
    }

    # update the corresponding prior distribution back into the prior list
    # (and forward attributes to lower level components in the case of spike and slab and mixture priors)
    attr(this_prior, "random_sd")     <- TRUE
    attr(this_prior, "random_factor") <- grouping_factor
    if(is.prior.spike_and_slab(this_prior) || is.prior.mixture(this_prior)){
      for(p in seq_along(this_prior)){
        attr(this_prior, "levels")            -> attr(this_prior[[p]], "levels")
        attr(this_prior, "level_names")       -> attr(this_prior[[p]], "level_names")
        attr(this_prior, "interaction")       -> attr(this_prior[[p]], "interaction")
        attr(this_prior, "interaction_terms") -> attr(this_prior[[p]], "interaction_terms")
      }
      this_prior -> new_prior_list[[paste0(parameter_suffix, "_", model_terms[i])]]
      random_scale_terms[[paste0(parameter_suffix, "_", model_terms[i])]] <- model_terms[i]
    }else{
      this_prior -> new_prior_list[[paste0(parameter_suffix, "_", model_terms[i])]]
      random_scale_terms[[paste0(parameter_suffix, "_", model_terms[i])]] <- model_terms[i]
    }

  }

  # step 3
  random_syntax <- c(random_syntax, paste0(
    " for(i in 1:",n_par,"){\n",
    "   ",paste0(parameter, "_xRE_COEFx"),"[1:",n_id,",i] = ",paste0(parameter, "_xRE_Zx"),"[1:",n_id,",i] * ",paste0(parameter, "_xRE_STDx"),"[i]\n",
    " }\n"
  ))

  # step 4
  random_syntax <- c(random_syntax, paste0(
    " for(i in 1:",nrow(model_matrix),"){\n",
    "   ",parameter,"[i] = inprod(", paste0(parameter, "_xRE_COEFx[", paste0(parameter, "_xRE_MAPx[i]"),", 1:",n_par,"]"), ", ", paste0(parameter, "_xRE_DATAx[i,1:", n_par,"]"),")\n",
    " }\n"
  ))

  # create the JAGS data list
  JAGS_data[[paste0(parameter, "_xRE_DATAx")]] <- model_matrix
  JAGS_data[[paste0(parameter, "_xRE_MAPx")]]  <- grouping_mapping

  return(list(
    random_syntax  = random_syntax,
    formula_term   = paste0(parameter,"[i]"),
    data           = JAGS_data,
    prior_list     = new_prior_list,
    random_scale_terms = random_scale_terms,
    formula        = formula
  ))
}

# formula helper functions
.remove_response        <- function(formula){
  # removes response from the expression
  # (prevents crash on formula evaluations)
  if(attr(stats::terms(formula), "response")  == 1){
    formula[2] <- NULL
  }
  return(formula)
}
.has_expression         <- function(formula){
  # check if there is any expression in the formula
  return(any(grepl("expression\\(", deparse(formula))))
}
.extract_expressions    <- function(formula){
  # extract all expressions from the formula

  # Convert the formula to a character string
  formula_string <- deparse(formula)

  # Use a regex to find all instances of "expression(...)"
  matches     <- gregexpr("expression\\(.*?\\)", formula_string)
  expressions <- regmatches(formula_string, matches)[[1]]

  # Use a regex to remove "expression(" and the closing ")"
  expressions <- lapply(expressions, .clean_from_expression)

  return(expressions)
}
.clean_from_expression  <- function(x){
  # expression to character

  return(sub("expression\\((.*)\\)", "\\1", x))
}
.remove_expressions     <- function(formula){
  # remove all expressions from the formula

  # Convert the formula to a character string
  formula_string <- paste0(deparse(formula), collapse = " ")

  # Use a regex to remove all instances of "+ expression(...)" or "expression(...) +", considering spaces and newlines
  formula_string_clean <- gsub("\\+\\s*expression\\(.*?\\)\\s*", "", formula_string)
  formula_string_clean <- gsub("\\s*expression\\(.*?\\)\\s*\\+", "", formula_string_clean)

  # Handle the case where the expression is the first term in the formula
  formula_string_clean <- gsub("^\\s*expression\\(.*?\\)\\s*", "", formula_string_clean)

  # Handle the case where the formula reduces to just "y ~ expression(...)"
  if(grepl("^\\s*[a-zA-Z0-9._]+\\s*~\\s*expression\\(.*?\\)\\s*$", formula_string_clean)){
    formula_string_clean <- gsub("expression\\(.*?\\)", "1", formula_string_clean)
  }

  # Reconvert the cleaned string back to a formula
  return(stats::as.formula(formula_string_clean))
}
.has_random_effects     <- function(formula){
  # Convert the formula to a character string
  formula_str <- paste(deparse(formula), collapse = " ")

  # Regular expression to match `( ... | ... )` patterns
  has_random <- grepl("\\([^\\)]+\\|[^\\)]+\\)", formula_str)

  # Return TRUE if at least one match is found, otherwise FALSE
  return(has_random)
}
.extract_random_effects <- function(formula) {
  # Convert the formula to a character string
  formula_str <- paste(deparse(formula), collapse = " ")

  # Regular expression to match `( ... | ... )` or `( ... || ... )` patterns
  random_effects <- gregexpr("\\([^\\)]+\\|{1,2}[^\\)]+\\)", formula_str)

  # Extract matches
  matches <- regmatches(formula_str, random_effects)

  # Clean up the parentheses and remove unnecessary spaces
  clean_matches <- lapply(unlist(matches), function(x) gsub("^\\(|\\)$", "", x))  # Remove outer parentheses
  clean_matches <- lapply(clean_matches, trimws)  # Remove extra spaces from each match

  # Add random effects information
  for (i in seq_along(clean_matches)) {
    # Extract the grouping factor (right-hand side of | or ||)
    grouping_factor <- trimws(sub(".*\\|{1,2}\\s*", "", clean_matches[[i]]))
    attr(clean_matches[[i]], "grouping_factor") <- grouping_factor

    # Detect whether independent `||` or correlated `|` random effects are used
    independent <- grepl("\\|\\|", clean_matches[[i]])  # Check if `||` is used
    attr(clean_matches[[i]], "independent") <- independent
  }

  # Return the cleaned random effects as a list
  return(as.list(clean_matches))
}

.remove_random_effects  <- function(formula){
  # Convert the formula to a character string
  formula_str <- paste(deparse(formula), collapse = " ")

  # Regular expression to match and remove `( ... | ... )` patterns
  cleaned_formula <- gsub("\\+?\\s*\\([^\\)]+\\|[^\\)]+\\)", "", formula_str)

  # Normalize spacing around '+' and remove any leading '+'
  cleaned_formula <- gsub("\\s*\\+\\s*", " + ", cleaned_formula)  # Normalize '+' spacing
  cleaned_formula <- gsub("^\\s*~\\s*\\+\\s*", "~ ", cleaned_formula)  # Remove leading '+'

  # Ensure no excessive spaces remain
  cleaned_formula <- gsub("\\s{2,}", " ", cleaned_formula)  # Replace multiple spaces with a single space
  cleaned_formula <- trimws(cleaned_formula)  # Trim leading/trailing whitespace

  # Ensure at least "1" remains if formula is empty after cleaning
  if (grepl("^\\s*~\\s*$", cleaned_formula)) {
    cleaned_formula <- "~ 1"
  }

  # Return as a formula
  return(stats::as.formula(cleaned_formula))
}
.get_grouping_factor    <- function(x){
  has_grouping            <- grepl("\\|", x)
  grouping                <- rep("", length(x))
  grouping[has_grouping]  <- trimws(sub(".*\\|\\s*", "", x[has_grouping]))
  return(grouping)
}
.remove_grouping_factor <- function(formula){
  return(trimws(sub("\\|.*$", "", formula)))
}
#' @title Add an Intercept to a Formula
#'
#' @description Converts a no-intercept formula to the corresponding formula
#' with an intercept while preserving the formula environment. Top-level
#' no-intercept encodings such as \code{- 1}, \code{+ 0}, and \code{0 +} are
#' removed without editing transformed calls such as \code{I(x - 1)} or
#' \code{offset(x - 1)}.
#'
#' @param formula a formula object.
#'
#' @return A formula object with an intercept.
#'
#' @export
formula_add_intercept <- function(formula){

  if(!inherits(formula, "formula")){
    stop("'formula' must be a formula.", call. = FALSE)
  }

  if(attr(stats::terms(formula), "intercept") == 1L){
    return(formula)
  }

  formula_env   <- environment(formula)
  formula_attrs <- attributes(formula)
  rhs_index     <- if(length(formula) == 3L) 3L else 2L
  rhs           <- .formula_strip_no_intercept(formula[[rhs_index]])

  if(is.null(rhs)){
    rhs <- 1
  }

  out <- formula
  out[[rhs_index]] <- rhs
  environment(out) <- formula_env

  for(attribute in setdiff(names(formula_attrs), c("class", ".Environment", "names"))){
    attr(out, attribute) <- formula_attrs[[attribute]]
  }

  if(attr(stats::terms(out), "intercept") == 0L){
    out[[rhs_index]] <- call("+", 1, out[[rhs_index]])
    environment(out) <- formula_env
  }

  return(out)
}
.add_intercept_to_formula <- formula_add_intercept

.formula_strip_no_intercept <- function(expr){

  if(.formula_is_no_intercept_additive_term(expr)){
    return(NULL)
  }

  if(is.call(expr) && identical(expr[[1L]], as.name("+")) && length(expr) == 3L){
    lhs <- .formula_strip_no_intercept(expr[[2L]])
    rhs <- .formula_strip_no_intercept(expr[[3L]])

    if(is.null(lhs)){
      return(rhs)
    }
    if(is.null(rhs)){
      return(lhs)
    }
    return(call("+", lhs, rhs))
  }

  if(is.call(expr) && identical(expr[[1L]], as.name("-")) && length(expr) == 3L &&
     .formula_is_numeric_constant(expr[[3L]], 1)){
    return(.formula_strip_no_intercept(expr[[2L]]))
  }

  return(expr)
}

.formula_is_no_intercept_additive_term <- function(expr){

  .formula_is_numeric_constant(expr, 0) || .formula_is_negative_one(expr)
}

.formula_is_negative_one <- function(expr){

  (is.numeric(expr) && length(expr) == 1L && identical(as.numeric(expr), -1)) ||
    (is.call(expr) && identical(expr[[1L]], as.name("-")) && length(expr) == 2L &&
       .formula_is_numeric_constant(expr[[2L]], 1))
}

.formula_is_numeric_constant <- function(expr, value){

  if(is.call(expr) && identical(expr[[1L]], as.name("(")) && length(expr) == 2L){
    return(.formula_is_numeric_constant(expr[[2L]], value))
  }

  is.numeric(expr) && length(expr) == 1L && identical(as.numeric(expr), as.numeric(value))
}

#' @title Evaluate JAGS formula using posterior samples
#'
#' @description Evaluates a JAGS formula on a posterior distribution
#' obtained from a fitted model.
#'
#' @param fit model fitted with either \link[runjags]{runjags} posterior
#' samples obtained with \link[rjags]{rjags-package}
#' @param formula formula specifying the right hand side of the assignment (the
#' left hand side is ignored). If the formula has a \code{"log(intercept)"}
#' attribute set to \code{TRUE}, the intercept values will be log-transformed
#' before computing the linear predictor.
#' @param parameter name of the parameter created with the formula
#' @param data data.frame containing predictors included in the formula
#' @param prior_list named list of prior distribution of parameters specified within
#' the \code{formula}
#'
#'
#' @return \code{JAGS_evaluate_formula} returns a matrix of the evaluated posterior samples on
#' the supplied data.
#'
#' @seealso [JAGS_fit()] [JAGS_formula()]
#' @export
JAGS_evaluate_formula <- function(fit, formula, parameter, data, prior_list){

  if(!is.language(formula))
    stop("'formula' must be a formula")
  if(!is.data.frame(data))
    stop("'data' must be a data.frame")
  check_char(parameter, "parameter")
  check_list(prior_list, "prior_list")
  if(any(!sapply(prior_list, is.prior)))
    stop("'prior_list' must be a list of priors.")

  # extract the posterior distribution
  posterior <- as.matrix(.fit_to_posterior(fit))

  # remove the specified response (would crash the model.frame if not included)
  formula <- .remove_response(formula)
  log_intercept <- isTRUE(attr(formula, "log(intercept)"))
  if(attr(stats::terms(formula), "intercept") == 0){
    formula <- formula_add_intercept(formula)
    if(log_intercept){
      attr(formula, "log(intercept)") <- TRUE
    }
  }

  # select priors corresponding to the prior distribution
  prior_parameter <- sapply(prior_list, function(p) if(is.null(attr(p, "parameter"))) "__none" else attr(p, "parameter"))
  if(!any(parameter %in% unique(prior_parameter)))
    stop("The specified parameter '", parameter, "' was not used in any of the prior distributions.")
  prior_list_formula <- prior_list[prior_parameter == parameter]
  names(prior_list_formula) <- format_parameter_names(names(prior_list_formula), formula_parameters = parameter, formula_prefix = FALSE)

  # extract the terms information from the formula
  formula_terms    <- stats::terms(formula)
  has_intercept    <- attr(formula_terms, "intercept") == 1
  predictors       <- as.character(attr(formula_terms, "variables"))[-1]
  model_terms      <- c(if(has_intercept) "intercept", attr(formula_terms, "term.labels"))

  # check that all predictors have data and prior distribution
  if(!all(predictors %in% colnames(data)))
    stop(paste0("The ", paste0("'", predictors[!predictors %in% colnames(data)], "'", collapse = ", ")," predictor variable is missing in the data."))
  if(!all(model_terms %in% names(prior_list_formula)))
    stop(paste0("The prior distribution for the ", paste0("'", predictors[!model_terms %in% format_parameter_names(names(prior_list_formula), formula_parameters = parameter, formula_prefix = FALSE)], "'", collapse = ", ")," term is missing in the prior_list."))

  # obtain predictors characteristics -- based on prior distributions used to fit the original model
  # (i.e., do not truest the supplied data -- probably passed by the user)
  model_terms_type <- sapply(model_terms, function(model_term){
    if(model_term == "intercept"){
      return("continuous")
    }else if(is.prior.factor(prior_list_formula[[model_term]]) || inherits(prior_list_formula[[model_term]], "prior.factor_mixture") || inherits(prior_list_formula[[model_term]], "prior.factor_spike_and_slab")){
      return("factor")
    }else if(is.prior.simple(prior_list_formula[[model_term]]) || inherits(prior_list_formula[[model_term]], "prior.simple_mixture") || inherits(prior_list_formula[[model_term]], "prior.simple_spike_and_slab")){
      return("continuous")
    } else {
      stop(paste0("Unrecognized prior distribution for the '", model_term, "' term."))
    }
  })
  predictors_type <- model_terms_type[predictors]

  # check that passed data correspond to the specified priors (factor levels etc...) and set the proper contrasts
  if(any(predictors_type == "factor")){

    # check the proper data input for each factor prior
    for(factor in names(predictors_type[predictors_type == "factor"])){

      # select the corresponding prior in the variable
      this_prior <- prior_list_formula[[factor]]

      if(is.factor(data[,factor])){
        if(all(levels(data[,factor]) %in% .get_prior_factor_level_names(this_prior))){
          # either the formatting is correct, or the supplied levels are a subset of the original levels
          # reformat to check ordering and etc...
          data[,factor] <- factor(data[,factor], levels = .get_prior_factor_level_names(this_prior))
        }else{
          # there are some additional levels
          stop(paste0("Levels specified in the '", factor, "' factor variable do not match the levels used for model specification."))
        }
      }else if(all(unique(data[,factor]) %in% .get_prior_factor_level_names(this_prior))){
        # the variable was not passed as a factor but the values matches the factor levels
        data[,factor] <- factor(data[,factor], levels = .get_prior_factor_level_names(this_prior))
      }else{
        # there are some additional mismatching values
        stop(paste0("Levels specified in the '", factor, "' factor variable do not match the levels used for model specification."))
      }

      # set the contrast
      if(is.prior.orthonormal(this_prior)){
        stats::contrasts(data[,factor]) <- "contr.orthonormal"
      }else if(is.prior.meandif(this_prior)){
        stats::contrasts(data[,factor]) <- "contr.meandif"
      }else if(is.prior.independent(this_prior)){
        stats::contrasts(data[,factor]) <- "contr.independent"
      }else if(is.prior.treatment(this_prior)){
        stats::contrasts(data[,factor]) <- "contr.treatment"
      }
    }
  }
  if(any(predictors_type == "continuous")){

    # check the proper data input for each continuous prior
    for(continuous in names(predictors_type[predictors_type == "continuous"])){

      # select the corresponding prior in the variable
      this_prior <- prior_list_formula[[continuous]]

      if(is.prior.factor(this_prior)|| is.prior.discrete(this_prior) || is.prior.PET(this_prior) || is.prior.PEESE(this_prior) || is.prior.weightfunction(this_prior)){
        stop(paste0("Unsupported prior distribution defined for '", continuous, "' continuous variable. See '?prior' for details."))
      }
    }

    # apply scaling if predictors were scaled during model fitting
    formula_scale <- attr(fit, "formula_scale")
    if(!is.null(formula_scale)){
      # Handle nested structure: formula_scale[[parameter]] contains the scaling info
      param_scale <- formula_scale[[parameter]]
      if(!is.null(param_scale)){
        for(continuous in names(predictors_type[predictors_type == "continuous"])){
          # check if this predictor was scaled (with parameter prefix)
          scaled_name <- paste0(parameter, "_", continuous)
          if(scaled_name %in% names(param_scale)){
            # apply the same scaling transformation
            scale_info <- param_scale[[scaled_name]]
            data[, continuous] <- (data[, continuous] - scale_info$mean) / scale_info$sd
          }
        }
      }
    }
  }

  # get the design matrix
  model_frame  <- stats::model.frame(formula, data = data)
  model_matrix <- stats::model.matrix(model_frame, formula = formula, data = data)

  ### evaluate the design matrix on the samples -> output[data, posterior]
  if(has_intercept){

    terms_indexes    <- attr(model_matrix, "assign") + 1
    terms_indexes[1] <- 0

    # check for scaling factors
    temp_multiply_by <- .get_parameter_scaling_factor_matrix(term = "intercept", prior_list = prior_list_formula, posterior = posterior, nrow = nrow(data), ncol = nrow(posterior))

    # get intercept values and apply log() transformation if log(intercept) attribute is set
    if(is.prior.point(prior_list_formula[["intercept"]])){
      intercept_values <- rep(
        prior_list_formula[["intercept"]]$parameters[["location"]],
        nrow(posterior)
      )
    }else{
      intercept_values <- posterior[, JAGS_parameter_names("intercept", formula_parameter = parameter)]
    }
    if(log_intercept){
      intercept_values <- log(intercept_values)
    }
    output           <- temp_multiply_by * matrix(intercept_values, nrow = nrow(data), ncol = nrow(posterior), byrow = TRUE)

  }else{

    terms_indexes    <- attr(model_matrix, "assign")
    output           <- matrix(0, nrow = nrow(data), ncol = nrow(posterior))

  }

  # add remaining terms (omitting the intercept indexed as NA)
  for(i in unique(terms_indexes[terms_indexes > 0])){

    # subset the model matrix
    temp_data <- model_matrix[,terms_indexes == i,drop = FALSE]

    # get the posterior (unless point prior was used)
    if(is.prior.point(prior_list_formula[[model_terms[i]]])){
      temp_posterior <- matrix(
        prior_list_formula[[model_terms[i]]]$parameters[["location"]],
        nrow = nrow(posterior),
        ncol = if(model_terms_type[i] == "factor") .get_prior_factor_levels(prior_list_formula[[model_terms[i]]]) else 1
      )
    }else{
      temp_posterior <- posterior[,paste0(
        JAGS_parameter_names(model_terms[i], formula_parameter = parameter),
        if(model_terms_type[i] == "factor" && .get_prior_factor_levels(prior_list_formula[[model_terms[i]]]) > 1) paste0("[", 1:.get_prior_factor_levels(prior_list_formula[[model_terms[i]]]), "]"))
        ,drop = FALSE]
    }

    # check for scaling factors
    temp_multiply_by <- .get_parameter_scaling_factor_matrix(term = model_terms[i], prior_list = prior_list_formula, posterior = posterior, nrow = nrow(data), ncol = nrow(posterior))

    output <- output + temp_multiply_by * (temp_data %*% t(temp_posterior))

  }

  return(output)
}


.get_parameter_scaling_factor_matrix <- function(term, prior_list, posterior, nrow, ncol){

  if(!is.null(attr(prior_list[[term]], "multiply_by"))){
    if(is.numeric(attr(prior_list[[term]], "multiply_by"))){
      temp_multiply_by <- matrix(attr(prior_list[[term]], "multiply_by"), nrow = nrow, ncol = ncol)
    }else{
      temp_multiply_by <- matrix(posterior[,JAGS_parameter_names(attr(prior_list[[term]], "multiply_by"))], nrow = nrow, ncol = ncol, byrow = TRUE)
    }
  }else{
    temp_multiply_by <- matrix(1, nrow = nrow, ncol = ncol)
  }

  return(temp_multiply_by)
}

.factor_level_list <- function(x){

  level_names <- attr(x, "level_names")
  if(is.null(level_names)){
    factor_terms <- attr(x, "factor_terms")
    if(!is.null(factor_terms) && length(factor_terms) > 1){
      return(NULL)
    }

    n_levels <- attr(x, "levels")
    if(is.null(n_levels)){
      return(NULL)
    }

    if(is.prior.factor(x)){
      level_names <- .get_prior_factor_level_names(x)
    }else if(isTRUE(attr(x, "independent"))){
      level_names <- seq_len(n_levels)
    }else{
      level_names <- seq_len(n_levels + 1)
    }
  }

  if(is.list(level_names)){
    factor_terms <- attr(x, "factor_terms")
    if(is.null(factor_terms)){
      factor_terms <- names(level_names)
    }
    if(is.null(factor_terms) || any(!nzchar(factor_terms))){
      factor_terms <- paste0("factor", seq_along(level_names))
    }
    level_names <- level_names[factor_terms]
    names(level_names) <- factor_terms
  }else{
    factor_terms <- attr(x, "factor_terms")
    if(is.null(factor_terms) || length(factor_terms) != 1){
      factor_terms <- ".factor"
    }
    level_names <- setNames(list(level_names), factor_terms)
  }

  level_names <- lapply(level_names, as.character)
  return(level_names)
}

.factor_cell_grid <- function(level_names){

  if(is.null(level_names) || length(level_names) == 0){
    return(data.frame())
  }

  expand.grid(level_names, KEEP.OUT.ATTRS = FALSE, stringsAsFactors = FALSE)
}

.factor_cell_labels <- function(level_names){

  level_grid <- .factor_cell_grid(level_names)
  if(ncol(level_grid) == 0){
    return(character(0))
  }

  if(ncol(level_grid) == 1){
    return(as.character(level_grid[[1]]))
  }

  apply(level_grid, 1, function(level_row) {
    paste0(names(level_row), "=", unname(level_row), collapse = ", ")
  })
}

.factor_contrast_parameter_names <- function(parameter, level_names, cell_names){

  if(is.null(level_names)){
    return(paste0(parameter, "[dif: ", cell_names, "]"))
  }

  level_grid <- .factor_cell_grid(level_names)
  if(nrow(level_grid) != length(cell_names)){
    return(paste0(parameter, "[dif: ", cell_names, "]"))
  }

  if(length(level_names) == 1){
    return(paste0(parameter, "[dif: ", level_grid[[1]], "]"))
  }

  parameter_terms <- strsplit(parameter, "__xXx__", fixed = TRUE)[[1]]
  factor_terms <- names(level_names)
  factor_positions <- vapply(factor_terms, function(factor_term) {
    factor_position <- which(
      parameter_terms == factor_term |
        endsWith(parameter_terms, paste0("_", factor_term))
    )

    if(length(factor_position) == 1){
      return(factor_position)
    }

    return(NA_integer_)
  }, integer(1))

  if(!all(!is.na(factor_positions))){
    return(paste0(parameter, "[dif: ", cell_names, "]"))
  }

  vapply(seq_len(nrow(level_grid)), function(level_i) {
    formatted_terms <- parameter_terms

    for(factor_term in factor_terms){
      factor_position <- factor_positions[[factor_term]]
      prefix_length <- nchar(formatted_terms[[factor_position]]) - nchar(factor_term)
      formatted_terms[[factor_position]] <- paste0(
        substr(formatted_terms[[factor_position]], 1, prefix_length),
        factor_term,
        "[dif: ",
        level_grid[[factor_term]][level_i],
        "]"
      )
    }

    paste0(formatted_terms, collapse = "__xXx__")
  }, character(1))
}

.factor_object_contrast_name <- function(x){

  if(is.prior.independent(x) || isTRUE(attr(x, "independent"))){
    return("contr.independent")
  }else if(is.prior.treatment(x) || isTRUE(attr(x, "treatment"))){
    return("contr.treatment")
  }else if(is.prior.orthonormal(x) || isTRUE(attr(x, "orthonormal"))){
    return("contr.orthonormal")
  }else if(is.prior.meandif(x) || isTRUE(attr(x, "meandif"))){
    return("contr.meandif")
  }

  return(NULL)
}

.factor_contrast_matrix <- function(level_names, contrast){

  switch(
    contrast,
    "contr.treatment"   = stats::contr.treatment(level_names),
    "contr.independent" = contr.independent(level_names),
    "contr.orthonormal" = contr.orthonormal(level_names),
    "contr.meandif"     = contr.meandif(level_names),
    stop("Unsupported factor contrast '", contrast, "'.", call. = FALSE)
  )
}

.add_factor_metadata_from_named_objects <- function(x, parameter, objects){

  level_names <- .factor_level_list(x)
  if(is.null(level_names)){
    return(x)
  }

  factor_terms <- names(level_names)
  if(is.null(attr(x, "factor_terms"))){
    attr(x, "factor_terms") <- factor_terms
  }

  factor_contrasts <- attr(x, "factor_contrasts")
  if(is.null(factor_contrasts)){
    factor_contrasts <- rep(NA_character_, length(factor_terms))
    names(factor_contrasts) <- factor_terms
  }else{
    factor_contrasts <- as.character(factor_contrasts)
    if(is.null(names(factor_contrasts))){
      names(factor_contrasts) <- factor_terms[seq_along(factor_contrasts)]
    }
    factor_contrasts <- factor_contrasts[factor_terms]
  }

  formula_parameter <- attr(x, "formula_parameter")
  if(is.null(formula_parameter)){
    formula_parameter <- attr(x, "parameter")
    if(!is.null(formula_parameter) && identical(formula_parameter, parameter)){
      formula_parameter <- NULL
    }
  }
  if(is.null(formula_parameter) && grepl("_", parameter, fixed = TRUE)){
    formula_parameter <- sub("_.*$", "", parameter)
  }

  for(factor_term in factor_terms[is.na(factor_contrasts)]){
    candidates <- factor_term
    if(!is.null(formula_parameter)){
      candidates <- c(paste0(formula_parameter, "_", factor_term), candidates)
    }
    candidates <- unique(candidates)
    candidates <- candidates[candidates %in% names(objects)]

    for(candidate in candidates){
      contrast <- .factor_object_contrast_name(objects[[candidate]])
      if(!is.null(contrast)){
        factor_contrasts[[factor_term]] <- contrast
        break
      }
    }
  }

  if(any(is.na(factor_contrasts)) && length(factor_terms) == 1){
    fallback_contrast <- .factor_object_contrast_name(x)
    if(!is.null(fallback_contrast)){
      factor_contrasts[is.na(factor_contrasts)] <- fallback_contrast
    }
  }

  attr(x, "factor_contrasts") <- factor_contrasts
  return(x)
}

.factor_term_design_from_formula <- function(formula, data, predictors, predictors_type, term_index, term_components, factor_terms, has_intercept){

  level_names <- lapply(factor_terms, function(factor_term) levels(data[[factor_term]]))
  names(level_names) <- factor_terms
  cell_grid <- .factor_cell_grid(level_names)

  grid_data <- data[rep(1, nrow(cell_grid)), predictors, drop = FALSE]
  rownames(grid_data) <- NULL

  for(predictor in predictors){
    if(predictors_type[[predictor]] == "factor"){
      predictor_values <- if(predictor %in% factor_terms){
        cell_grid[[predictor]]
      }else{
        rep(levels(data[[predictor]])[1], nrow(cell_grid))
      }
      grid_data[[predictor]] <- factor(predictor_values, levels = levels(data[[predictor]]))
      stats::contrasts(grid_data[[predictor]]) <- attr(data[[predictor]], "contrasts")
    }else{
      grid_data[[predictor]] <- if(predictor %in% term_components) 1 else 0
    }
  }

  grid_model_frame <- stats::model.frame(formula, data = grid_data)
  grid_model_matrix <- stats::model.matrix(grid_model_frame, formula = formula, data = grid_data)
  grid_terms_indexes <- attr(grid_model_matrix, "assign")
  if(has_intercept){
    grid_terms_indexes <- grid_terms_indexes + 1
    grid_terms_indexes[1] <- 0
  }

  design <- grid_model_matrix[, grid_terms_indexes == term_index, drop = FALSE]

  return(list(
    design     = unname(design),
    cell_grid  = cell_grid,
    cell_names = .factor_cell_labels(level_names),
    level_names = level_names
  ))
}

.factor_term_design_from_metadata <- function(x){

  factor_design <- attr(x, "factor_design")
  if(!is.null(factor_design)){
    factor_design <- as.matrix(factor_design)
    cell_names <- attr(x, "factor_cell_names")
    level_names <- .factor_level_list(x)
    if(is.null(cell_names)){
      if(is.null(level_names)){
        stop("Factor level names are missing and the factor contrast cannot be transformed.", call. = FALSE)
      }
      cell_names <- .factor_cell_labels(level_names)
    }
    return(list(
      design     = factor_design,
      cell_names = cell_names,
      level_names = level_names
    ))
  }

  level_names <- .factor_level_list(x)
  if(is.null(level_names)){
    stop("Factor level names are missing and the factor contrast cannot be transformed.", call. = FALSE)
  }

  factor_terms <- names(level_names)
  factor_contrasts <- attr(x, "factor_contrasts")
  if(is.null(factor_contrasts)){
    fallback_contrast <- .factor_object_contrast_name(x)
    if(is.null(fallback_contrast) || length(factor_terms) > 1){
      stop("Factor contrast metadata is missing and cannot be inferred.", call. = FALSE)
    }
    factor_contrasts <- rep(fallback_contrast, length(factor_terms))
    names(factor_contrasts) <- factor_terms
  }else{
    factor_contrasts <- as.character(factor_contrasts)
    if(is.null(names(factor_contrasts))){
      names(factor_contrasts) <- factor_terms[seq_along(factor_contrasts)]
    }
    factor_contrasts <- factor_contrasts[factor_terms]
  }

  if(any(is.na(factor_contrasts))){
    stop("Factor contrast metadata is incomplete and cannot be inferred.", call. = FALSE)
  }

  contrast_matrices <- lapply(factor_terms, function(factor_term) {
    .factor_contrast_matrix(level_names[[factor_term]], factor_contrasts[[factor_term]])
  })
  names(contrast_matrices) <- factor_terms

  level_grid <- expand.grid(lapply(level_names, seq_along), KEEP.OUT.ATTRS = FALSE)
  coef_grid <- expand.grid(lapply(contrast_matrices, function(contrast_matrix) seq_len(ncol(contrast_matrix))), KEEP.OUT.ATTRS = FALSE)

  design <- matrix(NA_real_, nrow = nrow(level_grid), ncol = nrow(coef_grid))
  for(row_i in seq_len(nrow(level_grid))){
    for(col_i in seq_len(nrow(coef_grid))){
      design[row_i, col_i] <- prod(vapply(factor_terms, function(factor_term) {
        contrast_matrices[[factor_term]][level_grid[[factor_term]][row_i], coef_grid[[factor_term]][col_i]]
      }, numeric(1)))
    }
  }

  return(list(
    design     = design,
    cell_names = .factor_cell_labels(level_names),
    level_names = level_names
  ))
}

.transform_factor_contrast_samples <- function(coefficient_samples, metadata, parameter, transformed_class){

  if(!is.matrix(coefficient_samples)){
    coefficient_samples <- matrix(coefficient_samples, ncol = 1)
  }

  design_info <- .factor_term_design_from_metadata(metadata)
  design <- design_info[["design"]]

  if(ncol(coefficient_samples) != ncol(design)){
    stop(
      "The factor contrast design for '", parameter, "' has ", ncol(design),
      " coefficient columns, but the samples contain ", ncol(coefficient_samples), ".",
      call. = FALSE
    )
  }

  transformed_samples <- coefficient_samples %*% t(design)
  colnames(transformed_samples) <- .factor_contrast_parameter_names(
    parameter = parameter,
    level_names = design_info[["level_names"]],
    cell_names = design_info[["cell_names"]]
  )

  old_attributes <- attributes(coefficient_samples)
  old_class <- class(coefficient_samples)
  old_attributes <- old_attributes[!names(old_attributes) %in% c("dim", "dimnames", "names", "class", "level_names")]
  attributes(transformed_samples) <- c(attributes(transformed_samples), old_attributes)
  attr(transformed_samples, "level_names")       <- design_info[["cell_names"]]
  attr(transformed_samples, "factor_cell_names") <- design_info[["cell_names"]]
  class(transformed_samples) <- unique(c(old_class, class(transformed_samples), transformed_class))

  return(transformed_samples)
}

#' @title Transform factor posterior samples into differences from the mean
#'
#' @description Transforms posterior samples from model-averaged posterior
#' distributions based on meandif/orthonormal prior distributions into differences from
#' the mean.
#'
#' @param samples (a list) of mixed posterior distributions created with
#' \code{mix_posteriors} function
#'
#' @return \code{transform_meandif_samples} returns a named list of mixed posterior
#' distributions (either a vector of matrix).
#'
#' @seealso [mix_posteriors] [transform_meandif_samples] [transform_meandif_samples] [transform_orthonormal_samples]
#'
#' @export
transform_factor_samples <- function(samples){

  check_list(samples, "samples", allow_NULL = TRUE)

  samples <- transform_meandif_samples(samples)
  samples <- transform_orthonormal_samples(samples)

  return(samples)
}

#' @title Transform meandif posterior samples into differences from the mean
#'
#' @description Transforms posterior samples from model-averaged posterior
#' distributions based on meandif prior distributions into differences from
#' the mean.
#'
#' @param samples (a list) of mixed posterior distributions created with
#' \code{mix_posteriors} function
#'
#' @return \code{transform_meandif_samples} returns a named list of mixed posterior
#' distributions (either a vector of matrix).
#'
#' @seealso [mix_posteriors] [contr.meandif]
#'
#' @export
transform_meandif_samples <- function(samples){

  check_list(samples, "samples", allow_NULL = TRUE)

  for(i in seq_along(samples)){
    if(!inherits(samples[[i]],"mixed_posteriors.meandif_transformed") && inherits(samples[[i]], "mixed_posteriors.factor") && isTRUE(attr(samples[[i]], "meandif"))){

      meandif_samples <- .add_factor_metadata_from_named_objects(samples[[i]], names(samples)[i], samples)
      samples[[i]] <- .transform_factor_contrast_samples(
        coefficient_samples = meandif_samples,
        metadata            = meandif_samples,
        parameter           = names(samples)[i],
        transformed_class   = "mixed_posteriors.meandif_transformed"
      )
    }
  }

  return(samples)
}

#' @title Transform orthonomal posterior samples into differences from the mean
#'
#' @description Transforms posterior samples from model-averaged posterior
#' distributions based on orthonormal prior distributions into differences from
#' the mean.
#'
#' @param samples (a list) of mixed posterior distributions created with
#' \code{mix_posteriors} function
#'
#' @return \code{transform_orthonormal_samples} returns a named list of mixed posterior
#' distributions (either a vector of matrix).
#'
#' @seealso [mix_posteriors] [contr.orthonormal]
#'
#' @export
transform_orthonormal_samples <- function(samples){

  check_list(samples, "samples", allow_NULL = TRUE)

  for(i in seq_along(samples)){
    if(!inherits(samples[[i]],"mixed_posteriors.orthonormal_transformed") && inherits(samples[[i]], "mixed_posteriors.factor") && isTRUE(attr(samples[[i]], "orthonormal"))){

      orthonormal_samples <- .add_factor_metadata_from_named_objects(samples[[i]], names(samples)[i], samples)
      samples[[i]] <- .transform_factor_contrast_samples(
        coefficient_samples = orthonormal_samples,
        metadata            = orthonormal_samples,
        parameter           = names(samples)[i],
        transformed_class   = "mixed_posteriors.orthonormal_transformed"
      )
    }
  }

  return(samples)
}

# not part of transform factor samples (as it's usefull only for marginal effects)
transform_treatment_samples <- function(samples){

  check_list(samples, "samples", allow_NULL = TRUE)

  for(i in seq_along(samples)){
    if(!inherits(samples[[i]],"mixed_posteriors.treatment_transformed") && inherits(samples[[i]], "mixed_posteriors.factor") && isTRUE(attr(samples[[i]], "treatment"))){

      treatment_samples <- .add_factor_metadata_from_named_objects(samples[[i]], names(samples)[i], samples)
      samples[[i]] <- .transform_factor_contrast_samples(
        coefficient_samples = treatment_samples,
        metadata            = treatment_samples,
        parameter           = names(samples)[i],
        transformed_class   = "mixed_posteriors.treatment_transformed"
      )
    }
  }

  return(samples)
}


# Helper: Parse a term name into its component variable names
# e.g., "mu_x1__xXx__x2" with prefix "mu" -> c("x1", "x2")
# e.g., "mu_intercept" -> character(0) (intercept has no components)
# e.g., "mu_x1" -> c("x1")
.parse_term_components <- function(term_name, prefix) {
  # Remove prefix
  term_part <- sub(paste0("^", prefix, "_"), "", term_name)

  # Check if it's the intercept
  if (term_part == "intercept") {
    return(character(0))
  }

  # Indexed factor terms attach the index to the full term name. Strip the
  # trailing index so scaled continuous components inside interactions are
  # still detected correctly (e.g., alloc__xXx__year[1] -> alloc, year).
  term_part <- sub("\\[[^]]+\\]$", "", term_part)

  # Split by interaction separator
  components <- strsplit(term_part, "__xXx__", fixed = TRUE)[[1]]
  return(components)
}


# Helper: Parse a term into scaled components and an unscaled identity string
# used to determine which coefficients can contribute to each other during
# de-standardization.
.parse_unscale_term_structure <- function(term_name, prefix, scaled_vars) {

  term_part <- sub(paste0("^", prefix, "_"), "", term_name)

  if (term_part == "intercept") {
    return(list(
      components  = character(0),
      scaled      = character(0),
      unscaled_id = ""
    ))
  }

  term_index <- ""
  if (grepl("\\[[^]]+\\]$", term_part)) {
    term_index <- sub("^.*(\\[[^]]+\\])$", "\\1", term_part)
    term_core  <- sub("\\[[^]]+\\]$", "", term_part)
  } else {
    term_core <- term_part
  }

  components <- strsplit(term_core, "__xXx__", fixed = TRUE)[[1]]
  is_scaled  <- components %in% scaled_vars

  unscaled_components <- components[!is_scaled]
  unscaled_id <- paste(unscaled_components, collapse = "__xXx__")

  if (nzchar(term_index) && length(unscaled_components) > 0) {
    unscaled_id <- paste0(unscaled_id, term_index)
  }

  return(list(
    components  = components,
    scaled      = components[is_scaled],
    unscaled_id = unscaled_id
  ))
}


# Helper: Check if set A is a subset of set B (including equality)
.is_subset <- function(A, B) {

  length(A) == 0 || all(A %in% B)
}


# Helper: Compare lower-order unscaled term identities. Two-level factor main
# effects are stored without [1], while their interactions are indexed.
.unscale_ids_match <- function(target_id, source_id) {

  identical(target_id, source_id) ||
    (nzchar(target_id) && !grepl("\\[[^]]+\\]$", target_id) && identical(paste0(target_id, "[1]"), source_id))
}


# Helper: Validate the nested formula_scale structure used for unscaling.
.check_formula_scale_info <- function(formula_scale, name = "formula_scale") {

  check_list(formula_scale, name)

  if(is.null(names(formula_scale)) || anyNA(names(formula_scale)) || any(names(formula_scale) == ""))
    stop(paste0("The '", name, "' argument must be a named nested list keyed by parameter name."), call. = FALSE)

  for(param_name in names(formula_scale)){
    param_scale <- formula_scale[[param_name]]
    check_list(param_scale, paste0(name, "[['", param_name, "]]"))

    if(length(param_scale) == 0)
      next

    if(is.null(names(param_scale)) || anyNA(names(param_scale)) || any(names(param_scale) == ""))
      stop(paste0("The '", name, "[['", param_name, "]]" ,"' entry must be a named list keyed by parameter term."), call. = FALSE)

    for(term_name in names(param_scale)){
      term_scale <- param_scale[[term_name]]

      check_list(
        term_scale,
        paste0(name, "[['", param_name, "]][['", term_name, "]]"),
        check_names = c("mean", "sd"),
        all_objects = TRUE,
        allow_other = TRUE
      )
      check_real(
        term_scale[["mean"]],
        paste0(name, "[['", param_name, "]][['", term_name, "]][['mean'] ]"),
        allow_NA = FALSE
      )
      check_real(
        term_scale[["sd"]],
        paste0(name, "[['", param_name, "]][['", term_name, "]][['sd'] ]"),
        lower = 0,
        allow_bound = FALSE,
        allow_NA = FALSE
      )
    }
  }

  invisible(NULL)
}


# Helper: warn when formula_scale entries do not map to posterior terms.
.warn_unused_formula_scale_terms <- function(term_names, formula_scale, prefix) {

  scaled_terms <- names(formula_scale)
  if(length(scaled_terms) == 0)
    return(invisible(NULL))

  scaled_vars <- sub(paste0("^", prefix, "_"), "", scaled_terms)
  term_components <- unique(unlist(lapply(term_names, .parse_term_components, prefix = prefix), use.names = FALSE))
  unused_terms <- scaled_terms[!scaled_vars %in% term_components]

  if(length(unused_terms) == 0)
    return(invisible(NULL))

  if(length(unused_terms) == length(scaled_terms)){
    warning(
      paste0(
        "Ignoring all formula_scale[['", prefix, "]] entries because none match posterior terms: '",
        paste0(unused_terms, collapse = "', '"),
        "'. Samples for this parameter prefix are returned unchanged."
      ),
      call. = FALSE,
      immediate. = TRUE
    )
  }else{
    warning(
      paste0(
        "Ignoring unused formula_scale[['", prefix, "]] entries: '",
        paste0(unused_terms, collapse = "', '"),
        "'. Matched entries are still applied."
      ),
      call. = FALSE,
      immediate. = TRUE
    )
  }

  invisible(NULL)
}


# Helper: Build the transformation matrix for unscaling coefficients
#
# For each target term T and source term S, computes the coefficient M[T,S] such that:
#   coef_orig[T] = sum over S of M[T,S] * coef_z[S]
#
# The formula is based on expanding products of (x_i - mu_i)/sigma_i terms.
# For S to contribute to T:
#   1. T_unscaled == S_unscaled (unscaled components must match exactly)
#   2. T_scaled is a subset of S_scaled
#
# The contribution is: (-1)^|extra| * prod(mu_extra) / prod(sigma_S_scaled)
# where extra = S_scaled \ T_scaled
#
# @param term_names Character vector of all term names in the posterior
# @param formula_scale Named list with scaling info (mean, sd) for scaled predictors
# @param prefix The parameter prefix (e.g., "mu")
# @return A square transformation matrix
.build_unscale_matrix <- function(term_names, formula_scale, prefix) {

  n_terms <- length(term_names)
  M <- diag(n_terms)  # Start with identity matrix
  rownames(M) <- colnames(M) <- term_names

  # Extract the variable names that are scaled (without prefix)
  scaled_vars <- sub(paste0("^", prefix, "_"), "", names(formula_scale))

  # Parse all terms into their scaled components and unscaled identity.
  term_structure <- lapply(
    term_names,
    .parse_unscale_term_structure,
    prefix = prefix,
    scaled_vars = scaled_vars
  )
  names(term_structure) <- term_names

  term_components <- lapply(term_structure, `[[`, "components")
  term_scaled <- lapply(term_structure, `[[`, "scaled")
  term_unscaled <- vapply(term_structure, `[[`, character(1), "unscaled_id")

  # Warn about high-order interactions
  max_order <- max(sapply(term_components, length))
  if (max_order >= 5) {
    warning("Model contains ", max_order, "-way or higher interactions. ",
            "Unscaling transformation may be computationally intensive.",
            immediate. = TRUE)
  }

  # Build the transformation matrix
  for (t_idx in seq_along(term_names)) {
    T_name <- term_names[t_idx]
    T_scaled <- term_scaled[[T_name]]
    T_unscaled <- term_unscaled[[T_name]]

    for (s_idx in seq_along(term_names)) {
      S_name <- term_names[s_idx]
      S_scaled <- term_scaled[[S_name]]
      S_unscaled <- term_unscaled[[S_name]]

      # Check contribution conditions
      # 1. Unscaled parts must match exactly
      if (!.unscale_ids_match(T_unscaled, S_unscaled)) next

      # 2. T_scaled must be a subset of S_scaled
      if (!.is_subset(T_scaled, S_scaled)) next

      # 3. S must have at least one scaled component (otherwise no transformation needed)
      if (length(S_scaled) == 0) {
        # No scaling for this source term - keep identity (already set)
        next
      }

      # Compute the coefficient
      extra_scaled <- setdiff(S_scaled, T_scaled)

      # Sign: (-1)^|extra|
      sign <- (-1)^length(extra_scaled)

      # Product of means for extra scaled components
      if (length(extra_scaled) > 0) {
        extra_params <- paste0(prefix, "_", extra_scaled)
        mean_product <- prod(sapply(extra_params, function(p) formula_scale[[p]]$mean))
      } else {
        mean_product <- 1
      }

      # Product of SDs for all scaled components in S
      S_scaled_params <- paste0(prefix, "_", S_scaled)
      sd_product <- prod(sapply(S_scaled_params, function(p) formula_scale[[p]]$sd))

      # Contribution coefficient
      M[t_idx, s_idx] <- sign * mean_product / sd_product
    }
  }

  return(M)
}


# Helper: Apply unscaling transformation to a matrix of posterior samples
#
# @param posterior Matrix with samples in rows, parameters in columns
# Apply the unscaling transformation to posterior samples
#
# @param posterior Matrix of posterior samples with parameter names as column names
# @param formula_scale Nested list with scaling info keyed by parameter name:
#   list(mu = list(mu_x1 = list(mean, sd)), log_sigma = list(log_sigma_x = list(mean, sd)))
# @return Transformed posterior matrix
.apply_unscale_transform <- function(posterior, formula_scale) {

  if (is.null(formula_scale) || length(formula_scale) == 0) {
    return(posterior)
  }

  # Handle nested structure: iterate over each parameter
  matched_prefix <- FALSE
  for (param_name in names(formula_scale)) {
    param_scale <- formula_scale[[param_name]]
    affected_cols <- grep(paste0("^", param_name, "_"), colnames(posterior), value = TRUE)
    if(length(affected_cols) == 0)
      next

    matched_prefix <- TRUE
    posterior <- .apply_unscale_transform_single(posterior, param_scale, prefix = param_name)
  }

  if(!matched_prefix){
    warning(
      "Ignoring formula_scale because none of its parameter prefixes match the posterior columns.",
      call. = FALSE,
      immediate. = TRUE
    )
  }

  return(posterior)
}

# Helper: Apply unscaling for a single parameter's predictors
# @param posterior Matrix of posterior samples
# @param formula_scale Flat list of scaling info: list(mu_x1 = list(mean, sd), mu_x2 = list(mean, sd))
# @param prefix Parameter prefix (e.g., "mu")
# @return Transformed posterior matrix
.apply_unscale_transform_single <- function(posterior, formula_scale, prefix) {
  
  if (is.null(formula_scale) || length(formula_scale) == 0) {
    return(posterior)
  }

  # Check if this parameter uses log(intercept)
  log_intercept <- isTRUE(attr(formula_scale, "log_intercept"))
  intercept_col <- paste0(prefix, "_intercept")

  # Identify which columns are affected by the transformation
  affected_cols <- grep(paste0("^", prefix, "_"), colnames(posterior), value = TRUE)
  random_sd_cols <- grep(paste0("^", prefix, "__xREx__"), affected_cols, value = TRUE)
  fixed_cols     <- setdiff(affected_cols, random_sd_cols)

  if (length(affected_cols) == 0) {
    return(posterior)
  }

  if(length(fixed_cols) > 0){
    .warn_unused_formula_scale_terms(fixed_cols, formula_scale, prefix)
  }

  # For log(intercept): transform to log scale before unscaling, then exp() back
  # This works because: log_sigma = log(intercept) + beta * x_z
  # is equivalent to: log_sigma = log_int + beta * x_z (standard additive form)
  # where log_int = log(intercept)
  if (length(fixed_cols) > 0 && log_intercept && intercept_col %in% colnames(posterior)) {
    posterior[, intercept_col] <- log(posterior[, intercept_col])
  }

  # Build and apply standard transformation matrix
  if(length(fixed_cols) > 0){
    M <- .build_unscale_matrix(fixed_cols, formula_scale, prefix)
    posterior[, fixed_cols] <- posterior[, fixed_cols, drop = FALSE] %*% t(M)
  }

  # Transform intercept back from log scale
  if (length(fixed_cols) > 0 && log_intercept && intercept_col %in% colnames(posterior)) {
    posterior[, intercept_col] <- exp(posterior[, intercept_col])
  }

  posterior <- .apply_random_sd_unscale(posterior, random_sd_cols, formula_scale, prefix)

  return(posterior)
}

.apply_random_sd_unscale <- function(posterior, random_sd_cols, formula_scale, prefix){

  if(length(random_sd_cols) == 0){
    return(posterior)
  }

  term_map <- .random_sd_term_map(random_sd_cols, formula_scale, prefix)
  if(length(term_map) == 0){
    return(posterior)
  }

  random_sd_cols <- names(term_map)
  group_keys <- vapply(random_sd_cols, .random_sd_group_key,
                       character(1), term_map = term_map, prefix = prefix)

  for(group_key in unique(group_keys)){
    group_cols <- random_sd_cols[group_keys == group_key]
    group_terms <- unname(term_map[group_cols])

    if(any(duplicated(group_terms))){
      next
    }

    pseudo_terms <- paste0(prefix, "_", group_terms)
    names(pseudo_terms) <- group_cols
    M <- .build_unscale_matrix(unname(pseudo_terms), formula_scale, prefix)

    source_sd <- posterior[, group_cols, drop = FALSE]
    transformed_sd <- matrix(NA_real_, nrow = nrow(source_sd), ncol = ncol(source_sd))
    colnames(transformed_sd) <- group_cols

    for(target_i in seq_along(group_cols)){
      transformed_var <- rowSums(t(t(source_sd^2) * (M[target_i, ]^2)))
      transformed_sd[, target_i] <- sqrt(transformed_var)
    }

    posterior[, group_cols] <- transformed_sd
  }

  posterior
}
.random_sd_term_map <- function(random_sd_cols, formula_scale, prefix){

  metadata <- attr(formula_scale, "random_effect_terms")
  if(!is.null(metadata) && length(metadata) > 0){
    base_cols <- sub("\\[[^]]+\\]$", "", random_sd_cols)
    term_map <- metadata[base_cols]
    names(term_map) <- random_sd_cols
    term_map <- term_map[!is.na(term_map)]
    if(length(term_map) > 0){
      return(term_map)
    }
  }

  scaled_vars <- sub(paste0("^", prefix, "_"), "", names(formula_scale))
  possible_terms <- c("intercept", scaled_vars)
  names(possible_terms) <- possible_terms

  term_map <- vapply(random_sd_cols, function(col){
    rest <- sub(paste0("^", prefix, "__xREx__"), "", sub("\\[[^]]+\\]$", "", col))
    candidates <- possible_terms[vapply(possible_terms, function(term){
      endsWith(rest, paste0("_", term))
    }, logical(1))]
    if(length(candidates) == 0){
      return(NA_character_)
    }
    candidates[which.max(nchar(candidates))]
  }, character(1))

  term_map[!is.na(term_map)]
}
.random_sd_group_key <- function(col, term_map, prefix){

  base_col <- sub("\\[[^]]+\\]$", "", col)
  rest <- sub(paste0("^", prefix, "__xREx__"), "", base_col)
  term <- unname(term_map[[col]])
  sub(paste0("_", term, "$"), "", rest)
}


#' @title Transform standardized posterior samples back to original scale
#'
#' @description Transforms posterior samples from standardized continuous
#' predictors back to the original scale. This function is used when predictors
#' were standardized during model fitting via the \code{formula_scale} parameter.
#'
#' @param fit a fitted model object with \code{formula_scale} attribute, or
#' a matrix of posterior samples
#' @param formula_scale nested list containing standardization information keyed by
#' parameter name. Each parameter entry contains scaling info (mean and sd) for
#' each standardized predictor, e.g., \code{list(mu = list(mu_x1 = list(mean = 0, sd = 1)))}.
#' If \code{fit} is provided and has a \code{formula_scale} attribute, this will be used automatically.
#'
#' @details The function transforms regression coefficients and intercepts
#' to account for predictor standardization using a combinatorial approach that
#' correctly handles interactions of any order.
#'
#' For a k-way interaction between standardized predictors, the expansion of
#' \eqn{\prod_{i} (x_i - \mu_i)/\sigma_i} contributes to all lower-order terms.
#' The contribution to a target term T from a source term S (where T is a subset
#' of S's scaled components) is:
#' \deqn{(-1)^{|extra|} \cdot \prod_{i \in extra} \mu_i / \prod_{i \in S_{scaled}} \sigma_i}
#' where \eqn{extra = S_{scaled} \setminus T_{scaled}}.
#'
#' @return \code{transform_scale_samples} returns posterior samples transformed
#' back to the original predictor scale.
#'
#' @seealso [JAGS_formula()] [JAGS_fit()]
#'
#' @export
transform_scale_samples <- function(fit, formula_scale = NULL){

  # extract formula_scale from fit if available
  if(is.null(formula_scale) && !is.null(attr(fit, "formula_scale"))){
    formula_scale <- attr(fit, "formula_scale")
  }

  if(is.null(formula_scale) || length(formula_scale) == 0){
    # no scaling information, return as is
    return(fit)
  }

  .check_formula_scale_info(formula_scale)

  # extract posterior samples
  if(inherits(fit, "runjags") || inherits(fit, "BayesTools_fit")){
    posterior <- as.matrix(.fit_to_posterior(fit))
  }else if(is.matrix(fit)){
    posterior <- fit
  }else{
    stop("'fit' must be a fitted model object or a matrix of posterior samples.")
  }

  # Apply the combinatorial unscaling transformation
  posterior <- .apply_unscale_transform(posterior, formula_scale)

  return(posterior)
}


#' @title Transform prior samples to original scale
#'
#' @description Generate prior samples and transform them using the same
#' matrix transformation as posterior samples. This is the correct approach for
#' visualizing priors on the original (unscaled) scale, especially for the intercept
#' which depends on contributions from multiple coefficient priors.
#'
#' @param fit a fitted model object with \code{prior_list} and optionally
#' \code{formula_scale} attributes
#' @param n_samples number of samples to generate (default: 10000)
#' @param seed random seed for reproducibility (optional)
#' @param formula_scale optional nested list containing standardization information.
#' If not provided, extracted from \code{fit} attribute.
#'
#' @details When models use auto-scaling (standardizing predictors), the posterior
#' samples are on the standardized scale. To correctly visualize priors on the
#' original scale, we cannot simply apply a linear transformation to individual
#' priors because the intercept on the original scale is a weighted sum of
#' multiple priors:
#'
#' \deqn{\beta_0^{orig} = \beta_0^* - \sum_i \frac{\mu_i}{\sigma_i} \beta_i^*}
#'

#' This function generates samples from ALL priors simultaneously and applies
#' the same matrix transformation used for posterior samples, which correctly
#' handles the intercept and all other parameters.
#'
#' @return A matrix of prior samples on the original (unscaled) scale, with
#' columns matching the structure of posterior samples.
#'
#' @seealso [transform_scale_samples()] [plot_posterior()]
#'
#' @examples
#' # With a fitted model that used formula_scale:
#' # prior_samples <- transform_prior_samples(fit, n_samples = 10000)
#' # This can then be used with density() or for custom plotting
#'
#' @export
transform_prior_samples <- function(fit, n_samples = 10000, seed = NULL, formula_scale = NULL){

  check_int(n_samples, "n_samples", lower = 1)
  check_int(seed, "seed", allow_NULL = TRUE)

  # Extract prior_list from fit

  prior_list <- attr(fit, "prior_list")

  if(is.null(prior_list)){
    stop("'fit' must have 'prior_list' attribute.")
  }

  # Extract formula_scale from fit if not provided
  if(is.null(formula_scale)){
    formula_scale <- attr(fit, "formula_scale")
  }

  # Get posterior column names for structure matching
  if(inherits(fit, "runjags") || inherits(fit, "BayesTools_fit")){
    posterior <- as.matrix(.fit_to_posterior(fit))
  }else{
    stop("'fit' must be a fitted model object.")
  }

  prior_samples <- .generate_transformed_prior_samples(
    prior_list   = prior_list,
    column_names = colnames(posterior),
    n_samples    = n_samples,
    seed         = seed,
    formula_scale = formula_scale
  )

  return(prior_samples)
}


# Helper: Generate prior samples and apply the same unscaling transform used
# for posterior samples.
#
# @param prior_list Named list of prior objects
# @param column_names Column names to match from the posterior structure
# @param n_samples Number of samples to generate
# @param seed Optional random seed
# @param formula_scale Optional nested scaling information
# @return Matrix with transformed prior samples
.generate_transformed_prior_samples <- function(prior_list, column_names, n_samples, seed = NULL, formula_scale = NULL){

  if(!is.null(formula_scale) && length(formula_scale) > 0){
    .check_formula_scale_info(formula_scale)
  }

  prior_samples <- .generate_prior_sample_matrix(
    prior_list     = prior_list,
    n_samples      = n_samples,
    column_names   = column_names,
    seed           = seed
  )

  if(!is.null(formula_scale) && length(formula_scale) > 0){
    prior_samples <- .apply_unscale_transform(prior_samples, formula_scale)
  }

  return(prior_samples)
}

.generate_factor_prior_sample_matrix <- function(prior, parameter, n_samples){

  K <- .get_prior_factor_levels(prior)
  if(is.null(K) || is.na(K)){
    stop("The number of factor coefficients for prior '", parameter, "' is unknown.", call. = FALSE)
  }

  if(is.prior.spike_and_slab(prior)){
    prior_variable  <- .get_spike_and_slab_variable(prior)
    prior_inclusion <- .get_spike_and_slab_inclusion(prior)
    samples <- .generate_factor_prior_sample_matrix(prior_variable, parameter, n_samples)
    inclusion <- stats::rbinom(n_samples, size = 1, prob = rng(prior_inclusion, n_samples))
    samples <- samples * inclusion
  }else if(is.prior.mixture(prior)){
    prior_weights <- attr(prior, "prior_weights")
    prior_weights <- prior_weights / sum(prior_weights)
    components <- sample(seq_along(prior_weights), size = n_samples, replace = TRUE, prob = prior_weights)
    samples <- matrix(NA_real_, nrow = n_samples, ncol = K)

    for(component in unique(components)){
      samples[components == component, ] <- .generate_factor_prior_sample_matrix(
        prior      = prior[[component]],
        parameter  = parameter,
        n_samples  = sum(components == component)
      )
    }
  }else if(is.prior.point(prior)){
    location <- prior$parameters[["location"]]
    samples <- matrix(rep(location, length.out = K), nrow = n_samples, ncol = K, byrow = TRUE)
  }else if(is.prior.orthonormal(prior) || is.prior.meandif(prior)){
    prior$parameters[["K"]] <- K
    samples <- rng(prior, n_samples, transform_factor_samples = FALSE)
    samples <- matrix(samples, nrow = n_samples, ncol = K)
  }else if(is.prior.treatment(prior) || is.prior.independent(prior)){
    samples <- replicate(K, rng(prior, n_samples, transform_factor_samples = FALSE))
    samples <- matrix(samples, nrow = n_samples, ncol = K)
  }else{
    samples <- rng(prior, n_samples, transform_factor_samples = FALSE)
    samples <- matrix(samples, nrow = n_samples, ncol = K)
  }

  colnames(samples) <- .JAGS_prior_factor_names(parameter, prior)
  return(samples)
}


# Helper: Generate a matrix of prior samples matching posterior structure
#
# @param prior_list Named list of prior objects
# @param n_samples Number of samples to generate
# @param column_names Optional vector of column names to match (filters output)
# @param seed Optional random seed
# @return Matrix with prior samples (rows = samples, columns = parameters)
.generate_prior_sample_matrix <- function(prior_list, n_samples, column_names = NULL, seed = NULL){

  if(!is.null(seed)){
    set.seed(seed)
  }

  # Determine which parameters to sample
  param_names <- names(prior_list)

  if(is.null(param_names) || length(param_names) == 0){
    stop("'prior_list' must be a named list of priors.")
  }

  # Initialize list to collect samples (handles varying column counts per prior)
  samples_list <- list()

  for(param_name in param_names){
    prior <- prior_list[[param_name]]

    if(is.null(prior)){
      # No prior for this parameter - use zeros
      samples_list[[param_name]] <- matrix(0, nrow = n_samples, ncol = 1)
      colnames(samples_list[[param_name]]) <- param_name

    }else if(is.prior.none(prior)){
      # No effect prior - use zeros
      samples_list[[param_name]] <- matrix(0, nrow = n_samples, ncol = 1)
      colnames(samples_list[[param_name]]) <- param_name

    }else if(is.prior.factor(prior) || inherits(prior, "prior.factor_mixture") || inherits(prior, "prior.factor_spike_and_slab")){
      samples_list[[param_name]] <- .generate_factor_prior_sample_matrix(prior, param_name, n_samples)

    }else if(is.prior.point(prior)){
      # Point prior - constant values
      samples_list[[param_name]] <- matrix(
        prior$parameters[["location"]],
        nrow = n_samples,
        ncol = 1
      )
      colnames(samples_list[[param_name]]) <- param_name

    }else if(is.prior.simple(prior)){
      # Simple priors - single column
      samples_list[[param_name]] <- matrix(
        rng(prior, n_samples),
        nrow = n_samples,
        ncol = 1
      )
      colnames(samples_list[[param_name]]) <- param_name

    }else if(is.prior.vector(prior)){
      # Vector priors return matrix from rng
      temp_samples <- rng(prior, n_samples)
      if(is.matrix(temp_samples)){
        n_cols <- ncol(temp_samples)
        col_names <- paste0(param_name, "[", 1:n_cols, "]")
        colnames(temp_samples) <- col_names
        samples_list[[param_name]] <- temp_samples
      }else{
        samples_list[[param_name]] <- matrix(temp_samples, nrow = n_samples, ncol = 1)
        colnames(samples_list[[param_name]]) <- param_name
      }

    }else{
      # Fallback for other prior types - try rng
      temp_samples <- tryCatch(
        rng(prior, n_samples),
        error = function(e){
          warning(sprintf("Could not generate samples for prior '%s': %s. Using zeros.",
                          param_name, e$message))
          rep(0, n_samples)
        }
      )

      if(is.matrix(temp_samples)){
        n_cols <- ncol(temp_samples)
        col_names <- paste0(param_name, "[", 1:n_cols, "]")
        colnames(temp_samples) <- col_names
        samples_list[[param_name]] <- temp_samples
      }else{
        samples_list[[param_name]] <- matrix(temp_samples, nrow = n_samples, ncol = 1)
        colnames(samples_list[[param_name]]) <- param_name
      }
    }
  }

  # Combine all samples into one matrix
  samples <- do.call(cbind, samples_list)

  # Filter to match column_names if provided
  if(!is.null(column_names)){
    available_cols <- intersect(column_names, colnames(samples))
    if(length(available_cols) > 0){
      samples <- samples[, available_cols, drop = FALSE]
    }
  }

  return(samples)
}


#' @title BayesTools Contrast Matrices
#'
#' @description BayesTools provides several contrast matrix functions for Bayesian factor analysis.
#' These functions create different types of contrast matrices that can be used with factor
#' variables in Bayesian models.
#'
#' @details
#' The package includes the following contrast functions:
#' \describe{
#'   \item{\code{contr.orthonormal}}{Return a matrix of orthonormal contrasts.
#'     Code is based on \code{stanova::contr.bayes} and corresponding to description
#'     by \insertCite{rouder2012default;textual}{BayesTools}. Returns a matrix with n rows and
#'     k columns, with k = n - 1 if \code{contrasts = TRUE} and k = n if \code{contrasts = FALSE}.}
#'   \item{\code{contr.meandif}}{Return a matrix of mean difference contrasts.
#'     This is an adjustment to the \code{contr.orthonormal} that ascertains that the prior
#'     distributions on difference between the gran mean and factor level are identical independent
#'     of the number of factor levels (which does not hold for the orthonormal contrast). Furthermore,
#'     the contrast is re-scaled so the specified prior distribution exactly corresponds to the prior
#'     distribution on difference between each factor level and the grand mean -- this is approximately
#'     twice the scale of \code{contr.orthonormal}. Returns a matrix with n rows and k columns,
#'     with k = n - 1 if \code{contrasts = TRUE} and k = n if \code{contrasts = FALSE}.}
#'   \item{\code{contr.independent}}{Return a matrix of independent contrasts -- a level for each term.
#'     Returns a matrix with n rows and k columns, with k = n if \code{contrasts = TRUE} and k = n
#'     if \code{contrasts = FALSE}.}
#' }
#'
#' @param n a vector of levels for a factor, or the number of levels
#' @param contrasts logical indicating whether contrasts should be computed
#'
#' @examples
#' # Orthonormal contrasts
#' contr.orthonormal(c(1, 2))
#' contr.orthonormal(c(1, 2, 3))
#'
#' # Mean difference contrasts
#' contr.meandif(c(1, 2))
#' contr.meandif(c(1, 2, 3))
#'
#' # Independent contrasts
#' contr.independent(c(1, 2))
#' contr.independent(c(1, 2, 3))
#'
#' @references
#' \insertAllCited{}
#'
#' @aliases contr.orthonormal contr.meandif contr.independent
#' @name contr.BayesTools
NULL

#' @rdname contr.BayesTools
#' @export
contr.orthonormal <- function(n, contrasts = TRUE){
  # based on: stanova::contr.bayes
  if(length(n) <= 1L){
    if(is.numeric(n) && length(n) == 1L && n > 1L){
      return(TRUE)
    }else{
      stop("Not enough degrees of freedom to define contrasts.")
    }
  }else{
    n <- length(n)
  }

  cont <- diag(n)
  if(contrasts){
    a       <- n
    I_a     <- diag(a)
    J_a     <- matrix(1, nrow = a, ncol = a)
    Sigma_a <- I_a - J_a/a
    cont    <- eigen(Sigma_a)$vectors[, seq_len(a - 1), drop = FALSE]
  }

  return(cont)
}

#' @rdname contr.BayesTools
#' @export
contr.meandif <- function(n, contrasts = TRUE){

  if(length(n) <= 1L){
    if(is.numeric(n) && length(n) == 1L && n > 1L){
      return(TRUE)
    }else{
      stop("Not enough degrees of freedom to define contrasts.")
    }
  }else{
    n <- length(n)
  }

  cont <- diag(n)
  if(contrasts){
    a       <- n
    I_a     <- diag(a)
    J_a     <- matrix(1, nrow = a, ncol = a)
    Sigma_a <- I_a - J_a/a
    cont    <- eigen(Sigma_a)$vectors[, seq_len(a - 1), drop = FALSE]
    cont    <- cont / (sqrt(1 - 1/n))
  }

  return(cont)
}


#' @rdname contr.BayesTools
#' @export
contr.independent <- function(n, contrasts = TRUE){

  if(length(n) <= 1L){
    if(is.numeric(n) && length(n) == 1L && n >= 1L){
      return(TRUE)
    }else{
      stop("Not enough degrees of freedom to define contrasts.")
    }
  }else{
    n <- length(n)
  }

  cont <- diag(x = 1, nrow = n, ncol = n)

  return(cont)
}


#' @title Clean parameter names from JAGS
#'
#' @description Removes additional formatting from parameter names outputted from
#' JAGS.
#'
#' @param parameters a vector of parameter names
#' @param formula_parameter a formula parameter prefix name
#' @param formula_parameters a vector of formula parameter prefix names
#' @param formula_random a vector of random effects grouping factors
#' @param formula_prefix whether the \code{formula_parameters} names should be
#' kept. Defaults to \code{TRUE}.
#' @param formula_scale optional nested list containing scaling info. When provided,
#' intercepts from parameters with \code{log_intercept = TRUE} attribute will be
#' renamed to \code{exp(intercept)}.
#'
#' @examples
#' format_parameter_names(c("mu_x_cont", "mu_x_fac3t", "mu_x_fac3t__xXx__x_cont"),
#'                        formula_parameters = "mu")
#'
#' @return A character vector with reformatted parameter names.
#'
#' @export format_parameter_names
#' @export JAGS_parameter_names
#' @name parameter_names
NULL

#' @rdname parameter_names
format_parameter_names <- function(parameters, formula_parameters = NULL, formula_random = NULL, formula_prefix = TRUE, formula_scale = NULL){

  check_char(parameters, "parameters", check_length = FALSE)
  check_char(formula_random, "formula_random", check_length = FALSE, allow_NULL = TRUE)
  check_char(formula_parameters, "formula_parameters", check_length = FALSE, allow_NULL = TRUE)
  check_bool(formula_prefix, "formula_prefix")
  check_list(formula_scale, "formula_scale", allow_NULL = TRUE)

  # rename intercept to exp(intercept) for parameters with log_intercept attribute
  if(!is.null(formula_scale)){
    for(param_name in names(formula_scale)){
      if(isTRUE(attr(formula_scale[[param_name]], "log_intercept"))){
        intercept_name <- paste0(param_name, "_intercept")
        if(intercept_name %in% parameters){
          parameters[parameters == intercept_name] <- paste0(param_name, "_exp(intercept)")
        }
      }
    }
  }

  for(i in seq_along(formula_parameters)){
    parameters[grep(paste0(formula_parameters[i], "_"), parameters)] <- gsub(
      paste0(formula_parameters[i], "_"),
      if(formula_prefix) paste0("(", formula_parameters[i], ") ") else "",
      parameters[grep(paste0(formula_parameters[i], "_"), parameters)])
  }

  for(i in seq_along(formula_random)){
    temp_which <- grepl(paste0("_xREx__", formula_random[i], "_"), parameters)
    temp_incl  <- grepl("(inclusion)", parameters)
    parameters[temp_which] <- gsub(
      paste0("_xREx__", formula_random[i], "_"),
      "",
      parameters[temp_which]
    )
    if(any(temp_which &  temp_incl)){
      parameters[temp_which &  temp_incl] <- paste0(gsub("(inclusion)", "", parameters[temp_which & temp_incl], fixed = TRUE), "|", formula_random[i], " (inclusion)")
    }
    if(any(temp_which & !temp_incl)){
      parameters[temp_which & !temp_incl] <- paste0("sd(", parameters[temp_which & !temp_incl], "|", formula_random[i], ")")
    }
  }

  parameters[grep("__xXx__", parameters)] <- gsub("__xXx__", ":", parameters[grep("__xXx__", parameters)])

  return(parameters)
}
#' @rdname parameter_names
JAGS_parameter_names   <- function(parameters, formula_parameter = NULL){

  check_char(parameters, "parameters", check_length = FALSE)
  check_char(formula_parameter, "formula_parameter", check_length = TRUE, allow_NULL = TRUE)

  if(!is.null(formula_parameter)){
    parameters <- paste0(formula_parameter, "_", parameters)
  }
  parameters <- gsub(":", "__xXx__", parameters)

  return(parameters)
}

.JAGS_prior_factor_names <- function(parameter, prior){

  levels <- .get_prior_factor_levels(prior)
  if((is.null(levels) || length(levels) == 0L || is.na(levels)) && "K" %in% names(prior[["parameters"]])){
    levels <- prior[["parameters"]][["K"]]
  }
  if(is.null(levels) || length(levels) == 0L || is.na(levels)){
    stop("Factor-prior dimensions must be available before constructing JAGS parameter names.", call. = FALSE)
  }

  if(levels == 1){
    par_names <- parameter
  }else{
    par_names <- paste0(parameter, "[", 1:levels, "]")
  }

  return(par_names)
}

#' @title Model-average marginal posterior distributions
#'
#' @description Creates marginal model-averages posterior distributions for a given
#' parameter based on model-averaged posterior samples and parameter name
#' (and formula with at specification).
#'
#' @param samples model-averaged posterior samples created by \code{mix_posteriors()}
#' @param parameter parameter of interest
#' @param formula model formula (needs to be specified if \code{parameter} was part of a formula)
#' @param at named list with predictor levels of the formula for which marginalization
#' should be performed. If a predictor level is missing, \code{0} is used for continuous
#' predictors, the baseline factor level is used for factors with \code{contrast = "treatment"} prior
#' distributions, and the parameter is completely omitted for for factors with \code{contrast = "meandif"},
#' @param prior_samples whether marginal prior distributions should be generated
#' \code{contrast = "orthonormal"}, and \code{contrast = "independent"} levels
#' @param use_formula whether the parameter should be evaluated as a part of supplied formula
#' @param n_samples controls the numerical grid used for model-averaged
#' prior densities
#' @inheritParams density.prior
#'
#' @return \code{marginal_posterior} returns a named list of mixed marginal posterior
#' distributions (either a vector of matrix).
#'#'
#' @export
marginal_posterior <- function(samples, parameter, formula = NULL, at = NULL, prior_samples = FALSE, use_formula = TRUE,
                               transformation = NULL, transformation_arguments = NULL, transformation_settings = FALSE,
                               n_samples = 10000, ...){

  check_list(samples, "samples")
  if(!inherits(samples, "mixed_posteriors"))
    stop("'samples' must be a be an object generated by 'mix_posteriors' function.")
  check_char(parameter, "parameter", allow_values = names(samples))
  if(!is.null(formula) && !is.language(formula))
    stop("'formula' must be a formula")
  if(!is.null(at) && !is.list(at))
    stop("'at' must be a list")
  check_bool(prior_samples, "prior_samples")
  check_bool(use_formula, "use_formula")
  .check_transformation_input(transformation, transformation_arguments, transformation_settings)


  # deal formula vs non-formula marginal posterior
  if(use_formula && inherits(samples[[parameter]], "mixed_posteriors.formula")){

      # remove the specified response (would crash the model.frame if not included)
      formula <- .remove_response(formula)
      formula_parameter <- attr(samples[[parameter]], "formula_parameter")

      ### extract the terms information from the formula
      formula_terms          <- stats::terms(formula)
      has_intercept          <- attr(formula_terms, "intercept") == 1
      predictors             <- as.character(attr(formula_terms, "variables"))[-1]
      model_terms            <- c(if(has_intercept) "intercept", attr(formula_terms, "term.labels"))

      JAGS_model_terms <- JAGS_parameter_names(parameters = model_terms, formula_parameter = formula_parameter)
      JAGS_predictors  <- JAGS_parameter_names(parameters = predictors, formula_parameter = formula_parameter)


      ### obtain posterior samples and check that all are present
      if(!all(JAGS_model_terms %in% names(samples)))
        stop(paste0("The posterior samples for the ", paste0("'", JAGS_model_terms[!JAGS_model_terms %in% names(samples)], "'", collapse = ", ")," term is missing in the samples."))


      ### obtain prior list and information
      prior_list <- lapply(names(samples), function(model_term) attr(samples[[model_term]], "prior_list"))
      names(prior_list) <- names(samples)

      # get parameter information
      priors_info <- lapply(names(prior_list), function(model_term) list(
        term              = model_term,
        intercept         = model_term == "intercept",
        factor            = inherits(samples[[model_term]], "mixed_posteriors.factor"),
        levels            = attr(samples[[model_term]], "levels"),
        level_names       = attr(samples[[model_term]], "level_names"),
        interaction       = attr(samples[[model_term]], "interaction"),
        interaction_terms = attr(samples[[model_term]], "interaction_terms"),
        treatment         = attr(samples[[model_term]], "treatment"),
        independent       = attr(samples[[model_term]], "independent"),
        orthonormal       = attr(samples[[model_term]], "orthonormal"),
        meandif           = attr(samples[[model_term]], "meandif")
        ))
      names(priors_info) <- names(prior_list)
      model_terms_type <- sapply(JAGS_model_terms, function(model_term){
        if(priors_info[[model_term]][["factor"]]){
          return("factor")
        }else{
          return("continuous")
        }
      })
      predictors_type  <- model_terms_type[JAGS_parameter_names(parameters = predictors, formula_parameter = formula_parameter)]


      ### prepare at specification
      # in case of an interaction, all levels need to be set
      if(!is.null(priors_info[[parameter]][["interaction"]]) && priors_info[[parameter]][["interaction"]]){
        at_manipulated <- JAGS_parameter_names(priors_info[[parameter]][["interaction_terms"]], formula_parameter = formula_parameter)
      }else{
        at_manipulated <- parameter
      }

      if(!all(names(at) %in% predictors))
        stop(paste0("The following values passed via the 'at' argument do not correspond to the specified model: ", paste0("'", names(at)[!names(at) %in% predictors], "'", collapse = ", ")))
      if(any(format_parameter_names(at_manipulated, formula_parameters = formula_parameter, formula_prefix = FALSE) %in% names(at)))
        stop("Values of the parameter of interested cannot be specified via the 'at' argument.")

      # fill in with default values if needed
      for(i in seq_along(predictors)){
        if(JAGS_predictors[i] %in% at_manipulated){
          # specify levels for the parameter of interest
          if(model_terms_type[[JAGS_predictors[i]]] == "continuous"){
            at[[predictors[i]]] <- c(-1, 0, 1)
          }else{
            at[[predictors[i]]] <- priors_info[[JAGS_predictors[i]]][["level_names"]]
          }
        }else if(is.null(at[[predictors[i]]])){
          # specify levels for the remaining parameters
          if(model_terms_type[[JAGS_predictors[i]]] == "continuous"){
            # fill in zeroes for unspecified continuous predictors
            at[[predictors[i]]] <- 0
          }else if(priors_info[[JAGS_predictors[i]]][["treatment"]]){
            # fill in the default category for unspecified treatment factors
            at[[predictors[i]]] <- priors_info[[JAGS_predictors[i]]][["level_names"]][1]
          }else{
            # fill in NA for any other factor type
            at[[predictors[i]]] <- NA
          }
        }
      }

      # transform to a data.frame
      data <- as.data.frame(expand.grid(at))

      # check the specified data
      if(any(predictors_type == "factor")){

        # check the proper data input for each factor variable
        for(i in seq_along(predictors_type)[predictors_type == "factor"]){

          if(is.factor(data[,predictors[i]])){
            if(all(levels(data[,predictors[i]]) %in% priors_info[[JAGS_predictors[i]]][["level_names"]])){
              # either the formatting is correct, or the supplied levels are a subset of the original levels
              # reformat to check ordering and etc...
              data[,predictors[i]] <- factor(data[,predictors[i]], levels = priors_info[[JAGS_predictors[i]]][["level_names"]])
            }else{
              # there are some additional levels
              stop(paste0("Levels specified in the '", predictors[i], "' factor variable do not match the levels used for model specification."))
            }
          }else if(all(stats::na.omit(unique(data[,predictors[i]])) %in% priors_info[[JAGS_predictors[i]]][["level_names"]])){
            # the variable was not passed as a factor but the values matches the factor levels
            data[,predictors[i]] <- factor(data[,predictors[i]], levels = priors_info[[JAGS_predictors[i]]][["level_names"]])
          }else{
            # there are some additional mismatching values
            stop(paste0("Levels specified in the '", predictors[i], "' factor variable do not match the levels used for model specification."))
          }

          # set the contrast
          if(priors_info[[JAGS_predictors[i]]][["orthonormal"]]){
            stats::contrasts(data[,predictors[i]]) <- "contr.orthonormal"
          }else if(priors_info[[JAGS_predictors[i]]][["meandif"]]){
            stats::contrasts(data[,predictors[i]]) <- "contr.meandif"
          }else if(priors_info[[JAGS_predictors[i]]][["independent"]]){
            stats::contrasts(data[,predictors[i]]) <- "contr.independent"
          }else if(priors_info[[JAGS_predictors[i]]][["treatment"]]){
            stats::contrasts(data[,predictors[i]]) <- "contr.treatment"
            if(anyNA(data[,predictors[i]]))
              stop("Unspecified levels in the '", predictors[i], "' factor (NAs not allowed for 'treatment' factors).")
          }
        }
      }
      if(any(predictors_type == "continuous")){

        # check the proper data input for each continuous variable
        for(i in seq_along(predictors_type)[predictors_type == "continuous"]){
          if(anyNA(data[,predictors[i]]))
            stop("Unspecified levels in the '", predictors[i], "' variable (NAs not allowed for continuous variables).")
          if(!is.numeric(data[,predictors[i]]))
            stop("Nonnumeric values in the '", predictors[i], "' continuous variable.")
        }
      }


      ### get the design matrix
      model_frame  <- stats::model.frame(formula, data = data, na.action = NULL)
      model_matrix <- stats::model.matrix(model_frame, data = model_frame, formula = formula)

      # replaces NAs by zero to omit the corresponding coefficients
      model_matrix[is.na(model_matrix)] <- 0


      ### prepare posterior samples and information
      for(i in seq_along(samples)){
        # de-name factor levels
        if(priors_info[[i]][["factor"]]){
          if(priors_info[[i]][["levels"]] == 1){
            colnames(samples[[i]]) <- priors_info[[i]][["term"]]
          }else{
            colnames(samples[[i]]) <- paste0(priors_info[[i]][["term"]], "[", 1:priors_info[[i]][["levels"]], "]")
          }
        }
      }
      posterior_samples_matrix <- do.call(cbind, samples)


      # obtain samples information
      models_ind <- do.call(cbind, lapply(c(if(has_intercept) "intercept", model_terms), function(x) attr(samples[[JAGS_parameter_names(x, formula_parameter = formula_parameter)]], "models_ind")))
      sample_ind <- do.call(cbind, lapply(c(if(has_intercept) "intercept", model_terms), function(x) attr(samples[[JAGS_parameter_names(x, formula_parameter = formula_parameter)]], "sample_ind")))
      if(!inherits(samples, "as_mixed_posteriors") && (!all(models_ind[,1] == models_ind) || !all(sample_ind[,1] == sample_ind)))
        stop("the posterior samples are not alligned across models/draws")
      models_ind <- models_ind[,1]


      ### evaluate the design matrix on the samples -> output[data, posterior]
      if(has_intercept){

        terms_indexes    <- attr(model_matrix, "assign") + 1
        terms_indexes[1] <- 0

        # get model/sample indices and check for scaling factors
        temp_multiply_by <- .get_combined_parameter_scaling_factor_matrix(
          JAGS_parameter_names("intercept", formula_parameter = formula_parameter),
          prior_list  = prior_list,
          posterior   = posterior_samples_matrix,
          models_ind  = models_ind,
          nrow        = nrow(data),
          simple_list = inherits(samples, "as_mixed_posteriors")
        )


        marginal_posterior_samples <- temp_multiply_by * matrix(posterior_samples_matrix[,JAGS_parameter_names("intercept", formula_parameter = formula_parameter)],
                                                       nrow = nrow(data), ncol = nrow(posterior_samples_matrix), byrow = TRUE)

      }else{

        terms_indexes <- attr(model_matrix, "assign")
        marginal_posterior_samples <- matrix(0, nrow = nrow(data), ncol = nrow(posterior_samples_matrix))

      }

      # add remaining terms (omitting the intercept indexed as 0)
      for(i in unique(terms_indexes[terms_indexes > 0])){

        # subset the model matrix
        temp_data <- model_matrix[,terms_indexes == i,drop = FALSE]

        temp_posterior <- posterior_samples_matrix[,paste0(
          JAGS_model_terms[i],
          if(model_terms_type[i] == "factor" && priors_info[[JAGS_model_terms[i]]][["levels"]] > 1) paste0("[", 1:priors_info[[JAGS_model_terms[i]]][["levels"]], "]"))
          ,drop = FALSE]

        # check for scaling factors
        temp_multiply_by <- .get_combined_parameter_scaling_factor_matrix(
          JAGS_model_terms[i],
          prior_list  = prior_list,
          posterior   = posterior_samples_matrix,
          models_ind  = models_ind,
          nrow        = nrow(data),
          simple_list = inherits(samples, "as_mixed_posteriors")
        )

        marginal_posterior_samples <- marginal_posterior_samples + temp_multiply_by * (temp_data %*% t(temp_posterior))

      }


      # apply transformations
      if(!is.null(transformation)){
        marginal_posterior_samples <- .density.prior_transformation_x(marginal_posterior_samples, transformation, transformation_arguments)
      }


      ### split the output into lists based on specification
      # create indexing and names for the manipulated predictors
      if(length(at_manipulated) == 1 && format_parameter_names(at_manipulated, formula_parameters = formula_parameter, formula_prefix = FALSE) == "intercept"){

        class(marginal_posterior_samples)             <- c(class(marginal_posterior_samples), "marginal_posterior.simple")
        attr(marginal_posterior_samples, "parameter") <- parameter
        attr(marginal_posterior_samples, "level")     <- "intercept"
        attr(marginal_posterior_samples, "data")      <- data

        marginal_posterior_samples <- list("intercept" = marginal_posterior_samples)

        attr(marginal_posterior_samples, "data")        <- data
        attr(marginal_posterior_samples, "level_at")    <- NULL
        attr(marginal_posterior_samples, "level_names") <- "intercept"
        attr(marginal_posterior_samples, "parameter")   <- parameter

      }else{

        manipulated_predictors <- format_parameter_names(at_manipulated, formula_parameters = formula_parameter, formula_prefix = FALSE)
        at_index_output        <- at[manipulated_predictors]
        at_index_output.names  <- at_index_output

        # rename continuous predictors levels
        for(i in seq_along(at_manipulated)){
          if(predictors_type[[at_manipulated[i]]] == "continuous"){
            at_index_output.names[[manipulated_predictors[i]]] <- paste0(at_index_output.names[[manipulated_predictors[i]]], "SD")
          }
        }

        at_index_output_frame       <- expand.grid(at_index_output)
        at_index_output.names_frame <- expand.grid(at_index_output.names)
        level_names                 <- apply(at_index_output.names_frame, 1, paste0, collapse = ", ")

        # split the output samples
        data_split <- lapply(1:nrow(at_index_output_frame), function(i){
          apply(do.call(cbind, lapply(colnames(at_index_output_frame), function(pred){
            data[, pred] == at_index_output_frame[i, pred]
          })), 1, all)
        })
        marginal_posterior_samples <- lapply(seq_along(data_split), function(lvl){
          temp_marginal_posterior_samples <- marginal_posterior_samples[data_split[[lvl]],]
          temp_data                       <- data[data_split[[lvl]],]
          class(temp_marginal_posterior_samples)             <- c(class(temp_marginal_posterior_samples), "marginal_posterior.simple")
          attr(temp_marginal_posterior_samples, "parameter") <- parameter
          attr(temp_marginal_posterior_samples, "level")     <- level_names[lvl]
          attr(temp_marginal_posterior_samples, "data")      <- temp_data
          return(temp_marginal_posterior_samples)
        })
        names(marginal_posterior_samples) <- level_names
        class(marginal_posterior_samples) <- c(class(marginal_posterior_samples), "marginal_posterior.factor")

        attr(marginal_posterior_samples, "data")        <- data
        attr(marginal_posterior_samples, "level_at")    <- at_index_output_frame
        attr(marginal_posterior_samples, "level_names") <- level_names
        attr(marginal_posterior_samples, "parameter")   <- parameter

      }


      # add priors
      if(prior_samples){

        if(sum(grepl(":", model_terms, fixed = TRUE)) > 5){
          warning(
            "Deterministic marginal prior densities with more than five interaction terms can be slow.",
            call. = FALSE
          )
        }

        prior_density_context <- attr(samples, "prior_density_context")
        if(is.null(prior_density_context)){
          prior_density_context <- .prior_density_build_context(
            prior_list   = prior_list,
            column_names = colnames(posterior_samples_matrix),
            n_grid       = max(16L, n_samples)
          )
        }

        linear_weights <- matrix(
          0,
          nrow = nrow(data),
          ncol = length(prior_density_context$column_names),
          dimnames = list(NULL, prior_density_context$column_names)
        )

        if(has_intercept){
          terms_indexes    <- attr(model_matrix, "assign") + 1
          terms_indexes[1] <- 0
          intercept_name <- JAGS_parameter_names("intercept", formula_parameter = formula_parameter)
          if(intercept_name %in% colnames(linear_weights)){
            linear_weights[, intercept_name] <- 1
          }
        }else{
          terms_indexes <- attr(model_matrix, "assign")
        }

        for(i in unique(terms_indexes[terms_indexes > 0])){
          temp_data <- model_matrix[, terms_indexes == i, drop = FALSE]
          temp_all_columns <- paste0(
            JAGS_model_terms[i],
            if(model_terms_type[i] == "factor" && priors_info[[JAGS_model_terms[i]]][["levels"]] > 1) paste0("[", 1:priors_info[[JAGS_model_terms[i]]][["levels"]], "]")
          )
          temp_columns_keep <- temp_all_columns %in% colnames(linear_weights)
          temp_columns <- temp_all_columns[temp_columns_keep]
          temp_data <- temp_data[, temp_columns_keep, drop = FALSE]
          if(length(temp_columns) == 0)
            next

          linear_weights[, temp_columns] <- linear_weights[, temp_columns, drop = FALSE] + temp_data[, seq_along(temp_columns), drop = FALSE]
        }

        if(length(at_manipulated) == 1 && format_parameter_names(at_manipulated, formula_parameters = formula_parameter, formula_prefix = FALSE) == "intercept"){

          prior_weights <- linear_weights
          prior_density <- .prior_density_from_context_rows(
            prior_density_context,
            prior_weights,
            output_transformation           = transformation,
            output_transformation_arguments = transformation_arguments
          )
          attr(marginal_posterior_samples[["intercept"]], "linear_weights") <- prior_weights
          attr(marginal_posterior_samples[["intercept"]], "prior_density") <- prior_density

        }else{

          for(lvl in seq_along(level_names)){
            prior_weights <- linear_weights[data_split[[lvl]], , drop = FALSE]
            prior_density <- .prior_density_from_context_rows(
              prior_density_context,
              prior_weights,
              output_transformation           = transformation,
              output_transformation_arguments = transformation_arguments
            )
            attr(marginal_posterior_samples[[level_names[lvl]]], "linear_weights") <- prior_weights
            attr(marginal_posterior_samples[[level_names[lvl]]], "prior_density") <- prior_density
          }
        }

      }

      attr(marginal_posterior_samples, "formula_parameter") <- formula_parameter
      class(marginal_posterior_samples) <- c(class(marginal_posterior_samples), "marginal_posterior.formula")

  }else{

    if(!is.null(formula))
      stop("'formula' is supposed to be NULL when dealing with simple posteriors")
    if(!is.null(at))
      stop("'at' is supposed to be NULL when dealing with simple posteriors")


    ### obtain prior list and information
    prior_list <- lapply(names(samples), function(model_term) attr(samples[[model_term]], "prior_list"))
    names(prior_list) <- names(samples)


    ### extract the corresponding samples
    if(inherits(samples[[parameter]], "mixed_posteriors.factor")){

      # transform factor levels
      marginal_posterior_samples <- transform_factor_samples(samples)
      marginal_posterior_samples <- transform_treatment_samples(marginal_posterior_samples)[[parameter]]
      marginal_factor_metadata <- marginal_posterior_samples

      level_names <- attr(marginal_posterior_samples, "level_names")
      if(is.null(level_names) || is.list(level_names)){
        level_names <- .factor_cell_labels(.factor_level_list(marginal_posterior_samples))
      }

      # apply transformations
      if(!is.null(transformation)){
        marginal_posterior_samples <- .density.prior_transformation_x(marginal_posterior_samples, transformation, transformation_arguments)
      }

      # create output object
      marginal_posterior_samples <- lapply(seq_along(level_names), function(lvl_i){
        temp_marginal_posterior_samples <- marginal_posterior_samples[,lvl_i]
        class(temp_marginal_posterior_samples) <- c(class(temp_marginal_posterior_samples), "marginal_posterior.factor")
        attr(temp_marginal_posterior_samples, "parameter")  <- parameter
        attr(temp_marginal_posterior_samples, "level_name") <- level_names[lvl_i]
        return(temp_marginal_posterior_samples)
      })
      names(marginal_posterior_samples) <- level_names
      attr(marginal_posterior_samples, "level_names") <- level_names
      class(marginal_posterior_samples) <- c(class(marginal_posterior_samples), "marginal_posterior.factor")

    }else if(inherits(samples[[parameter]], "mixed_posteriors.simple")){

      marginal_posterior_samples <- samples[[parameter]]

      # apply transformations
      if(!is.null(transformation)){
        marginal_posterior_samples <- .density.prior_transformation_x(marginal_posterior_samples, transformation, transformation_arguments)
      }

      class(marginal_posterior_samples) <- c(class(marginal_posterior_samples), "marginal_posterior.simple")

    }


    # add prior densities
    if(prior_samples){

      prior_density_context <- attr(samples, "prior_density_context")
      if(is.null(prior_density_context)){
        context_columns <- unique(unlist(lapply(names(prior_list), function(parameter_name){
          parameter_prior <- prior_list[[parameter_name]]
          if(is.prior(parameter_prior)){
            .prior_linear_prior_columns(parameter_name, parameter_prior)
          }else{
            .prior_linear_prior_columns(parameter_name, parameter_prior[[1]])
          }
        }), use.names = FALSE))
        prior_density_context <- .prior_density_build_context(
          prior_list   = prior_list,
          column_names = context_columns,
          n_grid       = max(16L, n_samples)
        )
      }

      if(inherits(samples[[parameter]], "mixed_posteriors.factor")){

        factor_weights <- .prior_factor_level_weight_matrix(
          sample_metadata = marginal_factor_metadata,
          parameter       = parameter,
          samples         = samples
        )

        for(lvl_i in seq_along(level_names)){
          weights <- rep(0, length(prior_density_context$column_names))
          names(weights) <- prior_density_context$column_names
          weights[colnames(factor_weights)] <- factor_weights[lvl_i, ]

          prior_density <- .prior_density_from_context(
            prior_density_context,
            weights,
            output_transformation           = transformation,
            output_transformation_arguments = transformation_arguments
          )
          attr(marginal_posterior_samples[[level_names[lvl_i]]], "linear_weights") <- weights
          attr(marginal_posterior_samples[[level_names[lvl_i]]], "prior_density") <- prior_density
        }

      }else if(inherits(samples[[parameter]], "mixed_posteriors.simple")){

        weights <- rep(0, length(prior_density_context$column_names))
        names(weights) <- prior_density_context$column_names
        weights[[parameter]] <- 1

        prior_density <- .prior_density_from_context(
          prior_density_context,
          weights,
          output_transformation           = transformation,
          output_transformation_arguments = transformation_arguments
        )
        attr(marginal_posterior_samples, "linear_weights") <- weights
        attr(marginal_posterior_samples, "prior_density") <- prior_density
      }

    }
  }

  class(marginal_posterior_samples) <- c(class(marginal_posterior_samples), "marginal_posterior")
  return(marginal_posterior_samples)
}

.marginal_posterior_parameter_samples <- function(samples, parameter){

  parameter_samples <- samples[[parameter]]

  if(is.list(parameter_samples)){
    out <- lapply(parameter_samples, as.numeric)
  }else{
    out <- list(as.numeric(parameter_samples))
    names(out) <- parameter
  }

  out
}

.marginal_posterior_parameter_prior_densities <- function(samples, parameter){

  parameter_samples <- samples[[parameter]]

  if(is.list(parameter_samples)){
    out <- lapply(parameter_samples, attr, which = "prior_density")
  }else{
    out <- list(attr(parameter_samples, "prior_density"))
    names(out) <- parameter
  }

  out
}

.get_combined_parameter_scaling_factor_matrix <- function(term, prior_list, posterior, models_ind, nrow, simple_list = FALSE){

  if(simple_list){
    temp_multiply_by <- .get_parameter_scaling_factor_matrix(term, prior_list, posterior, nrow = nrow, ncol = nrow(posterior))
  }else{
    model_samples <- table(models_ind)

    temp_multiply_by <- do.call(cbind, lapply(unique(models_ind), function(m){
      temp_prior_list <- lapply(prior_list, function(parameter_priors) parameter_priors[[m]])
      temp_posterior  <- posterior[models_ind == m,,drop=FALSE]
      return(.get_parameter_scaling_factor_matrix(term, temp_prior_list, temp_posterior, nrow = nrow, ncol = sum(models_ind == m)))
    }))
  }

  return(temp_multiply_by)
}

.mix_priors                <- function(prior_list, seed = NULL, n_samples = 10000){

  check_list(prior_list, "prior_list")
  for(i in seq_along(prior_list)){
    if(any(!sapply(prior_list[[i]], is.prior)))
      stop("'prior_list' must be a list of prior distributions")
  }
  check_real(seed, "seed", allow_NULL = TRUE)
  check_int(n_samples, "n_samples")

  ### get model indices
  prior_weights <- do.call(cbind, lapply(seq_along(prior_list), function(i){
    prior_weights <- sapply(prior_list[[i]], .prior_model_weight)
    prior_weights <- prior_weights / sum(prior_weights)
    return(prior_weights)
  }))
  if(!all(prior_weights[,1] == prior_weights))
    stop("the prior samples are not alligned across models/draws")
  prior_weights <- prior_weights[,1]

  # set seed only once at the beginning -- not in the individual draws as the priors will end up completely correlated
  if(is.null(seed)){
    seed <- sample(.Machine$integer.max, 1)
  }
  set.seed(seed)

  ### adapted from 'mix_posteriors'
  parameters <- names(prior_list)
  out        <- list()

  for(p in seq_along(parameters)){

    # prepare parameter specific values
    temp_parameter    <- parameters[p]
    temp_priors       <- prior_list[[temp_parameter]]

    if(any(sapply(temp_priors, is.prior.weightfunction)) && all(sapply(temp_priors, is.prior.weightfunction) | sapply(temp_priors, .is_prior_weightfunction_null) | sapply(temp_priors, is.null))){
      # weightfunctions:

      # replace missing priors with default prior: none
      for(i in 1:length(temp_priors)){
        if(is.null(temp_priors[[i]])){
          temp_priors[[i]] <- prior_none(prior_weights = prior_weights[i])
        }
      }

      out[[temp_parameter]] <- .mix_priors.weightfunction(temp_priors, temp_parameter, NULL, n_samples)

    }else if(any(sapply(temp_priors, is.prior.factor)) && all(sapply(temp_priors, is.prior.factor) | sapply(temp_priors, is.prior.point) | sapply(temp_priors, is.null))){
      # factor priors

      # replace missing priors with default prior: spike(0)
      for(i in 1:length(temp_priors)){
        if(is.null(temp_priors[[i]])){
          temp_priors[[i]] <- prior("spike", parameters = list("location" = 0), prior_weights = prior_weights[i])
        }
      }

      out[[temp_parameter]] <- .mix_priors.factor(temp_priors, temp_parameter, NULL, n_samples)

    }else if(any(sapply(temp_priors, is.prior.vector)) && all(sapply(temp_priors, is.prior.vector) | sapply(temp_priors, is.prior.point) | sapply(temp_priors, is.null))){
      # vector priors:

      # replace missing priors with default prior: spike(0)
      for(i in 1:length(temp_priors)){
        if(is.null(temp_priors[[i]])){
          temp_priors[[i]] <- prior("spike", parameters = list("location" = 0), prior_weights = prior_weights[i])
        }
      }

      out[[temp_parameter]] <- .mix_priors.vector(temp_priors, temp_parameter, NULL, n_samples)

    }else if(all(sapply(temp_priors, is.prior.simple) | sapply(temp_priors, is.prior.point) | sapply(temp_priors, is.null))){
      # simple priors:

      # replace missing priors with default prior: spike(0)
      for(i in 1:length(temp_priors)){
        if(is.null(temp_priors[[i]])){
          temp_priors[[i]] <- prior("spike", parameters = list("location" = 0), prior_weights = prior_weights[i])
        }
      }

      out[[temp_parameter]] <- .mix_priors.simple(temp_priors, temp_parameter, NULL, n_samples)

    }else{
      stop("The posterior samples cannot be mixed: unsupported mixture of prior distributions.")
    }

    # add formula relevant information
    if(!is.null(unique(unlist(lapply(temp_priors, attr, which = "parameter"))))){
      class(out[[temp_parameter]]) <- c(class(out[[temp_parameter]]), "mixed_posteriors.formula")
      attr(out[[temp_parameter]], "formula_parameter")  <- unique(unlist(lapply(temp_priors, attr, which = "parameter")))
    }
  }

  return(out)
}
.mix_priors.simple         <- function(priors, parameter, seed = NULL, n_samples = 10000){

  # check input
  check_list(priors, "priors")
  check_char(parameter, "parameter")
  check_real(seed, "seed", allow_NULL = TRUE)
  check_int(n_samples, "n_samples")
  if(!all(sapply(priors, is.prior.simple) | sapply(priors, is.prior.point)))
    stop("'priors' must be a list of simple priors")

  # get prior model probabilities
  prior_probs <- sapply(priors, .prior_model_weight)
  prior_probs <- prior_probs / sum(prior_probs)

  # do not set seed when sampling multiple priors for the same model -- they will end up completely correlated
  if(!is.null(seed)){
    set.seed(seed)
  }

  # prepare output objects
  samples <- NULL
  sample_ind <- NULL
  models_ind <- NULL

  # mix samples
  sample_counts <- .prior_mixture_sample_counts(prior_probs, n_samples)
  for(i in seq_along(priors)[sample_counts > 0]){

    # sample indexes
    temp_ind <- seq_len(sample_counts[i])

    # sample prior
    samples <- c(samples, rng(priors[[i]], length(temp_ind), transform_factor_samples = FALSE))

    sample_ind <- c(sample_ind, temp_ind)
    models_ind <- c(models_ind, rep(i, length(temp_ind)))
  }

  # assure the correct number of samples
  samples    <- samples[1:n_samples]
  sample_ind <- sample_ind[1:n_samples]
  models_ind <- models_ind[1:n_samples]

  samples <- unname(samples)
  attr(samples, "sample_ind") <- sample_ind
  attr(samples, "models_ind") <- models_ind
  attr(samples, "parameter")  <- parameter
  attr(samples, "prior_list") <- priors
  class(samples) <- c("mixed_posteriors", "mixed_posteriors.simple")

  return(samples)
}
.mix_priors.vector         <- function(priors, parameter, seed = NULL, n_samples = 10000){

  # check input
  check_list(priors, "priors")
  check_char(parameter, "parameter")
  check_real(seed, "seed", allow_NULL = TRUE)
  check_int(n_samples, "n_samples")
  if(!all(sapply(priors, is.prior.vector) | sapply(priors, is.prior.point)))
    stop("'priors' must be a list of vector priors")

  # get prior model probabilities
  prior_probs <- sapply(priors, .prior_model_weight)
  prior_probs <- prior_probs / sum(prior_probs)

  # do not set seed when sampling multiple priors for the same model -- they will end up completely correlated
  if(!is.null(seed)){
    set.seed(seed)
  }

  # prepare output objects
  K <- unique(sapply(priors[sapply(priors, is.prior.vector)], function(p) p$parameters[["K"]]))
  if(length(K) != 1)
    stop("all vector priors must be of the same length")

  samples    <- matrix(nrow = 0, ncol = K)
  sample_ind <- NULL
  models_ind <- NULL

  # mix samples
  sample_counts <- .prior_mixture_sample_counts(prior_probs, n_samples)
  for(i in seq_along(priors)[sample_counts > 0]){

    # sample indexes
    temp_ind <- seq_len(sample_counts[i])

    if(is.prior.point(priors[[i]]) & is.prior.simple(priors[[i]])){
      # not sampling the priors in case they were imputed (missing dimensions)
      samples <- rbind(samples, matrix(priors[[i]]$parameters[["location"]], nrow = length(temp_ind), ncol = K))
    }else if(K == 1){
      samples <- rbind(samples, matrix(rng(priors[[i]], length(temp_ind), transform_factor_samples = FALSE), nrow = length(temp_ind), ncol = K))
    }else{
      samples <- rbind(samples, rng(priors[[i]], length(temp_ind), transform_factor_samples = FALSE))
    }

    sample_ind <- c(sample_ind, temp_ind)
    models_ind <- c(models_ind, rep(i, length(temp_ind)))
  }

  # assure the correct number of samples
  samples    <- samples[1:n_samples,,drop=FALSE]
  sample_ind <- sample_ind[1:n_samples]
  models_ind <- models_ind[1:n_samples]

  rownames(samples) <- NULL
  colnames(samples) <- paste0(parameter,"[",1:K,"]")
  attr(samples, "sample_ind") <- sample_ind
  attr(samples, "models_ind") <- models_ind
  attr(samples, "parameter")  <- parameter
  attr(samples, "prior_list") <- priors
  class(samples) <- c("mixed_posteriors", "mixed_posteriors.vector")

  return(samples)
}
.mix_priors.factor         <- function(priors, parameter, seed = NULL, n_samples = 10000){

  # check input
  check_list(priors, "priors")
  check_char(parameter, "parameter")
  check_real(seed, "seed", allow_NULL = TRUE)
  check_int(n_samples, "n_samples")
  if(!all(sapply(priors, is.prior.factor) | sapply(priors, is.prior.point)))
    stop("'priors' must be a list of factor priors")

  # get prior model probabilities
  prior_probs <- sapply(priors, .prior_model_weight)
  prior_probs <- prior_probs / sum(prior_probs)

  # check the prior levels
  levels <- unique(sapply(priors[sapply(priors, is.prior.factor)], .get_prior_factor_levels))
  if(length(levels) != 1)
    stop("all factor priors must be of the same number of levels")

  # gather and check compatibility of prior distributions
  priors_info <- lapply(priors, function(p){
    if(is.prior.point(p) | is.prior.none(p)){
      return(FALSE)
    }else if(is.prior.factor(p)){
      return(list(
        "levels"      = .get_prior_factor_levels(p),
        "level_names" = .get_prior_factor_level_names(p),
        "interaction" = .is_prior_interaction(p),
        "interaction_terms" = attr(p, "interaction_terms"),
        "term_components"   = attr(p, "term_components"),
        "factor_terms"      = attr(p, "factor_terms"),
        "factor_contrasts"  = attr(p, "factor_contrasts"),
        "factor_design"     = attr(p, "factor_design"),
        "factor_cell_names" = attr(p, "factor_cell_names"),
        "treatment"   = is.prior.treatment(p),
        "independent" = is.prior.independent(p),
        "orthonormal" = is.prior.orthonormal(p),
        "meandif"     = is.prior.meandif(p)
      ))
    }else{
      stop("unsupported prior type")
    }
  })
  priors_info <- priors_info[!sapply(priors_info, isFALSE)]
  if(length(priors_info) >= 2 && any(!unlist(lapply(priors_info, function(i) all.equal(i, priors_info[[1]]))))){
    stop("non-matching prior factor type specifications")
  }
  priors_info <- priors_info[[1]]

  if(priors_info[["treatment"]]){

    if(levels == 1){

      samples <- .mix_priors.simple(priors, parameter, seed, n_samples)

      sample_ind <- attr(samples, "sample_ind")
      models_ind <- attr(samples, "models_ind")

      samples <- matrix(samples, ncol = 1)

    }else{

      # keep the same seed across levels
      if(is.null(seed)){
        seed <- sample(666666, 1)
      }

      samples <- lapply(1:levels, function(i) .mix_priors.simple(priors, paste0(parameter, "[", i, "]"), seed, n_samples))

      sample_ind <- attr(samples[[1]], "sample_ind")
      models_ind <- attr(samples[[1]], "models_ind")

      samples <- do.call(cbind, samples)

    }

    rownames(samples) <- NULL
    colnames(samples) <- paste0(parameter,"[",priors_info$level_names[-1],"]")
    attr(samples, "sample_ind") <- sample_ind
    attr(samples, "models_ind") <- models_ind
    attr(samples, "parameter")  <- parameter
    attr(samples, "prior_list") <- priors
    class(samples) <- c("mixed_posteriors", "mixed_posteriors.factor", "mixed_posteriors.vector")

  }else if(priors_info[["independent"]]){

    if(levels == 1){

      samples <- .mix_priors.simple(priors, parameter, seed, n_samples)

      sample_ind <- attr(samples, "sample_ind")
      models_ind <- attr(samples, "models_ind")

      samples <- matrix(samples, ncol = 1)

    }else{

      # keep the same seed across levels
      if(is.null(seed)){
        seed <- sample(666666, 1)
      }

      samples <- lapply(1:levels, function(i) .mix_priors.simple(priors, paste0(parameter, "[", i, "]"), seed, n_samples))

      sample_ind <- attr(samples[[1]], "sample_ind")
      models_ind <- attr(samples[[1]], "models_ind")

      samples <- do.call(cbind, samples)

    }

    rownames(samples) <- NULL
    colnames(samples) <- paste0(parameter,"[",priors_info$level_names,"]")
    attr(samples, "sample_ind") <- sample_ind
    attr(samples, "models_ind") <- models_ind
    attr(samples, "parameter")  <- parameter
    attr(samples, "prior_list") <- priors
    class(samples) <- c("mixed_posteriors", "mixed_posteriors.factor", "mixed_posteriors.vector")

  }else if(priors_info[["orthonormal"]] | priors_info[["meandif"]]){

    for(i in seq_along(priors)){
      if(is.prior.factor(priors[[i]])){
        priors[[i]]$parameters[["K"]] <- levels
      }
    }

    samples <- .mix_priors.vector(priors, parameter, seed, n_samples)
    class(samples) <- c(class(samples), "mixed_posteriors.factor")

  }

  attr(samples, "levels")      <- priors_info[["levels"]]
  attr(samples, "level_names") <- priors_info[["level_names"]]
  attr(samples, "interaction") <- priors_info[["interaction"]]
  attr(samples, "interaction_terms") <- priors_info[["interaction_terms"]]
  attr(samples, "term_components")   <- priors_info[["term_components"]]
  attr(samples, "factor_terms")      <- priors_info[["factor_terms"]]
  attr(samples, "factor_contrasts")  <- priors_info[["factor_contrasts"]]
  attr(samples, "factor_design")     <- priors_info[["factor_design"]]
  attr(samples, "factor_cell_names") <- priors_info[["factor_cell_names"]]
  attr(samples, "treatment")   <- priors_info[["treatment"]]
  attr(samples, "independent") <- priors_info[["independent"]]
  attr(samples, "orthonormal") <- priors_info[["orthonormal"]]
  attr(samples, "meandif")     <- priors_info[["meandif"]]

  return(samples)
}

.prior_mixture_sample_counts <- function(prior_probs, n_samples){

  .mixture_sample_counts(prior_probs, n_samples)
}

.mix_priors.weightfunction <- function(priors, parameter, seed = NULL, n_samples = 10000){

  # check input
  check_list(priors, "priors")
  check_char(parameter, "parameter")
  check_real(seed, "seed", allow_NULL = TRUE)
  check_int(n_samples, "n_samples")
  if(!all(sapply(priors, is.prior.weightfunction) | sapply(priors, .is_prior_weightfunction_null)))
    stop("'priors' must be a list of weightfunction priors or point(1)/none null priors")

  # get prior model probabilities
  prior_probs <- sapply(priors, .prior_model_weight)
  prior_probs <- prior_probs / sum(prior_probs)

  # do not set seed when sampling multiple priors for the same model -- they will end up completely correlated
  if(!is.null(seed)){
    set.seed(seed)
  }

  # obtain mapping for the weight coefficients
  omega_info    <- .weightfunction_mapping_info(priors)
  omega_mapping <- omega_info$mapping
  omega_cuts    <- omega_info$cuts
  omega_names   <- omega_info$names

  # prepare output objects
  samples    <- matrix(nrow = 0, ncol = length(omega_cuts) - 1)
  sample_ind <- NULL
  models_ind <- NULL

  # mix samples
  sample_counts <- .prior_mixture_sample_counts(prior_probs, n_samples)
  for(i in seq_along(priors)[sample_counts > 0]){

    # sample indexes
    temp_ind <- seq_len(sample_counts[i])

    if(.is_prior_weightfunction_null(priors[[i]])){
      samples <- rbind(samples, matrix(1, ncol = length(omega_cuts) - 1, nrow = length(temp_ind)))
    }else{
      # create temp samples so names can be matched by mapping
      temp_samples <- rng(priors[[i]], length(temp_ind))
      colnames(temp_samples) <- paste0("omega[",1:ncol(temp_samples),"]")
      samples <- rbind(samples, temp_samples[, paste0("omega[",omega_mapping[[i]],"]")])
    }

    sample_ind <- c(sample_ind, temp_ind)
    models_ind <- c(models_ind, rep(i, length(temp_ind)))
  }

  # assure the correct number of samples
  samples    <- samples[seq_len(n_samples),,drop=FALSE]
  sample_ind <- sample_ind[seq_len(n_samples)]
  models_ind <- models_ind[seq_len(n_samples)]

  rownames(samples) <- NULL
  colnames(samples) <- omega_names
  attr(samples, "sample_ind") <- sample_ind
  attr(samples, "models_ind") <- models_ind
  attr(samples, "parameter")  <- parameter
  attr(samples, "prior_list") <- priors
  samples <- .weightfunction_set_omega_context(samples, omega_info)
  class(samples) <- c("mixed_posteriors", "mixed_posteriors.weightfunction")

  return(samples)
}

.as_mixed_priors            <- function(prior_list, seed = NULL, n_samples = 10000, conditional = NULL, conditional_rule = NULL){

  check_list(prior_list, "prior_list")
  if(any(!sapply(prior_list, is.prior)))
    stop("'prior_list' must be a list of prior distributions")
  check_real(seed, "seed", allow_NULL = TRUE)
  check_int(n_samples, "n_samples")

  # set seed only once at the beginning -- not in the individual draws as the priors will end up completely correlated
  if(is.null(seed)){
    seed <- sample(.Machine$integer.max, 1)
  }
  set.seed(seed)

  # adapted from 'as_mixed_posteriors'
  parameters <- names(prior_list)
  out        <- list()

  # estimate the number of necessary samples for conditioning
  if(length(conditional) > 0){

    conditioning_probabilities <- sapply(conditional, function(parameter){

      temp_prior <- prior_list[[parameter]]

      if(is.prior.spike_and_slab(temp_prior)){

        return(mean(.get_spike_and_slab_inclusion(temp_prior)))

      }else if(is.prior.mixture(temp_prior)){

        components    <- attr(temp_prior, "components")
        prior_weights <- attr(temp_prior, "prior_weights")

        if(!all(components %in% c("null", "alternative")))
          stop("conditional mixture posterior distributions are available only for 'null' and 'alternative' components")

        return(sum(prior_weights[components == "alternative"]) / sum(prior_weights))

      }else{

        warning(sprintf("The parameter '%s' is not a conditional parameter. All samples are assumed to compe from the conditional posterior distribution.", parameter), call. = FALSE, immediate. = TRUE)
        return(1)
      }
    })
    # add a check when forwarding samples to marginal inference
    all_alternative <- all(sapply(conditional, function(parameter){

      temp_prior <- prior_list[[parameter]]

      if(is.prior.spike_and_slab(temp_prior)){
        return(mean(.get_spike_and_slab_inclusion(temp_prior)) == 1)
      }else if(is.prior.mixture(temp_prior)){
        return(all(attr(temp_prior, "components") == "alternative"))
      }else{
        return(TRUE)
      }
    }))

    # multiply by 1.25 to ensure that the requested number of samples is reached
    requested_samples <- n_samples
    if(conditional_rule == "AND"){
      n_samples <- round(n_samples / prod(conditioning_probabilities) * 1.25)
    }else if(conditional_rule == "OR"){
      n_samples  <- round(n_samples / (1 - prod(1 - conditioning_probabilities)) * 1.25)
    }
  }

  # create the samples
  for(p in seq_along(parameters)){

    # prepare parameter specific values
    temp_parameter <- parameters[p]
    temp_prior     <- prior_list[[temp_parameter]]

    if(is.prior.spike_and_slab(temp_prior)){
      # spike and slab priors
      out[[temp_parameter]] <- .as_mixed_priors.spike_and_slab(temp_prior, temp_parameter, NULL, n_samples)

    }else if(is.prior.mixture(temp_prior)){
      # mixture priors
      out[[temp_parameter]] <- .as_mixed_priors.mixture(temp_prior, temp_parameter, NULL, n_samples)

    }else if(is.prior.weightfunction(temp_prior)){
      # weightfunctions:
      out[[temp_parameter]] <- .as_mixed_priors.weightfunction(temp_prior, temp_parameter, NULL, n_samples)

    }else if(is_prior_phacking(temp_prior)){
      # p-hacking priors:
      out[[temp_parameter]] <- .as_mixed_priors.phacking(temp_prior, temp_parameter, NULL, n_samples)

    }else if(is_prior_bias(temp_prior)){
      # composed publication-bias priors:
      out[[temp_parameter]] <- .as_mixed_priors.bias(temp_prior, temp_parameter, NULL, n_samples)

    }else if(is.prior.factor(temp_prior)){
      # factor priors
      out[[temp_parameter]] <- .as_mixed_priors.factor(temp_prior, temp_parameter, NULL, n_samples)

    }else if(is.prior.vector(temp_prior)){
      # vector priors:
      out[[temp_parameter]] <- .as_mixed_priors.vector(temp_prior, temp_parameter, NULL, n_samples)

    }else if(is.prior.simple(temp_prior)){
      # simple priors:
      out[[temp_parameter]] <- .as_mixed_priors.simple(temp_prior, temp_parameter, NULL, n_samples)

    }else{
      stop("The posterior samples cannot be mixed: unsupported mixture of prior distributions.")
    }

    # add formula relevant information
    if(!is.null(attr(temp_prior, which = "parameter"))){
      class(out[[temp_parameter]]) <- c(class(out[[temp_parameter]]), "mixed_posteriors.formula")
      attr(out[[temp_parameter]], "formula_parameter")  <- attr(temp_prior, which = "parameter")
    }
  }


  # perform conditioning (and copy back with attributes)
  if(length(conditional) > 0){

    # obtain the indicator samples
    conditioning_samples <- do.call(cbind, lapply(conditional, function(parameter){

      temp_prior <- prior_list[[parameter]]

      if(is.prior.spike_and_slab(temp_prior)){

        return(attr(out[[parameter]], "models_ind") == 1)

      }else if(is.prior.mixture(temp_prior)){

        components <- attr(temp_prior, "components")
        return(attr(out[[parameter]], "models_ind") %in% which(components == "alternative"))

      }else{

        return(rep(TRUE, n_samples))
      }
    }))
    conditioning_samples <- apply(conditioning_samples, 1, ifelse(conditional_rule == "AND", all, any))

    # check enough samples were drawn (if too many remove the extra ones)
    if(sum(conditioning_samples) < requested_samples){
      warning(sprintf("Only %d samples were drawn from the prior distributions due to conditioning.", sum(conditioning_samples)))
    }else{
      conditioning_samples[which(conditioning_samples)[-(1:requested_samples)]] <- FALSE
    }

    # select the conditional samples (and copy attributes)
    for(p in seq_along(parameters)){
      temp <- attributes(out[[parameters[p]]])
      if(is.null(dim(out[[parameters[p]]]))){
        out[[parameters[p]]] <- out[[parameters[p]]][conditioning_samples]
        attributes(out[[parameters[p]]]) <- c(attributes(out[[parameters[p]]]), temp)
        attr(out[[parameters[p]]], "models_ind") <- attr(out[[parameters[p]]], "models_ind")[conditioning_samples]
      }else{
        out[[parameters[p]]] <- out[[parameters[p]]][conditioning_samples,,drop=FALSE]
        attributes(out[[parameters[p]]]) <- c(attributes(out[[parameters[p]]])[!names(attributes(out[[parameters[p]]])) %in% c("dimnames")], temp[!names(temp) %in% c("dim")])
        attr(out[[parameters[p]]], "models_ind") <- attr(out[[parameters[p]]], "models_ind")[conditioning_samples]
      }
    }

    # put a check whether all samples were conditional
    attr(out, "all_alternative") <- all_alternative
  }

  return(out)
}
.as_mixed_priors.simple         <- function(prior, parameter, seed = NULL, n_samples = 10000){

  # check input
  check_list(prior, "prior")
  check_char(parameter, "parameter")
  check_real(seed, "seed", allow_NULL = TRUE)
  check_int(n_samples, "n_samples")

  # do not set seed when sampling multiple prior for the same model -- they will end up completely correlated
  if(!is.null(seed)){
    set.seed(seed)
  }

  # prepare output objects
  samples <- rng(prior, n_samples, transform_factor_samples = FALSE)

  attr(samples, "sample_ind") <- FALSE
  attr(samples, "models_ind") <- FALSE
  attr(samples, "parameter")  <- parameter
  attr(samples, "prior_list") <- prior
  class(samples) <- c("mixed_posteriors", "mixed_posteriors.simple")

  return(samples)
}
.as_mixed_priors.vector         <- function(prior, parameter, seed = NULL, n_samples = 10000){

  # check input
  check_list(prior, "prior")
  check_char(parameter, "parameter")
  check_real(seed, "seed", allow_NULL = TRUE)
  check_int(n_samples, "n_samples")

  # do not set seed when sampling multiple prior for the same model -- they will end up completely correlated
  if(!is.null(seed)){
    set.seed(seed)
  }

  # prepare output objects
  K <- prior$parameters[["K"]]

  if(is.prior.point(prior) & is.prior.simple(prior)){
    # not sampling the prior in case they were imputed (missing dimensions)
    samples <- matrix(prior$parameters[["location"]], nrow = n_samples, ncol = K)
  }else if(K == 1){
    samples <- matrix(rng(prior, n_samples, transform_factor_samples = FALSE), nrow = n_samples, ncol = K)
  }else{
    samples <- rng(prior, n_samples, transform_factor_samples = FALSE)
  }

  rownames(samples) <- NULL
  colnames(samples) <- paste0(parameter,"[",1:K,"]")
  attr(samples, "sample_ind") <- FALSE
  attr(samples, "models_ind") <- FALSE
  attr(samples, "parameter")  <- parameter
  attr(samples, "prior_list") <- prior
  class(samples) <- c("mixed_posteriors", "mixed_posteriors.vector")

  return(samples)
}
.as_mixed_priors.factor         <- function(prior, parameter, seed = NULL, n_samples = 10000){

  # check input
  check_list(prior, "prior")
  check_char(parameter, "parameter")
  check_real(seed, "seed", allow_NULL = TRUE)
  check_int(n_samples, "n_samples")

  # check the prior levels
  levels <- .get_prior_factor_levels(prior)

  # gather and check compatibility of prior distributions
  prior_info <- list(
    "levels"      = .get_prior_factor_levels(prior),
    "level_names" = .get_prior_factor_level_names(prior),
    "interaction" = .is_prior_interaction(prior),
    "interaction_terms" = attr(prior, "interaction_terms"),
    "term_components"   = attr(prior, "term_components"),
    "factor_terms"      = attr(prior, "factor_terms"),
    "factor_contrasts"  = attr(prior, "factor_contrasts"),
    "factor_design"     = attr(prior, "factor_design"),
    "factor_cell_names" = attr(prior, "factor_cell_names"),
    "treatment"   = is.prior.treatment(prior),
    "independent" = is.prior.independent(prior),
    "orthonormal" = is.prior.orthonormal(prior),
    "meandif"     = is.prior.meandif(prior)
  )

  if(prior_info[["treatment"]]){

    if(levels == 1){

      samples <- .as_mixed_priors.simple(prior, parameter, seed, n_samples)
      samples <- matrix(samples, ncol = 1)

    }else{

      samples <- lapply(1:levels, function(i) .as_mixed_priors.simple(prior, paste0(parameter, "[", i, "]"), seed, n_samples))
      samples <- do.call(cbind, samples)

    }

    rownames(samples) <- NULL
    colnames(samples) <- paste0(parameter,"[",prior_info$level_names[-1],"]")
    attr(samples, "sample_ind") <- FALSE
    attr(samples, "models_ind") <- FALSE
    attr(samples, "parameter")  <- parameter
    attr(samples, "prior_list") <- prior
    class(samples) <- c("mixed_posteriors", "mixed_posteriors.factor", "mixed_posteriors.vector")

  }else if(prior_info[["independent"]]){

    if(levels == 1){

      samples <- .as_mixed_priors.simple(prior, parameter, seed, n_samples)
      samples <- matrix(samples, ncol = 1)

    }else{

      samples <- lapply(1:levels, function(i) .as_mixed_priors.simple(prior, paste0(parameter, "[", i, "]"), seed, n_samples))
      samples <- do.call(cbind, samples)

    }

    rownames(samples) <- NULL
    colnames(samples) <- paste0(parameter,"[",prior_info$level_names,"]")
    attr(samples, "sample_ind") <- FALSE
    attr(samples, "models_ind") <- FALSE
    attr(samples, "parameter")  <- parameter
    attr(samples, "prior_list") <- prior
    class(samples) <- c("mixed_posteriors", "mixed_posteriors.factor", "mixed_posteriors.vector")

  }else if(prior_info[["orthonormal"]] | prior_info[["meandif"]]){

    prior$parameters[["K"]] <- levels
    samples <- .as_mixed_priors.vector(prior, parameter, seed, n_samples)
    class(samples) <- c(class(samples), "mixed_posteriors.factor")

  }

  attr(samples, "levels")      <- prior_info[["levels"]]
  attr(samples, "level_names") <- prior_info[["level_names"]]
  attr(samples, "interaction") <- prior_info[["interaction"]]
  attr(samples, "interaction_terms") <- prior_info[["interaction_terms"]]
  attr(samples, "term_components")   <- prior_info[["term_components"]]
  attr(samples, "factor_terms")      <- prior_info[["factor_terms"]]
  attr(samples, "factor_contrasts")  <- prior_info[["factor_contrasts"]]
  attr(samples, "factor_design")     <- prior_info[["factor_design"]]
  attr(samples, "factor_cell_names") <- prior_info[["factor_cell_names"]]
  attr(samples, "treatment")   <- prior_info[["treatment"]]
  attr(samples, "independent") <- prior_info[["independent"]]
  attr(samples, "orthonormal") <- prior_info[["orthonormal"]]
  attr(samples, "meandif")     <- prior_info[["meandif"]]

  return(samples)
}
.as_mixed_priors.weightfunction <- function(prior, parameter, seed = NULL, n_samples = 10000){

  # check input
  check_list(prior, "prior")
  check_char(parameter, "parameter")
  check_real(seed, "seed", allow_NULL = TRUE)
  check_int(n_samples, "n_samples")

  # do not set seed when sampling multiple prior for the same model -- they will end up completely correlated
  if(!is.null(seed)){
    set.seed(seed)
  }

  # obtain mapping for the weight coefficients
  omega_mapping <- weightfunctions_mapping(list(prior))
  omega_cuts    <- weightfunctions_mapping(list(prior), cuts_only = TRUE)
  omega_names   <- sapply(1:(length(omega_cuts)-1), function(i)paste0("omega[",omega_cuts[i],",",omega_cuts[i+1],"]"))

  # prepare output objects
  samples <- rng(prior, n_samples)

  rownames(samples) <- NULL
  colnames(samples) <- omega_names
  attr(samples, "sample_ind") <- FALSE
  attr(samples, "models_ind") <- FALSE
  attr(samples, "parameter")  <- parameter
  attr(samples, "prior_list") <- prior
  class(samples) <- c("mixed_posteriors", "mixed_posteriors.weightfunction")

  return(samples)
}
.as_mixed_priors.phacking <- function(prior, parameter, seed = NULL, n_samples = 10000){

  # check input
  check_list(prior, "prior")
  check_char(parameter, "parameter")
  check_real(seed, "seed", allow_NULL = TRUE)
  check_int(n_samples, "n_samples")

  if(!is.null(seed)){
    set.seed(seed)
  }

  samples <- rng(prior, n_samples)
  par_names <- intersect(.phacking_report_parameter(prior), colnames(samples))
  samples <- samples[, par_names, drop = FALSE]

  rownames(samples) <- NULL
  attr(samples, "sample_ind") <- FALSE
  attr(samples, "models_ind") <- FALSE
  attr(samples, "parameter")  <- parameter
  attr(samples, "prior_list") <- prior
  class(samples) <- c("mixed_posteriors", "mixed_posteriors.phacking")

  return(samples)
}
.as_mixed_priors.bias <- function(prior, parameter, seed = NULL, n_samples = 10000){

  # check input
  check_list(prior, "prior")
  check_char(parameter, "parameter")
  check_real(seed, "seed", allow_NULL = TRUE)
  check_int(n_samples, "n_samples")

  if(!is.null(seed)){
    set.seed(seed)
  }

  spec          <- selection_backend_spec(prior)
  branch_info   <- .selection_prior_branch_info(prior)
  has_selection <- vapply(branch_info, function(x) !is.null(x$selection), logical(1))
  has_phacking  <- vapply(branch_info, function(x) !is.null(x$phacking),  logical(1))

  samples <- rng(prior, n_samples)

  out_names <- character()
  par_names <- character()

  if(any(has_selection)){
    omega_cuts  <- spec$step$breaks
    omega_names <- sapply(seq_len(length(omega_cuts) - 1L), function(i) paste0("omega[", omega_cuts[i], ",", omega_cuts[i + 1L], "]"))
    omega_par   <- paste0("omega[", seq_len(length(omega_cuts) - 1L), "]")
    out_names   <- c(out_names, omega_names)
    par_names   <- c(par_names, omega_par)
  }
  if(any(has_phacking)){
    phacking_priors <- lapply(branch_info[has_phacking], function(x) x$phacking)
    phacking_names <- .selection_phacking_report_parameters(phacking_priors)
    out_names <- c(out_names, phacking_names)
    par_names <- c(par_names, phacking_names)
  }

  keep <- par_names %in% colnames(samples)
  samples <- samples[, par_names[keep], drop = FALSE]
  colnames(samples) <- out_names[keep]

  rownames(samples) <- NULL
  attr(samples, "sample_ind") <- FALSE
  attr(samples, "models_ind") <- FALSE
  attr(samples, "parameter")  <- parameter
  attr(samples, "prior_list") <- prior
  class(samples) <- c("mixed_posteriors", "mixed_posteriors.bias")

  return(samples)
}
.as_mixed_priors.spike_and_slab <- function(prior, parameter, seed = NULL, n_samples = 10000){

  # check input
  check_list(prior, "prior")
  check_char(parameter, "parameter")
  check_real(seed, "seed", allow_NULL = TRUE)
  check_int(n_samples, "n_samples")

  # do not set seed when sampling multiple prior for the same model -- they will end up completely correlated
  if(!is.null(seed)){
    set.seed(seed)
  }

  prior_variable   <- .get_spike_and_slab_variable(prior)
  prior_inclusion  <- .get_spike_and_slab_inclusion(prior)

  inclusion <- stats::rbinom(n_samples, size = 1, prob = rng(prior_inclusion, n_samples))

  if(is.prior.factor(prior_variable)){

    samples <- .as_mixed_priors.factor(prior_variable, parameter, seed, n_samples)

  }else if(is.prior.simple(prior_variable)){

    samples <- .as_mixed_priors.simple(prior_variable, parameter, seed, n_samples)

  }

  # merge with names and attributes
  samples       <- samples * inclusion

  class(samples) <- c(class(samples), "mixed_posteriors.spike_and_slab")
  attr(samples, "sample_ind") <- FALSE
  attr(samples, "models_ind") <- inclusion
  attr(samples, "prior_list") <- prior

  return(samples)
}
.as_mixed_priors.mixture        <- function(prior, parameter, seed = NULL, n_samples = 10000){

  # check input
  check_list(prior, "prior")
  check_char(parameter, "parameter")
  check_real(seed, "seed", allow_NULL = TRUE)
  check_int(n_samples, "n_samples")

  is_PET            <- sapply(prior, is.prior.PET)
  is_PEESE          <- sapply(prior, is.prior.PEESE)
  is_weightfunction <- sapply(prior, is.prior.weightfunction)
  is_phacking       <- sapply(prior, is_prior_phacking)
  is_bias           <- sapply(prior, is_prior_bias)

  if(any(is_PET | is_PEESE | is_weightfunction | is_phacking | is_bias)){

    temp_samples <- .mix_priors.bias(prior, parameter = parameter, seed = seed, n_samples = n_samples)

  }else{

    is_simple <- sapply(prior, is.prior.simple)
    is_factor <- sapply(prior, is.prior.factor)

    if(any(is_factor)){
      temp_samples <- .mix_priors.factor(prior, parameter = parameter, seed = seed, n_samples = n_samples)
    }else{
      temp_samples <- .mix_priors.simple(prior, parameter = parameter, seed = seed, n_samples = n_samples)
    }

  }

  # the samples parameters need to be randomly shuffled
  # (the  .mix_priors.XXX functions generate the samples model by model to keep bridge-sampling model-averaging consistent structure,
  #  this however does not apply to the spike an slab priors)
  random_ind <- sample(n_samples)
  if(is.null(dim(temp_samples))){
    samples <- temp_samples[random_ind]
  }else{
    samples <- temp_samples[random_ind,,drop=FALSE]
  }
  attributes(samples) <- attributes(temp_samples)
  attr(samples, "sample_ind") <- FALSE
  attr(samples, "models_ind") <- attr(samples, "models_ind")[random_ind]

  # append classes and priors
  class(samples) <- c(class(samples), "mixed_posteriors.mixture")
  attr(samples, "prior_list") <- prior

  return(samples)
}
.mix_priors.bias <- function(priors, parameter, seed = NULL, n_samples = 10000){

  # check input
  check_list(priors, "priors")
  check_char(parameter, "parameter")
  check_real(seed, "seed", allow_NULL = TRUE)
  check_int(n_samples, "n_samples")

  allowed <- sapply(priors, function(prior){
    is.prior.none(prior) || is.prior.PET(prior) || is.prior.PEESE(prior) ||
      is.prior.weightfunction(prior) || is_prior_phacking(prior) || is_prior_bias(prior)
  })
  if(!all(allowed)){
    stop("'priors' must be a list of publication-bias priors.", call. = FALSE)
  }

  spec        <- selection_backend_spec(priors)
  branch_info <- lapply(priors, .selection_branch_info)

  is_PET       <- sapply(priors, is.prior.PET)
  is_PEESE     <- sapply(priors, is.prior.PEESE)
  has_selection <- vapply(branch_info, function(x) !is.null(x$selection), logical(1))
  has_phacking  <- vapply(branch_info, function(x) !is.null(x$phacking),  logical(1))

  prior_probs <- sapply(priors, .prior_model_weight)
  prior_probs <- prior_probs / sum(prior_probs)

  if(!is.null(seed)){
    set.seed(seed)
  }

  out_names <- character()
  if(any(has_selection)){
    omega_cuts  <- spec$step$breaks
    omega_names <- .weightfunction_omega_names(omega_cuts)
    out_names   <- c(out_names, omega_names)

    selection_priors <- lapply(branch_info[has_selection], function(x) x$selection)
    omega_mapping    <- .weightfunction_mapping_info(selection_priors, one_sided = TRUE)$mapping
    selection_index  <- which(has_selection)
  }
  if(any(has_phacking)){
    phacking_priors <- lapply(branch_info[has_phacking], function(x) x$phacking)
    phacking_names  <- .selection_phacking_report_parameters(phacking_priors)
    out_names <- c(out_names, phacking_names)
  }
  if(any(is_PET)){
    out_names <- c(out_names, "PET")
  }
  if(any(is_PEESE)){
    out_names <- c(out_names, "PEESE")
  }

  samples    <- matrix(nrow = 0, ncol = length(out_names))
  colnames(samples) <- out_names
  sample_ind <- NULL
  models_ind <- NULL

  sample_counts <- .prior_mixture_sample_counts(prior_probs, n_samples)
  for(i in seq_along(priors)[sample_counts > 0]){

    temp_ind <- seq_len(sample_counts[i])
    temp_samples <- matrix(0, nrow = length(temp_ind), ncol = length(out_names))
    colnames(temp_samples) <- out_names

    if(any(has_selection)){
      temp_samples[, omega_names] <- 1
    }

    if(has_selection[i]){
      selection_i <- match(i, selection_index)
      selection_samples <- rng(branch_info[[i]]$selection, length(temp_ind))
      temp_samples[, omega_names] <- selection_samples[, paste0("omega[", omega_mapping[[selection_i]], "]"), drop = FALSE]
    }

    if(has_phacking[i]){
      phacking_samples <- rng(branch_info[[i]]$phacking, length(temp_ind))
      phacking_names <- .phacking_report_parameter(branch_info[[i]]$phacking)
      temp_samples[, phacking_names] <- phacking_samples[, phacking_names, drop = FALSE]
    }

    if(is_PET[i]){
      temp_samples[, "PET"] <- rng(priors[[i]], length(temp_ind), transform_factor_samples = FALSE)
    }
    if(is_PEESE[i]){
      temp_samples[, "PEESE"] <- rng(priors[[i]], length(temp_ind), transform_factor_samples = FALSE)
    }

    samples    <- rbind(samples, temp_samples)
    sample_ind <- c(sample_ind, temp_ind)
    models_ind <- c(models_ind, rep(i, length(temp_ind)))
  }

  samples    <- samples[seq_len(n_samples), , drop = FALSE]
  sample_ind <- sample_ind[seq_len(n_samples)]
  models_ind <- models_ind[seq_len(n_samples)]

  rownames(samples) <- NULL
  attr(samples, "sample_ind") <- sample_ind
  attr(samples, "models_ind") <- models_ind
  attr(samples, "parameter")  <- parameter
  attr(samples, "prior_list") <- priors
  if(any(has_selection)){
    samples <- .weightfunction_set_omega_context(samples, list(
      mapping   = NULL,
      cuts      = omega_cuts,
      names     = omega_names,
      pars      = paste0("omega[", seq_len(length(omega_cuts) - 1L), "]"),
      one_sided = TRUE
    ))
  }
  class(samples) <- c("mixed_posteriors", "mixed_posteriors.bias")

  return(samples)
}

#' @title Compute Savage-Dickey inclusion Bayes factors
#'
#' @description Computes Savage-Dickey (density ratio) inclusion Bayes factors
#' based the change of height from prior to posterior distribution at the test value.
#'
#' @param posterior marginal posterior distribution generated via the
#' \code{marginal_posterior} function
#' @param null_hypothesis point null hypothesis to test. Defaults to \code{0}
#' @param normal_approximation whether the height of prior and posterior density should be
#' approximated via a normal distribution (rather than kernel density). Defaults to \code{FALSE}.
#' @param silent whether warnings should be returned silently. Defaults to \code{FALSE}
#'
#'
#' @return \code{Savage_Dickey_BF} returns a Bayes factor.
#'
#' @export
Savage_Dickey_BF <- function(posterior, null_hypothesis = 0, normal_approximation = FALSE, silent = FALSE){

  if(!inherits(posterior, "marginal_posterior"))
    stop("'BF_savage_dickey' function requires an object of class 'marginal_posteriors'")
  check_real(null_hypothesis, "null_hypothesis")
  check_bool(normal_approximation, "normal_approximation")
  check_bool(silent, "silent")

  if(is.list(posterior)){
    bf <- list()
    for(i in seq_along(posterior)){
      bf[[i]] <- .Savage_Dickey_BF.fun(posterior[[i]], null_hypothesis, normal_approximation, silent)
    }
    names(bf) <- names(posterior)
  }else{
    bf <- .Savage_Dickey_BF.fun(posterior, null_hypothesis, normal_approximation, silent)
  }

  return(bf)
}

.Savage_Dickey_BF.fun    <- function(posterior, null_hypothesis, normal_approximation, silent){

  if(is.null(attr(posterior, "prior_density")))
    stop("there are no prior densities for the posterior distribution", call. = FALSE)

  prior <- attr(posterior, "prior_density")

  warnings <- NULL

  if(mean(posterior == null_hypothesis) > 0.05){
    warnings <- c(warnings, "There is a considerable cluster of posterior samples at the exact null hypothesis values. The Savage-Dickey density ratio is likely to be invalid.")
  }
  if(.prior_linear_density_point_mass(prior, null_hypothesis) > 0.05){
    warnings <- c(warnings, "There is a considerable point mass in the prior at the exact null hypothesis value. The Savage-Dickey density ratio is likely to be invalid.")
  }

  prior_range <- range(c(
    if(!is.null(prior$density)) prior$density$x else NULL,
    if(!is.null(prior$points) && nrow(prior$points) > 0) prior$points$x else NULL
  ))
  if(null_hypothesis < prior_range[1] || null_hypothesis > prior_range[2]){
    warnings <- c(warnings, "Prior density does not span both sides of the null hypothesis. Check whether the prior distribution contains the null hypothesis in the first place. The Savage-Dickey density ratio is likely to be invalid.")
  }
  if(null_hypothesis < min(posterior) || null_hypothesis > max(posterior)){
    warnings <- c(warnings, "Posterior samples do not span both sides of the null hypothesis. The Savage-Dickey density ratio is likely to be overestimated.")
  }
  if(!silent && !is.null(warnings)){
    sapply(warnings, warning, call. = FALSE)
  }


  if(normal_approximation){
    posterior_height <- .Savage_Dickey_BF.normal(posterior, null_hypothesis)
  }else{
    posterior_height <- .Savage_Dickey_BF.kd(posterior, null_hypothesis)
  }
  prior_height <- .prior_linear_density_height(prior, null_hypothesis)

  BF <- exp(log(prior_height) - log(posterior_height))

  if(!is.null(warnings)){
    attr(BF, "warnings") <- warnings
  }

  return(BF)
}
.Savage_Dickey_BF.normal <- function(samples, null_hypothesis){

  height <- stats::dnorm(null_hypothesis, mean = mean(samples), sd = stats::sd(samples))

  return(height)
}
.Savage_Dickey_BF.kd     <- function(samples, null_hypothesis){

  if(null_hypothesis < min(samples) || null_hypothesis > max(samples)){

    # the test value is outside of the samples
    height  <- 0

  }else{

    # use linear approximation to find the point
    density_posterior <- stats::density(samples)
    density_posterior.x <- c(
      density_posterior$x[which.max(density_posterior$x > null_hypothesis) - 1],
      density_posterior$x[which.max(density_posterior$x > null_hypothesis)]
    )
    density_posterior.y <- c(
      density_posterior$y[which.max(density_posterior$x > null_hypothesis) - 1],
      density_posterior$y[which.max(density_posterior$x > null_hypothesis)]
    )

    dif.y <- density_posterior.y[2] - density_posterior.y[1]
    dif.x <- density_posterior.x[2] - density_posterior.x[1]

    height <- density_posterior.y[1] + dif.y * (null_hypothesis - density_posterior.x[1])/dif.x
  }

  return(height)
}


#' @title Model-average marginal posterior distributions and
#' marginal Bayes factors
#'
#' @description Creates marginal model-averaged and conditional
#' posterior distributions based on a list of models, vector of parameters,
#' formula, and a list of indicators of the null or alternative hypothesis models
#' for each parameter. Computes inclusion Bayes factors for each
#' marginal estimate via a Savage-Dickey density approximation.
#'
#' @param marginal_parameters parameters for which the the marginal summary
#' should be created
#' @param parameters all parameters included in the model_list that are
#' relevant for the formula (all of which need to have specification of
#' \code{is_null_list})
#' @param seed seed for random number generation
#' @inheritParams ensemble_inference
#' @inheritParams marginal_posterior
#' @inheritParams Savage_Dickey_BF
#'
#' @return \code{marginal_inference} returns an object of class 'marginal_inference'.
#'
#' @seealso [ensemble_inference] [mix_posteriors] [BayesTools_ensemble_tables]
#'
#' @export
marginal_inference <- function(model_list, marginal_parameters, parameters, is_null_list, formula,
                               null_hypothesis = 0, normal_approximation = FALSE,
                               n_samples = 10000, seed = NULL, silent = FALSE){

  # check input (majority of the checks performed within mix_posteriors)
  check_list(model_list, "model_list")
  check_char(parameters, "parameters", check_length = FALSE)
  check_char(marginal_parameters, "marginal_parameters", check_length = FALSE)
  check_list(is_null_list, "is_null_list", check_length = length(parameters))
  if(!all(unlist(sapply(model_list, function(m) sapply(attr(m[["fit"]], "prior_list"), function(p) is.prior(p))))))
    stop("model_list:priors must contain 'BayesTools' priors")


  # create one full model-averaged ensemble
  averaged_posterior <- mix_posteriors(
    model_list   = model_list,
    parameters   = parameters,
    is_null_list = is_null_list,
    seed         = seed,
    n_samples    = n_samples,
    conditional  = FALSE
  )

  # prepare output object
  out <- list(
    conditional = list(),
    averaged    = list(),
    inference   = list()
  )

  for(i in seq_along(marginal_parameters)){

    if(all(is_null_list[[marginal_parameters[i]]])){
      warning(paste0("parameter '", marginal_parameters[i], "' does not contain any alternative hypothesis models."), immediate. = TRUE, call. = FALSE)
      next
    }

    # obtain model-averaged posterior conditional on including the parameter of interest
    # (different from individual conditionals)
    temp_conditional_posterior <- mix_posteriors(
      model_list   = model_list[!is_null_list[[marginal_parameters[i]]]],
      parameters   = parameters,
      is_null_list = lapply(is_null_list, function(l) l[!is_null_list[[marginal_parameters[i]]]]),
      seed         = seed,
      n_samples    = n_samples,
      conditional  = FALSE
    )

    # compute the marginals
    out[["averaged"]][[marginal_parameters[i]]] <- marginal_posterior(
      samples           = averaged_posterior,
      parameter         = marginal_parameters[i],
      formula           = formula,
      prior_samples     = TRUE,
      n_samples         = n_samples
    )
    out[["conditional"]][[marginal_parameters[i]]] <- marginal_posterior(
      samples           = temp_conditional_posterior,
      parameter         = marginal_parameters[i],
      formula           = formula,
      prior_samples     = TRUE,
      n_samples         = n_samples
    )

    # and inclusion Bayes factor
    out[["inference"]][[marginal_parameters[i]]] <- Savage_Dickey_BF(
      posterior            = out[["conditional"]][[marginal_parameters[i]]],
      null_hypothesis      = null_hypothesis,
      normal_approximation = normal_approximation,
      silent               = silent
    )
  }

  attr(out, "null_hypothesis")      <- null_hypothesis
  attr(out, "normal_approximation") <- normal_approximation
  class(out) <- c(class(out), "marginal_inference")
  return(out)
}


#' @title Model-average marginal posterior distributions and
#' marginal Bayes factors based on BayesTools JAGS model via \code{marginal_inference}
#'
#' @description Creates marginal model-averaged and conditional
#' posterior distributions based on a BayesTools JAGS model, vector of parameters,
#' formula, and a list of conditional specifications for each parameter.
#' Computes inclusion Bayes factors for each marginal estimate via a Savage-Dickey
#' density approximation.
#'
#' @param marginal_parameters parameters for which the the marginal summary
#' should be created
#' @param conditional_list list of conditional parameters for each marginal parameter
#' @param parameters all parameters included in the model_list that are
#' relevant for the formula (all of which need to have specification of
#' \code{is_null_list})
#' @inheritParams as_mixed_posteriors
#' @inheritParams marginal_inference
#' @inheritParams Savage_Dickey_BF
#'
#' @return \code{as_marginal_inference} returns an object of class 'marginal_inference'.
#'
#' @seealso [marginal_inference] [as_mixed_posteriors]
#'
#' @export
as_marginal_inference <- function(model, marginal_parameters, parameters, conditional_list, conditional_rule, formula,
                                  null_hypothesis = 0, normal_approximation = FALSE,
                                  n_samples = 10000, silent = FALSE, force_plots = FALSE){

  # check input (majority of the checks performed within mix_posteriors)
  # check input
  if(!inherits(model, "BayesTools_fit"))
    stop("'model' must be a 'BayesTools_fit'")
  check_char(parameters, "parameters", check_length = FALSE)
  check_char(marginal_parameters, "marginal_parameters", check_length = FALSE)
  check_list(conditional_list, "conditional_list", check_length = length(marginal_parameters))
  check_char(conditional_rule, "conditional_rule")

  priors <- attr(model, "prior_list")


  # create one full model-averaged ensemble
  averaged_posterior <- as_mixed_posteriors(
    model        = model,
    parameters   = parameters
  )

  # prepare output object
  out <- list(
    conditional = list(),
    averaged    = list(),
    inference   = list()
  )

  for(i in seq_along(marginal_parameters)){

    check_char(conditional_list[[marginal_parameters[i]]], sprintf("conditional_list[[%1$s]]", marginal_parameters[i]), check_length = FALSE, allow_values = parameters, allow_NULL = TRUE)

    # compute the marginals
    out[["averaged"]][[marginal_parameters[i]]] <- marginal_posterior(
      samples           = averaged_posterior,
      parameter         = marginal_parameters[i],
      formula           = formula,
      prior_samples     = TRUE,
      n_samples         = n_samples
    )

    out[["conditional"]][[marginal_parameters[i]]] <- .marginal_inference_conditional_posterior(
      model              = model,
      parameters         = parameters,
      marginal_parameter = marginal_parameters[i],
      formula            = formula,
      averaged_marginal  = out[["averaged"]][[marginal_parameters[i]]],
      prior_list         = attr(averaged_posterior, "prior_list"),
      conditional        = conditional_list[[marginal_parameters[i]]],
      conditional_rule   = conditional_rule,
      n_samples          = n_samples,
      force_plots        = force_plots
    )

    if(length(out[["conditional"]][[marginal_parameters[i]]]) == 0){
      out[["averaged"]][[marginal_parameters[i]]] <- NULL
      next
    }

    # and inclusion Bayes factor
    out[["inference"]][[marginal_parameters[i]]] <- Savage_Dickey_BF(
      posterior            = out[["conditional"]][[marginal_parameters[i]]],
      null_hypothesis      = null_hypothesis,
      normal_approximation = normal_approximation,
      silent               = silent
    )
  }

  attr(out, "null_hypothesis")      <- null_hypothesis
  attr(out, "normal_approximation") <- normal_approximation
  class(out) <- c(class(out), "marginal_inference")
  return(out)
}

.marginal_inference_condition_key <- function(conditional){

  paste0(c(length(conditional), conditional), collapse = "\r")
}

.marginal_inference_level_conditionals <- function(marginal, prior_list, conditional){

  levels <- names(marginal)
  conditionals <- lapply(levels, function(level){
    weights <- attr(marginal[[level]], "linear_weights")
    if(is.null(weights)){
      return(conditional)
    }
    if(!is.null(dim(weights))){
      active <- unique(unlist(apply(weights, 1, function(row_weights){
        .prior_linear_active_conditionals(
          prior_list  = prior_list,
          weights     = row_weights,
          conditional = conditional
        )
      }), use.names = FALSE))
      return(conditional[conditional %in% active])
    }
    .prior_linear_active_conditionals(
      prior_list   = prior_list,
      weights      = weights,
      conditional  = conditional
    )
  })
  names(conditionals) <- levels

  conditionals
}

.marginal_inference_conditional_posterior <- function(model, parameters, marginal_parameter,
                                                      formula, averaged_marginal, prior_list,
                                                      conditional, conditional_rule, n_samples,
                                                      force_plots){

  level_conditionals <- .marginal_inference_level_conditionals(
    marginal    = averaged_marginal,
    prior_list  = prior_list,
    conditional = conditional
  )

  conditional_marginal <- averaged_marginal
  marginal_cache <- list()

  for(level in names(level_conditionals)){
    level_conditional <- level_conditionals[[level]]
    key <- .marginal_inference_condition_key(level_conditional)

    if(is.null(marginal_cache[[key]])){
      conditional_posterior <- as_mixed_posteriors(
        model            = model,
        parameters       = parameters,
        conditional      = if(length(level_conditional) == 0) NULL else level_conditional,
        conditional_rule = conditional_rule,
        force_plots      = force_plots
      )

      if(length(conditional_posterior) == 0){
        return(list())
      }else{
        marginal_cache[[key]] <- marginal_posterior(
          samples       = conditional_posterior,
          parameter     = marginal_parameter,
          formula       = formula,
          prior_samples = TRUE,
          n_samples     = n_samples
        )
      }
    }

    if(length(marginal_cache[[key]]) == 0){
      return(list())
    }

    conditional_marginal[[level]] <- marginal_cache[[key]][[level]]
    attr(conditional_marginal[[level]], "effective_conditional") <- level_conditional
  }

  conditional_marginal
}

#' @title Kitchen Rolls data from
#' \insertCite{wagenmakers2015turning;textual}{BayesTools} replication study.
#'
#' @description The data set contains mean NEO PI-R scores for two groups of students.
#' Each of them filled a personality questionnaire while rotating a kitchen roll either
#' clock or counter-clock wise. See \insertCite{wagenmakers2015turning;textual}{BayesTools}
#' for more details about the replication study and the https://osf.io/uszvx/ for the
#' original data.
#'
#' @format A data.frame with 2 columns and 102 observations.
#'
#' @return a data.frame.
#'
#' @references
#' \insertAllCited{}
"kitchen_rolls"

#' @title Helper functions for extracting and formatting posterior distributions
#'
#' @description Internal helper functions to extract posterior samples from JAGS
#' fits and reformat them for further processing (summary tables, diagnostics, plots).
#' These functions consolidate common logic that was duplicated across
#' \code{runjags_estimates_table}, \code{.diagnostics_plot_data}, and plotting functions.
#'
#' @name posterior_extraction_helpers
#' @keywords internal
NULL


#' @rdname posterior_extraction_helpers
#' @param fit a JAGS model fit object
#' @param as_list whether to return samples as mcmc.list (TRUE) or merged matrix (FALSE)
#' @return matrix or mcmc.list of posterior samples
.extract_posterior_samples <- function(fit, as_list = FALSE) {

  if (inherits(fit, "runjags")) {
    .check_runjags()
  }
  
  if (as_list) {
    # Use generic function to allow S3 method dispatch (runjags has its own as.mcmc.list method)
    model_samples <- coda::as.mcmc.list(fit)
  } else {
    # Use generic function to allow S3 method dispatch (runjags has its own as.mcmc method)
    model_samples <- suppressWarnings(coda::as.mcmc(fit))
  }
  
  return(model_samples)
}


#' @rdname posterior_extraction_helpers
#' @param model_samples matrix of posterior samples
#' @param prior_list list of prior objects
#' @param remove_parameters character vector of parameter names to remove
#' @return list with cleaned model_samples and updated prior_list
.remove_auxiliary_parameters <- function(model_samples, prior_list, remove_parameters = NULL) {
  
  for (i in rev(seq_along(prior_list))) {
    
    par_name <- names(prior_list)[i]
    
    # invgamma support parameter
    if (is.prior.simple(prior_list[[i]]) && prior_list[[i]][["distribution"]] == "invgamma") {
      aux_names <- if(is.prior.factor(prior_list[[i]])){
        paste0("inv_", .JAGS_prior_factor_names(par_name, prior_list[[i]]))
      }else{
        paste0("inv_", par_name)
      }
      model_samples <- model_samples[, !colnames(model_samples) %in% aux_names, drop = FALSE]
    }
    
    # weightfunction parameters
    if (is.prior.weightfunction(prior_list[[i]])) {
      private_parameters <- .JAGS_monitor_private.weightfunction(prior_list[[i]])
      if(length(private_parameters) > 0){
        private_pattern <- paste0("^(", paste(private_parameters, collapse = "|"), ")(\\[|$)")
        model_samples <- model_samples[, !grepl(private_pattern, colnames(model_samples)), drop = FALSE]
      }
      
      # rename the omegas
      omega_cuts      <- weightfunctions_mapping(prior_list[i], cuts_only = TRUE)
      omega_names_old <- paste0("omega[", 1:(length(omega_cuts) - 1), "]")
      omega_names     <- sapply(1:(length(omega_cuts) - 1), function(j) paste0("omega[", omega_cuts[j], ",", omega_cuts[j + 1], "]"))
      omega_raw_names <- grep("^omega\\[[0-9]+\\]$", colnames(model_samples), value = TRUE)

      colnames(model_samples)[which(colnames(model_samples) %in% omega_names_old)] <- omega_names

      # Two-sided weightfunctions are expanded internally to mirrored one-sided
      # coefficients for JAGS. Only the public local bins belong in summaries.
      omega_extra_names <- setdiff(omega_raw_names, omega_names_old)
      if(length(omega_extra_names) > 0L){
        model_samples <- model_samples[, !colnames(model_samples) %in% omega_extra_names, drop = FALSE]
      }

      # remove omegas if requested
      if ("omega" %in% remove_parameters) {
        model_samples <- model_samples[, !colnames(model_samples) %in% omega_names, drop = FALSE]
        prior_list[i] <- NULL
      }

    } else if (is_prior_phacking(prior_list[[i]])) {
      drop_phacking <- .phacking_unreported_parameters(prior_list[[i]])
      model_samples <- model_samples[, !colnames(model_samples) %in% drop_phacking, drop = FALSE]

      if ("omega" %in% remove_parameters) {
        omega_names <- colnames(model_samples)[grepl("^omega\\[", colnames(model_samples))]
        model_samples <- model_samples[, !colnames(model_samples) %in% omega_names, drop = FALSE]
      }

      if (par_name %in% remove_parameters) {
        model_samples <- .remove_parameter_columns(model_samples, prior_list[[i]], par_name)
        prior_list[i] <- NULL
      }

    } else if (is_prior_bias(prior_list[[i]])) {
      drop_phacking <- .selection_phacking_unreported_parameters(.selection_prior_phacking_priors(prior_list[[i]]))
      model_samples <- model_samples[, !colnames(model_samples) %in% drop_phacking, drop = FALSE]

      if (.selection_prior_has_selection(prior_list[[i]])) {
        omega_cuts      <- weightfunctions_mapping(.selection_prior_selection_priors(prior_list[[i]]), cuts_only = TRUE, one_sided = TRUE)
        omega_names_old <- paste0("omega[", 1:(length(omega_cuts) - 1), "]")
        omega_names     <- sapply(1:(length(omega_cuts) - 1), function(j) paste0("omega[", omega_cuts[j], ",", omega_cuts[j + 1], "]"))
        colnames(model_samples)[which(colnames(model_samples) %in% omega_names_old)] <- omega_names
      } else {
        omega_names <- colnames(model_samples)[grepl("^omega\\[", colnames(model_samples))]
      }

      if ("omega" %in% remove_parameters) {
        model_samples <- model_samples[, !colnames(model_samples) %in% omega_names, drop = FALSE]
      }

      if (par_name %in% remove_parameters) {
        model_samples <- .remove_parameter_columns(model_samples, prior_list[[i]], par_name)
        prior_list[i] <- NULL
      }
      
    } else if (par_name %in% remove_parameters) {
      # remove parameters to be excluded (note: spike_0 removal is handled by caller)
      model_samples <- .remove_parameter_columns(model_samples, prior_list[[i]], par_name)
      prior_list[i] <- NULL
    }
  }
  
  return(list(model_samples = model_samples, prior_list = prior_list))
}


#' @rdname posterior_extraction_helpers
#' @description Helper to remove all columns associated with a parameter
#' @param model_samples matrix of posterior samples
#' @param prior prior object for the parameter
#' @param par_name name of the parameter
#' @return updated model_samples matrix
.remove_parameter_columns <- function(model_samples, prior, par_name) {
  
  # collect all column patterns to remove
  cols_to_remove <- character(0)
  
  if (is.prior.spike_and_slab(prior)) {
    # spike and slab: remove main parameter, indicator, inclusion, variable
    cols_to_remove <- c(
      par_name,
      paste0(par_name, "_indicator"),
      paste0(par_name, "_inclusion"),
      paste0(par_name, "_variable")
    )
    # also handle factor spike and slab with indexed columns
    cols_to_remove <- c(cols_to_remove, 
      colnames(model_samples)[grepl(paste0("^", par_name, "\\["), colnames(model_samples))],
      colnames(model_samples)[grepl(paste0("^", par_name, "_variable\\["), colnames(model_samples))]
    )
    
  } else if (is.prior.mixture(prior)) {
    # mixture: remove main parameter, indicator, and component-specific columns
    cols_to_remove <- c(
      par_name,
      paste0(par_name, "_indicator")
    )
    # handle factor mixture with indexed columns
    cols_to_remove <- c(cols_to_remove,
      colnames(model_samples)[grepl(paste0("^", par_name, "\\["), colnames(model_samples))]
    )
    
    # check for bias mixture (PET, PEESE, omega, p-hacking)
    if (inherits(prior, "prior.bias_mixture")) {
      cols_to_remove <- c(cols_to_remove, .selection_bias_parameter_names(prior))
      cols_to_remove <- c(cols_to_remove,
        colnames(model_samples)[grepl("^omega\\[", colnames(model_samples))]
      )
    }
    
  } else if (is_prior_phacking(prior)) {
    cols_to_remove <- c(par_name, "omega", "alpha", "pi_null", "phack_kind")
    cols_to_remove <- c(cols_to_remove,
      colnames(model_samples)[grepl("^omega\\[", colnames(model_samples))]
    )

  } else if (is_prior_bias(prior)) {
    cols_to_remove <- c(par_name, .selection_bias_parameter_names(prior), "alpha", "pi_null")
    cols_to_remove <- c(cols_to_remove,
      colnames(model_samples)[grepl("^omega\\[", colnames(model_samples))]
    )

  } else if (is.prior.factor(prior)) {
    # factor prior: remove all indexed columns
    cols_to_remove <- .JAGS_prior_factor_names(par_name, prior)
    
  } else if (is.prior.PET(prior)) {
    # PET prior: remove the PET column (samples are stored as "PET", not par_name)
    cols_to_remove <- c(par_name, "PET")
    
  } else if (is.prior.PEESE(prior)) {
    # PEESE prior: remove the PEESE column (samples are stored as "PEESE", not par_name)
    cols_to_remove <- c(par_name, "PEESE")
    
  } else {
    # simple prior: just remove the main column
    cols_to_remove <- par_name
  }
  
  # remove duplicates and filter to existing columns
  cols_to_remove <- unique(cols_to_remove)
  cols_to_remove <- cols_to_remove[cols_to_remove %in% colnames(model_samples)]
  
  model_samples <- model_samples[, !colnames(model_samples) %in% cols_to_remove, drop = FALSE]
  
  return(model_samples)
}


#' @rdname posterior_extraction_helpers
#' @param remove_parameters character vector of parameter names to remove, or TRUE to remove all non-formula parameters.
#' If "bias" is specified and the bias prior contains PET, PEESE, or weightfunction priors,
#' the corresponding parameters (PET, PEESE, omega) are also added to the removal list.
#' @param remove_formulas character vector of formula names whose parameters should be removed
#' @param keep_parameters character vector of parameter names to keep (all others removed unless in keep_formulas).
#' If "bias" is specified and the bias prior contains PET, PEESE, or weightfunction priors,
#' the corresponding parameters (PET, PEESE, omega) are also added to the keep list.
#' @param keep_formulas character vector of formula names whose parameters should be kept (all others removed unless in keep_parameters)
#' @param remove_spike_0 whether to remove spike at 0 priors
#' @return list with filtered model_samples and prior_list
.filter_parameters <- function(prior_list, remove_parameters = NULL, remove_formulas = NULL,
                               keep_parameters = NULL, keep_formulas = NULL, remove_spike_0 = TRUE) {
  
  # get formula parameter for each prior
  prior_formulas <- sapply(prior_list, function(p) {
    form <- attr(p, "parameter")
    if (is.null(form)) "__none" else form

  })
  
  # helper function to get bias-related parameters from a bias prior
  .get_bias_params <- function(prior_list, bias_name = "bias") {
    bias_params <- character(0)
    if (bias_name %in% names(prior_list)) {
      bias_params <- .selection_bias_parameter_names(prior_list[[bias_name]])
    }
    return(bias_params)
  }
  
  # initialize parameters to remove
  params_to_remove <- character(0)
  
  # handle remove_spike_0
  if (remove_spike_0) {
    spike_0_params <- names(prior_list)[sapply(seq_along(prior_list), function(i) {
      is.prior.point(prior_list[[i]]) && prior_list[[i]][["parameters"]][["location"]] == 0
    })]
    params_to_remove <- c(params_to_remove, spike_0_params)
  }
  
  # handle remove_parameters
  if (is.logical(remove_parameters) && isTRUE(remove_parameters)) {
    # remove all non-formula parameters
    non_formula_params <- names(prior_list)[prior_formulas == "__none"]
    params_to_remove <- c(params_to_remove, non_formula_params)
  } else if (is.character(remove_parameters)) {
    params_to_remove <- c(params_to_remove, remove_parameters)
    # if "bias" is in remove_parameters, also add corresponding bias-related parameters
    if ("bias" %in% remove_parameters) {
      params_to_remove <- c(params_to_remove, .get_bias_params(prior_list, "bias"))
    }
  }
  
  # handle remove_formulas
  if (!is.null(remove_formulas)) {
    formula_params <- names(prior_list)[prior_formulas %in% remove_formulas]
    params_to_remove <- c(params_to_remove, formula_params)
  }
  
  # handle keep_parameters and keep_formulas (these define what to keep, everything else is removed)
  if (!is.null(keep_parameters) || !is.null(keep_formulas)) {
    # start with all parameters as candidates for removal
    all_params <- names(prior_list)
    
    # determine which parameters to keep
    params_to_keep <- character(0)
    
    if (!is.null(keep_parameters)) {
      params_to_keep <- c(params_to_keep, keep_parameters)
      # if "bias" is in keep_parameters, also add corresponding bias-related parameters
      if ("bias" %in% keep_parameters) {
        params_to_keep <- c(params_to_keep, .get_bias_params(prior_list, "bias"))
      }
    }
    
    if (!is.null(keep_formulas)) {
      formula_params_to_keep <- names(prior_list)[prior_formulas %in% keep_formulas]
      params_to_keep <- c(params_to_keep, formula_params_to_keep)
    }
    
    # add parameters not in keep list to removal list
    params_not_kept <- all_params[!all_params %in% params_to_keep]
    params_to_remove <- c(params_to_remove, params_not_kept)
    
    # if "bias" is in params_not_kept, also add corresponding bias-related parameters
    if ("bias" %in% params_not_kept) {
      params_to_remove <- c(params_to_remove, .get_bias_params(prior_list, "bias"))
    }
  }
  
  # remove duplicates

  params_to_remove <- unique(params_to_remove)
  
  return(params_to_remove)
}


#' @rdname posterior_extraction_helpers
#' @param par parameter name
#' @param conditional whether to compute conditional summary
#' @param remove_inclusion whether to remove inclusion indicators
#' @param warnings character vector for collecting warnings
#' @return list with updated model_samples, prior_list, and warnings
.process_spike_and_slab <- function(model_samples, prior_list, par, conditional = FALSE, remove_inclusion = FALSE, warnings = NULL) {
  
  # prepare parameter names
  if (is.prior.factor(.get_spike_and_slab_variable(prior_list[[par]]))) {
    if (.get_prior_factor_levels(.get_spike_and_slab_variable(prior_list[[par]])) == 1) {
      par_names <- par
    } else {
      par_names <- paste0(par, "[", 1:.get_prior_factor_levels(.get_spike_and_slab_variable(prior_list[[par]])), "]")
    }
  } else {
    par_names <- par
  }
  
  # change the samples between conditional/averaged based on the preferences
  if (conditional) {
    # compute the number of conditional samples
    n_conditional_samples <- sum(model_samples[, colnames(model_samples) == paste0(par, "_indicator")] == 1)
    
    # replace null samples with NAs (important for later transformations)
    model_samples[model_samples[, colnames(model_samples) == paste0(par, "_indicator")] != 1, par_names] <- NA
    
    # add warnings about conditional summary
    warnings <- c(warnings, .runjags_conditional_warning(par_names, n_conditional_samples))
  }
  
  # remove the inclusion
  model_samples <- model_samples[, colnames(model_samples) != paste0(par, "_inclusion"), drop = FALSE]
  
  # remove the latent variable
  model_samples <- model_samples[, !colnames(model_samples) %in% gsub(par, paste0(par, "_variable"), par_names), drop = FALSE]
  
  # remove/rename the inclusions probabilities
  if (remove_inclusion) {
    model_samples <- model_samples[, colnames(model_samples) != paste0(par, "_indicator"), drop = FALSE]
  } else {
    colnames(model_samples)[colnames(model_samples) == paste0(par, "_indicator")] <- paste0(par, " (inclusion)")
  }
  
  # modify the parameter list (forward the parameter attribute)
  variable_component <- .get_spike_and_slab_variable(prior_list[[par]])
  attr(variable_component, "parameter") <- attr(prior_list[[par]], "parameter")
  prior_list[[par]] <- variable_component
  
  return(list(model_samples = model_samples, prior_list = prior_list, warnings = warnings))
}


#' @rdname posterior_extraction_helpers
#' @param transformations list of transformations to apply
#' @param transform_factors whether orthonormal/meandif will be transformed later
#' @return updated model_samples matrix
.apply_parameter_transformations <- function(model_samples, transformations, prior_list, transform_factors = FALSE) {
  
  if (is.null(transformations)) {
    return(model_samples)
  }
  
  for (par in names(transformations)) {
    if (!is.prior.factor(prior_list[[par]])) {
      # non-factor priors
      model_samples[, par] <- do.call(transformations[[par]][["fun"]], c(list(model_samples[, par]), transformations[[par]][["arg"]]))
    } else if ((!transform_factors && (is.prior.orthonormal(prior_list[[par]]) || is.prior.meandif(prior_list[[par]]))) || is.prior.treatment(prior_list[[par]])) {
      # treatment priors, or orthonormal/meandif that won't be transformed to differences
      par_names <- .JAGS_prior_factor_names(par, prior_list[[par]])
      
      for (i in seq_along(par_names)) {
        model_samples[, par_names[i]] <- do.call(transformations[[par]][["fun"]], c(list(model_samples[, par_names[i]]), transformations[[par]][["arg"]]))
      }
    }
  }
  
  return(model_samples)
}


#' @rdname posterior_extraction_helpers
#' @param transform_factors whether to transform orthonormal/meandif to differences
#' @return updated model_samples matrix
.transform_factor_contrasts <- function(model_samples, prior_list, transform_factors = FALSE, transformations = NULL) {
  
  factor_parameters <- names(prior_list)[sapply(prior_list, function(x) is.prior.orthonormal(x) | is.prior.meandif(x))]

  if (!transform_factors || length(factor_parameters) == 0) {
    return(model_samples)
  }

  if (any(factor_parameters %in% names(transformations))) {
    message("The transformation was applied to the differences from the mean. Note that non-linear transformations do not map from the orthonormal/meandif contrasts to the differences from the mean.")
  }

  for (par in factor_parameters) {
    
    prior_list[[par]] <- .add_factor_metadata_from_named_objects(prior_list[[par]], par, prior_list)
    par_names <- .JAGS_prior_factor_names(par, prior_list[[par]])
    
    temp_position <- min(which(colnames(model_samples) %in% par_names))
    temp_samples  <- model_samples[, colnames(model_samples) %in% par_names, drop = FALSE]
    model_samples <- model_samples[, !colnames(model_samples) %in% par_names, drop = FALSE]
    
    transformed_samples <- .transform_factor_contrast_samples(
      coefficient_samples = temp_samples,
      metadata            = prior_list[[par]],
      parameter           = par,
      transformed_class   = if(is.prior.orthonormal(prior_list[[par]])) {
        "mixed_posteriors.orthonormal_transformed"
      }else{
        "mixed_posteriors.meandif_transformed"
      }
    )
    
    # apply transformation if specified
    if (!is.null(transformations[[par]])) {
      for (i in seq_len(ncol(transformed_samples))) {
        transformed_samples[, i] <- do.call(transformations[[par]][["fun"]], c(list(transformed_samples[, i]), transformations[[par]][["arg"]]))
      }
    }
    
    # place the transformed samples back
    model_samples <- cbind(
      if (temp_position > 1) model_samples[, 1:(temp_position - 1), drop = FALSE],
      transformed_samples,
      if (temp_position <= ncol(model_samples)) model_samples[, temp_position:ncol(model_samples), drop = FALSE]
    )
  }
  
  return(model_samples)
}


# Helper: Attach factor levels to the factor components inside an interaction
# term instead of appending the level to the full interaction string.
.format_factor_level_parameter_names <- function(parameter, level_names, n_parameters = NULL) {

  if (!is.list(level_names)) {
    return(paste0(parameter, "[", level_names, "]"))
  }

  parameter_terms <- strsplit(parameter, "__xXx__", fixed = TRUE)[[1]]
  factor_terms <- names(level_names)
  factor_positions <- vapply(factor_terms, function(factor_term) {
    factor_position <- which(
      parameter_terms == factor_term |
        endsWith(parameter_terms, paste0("_", factor_term))
    )

    if (length(factor_position) == 1) {
      return(factor_position)
    }

    return(NA_integer_)
  }, integer(1))

  if (all(!is.na(factor_positions))) {
    level_grid <- expand.grid(level_names, KEEP.OUT.ATTRS = FALSE, stringsAsFactors = FALSE)
    formatted_names <- vapply(seq_len(nrow(level_grid)), function(i) {
      formatted_terms <- parameter_terms

      for (factor_term in factor_terms) {
        formatted_terms[factor_positions[[factor_term]]] <- sub(
          paste0(factor_term, "$"),
          paste0(factor_term, "[", level_grid[[factor_term]][i], "]"),
          formatted_terms[factor_positions[[factor_term]]]
        )
      }

      paste0(formatted_terms, collapse = "__xXx__")
    }, character(1))

    if (is.null(n_parameters) || length(formatted_names) == n_parameters) {
      return(formatted_names)
    }
  }

  fallback_names <- paste0(parameter, "[", unlist(level_names, use.names = FALSE), "]")
  if (!is.null(n_parameters) && length(fallback_names) != n_parameters) {
    return(paste0(parameter, "[", seq_len(n_parameters), "]"))
  }

  return(fallback_names)
}


#' @rdname posterior_extraction_helpers
#' @return updated model_samples matrix with renamed columns
.rename_factor_levels <- function(model_samples, prior_list) {
  
  # rename treatment factor levels
  if (any(sapply(prior_list, is.prior.treatment))) {
    for (par in names(prior_list)[sapply(prior_list, is.prior.treatment)]) {
      if (!.is_prior_interaction(prior_list[[par]])) {
        renamed_levels <- .format_factor_level_parameter_names(
          par,
          .get_prior_factor_level_names(prior_list[[par]])[-1],
          .get_prior_factor_levels(prior_list[[par]])
        )
        if (.get_prior_factor_levels(prior_list[[par]]) == 1) {
          colnames(model_samples)[colnames(model_samples) == par] <-
            renamed_levels[1]
        } else {
          colnames(model_samples)[colnames(model_samples) %in% paste0(par, "[", 1:.get_prior_factor_levels(prior_list[[par]]), "]")] <-
            renamed_levels
        }
      } else if (length(attr(prior_list[[par]], "levels")) == 1) {
        interaction_level_names <- .get_prior_factor_level_names(prior_list[[par]])
        interaction_level_names <- lapply(interaction_level_names, function(level_name) level_name[-1])
        colnames(model_samples)[colnames(model_samples) %in% paste0(par, "[", 1:.get_prior_factor_levels(prior_list[[par]]), "]")] <-
          .format_factor_level_parameter_names(par, interaction_level_names, .get_prior_factor_levels(prior_list[[par]]))
      }
    }
  }
  
  # rename independent factor levels
  if (any(sapply(prior_list, is.prior.independent))) {
    for (par in names(prior_list)[sapply(prior_list, is.prior.independent)]) {
      if (!.is_prior_interaction(prior_list[[par]])) {
        renamed_levels <- .format_factor_level_parameter_names(
          par,
          .get_prior_factor_level_names(prior_list[[par]]),
          .get_prior_factor_levels(prior_list[[par]])
        )
        if (.get_prior_factor_levels(prior_list[[par]]) == 1) {
          colnames(model_samples)[colnames(model_samples) == par] <-
            renamed_levels[1]
        } else {
          colnames(model_samples)[colnames(model_samples) %in% paste0(par, "[", 1:.get_prior_factor_levels(prior_list[[par]]), "]")] <-
            renamed_levels
        }
      } else if (length(attr(prior_list[[par]], "levels")) == 1) {
        colnames(model_samples)[colnames(model_samples) %in% paste0(par, "[", 1:.get_prior_factor_levels(prior_list[[par]]), "]")] <-
          .format_factor_level_parameter_names(par, .get_prior_factor_level_names(prior_list[[par]]), .get_prior_factor_levels(prior_list[[par]]))
      }
    }
  }
  
  return(model_samples)
}

#' @title Prior density
#'
#' @description Computes density of a prior
#' distribution across a range of values.
#'
#' @param x a prior
#' @param x_seq sequence of x coordinates
#' @param x_range vector of length two with
#' lower and upper range for the support
#' (used if \code{x_seq} is unspecified)
#' @param x_range_quant quantile used for
#' automatically obtaining \code{x_range}
#' if both \code{x_range} and \code{x_seq}
#' are unspecified. Defaults to \code{0.005}
#' for all but Cauchy, Student-t, Gamma, and
#' Inverse-gamme distributions that use
#' \code{0.010}.
#' @param n_points number of equally spaced points
#' in the \code{x_range} if \code{x_seq} is unspecified
#' @param n_samples number of samples from the prior
#' distribution if the density cannot be obtained
#' analytically (or if samples are forced with
#' \code{force_samples = TRUE})
#' @param force_samples should prior be sampled instead
#' of obtaining analytic solution whenever possible
#' @param individual should individual densities be returned
#' (e.g., in case of weightfunction)
#' @param transformation transformation to be applied
#' to the prior distribution. Either a character
#' specifying one of the prepared transformations:
#' \describe{
#'   \item{lin}{linear transformation in form of \code{a + b*x}}
#'   \item{tanh}{hyperbolic tangent transformation}
#'   \item{exp}{exponential transformation}
#' }, or a list containing the transformation function \code{fun},
#' inverse transformation function \code{inv}, and derivative of the
#' transformation \code{jac}, evaluated on the original support. See examples
#' for details.
#' @param transformation_arguments a list with named arguments for
#' the \code{transformation}
#' @param transformation_settings boolean indicating whether the
#' settings the \code{x_seq} or \code{x_range} was specified on
#' the transformed support
#' @param truncate_end whether the density should be set to zero in
#' for the endpoints of truncated distributions
#' @param ... additional arguments
#'
#' @return \code{density.prior} returns an object of class 'density'.
#'
#' @importFrom stats density
#' @seealso [prior()]
#' @rdname density.prior
#' @export
density.prior <- function(x,
                          x_seq = NULL, x_range = NULL, x_range_quant = NULL, n_points = 1000,
                          n_samples = 10000, force_samples = FALSE, individual = FALSE,
                          transformation = NULL, transformation_arguments = NULL, transformation_settings = FALSE, truncate_end = TRUE, ...){

  # input check
  .check_prior(x, "x")
  check_real(x_seq, "x_seq", check_length = 0, allow_NULL = TRUE)
  check_real(x_range, "x_range", check_length = 2, allow_NULL = TRUE)
  if(!is.null(x_range) && x_range[1] > x_range[2])
    stop("The lower range limit must be lower than the upper range limit.")
  check_real(x_range_quant, "x_range_quant", lower = 0, upper = 1, allow_NULL = TRUE)
  check_int(n_points, "n_points",  lower = 2)
  check_int(n_samples, "n_samples", lower = 1)
  check_bool(force_samples, "force_samples")
  check_bool(individual, "individual")
  .check_transformation_input(transformation, transformation_arguments, transformation_settings)
  check_bool(truncate_end, "truncate_end")


  ### setting the range
  # get plotting range if not specified
  if(is.null(x_range)){
    if(!is.null(x_seq)){
      x_range <- range(x_seq)
    }else{
      if(is_prior_phacking(x) || is_prior_bias(x)){
        .selection_prior_stop_unsupported_generic("density", x)
      }else if(!individual & (is.prior.PET(x) | is.prior.PEESE(x))){
        x_range <- c(0, 1)
      }else if(!individual & is.prior.weightfunction(x)){
        x_range <- c(0, 1)
      }else if(is.prior.spike_and_slab(x)){
        x_range <- range(c(range(.get_spike_and_slab_variable(x), if(is.null(x_range_quant)) .range.prior_quantile_default(.get_spike_and_slab_variable(x)) else x_range_quant), 0))
      }else if(is.prior.discrete(x)){
        x_range <- c(x[["truncation"]][["lower"]], x[["truncation"]][["upper"]])
      }else{
        x_range <- range(x, if(is.null(x_range_quant)) .range.prior_quantile_default(x) else x_range_quant)
      }
    }
  }

  # get the x_seq for plotting
  if(is.null(x_seq)){
    if(is.prior.discrete(x)){
      x_seq <- seq(x_range[1], x_range[2], by = 1)
    }else{
      x_seq <- seq(x_range[1], x_range[2], length.out = n_points)
    }
  }

  # specify it on the transformed range if requested
  if(transformation_settings & !is.null(transformation)){
    x_seq   <- .density.prior_transformation_inv_x(x_seq,   transformation, transformation_arguments)
    x_range <- .density.prior_transformation_inv_x(x_range, transformation, transformation_arguments)
  }


  # use the corresponding density subfunction
  if(is.prior.weightfunction(x)){
    out <- .density.prior.weightfunction(x, x_seq, x_range, n_points, n_samples, force_samples, individual)
  }else if(is_prior_phacking(x) || is_prior_bias(x)){
    .selection_prior_stop_unsupported_generic("density", x)
  }else if(is.prior.PET(x) | is.prior.PEESE(x)){
    out <- .density.prior.PETPEESE(x, x_seq, x_range, n_points, n_samples, force_samples, individual, transformation, transformation_arguments, truncate_end)
  }else if(is.prior.spike_and_slab(x)){
    out <- .density.prior.spike_and_slab(x, x_seq, x_range, n_points, n_samples, force_samples, transformation, transformation_arguments, truncate_end)
  }else if(is.prior.point(x)){
    out <- .density.prior.point(x, x_seq, x_range, n_points, n_samples, force_samples, transformation, transformation_arguments)
  }else if(is.prior.orthonormal(x) | is.prior.meandif(x)){
    out <- .density.prior.orthonormal_or_meandif(x, x_seq, x_range, n_points, n_samples, force_samples, transformation, transformation_arguments, truncate_end)
  }else if(is.prior.simple(x)){
    out <- .density.prior.simple(x, x_seq, x_range, n_points, n_samples, force_samples, transformation, transformation_arguments, truncate_end)
  }

  if(!is.null(transformation)){
    attr(out, "transformation") <- transformation
  }

  return(out)
}

.density.prior.simple                 <- function(x, x_seq, x_range, n_points, n_samples, force_samples, transformation, transformation_arguments, truncate_end){

  # get the samples to estimate density / obtain the density directly
  if(force_samples | .density.prior_need_samples(x)){
    x_sam <- rng(x, n_samples)
    if(is.prior.discrete(x)){
      x_seq <- unique(round(x_seq))
      x_den <- vapply(x_seq, function(x_i) mean(x_sam == x_i), numeric(1))
    }else{
      x_density <- stats::density(x_sam, n = n_points, from = x_range[1], to = x_range[2])
      x_seq     <- x_density$x
      x_den     <- x_density$y
    }
  }else{

    if(is.prior.discrete(x)){
      x_seq <- unique(round(x_seq))
    }

    x_den <- mpdf(x, x_seq)
    x_sam <- NULL
  }


  # set the endpoints to zero if they correspond to truncation
  if(truncate_end){
    if(isTRUE(all.equal(x$truncation[["lower"]], x_seq[1])) | x$truncation[["lower"]] >= x_seq[1]){
      x_den <- c(0, x_den)
      x_seq <- c(x_seq[1], x_seq)
    }
    if(isTRUE(all.equal(x$truncation[["upper"]], x_seq[length(x_seq)])) | x$truncation[["upper"]] <= x_seq[length(x_seq)]){
      x_den <- c(x_den, 0)
      x_seq <- c(x_seq, x_seq[length(x_seq)])
    }
  }



  # transform the output, if requested
  if(!is.null(transformation)){
    x_seq   <- .density.prior_transformation_x(x_seq,   transformation, transformation_arguments)
    x_range <- .density.prior_transformation_x(x_range, transformation, transformation_arguments)
    if(!is.null(x_sam)){
      x_sam <- .density.prior_transformation_x(x_sam,   transformation, transformation_arguments)
    }
    x_den   <- .density.prior_transformation_y(x_seq, x_den, transformation, transformation_arguments)
  }


  # create the output object
  out <- list(
    call    = call("density", print(x, silent = TRUE)),
    bw      = NULL,
    n       = n_points,
    x       = x_seq,
    y       = x_den,
    samples = x_sam
  )


  class(out) <- c("density", "density.prior", "density.prior.simple")
  attr(out, "x_range") <- x_range
  attr(out, "y_range") <- c(0, max(x_den))

  return(out)
}
.density.prior.point                  <- function(x, x_seq, x_range, n_points, n_samples, force_samples, transformation, transformation_arguments){

  # return the samples if requested
  if(force_samples){
    x_sam <- rng(x, n_samples)
  }else{
    x_sam <- NULL
  }

  x_seq <- x$parameters[["location"]]
  x_den <- 1


  # transform the output, if requested
  if(!is.null(transformation)){
    x_seq   <- .density.prior_transformation_x(x_seq,   transformation, transformation_arguments)
    x_range <- .density.prior_transformation_x(x_range, transformation, transformation_arguments)
    if(!is.null(x_sam)){
      x_sam <- .density.prior_transformation_x(x_sam,   transformation, transformation_arguments)
    }
  }


  # create the output object
  out <- list(
    call    = call("density", print(x, silent = TRUE)),
    bw      = NULL,
    n       = n_points,
    x       = x_seq,
    y       = x_den,
    samples = x_sam
  )


  class(out) <- c("density", "density.prior", "density.prior.point", if(is.prior.orthonormal(x)) "density.prior.orthonormal" else if(is.prior.meandif(x)) "density.prior.meandif")
  attr(out, "x_range") <- x_range
  attr(out, "y_range") <- c(0, max(x_den))

  return(out)
}
.density.prior.weightfunction         <- function(x, x_seq, x_range, n_points, n_samples, force_samples, individual){

  # create either distribution for the individual weights or the whole weightfunction
  if(individual){

    if(force_samples | .density.prior_need_samples(x)){
      x_sam <- rng(x, n_samples)
      densities <- lapply(1:ncol(x_sam), function(i) stats::density(x_sam[,i], n = n_points, from = x_range[1], to = x_range[2]))
      x_seq     <- densities[[1]]$x
      x_den     <- do.call(cbind, lapply(densities, `[[`, "y"))
    }else{
      x_den <- mpdf(x, x_seq)
      x_sam <- NULL
    }

    # set the endpoints to zero if they correspond to truncation
    if(isTRUE(all.equal(x$truncation[["lower"]], x_seq[1])) | x$truncation[["lower"]] >= x_seq[1]){
      x_den <- rbind(0, x_den)
      x_seq <- c(x_seq[1], x_seq)
    }
    if(isTRUE(all.equal(x$truncation[["upper"]], x_seq[length(x_seq)])) | x$truncation[["upper"]] <= x_seq[length(x_seq)]){
      x_den <- rbind(x_den, 0)
      x_seq <- c(x_seq, x_seq[length(x_seq)])
    }

    out <- list()
    out_types <- .density.prior_type(x)
    components <- .weightfunction_marginal_components(x)

    for(i in 1:ncol(x_den)){

      temp_samples <- if(is.null(x_sam)) NULL else x_sam[,i]

      # create the output object
      if(out_types[i] == "point"){
        temp_out <- list(
          call    = call("density", print(x, silent = TRUE)),
          bw      = NULL,
          n       = n_points,
          x       = components[[i]]$location,
          y       = 1,
          samples = temp_samples
        )
      }else{
        temp_out <- list(
          call    = call("density", print(x, silent = TRUE)),
          bw      = NULL,
          n       = n_points,
          x       = x_seq,
          y       = x_den[,i],
          samples = temp_samples
        )
      }


      class(temp_out) <- c("density", "density.prior", paste0("density.prior.",out_types[i]))
      attr(temp_out, "x_range") <- x_range
      attr(temp_out, "y_range") <- if(out_types[i] == "point") c(0, 1) else c(0, max(x_den[,i]))
      attr(temp_out, "steps")   <- c(x$bins$lower[i], x$bins$upper[i])

      out[[i]] <- temp_out
    }

  }else{

    # weightfunction specific stuff
    x_seq     <- .weightfunction_local_cuts(x)
    x_seq_rep <- c(1, sort(rep(2:(length(x_seq)-1), 2)) ,length(x_seq))
    x_val_rep <- sort(rep(1:(length(x_seq)-1), 2))
    if(force_samples | .density.prior_need_samples(x)){
      x_sam  <- rng(x, n_samples)
      x_lCI  <- apply(x_sam, 2, stats::quantile, probs = .025)
      x_uCI  <- apply(x_sam, 2, stats::quantile, probs = .975)
      x_mean <- apply(x_sam, 2, mean)
    }else{
      x_sam  <- NULL
      x_lCI  <- mquant(x, .025)
      x_uCI  <- mquant(x, .975)
      x_mean <- mean(x)
    }

    out <- list(
      call    = call("density", print(x, silent = TRUE)),
      bw      = NULL,
      n       = n_points,
      x       = x_seq[x_seq_rep],
      y       = x_mean[x_val_rep],
      y_lCI   = x_lCI[x_val_rep],
      y_uCI   = x_uCI[x_val_rep],
      samples = x_sam
    )


    class(out) <- c("density", "density.prior", "density.prior.weightfunction")
    attr(out, "x_range") <- c(0, 1)
    attr(out, "y_range") <- c(0, max(1, x_mean, x_lCI, x_uCI, na.rm = TRUE))
  }

  return(out)
}
.density.prior.PETPEESE               <- function(x, x_seq, x_range, n_points, n_samples, force_samples, individual, transformation, transformation_arguments, truncate_end){

  # create either distribution for the parameter or the PET/PEESE function
  if(individual){

    if(is.prior.point(x)){
      out <- .density.prior.point(x, x_seq, x_range, n_points, n_samples, force_samples, transformation, transformation_arguments)
    }else if(is.prior.simple(x)){
      out <- .density.prior.simple(x, x_seq, x_range, n_points, n_samples, force_samples, transformation, transformation_arguments, truncate_end)
    }

  }else{

    # get the samples to estimate density / obtain the density directly
    if(force_samples | .density.prior_need_samples(x)){
      x_sam  <- rng(x, n_samples)
      x_med  <- stats::quantile(x_sam, .500)
      x_lCI  <- stats::quantile(x_sam, .025)
      x_uCI  <- stats::quantile(x_sam, .975)
    }else{
      x_med  <- quant(x, .500)
      x_lCI  <- quant(x, .025)
      x_uCI  <- quant(x, .975)
      x_sam  <- NULL
    }


    # transform the output, if requested
    if(!is.null(transformation)){
      x_med   <- .density.prior_transformation_x(x_med,   transformation, transformation_arguments)
      x_lCI   <- .density.prior_transformation_x(x_lCI,   transformation, transformation_arguments)
      x_uCI   <- .density.prior_transformation_x(x_uCI,   transformation, transformation_arguments)
      x_range <- .density.prior_transformation_x(x_range, transformation, transformation_arguments)
      if(!is.null(x_sam)){
        x_sam <- .density.prior_transformation_x(x_sam,   transformation, transformation_arguments)
      }
    }


    # compute the PET/PEESE
    if(is.prior.PET(x)){
      y_med   = x_med  * x_seq
      y_lCI   = x_lCI  * x_seq
      y_uCI   = x_uCI  * x_seq
    }else if(is.prior.PEESE(x)){
      y_med   = x_med  * x_seq^2
      y_lCI   = x_lCI  * x_seq^2
      y_uCI   = x_uCI  * x_seq^2
    }


    out <- list(
      call    = call("density", print(x, silent = TRUE)),
      bw      = NULL,
      n       = n_points,
      x       = x_seq,
      y       = y_med,
      y_lCI   = y_lCI,
      y_uCI   = y_uCI,
      samples = x_sam
    )


    class(out) <- c("density", "density.prior", if(is.prior.PET(x)) "density.prior.PET" else if(is.prior.PEESE(x)) "density.prior.PEESE")
    attr(out, "x_range") <- range(x_seq)
    attr(out, "y_range") <- range(y_med)
  }

  return(out)
}
.density.prior.orthonormal_or_meandif <- function(x, x_seq, x_range, n_points, n_samples, force_samples, transformation, transformation_arguments, truncate_end){

  # get the samples to estimate density / obtain the density directly
  if(force_samples | .density.prior_need_samples(x)){

    if(is.na(x$parameters[["K"]]) && !is.null(attr(x, "levels"))){
      x$parameters[["K"]] <- .get_prior_factor_levels(x)
    }else if(is.na(x$parameters[["K"]])){
      x$parameters[["K"]] <- 1
      warning("number of factor levels / dimensionality of the prior distribution was not specified -- assuming two factor levels")
    }

    x_sam <- rng(x, n_samples)
    x_sam <- as.vector(x_sam)
    x_density <- stats::density(x_sam, n = n_points, from = x_range[1], to = x_range[2])
    x_seq     <- x_density$x
    x_den     <- x_density$y

  }else{
    x_den <- mpdf(x, x_seq)
    x_sam <- NULL
  }


  # set the endpoints to zero if they correspond to truncation
  if(truncate_end){
    if(isTRUE(all.equal(x$truncation[["lower"]], x_seq[1])) | x$truncation[["lower"]] >= x_seq[1]){
      x_den <- c(0, x_den)
      x_seq <- c(x_seq[1], x_seq)
    }
    if(isTRUE(all.equal(x$truncation[["upper"]], x_seq[length(x_seq)])) | x$truncation[["upper"]] <= x_seq[length(x_seq)]){
      x_den <- c(x_den, 0)
      x_seq <- c(x_seq, x_seq[length(x_seq)])
    }
  }


  # transform the output, if requested
  if(!is.null(transformation)){
    message("The transformation was applied to the differences from the mean. Note that non-linear transformations do not map from the orthonormal/meandif contrasts to the differences from the mean.")
    x_seq   <- .density.prior_transformation_x(x_seq,   transformation, transformation_arguments)
    x_range <- .density.prior_transformation_x(x_range, transformation, transformation_arguments)
    if(!is.null(x_sam)){
      x_sam <- .density.prior_transformation_x(x_sam,   transformation, transformation_arguments)
    }
    x_den   <- .density.prior_transformation_y(x_seq, x_den, transformation, transformation_arguments)
  }


  # create the output object
  out <- list(
    call    = call("density", print(x, silent = TRUE)),
    bw      = NULL,
    n       = n_points,
    x       = x_seq,
    y       = x_den,
    samples = x_sam
  )


  class(out) <- c("density", "density.prior", if(is.prior.orthonormal(x)) "density.prior.orthonormal" else if(is.prior.meandif(x)) "density.prior.meandif")
  attr(out, "x_range") <- x_range
  attr(out, "y_range") <- c(0, max(x_den))

  return(out)
}
.density.prior.spike_and_slab         <- function(x, x_seq, x_range, n_points, n_samples, force_samples, transformation, transformation_arguments, truncate_end){

  density_variable  <- .density.prior.simple(.get_spike_and_slab_variable(x), x_seq, x_range, n_points, n_samples, force_samples, transformation, transformation_arguments, truncate_end)
  density_inclusion <- .density.prior.point(prior(distribution = "spike", parameters = list(location = 0)), x_seq, x_range, n_points, n_samples, force_samples, transformation, transformation_arguments)

  inclusion_prob <- mean(.get_spike_and_slab_inclusion(x))
  density_variable$y  <- density_variable[["y"]]  * inclusion_prob
  density_inclusion$y <- density_inclusion[["y"]] * (1 - inclusion_prob)

  attr(density_variable,  "y_range") <- attr(density_variable, "y_range")  * inclusion_prob
  attr(density_inclusion, "y_range") <- attr(density_inclusion, "y_range") * (1 - inclusion_prob)

  # create the output object
  out <- list(
    call      = call("density", print(x, silent = TRUE)),
    variable  = density_variable,
    inclusion = density_inclusion
  )


  class(out) <- c("density", "density.prior.spike_and_slab")
  attr(out, "x_range") <- range(c(attr(density_variable, "x_range"), attr(density_inclusion, "x_range")))
  attr(out, "y_range_variable")  <- attr(density_variable,  "y_range")
  attr(out, "y_range_inclusion") <- attr(density_inclusion, "y_range")

  return(out)
}


#' @title Prior range
#'
#' @description Computes range of a prior
#' distribution (if the prior distribution is
#' unbounded range from \code{quantiles} to
#' \code{1 -quantiles}) is returned.
#'
#' @param x a prior
#' @param quantiles quantile to be returned in
#' case of unbounded distribution.
#' @param ... additional arguments
#' @param na.rm unused
#'
#' @return \code{range.prior} returns a numeric vector of
#' length with a plotting range of a prior distribution.
#'
#' @seealso [prior()]
#' @rdname range.prior
#' @export
range.prior  <- function(x, quantiles = NULL, ..., na.rm = FALSE){

  .check_prior(x)
  if(!is.null(quantiles)){
    check_real(quantiles, "quantiles", upper = 0.5, allow_bound = FALSE)
  }else{
    quantiles <- .range.prior_quantile_default(x)
  }


  x_range <- c(NA, NA)

  if(is.prior.weightfunction(x)){
    return(.weightfunction_range(x, quantiles))
  }
  if(is_prior_phacking(x) || is_prior_bias(x)){
    .selection_prior_stop_unsupported_generic("range", x)
  }

  if(is.infinite(x[["truncation"]][["lower"]])){
    x_range[1] <- mquant(x, quantiles)
  }else{
    x_range[1] <- x[["truncation"]][["lower"]]
  }

  if(is.infinite(x[["truncation"]][["upper"]])){
    x_range[2] <- mquant(x, 1 - quantiles)
  }else{
    x_range[2] <- x[["truncation"]][["upper"]]
  }

  return(x_range)
}



# helper functions
.density.prior_need_samples   <- function(prior){

  return(FALSE)
}
.density.prior_type           <- function(prior){
  if(is.prior.point(prior)){
    return("point")
  }else if(is.prior.simple(prior)){
    return("simple")
  }else if(is.prior.weightfunction(prior)){
    components <- .weightfunction_marginal_components(prior)
    return(vapply(components, function(component){
      if(component$type == "point") "point" else "simple"
    }, character(1)))
  }else if(is_prior_phacking(prior) || is_prior_bias(prior)){
    .selection_prior_stop_unsupported_generic("density", prior)
  }else if(is.prior.orthonormal(prior)){
    return("orthonormal")
  }else if(is.prior.meandif(prior)){
    return("meandif")
  }
}
.range.prior_quantile_default <- function(prior){

  switch(
    prior[["distribution"]],
    "normal"    = .005,
    "lognormal" = .005,
    "t"         = .010,
    "gamma"     = .010,
    "invgamma"  = .010,
    "beta"      = .005,
    "exp"       = .005,
    "uniform"   = .005,
    "point"     = .005,
    "weightfunction" = .005,
    "mnormal"    = .005,
    "mt"         = .010
  )

}

# transformation functions
.density.prior_transformation_x         <- function(x, transformation, transformation_arguments = NULL){

  arg <- list(x = x)
  for(i in seq_along(transformation_arguments)){
    arg[[names(transformation_arguments)[i]]] <- transformation_arguments[[i]]
  }

  do.call(.density.prior_transformation_functions(transformation)$fun, arg)
}
.density.prior_transformation_inv_x     <- function(x, transformation, transformation_arguments = NULL){

  arg <- list(x = x)
  for(i in seq_along(transformation_arguments)){
    arg[[names(transformation_arguments)[i]]] <- transformation_arguments[[i]]
  }

  do.call(.density.prior_transformation_functions(transformation)$inv, arg)
}
.density.prior_transformation_y         <- function(x, y, transformation, transformation_arguments = NULL){

  x_inv <- .density.prior_transformation_inv_x(x, transformation, transformation_arguments)
  arg <- list(x = x_inv)
  for(i in seq_along(transformation_arguments)){
    arg[[names(transformation_arguments)[i]]] <- transformation_arguments[[i]]
  }

  y / abs(do.call(.density.prior_transformation_functions(transformation)$jac, arg))
}
.density.prior_transformation_functions <- function(transformation){

  if(is.character(transformation) & length(transformation) == 1){

    return(switch(
      transformation,
      "lin" = list(
        fun = function(x, a = 0, b = 1)a + b * x,
        inv = function(x, a = 0, b = 1)(x - a) / b,
        jac = function(x, a = 0, b = 1)b
      ),
      "exp_lin" = list(
        # Exponential-linear transformation: exp(a + b * log(x))
        # Used for log-intercept unscaling where: intercept_orig = exp(log(intercept_z) * b + a)
        # When a = 0 and b = 1, this is identity: exp(log(x)) = x
        fun = function(x, a = 0, b = 1) exp(a + b * log(x)),
        inv = function(x, a = 0, b = 1) exp((log(x) - a) / b),
        jac = function(x, a = 0, b = 1) b * exp(a + b * log(x)) / x
      ),
      "tanh" = list(
        fun = tanh,
        inv = atanh,
        jac = function(x)1 - tanh(x)^2
      ),
      "exp"  = list(
        fun = exp,
        inv = log,
        jac = exp
      )
    ))

  }else if(is.list(transformation) & length(transformation) == 3 & all(names(transformation) %in% c("fun", "inv", "jac"))){

    return(transformation)

  }else{

    stop("Transformation must be either a character vector of length 1 corresponding to one of known transformations ('lin' = linear, 'exp_lin' = exponential-linear for log-intercept, 'tanh' = hyperbolic tangent, 'exp' = exponential) or a list of three functions (fun = transformation function, inv = inverse transformation, jac = derivative of the transformation).")

  }

}

#' @title Plot diagnostics of a 'JAGS' model
#'
#' @description Creates density plots, trace plots, and autocorrelation plots
#' for a given parameter of a JAGS model.
#'
#' @param fit a JAGS model fitted via [JAGS_fit()]
#' @param parameter parameter to be plotted
#' @param type what type of model diagnostic should be plotted. The available
#' options are \code{"density"}, \code{"trace"}, and \code{"autocorrelation"}
#' @param plot_type whether to use a base plot \code{"base"}
#' or ggplot2 \code{"ggplot"} for plotting.
#' @param xlim x plotting range
#' @param ylim y plotting range
#' @param lags number of lags to be shown for the autocorrelation plot.
#' Defaults to \code{30}.
#' @param ... additional arguments
#' @inheritParams density.prior
#' @inheritParams plot.prior
#' @inheritParams print.prior
#' @inheritParams BayesTools_model_tables
#'
#' @seealso [JAGS_fit()] [JAGS_check_convergence()]
#'
#' @return \code{diagnostics} returns either \code{NULL} if \code{plot_type = "base"}
#' or an object/list of objects (depending on the number of parameters to be plotted)
#' of class 'ggplot2' if \code{plot_type = "ggplot2"}.
#'
#' @name JAGS_diagnostics
#' @aliases JAGS_diagnostics_density JAGS_diagnostics_autocorrelation JAGS_diagnostics_trace
#' @export JAGS_diagnostics
#' @export JAGS_diagnostics_density
#' @export JAGS_diagnostics_autocorrelation
#' @export JAGS_diagnostics_trace

#' @rdname JAGS_diagnostics
JAGS_diagnostics                 <- function(fit, parameter, type, plot_type = "base",
                                             xlim = NULL, ylim = NULL, lags = 30, n_points = 1000,
                                             transformations = NULL, transform_factors = FALSE, transform_orthonormal = FALSE,
                                             short_name = FALSE, parameter_names = FALSE, formula_prefix = TRUE, ...){


  # check fits
  if(!inherits(fit, "runjags"))
    stop("'fit' must be a runjags fit")
  if(!inherits(fit, "BayesTools_fit"))
    stop("'fit' must be a BayesTools fit")

  check_char(type, "type", allow_values = c("density", "trace", "autocorrelation"))
  check_char(plot_type, "plot_type", allow_values = c("base", "ggplot"))
  prior_list <- attr(fit, "prior_list")
  check_list(prior_list, "prior_list")
  if(!all(sapply(prior_list, is.prior)))
    stop("'prior_list' must be a list of priors.")
  check_char(parameter, "parameter", allow_values = c(names(prior_list), if(any(names(prior_list) == "bias")) c("PET", "PEESE", "omega", "alpha", "pi_null")))
  if(!is.null(transformations))
    check_char(names(transformations), "names(transformations)", allow_values = parameter)
  if(parameter == "omega" && !is.null(prior_list[["bias"]]) && !.selection_prior_has_selection(prior_list[["bias"]]))
    stop("'omega' diagnostics require at least one weightfunction component in the 'bias' prior.", call. = FALSE)
  if(parameter %in% c("alpha", "pi_null") && !is.null(prior_list[["bias"]]) && !.selection_prior_has_phacking(prior_list[["bias"]]))
    stop("'alpha' and 'pi_null' diagnostics require at least one p-hacking component in the 'bias' prior.", call. = FALSE)

  # do not produce diagnostics for a spike prior
  if(is.prior.point(prior_list[[parameter]])){
    message("No diagnostic plots are produced for a spike prior distribution")
    return(NULL)
  }

  # depreciate
  transform_factors <- .depreciate.transform_orthonormal(transform_orthonormal, transform_factors)

  # prepare the plot data
  plot_data <- .diagnostics_plot_data(fit = fit, parameter = parameter, prior_list = prior_list, transformations = transformations, transform_factors = transform_factors)
  plot_data <- switch(
    type,
    "density"         = .diagnostics_plot_data_density(plot_data, n_points, xlim),
    "trace"           = .diagnostics_plot_data_trace(plot_data, n_points, ylim),
    "autocorrelation" = .diagnostics_plot_data_autocorrelation(plot_data, n_points, lags)
  )


  # prepare nice parameter names
  display_prior <- if(parameter %in% c("PET", "PEESE", "omega", "alpha", "pi_null") && !is.null(prior_list[["bias"]])){
    prior_list[["bias"]]
  }else{
    prior_list[[parameter]]
  }
  if(!is.null(attr(display_prior, "parameter"))){
    parameter_name <- format_parameter_names(attr(plot_data, "parameter_name"), attr(display_prior, "parameter"), formula_prefix = formula_prefix)
  }else{
    parameter_name <- attr(plot_data, "parameter_name")
  }

  # get default plot settings
  dots      <- list(...)
  main      <- if(parameter %in% c("alpha", "pi_null")){
    parameter
  }else{
    print(display_prior, plot = TRUE, short_name = short_name, parameter_names = parameter_names)
  }
  xlab      <- switch(
    type,
    "density"         = parameter_name,
    "trace"           = "Iteration",
    "autocorrelation" = "Lag"
  )
  ylab      <- switch(
    type,
    "density"         = "Density",
    "trace"           = parameter_name,
    "autocorrelation" = paste0("Autocorrelation(", parameter_name, ")")
  )


  if(is.null(dots[["main"]])) dots$main <-  main
  if(is.null(dots[["xlab"]])) dots$xlab <-  xlab
  if(is.null(dots[["ylab"]])) dots$ylab <-  ylab

  if(is.null(dots[["col"]]))  dots$col  <-  "black"
  if(is.null(dots[["lty"]]))  dots$lty  <-  1
  if(is.null(dots[["lwd"]]))  dots$lwd  <-  1

  if(length(dots[["col"]]) == 1)  dots$col  <- rep(dots$col, attr(plot_data, "chains"))
  if(length(dots[["lty"]]) == 1)  dots$lty  <- rep(dots$lty, attr(plot_data, "chains"))
  if(length(dots[["lwd"]]) == 1)  dots$lwd  <- rep(dots$lwd, attr(plot_data, "chains"))

  plots <- list()

  for(i in seq_along(plot_data)){

    temp_dots <- dots

    temp_dots[["xlab"]] <- if(length(temp_dots[["xlab"]]) > 1) temp_dots[["xlab"]][i] else temp_dots[["xlab"]]
    temp_dots[["ylab"]] <- if(length(temp_dots[["ylab"]]) > 1) temp_dots[["ylab"]][i] else temp_dots[["ylab"]]

    if(is.null(temp_dots[["xlim"]])) temp_dots$xlim <-  attr(plot_data[[i]], "x_range")
    if(is.null(temp_dots[["ylim"]])) temp_dots$ylim <-  attr(plot_data[[i]], "y_range")


    if(plot_type == "ggplot"){
      plots[[i]] <- .ggplot.prior_empty("simple", temp_dots)
    }else{
      .plot.prior_empty("simple", temp_dots)
    }

    for(j in seq_along(plot_data[[i]])){

      temp_args <- list()
      temp_args[["plot_data"]] <- plot_data[[i]][[j]]
      temp_args[["col"]]       <- dots[["col"]][j]
      temp_args[["lwd"]]       <- dots[["lwd"]][j]
      temp_args[["lty"]]       <- dots[["lty"]][j]

      if(plot_type == "ggplot"){
        plots[[i]] <- plots[[i]] + do.call(switch(
          type,
          "density"         = .geom_diagnostics.density,
          "trace"           = .geom_diagnostics.trace,
          "autocorrelation" = .geom_diagnostics.autocorrelation
        ), temp_args)
      }else{
        do.call(switch(
          type,
          "density"         = .lines_diagnostics.density,
          "trace"           = .lines_diagnostics.trace,
          "autocorrelation" = .lines_diagnostics.autocorrelation
        ), temp_args)
      }

    }
  }


  # return the plots
  if(plot_type == "base"){
    return(invisible())
  }else if(plot_type == "ggplot"){
    if(length(plots) == 1){
      plots <- plots[[1]]
    }
    return(plots)
  }


}

#' @rdname JAGS_diagnostics
JAGS_diagnostics_density         <- function(fit, parameter, plot_type = "base",
                                             xlim = NULL, n_points = 1000,
                                             transformations = NULL, transform_factors = FALSE, transform_orthonormal = FALSE,
                                             short_name = FALSE, parameter_names = FALSE, formula_prefix = TRUE, ...){

  JAGS_diagnostics(fit = fit, parameter = parameter, plot_type = plot_type, type = "density",
                   xlim = xlim, n_points = n_points,
                   transformations = transformations, transform_factors = transform_factors, transform_orthonormal = transform_orthonormal,
                   short_name = short_name, parameter_names = parameter_names, formula_prefix = formula_prefix, ...)
}

#' @rdname JAGS_diagnostics
JAGS_diagnostics_trace           <- function(fit, parameter, plot_type = "base",
                                             ylim = NULL,
                                             transformations = NULL, transform_factors = FALSE, transform_orthonormal = FALSE,
                                             short_name = FALSE, parameter_names = FALSE, formula_prefix = TRUE, ...){

  JAGS_diagnostics(fit = fit, parameter = parameter, plot_type = plot_type, type = "trace",
                   ylim = ylim,
                   transformations = transformations, transform_factors = transform_factors, transform_orthonormal = transform_orthonormal,
                   short_name = short_name, parameter_names = parameter_names, formula_prefix = formula_prefix, ...)
}

#' @rdname JAGS_diagnostics
JAGS_diagnostics_autocorrelation <- function(fit, parameter, plot_type = "base",
                                             lags = 30,
                                             transformations = NULL, transform_factors = FALSE, transform_orthonormal = FALSE,
                                             short_name = FALSE, parameter_names = FALSE, formula_prefix = TRUE, ...){

  JAGS_diagnostics(fit = fit, parameter = parameter, plot_type = plot_type, type = "autocorrelation",
                   lags = lags,
                   transformations = transformations, transform_factors = transform_factors, transform_orthonormal = transform_orthonormal,
                   short_name = short_name, parameter_names = parameter_names, formula_prefix = formula_prefix, ...)
}


.diagnostics_plot_data                 <- function(fit, parameter, prior_list, transformations, transform_factors){

  check_list(transformations, "transformations", allow_NULL = TRUE)
  if(!is.null(transformations) && any(!sapply(transformations, function(trans)is.function(trans[["fun"]]))))
    stop("'transformations' must be list of functions in the 'fun' element.")

  model_samples_list <- .extract_posterior_samples(fit, as_list = TRUE)
  samples_chain <- lapply(seq_along(model_samples_list), function(i) {
    return(rep(i, nrow(model_samples_list[[i]])))
  })
  samples_iter  <- lapply(seq_along(model_samples_list), function(i) {
    return(1:nrow(model_samples_list[[i]]))
  })
  model_samples <- do.call(rbind, model_samples_list)


  # extract the relevant parameters
  if(is.prior.spike_and_slab(prior_list[[parameter]])){

    if(sum(model_samples[,colnames(model_samples) == paste0(parameter, "_indicator")] != 0) < 10)
      stop("The parameter with a spike and slab prior did not result in enough samples under the slab for producing a diagnostic figure.")

    # change the samples between conditional
    model_samples[
      model_samples[,colnames(model_samples) == paste0(parameter, "_indicator")] == 0,
      colnames(model_samples) == parameter] <- NA

    # modify the parameter list
    prior_list[[parameter]] <- .get_spike_and_slab_variable(prior_list[[parameter]])
  }

  if(is.prior.factor(prior_list[[parameter]]) || (is.prior.mixture(prior_list[[parameter]]) && any(sapply(prior_list[[parameter]], is.prior.factor)))){

    if(is.prior.mixture(prior_list[[parameter]])){
      prior_list[[parameter]] <- prior_list[[parameter]][[which(sapply(prior_list[[parameter]], is.prior.factor))[1]]]
    }

    if(.get_prior_factor_levels(prior_list[[parameter]]) > 1){
      model_samples <- model_samples[,paste0(parameter, "[", 1:.get_prior_factor_levels(prior_list[[parameter]]), "]"),drop = FALSE]
    }else{
      model_samples <- model_samples[,parameter,drop = FALSE]
    }
  }else if(is.prior.weightfunction(prior_list[[parameter]])){
    omega_cuts <- weightfunctions_mapping(list(prior_list[[parameter]]), cuts_only = TRUE)
    model_samples <- model_samples[,paste0("omega", "[", 2:(length(omega_cuts) - 1), "]"),drop = FALSE]
  }else if(parameter == "omega" && !is.null(prior_list[["bias"]])){
    selection_priors <- .selection_prior_selection_priors(prior_list[["bias"]])
    model_samples <- model_samples[,paste0("omega", "[", 2:(length(weightfunctions_mapping(selection_priors, cuts_only = TRUE, one_sided = TRUE)) - 1), "]"),drop = FALSE]
  }else if(is_prior_phacking(prior_list[[parameter]])){
    model_samples <- model_samples[,intersect(.phacking_report_parameter(prior_list[[parameter]]), colnames(model_samples)),drop = FALSE]
  }else if(is_prior_bias(prior_list[[parameter]])){
    bias_parameters <- .selection_bias_parameter_names(prior_list[[parameter]], include_kind = FALSE)
    bias_parameters <- bias_parameters[bias_parameters != "omega"]
    if("omega" %in% .selection_bias_parameter_names(prior_list[[parameter]], include_kind = FALSE)){
      omega_cuts <- weightfunctions_mapping(.selection_prior_selection_priors(prior_list[[parameter]]), cuts_only = TRUE, one_sided = TRUE)
      bias_parameters <- c(paste0("omega", "[", 2:(length(omega_cuts) - 1), "]"), bias_parameters)
    }
    model_samples <- model_samples[,intersect(bias_parameters, colnames(model_samples)),drop = FALSE]
  }else if(parameter %in% c("alpha", "pi_null") && !is.null(prior_list[["bias"]])){
    model_samples <- model_samples[,parameter,drop = FALSE]
  }else if(is.prior.vector(prior_list[[parameter]])){
    if(prior_list[[parameter]]$parameters[["K"]] > 1){
      model_samples <- model_samples[,paste0(parameter, "[", 1:prior_list[[parameter]]$parameters[["K"]], "]"),drop = FALSE]
    }else{
      model_samples <- model_samples[,parameter,drop = FALSE]
    }
  }else{
    if(is.prior.PET(prior_list[[parameter]])){
      parameter <- "PET"
    }else if(is.prior.PEESE(prior_list[[parameter]])){
      parameter <- "PEESE"
    }
    model_samples <- model_samples[,parameter,drop = FALSE]
  }

  # deal with bias mixture dispatching
  if(parameter %in% c("PET", "PEESE", "omega", "alpha", "pi_null") && !is.null(prior_list[["bias"]])){
    prior_list      <- prior_list[["bias"]]
    parameter_names <- parameter
  }else{
    prior_list      <- prior_list[parameter]
    parameter_names <- parameter
  }


  # mostly adapted from runjags_estimates_table
  # apply transformations
  model_samples <- .apply_parameter_transformations(model_samples, transformations, prior_list, transform_factors)

  # transform meandif and orthonormal factors to differences
  model_samples <- .transform_factor_contrasts(model_samples, prior_list, transform_factors, transformations)
  
  # rename factor levels (treatment, independent)
  model_samples <- .rename_factor_levels(model_samples, prior_list)
  
  # extract parameter names from column names after transformations and renaming
  parameter_names <- colnames(model_samples)

  # rename weightfunctions factor levels (special case that overrides)
  if(any(sapply(prior_list, is.prior.weightfunction)) && !is.prior.mixture(prior_list)){
    for(par in names(prior_list)[sapply(prior_list, is.prior.weightfunction)]){
      omega_cuts      <- weightfunctions_mapping(prior_list[par], cuts_only = TRUE)
      parameter_names <- sapply(1:(length(omega_cuts)-1), function(i)paste0("omega[",omega_cuts[i],",",omega_cuts[i+1],"]"))
      parameter_names <- parameter_names[-1]
    }
  }

  if(is.prior.mixture(prior_list) && parameter == "omega"){
    omega_cuts      <- weightfunctions_mapping(.selection_prior_selection_priors(prior_list), cuts_only = TRUE, one_sided = TRUE)
    parameter_names <- sapply(2:(length(omega_cuts)-1), function(i)paste0("omega[",omega_cuts[i],",",omega_cuts[i+1],"]"))
  }

  if(any(sapply(prior_list, is_prior_bias))){
    bias_prior <- prior_list[[which(sapply(prior_list, is_prior_bias))[1]]]
    if(.selection_prior_has_selection(bias_prior)){
      omega_cuts      <- weightfunctions_mapping(.selection_prior_selection_priors(bias_prior), cuts_only = TRUE, one_sided = TRUE)
      omega_names_old <- paste0("omega[", 1:(length(omega_cuts)-1), "]")
      omega_names     <- sapply(1:(length(omega_cuts)-1), function(i)paste0("omega[",omega_cuts[i],",",omega_cuts[i+1],"]"))
      parameter_names[parameter_names %in% omega_names_old] <- omega_names[match(parameter_names[parameter_names %in% omega_names_old], omega_names_old)]
    }
  }

  # attach the relevant attributes
  colnames(model_samples)          <- parameter_names
  attr(model_samples, "chain")     <- do.call(c, samples_chain)
  attr(model_samples, "iter")      <- do.call(c, samples_iter)
  attr(model_samples, "parameter") <- parameter
  attr(model_samples, "prior")     <- if(is.prior.mixture(prior_list)) prior_list else prior_list[[parameter]]

  return(model_samples)
}
.diagnostics_plot_data_density         <- function(plot_data, n_points, xlim){

  chain <- attr(plot_data, "chain")
  prior <- attr(plot_data, "prior")
  bounds <- .diagnostics_prior_bounds(prior, attr(plot_data, "parameter"))

  prior_lower <- bounds[["lower"]]
  prior_upper <- bounds[["upper"]]

  out   <- list()

  for(i in 1:ncol(plot_data)){

    if(is.null(xlim)){
      x_range <- range(plot_data[,i], na.rm = TRUE)
    }else{
      x_range <- xlim
    }

    for(j in seq_along(unique(chain))){

      temp_args    <- list(x = plot_data[chain == j,i], n = n_points, from = x_range[1], to = x_range[2], na.rm = TRUE)
      temp_density <- do.call(stats::density, temp_args)

      x_den    <- temp_density$x
      y_den    <- temp_density$y

      # check for truncation
      if(isTRUE(all.equal(prior_lower, x_den[1])) | prior_lower >= x_den[1]){
        y_den <- c(0, y_den)
        x_den <- c(x_den[1], x_den)
      }
      if(isTRUE(all.equal(prior_upper, x_den[length(x_den)])) | prior_upper <= x_den[length(x_den)]){
        y_den <- c(y_den, 0)
        x_den <- c(x_den, x_den[length(x_den)])
      }

      temp_density    <- list(
        call    = call("density"),
        bw      = NULL,
        n       = n_points,
        x       = x_den,
        y       = y_den
      )

      class(temp_density) <- c("density")
      attr(temp_density, "x_range")        <- range(x_den)
      attr(temp_density, "y_range")        <- c(0, max(y_den))
      attr(temp_density, "chain")          <- j
      attr(temp_density, "parameter")      <- attr(plot_data, "parameter")
      attr(temp_density, "parameter_name") <- colnames(plot_data)[i]

      out[[colnames(plot_data)[[i]]]][[j]] <- temp_density
    }

    attr(out[[colnames(plot_data)[[i]]]], "x_range")        <- x_range
    attr(out[[colnames(plot_data)[[i]]]], "y_range")        <- c(0, max(sapply(out[[i]], function(x) attr(x, "y_range"))))
    attr(out[[colnames(plot_data)[[i]]]], "chains")         <- length(unique(chain))
    attr(out[[colnames(plot_data)[[i]]]], "parameter")      <- attr(plot_data, "parameter")
    attr(out[[colnames(plot_data)[[i]]]], "parameter_name") <- colnames(plot_data)[i]
  }

  attr(out, "chains")         <- length(unique(chain))
  attr(out, "parameter")      <- attr(plot_data, "parameter")
  attr(out, "parameter_name") <- colnames(plot_data)

  return(out)
}

.diagnostics_prior_bounds <- function(prior, parameter){

  is_selection_prior <- .diagnostics_is_selection_prior(prior)

  if(is.prior.weightfunction(prior)){
    return(.weightfunction_weights_truncation(prior$weights))
  }

  if(is_selection_prior && parameter == "omega"){
    selection_priors <- .selection_prior_selection_priors(prior)
    if(length(selection_priors) == 0L){
      return(list(lower = 1, upper = 1))
    }
    upper <- max(vapply(selection_priors, function(x){
      .weightfunction_weights_truncation(x$weights)$upper
    }, numeric(1)))
    return(list(lower = 0, upper = upper))
  }

  if(is_selection_prior && parameter == "alpha"){
    return(list(lower = 0, upper = 1))
  }

  if(is_prior_phacking(prior) && !parameter %in% c("alpha", "pi_null")){
    parameter <- .phacking_report_parameter(prior)
  }

  if(is_selection_prior && parameter == "pi_null"){
    phacking_priors <- .selection_prior_phacking_priors(prior)
    if(length(phacking_priors) > 0L){
      upper <- max(vapply(phacking_priors, function(x){
        phack_backend_constants(x$form, x$source, x$destination, target = x$target)$pi_null_per_alpha
      }, numeric(1)))
      return(list(lower = 0, upper = upper))
    }
    return(list(lower = 0, upper = Inf))
  }

  if(is_prior_phacking(prior) || is_prior_bias(prior)){
    return(list(lower = 0, upper = 1))
  }

  if(is.prior.mixture(prior)){
    simple_priors <- prior[vapply(prior, function(x) is.prior.simple(x) || is.prior.point(x), logical(1))]
    if(length(simple_priors) == 0L){
      return(list(lower = -Inf, upper = Inf))
    }
    prior_lower <- min(vapply(simple_priors, function(x){
      if(is.prior.point(x)) x$parameters[["location"]] else x$truncation[["lower"]]
    }, numeric(1)))
    prior_upper <- max(vapply(simple_priors, function(x){
      if(is.prior.point(x)) x$parameters[["location"]] else x$truncation[["upper"]]
    }, numeric(1)))
    return(list(lower = prior_lower, upper = prior_upper))
  }

  if(is.prior.point(prior)){
    return(list(lower = prior$parameters[["location"]], upper = prior$parameters[["location"]]))
  }

  return(list(lower = prior$truncation[["lower"]], upper = prior$truncation[["upper"]]))
}

.diagnostics_is_selection_prior <- function(prior){

  if(is.prior.weightfunction(prior) || is_prior_phacking(prior) || is_prior_bias(prior)){
    return(TRUE)
  }

  if(is.prior.mixture(prior)){
    return(any(vapply(prior, function(x){
      is.prior.none(x) || is.prior.PET(x) || is.prior.PEESE(x) ||
        is.prior.weightfunction(x) || is_prior_phacking(x) || is_prior_bias(x)
    }, logical(1))))
  }

  FALSE
}
.diagnostics_plot_data_trace           <- function(plot_data, n_points, ylim){

  chain <- attr(plot_data, "chain")
  iter  <- attr(plot_data, "iter")
  prior <- attr(plot_data, "prior")

  out   <- list()

  for(i in 1:ncol(plot_data)){

    if(is.null(ylim)){
      y_range <- range(plot_data[,i], na.rm = TRUE)
    }else{
      y_range <- ylim
    }
    x_range <- range(iter)

    for(j in seq_along(unique(chain))){

      temp_x       <- iter[chain == j]
      temp_y       <- plot_data[chain == j,i]

      temp_chain <- list(
        x       = temp_x,
        y       = temp_y
      )

      class(temp_chain) <- c("BayesTools_chain")
      attr(temp_chain, "x_range")        <- x_range
      attr(temp_chain, "y_range")        <- y_range
      attr(temp_chain, "chain")          <- j
      attr(temp_chain, "parameter")      <- attr(plot_data, "parameter")
      attr(temp_chain, "parameter_name") <- colnames(plot_data)[i]

      out[[colnames(plot_data)[[i]]]][[j]] <- temp_chain
    }

    attr(out[[colnames(plot_data)[[i]]]], "x_range")        <- x_range
    attr(out[[colnames(plot_data)[[i]]]], "y_range")        <- y_range
    attr(out[[colnames(plot_data)[[i]]]], "chains")         <- length(unique(chain))
    attr(out[[colnames(plot_data)[[i]]]], "parameter")      <- attr(plot_data, "parameter")
    attr(out[[colnames(plot_data)[[i]]]], "parameter_name") <- colnames(plot_data)[i]
  }

  attr(out, "chains")         <- length(unique(chain))
  attr(out, "parameter")      <- attr(plot_data, "parameter")
  attr(out, "parameter_name") <- colnames(plot_data)

  return(out)
}
.diagnostics_plot_data_autocorrelation <- function(plot_data, n_points, lags){

  chain <- attr(plot_data, "chain")
  iter  <- attr(plot_data, "iter")
  prior <- attr(plot_data, "prior")

  out   <- list()

  for(i in 1:ncol(plot_data)){

    x_range <-

    for(j in seq_along(unique(chain))){

      temp_x  <- 0:lags
      temp_y  <- stats::acf(plot_data[chain == j,i], lag.max = lags, plot = FALSE, na.action = stats::na.pass)$acf[, , 1L]


      temp_autocor <- list(
        x       = temp_x,
        y       = temp_y
      )

      class(temp_autocor) <- c("BayesTools_autocorrelation")
      attr(temp_autocor, "x_range")        <- c(0, lags)
      attr(temp_autocor, "y_range")        <- range(c(0, temp_y))
      attr(temp_autocor, "chain")          <- j
      attr(temp_autocor, "parameter")      <- attr(plot_data, "parameter")
      attr(temp_autocor, "parameter_name") <- colnames(plot_data)[i]

      out[[colnames(plot_data)[[i]]]][[j]] <- temp_autocor
    }

    attr(out[[colnames(plot_data)[[i]]]], "x_range")        <- c(0, lags)
    attr(out[[colnames(plot_data)[[i]]]], "y_range")        <- c(0, max(sapply(out[[i]], function(x) attr(x, "y_range"))))
    attr(out[[colnames(plot_data)[[i]]]], "chains")         <- length(unique(chain))
    attr(out[[colnames(plot_data)[[i]]]], "parameter")      <- attr(plot_data, "parameter")
    attr(out[[colnames(plot_data)[[i]]]], "parameter_name") <- colnames(plot_data)[i]
  }

  attr(out, "chains")         <- length(unique(chain))
  attr(out, "parameter")      <- attr(plot_data, "parameter")
  attr(out, "parameter_name") <- colnames(plot_data)

  return(out)
}


.lines_diagnostics.density         <- function(plot_data, ...){

  dots      <- list(...)
  col       <- if(!is.null(dots[["col"]]))      dots[["col"]]      else .plot.prior_settings()[["col"]]
  lwd       <- if(!is.null(dots[["lwd"]]))      dots[["lwd"]]      else .plot.prior_settings()[["lwd"]]
  lty       <- if(!is.null(dots[["lty"]]))      dots[["lty"]]      else .plot.prior_settings()[["lty"]]


  graphics::lines(x = plot_data$x, y = plot_data$y, type = "l", lwd = lwd, lty = lty, col = col)

  return(invisible())
}
.lines_diagnostics.trace           <- function(plot_data, ...){

  dots      <- list(...)
  col       <- if(!is.null(dots[["col"]]))      dots[["col"]]      else .plot.prior_settings()[["col"]]
  lwd       <- if(!is.null(dots[["lwd"]]))      dots[["lwd"]]      else .plot.prior_settings()[["lwd"]]
  lty       <- if(!is.null(dots[["lty"]]))      dots[["lty"]]      else .plot.prior_settings()[["lty"]]


  graphics::lines(x = plot_data$x, y = plot_data$y, type = "l", lwd = lwd, lty = lty, col = col)

  return(invisible())
}
.lines_diagnostics.autocorrelation <- function(plot_data, ...){

  dots      <- list(...)
  col       <- if(!is.null(dots[["col"]])) dots[["col"]] else .plot.prior_settings()[["col"]]

  graphics::rect(
    xleft   = plot_data$x + 0.075 - 0.5,
    ybottom = 0,
    xright  = plot_data$x + 0.925 - 0.5,
    ytop    = plot_data$y,
    col     = col)

  return(invisible())
}

.geom_diagnostics.density         <- function(plot_data, ...){

  dots      <- list(...)
  col       <- if(!is.null(dots[["col"]]))      dots[["col"]]      else .plot.prior_settings()[["col"]]
  lwd       <- if(!is.null(dots[["size"]]))     dots[["size"]]     else  if(!is.null(dots[["lwd"]])) dots[["lwd"]] else .plot.prior_settings()[["lwd"]]
  lty       <- if(!is.null(dots[["linetype"]])) dots[["linetype"]] else  if(!is.null(dots[["lty"]])) dots[["lty"]] else .plot.prior_settings()[["lty"]]

  geom <- ggplot2::geom_line(
    data    = data.frame(
      x = plot_data$x,
      y = plot_data$y),
    mapping = ggplot2::aes(
      x = .data[["x"]],
      y = .data[["y"]]),
    linewidth = lwd, linetype = lty, color = col)

  return(geom)
}
.geom_diagnostics.trace           <- function(plot_data, ...){

  dots      <- list(...)
  col       <- if(!is.null(dots[["col"]]))      dots[["col"]]      else .plot.prior_settings()[["col"]]
  lwd       <- if(!is.null(dots[["size"]]))     dots[["size"]]     else  if(!is.null(dots[["lwd"]])) dots[["lwd"]] else .plot.prior_settings()[["lwd"]]
  lty       <- if(!is.null(dots[["linetype"]])) dots[["linetype"]] else  if(!is.null(dots[["lty"]])) dots[["lty"]] else .plot.prior_settings()[["lty"]]

  geom <- ggplot2::geom_line(
    data    = data.frame(
      x = plot_data$x,
      y = plot_data$y),
    mapping = ggplot2::aes(
      x = .data[["x"]],
      y = .data[["y"]]),
    linewidth = lwd, linetype = lty, color = col)

  return(geom)
}
.geom_diagnostics.autocorrelation <- function(plot_data, ...){

  dots      <- list(...)
  col       <- if(!is.null(dots[["col"]])) dots[["col"]] else .plot.prior_settings()[["col"]]

  geom <- ggplot2::geom_bar(
    data    = data.frame(
      x = plot_data$x,
      y = plot_data$y),
    mapping = ggplot2::aes(
      x      = .data[["x"]],
      weight = .data[["y"]]),
    color = col, fill = col)

  return(geom)
}

#' @title Plot a list of prior distributions
#'
#' @param prior_list list of prior distributions
#' @param prior_list_mu list of priors for the mu parameter
#' required when plotting PET-PEESE
#' @param effect_direction direction of the effect for PET-PEESE
#' regression. Use \code{"positive"} (default) for
#' \code{mu + PET*se + PEESE*se^2} or \code{"negative"} for
#' \code{mu - PET*se - PEESE*se^2}.
#' @param legend whether factor legends should be drawn.
#' @param legend_title optional title for factor legends.
#' @param legend_labels optional labels for factor legend levels.
#' @param legend_position optional legend position for factor legends.
#' @param ... additional arguments
#' @inheritParams density.prior
#' @inheritParams plot.prior
#'
#' @return \code{plot_prior_list} returns either \code{NULL} or
#' an object of class 'ggplot' if plot_type is \code{plot_type = "ggplot"}.
#'
#' @seealso [prior()] [lines_prior_list()]  [geom_prior_list()]
#' @export
plot_prior_list <- function(prior_list, plot_type = "base",
                            x_seq = NULL, xlim = NULL, x_range_quant = NULL, n_points = 500,
                            n_samples = 10000, force_samples = FALSE,
                            individual = FALSE, show_figures = if(individual) 1 else NULL,
                            transformation = NULL, transformation_arguments = NULL, transformation_settings = FALSE,
                            rescale_x = FALSE, par_name = NULL, prior_list_mu = NULL, effect_direction = "positive",
                            legend = TRUE, legend_title = NULL, legend_labels = NULL, legend_position = NULL, ...){

  # check input (most arguments are checked within density)
  check_list(prior_list, "prior_list")
  if(is.prior(prior_list) | !all(sapply(prior_list, is.prior)))
    stop("'prior_list' must be a list of priors.")
  check_char(plot_type, "plot_type", allow_values = c("base", "ggplot"))
  check_bool(individual, "individual")
  check_bool(rescale_x, "rescale_x")
  check_bool(legend, "legend", allow_NA = FALSE)
  check_int(show_figures, "show_figures", allow_NULL = TRUE)
  check_char(effect_direction, "effect_direction", allow_values = c("positive", "negative"))
  # check that there is no mixing of PET-PEESE and weightfunctions
  if(any(sapply(prior_list, is.prior.weightfunction)) & (any(sapply(prior_list, is.prior.PET)) | any(sapply(prior_list, is.prior.PEESE))))
    stop("weightfunction and PET-PEESE priors cannot be mixed within a 'prior_list'.")


  # get the plotting type
  if(any(sapply(prior_list, is.prior.weightfunction))){
    prior_type <- "weightfunction"
  }else if(any(sapply(prior_list, is.prior.PET)) | any(sapply(prior_list, is.prior.PEESE))){
    prior_type <- "PETPEESE"
  }else if(any(sapply(prior_list, is.prior.orthonormal))){
    prior_type <- "orthonormal"
  }else if(any(sapply(prior_list, is.prior.meandif))){
    prior_type <- "meandif"
  }else{
    prior_type <- "simple"

  }

  if(prior_type == "PETPEESE" && !individual){
    check_list(prior_list_mu, "prior_list_mu", check_length = length(prior_list))
    if(is.prior(prior_list_mu) | !all(sapply(prior_list_mu, is.prior)))
      stop("'prior_list_mu' must be a list of priors (priors for the mu parameter are required for plotting PET-PEESE).")
  }else{
    if(!is.null(prior_list_mu))
      stop("'prior_list_mu' is required only for PET-PEESE plots.")
  }


  # get the plotting range
  if(is.null(xlim) & is.null(x_seq)){
    if(prior_type %in% c("weightfunction", "PETPEESE") & !individual){
      xlim      <- c(0, 1)
    }else if(prior_type %in% c("simple", "orthonormal", "meandif")){
      xlim   <- do.call(rbind, lapply(prior_list, range, quantiles = x_range_quant))
      xlim   <- range(pretty(range(as.vector(xlim))))
    }
  }


  # get the plotting data
  if(prior_type == "weightfunction" && !individual){
    # special dispatching for visualizing the whole weightfunction

    # use analytical marginal summaries of the mapped cumulative Dirichlet weights
    plot_data <- .plot_data_prior_list.weightfunction(prior_list, x_seq = x_seq, x_range = xlim, x_range_quant = x_range_quant,
                                                      n_points = n_points, n_samples = n_samples)
    plot <- .plot.prior.weightfunction(prior_list, plot_type = plot_type, plot_data = plot_data, rescale_x = rescale_x, par_name = par_name, ...)

  }else if(prior_type == "PETPEESE" && !individual){
    # special dispatching for visualizing the PET-PEESE regression

    # use deterministic linear-combination summaries when supported, with a sampling fallback
    plot_data <- .plot_data_prior_list.PETPEESE(prior_list, x_seq = x_seq, x_range = xlim, x_range_quant = x_range_quant,
                                                n_points = n_points, n_samples = n_samples,
                                                transformation = transformation, transformation_arguments = transformation_arguments,
                                                transformation_settings = transformation_settings, prior_list_mu = prior_list_mu,
                                                effect_direction = effect_direction)
    plot <- .plot.prior.PETPEESE(prior_list, plot_type = plot_type, plot_data = plot_data, par_name = par_name, ...)

  }else if(prior_type %in% c("simple", "orthonormal", "meandif")){
    # regular prior distributions (or individual plots for parameters from weightfunctions/PET-PEESE)

    # solve analytically
    plot_data <- .plot_data_prior_list.simple(prior_list, x_seq = x_seq, x_range = xlim, x_range_quant = x_range_quant,
                                              n_points = n_points, n_samples = n_samples, force_samples = force_samples, individual = individual,
                                              transformation = transformation, transformation_arguments = transformation_arguments,
                                              transformation_settings = transformation_settings)
    if(any(sapply(plot_data, inherits, what = "density.prior.factor"))){
      plot <- .plot_prior_list.factor(
        plot_data = plot_data, plot_type = plot_type, par_name = par_name,
        legend = legend, legend_title = legend_title, legend_labels = legend_labels,
        legend_position = legend_position, ...
      )
    }else{
      plot <- .plot_prior_list.both(plot_data = plot_data, plot_type = plot_type, par_name = par_name, ...)
    }

  }


  if(plot_type == "ggplot"){
    return(plot)
  }else{
    return(invisible())
  }
}

.plot_prior_list.both             <- function(plot_data, plot_type, par_name = NULL, scale_y2 = NULL, add = FALSE, ...){

  # get default plot settings
  dots      <- list(...)

  xlim      <- range(as.vector(sapply(plot_data, attr, which = "x_range")))

  main      <- ""
  xlab      <- if(!is.null(par_name)) par_name else ""

  if(is.null(scale_y2)) scale_y2  <- .get_scale_y2(plot_data, dots)

  if(any(sapply(plot_data, inherits, what = "density.prior.simple")) & any(sapply(plot_data, inherits, what = "density.prior.point"))){
    type  <- "both"
    ylab  <- "Density"
    ylab2 <- "Probability"
    ylim  <- range(as.vector(sapply(plot_data[sapply(plot_data, inherits, what = "density.prior.simple")], attr, which = "y_range")))
    ylim2 <- range(as.vector(sapply(plot_data[sapply(plot_data, inherits, what = "density.prior.point")],  attr, which = "y_range")))
  }else if(any(sapply(plot_data, inherits, what = "density.prior.simple"))){
    type  <- "simple"
    ylab  <- "Density"
    ylim  <- range(as.vector(sapply(plot_data[sapply(plot_data, inherits, what = "density.prior.simple")], attr, which = "y_range")))
    ylab2 <- NULL
    ylim2 <- NULL
  }else if(any(sapply(plot_data, inherits, what = "density.prior.point"))){
    type  <- "point"
    ylab  <- "Probability"
    ylim  <- range(as.vector(sapply(plot_data[sapply(plot_data, inherits, what = "density.prior.point")],  attr, which = "y_range")))
    ylab2 <- NULL
    ylim2 <- NULL
  }


  # add it to the user input if desired
  if(is.null(dots[["main"]]))  dots$main  <-  main
  if(is.null(dots[["xlab"]]))  dots$xlab  <-  xlab
  if(is.null(dots[["ylab"]]))  dots$ylab  <-  ylab
  if(is.null(dots[["ylab2"]])) dots$ylab2 <-  ylab2
  if(is.null(dots[["xlim"]]))  dots$xlim  <-  xlim
  if(is.null(dots[["ylim"]]))  dots$ylim  <-  ylim
  if(is.null(dots[["ylim2"]])) dots$ylim2 <-  ylim2


  if(plot_type == "base"){

    if(!add){
      .plot.prior_empty(type, dots)
    }

    for(i in seq_along(plot_data)){
      if(inherits(plot_data[[i]], what = "density.prior.simple")){
        args           <- dots
        args$plot_data <- plot_data[[i]]
        do.call(.lines.prior.simple, args)
      }else if(inherits(plot_data[[i]], what = "density.prior.point")){
        args           <- dots
        args$scale_y2  <- scale_y2
        args$plot_data <- plot_data[[i]]
        do.call(.lines.prior.point, args)
      }
    }
    plot <- list(scale_y2 = scale_y2)

  }else if(plot_type == "ggplot"){

    plot <- list()

    for(i in seq_along(plot_data)){
      if(inherits(plot_data[[i]], what = "density.prior.simple")){
        args           <- dots
        args$plot_data <- plot_data[[i]]
        plot           <- c(plot, do.call(.geom_prior.simple, args))
      }else if(inherits(plot_data[[i]], what = "density.prior.point")){
        args           <- dots
        args$scale_y2  <- scale_y2
        args$plot_data <- plot_data[[i]]
        plot           <- c(plot, do.call(.geom_prior.point, args))
      }
    }

    if(!add){
      plot <- .ggplot.prior_empty(type, dots) + plot
    }

  }

  # return the plots
  if(plot_type == "base"){
    return(invisible(plot))
  }else if(plot_type == "ggplot"){
    return(plot)
  }
}
.plot_prior_factor_component_level_name <- function(component){

  level_name <- attr(component, "level_name")
  if(!is.null(level_name) && length(level_name) == 1L && !is.na(level_name)){
    return(as.character(level_name))
  }

  if(inherits(component, "density.prior.factor")){
    level <- attr(component, "level")
    if(!is.null(level) && length(level) == 1L && !is.na(level)){
      return(as.character(level))
    }
  }

  NA_character_
}
.plot_prior_factor_format_level_names <- function(level_names){

  if(length(level_names) == 0L){
    return(character())
  }

  level_labels <- level_names
  dif_matches  <- gregexpr("[dif:", level_labels, fixed = TRUE)
  has_dif       <- vapply(dif_matches, function(x) x[1] != -1L, logical(1))
  multi_dif     <- vapply(dif_matches, function(x) x[1] != -1L && length(x) > 1, logical(1))
  if(any(multi_dif)){
    level_labels[multi_dif] <- gsub("__xXx__", ":", level_labels[multi_dif], fixed = TRUE)
  }
  if(any(has_dif & !multi_dif)){
    single_dif <- has_dif & !multi_dif
    level_labels[single_dif] <- substr(
      level_labels[single_dif],
      regexpr("[dif:", level_labels[single_dif], fixed = TRUE) + 5,
      regexpr("]", level_labels[single_dif], fixed = TRUE) - 1
    )
  }
  no_dif <- !has_dif
  if(any(no_dif & grepl("[", level_labels, fixed = TRUE))){
    bracket_names <- no_dif & grepl("[", level_labels, fixed = TRUE)
    level_labels[bracket_names] <- substr(
      level_labels[bracket_names],
      regexpr("[", level_labels[bracket_names], fixed = TRUE) + 1,
      regexpr("]", level_labels[bracket_names], fixed = TRUE) - 1
    )
  }

  level_labels
}
.plot_prior_factor_normalize_data <- function(plot_data){

  is_point  <- vapply(plot_data, inherits, logical(1), what = "density.prior.point")
  is_factor <- vapply(plot_data, inherits, logical(1), what = "density.prior.factor")

  component_levels <- vapply(plot_data, .plot_prior_factor_component_level_name, character(1))
  level_names_raw  <- unique(component_levels[is_factor & !is.na(component_levels)])
  level_names      <- .plot_prior_factor_format_level_names(level_names_raw)

  plot_data <- lapply(seq_along(plot_data), function(i){
    component <- plot_data[[i]]
    level_id  <- match(component_levels[i], level_names_raw)

    if(is.na(level_id)){
      level_id <- NA_integer_
    }

    attr(component, "plot_component") <- if(is_point[i]) "point" else if(is_factor[i]) "density" else "other"
    attr(component, "component_id")    <- i
    attr(component, "level_id")       <- level_id
    attr(component, "level_label")    <- if(!is.na(level_id)) level_names[level_id] else NA_character_

    component
  })
  names(plot_data) <- names(is_point)

  list(
    plot_data       = plot_data,
    points          = plot_data[is_point],
    densities       = plot_data[is_factor & !is_point],
    level_names_raw = level_names_raw,
    level_names     = level_names
  )
}
.plot_prior_factor_style_value <- function(values, level_id, component_id, n_levels, default = NULL){

  if(is.null(values)){
    return(default)
  }

  if(length(values) == 1L){
    value <- values[[1L]]
  }else if(length(values) == n_levels && length(level_id) == 1L && !is.na(level_id) && is.finite(level_id) && level_id >= 1L && length(values) >= level_id){
    value <- values[[level_id]]
  }else if(length(component_id) == 1L && !is.na(component_id) && is.finite(component_id) && component_id >= 1L && length(values) >= component_id){
    value <- values[[component_id]]
  }else if(length(level_id) == 1L && !is.na(level_id) && is.finite(level_id) && level_id >= 1L && length(values) >= level_id){
    value <- values[[level_id]]
  }else{
    return(default)
  }

  if(is.null(value) || length(value) == 0L || all(is.na(value))){
    return(default)
  }

  value
}
.plot_prior_factor_level_style_values <- function(values, plot_data, n_levels, default = NULL){

  out <- rep(default, n_levels)
  if(n_levels == 0L){
    return(out)
  }

  for(level_id in seq_len(n_levels)){
    component_ind <- which(vapply(plot_data, function(component){
      identical(attr(component, "level_id"), level_id)
    }, logical(1)))[1]

    component_id <- if(!is.na(component_ind)) attr(plot_data[[component_ind]], "component_id") else NA_integer_
    out[level_id] <- .plot_prior_factor_style_value(
      values       = values,
      level_id     = level_id,
      component_id = component_id,
      n_levels     = n_levels,
      default      = default
    )
  }

  out
}
.plot_prior_list.factor           <- function(plot_data, plot_type, par_name = NULL, scale_y2 = NULL, add = FALSE, ...){

  # get default plot settings
  dots      <- list(...)

  xlim      <- range(as.vector(sapply(plot_data, attr, which = "x_range")))

  main      <- ""
  xlab      <- if(!is.null(par_name)) par_name else ""

  if(is.null(scale_y2)) scale_y2  <- .get_scale_y2(plot_data, dots)

  if(any(sapply(plot_data, inherits, what = "density.prior.simple")) & any(sapply(plot_data, inherits, what = "density.prior.point"))){
    type  <- "both"
    ylab  <- "Density"
    ylab2 <- "Probability"
    ylim  <- range(as.vector(sapply(plot_data[sapply(plot_data, inherits, what = "density.prior.simple")], attr, which = "y_range")))
    ylim2 <- range(as.vector(sapply(plot_data[sapply(plot_data, inherits, what = "density.prior.point")],  attr, which = "y_range")))
  }else if(any(sapply(plot_data, inherits, what = "density.prior.simple"))){
    type  <- "simple"
    ylab  <- "Density"
    ylim  <- range(as.vector(sapply(plot_data[sapply(plot_data, inherits, what = "density.prior.simple")], attr, which = "y_range")))
    ylab2 <- NULL
    ylim2 <- NULL
  }else if(any(sapply(plot_data, inherits, what = "density.prior.point"))){
    type  <- "point"
    ylab  <- "Probability"
    ylim  <- range(as.vector(sapply(plot_data[sapply(plot_data, inherits, what = "density.prior.point")],  attr, which = "y_range")))
    ylab2 <- NULL
    ylim2 <- NULL
  }


  # add it to the user input if desired
  if(is.null(dots[["main"]]))  dots$main  <-  main
  if(is.null(dots[["xlab"]]))  dots$xlab  <-  xlab
  if(is.null(dots[["ylab"]]))  dots$ylab  <-  ylab
  if(is.null(dots[["ylab2"]])) dots$ylab2 <-  ylab2
  if(is.null(dots[["xlim"]]))  dots$xlim  <-  xlim
  if(is.null(dots[["ylim"]]))  dots$ylim  <-  ylim
  if(is.null(dots[["ylim2"]])) dots$ylim2 <-  ylim2

  # normalize factor component metadata before rendering
  plot_data_normalized <- .plot_prior_factor_normalize_data(plot_data)
  level_names          <- plot_data_normalized[["level_names"]]
  if(!is.null(dots[["legend_labels"]])){
    if(length(dots[["legend_labels"]]) != length(level_names)){
      stop("'legend_labels' must have the same length as the number of factor levels.", call. = FALSE)
    }
    level_names <- as.character(dots[["legend_labels"]])
    plot_data_normalized[["level_names"]] <- level_names
    plot_data_normalized[["plot_data"]] <- lapply(plot_data_normalized[["plot_data"]], function(component){
      level_id <- attr(component, "level_id")
      if(length(level_id) == 1L && !is.na(level_id)){
        attr(component, "level_label") <- level_names[level_id]
      }
      component
    })
    plot_data_normalized[["points"]] <- plot_data_normalized[["plot_data"]][
      vapply(plot_data_normalized[["plot_data"]], inherits, logical(1), what = "density.prior.point")
    ]
    plot_data_normalized[["densities"]] <- plot_data_normalized[["plot_data"]][
      vapply(plot_data_normalized[["plot_data"]], function(component){
        inherits(component, "density.prior.factor") && !inherits(component, "density.prior.point")
      }, logical(1))
    ]
  }
  plot_data_points     <- plot_data_normalized[["points"]]
  plot_data_factors    <- plot_data_normalized[["densities"]]
  style_components     <- if(length(plot_data_factors) > 0L) plot_data_factors else plot_data_points


  # prepare legend information
  if(is.null(dots[["legend"]])){
    dots[["legend"]] <- TRUE
  }else{
    check_bool(dots[["legend"]], "legend", allow_NA = FALSE)
  }
  draw_legend <- isTRUE(dots[["legend"]])
  if(identical(dots[["legend_position"]], "none")){
    draw_legend <- FALSE
  }
  dots[["legend"]] <- draw_legend

  if(is.null(dots[["col"]]) & (is.null(dots[["lty"]]) | is.null(dots[["linetype"]]))){
    dots$col <- grDevices::palette.colors(n = length(level_names) + 1)[-1]
  }
  if(length(dots[["col"]]) == 1)      dots[["col"]]      <- rep(dots[["col"]],      length(level_names))
  if(length(dots[["lty"]]) == 1)      dots[["lty"]]      <- rep(dots[["lty"]],      length(level_names))
  if(length(dots[["linetype"]]) == 1) dots[["linetype"]] <- rep(dots[["linetype"]], length(level_names))

  level_col <- .plot_prior_factor_level_style_values(
    dots[["col"]],
    style_components,
    length(level_names),
    .plot.prior_settings()[["col"]]
  )
  level_lty <- .plot_prior_factor_level_style_values(
    dots[["lty"]],
    style_components,
    length(level_names),
    .plot.prior_settings()[["lty"]]
  )
  level_linetype <- .plot_prior_factor_level_style_values(
    dots[["linetype"]],
    style_components,
    length(level_names),
    .plot.prior_settings()[["lty"]]
  )


  if(plot_type == "base"){

    if(!add){
      .plot.prior_empty(type, dots)
    }

    # plot points
    for(i in seq_along(plot_data_points)){
      args           <- dots
      args$scale_y2  <- scale_y2
      args$plot_data <- plot_data_points[[i]]
      point_level    <- attr(plot_data_points[[i]], "level_id")
      point_component <- attr(plot_data_points[[i]], "component_id")
      args$col       <- .plot_prior_factor_style_value(
        values       = dots[["col"]],
        level_id     = point_level,
        component_id = point_component,
        n_levels     = length(level_names),
        default      = if(length(dots[["col"]]) > 1) .plot.prior_settings()[["col"]]
      )
      args$lty <- .plot_prior_factor_style_value(
        values       = dots[["lty"]],
        level_id     = point_level,
        component_id = point_component,
        n_levels     = length(level_names),
        default      = if(length(dots[["lty"]]) > 1) .plot.prior_settings()[["lty"]]
      )
      do.call(.lines.prior.point, args)
    }

    # plot factor levels
    for(i in seq_along(plot_data_factors)){
      args           <- dots
      args$plot_data <- plot_data_factors[[i]]
      args$level     <- attr(plot_data_factors[[i]], "level_id")
      if(is.na(args$level)) args$level <- i
      args$col       <- level_col
      args$lty       <- level_lty
      do.call(.lines.prior.factor, args)
    }

    if(draw_legend && length(level_names) > 0){
      graphics::legend(
        if(is.null(dots[["legend_position"]])) "topright" else dots[["legend_position"]],
        legend = level_names,
        col    = level_col,
        lty    = level_lty,
        lwd    = if(!is.null(dots[["lwd"]])) dots[["lwd"]] else rep(.plot.prior_settings()[["lwd"]], length(level_names)),
        title  = dots[["legend_title"]],
        bty    = "n")
    }

    plot <- list(scale_y2 = scale_y2)

  }else if(plot_type == "ggplot"){

    plot <- list()

    # plot points
    for(i in seq_along(plot_data_points)){
      args           <- dots
      args$scale_y2  <- scale_y2
      args$plot_data <- plot_data_points[[i]]
      point_level    <- attr(plot_data_points[[i]], "level_id")
      point_component <- attr(plot_data_points[[i]], "component_id")
      args$col       <- .plot_prior_factor_style_value(
        values       = dots[["col"]],
        level_id     = point_level,
        component_id = point_component,
        n_levels     = length(level_names),
        default      = if(length(dots[["col"]]) > 1) .plot.prior_settings()[["col"]]
      )
      args$lty <- .plot_prior_factor_style_value(
        values       = if(!is.null(dots[["linetype"]])) dots[["linetype"]] else dots[["lty"]],
        level_id     = point_level,
        component_id = point_component,
        n_levels     = length(level_names),
        default      = .plot.prior_settings()[["lty"]]
      )
      plot           <- c(plot, do.call(.geom_prior.point, args))
    }

    # plot factor levels
    if(length(plot_data_factors) > 0){
      plot_data_factors <- data.frame(
        x     = do.call(c, lapply(plot_data_factors, function(x) x$x)),
        y     = do.call(c, lapply(plot_data_factors, function(x) x$y)),
        level = do.call(c, lapply(seq_along(plot_data_factors), function(i) {
          level <- attr(plot_data_factors[[i]], "level_label")
          if(length(level) != 1L || is.na(level)){
            level <- as.character(i)
          }
          rep(level, length(plot_data_factors[[i]]$x))
        }))
      )

      args             <- dots
      args$level_names <- level_names
      args$plot_data   <- plot_data_factors
      args$col         <- level_col
      args$lty         <- if(!is.null(dots[["linetype"]])) level_linetype else level_lty
      args$linetype    <- if(!is.null(dots[["linetype"]])) level_linetype else NULL

      plot <- c(plot, do.call(.geom_prior.factors, args))
    }


    if(draw_legend && length(level_names) > 0){
      plot <- c(plot, list(ggplot2::theme(
        legend.position = if(is.null(dots[["legend_position"]])) "right" else dots[["legend_position"]])))
    }

    if(!add){
      plot <- .ggplot.prior_empty(type, dots) + plot
    }

  }

  # return the plots
  if(plot_type == "base"){
    return(invisible(plot))
  }else if(plot_type == "ggplot"){
    return(plot)
  }
}

.plot_data_prior_list.weightfunction <- function(prior_list, x_seq, x_range, x_range_quant, n_points, n_samples){

  context <- .weightfunction_prior_list_context(prior_list)

  omega_cuts <- context$omega_cuts
  x_mean     <- numeric(length(omega_cuts) - 1)
  x_lCI      <- numeric(length(omega_cuts) - 1)
  x_uCI      <- numeric(length(omega_cuts) - 1)

  for(i in seq_len(length(omega_cuts) - 1)){
    components <- .weightfunction_prior_marginal_components(context, i)

    x_mean[i] <- .weightfunction_mixture_mean(components)
    x_lCI[i]  <- .weightfunction_mixture_quantile(components, .025)
    x_uCI[i]  <- .weightfunction_mixture_quantile(components, .975)
  }

  x_seq     <- omega_cuts
  if(length(x_seq) > 2){
    x_seq_rep <- c(1, sort(rep(2:(length(x_seq)-1), 2)) ,length(x_seq))
    x_val_rep <- sort(rep(1:(length(x_seq)-1), 2))
  }else{
    x_seq_rep <- c(1, 2)
    x_val_rep <- c(1, 1)
  }


  out <- list(
    call    = call("density", "weightfunction list"),
    bw      = NULL,
    n       = n_points,
    x       = x_seq[x_seq_rep],
    y       = x_mean[x_val_rep],
    y_lCI   = x_lCI[x_val_rep],
    y_uCI   = x_uCI[x_val_rep],
    samples = NULL
  )


  class(out) <- c("density", "density.prior", "density.prior.weightfunction")
  attr(out, "x_range") <- c(0, 1)
  attr(out, "y_range") <- c(0, max(1, x_mean, x_lCI, x_uCI, na.rm = TRUE))

  return(out)
}
.plot_data_prior_list.weightparameter<- function(prior_list, parameter, n_points, n_samples){

  context       <- .weightfunction_prior_list_context(prior_list)
  parameter_ind <- match(parameter, context$omega_names)

  if(is.na(parameter_ind)){
    stop(paste0("Parameter '", parameter, "' not found in the weightfunction prior."), call. = FALSE)
  }

  components <- .weightfunction_prior_marginal_components(context, parameter_ind)
  out        <- .plot_data_prior_weightparameter_components(components, parameter, n_points)

  return(out)
}
.weightfunction_prior_list_context <- function(prior_list, one_sided = NULL){

  omega_context <- attr(prior_list, "omega_context")
  if(is.null(one_sided)){
    one_sided <- if(!is.null(omega_context) && !is.null(omega_context$one_sided)){
      isTRUE(omega_context$one_sided)
    }else{
      .weightfunction_prior_context_uses_selection_mapping(prior_list)
    }
  }else{
    check_bool(one_sided, "one_sided")
  }

  prior_list <- .weightfunction_expand_bias_mixture_priors(prior_list)
  prior_list <- .simplify_prior_list(prior_list)
  prior_list <- .weightfunction_expand_bias_mixture_priors(prior_list)

  prior_weights <- sapply(prior_list, .prior_model_weight)
  keep          <- is.finite(prior_weights) & prior_weights > 0
  prior_list    <- prior_list[keep]
  prior_weights <- prior_weights[keep]

  if(length(prior_list) == 0){
    stop("At least one weightfunction prior must have positive prior weight.", call. = FALSE)
  }

  # Non-weightfunction bias components imply no selection adjustment and
  # therefore correspond to publication weights fixed at one.
  for(i in seq_along(prior_list)){
    if(.weightfunction_prior_has_selection(prior_list[[i]])){
      selection_priors <- .selection_prior_selection_priors(prior_list[[i]])
      selection_prior  <- selection_priors[[1L]]
      selection_prior$prior_weights <- prior_weights[i]
      prior_list[[i]] <- selection_prior
    }else if(!(is.prior.weightfunction(prior_list[[i]]) | is.prior.none(prior_list[[i]]))){
      prior_list[[i]] <- prior_none(prior_weights = prior_weights[i])
    }
  }

  prior_weights <- sapply(prior_list, .prior_model_weight)
  model_weights <- prior_weights / sum(prior_weights)
  omega_info    <- .weightfunction_mapping_info(prior_list, one_sided = one_sided)
  omega_mapping <- omega_info$mapping
  omega_cuts    <- omega_info$cuts
  omega_names   <- omega_info$names

  list(
    prior_list    = prior_list,
    model_weights = model_weights,
    omega_mapping = omega_mapping,
    omega_cuts    = omega_cuts,
    omega_names   = omega_names
  )
}
.weightfunction_prior_context_uses_selection_mapping <- function(prior_list){

  if(inherits(prior_list, "prior.bias_mixture") || is_prior_bias(prior_list)){
    return(TRUE)
  }
  if(is.prior(prior_list)){
    return(FALSE)
  }
  if(is.list(prior_list)){
    return(any(vapply(prior_list, function(prior){
      inherits(prior, "prior.bias_mixture") || is_prior_bias(prior)
    }, logical(1))))
  }

  FALSE
}
.weightfunction_expand_bias_mixture_priors <- function(prior_list){

  if(is.prior.mixture(prior_list)){
    class(prior_list) <- NULL
    return(prior_list)
  }
  if(is.prior(prior_list)){
    return(list(prior_list))
  }

  expanded <- list()
  for(i in seq_along(prior_list)){
    if(inherits(prior_list[[i]], "prior.bias_mixture")){
      prior_mixture_components <- prior_list[[i]]
      class(prior_mixture_components) <- NULL
      expanded <- c(expanded, prior_mixture_components)
    }else{
      expanded[[length(expanded) + 1L]] <- prior_list[[i]]
    }
  }

  expanded
}
.weightfunction_prior_has_selection <- function(prior){

  if(is.prior.weightfunction(prior) || is_prior_bias(prior) || inherits(prior, "prior.bias_mixture")){
    return(.selection_prior_has_selection(prior))
  }

  FALSE
}
.weightfunction_prior_marginal_components <- function(context, parameter_ind){

  components <- list()

  for(i in seq_along(context$prior_list)){
    prior <- context$prior_list[[i]]

    if(is.prior.weightfunction(prior)){
      component <- .weightfunction_prior_component(
        prior  = prior,
        index  = context$omega_mapping[[i]][parameter_ind],
        weight = context$model_weights[i]
      )
    }else{
      component <- list(
        type     = "point",
        weight   = context$model_weights[i],
        location = 1
      )
    }

    components[[length(components) + 1L]] <- component
  }

  components <- components[vapply(components, function(component){
    is.finite(component$weight) && component$weight > 0
  }, logical(1))]
  total_weight <- sum(vapply(components, function(component) component$weight, numeric(1)))

  lapply(components, function(component){
    component$weight <- component$weight / total_weight
    component
  })
}
.weightfunction_prior_component <- function(prior, index, weight){

  if(prior$weights$type == "fixed"){
    return(list(
      type     = "point",
      weight   = weight,
      location = prior$weights[["omega"]][index]
    ))
  }

  if(prior$weights$type == "cumulative"){
    return(.weightfunction_prior_component_cumdirichlet(prior$weights[["alpha"]], index, weight))
  }

  if(prior$weights$type == "independent"){
    if(index == 1L){
      return(list(
        type     = "point",
        weight   = weight,
        location = 1
      ))
    }
    return(list(
      type   = "prior",
      weight = weight,
      prior  = prior$weights$prior,
      scale  = prior$weights$scale
    ))
  }

  stop("Unsupported weightfunction prior specification.", call. = FALSE)
}
.weightfunction_prior_component_cumdirichlet <- function(alpha, index, weight){

  if(index <= 1L){
    return(list(
      type     = "point",
      weight   = weight,
      location = 1
    ))
  }

  list(
    type   = "beta",
    weight = weight,
    alpha  = sum(alpha[index:length(alpha)]),
    beta   = sum(alpha[seq_len(index - 1L)])
  )
}
.weightfunction_component_mean <- function(component){

  switch(
    component$type,
    "point" = component$location,
    "beta"  = component$alpha / (component$alpha + component$beta),
    "prior" = if(component$scale == "omega"){
      mean(component$prior)
    }else{
      stats::integrate(
        f     = function(x, prior) {
          y <- exp(x) * pdf(prior, x)
          y[!is.finite(y)] <- 0
          y
        },
        lower = component$prior$truncation[["lower"]],
        upper = component$prior$truncation[["upper"]],
        prior = component$prior
      )$value
    },
    "one_minus_product_beta" = {
      1 - (component$u_alpha / (component$u_alpha + component$u_beta)) *
        (component$v_alpha / (component$v_alpha + component$v_beta))
    }
  )
}
.weightfunction_component_cdf <- function(component, q){

  switch(
    component$type,
    "point" = as.numeric(q >= component$location),
    "beta"  = stats::pbeta(q, shape1 = component$alpha, shape2 = component$beta),
    "prior" = if(component$scale == "omega"){
      mcdf(component$prior, q)
    }else{
      p <- numeric(length(q))
      p[q <= 0] <- 0
      inside <- q > 0
      p[inside] <- mcdf(component$prior, log(q[inside]))
      p
    },
    "one_minus_product_beta" = .weightfunction_one_minus_product_beta_cdf(
      q       = q,
      u_alpha = component$u_alpha,
      u_beta  = component$u_beta,
      v_alpha = component$v_alpha,
      v_beta  = component$v_beta
    )
  )
}
.weightfunction_component_pdf <- function(component, x){

  switch(
    component$type,
    "point" = rep(0, length(x)),
    "beta"  = stats::dbeta(x, shape1 = component$alpha, shape2 = component$beta),
    "prior" = if(component$scale == "omega"){
      mpdf(component$prior, x)
    }else{
      y <- numeric(length(x))
      inside <- x > 0
      y[inside] <- mpdf(component$prior, log(x[inside])) / x[inside]
      y
    },
    "one_minus_product_beta" = .weightfunction_one_minus_product_beta_pdf(
      x       = x,
      u_alpha = component$u_alpha,
      u_beta  = component$u_beta,
      v_alpha = component$v_alpha,
      v_beta  = component$v_beta
    )
  )
}
.weightfunction_component_range <- function(component, quantiles = .005){

  switch(
    component$type,
    "point" = c(component$location, component$location),
    "beta"  = c(0, 1),
    "prior" = {
      if(component$scale == "omega"){
        lower <- component$prior$truncation[["lower"]]
        upper <- component$prior$truncation[["upper"]]

        lower <- if(is.infinite(lower)) mquant(component$prior, quantiles) else lower
        upper <- if(is.infinite(upper)) mquant(component$prior, 1 - quantiles) else upper
        c(lower, upper)
      }else{
        lower <- component$prior$truncation[["lower"]]
        upper <- component$prior$truncation[["upper"]]

        lower <- if(is.infinite(lower)) 0 else exp(lower)
        upper <- if(is.infinite(upper)) exp(mquant(component$prior, 1 - quantiles)) else exp(upper)
        c(lower, upper)
      }
    },
    "one_minus_product_beta" = c(0, 1)
  )
}
.weightfunction_components_range <- function(components, samples = NULL, quantiles = .005){

  ranges <- do.call(rbind, lapply(components, .weightfunction_component_range, quantiles = quantiles))
  values <- as.vector(ranges)
  if(!is.null(samples)){
    values <- c(values, samples)
  }

  x_range <- range(values, finite = TRUE)
  if(!all(is.finite(x_range))){
    x_range <- c(0, 1)
  }
  if(x_range[1] == x_range[2]){
    x_range <- range(c(0, 1, values), finite = TRUE)
  }
  x_range[1] <- max(0, x_range[1])

  x_range
}
.weightfunction_component_quantile <- function(component, p){

  switch(
    component$type,
    "point" = component$location,
    "beta"  = stats::qbeta(p, shape1 = component$alpha, shape2 = component$beta),
    "prior" = if(component$scale == "omega"){
      mquant(component$prior, p)
    }else{
      exp(mquant(component$prior, p))
    },
    "one_minus_product_beta" = {
      lower <- 0
      upper <- 1
      for(iter in seq_len(100L)){
        mid <- lower / 2 + upper / 2
        if(.weightfunction_component_cdf(component, mid) >= p){
          upper <- mid
        }else{
          lower <- mid
        }
        if(abs(upper - lower) <= 1e-8){
          break
        }
      }
      upper
    }
  )
}
.weightfunction_mixture_mean <- function(components){

  sum(vapply(components, function(component){
    component$weight * .weightfunction_component_mean(component)
  }, numeric(1)))
}
.weightfunction_mixture_cdf <- function(components, q){

  p <- sum(vapply(components, function(component){
    component$weight * .weightfunction_component_cdf(component, q)
  }, numeric(1)))

  pmin(pmax(p, 0), 1)
}
.weightfunction_mixture_quantile <- function(components, p){

  if(p <= 0){
    return(0)
  }
  if(p >= 1){
    return(max(.weightfunction_components_range(components)))
  }

  lower <- 0
  upper <- max(vapply(components, .weightfunction_component_quantile, numeric(1), p = p))
  if(!is.finite(upper) || upper <= lower){
    upper <- max(.weightfunction_components_range(components))
  }

  for(iter in seq_len(100L)){
    mid <- lower / 2 + upper / 2
    if(.weightfunction_mixture_cdf(components, mid) >= p){
      upper <- mid
    }else{
      lower <- mid
    }

    if(abs(upper - lower) <= 1e-8){
      break
    }
  }

  upper
}
.plot_data_prior_weightparameter_components <- function(components, parameter, n_points){

  x_range <- .weightfunction_components_range(components)
  x_den <- seq(x_range[1], x_range[2], length.out = n_points)
  y_den <- rep(0, length(x_den))

  for(component in components){
    if(component$type != "point"){
      y_component <- .weightfunction_component_pdf(component, x_den)
      y_component[!is.finite(y_component)] <- 0
      y_den <- y_den + component$weight * y_component
    }
  }

  point_components <- components[vapply(components, function(component){
    component$type == "point"
  }, logical(1))]

  x_points <- NULL
  y_points <- NULL
  if(length(point_components) > 0){
    point_locations <- vapply(point_components, function(component) component$location, numeric(1))
    point_keys      <- as.character(signif(point_locations, 15))
    point_groups    <- split(seq_along(point_components), point_keys)

    x_points <- unname(vapply(point_groups, function(ind) point_locations[ind[1]], numeric(1)))
    y_points <- unname(vapply(point_groups, function(ind){
      sum(vapply(point_components[ind], function(component) component$weight, numeric(1)))
    }, numeric(1)))
  }

  out <- list()

  if(any(y_den > 0)){
    out_den <- list(
      call    = call("density", "weightfunction prior"),
      bw      = NULL,
      n       = n_points,
      x       = x_den,
      y       = y_den,
      samples = NULL
    )

    class(out_den) <- c("density", "density.prior", "density.prior.simple")
    attr(out_den, "x_range") <- x_range
    attr(out_den, "y_range") <- c(0, max(y_den))
    attr(out_den, "parameter") <- parameter

    out[["density"]] <- out_den
  }

  if(!is.null(y_points)){
    for(i in seq_along(y_points)){
      temp_points <- list(
        call    = call("density", paste0("point", i)),
        bw      = NULL,
        n       = n_points,
        x       = x_points[i],
        y       = y_points[i],
        samples = NULL
      )

      class(temp_points) <- c("density", "density.prior", "density.prior.point")
      attr(temp_points, "x_range") <- x_range
      attr(temp_points, "y_range") <- c(0, max(y_points[i]))
      attr(temp_points, "parameter") <- parameter

      out[[paste0("points",i)]] <- temp_points
    }
  }

  out
}
.weightfunction_one_minus_product_beta_cdf <- function(q, u_alpha, u_beta, v_alpha, v_beta){

  vapply(q, function(q_i){
    if(q_i <= 0){
      return(0)
    }
    if(q_i >= 1){
      return(1)
    }

    product_lower <- 1 - q_i
    integration <- stats::integrate(
      f = function(u){
        stats::dbeta(u, shape1 = u_alpha, shape2 = u_beta) *
          stats::pbeta(product_lower / u, shape1 = v_alpha, shape2 = v_beta, lower.tail = FALSE)
      },
      lower         = product_lower,
      upper         = 1,
      subdivisions  = 200L,
      rel.tol       = 1e-7,
      stop.on.error = FALSE
    )

    if(!is.finite(integration$value)){
      stop("Weightfunction prior CDF integration failed.", call. = FALSE)
    }

    pmin(pmax(integration$value, 0), 1)
  }, numeric(1))
}
.weightfunction_one_minus_product_beta_pdf <- function(x, u_alpha, u_beta, v_alpha, v_beta){

  vapply(x, function(x_i){
    if(x_i <= 0 || x_i >= 1){
      return(0)
    }

    product_value <- 1 - x_i
    integration <- stats::integrate(
      f = function(u){
        stats::dbeta(u, shape1 = u_alpha, shape2 = u_beta) *
          stats::dbeta(product_value / u, shape1 = v_alpha, shape2 = v_beta) / u
      },
      lower         = product_value,
      upper         = 1,
      subdivisions  = 200L,
      rel.tol       = 1e-7,
      stop.on.error = FALSE
    )

    if(!is.finite(integration$value)){
      return(NA_real_)
    }

    pmax(integration$value, 0)
  }, numeric(1))
}
.plot_data_prior_list.PETPEESE       <- function(prior_list, x_seq, x_range, x_range_quant, n_points, n_samples,
                                                 transformation, transformation_arguments, transformation_settings, prior_list_mu,
                                                 effect_direction = "positive"){

  if(is.null(x_seq)){
    x_seq <- seq(x_range[1], x_range[2], length.out = n_points)
  }

  # specify it on the transformed range if requested
  if(transformation_settings & !is.null(transformation)){
    x_seq   <- .density.prior_transformation_x(x_seq,   transformation, transformation_arguments)
    x_range <- .density.prior_transformation_x(x_range, transformation, transformation_arguments)
  }

  deterministic <- tryCatch(
    .plot_data_prior_list.PETPEESE_deterministic(
      prior_list               = prior_list,
      x_seq                    = x_seq,
      n_points                 = n_points,
      transformation           = transformation,
      transformation_arguments = transformation_arguments,
      prior_list_mu            = prior_list_mu,
      effect_direction         = effect_direction
    ),
    error = function(e) NULL
  )

  if(!is.null(deterministic)){
    return(deterministic)
  }

  .plot_data_prior_list.PETPEESE_sampled(
    prior_list               = prior_list,
    x_seq                    = x_seq,
    n_points                 = n_points,
    n_samples                = n_samples,
    transformation           = transformation,
    transformation_arguments = transformation_arguments,
    prior_list_mu            = prior_list_mu,
    effect_direction         = effect_direction
  )
}
.plot_data_prior_list.PETPEESE_deterministic <- function(prior_list, x_seq, n_points, transformation, transformation_arguments,
                                                         prior_list_mu, effect_direction = "positive"){

  if(is.list(transformation)){
    stop("Custom transformations are handled by sampled PET-PEESE prior summaries.", call. = FALSE)
  }

  prior_weights <- sapply(prior_list, .prior_model_weight)
  keep <- is.finite(prior_weights) & prior_weights > 0
  prior_list    <- prior_list[keep]
  prior_list_mu <- prior_list_mu[keep]
  prior_weights <- prior_weights[keep]

  if(length(prior_list) == 0){
    stop("At least one PET-PEESE prior must have positive prior weight.", call. = FALSE)
  }

  model_weights <- prior_weights / sum(prior_weights)
  context <- .petpeese_prior_cdf_context(
    prior_list       = prior_list,
    prior_list_mu    = prior_list_mu,
    model_weights    = model_weights,
    effect_direction = effect_direction
  )

  quantiles <- vapply(x_seq, function(se){
    .petpeese_prior_cdf_quantile(context, se, c(.500, .025, .975))
  }, numeric(3))

  if(!is.null(transformation)){
    quantiles <- .petpeese_transform_quantiles(
      quantiles,
      transformation,
      transformation_arguments
    )
  }

  out <- list(
    call    = call("density", "PET-PEESE list"),
    bw      = NULL,
    n       = n_points,
    x       = x_seq,
    y       = quantiles[1,],
    y_lCI   = quantiles[2,],
    y_uCI   = quantiles[3,],
    samples = NULL
  )

  class(out) <- c("density", "density.prior", "density.prior.PETPEESE")
  attr(out, "x_range") <- range(x_seq)
  attr(out, "y_range") <- range(out$y)

  return(out)
}
.petpeese_prior_cdf_context <- function(prior_list, prior_list_mu, model_weights,
                                        effect_direction = "positive"){

  direction_sign <- if(effect_direction == "negative") -1 else 1

  models <- vector("list", length(prior_list))
  for(i in seq_along(prior_list)){
    bias_prior <- prior_list[[i]]
    if(is.prior.PET(bias_prior)){
      bias_type <- "PET"
      bias_components <- .petpeese_prior_components(bias_prior)
    }else if(is.prior.PEESE(bias_prior)){
      bias_type <- "PEESE"
      bias_components <- .petpeese_prior_components(bias_prior)
    }else{
      bias_type <- "none"
      bias_components <- .petpeese_prior_components(prior("point", list(location = 0)))
    }

    models[[i]] <- list(
      weight = model_weights[i],
      mu     = .petpeese_prior_components(prior_list_mu[[i]]),
      bias   = bias_components,
      type   = bias_type
    )
  }

  list(
    models         = models,
    direction_sign = direction_sign
  )
}
.petpeese_prior_components <- function(prior){

  if(is.null(prior) || is.prior.none(prior)){
    return(.petpeese_prior_components_normalize(list(
      list(weight = 1, type = "atom", x = 0)
    )))
  }

  if(is.prior.spike_and_slab(prior)){
    inclusion <- mean(.get_spike_and_slab_inclusion(prior))
    if(!is.finite(inclusion) || inclusion < 0 || inclusion > 1){
      stop("Spike-and-slab inclusion prior must have a finite mean in [0, 1].", call. = FALSE)
    }
    variable_components <- .petpeese_prior_components(.get_spike_and_slab_variable(prior))
    variable_components <- lapply(variable_components, function(component){
      component$weight <- component$weight * inclusion
      component
    })
    spike_component <- list(list(weight = 1 - inclusion, type = "atom", x = 0))
    return(.petpeese_prior_components_normalize(c(variable_components, spike_component)))
  }

  if(is.prior.mixture(prior)){
    weights <- attr(prior, "prior_weights")
    if(is.null(weights)){
      weights <- sapply(prior, .prior_model_weight)
    }
    weights <- weights / sum(weights)

    components <- list()
    for(i in seq_along(prior)){
      component_components <- .petpeese_prior_components(prior[[i]])
      component_components <- lapply(component_components, function(component){
        component$weight <- component$weight * weights[i]
        component
      })
      components <- c(components, component_components)
    }
    return(.petpeese_prior_components_normalize(components))
  }

  if(is.prior.point(prior)){
    return(.petpeese_prior_components_normalize(list(
      list(weight = 1, type = "atom", x = prior$parameters[["location"]])
    )))
  }

  if(is.prior.discrete(prior)){
    support <- switch(
      prior[["distribution"]],
      "bernoulli" = c(0, 1),
      stop("Unsupported discrete PET-PEESE prior distribution.", call. = FALSE)
    )
    probabilities <- mpdf(prior, support)
    keep <- is.finite(probabilities) & probabilities > 0
    if(!any(keep)){
      stop("Discrete PET-PEESE prior has zero probability mass.", call. = FALSE)
    }
    support <- support[keep]
    probabilities <- probabilities[keep] / sum(probabilities[keep])
    return(.petpeese_prior_components_normalize(lapply(seq_along(support), function(i){
      list(weight = probabilities[i], type = "atom", x = support[i])
    })))
  }

  if(is.prior.simple(prior)){
    prior_functions <- .petpeese_prior_simple_functions(prior)
    return(.petpeese_prior_components_normalize(list(
      list(
        weight = 1,
        type   = "continuous",
        prior  = prior,
        cdf    = prior_functions$cdf,
        ccdf   = prior_functions$ccdf,
        pdf    = prior_functions$pdf,
        quant  = prior_functions$quant
      )
    )))
  }

  stop("Unsupported PET-PEESE prior type for deterministic CDF plotting.", call. = FALSE)
}
.petpeese_prior_simple_functions <- function(prior){

  default_range <- .is_prior_default_range(prior)
  if(default_range){
    return(list(
      cdf = function(q) .prior_simple_base_p(prior, q, lower.tail = TRUE),
      ccdf = function(q) .prior_simple_base_p(prior, q, lower.tail = FALSE),
      pdf = function(x) .prior_simple_base_d(prior, x, log = FALSE),
      quant = function(p) .prior_simple_base_q(prior, p)
    ))
  }

  C1 <- .prior_C1(prior)
  C2 <- .prior_C2(prior)
  C  <- C2 - C1
  lower <- prior$truncation[["lower"]]
  upper <- prior$truncation[["upper"]]

  list(
    cdf = function(q){
      p <- numeric(length(q))
      q_lower  <- q < lower
      q_higher <- q > upper
      q_inside <- !q_lower & !q_higher

      p[q_lower]  <- 0
      p[q_higher] <- 1
      if(any(q_inside)){
        p[q_inside] <- (.prior_simple_base_p(prior, q[q_inside], lower.tail = TRUE) - C1) / C
      }
      p
    },
    ccdf = function(q){
      p <- numeric(length(q))
      q_lower  <- q < lower
      q_higher <- q > upper
      q_inside <- !q_lower & !q_higher

      p[q_lower]  <- 1
      p[q_higher] <- 0
      if(any(q_inside)){
        p[q_inside] <- (.prior_simple_base_p(prior, q[q_inside], lower.tail = FALSE) - (1 - C2)) / C
      }
      p
    },
    pdf = function(x){
      y <- .prior_simple_base_d(prior, x, log = FALSE)
      y[x < lower | x > upper] <- 0
      y / C
    },
    quant = function(p){
      .prior_simple_base_q(prior, C1 + p * C)
    }
  )
}
.petpeese_prior_components_normalize <- function(components){

  components <- components[vapply(components, function(component){
    is.finite(component$weight) && component$weight > 0
  }, logical(1))]

  if(length(components) == 0){
    stop("PET-PEESE prior components have zero total weight.", call. = FALSE)
  }

  total_weight <- sum(vapply(components, function(component) component$weight, numeric(1)))
  lapply(components, function(component){
    component$weight <- component$weight / total_weight
    component
  })
}
.petpeese_prior_cdf_quantile <- function(context, se, probs){

  vapply(probs, function(p){
    .petpeese_prior_cdf_one_quantile(context, se, p)
  }, numeric(1))
}
.petpeese_prior_cdf_one_quantile <- function(context, se, p){

  if(p <= 0){
    return(.petpeese_prior_cdf_range(context, se, tail_prob = .Machine$double.eps)[1])
  }
  if(p >= 1){
    return(.petpeese_prior_cdf_range(context, se, tail_prob = .Machine$double.eps)[2])
  }

  fast_quantile <- .petpeese_prior_fast_quantile(context, se, p)
  if(is.finite(fast_quantile)){
    return(fast_quantile)
  }

  bounds <- .petpeese_prior_cdf_range(context, se)
  lower <- bounds[1]
  upper <- bounds[2]

  if(!is.finite(lower) || !is.finite(upper)){
    lower <- -1
    upper <-  1
  }

  if(isTRUE(all.equal(lower, upper))){
    return(lower)
  }

  cdf_lower <- .petpeese_prior_cdf(context, lower, se)
  cdf_upper <- .petpeese_prior_cdf(context, upper, se)
  width <- max(1, upper - lower, abs(lower), abs(upper))

  iter <- 0L
  while(is.finite(cdf_lower) && cdf_lower >= p && iter < 80L){
    upper <- lower
    lower <- lower - width
    width <- width * 2
    cdf_lower <- .petpeese_prior_cdf(context, lower, se)
    iter <- iter + 1L
  }

  iter <- 0L
  while(is.finite(cdf_upper) && cdf_upper < p && iter < 80L){
    lower <- upper
    upper <- upper + width
    width <- width * 2
    cdf_upper <- .petpeese_prior_cdf(context, upper, se)
    iter <- iter + 1L
  }

  if(!is.finite(cdf_lower) || !is.finite(cdf_upper) || cdf_lower >= p || cdf_upper < p){
    stop("Could not bracket PET-PEESE prior quantile.", call. = FALSE)
  }

  if(!.petpeese_prior_cdf_has_atoms(context, se)){
    root <- tryCatch(
      stats::uniroot(
        f        = function(q) .petpeese_prior_cdf(context, q, se) - p,
        interval = c(lower, upper),
        tol      = 1e-8 * max(1, abs(upper - lower))
      )$root,
      error = function(e) NA_real_
    )
    if(is.finite(root)){
      return(root)
    }
  }

  for(iter in seq_len(100L)){
    mid <- lower / 2 + upper / 2
    cdf_mid <- .petpeese_prior_cdf(context, mid, se)
    if(!is.finite(cdf_mid)){
      stop("PET-PEESE prior CDF returned a non-finite value.", call. = FALSE)
    }

    if(cdf_mid >= p){
      upper <- mid
    }else{
      lower <- mid
    }

    if(abs(upper - lower) <= 1e-8 * max(1, abs(lower), abs(upper))){
      break
    }
  }

  upper
}
.petpeese_prior_cdf_has_atoms <- function(context, se){

  any(vapply(context$models, function(model){
    if(model$weight <= 0){
      return(FALSE)
    }
    scale <- switch(
      model$type,
      "PET"   = context$direction_sign * se,
      "PEESE" = context$direction_sign * se^2,
      "none"  = 0
    )

    any(vapply(model$mu, function(mu_component){
      any(vapply(model$bias, function(bias_component){
        .petpeese_prior_sum_has_atom(mu_component, bias_component, scale)
      }, logical(1)))
    }, logical(1)))
  }, logical(1)))
}
.petpeese_prior_sum_has_atom <- function(mu_component, bias_component, scale){

  if(abs(scale) <= .prior_linear_density_zero_tol()){
    return(mu_component$type == "atom")
  }

  mu_component$type == "atom" && bias_component$type == "atom"
}
.petpeese_prior_fast_quantile <- function(context, se, p){

  if(length(context$models) != 1L){
    return(NA_real_)
  }

  model <- context$models[[1]]
  if(length(model$mu) != 1L || length(model$bias) != 1L){
    return(NA_real_)
  }

  scale <- switch(
    model$type,
    "PET"   = context$direction_sign * se,
    "PEESE" = context$direction_sign * se^2,
    "none"  = 0
  )

  .petpeese_prior_sum_quantile(model$mu[[1]], model$bias[[1]], scale, p)
}
.petpeese_prior_sum_quantile <- function(mu_component, bias_component, scale, p){

  if(abs(scale) <= .prior_linear_density_zero_tol()){
    return(.petpeese_prior_component_quantile(mu_component, p))
  }

  if(mu_component$type == "atom" && bias_component$type == "atom"){
    return(mu_component$x + scale * bias_component$x)
  }

  if(mu_component$type == "atom"){
    bias_p <- if(scale > 0) p else 1 - p
    return(mu_component$x + scale * .petpeese_prior_component_quantile(bias_component, bias_p))
  }

  if(bias_component$type == "atom"){
    return(.petpeese_prior_component_quantile(mu_component, p) + scale * bias_component$x)
  }

  NA_real_
}
.petpeese_prior_cdf <- function(context, q, se){

  cdf <- sum(vapply(context$models, function(model){
    model$weight * .petpeese_prior_model_cdf(model, q, se, context$direction_sign)
  }, numeric(1)))

  pmin(pmax(cdf, 0), 1)
}
.petpeese_prior_model_cdf <- function(model, q, se, direction_sign){

  scale <- switch(
    model$type,
    "PET"   = direction_sign * se,
    "PEESE" = direction_sign * se^2,
    "none"  = 0
  )

  cdf <- 0
  for(mu_component in model$mu){
    for(bias_component in model$bias){
      cdf <- cdf + mu_component$weight * bias_component$weight *
        .petpeese_prior_sum_cdf(mu_component, bias_component, scale, q)
    }
  }

  cdf
}
.petpeese_prior_sum_cdf <- function(mu_component, bias_component, scale, q){

  if(abs(scale) <= .prior_linear_density_zero_tol()){
    return(.petpeese_prior_component_cdf(mu_component, q))
  }

  if(mu_component$type == "atom" && bias_component$type == "atom"){
    return(as.numeric(q >= mu_component$x + scale * bias_component$x))
  }

  if(mu_component$type == "atom"){
    threshold <- (q - mu_component$x) / scale
    if(scale > 0){
      return(.petpeese_prior_component_cdf(bias_component, threshold))
    }else{
      return(.petpeese_prior_component_ccdf(bias_component, threshold))
    }
  }

  if(bias_component$type == "atom"){
    return(.petpeese_prior_component_cdf(mu_component, q - scale * bias_component$x))
  }

  integration <- stats::integrate(
    f = function(b){
      .petpeese_prior_component_cdf(mu_component, q - scale * b) *
        .petpeese_prior_component_pdf(bias_component, b)
    },
    lower        = bias_component$prior$truncation[["lower"]],
    upper        = bias_component$prior$truncation[["upper"]],
    subdivisions = 200L,
    rel.tol      = 1e-7,
    stop.on.error = FALSE
  )

  if(!isTRUE(integration$message == "OK") && !is.finite(integration$value)){
    stop("PET-PEESE prior CDF integration failed.", call. = FALSE)
  }

  pmin(pmax(integration$value, 0), 1)
}
.petpeese_prior_component_cdf <- function(component, q){

  if(component$type == "atom"){
    return(as.numeric(q >= component$x))
  }

  component$cdf(q)
}
.petpeese_prior_component_ccdf <- function(component, q){

  if(component$type == "atom"){
    return(as.numeric(q <= component$x))
  }

  component$ccdf(q)
}
.petpeese_prior_component_pdf <- function(component, x){

  if(component$type == "atom"){
    return(ifelse(x == component$x, Inf, 0))
  }

  component$pdf(x)
}
.petpeese_prior_component_quantile <- function(component, p){

  if(component$type == "atom"){
    return(component$x)
  }

  component$quant(p)
}
.petpeese_prior_cdf_range <- function(context, se, tail_prob = .prior_linear_density_tail_prob()){

  ranges <- do.call(rbind, lapply(context$models, function(model){
    .petpeese_prior_model_range(model, se, context$direction_sign, tail_prob)
  }))

  range(ranges[,1], ranges[,2], finite = TRUE)
}
.petpeese_prior_model_range <- function(model, se, direction_sign, tail_prob){

  scale <- switch(
    model$type,
    "PET"   = direction_sign * se,
    "PEESE" = direction_sign * se^2,
    "none"  = 0
  )

  ranges <- list()
  for(mu_component in model$mu){
    mu_range <- .petpeese_prior_component_range(mu_component, tail_prob)
    for(bias_component in model$bias){
      if(abs(scale) <= .prior_linear_density_zero_tol()){
        ranges[[length(ranges) + 1L]] <- mu_range
      }else{
        bias_range <- .petpeese_prior_component_range(bias_component, tail_prob)
        scaled_bias_range <- sort(scale * bias_range)
        ranges[[length(ranges) + 1L]] <- c(
          mu_range[1] + scaled_bias_range[1],
          mu_range[2] + scaled_bias_range[2]
        )
      }
    }
  }

  ranges <- do.call(rbind, ranges)
  range(ranges[,1], ranges[,2], finite = TRUE)
}
.petpeese_prior_component_range <- function(component, tail_prob){

  if(component$type == "atom"){
    return(rep(component$x, 2))
  }

  component$quant(c(tail_prob, 1 - tail_prob))
}
.petpeese_transform_quantiles <- function(quantiles, transformation, transformation_arguments){

  transformed <- .density.prior_transformation_x(
    as.vector(quantiles),
    transformation,
    transformation_arguments
  )
  transformed <- matrix(transformed, nrow = nrow(quantiles), ncol = ncol(quantiles))

  if(any(!is.finite(transformed))){
    stop("PET-PEESE transformed prior quantiles are non-finite.", call. = FALSE)
  }

  rbind(
    transformed[1,],
    pmin(transformed[2,], transformed[3,]),
    pmax(transformed[2,], transformed[3,])
  )
}
.plot_data_prior_list.PETPEESE_sampled <- function(prior_list, x_seq, n_points, n_samples,
                                                   transformation, transformation_arguments, prior_list_mu,
                                                   effect_direction = "positive"){

  prior_weights  <- sapply(prior_list, .prior_model_weight)
  mixing_prop    <- prior_weights / sum(prior_weights)

  prior_list     <- prior_list[round(n_samples * mixing_prop) > 0]
  prior_list_mu  <- prior_list_mu[round(n_samples * mixing_prop) > 0]
  mixing_prop    <- mixing_prop[round(n_samples * mixing_prop) > 0]

  # get the samples
  samples_list <- list()
  for(i in seq_along(prior_list)){
    if(is.prior.PET(prior_list[[i]])){
      samples_list[[i]] <- cbind(rng(prior_list_mu[[i]], round(n_samples * mixing_prop[i])), rng(prior_list[[i]], round(n_samples * mixing_prop[i])), rep(0, length = round(n_samples * mixing_prop[i])))
    }else if(is.prior.PEESE(prior_list[[i]])){
      samples_list[[i]] <- cbind(rng(prior_list_mu[[i]], round(n_samples * mixing_prop[i])), rep(0, length = round(n_samples * mixing_prop[i])), rng(prior_list[[i]], round(n_samples * mixing_prop[i])))
    }else{
      samples_list[[i]] <- cbind(rng(prior_list_mu[[i]], round(n_samples * mixing_prop[i])), matrix(0, nrow = round(n_samples * mixing_prop[i]), ncol = 2))
    }
  }
  samples <- do.call(rbind, samples_list)

  summary <- .petpeese_line_summary_from_samples(
    samples                  = samples,
    x_seq                    = x_seq,
    transformation           = transformation,
    transformation_arguments = transformation_arguments,
    effect_direction         = effect_direction
  )


  out <- list(
    call    = call("density", "PET-PEESE list"),
    bw      = NULL,
    n       = n_points,
    x       = x_seq,
    y       = summary$median,
    y_lCI   = summary$lCI,
    y_uCI   = summary$uCI,
    samples = summary$samples
  )


  class(out) <- c("density", "density.prior", "density.prior.PETPEESE")
  attr(out, "x_range") <- range(x_seq)
  attr(out, "y_range") <- range(summary$median)

  return(out)
}
.plot_data_prior_list.simple         <- function(prior_list, x_seq, x_range, x_range_quant, n_points, n_samples, force_samples, individual,
                                                 transformation, transformation_arguments, transformation_settings){

  if(is.prior.spike_and_slab(prior_list))
    prior_list <- list(prior_list)

  # dispatching for spike and slab priors
  if(length(prior_list) == 1 && is.prior.spike_and_slab(prior_list[[1]])){

    prior_inclusion   <- .get_spike_and_slab_inclusion(prior_list[[1]])
    prior_variable    <- .get_spike_and_slab_variable(prior_list[[1]])

    if(mean(prior_inclusion) < 1 && mean(prior_inclusion) > 0){
      # create a dummy list for the simple mixture
      prior_null                        <- prior("spike", list(0), prior_weights = 1-mean(prior_inclusion))
      prior_variable[["prior_weights"]] <- mean(prior_inclusion)

      prior_list <- list(
        prior_variable,
        prior_null
      )
    }else if(mean(prior_inclusion) >= 1){
      prior_list <- list(prior_variable)
    }else if(mean(prior_inclusion) <= 0){
      prior_list <- list(prior("spike", list(0)))
    }
  }

  # join the same priors
  prior_list <- .simplify_prior_list(prior_list)

  # get common range to ascertain that all priors are aligned
  if(is.null(x_range)){
    if(!is.null(x_seq)){
      x_range <- range(x_seq)
    }else{
      x_range <- range(as.vector(do.call(rbind, lapply(prior_list, function(p) range(p, if(is.null(x_range_quant)) .range.prior_quantile_default(p) else x_range_quant)))))
    }
  }

  prior_weights  <- sapply(prior_list, .prior_model_weight)
  mixing_prop    <- prior_weights / sum(prior_weights)

  prior_list  <- prior_list[round(n_samples * mixing_prop) > 1]
  mixing_prop <- mixing_prop[round(n_samples * mixing_prop) > 0]

  plot_data <- list()
  for(i in seq_along(prior_list)){
    plot_data[[i]] <- density(prior_list[[i]], x_seq = x_seq, x_range = x_range, x_range_quant = x_range_quant,
                              n_points = n_points, n_samples = round(n_samples * mixing_prop[i]), force_samples = force_samples,
                              transformation = transformation, transformation_arguments = transformation_arguments,
                              transformation_settings = transformation_settings, individual = individual, truncate_end = FALSE)
  }

  # the complete samples are added to each output object
  x_sam    <- NULL
  x_points <- NULL
  y_points <- NULL
  x_den    <- NULL
  y_den    <- NULL

  for(i in seq_along(plot_data)){

    if(force_samples){
      x_sam <- c(x_sam, plot_data[[i]]$samples)
    }

    # align points and densities
    if(inherits(plot_data[[i]], "density.prior.point")){
      x_points <- c(x_points, plot_data[[i]]$x[plot_data[[i]]$y != 0])
      y_points <- c(y_points, mixing_prop[i])
    }else if(inherits(plot_data[[i]], "density.prior.simple") | inherits(plot_data[[i]], "density.prior.orthonormal") | inherits(plot_data[[i]], "density.prior.meandif")){
      x_den <- rbind(x_den, plot_data[[i]]$x)
      y_den <- rbind(y_den, plot_data[[i]]$y * mixing_prop[i])
    }
  }

  # deal with continuous densities
  if(!is.null(y_den)){
    y_den <- apply(y_den, 2, sum)
    if(any(sapply(1:nrow(x_den), function(i) !isTRUE(all.equal(x_den[1,], x_den[i,])))))
      stop("non-matching x-coordinates")
    x_den <- x_den[1,]

    # set the endpoints to zero if they correspond to truncation
    prior_list_simple <- prior_list[!sapply(prior_list, is.prior.point)]
    prior_list_simple_lower <- min(sapply(prior_list_simple, function(p) p$truncation[["lower"]]))
    prior_list_simple_upper <- max(sapply(prior_list_simple, function(p) p$truncation[["upper"]]))
    if(!is.null(transformation)){
      prior_list_simple_lower   <- .density.prior_transformation_x(prior_list_simple_lower, transformation, transformation_arguments)
      prior_list_simple_upper   <- .density.prior_transformation_x(prior_list_simple_upper, transformation, transformation_arguments)
    }
    if(isTRUE(all.equal(prior_list_simple_lower, x_den[1])) | prior_list_simple_lower >= x_den[1]){
      y_den <- c(0, y_den)
      x_den <- c(x_den[1], x_den)
    }
    if(isTRUE(all.equal(prior_list_simple_upper, x_den[length(x_den)])) | prior_list_simple_upper <= x_den[length(x_den)]){
      y_den <- c(y_den, 0)
      x_den <- c(x_den, x_den[length(x_den)])
    }
  }


  # create the output object
  out <- list()

  # add continuous densities
  if(!is.null(y_den)){
    out_den    <- list(
      call    = call("density", "list priors"),
      bw      = NULL,
      n       = n_points,
      x       = x_den,
      y       = y_den,
      samples = x_sam
    )

    class(out_den) <- c("density", "density.prior", "density.prior.simple")
    attr(out_den, "x_range") <- range(x_den)
    attr(out_den, "y_range") <- c(0, max(y_den))

    out[["density"]] <- out_den
  }

  # add spikes
  if(!is.null(y_points)){
    for(i in seq_along(y_points)){
      temp_points <- list(
        call    = call("density", paste0("point", i)),
        bw      = NULL,
        n       = n_points,
        x       = x_points[i],
        y       = y_points[i],
        samples = x_sam
      )

      class(temp_points) <- c("density", "density.prior", "density.prior.point")
      attr(temp_points, "x_range") <- range(x_points)
      attr(temp_points, "y_range") <- c(0, max(y_points[i]))

      out[[paste0("points",i)]] <- temp_points
    }
  }

  return(out)
}

.simplify_prior_list <- function(prior_list){

  # as_mixed_priors stores non-listed priors
  # (needs to be kept for marginal_ etc)
  if(is.prior.mixture(prior_list)){
    class(prior_list) <- NULL
  } else if(is.prior(prior_list)){
    prior_list <- list(prior_list)
  }


  # return the input with fewer than 2 priors
  if(length(prior_list) < 2){
    return(prior_list)
  }

  new_prior_list <- prior_list
  for(i in seq_along(new_prior_list)){
    new_prior_list[[i]][["prior_weights"]] <- NULL
  }

  # remove all attributes but names and class
  for(i in seq_along(new_prior_list)){
    attributes(new_prior_list[[i]])[!names(attributes(new_prior_list[[i]])) %in% c("names", "class")] <- NULL
  }

  # remove identical priors
  are_equal <- do.call(rbind, lapply(new_prior_list, function(p) sapply(new_prior_list, identical, y = p)))
  are_equal <- are_equal[!duplicated(are_equal) & apply(are_equal, 1, sum) > 1,,drop = FALSE]

  # return the input with no matches
  if(nrow(are_equal) == 0){
    return(prior_list)
  }

  # find the duplicates and collect prior odds
  prior_weights <- unname(sapply(prior_list, .prior_model_weight))
  to_remove  <- NULL
  for(i in 1:nrow(are_equal)){
    this_ind    <- c(1:ncol(are_equal))[are_equal[i,]]
    this_unique <- this_ind[1]
    prior_weights[this_unique] <- sum(prior_weights[this_ind])
    to_remove   <- c(to_remove, this_ind[-1])
  }

  # return prior odds
  for(i in seq_along(prior_list)){
    prior_list[[i]] <- .set_prior_model_weight(prior_list[[i]], prior_weights[i])
  }

  # remove the duplicates
  prior_list[to_remove] <- NULL

  return(prior_list)
}


#' @title Add list of prior objects to a plot
#'
#' @param scale_y2 scaling factor for a secondary axis
#' @param ... additional arguments
#' @inheritParams plot_prior_list
#' @inheritParams density.prior
#'
#' @return \code{lines_prior_list} returns \code{NULL}.
#'
#' @seealso [plot_prior_list()] [geom_prior_list()]
#' @rdname lines_prior_list
#' @export
lines_prior_list <- function(prior_list, xlim = NULL, x_seq = NULL, x_range_quant = NULL, n_points = 500,
                             n_samples = 10000, force_samples = FALSE,
                             individual = FALSE, show_figures = if(individual) 1 else NULL,
                             transformation = NULL, transformation_arguments = NULL, transformation_settings = FALSE,
                             rescale_x = FALSE, scale_y2 = NULL, prior_list_mu = NULL, effect_direction = "positive", ...){

  # check input (most arguments are checked within density)
  check_list(prior_list, "prior_list")
  if(!all(sapply(prior_list, is.prior)))
    stop("'prior_list' must be a list of priors.")
  check_bool(individual, "individual")
  check_bool(rescale_x, "rescale_x")
  check_int(show_figures, "show_figures", allow_NULL = TRUE)
  check_real(scale_y2, "scale_y2", lower = 0, allow_NULL = TRUE)
  check_char(effect_direction, "effect_direction", allow_values = c("positive", "negative"))


  # get the plotting type
  if(any(sapply(prior_list, is.prior.weightfunction))){
    prior_type <- "weightfunction"
  }else if(any(sapply(prior_list, is.prior.PET)) | any(sapply(prior_list, is.prior.PEESE))){
    prior_type <- "PETPEESE"
  }else{
    prior_type <- "simple"
  }

  if(prior_type == "PETPEESE"){
    check_list(prior_list_mu, "prior_list_mu", check_length = length(prior_list))
    if(is.prior(prior_list_mu) | !all(sapply(prior_list_mu, is.prior)))
      stop("'prior_list_mu' must be a list of priors (priors for the mu parameter are required for plotting PET-PEESE).")
  }else{
    if(!is.null(prior_list_mu))
      stop("'prior_list_mu' is required only for PET-PEESE plots.")
  }


  # get the plotting range
  if(is.null(xlim) & is.null(x_seq)){
    if(prior_type %in% c("weightfunction", "PETPEESE") & !individual){
      xlim      <- c(0, 1)
    }else if(prior_type == "simple"){
      xlim   <- do.call(rbind, lapply(prior_list, range, quantiles = x_range_quant))
      xlim   <- range(pretty(range(as.vector(xlim))))
    }
  }


  # get the plotting data
  if(prior_type == "weightfunction"){

    # use analytical marginal summaries of the mapped cumulative Dirichlet weights
    plot_data <- .plot_data_prior_list.weightfunction(prior_list, x_seq = x_seq, x_range = xlim, x_range_quant = x_range_quant,
                                                      n_points = n_points, n_samples = n_samples)
    .lines.prior.weightfunction(plot_data = plot_data, rescale_x = rescale_x, ...)

  }else if(prior_type == "PETPEESE"){

    # use deterministic linear-combination summaries when supported, with a sampling fallback
    plot_data <- .plot_data_prior_list.PETPEESE(prior_list, x_seq = x_seq, x_range = xlim, x_range_quant = x_range_quant,
                                                n_points = n_points, n_samples = n_samples,
                                                transformation = transformation, transformation_arguments = transformation_arguments,
                                                transformation_settings = transformation_settings, prior_list_mu = prior_list_mu,
                                                effect_direction = effect_direction)
    .lines.prior.PETPEESE(plot_data = plot_data, ...)

  }else if(prior_type == "simple"){

    # solve analytically
    plot_data <- .plot_data_prior_list.simple(prior_list, x_seq = x_seq, x_range = xlim, x_range_quant = x_range_quant,
                                              n_points = n_points, n_samples = n_samples, force_samples = force_samples, individual = individual,
                                              transformation = transformation, transformation_arguments = transformation_arguments,
                                              transformation_settings = transformation_settings)

    if(is.null(scale_y2)){
      scale_y2 <- .get_scale_y2(plot_data, ...)
    }
    for(i in seq_along(plot_data)){
      if(inherits(plot_data[[i]], what = "density.prior.simple")){
        .lines.prior.simple(plot_data[[i]], ...)
      }else if(inherits(plot_data[[i]], what = "density.prior.point")){
        .lines.prior.point(plot_data[[i]], scale_y2 = scale_y2, ...)
      }
    }

  }

  return(invisible())
}

#' @title Add list of prior objects to a plot
#'
#' @inheritParams lines_prior_list
#' @inheritParams density.prior
#' @inheritParams plot_prior_list
#'
#' @return \code{geom_prior_list} returns an object of class 'ggplot'.
#'
#' @seealso [plot_prior_list()] [lines_prior_list()]
#' @rdname geom_prior_list
#' @export
geom_prior_list  <- function(prior_list, xlim = NULL, x_seq = NULL, x_range_quant = NULL, n_points = 500,
                             n_samples = 10000, force_samples = FALSE,
                             individual = FALSE, show_figures = if(individual) 1 else NULL,
                             transformation = NULL, transformation_arguments = NULL, transformation_settings = FALSE,
                             rescale_x = FALSE, scale_y2 = NULL, prior_list_mu = NULL, effect_direction = "positive", ...){

  # check input (most arguments are checked within density)
  check_list(prior_list, "prior_list")
  if(is.prior(prior_list) | !all(sapply(prior_list, is.prior)))
    stop("'prior_list' must be a list of priors.")
  check_bool(individual, "individual")
  check_bool(rescale_x, "rescale_x")
  check_int(show_figures, "show_figures", allow_NULL = TRUE)
  check_real(scale_y2, "scale_y2", lower = 0, allow_NULL = TRUE)
  check_char(effect_direction, "effect_direction", allow_values = c("positive", "negative"))


  # get the plotting type
  if(any(sapply(prior_list, is.prior.weightfunction))){
    prior_type <- "weightfunction"
  }else if(any(sapply(prior_list, is.prior.PET)) | any(sapply(prior_list, is.prior.PEESE))){
    prior_type <- "PETPEESE"
  }else{
    prior_type <- "simple"
  }

  if(prior_type == "PETPEESE"){
    check_list(prior_list_mu, "prior_list_mu", check_length = length(prior_list))
    if(is.prior(prior_list_mu) | !all(sapply(prior_list_mu, is.prior)))
      stop("'prior_list_mu' must be a list of priors (priors for the mu parameter are required for plotting PET-PEESE).")
  }else{
    if(!is.null(prior_list_mu))
      stop("'prior_list_mu' is required only for PET-PEESE plots.")
  }

  # get the plotting range
  if(is.null(xlim) & is.null(x_seq)){
    if(prior_type %in% c("weightfunction", "PETPEESE") & !individual){
      xlim      <- c(0, 1)
    }else if(prior_type == "simple"){
      xlim   <- do.call(rbind, lapply(prior_list, range, quantiles = x_range_quant))
      xlim   <- range(pretty(range(as.vector(xlim))))
    }
  }


  # get the plotting data
  if(prior_type == "weightfunction"){

    # use analytical marginal summaries of the mapped cumulative Dirichlet weights
    plot_data <- .plot_data_prior_list.weightfunction(prior_list, x_seq = x_seq, x_range = xlim, x_range_quant = x_range_quant,
                                                      n_points = n_points, n_samples = n_samples)
    geom <- .geom_prior.weightfunction(plot_data = plot_data, rescale_x = rescale_x, ...)

  }else if(prior_type == "PETPEESE"){

    # use deterministic linear-combination summaries when supported, with a sampling fallback
    plot_data <- .plot_data_prior_list.PETPEESE(prior_list, x_seq = x_seq, x_range = xlim, x_range_quant = x_range_quant,
                                                n_points = n_points, n_samples = n_samples,
                                                transformation = transformation, transformation_arguments = transformation_arguments,
                                                transformation_settings = transformation_settings, prior_list_mu = prior_list_mu,
                                                effect_direction = effect_direction)
    geom <- .geom_prior.PETPEESE(plot_data = plot_data, ...)

  }else if(prior_type == "simple"){

    # solve analytically
    plot_data <- .plot_data_prior_list.simple(prior_list, x_seq = x_seq, x_range = xlim, x_range_quant = x_range_quant,
                                              n_points = n_points, n_samples = n_samples, force_samples = force_samples, individual = individual,
                                              transformation = transformation, transformation_arguments = transformation_arguments,
                                              transformation_settings = transformation_settings)
    geom     <- NULL
    if(is.null(scale_y2)){
      scale_y2 <- .get_scale_y2(plot_data, ...)
    }
    for(i in seq_along(plot_data)){
      if(inherits(plot_data[[i]], what = "density.prior.simple")){
        geom <- c(geom, list(.geom_prior.simple(plot_data[[i]], ...)))
      }else if(inherits(plot_data[[i]], what = "density.prior.point")){
        geom <- c(geom, list(.geom_prior.point(plot_data[[i]], scale_y2 = scale_y2, ...)))
      }
    }

  }

  return(geom)
}


#' @title Plot samples from the mixed posterior distributions
#'
#' @param samples samples from a posterior distribution for a
#' parameter generated by [mix_posteriors] or [as_mixed_posteriors].
#' @param parameter parameter name to be plotted. Use \code{"PETPEESE"}
#' for PET-PEESE plot with parameters \code{"PET"} and \code{"PEESE"},
#' and \code{"weightfunction"} for plotting a weightfunction with
#' parameters \code{"omega"}.
#' @param prior whether prior distribution should be added to the figure.
#' When samples were prepared with \code{as_mixed_posteriors(..., transform_scaled = TRUE)},
#' the transformed prior samples are automatically used.
#' @param effect_direction direction of the effect for PET-PEESE
#' regression. Use \code{"positive"} (default) for
#' \code{mu + PET*se + PEESE*se^2} or \code{"negative"} for
#' \code{mu - PET*se - PEESE*se^2}.
#' @param dots_prior additional arguments for the prior distribution plot
#' @param data optional numeric vector of observed p-values in \code{[0, 1]}, or
#' a data frame with a \code{p} column. Used only for weightfunction plots.
#' @param show_data whether observed p-values should be shown as rug marks on
#' the weightfunction x-axis.
#' @param dots_data additional styling arguments for observed p-value rug marks.
#' Supports \code{col}/\code{color}, \code{alpha}, \code{lwd}/\code{linewidth},
#' \code{side}/\code{rug_side}, and \code{height}/\code{rug_height}.
#' @param legend whether factor legends should be drawn.
#' @param legend_title optional title for factor legends.
#' @param legend_labels optional labels for factor legend levels.
#' @param legend_position optional legend position for factor legends.
#' @param ... additional arguments
#' @inheritParams density.prior
#' @inheritParams plot.prior
#'
#' @details
#' When using scaled predictors (via \code{formula_scale_list} in [JAGS_fit]),
#' you can plot posteriors on the original (unscaled) scale by preparing samples with
#' \code{as_mixed_posteriors(..., transform_scaled = TRUE)}. The function automatically
#' detects this and uses the pre-computed transformed prior samples when \code{prior = TRUE}.
#'
#' @return \code{plot_posterior} returns either \code{NULL} or
#' an object of class 'ggplot' if plot_type is \code{plot_type = "ggplot"}.
#'
#' @seealso [prior()] [lines_prior_list()]  [geom_prior_list()]
#' @export
plot_posterior <- function(samples, parameter, plot_type = "base", prior = FALSE,
                           n_points = 1000, n_samples = 10000, force_samples = FALSE,
                           individual   = FALSE, show_figures = NULL,
                           transformation = NULL, transformation_arguments = NULL, transformation_settings = FALSE,
                           rescale_x = FALSE, par_name = NULL, effect_direction = "positive",
                           dots_prior = list(), data = NULL, show_data = FALSE, dots_data = list(),
                           legend = TRUE, legend_title = NULL, legend_labels = NULL, legend_position = NULL, ...){

  # check input
  check_list(samples, "prior_list")
  if(any(!sapply(samples, inherits, what = "mixed_posteriors")))
    stop("'samples' must be a be an object generated by 'mix_posteriors' function.")
  check_char(parameter, "parameter")
  check_char(plot_type, "plot_type", allow_values = c("base", "ggplot"))
  check_bool(individual, "individual")
  check_bool(rescale_x, "rescale_x")
  check_bool(legend, "legend", allow_NA = FALSE)
  check_bool(show_data, "show_data")
  check_list(dots_data, "dots_data", allow_NULL = TRUE)
  check_int(show_figures, "show_figures", allow_NULL = TRUE, lower = 0)
  check_char(effect_direction, "effect_direction", allow_values = c("positive", "negative"))
  .check_transformation_input(transformation, transformation_arguments, transformation_settings)

  # deal with bad parameter names for PET-PEESE, weightfunction
  if(tolower(gsub("-", "", gsub("_", "", gsub(".", "", parameter, fixed = TRUE),fixed = TRUE), fixed = TRUE)) %in% c("weightfunction", "weigthfunction", "omega")){
    parameter <- "omega"
  }else if(tolower(gsub("-", "", gsub("_", "", gsub(".", "", parameter, fixed = TRUE),fixed = TRUE), fixed = TRUE)) %in% "petpeese"){
    parameter <- "PETPEESE"
  }else if(tolower(gsub("-", "", gsub("_", "", gsub(".", "", parameter, fixed = TRUE),fixed = TRUE), fixed = TRUE)) %in% "pet"){
    parameter <- "PET"
  }else if(tolower(gsub("-", "", gsub("_", "", gsub(".", "", parameter, fixed = TRUE),fixed = TRUE), fixed = TRUE)) %in% "peese"){
    parameter <- "PEESE"
  }

  if(show_data && !identical(parameter, "omega")){
    stop("'show_data' is currently supported only for weightfunction posterior plots.", call. = FALSE)
  }
  if(show_data && individual){
    stop("'show_data' is supported only for the full weightfunction posterior plot.", call. = FALSE)
  }
  data_p <- .weightfunction_plot_data_pvalues(data, show_data)

  # auto-detect transform_scaled from samples attribute
  transform_scaled <- isTRUE(attr(samples, "transform_scaled"))

  # handle transform_scaled: check for pre-computed prior densities
  prior_densities_transformed <- NULL
  prior_density_context <- attr(samples, "prior_density_context")
  if(transform_scaled && prior){
    prior_densities_transformed <- attr(samples, "prior_densities")
    if(is.null(prior_densities_transformed)){
      stop("Samples were prepared with 'transform_scaled = TRUE' but no prior densities found. ",
           "This should not happen - please report this as a bug.")
    }
  }

  # get the plotting range
  dots <- list(...)
  xlim <- dots[["xlim"]]
  if(is.null(xlim)){
    if(parameter %in% c("PET", "PEESE", "PETPEESE") & !individual){
      xlim <- c(0, 1)
    }else if(parameter == "omega"){
      xlim <- c(0, 1)
    }else{
      # use the data range otherwise
      xlim <- NULL
    }
  }


  if(is.element(parameter, "omega")){

    if (individual) {

      # bias plot parameters require special extraction
      if (!is.null(samples[["bias"]]) && inherits(samples[["bias"]], "mixed_posteriors.bias")) {
        samples <- .simplify_as_mixed_posterior_bias(samples, parameter)
      }
      prior_list  <- attr(samples[[parameter]], "prior_list")
      prior_list  <- .simplify_prior_list(prior_list)


      # plot the individual weight parameters
      out_list  <- list()
      par_names <- colnames(samples[["omega"]])

      if (!is.null(show_figures)) {
        if (show_figures > length(par_names)) {
          stop("'show_figures' corresponds to a number larger than the number of weight function parameters.")
        } else {
          par_names <- par_names[show_figures]
        }
      }

      for (par in par_names) {

        plot_data <- .plot_data_samples.weightparameter(samples, parameter = par, n_points = n_points)

        # add priors, if requested
        if(prior){

          plot_data_prior <- .plot_data_prior_list.weightparameter(prior_list, parameter = par, n_points = n_points, n_samples = n_samples)

          # transplant common xlim and ylim
          plot_data_joined <- c(plot_data_prior, plot_data)

          xlim <- range(as.vector(sapply(plot_data_joined, attr, which = "x_range")))
          attr(plot_data_prior[[1]], "x_range") <- xlim

          if(any(sapply(plot_data_prior, inherits, what = "density.prior.simple")) & any(sapply(plot_data_prior, inherits, what = "density.prior.point"))){
            ylim  <- range(as.vector(sapply(plot_data_joined[sapply(plot_data_joined, inherits, what = "density.prior.simple")], attr, which = "y_range")))
            ylim2 <- range(as.vector(sapply(plot_data_joined[sapply(plot_data_joined, inherits, what = "density.prior.point")],  attr, which = "y_range")))
            attr(plot_data_prior[[which.max(sapply(plot_data_prior, inherits, what = "density.prior.simple"))]], "y_range") <- ylim
            attr(plot_data_prior[[which.max(sapply(plot_data_prior, inherits, what = "density.prior.point"))]],  "y_range") <- ylim2
          }else if(any(sapply(plot_data_prior, inherits, what = "density.prior.simple"))){
            ylim  <- range(as.vector(sapply(plot_data_joined[sapply(plot_data_joined, inherits, what = "density.prior.simple")], attr, which = "y_range")))
            attr(plot_data_prior[[which.max(sapply(plot_data_prior, inherits, what = "density.prior.simple"))]], "y_range") <- ylim
          }else if(any(sapply(plot_data_prior, inherits, what = "density.prior.point"))){
            ylim  <- range(as.vector(sapply(plot_data_joined[sapply(plot_data_joined, inherits, what = "density.prior.point")],  attr, which = "y_range")))
            attr(plot_data_prior[[which.max(sapply(plot_data_prior, inherits, what = "density.prior.point"))]], "y_range") <- ylim
          }

          scale_y2   <- .get_scale_y2(plot_data_prior, ...)
          dots_prior <- .transfer_dots(dots_prior, ...)


          # set the y/x ranges
          for(i in seq_along(plot_data)){
            if(inherits(plot_data[[i]], what = "density.prior.point")){
              attr(plot_data[[i]], which = "y_range") <- if(any(sapply(plot_data_prior, inherits, what = "density.prior.simple")) & any(sapply(plot_data_prior, inherits, what = "density.prior.point"))) ylim2 else ylim
            }else{
              attr(plot_data[[i]], which = "y_range") <- ylim
              attr(plot_data[[i]], which = "x_range") <- xlim
            }
          }

          # plot prior
          args_prior           <- dots_prior
          args_prior$plot_data <- plot_data_prior
          args_prior$plot_type <- plot_type
          args_prior$par_name  <- par
          args_prior$scale_y2  <- scale_y2

          plot <- do.call(.plot_prior_list.both, args_prior)


          # plot posterior
          args           <- list(...)
          args$plot_data <- plot_data
          args$plot_type <- plot_type
          args$par_name  <- par
          args$scale_y2  <- scale_y2
          args$add       <- TRUE

          if(plot_type == "base"){
            plot <- do.call(.plot_prior_list.both, args)
          }else if(plot_type == "ggplot"){
            plot <- plot + do.call(.plot_prior_list.both, args)
            out_list[[par]] <- plot
          }

        }else{

          # plot just posterior otherwise
          plot <- .plot_prior_list.both(plot_data = plot_data, plot_type = plot_type, par_name = par, ...)
          out_list[[par]] <- plot

        }
      }

      plot <- out_list




    } else {

      # special dispatching for visualizing the whole weightfunction

      plot_data <- .plot_data_samples.weightfunction(samples, x_seq = NULL, x_range = xlim, x_range_quant = NULL, n_points = n_points)

      # add priors, if requested
      if(prior){

        # extract the correct weightfunction samples
        if(!is.null(samples[[parameter]])){
          prior_list <- attr(samples[[parameter]], "prior_list")
        }else if(!is.null(samples[["bias"]])){
          prior_list <- attr(samples[["bias"]], "prior_list")
        }else{
          stop("No 'omega' or 'bias' samples found.")
        }

        prior_list      <- .simplify_prior_list(prior_list)
        plot_data_prior <- .plot_data_prior_list.weightfunction(prior_list, x_seq = NULL, x_range = xlim, x_range_quant = NULL,
                                                                n_points = n_points, n_samples = n_samples)

        # transplant common xlim and ylim
        plot_data_joined <- list(plot_data_prior, plot_data)

        xlim <- range(as.vector(sapply(plot_data_joined, attr, which = "x_range")))
        ylim <- range(as.vector(sapply(plot_data_joined, attr, which = "y_range")))
        attr(plot_data_prior, "x_range") <- xlim
        attr(plot_data_prior, "y_range") <- ylim
        dots_prior <- .transfer_dots(dots_prior, ...)

        args           <- dots_prior
        args$x         <- prior_list
        args$plot_data <- plot_data_prior
        args$rescale_x <- rescale_x
        args$plot_type <- plot_type
        args$par_name  <- par_name
        plot           <- do.call(.plot.prior.weightfunction, args)

        if(plot_type == "ggplot"){
          plot <- plot + .geom_prior.weightfunction(plot_data, rescale_x = rescale_x, ...)
          if(show_data){
            plot <- plot + .geom.weightfunction_data(data_p, plot_data, rescale_x, dots_data)
          }
        }else{
          .lines.prior.weightfunction(plot_data, rescale_x = rescale_x, ...)
          if(show_data){
            .lines.weightfunction_data(data_p, plot_data, rescale_x, dots_data)
          }
        }

      }else{

        # plot just posterior otherwise
        plot <- .plot.prior.weightfunction(NULL, plot_data = plot_data, plot_type = plot_type, rescale_x = rescale_x, par_name = par_name, ...)
        if(show_data){
          if(plot_type == "ggplot"){
            plot <- plot + .geom.weightfunction_data(data_p, plot_data, rescale_x, dots_data)
          }else{
            .lines.weightfunction_data(data_p, plot_data, rescale_x, dots_data)
          }
        }

      }
    }

  }else if(is.element(parameter, c("PET", "PEESE", "PETPEESE")) && !individual){
    # special dispatching for visualizing the PET-PEESE regression

    plot_data <- .plot_data_samples.PETPEESE(samples, x_seq = NULL, x_range = xlim, x_range_quant = NULL, n_points = n_points,
                                             transformation = transformation, transformation_arguments = transformation_arguments, transformation_settings = transformation_settings,
                                             effect_direction = effect_direction)

    # add priors, if requested
    if(prior){

      if(is.null(samples[["mu"]]) && is.null(samples[["mu_intercept"]]))
        stop("'mu' or 'mu_intercept' samples are required for plotting PET-PEESE.")

      if(!is.null(samples[["mu"]])){
        prior_list_mu <- attr(samples[["mu"]], "prior_list")
      }else if(!is.null(samples[["mu_intercept"]])){
        prior_list_mu <- attr(samples[["mu_intercept"]], "prior_list")
      }
      if(is.prior.simple(prior_list_mu)){
        prior_list_mu <- list(prior_list_mu)
      }

      if (is.null(samples[["bias"]])){
        # TODO: a bit of a hack - removing priors that were added as a fill for sampling
        if(!is.null(samples[["PET"]]) & !is.null(samples[["PEESE"]])){
          prior_list_PET   <- attr(samples[["PET"]],   "prior_list")
          prior_list_PEESE <- attr(samples[["PEESE"]], "prior_list")
          prior_fill       <- seq_along(prior_list_PET)[!sapply(prior_list_PET, is.prior.PET) & !sapply(prior_list_PEESE, is.prior.PEESE)]
          prior_list       <- c(prior_list_PET[sapply(prior_list_PET, is.prior.PET)], prior_list_PEESE[sapply(prior_list_PEESE, is.prior.PEESE)],
                                prior_list_PET[prior_fill])
          prior_list_mu    <- prior_list_mu[c(c(1:length(prior_list_mu))[sapply(prior_list_PET, is.prior.PET)], c(1:length(prior_list_mu))[sapply(prior_list_PEESE, is.prior.PEESE)], c(1:length(prior_list_mu))[prior_fill])]
        }else if(is.null(samples[["PET"]]) & !is.null(samples[["PEESE"]])){
          prior_list <- attr(samples[["PEESE"]], "prior_list")
        }else if(!is.null(samples[["PET"]]) & is.null(samples[["PEESE"]])){
          prior_list <- attr(samples[["PET"]], "prior_list")
        }else{
          stop("Either PET or PEESE samples need to be provided.")
        }
        if(is.prior.simple(prior_list)){
          prior_list <- list(prior_list)
        }
      } else {
        prior_list <- attr(samples[["bias"]], "prior_list")
        prior_list <- prior_list[sapply(prior_list, \(x) is.prior.PET(x) || is.prior.PEESE(x) || is.prior.none(x) || is.prior.point(x))]

        # make cross product of the mixture priors
        priors_grid <- expand.grid(
          "mu" = prior_list_mu,
          "PP" = prior_list
        )
        prior_list_mu <- priors_grid[["mu"]]
        prior_list    <- priors_grid[["PP"]]
      }

      # cannot simplify prior_list - it would break the dependency with mu
      plot_data_prior <- .plot_data_prior_list.PETPEESE(prior_list, x_seq = NULL, x_range = xlim, x_range_quant = NULL,
                                                  n_points = n_points, n_samples = n_samples,
                                                  transformation = transformation, transformation_arguments = transformation_arguments,
                                                  transformation_settings = transformation_settings, prior_list_mu = prior_list_mu,
                                                  effect_direction = effect_direction)

      # transplant common xlim and ylim
      plot_data_joined <- list(plot_data_prior, plot_data)

      xlim <- range(as.vector(sapply(plot_data_joined, attr, which = "x_range")))
      ylim <- range(as.vector(sapply(plot_data_joined, attr, which = "y_range")))
      # make sure y-range does not collapse
      if(all(ylim < .01)){
        ylim <- c(0, 1)
      }
      attr(plot_data_prior, "x_range") <- xlim
      attr(plot_data_prior, "y_range") <- ylim
      dots_prior <- .transfer_dots(dots_prior, ...)

      args           <- dots_prior
      args$x         <- prior_list
      args$plot_data <- plot_data_prior
      args$plot_type <- plot_type
      args$par_name  <- par_name
      plot           <- do.call(.plot.prior.PETPEESE, args)

      if(plot_type == "ggplot"){
        plot <- plot + .geom_prior.PETPEESE(plot_data, ...)
      }else{
        .lines.prior.PETPEESE(plot_data, ...)
      }

    }else{

      # plot just posterior otherwise
      plot <- .plot.prior.PETPEESE(NULL, plot_data = plot_data, plot_type = plot_type, par_name = par_name, ...)

    }


  }else{
    # regular prior distributions (or individual plots for parameters PET-PEESE)

    # bias plot parameters require special extraction
    if (is.element(parameter, c("PET", "PEESE", "PETPEESE")) && !is.null(samples[["bias"]]) && inherits(samples[["bias"]], "mixed_posteriors.bias")) {
      samples <- .simplify_as_mixed_posterior_bias(samples, parameter)
    }
    prior_list  <- attr(samples[[parameter]], "prior_list")
    prior_list  <- .simplify_prior_list(prior_list)


    if(any(sapply(prior_list, is.prior.factor))){
      plot_data <- .plot_data_samples.factor(samples, parameter = parameter, n_points = n_points,
                                             transformation = transformation, transformation_arguments = transformation_arguments, transformation_settings = transformation_settings)
    }else{
      plot_data <- .plot_data_samples.simple(samples, parameter = parameter, n_points = n_points,
                                             transformation = transformation, transformation_arguments = transformation_arguments, transformation_settings = transformation_settings)
    }



    # add priors, if requested
    if(prior){

      # use transformed prior densities if available (from transform_scaled)
      plot_data_prior <- NULL
      if(transform_scaled && !is.null(prior_density_context) && any(sapply(prior_list, is.prior.factor))){
        plot_data_prior <- .plot_data_prior_factor_density_transformed(
          prior_density_context      = prior_density_context,
          samples                   = samples,
          parameter                 = parameter,
          prior_list                = prior_list,
          n_points                  = n_points,
          x_range                   = xlim,
          transformation            = transformation,
          transformation_arguments  = transformation_arguments,
          transformation_settings   = transformation_settings
        )
      }else if(transform_scaled && !is.null(prior_densities_transformed) && parameter %in% names(prior_densities_transformed)){
        plot_data_prior <- .prior_linear_density_to_plot_data(
          prior_densities_transformed[[parameter]],
          n_points                  = n_points,
          x_range                   = xlim,
          transformation            = transformation,
          transformation_arguments  = transformation_arguments,
          transformation_settings   = transformation_settings
        )
      }

      if(is.null(plot_data_prior)){
        plot_data_prior <- .plot_data_prior_list.simple(prior_list, x_seq = NULL, x_range = xlim, x_range_quant = NULL,
                                                  n_points = n_points, n_samples = n_samples, force_samples = force_samples, individual = individual,
                                                  transformation = transformation, transformation_arguments = transformation_arguments,
                                                  transformation_settings = transformation_settings)
      }

      # transplant common xlim and ylim
      plot_data_joined <- c(plot_data_prior, plot_data)

      xlim <- range(as.vector(sapply(plot_data_joined, attr, which = "x_range")))
      attr(plot_data_prior[[1]], "x_range") <- xlim

      if(any(sapply(plot_data_prior, inherits, what = "density.prior.simple")) & any(sapply(plot_data_prior, inherits, what = "density.prior.point"))){
        ylim  <- range(as.vector(sapply(plot_data_joined[sapply(plot_data_joined, inherits, what = "density.prior.simple")], attr, which = "y_range")))
        ylim2 <- range(as.vector(sapply(plot_data_joined[sapply(plot_data_joined, inherits, what = "density.prior.point")],  attr, which = "y_range")))
        attr(plot_data_prior[[which.max(sapply(plot_data_prior, inherits, what = "density.prior.simple"))]], "y_range") <- ylim
        attr(plot_data_prior[[which.max(sapply(plot_data_prior, inherits, what = "density.prior.point"))]],  "y_range") <- ylim2
      }else if(any(sapply(plot_data_prior, inherits, what = "density.prior.simple"))){
        ylim  <- range(as.vector(sapply(plot_data_joined[sapply(plot_data_joined, inherits, what = "density.prior.simple")], attr, which = "y_range")))
        attr(plot_data_prior[[which.max(sapply(plot_data_prior, inherits, what = "density.prior.simple"))]], "y_range") <- ylim
      }else if(any(sapply(plot_data_prior, inherits, what = "density.prior.point"))){
        ylim  <- range(as.vector(sapply(plot_data_joined[sapply(plot_data_joined, inherits, what = "density.prior.point")],  attr, which = "y_range")))
        attr(plot_data_prior[[which.max(sapply(plot_data_prior, inherits, what = "density.prior.point"))]], "y_range") <- ylim
      }

      scale_y2   <- .get_scale_y2(plot_data_prior, ...)
      dots_prior <- .transfer_dots(dots_prior, ...)


      # set the y/x ranges
      for(i in seq_along(plot_data)){
        if(inherits(plot_data[[i]], what = "density.prior.point")){
          attr(plot_data[[i]], which = "y_range") <- if(any(sapply(plot_data_prior, inherits, what = "density.prior.simple")) & any(sapply(plot_data_prior, inherits, what = "density.prior.point"))) ylim2 else ylim
        }else{
          attr(plot_data[[i]], which = "y_range") <- ylim
          attr(plot_data[[i]], which = "x_range") <- xlim
        }
      }

      # plot prior
      args_prior           <- dots_prior
      args_prior$plot_data <- plot_data_prior
      args_prior$plot_type <- plot_type
      args_prior$par_name  <- par_name
      args_prior$scale_y2  <- scale_y2
      if(plot_type == "ggplot"){
        args_prior$hardcode <- TRUE
      }
      args_prior$legend <- FALSE
      args_prior$legend_title <- legend_title
      args_prior$legend_labels <- legend_labels
      args_prior$legend_position <- legend_position

      if(any(sapply(plot_data_prior, inherits, what = "density.prior.factor"))){
        plot <- do.call(.plot_prior_list.factor, args_prior)
      }else{
        plot <- do.call(.plot_prior_list.both, args_prior)
      }


      # plot posterior
      args           <- list(...)
      args$plot_data <- plot_data
      args$plot_type <- plot_type
      args$par_name  <- par_name
      args$scale_y2  <- scale_y2
      args$add       <- TRUE
      args$legend    <- legend
      args$legend_title <- legend_title
      args$legend_labels <- legend_labels
      args$legend_position <- legend_position

      if(plot_type == "base"){
        if(any(sapply(prior_list, is.prior.factor))){
          plot <- do.call(.plot_prior_list.factor, args)
        }else{
          plot <- do.call(.plot_prior_list.both, args)
        }
      }else if(plot_type == "ggplot"){
        if(any(sapply(prior_list, is.prior.factor))){
          plot <- plot + do.call(.plot_prior_list.factor, args)
        }else{
          plot <- plot + do.call(.plot_prior_list.both, args)
        }
      }


    }else{

      # plot just posterior otherwise
      if(any(sapply(prior_list, is.prior.factor))){
        plot <- .plot_prior_list.factor(
          plot_data = plot_data, plot_type = plot_type, par_name = par_name,
          legend = legend, legend_title = legend_title, legend_labels = legend_labels,
          legend_position = legend_position, ...
        )
      }else{
        plot <- .plot_prior_list.both(plot_data = plot_data, plot_type = plot_type, par_name = par_name, ...)
      }

    }


  }


  if(plot_type == "ggplot"){
    return(plot)
  }else{
    return(invisible())
  }
}

# Helper function to create factor plot data from transformed prior densities.
.plot_data_prior_factor_density_transformed <- function(prior_density_context, samples, parameter, prior_list, n_points, x_range = NULL,
                                                        transformation = NULL, transformation_arguments = NULL,
                                                        transformation_settings = FALSE){

  if(is.null(samples[[parameter]]) || !inherits(samples[[parameter]], "mixed_posteriors.factor")){
    return(NULL)
  }

  factor_weights <- .prior_factor_level_weight_matrix(
    sample_metadata = samples[[parameter]],
    parameter       = parameter,
    samples         = samples
  )

  plot_data <- list()
  for(level_i in seq_len(nrow(factor_weights))){
    weights <- rep(0, length(prior_density_context$column_names))
    names(weights) <- prior_density_context$column_names
    weights[colnames(factor_weights)] <- factor_weights[level_i, ]

    level_density <- .prior_density_from_context(prior_density_context, weights)
    level_plot_data <- .prior_linear_density_to_plot_data(
      level_density,
      n_points                  = n_points,
      x_range                   = x_range,
      transformation            = transformation,
      transformation_arguments  = transformation_arguments,
      transformation_settings   = transformation_settings,
      factor                    = TRUE,
      level                     = level_i,
      level_name                = rownames(factor_weights)[level_i]
    )

    for(data_i in seq_along(level_plot_data)){
      plot_data[[paste0("level", level_i, "_", names(level_plot_data)[data_i])]] <- level_plot_data[[data_i]]
    }
  }

  return(plot_data)
}

.plot_data_samples.simple         <- function(samples, parameter, n_points, transformation, transformation_arguments, transformation_settings){

  check_list(samples, "samples", check_names = parameter, allow_other = TRUE)

  x_points <- NULL
  y_points <- NULL
  x_den    <- NULL
  y_den    <- NULL

  # extract the relevant data
  samples    <- samples[[parameter]]
  prior_list <- attr(samples, "prior_list")
  if (!(is.prior.mixture(prior_list) || is.prior.spike_and_slab(prior_list)) && is.prior(prior_list))
    prior_list <- list(prior_list)

  # deal with spikes
  if(any(sapply(prior_list, is.prior.point))){

    # aggregate samples across spikes
    spikes_simplified <- .simplify_spike_samples(samples, prior_list)

    if(nrow(spikes_simplified) > 0){
      x_points <- spikes_simplified[,"location"]
      y_points <- spikes_simplified[,"probability"]
    }else{
      x_points <- NULL
      y_points <- NULL
    }

    # apply transformations
    if(!is.null(transformation)){
      x_points <- .density.prior_transformation_x(x_points, transformation, transformation_arguments)
    }
  }

  # deal with the densities
  if(any(!sapply(prior_list, is.prior.point))){

    samples_density   <- samples[attr(samples, "models_ind") %in% which(!sapply(prior_list, is.prior.point))]

    if(length(samples_density) > 0){

      args <- list(x = samples_density, n = n_points)

      # set the endpoints for possible truncation
      prior_list_simple <- prior_list[!sapply(prior_list, is.prior.point)]
      prior_list_simple_lower <- min(sapply(prior_list_simple, function(p) p$truncation[["lower"]]))
      prior_list_simple_upper <- max(sapply(prior_list_simple, function(p) p$truncation[["upper"]]))
      if(!is.infinite(prior_list_simple_lower)){ # adding a small number for possible transformations (0 -> -Inf)
        args <- c(args, from = prior_list_simple_lower + if(!is.null(transformation)) 1e-5 else 0)
      }
      if(!is.infinite(prior_list_simple_upper)){
        args <- c(args, to = prior_list_simple_upper   - if(!is.null(transformation)) 1e-5 else 0)
      }

      # get the density estimate
      density_continuous <- do.call(stats::density, args)
      x_den    <- density_continuous$x
      y_den    <- density_continuous$y * (length(samples_density) / length(samples))

      # check for truncation
      if(!is.null(transformation)){
        if(isTRUE(all.equal(prior_list_simple_lower + 1e-5, x_den[1])) | prior_list_simple_lower + 1e-5 >= x_den[1]){
          y_den <- c(0, y_den)
          x_den <- c(x_den[1], x_den)
        }
        if(isTRUE(all.equal(prior_list_simple_upper - 1e-5, x_den[length(x_den)])) | prior_list_simple_upper + 1e-5 <= x_den[length(x_den)]){
          y_den <- c(y_den, 0)
          x_den <- c(x_den, x_den[length(x_den)])
        }
      }else{
        if(isTRUE(all.equal(prior_list_simple_lower, x_den[1])) | prior_list_simple_lower >= x_den[1]){
          y_den <- c(0, y_den)
          x_den <- c(x_den[1], x_den)
        }
        if(isTRUE(all.equal(prior_list_simple_upper, x_den[length(x_den)])) | prior_list_simple_upper <= x_den[length(x_den)]){
          y_den <- c(y_den, 0)
          x_den <- c(x_den, x_den[length(x_den)])
        }
      }

      # apply transformations
      if(!is.null(transformation)){
        x_den   <- .density.prior_transformation_x(x_den,   transformation, transformation_arguments)
        y_den   <- .density.prior_transformation_y(x_den, y_den, transformation, transformation_arguments)
        samples_density <- .density.prior_transformation_x(samples_density,   transformation, transformation_arguments)
      }

    }
  }


  # create the output object
  out <- list()

  # add continuous densities
  if(!is.null(y_den)){
    out_den    <- list(
      call    = call("density", "mixed samples"),
      bw      = NULL,
      n       = n_points,
      x       = x_den,
      y       = y_den,
      samples = samples_density
    )

    class(out_den) <- c("density", "density.prior", "density.prior.simple")
    attr(out_den, "x_range") <- range(x_den)
    attr(out_den, "y_range") <- c(0, max(y_den))

    out[["density"]] <- out_den
  }

  # add spikes
  if(!is.null(y_points)){
    for(i in seq_along(y_points)){
      temp_points <- list(
        call    = call("density", paste0("point", i)),
        bw      = NULL,
        n       = n_points,
        x       = x_points[i],
        y       = y_points[i],
        samples = NULL
      )

      class(temp_points) <- c("density", "density.prior", "density.prior.point")
      attr(temp_points, "x_range") <- range(x_points[i])
      attr(temp_points, "y_range") <- c(0, max(y_points[i]))

      out[[paste0("points",i)]] <- temp_points
    }
  }

  return(out)
}
.plot_data_samples.PETPEESE       <- function(samples, x_seq, x_range, x_range_quant, n_points, transformation, transformation_arguments, transformation_settings, effect_direction = "positive"){

  check_list(samples, "samples")
  if (is.null(samples[["mu"]]) && is.null(samples[["mu_intercept"]]))
    stop("'mu' or 'mu_intercept' samples need to be present.")

  if (!is.null(samples[["bias"]])) {

    if(length(c("PET", "PEESE") %in% samples[["bias"]]) == 0)
      stop("At least one 'PET' or 'PEESE' model needs to be specified.")

    # create mu-PET-PEESE samples matrix
    new_samples <- matrix(if(!is.null(samples[["mu"]])) samples[["mu"]] else samples[["mu_intercept"]], ncol = 1)
    for (par in c("PET", "PEESE")) {
      if (is.element(par, colnames(samples[["bias"]]))) {
        new_samples <- cbind(new_samples, samples[["bias"]][,par])
      } else {
        new_samples <- cbind(new_samples, 0)
      }
    }

  } else {

    if(is.null(samples[["PET"]]) & is.null(samples[["PEESE"]]))
      stop("At least one 'PET' or 'PEESE' model needs to be specified.")

    # create mu-PET-PEESE samples matrix
    new_samples <- matrix(if(!is.null(samples[["mu"]])) samples[["mu"]] else samples[["mu_intercept"]], ncol = 1)
    for (par in c("PET", "PEESE")) {
      if (!is.null(samples[[par]])) {
        new_samples <- cbind(new_samples, samples[[par]])
      } else {
        new_samples <- cbind(new_samples, 0)
      }
    }
  }

  # get the plotting range
  if(is.null(x_range)){
    x_range <- c(0, 1)
  }
  if(is.null(x_seq)){
    x_seq   <- seq(x_range[1], x_range[2], length.out = n_points)
  }

  summary <- .petpeese_line_summary_from_samples(
    samples                  = new_samples,
    x_seq                    = x_seq,
    transformation           = transformation,
    transformation_arguments = transformation_arguments,
    effect_direction         = effect_direction
  )


  out <- list(
    call    = call("density", "PET-PEESE list"),
    bw      = NULL,
    n       = n_points,
    x       = x_seq,
    y       = summary$median,
    y_lCI   = summary$lCI,
    y_uCI   = summary$uCI,
    samples = summary$samples
  )


  class(out) <- c("density", "density.prior", "density.prior.PETPEESE")
  attr(out, "x_range") <- range(x_seq)
  attr(out, "y_range") <- range(summary$median)

  return(out)
}
.petpeese_line_summary_from_samples <- function(samples, x_seq, transformation, transformation_arguments,
                                                effect_direction = "positive"){

  samples <- as.matrix(samples)
  if(ncol(samples) != 3){
    stop("'samples' must contain mu, PET, and PEESE columns.", call. = FALSE)
  }

  direction_sign <- if(effect_direction == "negative") -1 else 1
  n_samples <- nrow(samples)

  x_sam <- matrix(samples[,1], nrow = n_samples, ncol = length(x_seq)) +
    direction_sign * matrix(samples[,2], nrow = n_samples, ncol = length(x_seq)) *
      matrix(x_seq, nrow = n_samples, ncol = length(x_seq), byrow = TRUE) +
    direction_sign * matrix(samples[,3], nrow = n_samples, ncol = length(x_seq)) *
      matrix(x_seq^2, nrow = n_samples, ncol = length(x_seq), byrow = TRUE)

  if(!is.null(transformation)){
    x_sam <- .density.prior_transformation_x(x_sam, transformation, transformation_arguments)
  }

  quantiles <- apply(x_sam, 2, stats::quantile, probs = c(.500, .025, .975), names = FALSE)
  quantiles <- matrix(quantiles, nrow = 3)

  list(
    median  = quantiles[1,],
    lCI     = quantiles[2,],
    uCI     = quantiles[3,],
    samples = x_sam
  )
}
.plot_data_samples.weightfunction <- function(samples, x_seq, x_range, x_range_quant, n_points){

  check_list(samples, "samples", check_names = "omega", allow_other = TRUE)
  if(!is.null(samples[["omega"]])){
    samples <- samples[["omega"]]
  }else if(!is.null(samples[["bias"]])){
    samples <- samples[["bias"]]
  }else{
    stop("No 'omega' or 'bias' samples found.")
  }

  prior_list <- attr(samples, "prior_list")
  if (!(is.prior.mixture(prior_list) || is.prior.spike_and_slab(prior_list)) && is.prior(prior_list))
    prior_list <- list(prior_list)

  # get the plotting range
  if(is.null(x_range)){
    x_range <- c(0, 1)
  }
  if(is.null(x_seq)){
    x_seq   <- seq(x_range[1], x_range[2], length.out = n_points)
  }

  context    <- .weightfunction_prior_list_context(prior_list)
  omega_cuts <- context$omega_cuts

  omega_columns <- grepl("^omega\\[", colnames(samples))
  if(any(omega_columns)){
    samples <- samples[, omega_columns, drop = FALSE]
  }

  x_lCI  <- apply(samples, 2, stats::quantile, probs = .025)
  x_uCI  <- apply(samples, 2, stats::quantile, probs = .975)
  x_mean <- apply(samples, 2, mean)

  x_seq     <- omega_cuts
  x_seq_rep <- c(1, sort(rep(2:(length(x_seq)-1), 2)) ,length(x_seq))
  x_val_rep <- sort(rep(1:(length(x_seq)-1), 2))


  out <- list(
    call    = call("density", "weightfunction list"),
    bw      = NULL,
    n       = n_points,
    x       = x_seq[x_seq_rep],
    y       = x_mean[x_val_rep],
    y_lCI   = x_lCI[x_val_rep],
    y_uCI   = x_uCI[x_val_rep],
    samples = samples
  )


  class(out) <- c("density", "density.prior", "density.prior.weightfunction")
  attr(out, "x_range") <- c(0, 1)
  attr(out, "y_range") <- c(0, max(1, x_mean, x_lCI, x_uCI, na.rm = TRUE))

  return(out)
}

.weightfunction_plot_data_pvalues <- function(data, show_data){

  if(!show_data){
    return(numeric())
  }
  if(is.null(data)){
    stop("'data' must be supplied when 'show_data = TRUE'.", call. = FALSE)
  }

  if(is.data.frame(data)){
    if(!"p" %in% names(data)){
      stop("'data' must be a numeric vector of p-values or a data frame with a 'p' column.", call. = FALSE)
    }
    data <- data[["p"]]
  }

  check_real(data, "data", lower = 0, upper = 1, check_length = 0, allow_NA = FALSE)
  data <- data[is.finite(data)]

  return(data)
}

.weightfunction_plot_data_x <- function(p, plot_data, rescale_x){

  if(length(p) == 0L || !rescale_x){
    return(p)
  }

  cuts <- unique(plot_data$x)
  scaled_cuts <- seq(0, 1, length.out = length(cuts))

  stats::approx(
    x    = cuts,
    y    = scaled_cuts,
    xout = p,
    rule = 2,
    ties = "ordered"
  )$y
}

.weightfunction_data_col <- function(dots_data){

  col <- if(!is.null(dots_data[["col"]])){
    dots_data[["col"]]
  }else if(!is.null(dots_data[["color"]])){
    dots_data[["color"]]
  }else{
    "black"
  }

  if(!is.null(dots_data[["alpha"]])){
    col <- grDevices::adjustcolor(col, alpha.f = dots_data[["alpha"]])
  }

  col
}

.weightfunction_data_lwd <- function(dots_data){

  if(!is.null(dots_data[["lwd"]])){
    dots_data[["lwd"]]
  }else if(!is.null(dots_data[["linewidth"]])){
    dots_data[["linewidth"]]
  }else if(!is.null(dots_data[["size"]])){
    dots_data[["size"]]
  }else{
    .5
  }
}

.weightfunction_data_side <- function(dots_data, ggplot = FALSE){

  side <- if(!is.null(dots_data[["side"]])){
    dots_data[["side"]]
  }else if(!is.null(dots_data[["rug_side"]])){
    dots_data[["rug_side"]]
  }else{
    if(ggplot) "b" else 1
  }

  if(ggplot){
    side <- as.character(side)
    if(side %in% c("1", "bottom", "b")){
      return("b")
    }
    if(side %in% c("3", "top", "t")){
      return("t")
    }
    return(side)
  }

  if(is.character(side)){
    if(side %in% c("bottom", "b")){
      return(1)
    }
    if(side %in% c("top", "t")){
      return(3)
    }
  }

  side
}

.weightfunction_data_height <- function(dots_data){

  if(!is.null(dots_data[["height"]])){
    dots_data[["height"]]
  }else if(!is.null(dots_data[["rug_height"]])){
    dots_data[["rug_height"]]
  }else if(!is.null(dots_data[["ticksize"]])){
    dots_data[["ticksize"]]
  }else{
    .03
  }
}

.lines.weightfunction_data <- function(p, plot_data, rescale_x, dots_data = list()){

  if(length(p) == 0L){
    return(invisible())
  }

  p <- .weightfunction_plot_data_x(p, plot_data, rescale_x)

  graphics::rug(
    p,
    side     = .weightfunction_data_side(dots_data, ggplot = FALSE),
    ticksize = .weightfunction_data_height(dots_data),
    col      = .weightfunction_data_col(dots_data),
    lwd      = .weightfunction_data_lwd(dots_data)
  )

  return(invisible())
}

.geom.weightfunction_data <- function(p, plot_data, rescale_x, dots_data = list()){

  if(length(p) == 0L){
    return(NULL)
  }

  p <- .weightfunction_plot_data_x(p, plot_data, rescale_x)

  ggplot2::geom_rug(
    data        = data.frame(p = p),
    mapping     = ggplot2::aes(x = .data[["p"]]),
    inherit.aes = FALSE,
    sides       = .weightfunction_data_side(dots_data, ggplot = TRUE),
    length      = grid::unit(.weightfunction_data_height(dots_data), "npc"),
    color       = .weightfunction_data_col(dots_data),
    linewidth   = .weightfunction_data_lwd(dots_data)
  )
}
.plot_data_samples.weightparameter<- function(samples, parameter, n_points){

  check_list(samples, "samples", check_names = "omega", allow_other = TRUE)
  if(!is.null(samples[["omega"]])){
    samples <- samples[["omega"]]
  }else if(!is.null(samples[["bias"]])){
    samples <- samples[["bias"]]
  }else{
    stop("No 'omega' or 'bias' samples found.")
  }

  x_points <- NULL
  y_points <- NULL
  x_den    <- NULL
  y_den    <- NULL

  # extract the relevant data
  prior_list <- attr(samples, "prior_list")
  models_ind <- attr(samples, "models_ind")
  samples    <- samples[,parameter]
  n_samples_total <- length(samples)
  if (!(is.prior.mixture(prior_list) || is.prior.spike_and_slab(prior_list)) && is.prior(prior_list))
    prior_list <- list(prior_list)

  # replace prior_none with spike(1)
  for (i in seq_along(prior_list)) {
    if (is.prior.none(prior_list[[i]])) {
      temp_weight <- prior_list[[i]][["prior_weights"]]
      prior_list[[i]] <- prior("spike", parameters = list(location = 1))
      prior_list[[i]][["prior_weights"]] <- temp_weight
    }
  }

  # deal with spikes
  samples_is_1 <- abs(samples - 1) < 1e-6

  if(any(samples_is_1)){
    x_points <- 1
    y_points <- mean(samples_is_1)

    # remove the used samples so they are not re-used in density
    # (since they might be forced to one even in non-null models due to cummulativness)
    models_ind <- models_ind[!samples_is_1]
    samples    <- samples[!samples_is_1]

  }else{
    x_points <- NULL
    y_points <- NULL
  }

  # deal with the densities
  if (!all(sapply(prior_list, \(x) is.prior.point(x) || is.prior.none(x)))) {

    samples_density   <- samples[models_ind %in% which(!sapply(prior_list, is.prior.point))]

    if(length(samples_density) > 0){

      args <- list(x = samples_density, n = n_points)

      # set the endpoints for possible truncation
      prior_list_simple <- prior_list[!sapply(prior_list, is.prior.point)]
      context <- .weightfunction_prior_list_context(prior_list)
      parameter_ind <- match(parameter, context$omega_names)
      if(is.na(parameter_ind)){
        prior_list_simple_lower <- min(c(0, samples_density), na.rm = TRUE)
        prior_list_simple_upper <- max(c(1, samples_density), na.rm = TRUE)
      }else{
        components <- .weightfunction_prior_marginal_components(context, parameter_ind)
        prior_range <- .weightfunction_components_range(components, samples = samples_density)
        prior_list_simple_lower <- prior_range[1]
        prior_list_simple_upper <- prior_range[2]
      }

      if(!is.infinite(prior_list_simple_lower)){
        args <- c(args, from = prior_list_simple_lower)
      }
      if(!is.infinite(prior_list_simple_upper)){
        args <- c(args, to = prior_list_simple_upper)
      }

      # get the density estimate
      density_continuous <- do.call(stats::density, args)
      x_den    <- density_continuous$x
      y_den    <- density_continuous$y * (length(samples_density) / n_samples_total)

      # check for truncation
      if(isTRUE(all.equal(prior_list_simple_lower, x_den[1])) | prior_list_simple_lower >= x_den[1]){
        y_den <- c(0, y_den)
        x_den <- c(x_den[1], x_den)
      }
      if(isTRUE(all.equal(prior_list_simple_upper, x_den[length(x_den)])) | prior_list_simple_upper <= x_den[length(x_den)]){
        y_den <- c(y_den, 0)
        x_den <- c(x_den, x_den[length(x_den)])
      }
    }
  }


  # create the output object
  out <- list()

  # add continuous densities
  if(!is.null(y_den)){
    out_den    <- list(
      call    = call("density", "mixed samples"),
      bw      = NULL,
      n       = n_points,
      x       = x_den,
      y       = y_den,
      samples = samples_density
    )

    class(out_den) <- c("density", "density.prior", "density.prior.simple")
    attr(out_den, "x_range") <- range(x_den)
    attr(out_den, "y_range") <- c(0, max(y_den))
    attr(out_den, "parameter") <- parameter

    out[["density"]] <- out_den
  }

  # add spikes
  if(!is.null(y_points)){
    for(i in seq_along(y_points)){
      temp_points <- list(
        call    = call("density", paste0("point", i)),
        bw      = NULL,
        n       = n_points,
        x       = x_points[i],
        y       = y_points[i],
        samples = NULL
      )

      class(temp_points) <- c("density", "density.prior", "density.prior.point")
      attr(temp_points, "x_range") <- if(!is.null(x_den)) range(x_den) else range(c(0, 1, x_points[i]))
      attr(temp_points, "y_range") <- c(0, max(y_points[i]))
      attr(temp_points, "parameter") <- parameter

      out[[paste0("points",i)]] <- temp_points
    }
  }

  return(out)
}
.plot_data_samples.factor         <- function(samples, parameter, n_points, transformation, transformation_arguments, transformation_settings){

  check_list(samples, "samples", check_names = parameter, allow_other = TRUE)

  x_points <- NULL
  y_points <- NULL
  x_den    <- NULL
  y_den    <- NULL

  # transform & extract the relevant data
  prior_list <- attr(samples[[parameter]], "prior_list")
  if (!(is.prior.mixture(prior_list) || is.prior.spike_and_slab(prior_list)) && is.prior(prior_list))
    prior_list <- list(prior_list)

  if(any(sapply(prior_list, function(x) is.prior.orthonormal(x) | is.prior.meandif(x)))){
    samples  <- transform_factor_samples(samples)
    if(!is.null(transformation)){
      message("The transformation was applied to the differences from the mean. Note that non-linear transformations do not map from the meandif/orthonormal contrasts to the differences from the mean.")
    }
  }

  samples    <- samples[[parameter]]

  # create the output object
  out <- list()

  # deal with spikes
  if(any(sapply(prior_list, is.prior.point))){

    # aggregate samples across spikes
    spikes_simplified <- .simplify_spike_samples(samples, prior_list)

    if(nrow(spikes_simplified) > 0){
      x_points <- spikes_simplified[,"location"]
      y_points <- spikes_simplified[,"probability"]
    }else{
      x_points <- NULL
      y_points <- NULL
    }

    # apply transformations
    if(!is.null(transformation)){
      x_points <- .density.prior_transformation_x(x_points, transformation, transformation_arguments)
    }

    for(i in seq_along(y_points)){
      temp_points <- list(
        call    = call("density", paste0("point", i)),
        bw      = NULL,
        n       = n_points,
        x       = x_points[i],
        y       = y_points[i],
        samples = NULL
      )

      class(temp_points) <- c("density", "density.prior", "density.prior.point")
      attr(temp_points, "x_range") <- range(x_points[i])
      attr(temp_points, "y_range") <- c(0, max(y_points[i]))

      out[[paste0("points",i)]] <- temp_points
    }
  }

  # deal with the densities
  if(any(!sapply(prior_list, is.prior.point))){

    samples_density <- samples[attr(samples, "models_ind") %in% which(!sapply(prior_list, is.prior.point)),,drop=FALSE]

    if(nrow(samples_density) > 0){
      for(i in 1:ncol(samples_density)){

        args <- list(x = samples_density[,i], n = n_points)

        # set the endpoints for possible truncation
        prior_list_simple <- prior_list[!sapply(prior_list, is.prior.point)]
        prior_list_simple_lower <- min(sapply(prior_list_simple, function(p) p$truncation[["lower"]]))
        prior_list_simple_upper <- max(sapply(prior_list_simple, function(p) p$truncation[["upper"]]))
        if(!is.infinite(prior_list_simple_lower)){
          args <- c(args, from = prior_list_simple_lower)
        }
        if(!is.infinite(prior_list_simple_upper)){
          args <- c(args, to = prior_list_simple_upper)
        }

        # get the density estimate
        density_continuous <- do.call(stats::density, args)
        x_den    <- density_continuous$x
        y_den    <- density_continuous$y * (nrow(samples_density) / nrow(samples))

        # check for truncation
        if(isTRUE(all.equal(prior_list_simple_lower, x_den[1])) | prior_list_simple_lower >= x_den[1]){
          y_den <- c(0, y_den)
          x_den <- c(x_den[1], x_den)
        }
        if(isTRUE(all.equal(prior_list_simple_upper, x_den[length(x_den)])) | prior_list_simple_upper <= x_den[length(x_den)]){
          y_den <- c(y_den, 0)
          x_den <- c(x_den, x_den[length(x_den)])
        }

        # apply transformations
        if(!is.null(transformation)){
          x_den   <- .density.prior_transformation_x(x_den,   transformation, transformation_arguments)
          y_den   <- .density.prior_transformation_y(x_den, y_den, transformation, transformation_arguments)
          samples_density[,i] <- .density.prior_transformation_x(samples_density[,i],   transformation, transformation_arguments)
        }

        out_den    <- list(
          call    = call("density", "mixed samples"),
          bw      = NULL,
          n       = n_points,
          x       = x_den,
          y       = y_den,
          samples = samples_density
        )

        class(out_den) <- c("density", "density.prior", "density.prior.factor", "density.prior.simple")
        attr(out_den, "x_range")    <- range(x_den)
        attr(out_den, "y_range")    <- c(0, max(y_den))
        attr(out_den, "level")      <- i
        attr(out_den, "level_name") <- colnames(samples_density)[i]

        out[[paste0("density", i)]] <- out_den

      }
    }
  }



  return(out)
}


#' @title Plot estimates from models
#'
#' @details Plots prior and posterior estimates of the same parameter
#' across multiple models (prior distributions with orthonormal/meandif contrast
#' are always plotted as differences from the grand mean).
#'
#' @param parameter parameter name to be plotted. Does not support
#' PET-PEESE and weightfunction.
#' @param inference object created by [ensemble_inference] function
#' @param conditional whether conditional models should be displayed
#' @param order list specifying ordering of the models. The first
#' element describes whether the ordering should be \code{"increasing"}
#' or \code{"decreasing"} and the second element describes whether
#' the ordering should be based \code{"model"} order, \code{"estimate"}
#' size, posterior \code{"probability"}, or the inclusion \code{"BF"}.
#' @param ... additional arguments. E.g.:
#' \describe{
#'   \item{\code{"show_updating"}}{whether Bayes factors and change from
#'   prior to posterior odds should be shown on the secondary y-axis}
#'   \item{\code{"show_estimates"}}{whether posterior estimates and 95% CI
#'   should be shown on the secondary y-axis}
#'   \item{\code{"y_axis2"}}{whether the secondary y-axis should be shown}
#' }
#' @inheritParams ensemble_inference
#' @inheritParams plot.prior
#' @inheritParams plot_posterior
#' @inheritParams format_parameter_names
#'
#' @return \code{plot_models} returns either \code{NULL} or
#' an object of class 'ggplot' if plot_type is \code{plot_type = "ggplot"}.
#'
#' @seealso [prior()] [lines_prior_list()]  [geom_prior_list()]
#' @export
plot_models <- function(model_list, samples, inference, parameter, plot_type = "base", prior = FALSE, conditional = FALSE,
                        order = NULL,
                        transformation = NULL, transformation_arguments = NULL, transformation_settings = FALSE,
                        par_name = NULL, formula_prefix = TRUE, ...){

  # check input
  check_list(model_list, "model_list")
  check_char(parameter, "parameter")
  sapply(model_list, function(m)check_list(m, "model_list:model", check_names = "fit_summary", all_objects = TRUE, allow_other = TRUE))
  if(!all(unlist(sapply(model_list, function(m) sapply(attr(m[["fit"]], "prior_list"), function(p) is.prior(p))))))
    stop("model_list:priors must contain 'BayesTools' priors")
  if(!all(sapply(model_list, function(m)inherits(m[["fit_summary"]], what = "BayesTools_runjags_summary"))))
    stop("model_list:fit_summary must contain 'BayesTools' fit_summary")
  check_list(samples, "samples")
  if(any(!sapply(samples, inherits, what = "mixed_posteriors")))
    stop("'samples' must be a be an object generated by 'mix_posteriors' function.")
  check_list(order, "order", allow_NULL = TRUE)
  if(!is.null(order)){
    check_list(order, "order", check_length = 2)
    check_char(order[[1]], allow_values = c("decreasing", "increasing"))
    check_char(order[[2]], allow_values = c("model", "estimate", "probability", "BF"))
  }

  # extract the objects
  models_summary   <- lapply(model_list, function(m)m[["fit_summary"]])
  models_inference <- lapply(model_list, function(m)m[["inference"]])
  total_inference  <- inference[[parameter]]
  total_samples    <- samples[[parameter]]
  prior_list       <- attr(total_samples, "prior_list")

  # deal with factors
  if(inherits(total_samples, "mixed_posteriors.factor")){

    # make sure that the orthonormal/meandif priors are transformed to differences from the mean
    if(attr(total_samples, "orthonormal") | attr(total_samples, "meandif")){

      # transform the samples
      if(ncol(total_samples) == attr(total_samples, "levels")){
        total_samples <- transform_factor_samples(samples)[[parameter]]
      }
      # transform the model summaries
      if(!any(sapply(models_summary, function(m) all(colnames(total_samples) %in% attr(m, "parameters"))))){
        models_summary <- lapply(model_list, function(m) runjags_estimates_table(m[["fit"]], transform_factors = TRUE))
      }

    }

    parameter <- colnames(total_samples)
  }

  # prepare nice parameter names
  if(is.null(par_name)){
    par_name  <- format_parameter_names(parameter, attr(total_samples, "formula_parameter"), formula_prefix = formula_prefix)
  }

  plot <- list()
  for(i in seq_along(parameter)){
    plot[[i]] <- .plot_models.simple(models_summary = models_summary, models_inference = models_inference, total_inference = total_inference, total_samples = total_samples,
                                     prior_list = prior_list, parameter = parameter[i], par_name = par_name[i],
                                     plot_type = plot_type, prior = prior, conditional = conditional, order = order,
                                     transformation = transformation, transformation_arguments = transformation_arguments, transformation_settings = transformation_settings, ...)
  }


  # return the plots
  if(plot_type == "base"){
    return(invisible())
  }else if(plot_type == "ggplot"){
    if(length(parameter) == 1){
      plot <- plot[[i]]
    }
    return(plot)
  }
}

.plot_models_data_prior     <- function(prior_list, models_inference){
  return(data.frame(
    model      = 1:length(prior_list),
    y          = sapply(prior_list, mean),
    y_lCI      = sapply(prior_list, mquant, .025),
    y_uCI      = sapply(prior_list, mquant, .975),
    prior_prob = sapply(models_inference, function(m)m[["prior_prob"]]),
    post_prob  = sapply(models_inference, function(m)m[["post_prob"]]),
    BF         = sapply(models_inference, function(m)m[["inclusion_BF"]])
  ))
}
.plot_models_data_posterior <- function(models_summary, parameter, prior_list, models_inference){
  return(data.frame(
    model      = 1:length(models_summary),
    y          = sapply(1:length(models_summary), function(i){
      if(any(attr(models_summary[[i]], "parameters") == parameter)){
        return(models_summary[[i]][attr(models_summary[[i]], "parameters") == parameter, "Mean"])
      }else if(is.prior.point(prior_list[[i]])){
        return(prior_list[[i]]$parameters[["location"]])
      }else{
        stop(paste0("Posterior distribution summary for '", parameter, "' is not available."))
      }
    }),
    y_lCI      = sapply(1:length(models_summary), function(i){
      if(any(attr(models_summary[[i]], "parameters") == parameter)){
        return(models_summary[[i]][attr(models_summary[[i]], "parameters") == parameter, "0.025"])
      }else if(is.prior.point(prior_list[[i]])){
        return(prior_list[[i]]$parameters[["location"]])
      }else{
        stop(paste0("Posterior distribution for '", parameter, "' is not available."))
      }
    }),
    y_uCI      = sapply(1:length(models_summary), function(i){
      if(any(attr(models_summary[[i]], "parameters") == parameter)){
        return(models_summary[[i]][attr(models_summary[[i]], "parameters") == parameter, "0.975"])
      }else if(is.prior.point(prior_list[[i]])){
        return(prior_list[[i]]$parameters[["location"]])
      }else{
        stop(paste0("Posterior distribution for '", parameter, "' is not available."))
      }
    }),
    prior_prob = sapply(models_inference, function(m) m[["prior_prob"]]),
    post_prob  = sapply(models_inference, function(m) m[["post_prob"]]),
    BF         = sapply(models_inference, function(m) m[["inclusion_BF"]])
  ))
}
.plot_models.simple         <- function(models_summary, models_inference, total_inference, total_samples,
                                        prior_list, parameter, par_name,
                                        plot_type, prior, conditional, order,
                                        transformation, transformation_arguments, transformation_settings, ...){

  # create a table with results
  prior_data     <- .plot_models_data_prior(prior_list, models_inference)
  posterior_data <- .plot_models_data_posterior(models_summary, parameter, prior_list, models_inference)

  # remove null models if requested (assuming that the overall estimate is already supplied accordingly)
  if(conditional){
    prior_data     <- prior_data[!attr(total_inference, "is_null"),]
    posterior_data <- posterior_data[!attr(total_inference, "is_null"),]
  }

  # apply ordering
  if(!is.null(order)){
    prior_data <- switch(
      order[[2]],
      "model"       = prior_data[order(posterior_data$model,      decreasing = order[[1]] == "decreasing"),],
      "estimate"    = prior_data[order(posterior_data$y,          decreasing = order[[1]] == "decreasing"),],
      "probability" = prior_data[order(posterior_data$post_prob,  decreasing = order[[1]] == "decreasing"),],
      "BF"          = prior_data[order(posterior_data$BF,         decreasing = order[[1]] == "decreasing"),]
    )
    posterior_data <- switch(
      order[[2]],
      "model"       = posterior_data[order(posterior_data$model,      decreasing = order[[1]] == "decreasing"),],
      "estimate"    = posterior_data[order(posterior_data$y,          decreasing = order[[1]] == "decreasing"),],
      "probability" = posterior_data[order(posterior_data$post_prob,  decreasing = order[[1]] == "decreasing"),],
      "BF"          = posterior_data[order(posterior_data$BF,         decreasing = order[[1]] == "decreasing"),]
    )
  }

  # compute overall estimate
  overal_mean <- mean(total_samples)
  overal_lCI  <- unname(stats::quantile(total_samples, .025))
  overal_uCI  <- unname(stats::quantile(total_samples, .975))
  vertical_0  <- 0

  # apply transformations
  if(!is.null(transformation)){
    prior_data[,c("y", "y_lCI","y_uCI")]     <- .density.prior_transformation_x(prior_data[,c("y", "y_lCI","y_uCI")],   transformation, transformation_arguments)
    posterior_data[,c("y", "y_lCI","y_uCI")] <- .density.prior_transformation_x(posterior_data[,c("y", "y_lCI","y_uCI")],   transformation, transformation_arguments)
    overal_mean <- .density.prior_transformation_x(overal_mean, transformation, transformation_arguments)
    overal_lCI  <- .density.prior_transformation_x(overal_lCI,  transformation, transformation_arguments)
    overal_uCI  <- .density.prior_transformation_x(overal_uCI , transformation, transformation_arguments)
    vertical_0  <- .density.prior_transformation_x(vertical_0 , transformation, transformation_arguments)
  }

  # add names and locations
  prior_data$x     <- 3:(nrow(prior_data) + 2)     + .25
  posterior_data$x <- 3:(nrow(posterior_data) + 2)

  posterior_data$y_labels2  <- paste0(
    format(round(posterior_data$y, 2), nsmall = 2),
    " [", format(round(posterior_data$y_lCI, 2), nsmall = 2), ", ",
    format(round(posterior_data$y_uCI, 2), nsmall = 2), "]")
  prior_data$y_labels2      <- paste0(
    "BF = ", format(round(prior_data$BF, 2), nsmall = 2),
    " [", format(round(prior_data$prior_prob, 2), nsmall = 2), " -> ", format(round(prior_data$post_prob, 2), nsmall = 2),"]")
  overal_y_label  <- "Model-Averaged"
  overal_y_label2 <- paste0(
    format(round(overal_mean, 2), nsmall = 2), " [", format(round(overal_lCI, 2), nsmall = 2), ", ",
    format(round(overal_uCI, 2), nsmall = 2), "]")


  # set the plotting values
  dots      <- list(...)

  show_updating  <- is.null(dots[["show_updating"]])  || dots[["show_updating"]]
  show_estimates <- is.null(dots[["show_estimates"]]) || dots[["show_estimates"]]

  y_at      <- c(1,               posterior_data$x)
  y_labels  <- c(overal_y_label,  paste0("Model ", posterior_data$model))
  y_at2     <- c(1,               posterior_data$x[show_estimates],         prior_data$x[show_updating])
  y_labels2 <- c(overal_y_label2, posterior_data$y_labels2[show_estimates], prior_data$y_labels2[show_updating])
  ylim      <- c(0, max(posterior_data$x) + 1)


  if(!is.null(dots[["xlim"]])){
    xlim     <- dots[["xlim"]]
    x_labels <- pretty(xlim)
  }else{
    x_labels  <- pretty(range(c(posterior_data$y_lCI, posterior_data$y_uCI, overal_lCI, overal_uCI, if(prior) c(prior_data$y_lCI, prior_data$y_uCI))))
    xlim      <- range(x_labels)
  }
  if(!is.null(dots[["xlab"]])){
    xlab <- dots[["xlab"]]
  }else{
    xlab <- if(is.null(par_name)) parameter else bquote(.(par_name))
  }
  if(!is.null(dots[["cex"]])){
    cex  <- dots[["cex"]]
  }else{
    cex  <- 4
  }
  if(!is.null(dots[["col2"]])){
    col2  <- dots[["col2"]]
  }else{
    col2  <- "grey80"
  }
  if(!is.null(dots[["col"]])){
    col   <- dots[["col"]]
  }else{
    col   <- "black"
  }


  ### do the plotting
  if(plot_type == "base"){

    oldpar <- graphics::par(no.readonly = TRUE)
    on.exit(graphics::par(mar = oldpar[["mar"]]))

    # set up margins
    if(length(dots[["mar"]]) == 0){
      graphics::par(mar = c(4, max(nchar(y_labels)) * 2/3, 0, max(nchar(y_labels2)) * 1/2))
    }else{
      graphics::par(mar = dots[["mar"]])
    }

    graphics::plot(NA, bty = "n", las = 1, xlab = xlab, ylab = "", main = "", yaxt = "n", ylim = ylim, xlim = xlim)
    graphics::axis(2, at = y_at,  labels = y_labels,  las = 1, col = NA)
    if(is.null(dots[["y_axis2"]]) || (!is.null(dots[["y_axis2"]]) && dots[["y_axis2"]])){
      graphics::axis(4, at = y_at2, labels = y_labels2, las = 1, col = NA, hadj = 0)
    }
    graphics::abline(v = vertical_0, lty = 3)

    graphics::arrows(
      x0 = posterior_data$y_lCI[posterior_data$y_uCI - posterior_data$y_lCI > 0],
      x1 = posterior_data$y_uCI[posterior_data$y_uCI - posterior_data$y_lCI > 0],
      y0 = posterior_data$x[posterior_data$y_uCI - posterior_data$y_lCI > 0],
      code = 3, angle = 90, length = 0.1, col = col)
    graphics::points(posterior_data$y, posterior_data$x, pch = 15, cex = cex*prior_data$post_prob, col = col)

    if(prior){
      graphics::arrows(
        x0 = prior_data$y_lCI[prior_data$y_uCI - prior_data$y_lCI > 0],
        x1 = prior_data$y_uCI[prior_data$y_uCI - prior_data$y_lCI > 0],
        y0 = prior_data$x[prior_data$y_uCI - prior_data$y_lCI > 0],
        code = 3, angle = 90, length = 0.1, col = col2)
      graphics::points(prior_data$y, prior_data$x, pch = 15, cex = cex*prior_data$prior_prob, col = col2)
    }


    graphics::polygon(
      x = c(overal_lCI, overal_mean , overal_uCI, overal_mean),
      y = c(1, 1.25, 1, 0.75),
      col = col
    )

  }else if(plot_type == "ggplot"){

    plot <- ggplot2::ggplot()

    # add the studies
    plot <- plot +ggplot2::geom_errorbar(
      mapping = ggplot2::aes(
        xmin   = posterior_data[posterior_data$y_uCI - posterior_data$y_lCI > 0,]$y_lCI,
        xmax   = posterior_data[posterior_data$y_uCI - posterior_data$y_lCI > 0,]$y_uCI,
        y      = posterior_data[posterior_data$y_uCI - posterior_data$y_lCI > 0,]$x),
      color   = col,
      width   = .25)
    plot <- plot +ggplot2::geom_point(
      mapping = ggplot2::aes(
        x    = posterior_data$y,
        y    = posterior_data$x,
        size = cex * posterior_data$post_prob),
      col   = col,
      shape = 15)

    if(prior){
      plot <- plot +ggplot2::geom_errorbar(
        mapping = ggplot2::aes(
          xmin   = prior_data[prior_data$y_uCI - prior_data$y_lCI > 0,]$y_lCI,
          xmax   = prior_data[prior_data$y_uCI - prior_data$y_lCI > 0,]$y_uCI,
          y      = prior_data[prior_data$y_uCI - prior_data$y_lCI > 0,]$x),
        color   = col2,
        width   = .25)
      plot <- plot +ggplot2::geom_point(
        mapping = ggplot2::aes(
          x    = prior_data$y,
          y    = prior_data$x,
          size = cex * prior_data$prior_prob),
        shape = 15,
        col   = col2)
    }

    # add the overall estimate
    plot <- plot + ggplot2::geom_polygon(
      mapping = ggplot2::aes(
        x = c(overal_lCI, overal_mean , overal_uCI, overal_mean),
        y = c(1, 1.25, 1, 0.75)),
      fill = col)

    # add the vertical line
    plot <- plot + ggplot2::geom_line(
      mapping = ggplot2::aes(
        x = c(vertical_0, vertical_0),
        y = ylim),
      linetype = "dotted")

    # add all the other stuff
    if(is.null(dots[["y_axis2"]]) || (!is.null(dots[["y_axis2"]]) && dots[["y_axis2"]])){
      plot <- plot + ggplot2::scale_y_continuous(
        name = "", breaks = y_at, labels = y_labels, limits = ylim,
        sec.axis = ggplot2::sec_axis( ~ ., breaks = y_at2, labels = y_labels2))
      attr(plot, "sec_axis") <- TRUE
    }else{
      plot <- plot + ggplot2::scale_y_continuous(
        name = "", breaks = y_at, labels = y_labels, limits = ylim)
      attr(plot, "sec_axis") <- FALSE
    }
    plot <- plot + ggplot2::scale_x_continuous(
      name = xlab, breaks = x_labels, labels = x_labels, limits = xlim)
    plot <- plot + ggplot2::theme(
      axis.title.y      = ggplot2::element_blank(),
      axis.line.y       = ggplot2::element_blank(),
      axis.ticks.y      = ggplot2::element_blank(),
      axis.text.y       = ggplot2::element_text(hjust = 0, color = "black"),
      axis.text.y.right = ggplot2::element_text(hjust = 1, color = "black"),
      legend.position   = "none")

  }

  return(plot)
}

.simplify_spike_samples           <- function(samples, prior_list){

  # Check if we're dealing with spike_and_slab or mixture (which are single priors) vs list of priors
  is_spike_and_slab <- is.prior.spike_and_slab(prior_list)
  is_mixture <- is.prior.mixture(prior_list)

  # If we have a spike_and_slab or mixture prior, we need to iterate over their components
  # Otherwise, we have a list of individual priors
  if(is_spike_and_slab || is_mixture) {
    # For spike_and_slab and mixture, iterate over the components
    components_to_iterate <- prior_list
    component_indices <- seq_along(prior_list)
  } else {
    # For lists of priors, iterate over the list
    components_to_iterate <- prior_list
    component_indices <- seq_along(prior_list)
  }

  # aggregate for each spike
  priors_point_map <- data.frame(do.call(rbind, lapply(component_indices, function(i) {
    current_component <- components_to_iterate[[i]]
    if(is.prior.point(current_component)){
      if(is_spike_and_slab) {
        # For spike_and_slab: dbern() generates 0 (null) and 1 (alternative)
        # We need to determine which component this is
        component_name <- attr(current_component, "component")
        model_index <- if(component_name == "null") 0 else 1
      } else {
        # For mixture or list of priors: dcat() generates 1, 2, 3... so index i maps to JAGS index i
        model_index <- i
      }
      c("location" = current_component$parameters[["location"]], "frequency" = sum(attr(samples, "models_ind") == model_index))
    }
  })))


  # return the input with fewer than 2 inputs
  if(nrow(priors_point_map) < 2){
    spike_probability = data.frame(cbind(
      "location"    = priors_point_map[, "location"],
      "probability" = priors_point_map[, "frequency"] / if(!is.matrix(samples)) length(samples) else nrow(samples) ))
    spike_probability <- spike_probability[priors_point_map[, "frequency"] != 0, ]
    return(spike_probability)
  }

  # find unique spikes
  unique_map <- cbind("location" = unique(priors_point_map[, "location"]), "frequency" = 0)

  # collect them
  for(i in 1:nrow(unique_map)){
    unique_map[i, "frequency"] <- sum(priors_point_map[sapply(priors_point_map[, "location"], function(l) isTRUE(all.equal(l, unname(unique_map[i, "location"])))), "frequency"])
  }

  spike_probability = data.frame(cbind(
    "location"    = unique_map[, "location"],
    "probability" = unique_map[, "frequency"] / if(!is.matrix(samples)) length(samples) else nrow(samples) ))
  spike_probability <- spike_probability[unique_map[, "frequency"] != 0, ]

  return(spike_probability)
}
.simplify_as_mixed_posterior_bias <- function(samples, parameter) {

  ### replace all remaining priors by null prior
  prior_list <- attr(samples[["bias"]], "prior_list")
  prior_list <- .weightfunction_expand_bias_mixture_priors(prior_list)

  if (parameter == "PET") {
    prior_ind <- which(sapply(prior_list, \(x) !is.prior.PET(x)))
  } else if (parameter == "PEESE") {
    prior_ind <- which(sapply(prior_list, \(x) !is.prior.PEESE(x)))
  } else if (parameter == "omega") {
    prior_ind <- which(sapply(prior_list, \(x) !.weightfunction_prior_has_selection(x)))
  }
  if (length(prior_ind) > 0) {
    for (i in prior_ind) {
      temp_weight     <- prior_list[[i]][["prior_weights"]]
      prior_list[[i]] <- if (parameter == "omega") prior_none() else prior("point", parameters = list(0))
      prior_list[[i]][["prior_weights"]] <- temp_weight
    }
  }

  ### create new samples
  new_samples <- samples[["bias"]][, grepl(parameter, colnames(samples[["bias"]])),drop=FALSE]

  ### store attribute
  std_attrs  <- c("dim", "dimnames", "names", "prior_list", "mcpar")
  all_attrs  <- attributes(samples[["bias"]])
  to_restore <- setdiff(names(all_attrs), std_attrs)

  ### re-assign attributes
  for (a in to_restore) {
    attr(new_samples, a) <- all_attrs[[a]]
  }

  # remove `mixed_posteriors.bias` class
  class(new_samples) <- class(new_samples)[!class(new_samples) %in% "mixed_posteriors.bias"]

  ### assign prior list and model indicator
  attr(prior_list, "omega_context") <- attr(samples[["bias"]], "omega_context")
  attr(new_samples, "prior_list") <- prior_list

  ### remove the old samples & store new samples
  samples[["bias"]]    <- NULL
  samples[[parameter]] <- new_samples

  return(samples)
}

#' @title Plot samples from the marginal posterior distributions
#'
#' @param samples samples from a posterior distribution for a
#' parameter generated by [marginal_inference].
#' @param parameter parameter name to be plotted.
#' @param ... additional arguments
#' @inheritParams plot_posterior
#' @inheritParams density.prior
#' @inheritParams plot.prior
#'
#' @return \code{plot_marginal} returns either \code{NULL} or
#' an object of class 'ggplot' if plot_type is \code{plot_type = "ggplot"}.
#'
#' @seealso [prior()] [marginal_inference()]  [plot_posterior()]
#' @export
plot_marginal <- function(samples, parameter, plot_type = "base", prior = FALSE,
                          n_points = 1000,
                          transformation = NULL, transformation_arguments = NULL, transformation_settings = FALSE,
                          rescale_x = FALSE, par_name = NULL, dots_prior = list(),
                          legend = TRUE, legend_title = NULL, legend_labels = NULL, legend_position = NULL, ...){

  # check input
  if(any(!sapply(samples, inherits, what = "marginal_posterior")))
    stop("'samples' must be a be an object generated by 'marginal_posterior' function.")
  check_char(parameter, "parameter")
  check_char(plot_type, "plot_type", allow_values = c("base", "ggplot"))
  check_bool(legend, "legend", allow_NA = FALSE)
  .check_transformation_input(transformation, transformation_arguments, transformation_settings)


  # get the plotting range
  dots <- list(...)
  xlim <- dots[["xlim"]]


  plot_data <- .plot_data_marginal_samples(samples, parameter = parameter, prior = prior, n_points = n_points,
                                           transformation = transformation, transformation_arguments = transformation_arguments, transformation_settings = transformation_settings)



  # add priors, if requested
  if(prior){

    plot_data_prior <- unlist(lapply(plot_data, attr, which = "prior"), recursive = FALSE)

    # transplant common xlim and ylim
    plot_data_joined <- c(plot_data, plot_data_prior)

    xlim <- range(as.vector(sapply(plot_data_joined, attr, which = "x_range")))
    attr(plot_data_prior[[1]], "x_range") <- xlim

    xlim <- range(as.vector(sapply(plot_data_joined, attr, which = "y_range")))
    attr(plot_data_prior[[1]], "y_range") <- xlim

    dots_prior <- .transfer_dots(dots_prior, ...)


    # plot prior
    args_prior           <- dots_prior
    args_prior$plot_data <- plot_data_prior
    args_prior$plot_type <- plot_type
    args_prior$par_name  <- par_name
    args_prior$hardcode  <- TRUE
    args_prior$legend    <- FALSE
    args_prior$legend_title <- legend_title
    args_prior$legend_labels <- legend_labels
    args_prior$legend_position <- legend_position

    plot <- do.call(.plot_prior_list.factor, args_prior)


    # plot posterior
    args           <- list(...)
    args$plot_data <- plot_data
    args$plot_type <- plot_type
    args$par_name  <- par_name
    args$add       <- TRUE
    args$legend    <- legend
    args$legend_title <- legend_title
    args$legend_labels <- legend_labels
    args$legend_position <- legend_position

    if(plot_type == "base"){
      plot <- do.call(.plot_prior_list.factor, args)
    }else if(plot_type == "ggplot"){
      plot <- plot + do.call(.plot_prior_list.factor, args)
    }

  }else{

    # plot just posterior otherwise
    plot <- .plot_prior_list.factor(
      plot_data = plot_data, plot_type = plot_type, par_name = par_name,
      legend = legend, legend_title = legend_title, legend_labels = legend_labels,
      legend_position = legend_position, ...
    )

  }


  if(plot_type == "ggplot"){
    return(plot)
  }else{
    return(invisible())
  }

}

.plot_data_marginal_samples     <- function(samples, parameter, prior, n_points, transformation, transformation_arguments, transformation_settings){

  check_list(samples, "samples", check_names = parameter, allow_other = TRUE)

  # extract the relevant information
  if(is.list(samples[[parameter]]) && length(samples[[parameter]]) > 1){
    posterior_samples <- .marginal_posterior_parameter_samples(samples, parameter)
    prior_densities   <- .marginal_posterior_parameter_prior_densities(samples, parameter)
    if(prior){
      if(any(vapply(prior_densities, is.null, logical(1))))
        stop("'samples' did not contain prior densities")
    }
  }else{
    posterior_samples <- .marginal_posterior_parameter_samples(samples, parameter)
    prior_densities   <- .marginal_posterior_parameter_prior_densities(samples, parameter)
    if(prior){
      if(is.null(prior_densities[[1]]))
        stop("'samples' did not contain prior densities")
    }
  }


  # create the output object
  out <- list()


  # deal with the densities
  for(i in seq_along(posterior_samples)){

    out_level <- .plot_data_marginal_samples.den(
      posterior_samples[[i]],
      n_points,
      transformation,
      transformation_arguments,
      transformation_settings,
      prior_density = prior_densities[[i]]
    )

    for(j in seq_along(out_level)){
      attr(out_level[[j]], "level")      <- i
      attr(out_level[[j]], "level_name") <- names(posterior_samples)[i]
    }

    if(prior){
      out_den.prior <- .prior_linear_density_to_plot_data(
        prior_densities[[i]],
        n_points                  = n_points,
        transformation            = transformation,
        transformation_arguments  = transformation_arguments,
        transformation_settings   = transformation_settings,
        factor                    = TRUE,
        level                     = i,
        level_name                = names(posterior_samples)[i]
      )

      attr(out_level[[1]], "prior") <- out_den.prior
    }

    for(j in seq_along(out_level)){
      out[[paste0(names(out_level)[j], i)]] <- out_level[[j]]
    }

  }

  return(out)
}
.plot_data_marginal_samples.den <- function(x, n_points, transformation, transformation_arguments, transformation_settings,
                                            prior_density = NULL){

  point_locations <- .plot_data_marginal_samples_point_locations(prior_density)
  point_samples   <- rep(FALSE, length(x))
  x_points        <- NULL
  y_points        <- NULL

  if(length(point_locations) > 0){
    point_counts <- numeric(length(point_locations))
    for(i in seq_along(point_locations)){
      tol <- sqrt(.Machine$double.eps) * max(1, abs(point_locations[i]))
      point_matches <- !point_samples & abs(x - point_locations[i]) <= tol
      point_counts[i] <- sum(point_matches)
      point_samples <- point_samples | point_matches
    }

    point_keep <- point_counts > 0
    x_points <- point_locations[point_keep]
    y_points <- point_counts[point_keep] / length(x)
  }

  samples_density <- x[!point_samples]

  # create the output object
  out <- list()

  # get the density estimate
  if(length(samples_density) > 1 && diff(range(samples_density)) > 0){
    args <- list(x = samples_density, n = n_points)

    density_continuous <- do.call(stats::density, args)
    x_den    <- density_continuous$x
    y_den    <- density_continuous$y * (length(samples_density) / length(x))

    # apply transformations
    if(!is.null(transformation)){
      x_den           <- .density.prior_transformation_x(x_den, transformation, transformation_arguments)
      y_den           <- .density.prior_transformation_y(x_den, y_den, transformation, transformation_arguments)
      samples_density <- .density.prior_transformation_x(samples_density, transformation, transformation_arguments)
    }

    out_den <- list(
      call    = call("density", "mixed samples"),
      bw      = NULL,
      n       = n_points,
      x       = x_den,
      y       = y_den,
      samples = samples_density
    )

    class(out_den) <- c("density", "density.prior", "density.prior.factor", "density.prior.simple")
    attr(out_den, "x_range")    <- range(x_den)
    attr(out_den, "y_range")    <- c(0, max(y_den))

    out[["density"]] <- out_den

  }else if(length(samples_density) > 0){
    x_points <- c(x_points, unique(samples_density))
    y_points <- c(y_points, tabulate(match(samples_density, unique(samples_density))) / length(x))
  }

  if(!is.null(y_points)){
    if(!is.null(transformation)){
      x_points <- .density.prior_transformation_x(x_points, transformation, transformation_arguments)
    }

    for(i in seq_along(y_points)){
      temp_points <- list(
        call    = call("density", paste0("point", i)),
        bw      = NULL,
        n       = n_points,
        x       = x_points[i],
        y       = y_points[i],
        samples = NULL
      )

      class(temp_points) <- c("density", "density.prior", "density.prior.point", "density.prior.factor")
      attr(temp_points, "x_range") <- range(x_points)
      attr(temp_points, "y_range") <- c(0, max(y_points))

      out[[paste0("points", i)]] <- temp_points
    }
  }

  return(out)
}
.plot_data_marginal_samples_point_locations <- function(prior_density){

  if(!inherits(prior_density, "prior_linear_density") || is.null(prior_density$points) || nrow(prior_density$points) == 0){
    return(numeric())
  }

  points <- prior_density$points
  points <- points[points$p > .prior_linear_density_zero_tol(), , drop = FALSE]
  if(nrow(points) == 0){
    return(numeric())
  }

  points$x
}

#' @title Plots a prior object
#'
#' @param x a prior
#' @param plot_type whether to use a base plot \code{"base"}
#' or ggplot2 \code{"ggplot"} for plotting.
#' @param xlim x plotting range
#' @param par_name a type of parameter for which the prior is
#' specified. Only relevant if the prior corresponds to a mu
#' parameter that needs to be transformed.
#' @param rescale_x allows to rescale x-axis in case a
#' weightfunction is plotted.
#' @param show_figures which figures should be returned in case of
#' multiple plots are generated. Useful when priors for the omega
#' parameter are plotted and \code{individual = TRUE}.
#' @param ... additional arguments
#' @inheritParams density.prior
#'
#' @examples
#' # create some prior distributions
#' p0 <- prior(distribution = "point",  parameters = list(location = 0))
#' p1 <- prior(distribution = "normal", parameters = list(mean = 0, sd = 1))
#' p2 <- prior(distribution = "normal", parameters = list(mean = 0, sd = 1), truncation = list(0, Inf))
#'
#' # a default plot
#' plot(p0)
#'
#' # manipulate line thickness and color, change the parameter name
#' plot(p1, lwd = 2, col = "blue", par_name = bquote(mu))
#'
#' # use ggplot
#' plot(p2, plot_type = "ggplot")
#'
#' # utilize the ggplot prior geom
#' plot(p2, plot_type = "ggplot", xlim = c(-2, 2)) + geom_prior(p1, col = "red", lty = 2)
#'
#' # apply transformation
#' plot(p1, transformation = "exp")
#'
#' @return \code{plot.prior} returns either \code{NULL} or
#' an object of class 'ggplot' if plot_type is \code{plot_type = "ggplot"}.
#'
#' @seealso [prior()] [lines.prior()]  [geom_prior()]
#' @rdname plot.prior
#' @export
plot.prior <- function(x, plot_type = "base",
                       x_seq = NULL, xlim = NULL, x_range_quant = NULL, n_points = 1000,
                       n_samples = 10000, force_samples = FALSE,
                       transformation = NULL, transformation_arguments = NULL, transformation_settings = FALSE,
                       show_figures = if(individual) -1 else NULL, individual = FALSE, rescale_x = FALSE, par_name = NULL, ...){

  # check input (most arguments are checked within density)
  .check_prior(x)
  check_char(plot_type, "plot_type", allow_values = c("base", "ggplot"))
  check_bool(individual, "individual")
  check_bool(rescale_x, "rescale_x")
  check_int(show_figures, "show_figures", allow_NULL = TRUE)

  if(is.prior.mixture(x)){
    class(x) <- NULL
    return(plot_prior_list(x, plot_type = plot_type, x_seq = x_seq, xlim = xlim, x_range_quant = x_range_quant, n_points = n_points,
                           n_samples = n_samples, force_samples = force_samples, transformation = transformation,
                           transformation_arguments = transformation_arguments, transformation_settings = transformation_settings,
                           show_figures = show_figures, individual = individual, rescale_x = rescale_x, par_name = par_name, ...))
  }


  # get the plotting data
  if(is.null(xlim) & is.null(x_seq)){
    if((is.prior.PET(x) | is.prior.PEESE(x) | is.prior.weightfunction(x)) & !individual){
      xlim   <- c(0, 1)
    }else if(is.prior.spike_and_slab(x)){
      xlim   <- range(c(0,range(.get_spike_and_slab_variable(x), quantiles = x_range_quant)))
      xlim   <- range(pretty(xlim))
    }else{
      xlim   <- range(x, quantiles = x_range_quant)
      xlim   <- range(pretty(xlim))
    }
  }
  plot_data <- density(x = x, x_seq = x_seq, x_range = xlim, x_range_quant = x_range_quant,
                       n_points = n_points, n_samples = n_samples, force_samples = force_samples,
                       transformation = transformation, transformation_arguments = transformation_arguments,
                       transformation_settings = transformation_settings, individual = individual)


  # plot a weightfunction
  if(is.prior.weightfunction(x)){

    if(!individual){
      plot <- .plot.prior.weightfunction(x = x, plot_type = plot_type, plot_data = plot_data, rescale_x = rescale_x, par_name = par_name, ...)
      if(plot_type == "ggplot"){
        return(plot)
      }else{
        return(invisible())
      }
    }

    # plot individual weights
    if(individual){
      # deal with the multiple figures
      if(is.null(show_figures)){
        plots_ind <- c(1:length(plot_data))
      }else{
        plots_ind <- c(1:length(plot_data))[show_figures]
      }

      # reverse the order
      plot_data <- plot_data[length(plot_data):1]

      plots <- list()
      for(figure in plots_ind){
        if(inherits(plot_data[[figure]], "density.prior.simple")){
          plots[[figure]] <- .plot.prior.simple(x = x, plot_type = plot_type, plot_data = plot_data[[figure]], par_name = par_name, ...)
        }else if(inherits(plot_data[[figure]], "density.prior.point")){
          plots[[figure]] <- .plot.prior.point( x = x, plot_type = plot_type, plot_data = plot_data[[figure]], par_name = par_name, ...)
        }
      }
      if(plot_type == "ggplot"){
        if(length(plots_ind) == 1){
          plots <- plots[[plots_ind]]
        }
      }

      if(plot_type == "ggplot"){
        return(plots)
      }else{
        return(invisible())
      }
    }

  }

  # plot PET-PEESE
  if((is.prior.PET(x) | is.prior.PEESE(x)) & !individual){
    plots <- .plot.prior.PETPEESE(x = x, plot_type = plot_type, plot_data = plot_data, par_name = par_name, ...)
    if(plot_type == "ggplot"){
      return(plots)
    }else{
      return(invisible())
    }
  }

  # spike and slab prior plots
  if(is.prior.spike_and_slab(x)){
    plots <- .plot.prior.spike_and_slab(x = x, plot_type = plot_type, plot_data = plot_data, par_name = par_name, ...)
    if(plot_type == "ggplot"){
      return(plots)
    }else{
      return(invisible())
    }
  }

  # point prior plots
  if(is.prior.point(x)){
    plots <- .plot.prior.point( x = x, plot_type = plot_type, plot_data = plot_data, par_name = par_name, ...)
    if(plot_type == "ggplot"){
      return(plots)
    }else{
      return(invisible())
    }
  }

  # plot orthonormal and meandif priors
  if(is.prior.orthonormal(x) | is.prior.meandif(x)){
    plots <- .plot.prior.orthonormal_or_meandif(x = x, plot_type = plot_type, plot_data = plot_data, par_name = par_name, ...)
    if(plot_type == "ggplot"){
      return(plots)
    }else{
      return(invisible())
    }
  }

  # discrete prior plots
  if(is.prior.discrete(x)){
    plots <- .plot.prior.discrete(x = x, plot_type = plot_type, plot_data = plot_data, par_name = par_name, ...)
    if(plot_type == "ggplot"){
      return(plots)
    }else{
      return(invisible())
    }
  }

  # default prior plots
  if(is.prior.simple(x)){
    plots <- .plot.prior.simple(x = x, plot_type = plot_type, plot_data = plot_data, par_name = par_name, ...)
    if(plot_type == "ggplot"){
      return(plots)
    }else{
      return(invisible())
    }
  }
}

.plot.prior.point          <- function(x, plot_type, plot_data, par_name = NULL, ...){

  # get default plot settings
  dots      <- list(...)

  xlim      <- attr(plot_data, "x_range")
  ylim      <- c(0, 1)

  short_name      <- if(is.null(dots[["short_name"]]))      FALSE else dots[["short_name"]]
  parameter_names <- if(is.null(dots[["parameter_names"]])) FALSE else dots[["parameter_names"]]
  prior_label     <- .plot.prior_label(x, plot_data, short_name, parameter_names)

  main      <- if(!is.null(attr(plot_data, "steps"))) prior_label else ""
  xlab      <- if(!is.null(attr(plot_data, "steps"))) bquote(omega["["*.(attr(plot_data, "steps")[1])*","~.(attr(plot_data, "steps")[2])*"]"]) else bquote(.(if(is.prior.orthonormal(x) | is.prior.meandif(x))"dif")*.(if(!is.null(par_name)){bquote(.(par_name)~"~")})~.(prior_label))
  ylab      <- if(!is.null(dots[["ylab"]])) dots[["ylab"]] else "Probability"

  # add it to the user input if desired
  if(is.null(dots[["main"]])) dots$main <-  main
  if(is.null(dots[["xlab"]])) dots$xlab <-  xlab
  if(is.null(dots[["ylab"]])) dots$ylab <-  ylab
  if(is.null(dots[["xlim"]])) dots$xlim <-  xlim
  if(is.null(dots[["ylim"]])) dots$ylim <-  ylim


  if(plot_type == "base"){

    .plot.prior_empty("point", dots)
    .lines.prior.point(plot_data, ...)

    plot <- NULL

  }else if(plot_type == "ggplot"){

    plot <- .ggplot.prior_empty("point", dots)
    plot <- plot + .geom_prior.point(plot_data, ...)

  }

  # return the plots
  if(plot_type == "base"){
    return(invisible())
  }else if(plot_type == "ggplot"){
    return(plot)
  }
}
.plot.prior.simple         <- function(x, plot_type, plot_data, par_name = NULL, ...){

  # get default plot settings
  dots      <- list(...)

  xlim      <- attr(plot_data, "x_range")
  ylim      <- attr(plot_data, "y_range")

  short_name      <- if(is.null(dots[["short_name"]]))      FALSE else dots[["short_name"]]
  parameter_names <- if(is.null(dots[["parameter_names"]])) FALSE else dots[["parameter_names"]]
  prior_label     <- .plot.prior_label(x, plot_data, short_name, parameter_names)

  main      <- if(!is.null(attr(plot_data, "steps"))) prior_label else ""
  xlab      <- if(!is.null(attr(plot_data, "steps"))) bquote(omega["["*.(attr(plot_data, "steps")[1])*","~.(attr(plot_data, "steps")[2])*"]"])  else bquote(.(if(!is.null(par_name)){bquote(.(par_name)~"~")})~.(prior_label))
  ylab      <- "Density"

  # add it to the user input if desired
  if(is.null(dots[["main"]])) dots$main <-  main
  if(is.null(dots[["xlab"]])) dots$xlab <-  xlab
  if(is.null(dots[["ylab"]])) dots$ylab <-  ylab
  if(is.null(dots[["xlim"]])) dots$xlim <-  xlim
  if(is.null(dots[["ylim"]])) dots$ylim <-  ylim


  if(plot_type == "base"){

    .plot.prior_empty("simple", dots)
    .lines.prior.simple(plot_data, ...)
    plot <- NULL

  }else if(plot_type == "ggplot"){

    plot <- .ggplot.prior_empty("simple", dots)
    plot <- plot + .geom_prior.simple(plot_data, ...)

  }

  # return the plots
  if(plot_type == "base"){
    return(invisible())
  }else if(plot_type == "ggplot"){
    return(plot)
  }
}
.plot.prior.discrete       <- function(x, plot_type, plot_data, par_name = NULL, ...){

  # get default plot settings
  dots      <- list(...)

  xlim      <- attr(plot_data, "x_range")
  ylim      <- attr(plot_data, "y_range")

  short_name      <- if(is.null(dots[["short_name"]]))      FALSE else dots[["short_name"]]
  parameter_names <- if(is.null(dots[["parameter_names"]])) FALSE else dots[["parameter_names"]]
  prior_label     <- .plot.prior_label(x, plot_data, short_name, parameter_names)

  main      <- if(!is.null(attr(plot_data, "steps"))) prior_label else ""
  xlab      <- if(!is.null(attr(plot_data, "steps"))) bquote(omega["["*.(attr(plot_data, "steps")[1])*","~.(attr(plot_data, "steps")[2])*"]"])  else bquote(.(if(!is.null(par_name)){bquote(.(par_name)~"~")})~.(prior_label))
  ylab      <- "Probability"

  # add it to the user input if desired
  if(is.null(dots[["main"]])) dots$main <-  main
  if(is.null(dots[["xlab"]])) dots$xlab <-  xlab
  if(is.null(dots[["ylab"]])) dots$ylab <-  ylab
  if(is.null(dots[["xlim"]])) dots$xlim <-  xlim
  if(is.null(dots[["ylim"]])) dots$ylim <-  ylim


  if(plot_type == "base"){

    .plot.prior_empty("simple", dots)
    .lines.prior.discrete(plot_data, ...)
    plot <- NULL

  }else if(plot_type == "ggplot"){

    plot <- .ggplot.prior_empty("simple", dots)
    plot <- plot + .geom_prior.discrete(plot_data, ...)

  }

  # return the plots
  if(plot_type == "base"){
    return(invisible())
  }else if(plot_type == "ggplot"){
    return(plot)
  }
}
.plot.prior.weightfunction <- function(x, plot_type, plot_data, rescale_x, par_name = NULL, ...){

  # get default plot settings
  dots      <- list(...)

  short_name      <- if(is.null(dots[["short_name"]]))      FALSE else dots[["short_name"]]
  parameter_names <- if(is.null(dots[["parameter_names"]])) FALSE else dots[["parameter_names"]]
  prior_label     <- .plot.prior_label(x, plot_data, short_name, parameter_names)

  xlab      <- if(!is.null(dots[["xlab"]])) dots[["xlab"]] else bquote(italic(p)*"-value")
  main      <- if(!is.null(dots[["main"]])) dots[["main"]] else if(is.prior(x)) bquote(.(if(!is.null(par_name)){bquote(.(par_name)~"~")})~.(prior_label)) else if(!is.null(par_name)) bquote(.(par_name)) else "Selection Models"
  ylab      <- if(!is.null(dots[["ylab"]])) dots[["ylab"]] else "Probability"

  xlim      <- attr(plot_data, "x_range")
  ylim      <- if(!is.null(dots[["ylim"]])) dots[["ylim"]] else attr(plot_data, "y_range")

  # weightfunction specific stuff (required for axes)
  x_cuts <- plot_data$x

  if(rescale_x){
    x_at <- seq(0, 1, length.out = length(unique(plot_data$x)))
    x_at <- x_at[c(1, sort(rep(2:(length(x_at)-1), 2)), length(x_at))]
  }else{
    x_at <- x_cuts
  }

  # add it to the user input if desired
  if(is.null(dots[["main"]])) dots$main <-  main
  if(is.null(dots[["xlab"]])) dots$xlab <-  xlab
  if(is.null(dots[["ylab"]])) dots$ylab <-  ylab
  if(is.null(dots[["xlim"]])) dots$xlim <-  xlim
  if(is.null(dots[["ylim"]])) dots$ylim <-  ylim
  dots$x_at     <- unique(x_at)
  dots$x_labels <- unique(x_cuts)


  if(plot_type == "base"){

    .plot.prior_empty("weightfunction", dots)
    .lines.prior.weightfunction(plot_data, rescale_x, ...)

    plot <- NULL

  }else if(plot_type == "ggplot"){

    plot <- .ggplot.prior_empty("weightfunction", dots)
    plot <- plot + .geom_prior.weightfunction(plot_data, rescale_x, ...)

  }

  # return the plots
  if(plot_type == "base"){
    return(invisible())
  }else if(plot_type == "ggplot"){
    return(plot)
  }
}
.plot.prior.PETPEESE       <- function(x, plot_type, plot_data, par_name = NULL, ...){

  # get default plot settings
  dots      <- list(...)

  short_name      <- if(is.null(dots[["short_name"]]))      FALSE else dots[["short_name"]]
  parameter_names <- if(is.null(dots[["parameter_names"]])) FALSE else dots[["parameter_names"]]
  prior_label     <- .plot.prior_label(x, plot_data, short_name, parameter_names)

  xlab      <- if(!is.null(dots[["xlab"]])) dots[["xlab"]] else "Standard error"
  main      <- if(!is.null(dots[["main"]])) dots[["main"]] else if(is.prior(x)) bquote(.(if(!is.null(par_name)){bquote(.(par_name)~"~")})~.(prior_label)) else if(!is.null(par_name)) bquote(.(par_name)) else "PET-PEESE"
  ylab      <- if(!is.null(dots[["ylab"]])) dots[["ylab"]] else "Effect size"

  xlim      <- attr(plot_data, "x_range")
  ylim      <- if(!is.null(dots[["ylim"]])) dots[["ylim"]] else attr(plot_data, "y_range")

  # add it to the user input if desired
  if(is.null(dots[["main"]])) dots$main <-  main
  if(is.null(dots[["xlab"]])) dots$xlab <-  xlab
  if(is.null(dots[["ylab"]])) dots$ylab <-  ylab
  if(is.null(dots[["xlim"]])) dots$xlim <-  xlim
  if(is.null(dots[["ylim"]])) dots$ylim <-  ylim


  if(plot_type == "base"){

    .plot.prior_empty("PETPEESE", dots)
    .lines.prior.PETPEESE(plot_data, ...)

    plot <- NULL

  }else if(plot_type == "ggplot"){


    plot <- .ggplot.prior_empty("PETPEESE", dots)
    plot <- plot + .geom_prior.PETPEESE(plot_data, ...)

  }

  # return the plots
  if(plot_type == "base"){
    return(invisible())
  }else if(plot_type == "ggplot"){
    return(plot)
  }
}
.plot.prior.orthonormal_or_meandif <- function(x, plot_type, plot_data, par_name = NULL, ...){

  # get default plot settings
  dots      <- list(...)

  xlim      <- attr(plot_data, "x_range")
  ylim      <- attr(plot_data, "y_range")

  short_name      <- if(is.null(dots[["short_name"]]))      FALSE else dots[["short_name"]]
  parameter_names <- if(is.null(dots[["parameter_names"]])) FALSE else dots[["parameter_names"]]
  prior_label     <- .plot.prior_label(x, plot_data, short_name, parameter_names)

  main      <- if(!is.null(attr(plot_data, "steps"))) prior_label else ""
  xlab      <- if(!is.null(attr(plot_data, "steps"))) bquote(omega["["*.(attr(plot_data, "steps")[1])*","~.(attr(plot_data, "steps")[2])*"]"])  else bquote("dif"*.(if(!is.null(par_name)){" "*bquote(.(par_name)~"~")})~.(prior_label))
  ylab      <- "Density"

  # add it to the user input if desired
  if(is.null(dots[["main"]])) dots$main <-  main
  if(is.null(dots[["xlab"]])) dots$xlab <-  xlab
  if(is.null(dots[["ylab"]])) dots$ylab <-  ylab
  if(is.null(dots[["xlim"]])) dots$xlim <-  xlim
  if(is.null(dots[["ylim"]])) dots$ylim <-  ylim


  if(plot_type == "base"){

    .plot.prior_empty("simple", dots)
    .lines.prior.orthonormal_or_meandif(plot_data, ...)
    plot <- NULL

  }else if(plot_type == "ggplot"){

    plot <- .ggplot.prior_empty("simple", dots)
    plot <- plot + .geom_prior.orthonormal_or_meandif(plot_data, ...)

  }

  # return the plots
  if(plot_type == "base"){
    return(invisible())
  }else if(plot_type == "ggplot"){
    return(plot)
  }
}
.plot.prior.spike_and_slab <- function(x, plot_type, plot_data, par_name = NULL, ...){

  # get default plot settings
  dots      <- list(...)

  xlim      <- attr(plot_data, "x_range")
  ylim      <- attr(plot_data, "y_range")

  short_name      <- if(is.null(dots[["short_name"]]))      FALSE else dots[["short_name"]]
  parameter_names <- if(is.null(dots[["parameter_names"]])) FALSE else dots[["parameter_names"]]
  prior_label     <- .plot.prior_label(x, plot_data, short_name, parameter_names)

  main      <- if(!is.null(attr(plot_data, "steps"))) prior_label else ""
  xlab      <- if(!is.null(attr(plot_data, "steps"))) bquote(omega["["*.(attr(plot_data, "steps")[1])*","~.(attr(plot_data, "steps")[2])*"]"])  else bquote(.(if(!is.null(par_name)){bquote(.(par_name)~"~")})~.(prior_label))
  ylab      <- "Density"

  # add it to the user input if desired
  if(is.null(dots[["main"]])) dots$main <-  main
  if(is.null(dots[["xlab"]])) dots$xlab <-  xlab
  if(is.null(dots[["ylab"]])) dots$ylab <-  ylab
  if(is.null(dots[["xlim"]])) dots$xlim <-  xlim
  if(is.null(dots[["ylim"]])) dots$ylim <-  ylim

  args_prior           <- dots
  args_prior$plot_data <- plot_data
  args_prior$plot_type <- plot_type

  plot <- do.call(.plot_prior_list.both, args_prior)

  # return the plots
  if(plot_type == "base"){
    return(invisible())
  }else if(plot_type == "ggplot"){
    return(plot)
  }
}

.plot.prior_empty    <- function(type, dots = list(), ...){

  dots      <- c(dots, list(...))

  main      <- if(!is.null(dots[["main"]]))     dots[["main"]]     else ""
  xlab      <- if(!is.null(dots[["xlab"]]))     dots[["xlab"]]     else ""
  ylab      <- if(!is.null(dots[["ylab"]]))     dots[["ylab"]]     else ""
  xlim      <- if(!is.null(dots[["xlim"]]))     dots[["xlim"]]     else c(0, 1)
  ylim      <- if(!is.null(dots[["ylim"]]))     dots[["ylim"]]     else c(0, 1)
  col.main  <- if(!is.null(dots[["col.main"]])) dots[["col.main"]] else .plot.prior_settings()[["col.main"]]
  cex.axis  <- if(!is.null(dots[["cex.axis"]])) dots[["cex.axis"]] else .plot.prior_settings()[["cex.axis"]]
  cex.lab   <- if(!is.null(dots[["cex.lab"]]))  dots[["cex.lab"]]  else .plot.prior_settings()[["cex.lab"]]
  cex.main  <- if(!is.null(dots[["cex.main"]])) dots[["cex.main"]] else .plot.prior_settings()[["cex.main"]]
  col.axis  <- if(!is.null(dots[["col.axis"]])) dots[["col.axis"]] else .plot.prior_settings()[["col.axis"]]
  col.lab   <- if(!is.null(dots[["col.lab"]]))  dots[["col.lab"]]  else .plot.prior_settings()[["col.lab"]]
  x_at      <- if(!is.null(dots[["x_at"]]))     dots[["x_at"]]     else NULL
  x_labels  <- if(!is.null(dots[["x_labels"]])) dots[["x_labels"]] else NULL

  if(type == "point"){

    graphics::plot(NA, type = "n", bty  = "n", las = 1, xlab = xlab, ylab = ylab, main = main,
                   xlim = xlim, ylim = ylim, axes = FALSE,
                   cex.axis = cex.axis, cex.lab = cex.lab, cex.main = cex.main,
                   col.axis = col.axis, col.lab = col.lab, col.main = col.main)
    graphics::axis(1, col = col.axis, cex = cex.axis)
    graphics::axis(2, at = ylim, labels = ylim, col = col.axis, cex = cex.axis, las = 1)

  }else if(type == "simple"){

    graphics::plot(NA, type = "n", bty  = "n", las = 1, xlab = xlab, ylab = ylab, main = main,
                   xlim = xlim, ylim = ylim,
                   cex.axis = cex.axis, cex.lab = cex.lab, cex.main = cex.main,
                   col.axis = col.axis, col.lab = col.lab, col.main = col.main)

  }else if(type == "weightfunction"){

    graphics::plot(NA, type = "n", bty  = "n", las = 1, xlab = xlab, ylab = ylab, main = main,
                   xlim = xlim, ylim = ylim, axes = FALSE,
                   cex.axis = cex.axis, cex.lab = cex.lab, cex.main = cex.main,
                   col.axis = col.axis, col.lab = col.lab, col.main = col.main)
    graphics::axis(1, at = x_at, labels = x_labels, col = col.axis, cex = cex.axis)
    graphics::axis(2, at = pretty(ylim), col = col.axis, cex = cex.axis, las = 1)

  }else if(type == "PETPEESE"){

    graphics::plot(NA, type = "n", bty  = "n", las = 1, xlab = xlab, ylab = ylab, main = main,
                   xlim = xlim, ylim = ylim,
                   cex.axis = cex.axis, cex.lab = cex.lab, cex.main = cex.main,
                   col.axis = col.axis, col.lab = col.lab, col.main = col.main)

  }else if(type == "both"){

    ylim2    <- if(!is.null(dots[["ylim2"]]))     dots[["ylim2"]]     else ylim
    ylab2    <- if(!is.null(dots[["ylab2"]]))     dots[["ylab2"]]     else ""
    scale_y2 <- if(!is.null(dots[["scale_y2"]]))  dots[["scale_y2"]]  else .plot.prior_settings()[["scale_y2"]]
    scale_y2 <- scale_y2 * max(pretty(ylim)) / max(pretty(ylim2))

    graphics::plot(NA, type = "n", bty  = "n", las = 1, xlab = xlab, ylab = ylab, main = main,
                   xlim = xlim, ylim = range(c(pretty(ylim), pretty(ylim2) * scale_y2)), axes = FALSE,
                   cex.axis = cex.axis, cex.lab = cex.lab, cex.main = cex.main,
                   col.axis = col.axis, col.lab = col.lab, col.main = col.main)
    graphics::axis(1, at = x_at, labels = x_labels, col = col.axis, cex = cex.axis)
    graphics::axis(2, at = pretty(ylim),             labels = pretty(ylim),  col = col.axis, cex = cex.axis, las = 1)
    graphics::axis(4, at = pretty(ylim2) * scale_y2, labels = pretty(ylim2), col = col.axis, cex = cex.axis, las = 1)
    graphics::mtext(ylab2, side = 4, line = 3)

    return(invisible(list(scale_y2 = scale_y2)))
  }

  return(invisible())
}
.ggplot.prior_empty  <- function(type, dots = list(), ...){

  dots      <- c(dots, list(...))

  main      <- if(!is.null(dots[["main"]]))     dots[["main"]]     else ""
  xlab      <- if(!is.null(dots[["xlab"]]))     dots[["xlab"]]     else ""
  ylab      <- if(!is.null(dots[["ylab"]]))     dots[["ylab"]]     else ""
  xlim      <- if(!is.null(dots[["xlim"]]))     dots[["xlim"]]     else c(0, 1)
  ylim      <- if(!is.null(dots[["ylim"]]))     dots[["ylim"]]     else c(0, 1)
  col.main  <- if(!is.null(dots[["col.main"]])) dots[["col.main"]] else .plot.prior_settings()[["col.main"]]
  cex.axis  <- if(!is.null(dots[["cex.axis"]])) dots[["cex.axis"]] else .plot.prior_settings()[["cex.axis"]]
  cex.lab   <- if(!is.null(dots[["cex.lab"]]))  dots[["cex.lab"]]  else .plot.prior_settings()[["cex.lab"]]
  cex.main  <- if(!is.null(dots[["cex.main"]])) dots[["cex.main"]] else .plot.prior_settings()[["cex.main"]]
  col.axis  <- if(!is.null(dots[["col.axis"]])) dots[["col.axis"]] else .plot.prior_settings()[["col.axis"]]
  col.lab   <- if(!is.null(dots[["col.lab"]]))  dots[["col.lab"]]  else .plot.prior_settings()[["col.lab"]]
  x_at      <- if(!is.null(dots[["x_at"]]))     dots[["x_at"]]     else NULL
  x_labels  <- if(!is.null(dots[["x_labels"]])) dots[["x_labels"]] else NULL

  if(type == "point"){

    plot <- ggplot2::ggplot()
    plot <- plot + ggplot2::ggtitle(main)
    plot <- plot + ggplot2::scale_x_continuous(name = xlab, breaks = pretty(xlim), limits = range(pretty(xlim)), oob = scales::oob_keep)
    plot <- plot + ggplot2::scale_y_continuous(name = ylab, breaks = pretty(ylim), limits = range(pretty(ylim)), oob = scales::oob_keep)

    attr(plot, "sec_axis") <- FALSE

  }else if(type == "simple"){

    plot <- ggplot2::ggplot()
    plot <- plot + ggplot2::ggtitle(main)
    plot <- plot + ggplot2::scale_x_continuous(name = xlab, breaks = pretty(xlim), limits = range(pretty(xlim)), oob = scales::oob_keep)
    plot <- plot + ggplot2::scale_y_continuous(name = ylab, breaks = pretty(ylim), limits = range(pretty(ylim)), oob = scales::oob_keep)

    attr(plot, "sec_axis") <- FALSE

  }else if(type == "weightfunction"){

    plot <- ggplot2::ggplot()
    plot <- plot + ggplot2::ggtitle(main)
    plot <- plot + ggplot2::scale_x_continuous(name = xlab, breaks = x_at, labels = x_labels, limits = xlim, oob = scales::oob_keep)
    plot <- plot + ggplot2::scale_y_continuous(name = ylab, breaks = pretty(ylim), limits = ylim,            oob = scales::oob_keep)

    attr(plot, "sec_axis") <- FALSE

  }else if(type == "PETPEESE"){

    plot <- ggplot2::ggplot()
    plot <- plot + ggplot2::ggtitle(main)
    plot <- plot + ggplot2::scale_x_continuous(name = xlab, breaks = pretty(xlim), limits = range(pretty(xlim)), oob = scales::oob_keep)
    plot <- plot + ggplot2::scale_y_continuous(name = ylab, breaks = pretty(ylim), limits = range(pretty(ylim)), oob = scales::oob_keep)

    attr(plot, "sec_axis") <- FALSE

  }else if(type == "both"){

    ylim2    <- if(!is.null(dots[["ylim2"]]))     dots[["ylim2"]]     else ylim
    ylab2    <- if(!is.null(dots[["ylab2"]]))     dots[["ylab2"]]     else ""
    scale_y2 <- if(!is.null(dots[["scale_y2"]]))  dots[["scale_y2"]]  else .plot.prior_settings()[["scale_y2"]]
    scale_y2 <- scale_y2 * max(pretty(ylim)) / max(pretty(ylim2))

    plot <- ggplot2::ggplot()
    plot <- plot + ggplot2::ggtitle(main)
    plot <- plot + ggplot2::scale_x_continuous(name = xlab, breaks = pretty(xlim), limits = range(pretty(xlim)), oob = scales::oob_keep)
    plot <- plot + ggplot2::scale_y_continuous(name = ylab, breaks = pretty(ylim), limits = range(c(pretty(ylim), pretty(ylim2) * scale_y2)), oob = scales::oob_keep, sec.axis = ggplot2::sec_axis(~ ., name = ylab2, breaks = pretty(ylim2) * scale_y2, labels = pretty(ylim2)))

    attr(plot, "scale_y2") <- scale_y2
    attr(plot, "sec_axis") <- TRUE

  }

  plot <- plot + ggplot2::theme(
    axis.text  = ggplot2::element_text(size = 10 * cex.axis, color = col.axis),
    axis.title = ggplot2::element_text(size = 10 * cex.lab,  color = col.lab),
    title      = ggplot2::element_text(size = 10 * cex.main, color = col.main))

  return(plot)
}

.plot.prior_label <- function(x, plot_data, short_name = FALSE, parameter_names = FALSE){

  if(!is.prior(x)){
    return(NULL)
  }

  label <- print(x, plot = TRUE, short_name = short_name, parameter_names = parameter_names)
  transformation <- attr(plot_data, "transformation")

  if(is.null(transformation)){
    return(label)
  }

  paste0(.plot.prior_transformation_label(transformation), "(", print(x, silent = TRUE), ")")
}
.plot.prior_transformation_label <- function(transformation){

  if(is.character(transformation) && length(transformation) == 1){
    return(switch(
      transformation,
      "lin"     = "linear",
      "exp_lin" = "exp-linear",
      "tanh"    = "tanh",
      "exp"     = "exp",
      "transformed"
    ))
  }

  "transformed"
}

.plot.prior_settings <- function(){
  return(list(
  cex       = 1,
  cex.axis  = 1,
  cex.lab   = 1,
  cex.main  = 1,
  col       = "black",
  col.axis  = "black",
  col.lab   = "black",
  col.main  = "black",
  col.fill  = "grey80",
  lwd       = 1,
  lty       = 1,
  scale_y2  = 1.10,
  width     = 0.20
  ))
}
.get_scale_y2        <- function(plot_data, ...){

  dots      <- list(...)

  if(any(sapply(plot_data, inherits, what = "density.prior.simple")) & any(sapply(plot_data, inherits, what = "density.prior.point"))){

    ylim  <- range(as.vector(sapply(plot_data[sapply(plot_data, inherits, what = "density.prior.simple")], attr, which = "y_range")))
    ylim2 <- range(as.vector(sapply(plot_data[sapply(plot_data, inherits, what = "density.prior.point")],  attr, which = "y_range")))

    scale_y2 <- if(!is.null(dots[["scale_y2"]]))  dots[["scale_y2"]]  else .plot.prior_settings()[["scale_y2"]]
    scale_y2 <- scale_y2 * max(pretty(ylim)) / max(pretty(ylim2))

  }else{

    scale_y2 <- 1

  }

  return(scale_y2)
}
.transfer_dots       <- function(dots, ...){

  dots_main <- list(...)

  dots$main      <- if(!is.null(dots_main[["main"]]))     dots_main[["main"]]
  dots$xlab      <- if(!is.null(dots_main[["xlab"]]))     dots_main[["xlab"]]
  dots$ylab      <- if(!is.null(dots_main[["ylab"]]))     dots_main[["ylab"]]
  dots$xlim      <- if(!is.null(dots_main[["xlim"]]))     dots_main[["xlim"]]
  dots$ylim      <- if(!is.null(dots_main[["ylim"]]))     dots_main[["ylim"]]
  dots$col.main  <- if(!is.null(dots_main[["col.main"]])) dots_main[["col.main"]]
  dots$cex.axis  <- if(!is.null(dots_main[["cex.axis"]])) dots_main[["cex.axis"]]
  dots$cex.lab   <- if(!is.null(dots_main[["cex.lab"]]))  dots_main[["cex.lab"]]
  dots$cex.main  <- if(!is.null(dots_main[["cex.main"]])) dots_main[["cex.main"]]
  dots$col.axis  <- if(!is.null(dots_main[["col.axis"]])) dots_main[["col.axis"]]
  dots$col.lab   <- if(!is.null(dots_main[["col.lab"]]))  dots_main[["col.lab"]]
  dots$x_at      <- if(!is.null(dots_main[["x_at"]]))     dots_main[["x_at"]]
  dots$x_labels  <- if(!is.null(dots_main[["x_labels"]])) dots_main[["x_labels"]]

  return(dots)
}



#' @title Add prior object to a plot
#'
#' @param x a prior
#' @param xlim plotting range of the prior
#' @param rescale_x allows to rescale x-axis in case a
#' weightfunction is plotted.
#' @param show_parameter which parameter should be returned in case of
#' multiple parameters per prior. Useful when priors for the omega
#' parameter are plotted and \code{individual = TRUE}.
#' @param scale_y2 scaling factor for a secondary axis
#' @param ... additional arguments
#' @inheritParams density.prior
#'
#' @return \code{lines.prior} returns \code{NULL}.
#'
#' @seealso [plot.prior()] [geom_prior()]
#' @rdname lines.prior
#' @export
lines.prior <- function(x, xlim = NULL, x_seq = NULL, x_range_quant = NULL, n_points = 1000,
                        n_samples = 10000, force_samples = FALSE,
                        transformation = NULL, transformation_arguments = NULL, transformation_settings = FALSE,
                        show_parameter = if(individual) 1 else NULL, individual = FALSE, rescale_x = FALSE, scale_y2 = 1, ...){

  # check input (most arguments are checked within density)
  .check_prior(x)
  check_bool(individual, "individual")
  check_bool(rescale_x, "rescale_x")
  check_int(show_parameter, "show_parameter", allow_NULL = TRUE)
  check_real(scale_y2, "scale_y2", lower = 0)

  if(is.prior.mixture(x)){
    return(lines_prior_list(x, xlim = xlim, x_seq = x_seq, x_range_quant = x_range_quant, n_points = n_points,
                            n_samples = n_samples, force_samples = force_samples,
                            transformation = transformation, transformation_arguments = transformation_arguments, transformation_settings = transformation_settings,
                            rescale_x = rescale_x, scale_y2 = scale_y2, ...))
  }



  # get the plotting data
  if(is.null(xlim) & is.null(x_seq)){
    if((is.prior.PET(x) | is.prior.PEESE(x) | is.prior.weightfunction(x)) & !individual){
      xlim   <- c(0, 1)
    }else if(is.prior.spike_and_slab(x)){
      xlim   <- range(c(0,range(.get_spike_and_slab_variable(x), quantiles = x_range_quant)))
      xlim   <- range(pretty(xlim))
    }else{
      xlim   <- range(x, quantiles = x_range_quant)
      xlim   <- range(pretty(xlim))
    }
  }
  plot_data <- density(x = x, x_seq = x_seq, x_range = xlim, x_range_quant = x_range_quant,
                       n_points = n_points, n_samples = n_samples, force_samples = force_samples,
                       transformation = transformation, transformation_arguments = transformation_arguments,
                       transformation_settings = transformation_settings, individual = individual)


  # plot a weightfunction
  if(is.prior.weightfunction(x) & !individual){
    .lines.prior.weightfunction(plot_data = plot_data, rescale_x = rescale_x, ...)
    return(invisible())
  }else if(is.prior.weightfunction(x) & individual){
    if(inherits(plot_data[[show_parameter]], "density.prior.simple")){
      .lines.prior.simple(plot_data, ...)
    }else if(inherits(plot_data[[show_parameter]], "density.prior.point")){
      .lines.prior.point(plot_data, scale_y2 = scale_y2, ...)
    }
    return(invisible())
  }

  # plot PET-PEESE
  if((is.prior.PET(x) | is.prior.PEESE(x))){
    if(!individual){
      .lines.prior.PETPEESE(plot_data, ...)
    }else if(inherits(plot_data, "density.prior.simple")){
      .lines.prior.simple(plot_data, ...)
    }else if(inherits(plot_data, "density.prior.point")){
      .lines.prior.point(plot_data, scale_y2 = scale_y2, ...)
    }
    return(invisible())
  }

  # plot spike and slab prior
  if(is.prior.spike_and_slab(x)){
    .lines.prior.spike_and_slab(plot_data, ...)
    return(invisible())
  }

  # point prior plots
  if(is.prior.point(x)){
    .lines.prior.point(plot_data, scale_y2 = scale_y2, ...)
    return(invisible())
  }

  # plot orthonormal and meandif plots
  if(is.prior.orthonormal(x) | is.prior.meandif(x)){
    .lines.prior.orthonormal_or_meandif(plot_data, ...)
    return(invisible())
  }

  # plot discrete prior
  if(is.prior.discrete(x)){
    .lines.prior.discrete(plot_data, ...)
    return(invisible())
  }

  # default prior plots
  if(is.prior.simple(x)){
    .lines.prior.simple(plot_data, ...)
    return(invisible())
  }

  return(invisible())
}

#' @title Add prior object to a ggplot
#'
#' @inheritParams lines.prior
#' @inheritParams density.prior
#'
#' @return \code{geom_prior_list} returns an object of class 'ggplot'.
#'
#' @seealso [plot.prior()] [lines.prior()]
#' @rdname geom_prior
#' @export
geom_prior  <- function(x, xlim = NULL, x_seq = NULL, x_range_quant = NULL, n_points = 1000,
                        n_samples = 10000, force_samples = FALSE,
                        transformation = NULL, transformation_arguments = NULL, transformation_settings = FALSE,
                        show_parameter = if(individual) 1 else NULL, individual = FALSE, rescale_x = FALSE, scale_y2 = 1, ...){

  # check input (most arguments are checked within density)
  .check_prior(x)
  check_bool(individual, "individual")
  check_bool(rescale_x, "rescale_x")
  check_int(show_parameter, "show_parameter", allow_NULL = TRUE)
  check_real(scale_y2, "scale_y2", lower = 0)

  if(is.prior.mixture(x)){
    class(x) <- NULL
    return(geom_prior_list(x, xlim = xlim, x_seq = x_seq, x_range_quant = x_range_quant, n_points = n_points,
                            n_samples = n_samples, force_samples = force_samples,
                            transformation = transformation, transformation_arguments = transformation_arguments, transformation_settings = transformation_settings,
                            rescale_x = rescale_x, scale_y2 = scale_y2, ...))
  }

  # get the plotting data
  if(is.null(xlim) & is.null(x_seq)){
    if((is.prior.PET(x) | is.prior.PEESE(x) | is.prior.weightfunction(x)) & !individual){
      xlim   <- c(0, 1)
    }else{
      xlim   <- range(x, quantiles = x_range_quant)
      xlim   <- range(pretty(xlim))
    }
  }
  plot_data <- density(x = x, x_seq = x_seq, x_range = xlim, x_range_quant = x_range_quant,
                       n_points = n_points, n_samples = n_samples, force_samples = force_samples,
                       transformation = transformation, transformation_arguments = transformation_arguments,
                       transformation_settings = transformation_settings, individual = individual)


  # plot a weightfunction
  if(is.prior.weightfunction(x)){
    if(!individual){
      geom <- .geom_prior.weightfunction(plot_data = plot_data, rescale_x = rescale_x, ...)
    }else if(inherits(plot_data[[show_parameter]], "density.prior.simple")){
      geom <- .geom_prior.simple(plot_data, ...)
    }else if(inherits(plot_data[[show_parameter]], "density.prior.point")){
      geom <- .geom_prior.point(plot_data, ...)
    }
    return(geom)
  }


  # plot PET-PEESE
  if((is.prior.PET(x) | is.prior.PEESE(x))){
    if(!individual){
      geom <- .geom_prior.PETPEESE(plot_data, ...)
    }else if(inherits(plot_data, "density.prior.simple")){
      .geom_prior.simple(plot_data, ...)
    }else if(inherits(plot_data, "density.prior.point")){
      geom <- .geom_prior.point(plot_data, ...)
    }
    return(geom)
  }


  # plot spike and slab prior
  if(is.prior.spike_and_slab(x)){
    geom <- .geom_prior.spike_and_slab(plot_data, ...)
    return(geom)
  }


  # plot point prior
  if(is.prior.point(x)){
    geom <- .geom_prior.point(plot_data, ...)
    return(geom)
  }


  # plot orthonormal and meandif prior
  if(is.prior.orthonormal(x) | is.prior.meandif(x)){
    geom <- .geom_prior.orthonormal_or_meandif(plot_data, ...)
    return(geom)
  }


  # plot discrete prior
  if(is.prior.discrete(x)){
    geom <- .geom_prior.discrete(plot_data, ...)
    return(geom)
  }


  # default prior plots
  if(is.prior.simple(x)){
    geom <- .geom_prior.simple(plot_data, ...)
    return(geom)
  }

  return(invisible())
}


# base plot prior plot elements
.lines.prior.simple          <- function(plot_data, ...){

  dots      <- list(...)
  col       <- if(!is.null(dots[["col"]]))      dots[["col"]]      else .plot.prior_settings()[["col"]]
  lwd       <- if(!is.null(dots[["lwd"]]))      dots[["lwd"]]      else .plot.prior_settings()[["lwd"]]
  lty       <- if(!is.null(dots[["lty"]]))      dots[["lty"]]      else .plot.prior_settings()[["lty"]]


  graphics::lines(x = plot_data$x, y = plot_data$y, type = "l", lwd = lwd, lty = lty, col = col)

  return(invisible())
}
.lines.prior.discrete        <- function(plot_data, ...){

  dots      <- list(...)
  col       <- if(!is.null(dots[["col"]]))   dots[["col"]]   else .plot.prior_settings()[["col"]]
  width     <- if(!is.null(dots[["width"]])) dots[["width"]] else .plot.prior_settings()[["width"]]
  lwd       <- if(!is.null(dots[["lwd"]]))   dots[["lwd"]]   else .plot.prior_settings()[["lwd"]]
  lty       <- if(!is.null(dots[["lty"]]))   dots[["lty"]]   else .plot.prior_settings()[["lty"]]

  graphics::rect(
    xleft   = plot_data$x - width/2,
    ybottom = 0,
    xright  = plot_data$x + width/2,
    ytop    = plot_data$y,
    lwd = lwd, lty = lty, col = col)

  return(invisible())
}
.lines.prior.point           <- function(plot_data, scale_y2 = 1, ...){

  dots      <- list(...)
  col       <- if(!is.null(dots[["col"]]))      dots[["col"]]      else .plot.prior_settings()[["col"]]
  lwd       <- if(!is.null(dots[["lwd"]]))      dots[["lwd"]]      else .plot.prior_settings()[["lwd"]]
  lty       <- if(!is.null(dots[["lty"]]))      dots[["lty"]]      else .plot.prior_settings()[["lty"]]

  if(!all(plot_data$y == 0)){
    graphics::arrows(
      x0 = plot_data$x[plot_data$y != 0],
      y0 = 0,
      y1 = plot_data$y[plot_data$y != 0] * scale_y2,
      lwd = 2*lwd, lty = lty, col = col)
  }

  return(invisible())
}
.lines.prior.weightfunction  <- function(plot_data, rescale_x, ...){

  dots      <- list(...)
  col       <- if(!is.null(dots[["col"]]))      dots[["col"]]      else .plot.prior_settings()[["col"]]
  col.fill  <- if(!is.null(dots[["col.fill"]])) dots[["col.fill"]] else .plot.prior_settings()[["col.fill"]]
  lwd       <- if(!is.null(dots[["lwd"]]))      dots[["lwd"]]      else .plot.prior_settings()[["lwd"]]
  lty       <- if(!is.null(dots[["lty"]]))      dots[["lty"]]      else .plot.prior_settings()[["lty"]]

  # weightfunction specific stuff
  x_cuts <- plot_data$x
  x_mean <- plot_data$y
  x_lCI  <- plot_data$y_lCI
  x_uCI  <- plot_data$y_uCI

  if(rescale_x){
    x_at <- seq(0, 1, length.out = length(unique(plot_data$x)))
    x_at <- x_at[c(1, sort(rep(2:(length(x_at)-1), 2)), length(x_at))]
  }else{
    x_at <- x_cuts
  }


  graphics::polygon(
    x   = c(x_at,  rev(x_at)),
    y   = c(x_lCI, rev(x_uCI)),
    col = col.fill, border = NA
  )
  graphics::lines(x_at, x_mean, lwd = lwd, lty = lty, col = col)


  return(invisible())
}
.lines.prior.PETPEESE        <- function(plot_data, ...){

  dots      <- list(...)
  col       <- if(!is.null(dots[["col"]]))      dots[["col"]]      else .plot.prior_settings()[["col"]]
  col.fill  <- if(!is.null(dots[["col.fill"]])) dots[["col.fill"]] else .plot.prior_settings()[["col.fill"]]
  lwd       <- if(!is.null(dots[["lwd"]]))      dots[["lwd"]]      else .plot.prior_settings()[["lwd"]]
  lty       <- if(!is.null(dots[["lty"]]))      dots[["lty"]]      else .plot.prior_settings()[["lty"]]


  graphics::polygon(
    x   = c(plot_data$x,     rev(plot_data$x)),
    y   = c(plot_data$y_lCI, rev(plot_data$y_uCI)),
    col = col.fill, border = NA
  )
  graphics::lines(plot_data$x, plot_data$y, lwd = lwd, lty = lty, col = col)


  return(invisible())
}
.lines.prior.orthonormal_or_meandif <- function(plot_data, ...){

  dots      <- list(...)
  col       <- if(!is.null(dots[["col"]]))      dots[["col"]]      else .plot.prior_settings()[["col"]]
  lwd       <- if(!is.null(dots[["lwd"]]))      dots[["lwd"]]      else .plot.prior_settings()[["lwd"]]
  lty       <- if(!is.null(dots[["lty"]]))      dots[["lty"]]      else .plot.prior_settings()[["lty"]]


  graphics::lines(x = plot_data$x, y = plot_data$y, type = "l", lwd = lwd, lty = lty, col = col)

  return(invisible())
}
.lines.prior.factor          <- function(plot_data, ...){

  dots <- list(...)
  col  <- if(!is.null(dots[["col"]][dots[["level"]]])) dots[["col"]][dots[["level"]]] else .plot.prior_settings()[["col"]]
  lty  <- if(!is.null(dots[["lty"]][dots[["level"]]])) dots[["lty"]][dots[["level"]]] else .plot.prior_settings()[["lty"]]
  lwd  <- if(!is.null(dots[["lwd"]]))                  dots[["lwd"]]                  else .plot.prior_settings()[["lwd"]]

  graphics::lines(x = plot_data$x, y = plot_data$y, type = "l", lwd = lwd, lty = lty, col = col)

  return(invisible())
}
.lines.prior.spike_and_slab  <- function(plot_data, ...){

  .lines.prior.simple(plot_data[["variable"]], ...)
  .lines.prior.point(plot_data[["inclusion"]], ...)

  return(invisible())
}

# ggplot prior plot elements
.geom_prior.simple           <- function(plot_data, ...){

  dots      <- list(...)
  col       <- if(!is.null(dots[["col"]]))      dots[["col"]]      else .plot.prior_settings()[["col"]]
  lwd       <- if(!is.null(dots[["size"]]))     dots[["size"]]     else  if(!is.null(dots[["lwd"]])) dots[["lwd"]] else .plot.prior_settings()[["lwd"]]
  lty       <- if(!is.null(dots[["linetype"]])) dots[["linetype"]] else  if(!is.null(dots[["lty"]])) dots[["lty"]] else .plot.prior_settings()[["lty"]]

  geom <- ggplot2::geom_line(
    data    = data.frame(
      x = plot_data$x,
      y = plot_data$y),
    mapping = ggplot2::aes(
      x = .data[["x"]],
      y = .data[["y"]]),
    linewidth = lwd, linetype = lty, color = col)

  return(geom)
}
.geom_prior.discrete         <- function(plot_data, ...){

  dots      <- list(...)
  col       <- if(!is.null(dots[["col"]]))   dots[["col"]]   else .plot.prior_settings()[["col"]]
  width     <- if(!is.null(dots[["width"]])) dots[["width"]] else .plot.prior_settings()[["width"]]
  lwd       <- if(!is.null(dots[["lwd"]]))   dots[["lwd"]]   else .plot.prior_settings()[["lwd"]]
  lty       <- if(!is.null(dots[["lty"]]))   dots[["lty"]]   else .plot.prior_settings()[["lty"]]

  geom <-   geom <- ggplot2::geom_bar(
    data    = data.frame(
      x = plot_data$x,
      y = plot_data$y),
    mapping = ggplot2::aes(
      x      = .data[["x"]],
      weight = .data[["y"]]),
    linewidth = lwd, linetype = lty, color = col, fill = col, width = width)

  return(geom)
}
.geom_prior.point            <- function(plot_data, scale_y2 = 1, ...){

  dots      <- list(...)
  col       <- if(!is.null(dots[["col"]]))      dots[["col"]]      else .plot.prior_settings()[["col"]]
  lwd       <- if(!is.null(dots[["size"]]))     dots[["size"]]     else  if(!is.null(dots[["lwd"]])) dots[["lwd"]] else .plot.prior_settings()[["lwd"]]
  lty       <- if(!is.null(dots[["linetype"]])) dots[["linetype"]] else  if(!is.null(dots[["lty"]])) dots[["lty"]] else .plot.prior_settings()[["lty"]]

  if(!all(plot_data$y == 0)){
    geom <- ggplot2::geom_segment(
      data    = data.frame(
        x    = unique(plot_data$x[plot_data$y != 0]),
        xend = unique(plot_data$x[plot_data$y != 0]),
        y    = 0,
        yend = plot_data$y[plot_data$y != 0] * scale_y2),
      mapping = ggplot2::aes(
        x    = .data[["x"]],
        xend = .data[["xend"]],
        y    = .data[["y"]],
        yend = .data[["yend"]]),
      arrow     = ggplot2::arrow(length = ggplot2::unit(0.5, "cm")),
      linewidth = 2*lwd, linetype = lty, color = col)
  }else{
    geom <- NULL
  }

  return(geom)
}
.geom_prior.weightfunction   <- function(plot_data, rescale_x, ...){

  dots      <- list(...)
  col       <- if(!is.null(dots[["col"]]))      dots[["col"]]      else .plot.prior_settings()[["col"]]
  col.fill  <- if(!is.null(dots[["col.fill"]])) dots[["col.fill"]] else .plot.prior_settings()[["col.fill"]]
  lwd       <- if(!is.null(dots[["size"]]))     dots[["size"]]     else  if(!is.null(dots[["lwd"]])) dots[["lwd"]] else .plot.prior_settings()[["lwd"]]
  lty       <- if(!is.null(dots[["linetype"]])) dots[["linetype"]] else  if(!is.null(dots[["lty"]])) dots[["lty"]] else .plot.prior_settings()[["lty"]]

  # weightfunction specific stuff
  x_cuts <- plot_data$x
  x_mean <- plot_data$y
  x_lCI  <- plot_data$y_lCI
  x_uCI  <- plot_data$y_uCI

  if(rescale_x){
    x_at <- seq(0, 1, length.out = length(unique(plot_data$x)))
    x_at <- x_at[c(1, sort(rep(2:(length(x_at)-1), 2)), length(x_at))]
  }else{
    x_at <- x_cuts
  }


  geom <- list(
    ggplot2::geom_polygon(
      data    = data.frame(
        x = c(x_at,  rev(x_at)),
        y = c(x_lCI, rev(x_uCI))),
      mapping = ggplot2::aes(
        x = .data[["x"]],
        y = .data[["y"]]),
      fill    = col.fill
    ),
    ggplot2::geom_line(
      data    = data.frame(
        x = x_at,
        y = x_mean),
      mapping = ggplot2::aes(
        x = .data[["x"]],
        y = .data[["y"]]),
      linewidth = lwd, linetype = lty, color = col)
  )

  return(geom)
}
.geom_prior.PETPEESE         <- function(plot_data, ...){

  dots      <- list(...)
  col       <- if(!is.null(dots[["col"]]))      dots[["col"]]      else .plot.prior_settings()[["col"]]
  col.fill  <- if(!is.null(dots[["col.fill"]])) dots[["col.fill"]] else .plot.prior_settings()[["col.fill"]]
  lwd       <- if(!is.null(dots[["size"]]))     dots[["size"]]     else  if(!is.null(dots[["lwd"]])) dots[["lwd"]] else .plot.prior_settings()[["lwd"]]
  lty       <- if(!is.null(dots[["linetype"]])) dots[["linetype"]] else  if(!is.null(dots[["lty"]])) dots[["lty"]] else .plot.prior_settings()[["lty"]]


  geom <-  list(
    ggplot2::geom_polygon(
      data    = data.frame(
        x = c(plot_data$x,  rev(plot_data$x)),
        y = c(plot_data$y_lCI, rev(plot_data$y_uCI))),
      mapping = ggplot2::aes(
        x = .data[["x"]],
        y = .data[["y"]]),
      fill    = col.fill
    ),
    ggplot2::geom_line(
      data    = data.frame(
        x = plot_data$x,
        y = plot_data$y),
      mapping = ggplot2::aes(
        x = .data[["x"]],
        y = .data[["y"]]),
      linewidth = lwd, linetype = lty, color = col)
  )

  return(geom)
}
.geom_prior.orthonormal_or_meandif <- function(plot_data, ...){

  dots      <- list(...)
  col       <- if(!is.null(dots[["col"]]))      dots[["col"]]      else .plot.prior_settings()[["col"]]
  lwd       <- if(!is.null(dots[["size"]]))     dots[["size"]]     else  if(!is.null(dots[["lwd"]])) dots[["lwd"]] else .plot.prior_settings()[["lwd"]]
  lty       <- if(!is.null(dots[["linetype"]])) dots[["linetype"]] else  if(!is.null(dots[["lty"]])) dots[["lty"]] else .plot.prior_settings()[["lty"]]

  geom <- ggplot2::geom_line(
    data    = data.frame(
      x = plot_data$x,
      y = plot_data$y),
    mapping = ggplot2::aes(
      x = .data[["x"]],
      y = .data[["y"]]),
    linewidth = lwd, linetype = lty, color = col)

  return(geom)
}
.geom_prior.factors          <- function(plot_data, ...){

  # this function notably differs from the .line_prior.factor counterpart
  # - it's so much more difficult to draw custom legend in ggplot2 ... :(

  dots <- list(...)
  col  <- if(!is.null(dots[["col"]]))      dots[["col"]]      else rep(.plot.prior_settings()[["col"]], length(dots[["level_names"]]))
  lty  <- if(!is.null(dots[["linetype"]])) dots[["linetype"]]
  else  if(!is.null(dots[["lty"]]))        dots[["lty"]]      else rep(.plot.prior_settings()[["lty"]], length(dots[["level_names"]]))
  lwd  <- if(!is.null(dots[["size"]]))     dots[["size"]]
  else  if(!is.null(dots[["lwd"]]))        dots[["lwd"]]      else .plot.prior_settings()[["lwd"]]
  legend_title <- if(!is.null(dots[["legend_title"]])) dots[["legend_title"]] else NULL

  names(col) <- dots[["level_names"]]
  names(lty) <- dots[["level_names"]]

  if(!is.null(dots[["hardcode"]]) && dots[["hardcode"]]){
    geom <- lapply(unique(plot_data$level), function(lvl){
      ggplot2::geom_line(
        data    = plot_data[plot_data$level == lvl,],
        mapping = ggplot2::aes(
          x        = .data[["x"]],
          y        = .data[["y"]]),
        linewidth = 1, show.legend = dots[["legend"]], color = col[lvl], linetype = lty[lvl])
    })
  }else{
    geom <- list(
      ggplot2::geom_line(
        data    = plot_data,
        mapping = ggplot2::aes(
          x        = .data[["x"]],
          y        = .data[["y"]],
          color    = .data[["level"]],
          linetype = .data[["level"]],
          group    = .data[["level"]]),
        linewidth = 1, show.legend = dots[["legend"]]),
      ggplot2::scale_linetype_manual(
        name   = legend_title,
        values = lty,
        breaks = dots[["level_names"]],
        labels = dots[["level_names"]]),
      ggplot2::scale_color_manual(
        name   = legend_title,
        values = col,
        breaks = dots[["level_names"]],
        labels = dots[["level_names"]]))
  }


  return(geom)
}
.geom_prior.spike_and_slab   <- function(plot_data, ...){

  geom <- list(
    .geom_prior.simple(plot_data[["variable"]], ...),
    .geom_prior.point(plot_data[["inclusion"]], ...)
  )

  return(geom)
}

.prior_linear_density_default_grid <- function(){
  4096L
}

.prior_linear_density_tail_prob <- function(){
  1e-4
}

.prior_linear_density_product_grid <- function(){
  1024L
}

.prior_linear_density_zero_tol <- function(){
  sqrt(.Machine$double.eps)
}
.prior_linear_source_transform <- function(source_transform){

  if(is.null(source_transform) || length(source_transform) == 0 || is.na(source_transform)){
    return(NULL)
  }

  unname(source_transform)
}

.prior_linear_density_empty_points <- function(){
  data.frame(x = numeric(), p = numeric())
}

.prior_linear_density_point <- function(x, p = 1){
  out <- list(
    density = NULL,
    points  = data.frame(x = x, p = p),
    n_grid  = 1L
  )
  class(out) <- c("prior_linear_density", "prior_density")
  return(.prior_linear_density_normalize(out))
}

.prior_linear_density_normalize <- function(dist){

  point_mass <- if(!is.null(dist$points) && nrow(dist$points) > 0) sum(dist$points$p) else 0
  density_mass <- if(!is.null(dist$density)) dist$density$mass else 0
  total_mass <- point_mass + density_mass

  if(!is.finite(total_mass) || total_mass <= 0){
    stop("The computed prior density has zero total mass.", call. = FALSE)
  }

  if(!is.null(dist$density)){
    dist$density$mass <- dist$density$mass / total_mass
  }
  if(!is.null(dist$points) && nrow(dist$points) > 0){
    dist$points$p <- dist$points$p / total_mass
  }

  class(dist) <- unique(c(class(dist), "prior_linear_density", "prior_density"))
  return(dist)
}

.prior_linear_density_fft_convolve <- function(a, b){

  n <- length(a) + length(b) - 1L
  n_fft <- 2^ceiling(log2(n))
  out <- fft(
    fft(c(a, rep(0, n_fft - length(a)))) *
      fft(c(b, rep(0, n_fft - length(b)))),
    inverse = TRUE
  ) / n_fft

  pmax(Re(out[seq_len(n)]), 0)
}

.prior_linear_density_aggregate_points <- function(points, dx){

  if(is.null(points) || nrow(points) == 0){
    return(.prior_linear_density_empty_points())
  }

  points <- points[is.finite(points$x) & is.finite(points$p) & points$p > 0, , drop = FALSE]
  if(nrow(points) == 0){
    return(.prior_linear_density_empty_points())
  }

  if(is.finite(dx) && dx > 0){
    key <- as.character(round(points$x / dx))
  }else{
    key <- format(signif(points$x, 14), scientific = FALSE)
  }

  split_points <- split(points, key)
  out <- do.call(rbind, lapply(split_points, function(p){
    data.frame(
      x = stats::weighted.mean(p$x, p$p),
      p = sum(p$p)
    )
  }))
  rownames(out) <- NULL
  out <- out[order(out$x), , drop = FALSE]

  return(out)
}

.prior_linear_density_coalesce <- function(densities = list(), points = NULL, dx, n_grid = NULL){

  points <- .prior_linear_density_aggregate_points(points, dx)

  densities <- densities[!vapply(densities, is.null, logical(1))]
  densities <- densities[vapply(densities, function(d){
    !is.null(d$x) && !is.null(d$y) && is.finite(d$mass) && d$mass > 0
  }, logical(1))]

  density <- NULL
  if(length(densities) > 0){
    x_min <- min(vapply(densities, function(d) min(d$x), numeric(1)))
    x_max <- max(vapply(densities, function(d) max(d$x), numeric(1)))

    if(!is.finite(dx) || dx <= 0){
      if(is.null(n_grid)){
        n_grid <- .prior_linear_density_default_grid()
      }
      dx <- (x_max - x_min) / max(1, n_grid - 1)
    }

    if(!is.finite(dx) || dx <= 0 || isTRUE(all.equal(x_min, x_max))){
      x <- x_min
      y_mass <- sum(vapply(densities, function(d) d$mass, numeric(1)))
    }else{
      x <- seq(x_min, x_max, by = dx)
      if(length(x) < 2){
        x <- seq(x_min, x_max, length.out = 2)
      }
      y_mass <- numeric(length(x))
      for(d in densities){
        y_mass <- y_mass + d$mass * stats::approx(d$x, d$y, xout = x, yleft = 0, yright = 0)$y
      }
    }

    density_mass <- sum(vapply(densities, function(d) d$mass, numeric(1)))
    area <- if(length(x) > 1) sum(y_mass) * (x[2] - x[1]) else density_mass
    if(is.finite(area) && area > 0 && density_mass > 0){
      y <- y_mass / area
    }else{
      y <- y_mass
    }

    density <- list(
      x    = x,
      y    = y,
      mass = density_mass
    )
  }

  out <- list(
    density = density,
    points  = points,
    n_grid  = if(is.null(n_grid)) length(if(!is.null(density)) density$x else points$x) else n_grid
  )
  class(out) <- c("prior_linear_density", "prior_density")
  return(.prior_linear_density_normalize(out))
}

.prior_linear_density_convolve <- function(lhs, rhs, dx){

  densities <- list()
  points <- .prior_linear_density_empty_points()

  if(!is.null(lhs$points) && nrow(lhs$points) > 0 && !is.null(rhs$points) && nrow(rhs$points) > 0){
    point_grid <- merge(lhs$points, rhs$points, by = NULL)
    points <- rbind(points, data.frame(
      x = point_grid$x.x + point_grid$x.y,
      p = point_grid$p.x * point_grid$p.y
    ))
  }

  if(!is.null(lhs$density) && !is.null(rhs$density)){
    y <- .prior_linear_density_fft_convolve(lhs$density$y, rhs$density$y) * dx
    area <- sum(y) * dx
    if(is.finite(area) && area > 0){
      y <- y / area
    }
    x <- lhs$density$x[1] + rhs$density$x[1] + dx * (seq_along(y) - 1)
    densities[[length(densities) + 1L]] <- list(
      x    = x,
      y    = y,
      mass = lhs$density$mass * rhs$density$mass
    )
  }

  if(!is.null(lhs$density) && !is.null(rhs$points) && nrow(rhs$points) > 0){
    for(i in seq_len(nrow(rhs$points))){
      densities[[length(densities) + 1L]] <- list(
        x    = lhs$density$x + rhs$points$x[i],
        y    = lhs$density$y,
        mass = lhs$density$mass * rhs$points$p[i]
      )
    }
  }

  if(!is.null(rhs$density) && !is.null(lhs$points) && nrow(lhs$points) > 0){
    for(i in seq_len(nrow(lhs$points))){
      densities[[length(densities) + 1L]] <- list(
        x    = rhs$density$x + lhs$points$x[i],
        y    = rhs$density$y,
        mass = rhs$density$mass * lhs$points$p[i]
      )
    }
  }

  .prior_linear_density_coalesce(densities = densities, points = points, dx = dx,
                                 n_grid = max(lhs$n_grid, rhs$n_grid))
}

.prior_linear_density_mix <- function(dists, weights, dx, n_grid = NULL){

  weights <- weights / sum(weights)
  densities <- list()
  points <- .prior_linear_density_empty_points()

  for(i in seq_along(dists)){
    dist <- dists[[i]]
    w <- weights[i]

    if(!is.null(dist$density)){
      densities[[length(densities) + 1L]] <- list(
        x    = dist$density$x,
        y    = dist$density$y,
        mass = dist$density$mass * w
      )
    }
    if(!is.null(dist$points) && nrow(dist$points) > 0){
      temp_points <- dist$points
      temp_points$p <- temp_points$p * w
      points <- rbind(points, temp_points)
    }
  }

  .prior_linear_density_coalesce(densities = densities, points = points, dx = dx, n_grid = n_grid)
}

.prior_linear_density_range <- function(dist){

  values <- numeric()
  if(!is.null(dist$density) && length(dist$density$x) > 0){
    values <- c(values, range(dist$density$x, finite = TRUE))
  }
  if(!is.null(dist$points) && nrow(dist$points) > 0){
    values <- c(values, dist$points$x[dist$points$p > .prior_linear_density_zero_tol()])
  }
  values <- values[is.finite(values)]

  if(length(values) == 0){
    return(c(0, 0))
  }
  range(values)
}

.prior_linear_density_dx <- function(dist){

  if(!is.null(dist$density) && length(dist$density$x) > 1){
    dx <- median(diff(dist$density$x))
    if(is.finite(dx) && dx > 0){
      return(dx)
    }
  }
  NA_real_
}

.prior_linear_density_regrid <- function(dist, dx, n_grid = NULL){

  densities <- list()
  if(!is.null(dist$density)){
    densities[[1]] <- dist$density
  }

  .prior_linear_density_coalesce(
    densities = densities,
    points    = dist$points,
    dx        = dx,
    n_grid    = if(is.null(n_grid)) dist$n_grid else n_grid
  )
}

.prior_linear_density_sum_independent <- function(dists, n_grid = NULL){

  dists <- dists[!vapply(dists, is.null, logical(1))]
  if(length(dists) == 0){
    return(.prior_linear_density_point(0))
  }
  if(length(dists) == 1){
    return(dists[[1]])
  }

  ranges <- do.call(rbind, lapply(dists, .prior_linear_density_range))
  target_width <- sum(pmax(0, ranges[, 2] - ranges[, 1]))
  if(is.null(n_grid)){
    n_grid <- max(vapply(dists, function(dist) dist$n_grid, integer(1)))
  }
  dx <- target_width / max(1, n_grid - 1)
  if(!is.finite(dx) || dx <= 0){
    dx_values <- vapply(dists, .prior_linear_density_dx, numeric(1))
    dx_values <- dx_values[is.finite(dx_values) & dx_values > 0]
    dx <- if(length(dx_values) > 0) min(dx_values) else 1
  }
  if(!is.finite(dx) || dx <= 0){
    dx <- 1
  }

  dist <- .prior_linear_density_point(0)
  for(component in dists){
    dist <- .prior_linear_density_convolve(
      .prior_linear_density_regrid(dist, dx, n_grid),
      .prior_linear_density_regrid(component, dx, n_grid),
      dx
    )
  }

  dist
}

.prior_linear_density_scaled <- function(dist, scale, mass = 1, dx = NA_real_, n_grid = NULL){

  if(abs(scale) <= .prior_linear_density_zero_tol()){
    out <- .prior_linear_density_point(0)
    out$points$p <- mass
    return(out)
  }

  densities <- list()
  if(!is.null(dist$density) && dist$density$mass > 0){
    x <- dist$density$x * scale
    y <- dist$density$y / abs(scale)
    ord <- order(x)
    densities[[1]] <- list(
      x    = x[ord],
      y    = y[ord],
      mass = dist$density$mass
    )
  }

  points <- .prior_linear_density_empty_points()
  if(!is.null(dist$points) && nrow(dist$points) > 0){
    points <- data.frame(
      x = dist$points$x * scale,
      p = dist$points$p
    )
  }

  out <- .prior_linear_density_coalesce(
    densities = densities,
    points    = points,
    dx        = dx,
    n_grid    = if(is.null(n_grid)) dist$n_grid else n_grid
  )
  if(!is.null(out$density)){
    out$density$mass <- out$density$mass * mass
  }
  if(!is.null(out$points) && nrow(out$points) > 0){
    out$points$p <- out$points$p * mass
  }
  out
}

.prior_linear_density_product_range <- function(lhs, rhs){

  lhs_range <- .prior_linear_density_range(lhs)
  rhs_range <- .prior_linear_density_range(rhs)
  products <- as.vector(outer(lhs_range, rhs_range, `*`))

  lhs_points <- if(!is.null(lhs$points) && nrow(lhs$points) > 0) lhs$points$x else numeric()
  rhs_points <- if(!is.null(rhs$points) && nrow(rhs$points) > 0) rhs$points$x else numeric()
  if(length(lhs_points) > 0){
    products <- c(products, as.vector(outer(lhs_points, rhs_range, `*`)))
  }
  if(length(rhs_points) > 0){
    products <- c(products, as.vector(outer(lhs_range, rhs_points, `*`)))
  }
  if(length(lhs_points) > 0 && length(rhs_points) > 0){
    products <- c(products, as.vector(outer(lhs_points, rhs_points, `*`)))
  }

  products <- products[is.finite(products)]
  if(length(products) == 0){
    return(c(0, 0))
  }
  range(products)
}

.prior_linear_density_product <- function(lhs, rhs, n_grid = NULL){

  if(is.null(n_grid)){
    n_grid <- max(lhs$n_grid, rhs$n_grid)
  }
  n_grid <- min(max(16L, n_grid), .prior_linear_density_product_grid())

  densities <- list()
  points <- .prior_linear_density_empty_points()
  dx <- NA_real_

  if(!is.null(lhs$points) && nrow(lhs$points) > 0 && !is.null(rhs$points) && nrow(rhs$points) > 0){
    point_grid <- merge(lhs$points, rhs$points, by = NULL)
    points <- rbind(points, data.frame(
      x = point_grid$x.x * point_grid$x.y,
      p = point_grid$p.x * point_grid$p.y
    ))
  }

  if(!is.null(lhs$density) && !is.null(rhs$points) && nrow(rhs$points) > 0){
    for(i in seq_len(nrow(rhs$points))){
      scaled <- .prior_linear_density_scaled(lhs, rhs$points$x[i],
                                             mass = rhs$points$p[i],
                                             n_grid = n_grid)
      if(!is.null(scaled$density)){
        densities[[length(densities) + 1L]] <- scaled$density
      }
      if(!is.null(scaled$points) && nrow(scaled$points) > 0){
        points <- rbind(points, scaled$points)
      }
    }
  }

  if(!is.null(rhs$density) && !is.null(lhs$points) && nrow(lhs$points) > 0){
    for(i in seq_len(nrow(lhs$points))){
      scaled <- .prior_linear_density_scaled(rhs, lhs$points$x[i],
                                             mass = lhs$points$p[i],
                                             n_grid = n_grid)
      if(!is.null(scaled$density)){
        densities[[length(densities) + 1L]] <- scaled$density
      }
      if(!is.null(scaled$points) && nrow(scaled$points) > 0){
        points <- rbind(points, scaled$points)
      }
    }
  }

  if(!is.null(lhs$density) && !is.null(rhs$density)){
    z_range <- .prior_linear_density_product_range(lhs, rhs)
    if(isTRUE(all.equal(z_range[1], z_range[2]))){
      points <- rbind(points, data.frame(
        x = z_range[1],
        p = lhs$density$mass * rhs$density$mass
      ))
    }else{
      z <- seq(z_range[1], z_range[2], length.out = n_grid)
      dx <- z[2] - z[1]

      x <- lhs$density$x
      fx <- lhs$density$y
      dx_x <- .prior_linear_density_dx(lhs)
      zero_tol <- .prior_linear_density_zero_tol() * max(1, max(abs(x), na.rm = TRUE))
      keep <- is.finite(x) & is.finite(fx) & fx > 0 & abs(x) > zero_tol
      x <- x[keep]
      fx <- fx[keep]

      y <- numeric(length(z))
      if(length(x) > 0 && is.finite(dx_x) && dx_x > 0){
        integrand_weight <- fx / abs(x)
        chunk_size <- max(1L, floor(5e6 / length(x)))
        for(start in seq(1L, length(z), by = chunk_size)){
          end <- min(length(z), start + chunk_size - 1L)
          source <- outer(1 / x, z[start:end], `*`)
          fy <- stats::approx(
            rhs$density$x,
            rhs$density$y,
            xout   = as.vector(source),
            yleft  = 0,
            yright = 0
          )$y
          fy <- matrix(fy, nrow = length(x))
          y[start:end] <- colSums(fy * integrand_weight) * dx_x
        }
      }

      y[!is.finite(y)] <- 0
      area <- sum(y) * dx
      if(is.finite(area) && area > 0){
        y <- y / area
        densities[[length(densities) + 1L]] <- list(
          x    = z,
          y    = y,
          mass = lhs$density$mass * rhs$density$mass
        )
      }
    }
  }

  if(is.na(dx)){
    dx_values <- vapply(densities, function(d){
      if(length(d$x) > 1) median(diff(d$x)) else NA_real_
    }, numeric(1))
    dx_values <- dx_values[is.finite(dx_values) & dx_values > 0]
    dx <- if(length(dx_values) > 0) min(dx_values) else NA_real_
  }
  if(!is.finite(dx) || dx <= 0){
    dx <- NA_real_
  }

  .prior_linear_density_coalesce(
    densities = densities,
    points    = points,
    dx        = dx,
    n_grid    = n_grid
  )
}

.prior_linear_is_factor_prior <- function(prior){
  is.prior.factor(prior) ||
    inherits(prior, "prior.factor_mixture") ||
    inherits(prior, "prior.factor_spike_and_slab")
}

.prior_linear_prior_dimension <- function(prior){

  if(.prior_linear_is_factor_prior(prior)){
    if(!is.null(attr(prior, "K")) && !is.na(attr(prior, "K"))){
      return(attr(prior, "K"))
    }
    if(is.prior.mixture(prior)){
      factor_components <- prior[vapply(prior, function(p) is.prior.factor(p) || inherits(p, "prior.factor_spike_and_slab"), logical(1))]
      if(length(factor_components) > 0){
        return(.get_prior_factor_levels(factor_components[[1]]))
      }
    }
    return(.get_prior_factor_levels(prior))
  }

  if(is.prior.vector(prior)){
    return(prior$parameters[["K"]])
  }

  1L
}

.prior_linear_prior_columns <- function(parameter, prior){

  K <- .prior_linear_prior_dimension(prior)
  if(length(K) != 1 || is.na(K) || K <= 1){
    return(parameter)
  }

  paste0(parameter, "[", seq_len(K), "]")
}

.prior_linear_representative_prior <- function(prior){

  if(is.prior(prior)){
    return(prior)
  }

  if(is.list(prior)){
    prior <- prior[vapply(prior, is.prior, logical(1))]
    if(length(prior) > 0){
      return(prior[[1]])
    }
  }

  NULL
}

.prior_linear_active_parameters <- function(prior_list, weights){

  active <- character()
  if(is.null(names(weights))){
    return(active)
  }

  weights <- weights[is.finite(weights)]
  weights <- weights[abs(weights) > .prior_linear_density_zero_tol()]
  if(length(weights) == 0){
    return(active)
  }

  for(parameter in names(prior_list)){
    prior <- .prior_linear_representative_prior(prior_list[[parameter]])
    if(is.null(prior)){
      next
    }
    columns <- .prior_linear_prior_columns(parameter, prior)
    present <- intersect(columns, names(weights))
    if(length(present) > 0){
      active <- c(active, parameter)
    }
  }

  active
}

.prior_linear_active_conditionals <- function(prior_list, weights, conditional){

  if(length(conditional) == 0){
    return(character())
  }

  active <- .prior_linear_active_parameters(prior_list, weights)
  conditional[conditional %in% active]
}

.prior_linear_weight_groups <- function(prior_list, weights){

  if(is.null(names(weights))){
    stop("'weights' must be a named numeric vector.", call. = FALSE)
  }

  groups <- list()
  matched <- rep(FALSE, length(weights))
  names(matched) <- names(weights)

  for(parameter in names(prior_list)){
    prior <- prior_list[[parameter]]
    if(is.null(prior)){
      next
    }
    if(!is.prior(prior)){
      stop("All entries of 'prior_list' must be prior objects.", call. = FALSE)
    }

    columns <- .prior_linear_prior_columns(parameter, prior)
    present <- intersect(columns, names(weights))
    present <- present[abs(weights[present]) > .prior_linear_density_zero_tol()]
    if(length(present) == 0){
      next
    }

    groups[[parameter]] <- list(
      parameter = parameter,
      prior     = prior,
      columns   = columns,
      weights   = weights[present],
      indices   = match(present, columns)
    )
    matched[present] <- TRUE
  }

  unmatched <- names(weights)[abs(weights) > .prior_linear_density_zero_tol() & !matched]
  if(length(unmatched) > 0){
    stop(
      "No prior distribution was found for coefficient column(s): '",
      paste0(unmatched, collapse = "', '"), "'.",
      call. = FALSE
    )
  }

  groups
}

.prior_linear_scalar_range <- function(prior, weight, tail_prob, source_transform = NULL){

  source_transform <- .prior_linear_source_transform(source_transform)

  if(abs(weight) <= .prior_linear_density_zero_tol() || is.prior.none(prior)){
    return(c(0, 0))
  }

  if(is.prior.point(prior)){
    location <- prior$parameters[["location"]]
    if(identical(source_transform, "log")){
      if(location <= 0){
        stop("A log-transformed prior source must have positive support.", call. = FALSE)
      }
      location <- log(location)
    }
    return(rep(weight * location, 2))
  }

  if(is.prior.discrete(prior)){
    support <- switch(
      prior[["distribution"]],
      "bernoulli" = c(0, 1),
      stop("Unsupported discrete prior distribution for linear-combination densities.", call. = FALSE)
    )
    if(identical(source_transform, "log")){
      if(any(support <= 0)){
        stop("A log-transformed prior source must have positive support.", call. = FALSE)
      }
      support <- log(support)
    }
    return(range(weight * support))
  }

  if(identical(source_transform, "log")){
    prior_range <- quant(prior, c(tail_prob, 1 - tail_prob))
    if(any(!is.finite(prior_range)) || any(prior_range <= 0)){
      prior_range <- range(prior, quantiles = tail_prob)
      prior_range[prior_range <= 0] <- min(prior_range[prior_range > 0], na.rm = TRUE)
    }
    if(any(!is.finite(prior_range)) || any(prior_range <= 0)){
      stop("A log-transformed prior source must have positive support.", call. = FALSE)
    }
    prior_range <- log(prior_range)
  }else{
    prior_range <- range(prior, quantiles = tail_prob)
  }

  range(weight * prior_range)
}

.prior_linear_vector_scalar_prior <- function(prior, weights){

  weights <- as.numeric(weights)
  norm_weight <- sqrt(sum(weights^2))

  if(norm_weight <= .prior_linear_density_zero_tol()){
    return(prior("point", list(location = 0)))
  }

  if(is.prior.point(prior)){
    location <- prior$parameters[["location"]]
    return(prior("point", list(location = sum(weights) * location)))
  }

  switch(
    prior[["distribution"]],
    "mnormal" = prior(
      "normal",
      list(
        mean = sum(weights) * prior$parameters[["mean"]],
        sd   = norm_weight * prior$parameters[["sd"]]
      )
    ),
    "mt" = prior(
      "t",
      list(
        location = sum(weights) * prior$parameters[["location"]],
        scale    = norm_weight * prior$parameters[["scale"]],
        df       = prior$parameters[["df"]]
      )
    ),
    "mpoint" = prior(
      "point",
      list(location = sum(weights) * prior$parameters[["location"]])
    ),
    stop("Unsupported vector prior distribution for linear-combination densities.", call. = FALSE)
  )
}

.prior_linear_group_range <- function(group, tail_prob, source_transforms = NULL){

  prior <- group$prior
  weights <- group$weights
  source_transforms <- source_transforms[names(weights)]

  if(is.prior.none(prior)){
    return(c(0, 0))
  }

  if(is.prior.spike_and_slab(prior)){
    variable_prior <- .get_spike_and_slab_variable(prior)
    variable_group <- group
    variable_group$prior <- variable_prior
    variable_range <- .prior_linear_group_range(variable_group, tail_prob, source_transforms)
    return(range(c(0, variable_range)))
  }

  if(is.prior.mixture(prior)){
    ranges <- do.call(rbind, lapply(prior, function(component){
      component_group <- group
      component_group$prior <- component
      .prior_linear_group_range(component_group, tail_prob, source_transforms)
    }))
    return(range(ranges))
  }

  if(is.prior.vector(prior) && !is.prior.treatment(prior) && !is.prior.independent(prior)){
    if(any(!is.na(source_transforms))){
      stop("Source transformations are only supported for scalar prior terms.", call. = FALSE)
    }
    scalar_prior <- .prior_linear_vector_scalar_prior(prior, weights)
    return(.prior_linear_scalar_range(scalar_prior, 1, tail_prob))
  }

  ranges <- do.call(rbind, Map(function(weight, source_transform){
    .prior_linear_scalar_range(prior, weight, tail_prob, source_transform)
  }, weights, source_transforms))

  c(sum(ranges[, 1]), sum(ranges[, 2]))
}

.prior_linear_scalar_distribution <- function(prior, weight, dx, tail_prob, source_transform = NULL, n_grid = NULL){

  source_transform <- .prior_linear_source_transform(source_transform)

  if(abs(weight) <= .prior_linear_density_zero_tol() || is.prior.none(prior)){
    return(.prior_linear_density_point(0))
  }

  if(is.prior.point(prior)){
    location <- prior$parameters[["location"]]
    if(identical(source_transform, "log")){
      if(location <= 0){
        stop("A log-transformed prior source must have positive support.", call. = FALSE)
      }
      location <- log(location)
    }
    return(.prior_linear_density_point(weight * location))
  }

  if(is.prior.discrete(prior)){
    support <- switch(
      prior[["distribution"]],
      "bernoulli" = c(0, 1),
      stop("Unsupported discrete prior distribution for linear-combination densities.", call. = FALSE)
    )
    probs <- mpdf(prior, support)
    if(identical(source_transform, "log")){
      if(any(support <= 0)){
        stop("A log-transformed prior source must have positive support.", call. = FALSE)
      }
      support <- log(support)
    }
    points <- data.frame(x = weight * support, p = probs / sum(probs))
    return(.prior_linear_density_coalesce(points = points, dx = dx, n_grid = n_grid))
  }

  x_range <- .prior_linear_scalar_range(prior, weight, tail_prob, source_transform)
  if(isTRUE(all.equal(x_range[1], x_range[2]))){
    return(.prior_linear_density_point(x_range[1]))
  }

  x <- seq(x_range[1], x_range[2], by = dx)
  if(length(x) < 3){
    x <- seq(x_range[1], x_range[2], length.out = 3)
  }

  source_x <- x / weight
  if(identical(source_transform, "log")){
    original_x <- exp(source_x)
    y <- mpdf(prior, original_x) * original_x / abs(weight)
  }else{
    y <- mpdf(prior, source_x) / abs(weight)
  }

  y[!is.finite(y)] <- 0
  area <- sum(y) * (x[2] - x[1])
  if(!is.finite(area) || area <= 0){
    stop("A continuous prior component evaluated to zero mass on its grid.", call. = FALSE)
  }
  y <- y / area

  .prior_linear_density_coalesce(
    densities = list(list(x = x, y = y, mass = 1)),
    dx = dx,
    n_grid = n_grid
  )
}

.prior_linear_group_distribution <- function(group, dx, tail_prob, source_transforms = NULL, n_grid = NULL){

  prior <- group$prior
  weights <- group$weights
  source_transforms <- source_transforms[names(weights)]

  if(is.prior.none(prior)){
    return(.prior_linear_density_point(0))
  }

  if(is.prior.spike_and_slab(prior)){
    variable_prior  <- .get_spike_and_slab_variable(prior)
    inclusion_prior <- .get_spike_and_slab_inclusion(prior)
    inclusion <- mean(inclusion_prior)
    inclusion <- min(max(inclusion, 0), 1)

    variable_group <- group
    variable_group$prior <- variable_prior

    return(.prior_linear_density_mix(
      dists = list(
        .prior_linear_group_distribution(variable_group, dx, tail_prob, source_transforms, n_grid),
        .prior_linear_density_point(0)
      ),
      weights = c(inclusion, 1 - inclusion),
      dx = dx,
      n_grid = n_grid
    ))
  }

  if(is.prior.mixture(prior)){
    prior_weights <- attr(prior, "prior_weights")
    prior_weights <- prior_weights / sum(prior_weights)
    dists <- lapply(prior, function(component){
      component_group <- group
      component_group$prior <- component
      .prior_linear_group_distribution(component_group, dx, tail_prob, source_transforms, n_grid)
    })
    return(.prior_linear_density_mix(dists, prior_weights, dx = dx, n_grid = n_grid))
  }

  if(is.prior.vector(prior) && !is.prior.treatment(prior) && !is.prior.independent(prior)){
    if(any(!is.na(source_transforms))){
      stop("Source transformations are only supported for scalar prior terms.", call. = FALSE)
    }
    scalar_prior <- .prior_linear_vector_scalar_prior(prior, weights)
    return(.prior_linear_scalar_distribution(scalar_prior, 1, dx, tail_prob, n_grid = n_grid))
  }

  dist <- .prior_linear_density_point(0)
  for(i in seq_along(weights)){
    source_dist <- .prior_linear_scalar_distribution(
      prior             = prior,
      weight            = weights[i],
      dx                = dx,
      tail_prob         = tail_prob,
      source_transform  = source_transforms[i],
      n_grid            = n_grid
    )
    dist <- .prior_linear_density_convolve(dist, source_dist, dx)
  }

  dist
}

.prior_linear_density_transform <- function(dist, transformation, transformation_arguments = NULL, n_grid = NULL){

  if(is.null(transformation)){
    return(dist)
  }

  densities <- list()
  if(!is.null(dist$density) && dist$density$mass > 0){
    x_old <- dist$density$x
    y_old <- dist$density$y
    x_new <- .density.prior_transformation_x(x_old, transformation, transformation_arguments)
    y_new <- .density.prior_transformation_y(x_new, y_old, transformation, transformation_arguments)

    keep <- is.finite(x_new) & is.finite(y_new)
    x_new <- x_new[keep]
    y_new <- y_new[keep]

    if(length(x_new) >= 2){
      ord <- order(x_new)
      x_new <- x_new[ord]
      y_new <- y_new[ord]

      keep <- c(TRUE, diff(x_new) > .prior_linear_density_zero_tol() * pmax(1, abs(x_new[-length(x_new)])))
      x_new <- x_new[keep]
      y_new <- y_new[keep]

      area <- if(length(x_new) > 1){
        sum(diff(x_new) * (head(y_new, -1) + tail(y_new, -1)) / 2)
      }else{
        0
      }
      if(is.finite(area) && area > 0){
        y_new <- y_new / area
      }
      densities[[1]] <- list(x = x_new, y = y_new, mass = dist$density$mass)
    }
  }

  points <- dist$points
  if(!is.null(points) && nrow(points) > 0){
    points$x <- .density.prior_transformation_x(points$x, transformation, transformation_arguments)
    points <- points[is.finite(points$x), , drop = FALSE]
  }

  if(length(densities) == 0){
    return(.prior_linear_density_coalesce(
      points = points,
      dx     = NA_real_,
      n_grid = if(is.null(n_grid)) dist$n_grid else n_grid
    ))
  }

  out <- list(
    density = densities[[1]],
    points  = .prior_linear_density_aggregate_points(points, NA_real_),
    n_grid  = if(is.null(n_grid)) dist$n_grid else n_grid
  )
  class(out) <- c("prior_linear_density", "prior_density")
  .prior_linear_density_normalize(out)
}

.prior_linear_split_multiply_groups <- function(prior_list, weights){

  additive_weights <- weights
  product_groups <- list()

  for(parameter in names(prior_list)){
    prior <- prior_list[[parameter]]
    if(is.null(prior)){
      next
    }
    if(!is.prior(prior)){
      stop("All entries of 'prior_list' must be prior objects.", call. = FALSE)
    }

    columns <- .prior_linear_prior_columns(parameter, prior)
    present <- intersect(columns, names(additive_weights))
    present <- present[abs(additive_weights[present]) > .prior_linear_density_zero_tol()]
    if(length(present) == 0){
      next
    }

    multiply_by <- attr(prior, "multiply_by")
    if(is.null(multiply_by)){
      next
    }

    if(is.numeric(multiply_by)){
      if(length(multiply_by) != 1L){
        stop("Numeric 'multiply_by' values must be scalar for deterministic prior densities.", call. = FALSE)
      }
      additive_weights[present] <- additive_weights[present] * multiply_by
      next
    }

    if(!is.character(multiply_by) || length(multiply_by) != 1L){
      stop("'multiply_by' must be either a scalar numeric value or a scalar parameter name.", call. = FALSE)
    }

    if(is.null(product_groups[[multiply_by]])){
      product_groups[[multiply_by]] <- list(
        multiplier = multiply_by,
        prior_list = list(),
        weights    = numeric()
      )
    }

    product_groups[[multiply_by]]$prior_list[[parameter]] <- prior
    product_groups[[multiply_by]]$weights <- c(
      product_groups[[multiply_by]]$weights,
      additive_weights[present]
    )
    additive_weights[present] <- 0
  }

  list(
    additive_weights = additive_weights,
    product_groups   = product_groups
  )
}

.prior_linear_additive_combination_density <- function(prior_list, weights,
                                                       n_grid = .prior_linear_density_default_grid(),
                                                       tail_prob = .prior_linear_density_tail_prob(),
                                                       source_transforms = NULL){

  weights <- weights[is.finite(weights)]
  weights <- weights[abs(weights) > .prior_linear_density_zero_tol()]

  if(length(weights) == 0){
    return(.prior_linear_density_point(0))
  }

  groups <- .prior_linear_weight_groups(prior_list, weights)
  if(length(groups) == 0){
    return(.prior_linear_density_point(0))
  }

  if(is.null(source_transforms)){
    source_transforms <- rep(NA_character_, length(weights))
    names(source_transforms) <- names(weights)
  }else{
    source_transforms <- source_transforms[names(weights)]
    source_transforms[is.na(source_transforms)] <- NA_character_
  }

  group_ranges <- do.call(rbind, lapply(groups, .prior_linear_group_range,
                                        tail_prob = tail_prob,
                                        source_transforms = source_transforms))
  target_range <- c(sum(group_ranges[, 1]), sum(group_ranges[, 2]))
  target_width <- diff(target_range)
  dx <- target_width / max(1, n_grid - 1)
  if(!is.finite(dx) || dx <= 0){
    dx <- 1
  }

  dist <- .prior_linear_density_point(0)
  for(group in groups){
    group_dist <- .prior_linear_group_distribution(
      group             = group,
      dx                = dx,
      tail_prob         = tail_prob,
      source_transforms = source_transforms,
      n_grid            = n_grid
    )
    dist <- .prior_linear_density_convolve(dist, group_dist, dx)
  }

  attr(dist, "weights") <- weights
  return(dist)
}

.prior_linear_combination_density <- function(prior_list, weights,
                                              n_grid = .prior_linear_density_default_grid(),
                                              tail_prob = .prior_linear_density_tail_prob(),
                                              source_transforms = NULL,
                                              output_transformation = NULL,
                                              output_transformation_arguments = NULL){

  check_list(prior_list, "prior_list")
  if(is.null(weights) || length(weights) == 0){
    weights <- numeric()
  }else{
    check_real(weights, "weights", check_length = 0)
  }
  check_int(n_grid, "n_grid", lower = 16)
  check_real(tail_prob, "tail_prob", lower = 0, upper = 0.5, allow_bound = FALSE)

  weights <- weights[is.finite(weights)]
  weights <- weights[abs(weights) > .prior_linear_density_zero_tol()]

  if(length(weights) == 0){
    dist <- .prior_linear_density_point(0)
    return(.prior_linear_density_transform(dist, output_transformation, output_transformation_arguments, n_grid))
  }

  if(is.null(source_transforms)){
    source_transforms <- rep(NA_character_, length(weights))
    names(source_transforms) <- names(weights)
  }else{
    source_transforms <- source_transforms[names(weights)]
    source_transforms[is.na(source_transforms)] <- NA_character_
  }

  split <- .prior_linear_split_multiply_groups(prior_list, weights)
  components <- list(
    .prior_linear_additive_combination_density(
      prior_list         = prior_list,
      weights            = split$additive_weights,
      n_grid             = n_grid,
      tail_prob          = tail_prob,
      source_transforms  = source_transforms
    )
  )

  if(length(split$product_groups) > 0){
    for(product_group in split$product_groups){
      multiplier <- product_group$multiplier
      if(!multiplier %in% names(prior_list)){
        stop("No prior distribution was found for 'multiply_by' parameter '", multiplier, "'.", call. = FALSE)
      }
      multiplier_prior <- prior_list[[multiplier]]
      if(is.null(multiplier_prior) || !is.prior(multiplier_prior)){
        stop("The 'multiply_by' parameter '", multiplier, "' does not have a supported prior distribution.", call. = FALSE)
      }
      if(!is.null(attr(multiplier_prior, "multiply_by"))){
        stop("Nested 'multiply_by' prior densities are not supported.", call. = FALSE)
      }
      if(.prior_linear_prior_dimension(multiplier_prior) != 1L){
        stop("The 'multiply_by' parameter '", multiplier, "' must have a scalar prior distribution.", call. = FALSE)
      }

      linear_dist <- .prior_linear_additive_combination_density(
        prior_list         = product_group$prior_list,
        weights            = product_group$weights,
        n_grid             = n_grid,
        tail_prob          = tail_prob,
        source_transforms  = source_transforms
      )

      multiplier_weights <- 1
      names(multiplier_weights) <- multiplier
      multiplier_dist <- .prior_linear_additive_combination_density(
        prior_list         = prior_list[multiplier],
        weights            = multiplier_weights,
        n_grid             = n_grid,
        tail_prob          = tail_prob,
        source_transforms  = source_transforms
      )

      components[[length(components) + 1L]] <- .prior_linear_density_product(
        linear_dist,
        multiplier_dist,
        n_grid = n_grid
      )
    }
  }

  dist <- .prior_linear_density_sum_independent(components, n_grid = n_grid)
  dist <- .prior_linear_density_transform(dist, output_transformation,
                                          output_transformation_arguments, n_grid)
  attr(dist, "weights") <- weights
  return(dist)
}

.prior_linear_density_to_plot_data <- function(x, n_points = 1000, x_range = NULL,
                                               transformation = NULL,
                                               transformation_arguments = NULL,
                                               transformation_settings = FALSE,
                                               factor = FALSE,
                                               level = NULL,
                                               level_name = NULL){

  if(!inherits(x, "prior_linear_density")){
    stop("'x' must be a prior linear density object.", call. = FALSE)
  }

  check_int(n_points, "n_points", lower = 2)
  check_real(x_range, "x_range", check_length = 2, allow_NULL = TRUE)
  .check_transformation_input(transformation, transformation_arguments, transformation_settings)

  dist <- x
  if(!is.null(transformation) && transformation_settings && !is.null(x_range)){
    x_range <- .density.prior_transformation_inv_x(x_range, transformation, transformation_arguments)
  }

  out <- list()

  if(!is.null(dist$density) && dist$density$mass > .prior_linear_density_zero_tol()){
    if(is.null(x_range)){
      x_den <- seq(min(dist$density$x), max(dist$density$x), length.out = n_points)
    }else{
      x_den <- seq(x_range[1], x_range[2], length.out = n_points)
    }
    y_den <- stats::approx(dist$density$x, dist$density$y, xout = x_den, yleft = 0, yright = 0)$y
    y_den <- y_den * dist$density$mass

    if(!is.null(transformation)){
      x_den <- .density.prior_transformation_x(x_den, transformation, transformation_arguments)
      y_den <- .density.prior_transformation_y(x_den, y_den, transformation, transformation_arguments)
    }

    out_den <- list(
      call    = call("density", "linear-combination prior"),
      bw      = NULL,
      n       = n_points,
      x       = x_den,
      y       = y_den,
      samples = NULL
    )
    class(out_den) <- c("density", "density.prior", "density.prior.simple",
                        if(factor) "density.prior.factor")
    attr(out_den, "x_range") <- range(x_den)
    attr(out_den, "y_range") <- c(0, max(y_den, 0, na.rm = TRUE))
    if(!is.null(level)) attr(out_den, "level") <- level
    if(!is.null(level_name)) attr(out_den, "level_name") <- level_name
    out[["density"]] <- out_den
  }

  points <- dist$points
  if(!is.null(points) && nrow(points) > 0){
    points <- points[points$p > .prior_linear_density_zero_tol(), , drop = FALSE]
    if(!is.null(x_range)){
      points <- points[points$x >= min(x_range) & points$x <= max(x_range), , drop = FALSE]
    }
    if(nrow(points) > 0 && !is.null(transformation)){
      points$x <- .density.prior_transformation_x(points$x, transformation, transformation_arguments)
    }

    for(i in seq_len(nrow(points))){
      out_point <- list(
        call    = call("density", paste0("point", i)),
        bw      = NULL,
        n       = n_points,
        x       = points$x[i],
        y       = points$p[i],
        samples = NULL
      )
      class(out_point) <- c("density", "density.prior", "density.prior.point",
                            if(factor) "density.prior.factor")
      attr(out_point, "x_range") <- range(points$x)
      attr(out_point, "y_range") <- c(0, max(points$p))
      if(!is.null(level)) attr(out_point, "level") <- level
      if(!is.null(level_name)) attr(out_point, "level_name") <- level_name
      out[[paste0("points", i)]] <- out_point
    }
  }

  return(out)
}

.prior_linear_density_height <- function(x, value){

  if(!inherits(x, "prior_linear_density")){
    stop("'x' must be a prior linear density object.", call. = FALSE)
  }

  height <- 0
  if(!is.null(x$density) && value >= min(x$density$x) && value <= max(x$density$x)){
    height <- height + stats::approx(
      x$density$x,
      x$density$y * x$density$mass,
      xout = value,
      yleft = 0,
      yright = 0
    )$y
  }

  height
}

.prior_linear_density_point_mass <- function(x, value){

  if(!inherits(x, "prior_linear_density") || is.null(x$points) || nrow(x$points) == 0){
    return(0)
  }

  sum(x$points$p[abs(x$points$x - value) <= .prior_linear_density_zero_tol()])
}

.prior_density_context <- function(prior_list, column_names, formula_scale = NULL,
                                   n_grid = .prior_linear_density_default_grid(),
                                   tail_prob = .prior_linear_density_tail_prob()){

  check_list(prior_list, "prior_list")
  check_char(column_names, "column_names", check_length = FALSE)
  check_list(formula_scale, "formula_scale", allow_NULL = TRUE)
  check_int(n_grid, "n_grid", lower = 16)
  check_real(tail_prob, "tail_prob", lower = 0, upper = 0.5, allow_bound = FALSE)

  if(!is.null(formula_scale) && length(formula_scale) > 0){
    .check_formula_scale_info(formula_scale)
  }

  transforms <- list()
  if(!is.null(formula_scale) && length(formula_scale) > 0){
    for(param_name in names(formula_scale)){
      affected_cols <- grep(paste0("^", param_name, "_"), column_names, value = TRUE)
      if(length(affected_cols) == 0){
        next
      }
      transforms[[param_name]] <- list(
        columns       = affected_cols,
        matrix        = .build_unscale_matrix(affected_cols, formula_scale[[param_name]], param_name),
        log_intercept = isTRUE(attr(formula_scale[[param_name]], "log_intercept")),
        intercept     = paste0(param_name, "_intercept")
      )
    }
  }

  out <- list(
    prior_list    = prior_list,
    column_names  = column_names,
    formula_scale = formula_scale,
    transforms    = transforms,
    n_grid        = n_grid,
    tail_prob     = tail_prob
  )
  class(out) <- "prior_density_context"
  return(out)
}

.prior_density_context_standardized_weights <- function(context, weights){

  if(!inherits(context, "prior_density_context")){
    stop("'context' must be a prior density context.", call. = FALSE)
  }
  if(is.null(names(weights))){
    stop("'weights' must be a named numeric vector.", call. = FALSE)
  }

  weights <- weights[intersect(names(weights), context$column_names)]
  out <- rep(0, length(context$column_names))
  names(out) <- context$column_names

  used <- rep(FALSE, length(weights))
  names(used) <- names(weights)

  for(transform in context$transforms){
    cols <- intersect(transform$columns, names(weights))
    if(length(cols) == 0){
      next
    }

    if(transform$log_intercept &&
       transform$intercept %in% cols &&
       abs(weights[[transform$intercept]]) > .prior_linear_density_zero_tol()){
      stop(
        "Linear-combination prior densities with log-intercept scaling are only available ",
        "for the transformed intercept coefficient itself.",
        call. = FALSE
      )
    }

    temp_weights <- rep(0, length(transform$columns))
    names(temp_weights) <- transform$columns
    temp_weights[cols] <- weights[cols]
    out[transform$columns] <- out[transform$columns] +
      as.numeric(temp_weights %*% transform$matrix)
    used[cols] <- TRUE
  }

  remaining <- names(weights)[!used]
  if(length(remaining) > 0){
    out[remaining] <- out[remaining] + weights[remaining]
  }

  out[abs(out) > .prior_linear_density_zero_tol()]
}

.prior_density_context_density <- function(context, weights,
                                           source_transforms = NULL,
                                           output_transformation = NULL,
                                           output_transformation_arguments = NULL){

  standardized_weights <- .prior_density_context_standardized_weights(context, weights)

  if(!is.null(source_transforms)){
    source_transforms <- source_transforms[names(standardized_weights)]
  }

  .prior_linear_combination_density(
    prior_list                       = context$prior_list,
    weights                          = standardized_weights,
    n_grid                           = context$n_grid,
    tail_prob                        = context$tail_prob,
    source_transforms                = source_transforms,
    output_transformation            = output_transformation,
    output_transformation_arguments  = output_transformation_arguments
  )
}

.prior_density_model_mixture_context <- function(prior_list, column_names,
                                                  n_grid = .prior_linear_density_default_grid(),
                                                  tail_prob = .prior_linear_density_tail_prob()){

  check_list(prior_list, "prior_list")
  check_char(column_names, "column_names", check_length = FALSE)

  prior_weights <- do.call(cbind, lapply(prior_list, function(parameter_priors){
    if(is.prior(parameter_priors)){
      return(.prior_model_weight(parameter_priors))
    }
    sapply(parameter_priors, .prior_model_weight)
  }))

  if(!all(prior_weights[, 1] == prior_weights)){
    stop("The model prior distributions are not aligned across parameters.", call. = FALSE)
  }

  model_weights <- prior_weights[, 1]
  model_weights <- model_weights / sum(model_weights)

  out <- list(
    prior_list   = prior_list,
    column_names = column_names,
    model_weights = model_weights,
    n_grid       = n_grid,
    tail_prob    = tail_prob
  )
  class(out) <- "prior_density_model_mixture_context"
  return(out)
}

.prior_density_condition_component <- function(prior){

  if(is.prior.spike_and_slab(prior)){
    components <- attr(prior, "components")
    if(!all(components %in% c("null", "alternative"))){
      stop("conditional mixture posterior distributions are available only for 'null' and 'alternative' components", call. = FALSE)
    }

    inclusion <- mean(.get_spike_and_slab_inclusion(prior))
    inclusion <- min(max(inclusion, 0), 1)
    probabilities <- ifelse(components == "alternative", inclusion, 1 - inclusion)

    return(lapply(seq_along(prior), function(i){
      list(
        prior       = prior[[i]],
        probability = probabilities[i],
        alternative = components[i] == "alternative"
      )
    }))
  }

  if(is.prior.mixture(prior)){
    components <- attr(prior, "components")
    if(!all(components %in% c("null", "alternative"))){
      stop("conditional mixture posterior distributions are available only for 'null' and 'alternative' components", call. = FALSE)
    }

    prior_weights <- attr(prior, "prior_weights")
    prior_weights <- prior_weights / sum(prior_weights)

    return(lapply(seq_along(prior), function(i){
      list(
        prior       = prior[[i]],
        probability = prior_weights[i],
        alternative = components[i] == "alternative"
      )
    }))
  }

  list(list(
    prior       = prior,
    probability = 1,
    alternative = TRUE
  ))
}

.prior_density_copy_parent_attributes <- function(component, parent){

  parent_attributes <- attributes(parent)
  skip <- c("class", "names", "components", "prior_weights", "inclusion_prior")

  for(attribute in setdiff(names(parent_attributes), skip)){
    if(is.null(attr(component, attribute, exact = TRUE))){
      attr(component, attribute) <- parent_attributes[[attribute]]
    }
  }

  component
}

.prior_density_condition_models <- function(prior_list, conditional, conditional_rule){

  conditional <- unique(conditional[conditional %in% names(prior_list)])
  if(length(conditional) == 0){
    return(NULL)
  }

  options <- lapply(conditional, function(parameter){
    .prior_density_condition_component(prior_list[[parameter]])
  })
  names(options) <- conditional

  option_grid <- expand.grid(lapply(options, seq_along))
  keep <- logical(nrow(option_grid))
  model_weights <- numeric(nrow(option_grid))

  for(i in seq_len(nrow(option_grid))){
    alternatives <- logical(length(conditional))
    probabilities <- numeric(length(conditional))

    for(j in seq_along(conditional)){
      option <- options[[j]][[option_grid[i, j]]]
      alternatives[j] <- option$alternative
      probabilities[j] <- option$probability
    }

    keep[i] <- if(conditional_rule == "AND") all(alternatives) else any(alternatives)
    model_weights[i] <- prod(probabilities)
  }

  option_grid <- option_grid[keep & model_weights > 0, , drop = FALSE]
  model_weights <- model_weights[keep & model_weights > 0]

  if(nrow(option_grid) == 0){
    return(list(prior_lists = list(), weights = numeric()))
  }

  prior_lists <- lapply(seq_len(nrow(option_grid)), function(i){
    model_prior_list <- prior_list
    for(j in seq_along(conditional)){
      parameter <- conditional[j]
      component <- options[[j]][[option_grid[i, j]]]$prior
      model_prior_list[[parameter]] <- .prior_density_copy_parent_attributes(
        component = component,
        parent    = prior_list[[parameter]]
      )
    }
    model_prior_list
  })

  list(
    prior_lists = prior_lists,
    weights     = model_weights / sum(model_weights)
  )
}

.prior_density_conditional_context <- function(prior_list, column_names, conditional,
                                               conditional_rule = "AND", formula_scale = NULL,
                                               n_grid = .prior_linear_density_default_grid(),
                                               tail_prob = .prior_linear_density_tail_prob()){

  condition_models <- .prior_density_condition_models(prior_list, conditional, conditional_rule)
  if(is.null(condition_models)){
    return(.prior_density_context(prior_list, column_names, formula_scale, n_grid, tail_prob))
  }

  out <- list(
    prior_list      = prior_list,
    column_names    = column_names,
    formula_scale   = formula_scale,
    prior_lists     = condition_models$prior_lists,
    model_weights   = condition_models$weights,
    n_grid          = n_grid,
    tail_prob       = tail_prob
  )
  class(out) <- "prior_density_conditional_context"
  return(out)
}

.prior_density_model_mixture_density <- function(context, weights,
                                                  output_transformation = NULL,
                                                  output_transformation_arguments = NULL){

  if(!inherits(context, "prior_density_model_mixture_context")){
    stop("'context' must be a prior density model-mixture context.", call. = FALSE)
  }

  dists <- vector("list", length(context$model_weights))
  for(model_i in seq_along(context$model_weights)){
    model_prior_list <- lapply(context$prior_list, function(parameter_priors){
      if(is.prior(parameter_priors)){
        return(parameter_priors)
      }
      parameter_priors[[model_i]]
    })
    names(model_prior_list) <- names(context$prior_list)

    for(parameter in names(model_prior_list)){
      if(is.null(model_prior_list[[parameter]])){
        model_prior_list[[parameter]] <- prior("point", list(location = 0))
      }
    }

    dists[[model_i]] <- .prior_linear_combination_density(
      prior_list = model_prior_list,
      weights    = weights,
      n_grid     = context$n_grid,
      tail_prob  = context$tail_prob
    )
  }

  dx <- min(vapply(dists, function(dist){
    if(!is.null(dist$density) && length(dist$density$x) > 1){
      return(dist$density$x[2] - dist$density$x[1])
    }
    Inf
  }, numeric(1)))
  if(!is.finite(dx)){
    dx <- NA_real_
  }

  dist <- .prior_linear_density_mix(
    dists   = dists,
    weights = context$model_weights,
    dx      = dx,
    n_grid  = context$n_grid
  )

  .prior_linear_density_transform(dist, output_transformation,
                                  output_transformation_arguments,
                                  n_grid = context$n_grid)
}

.prior_density_conditional_context_density <- function(context, weights,
                                                       source_transforms = NULL,
                                                       output_transformation = NULL,
                                                       output_transformation_arguments = NULL){

  if(!inherits(context, "prior_density_conditional_context")){
    stop("'context' must be a conditional prior density context.", call. = FALSE)
  }

  if(length(context$prior_lists) == 0){
    return(.prior_linear_density_point(0))
  }

  dists <- lapply(context$prior_lists, function(prior_list){
    if(!is.null(context$formula_scale) && length(context$formula_scale) > 0){
      component_context <- .prior_density_context(
        prior_list    = prior_list,
        column_names  = context$column_names,
        formula_scale = context$formula_scale,
        n_grid        = context$n_grid,
        tail_prob     = context$tail_prob
      )
      return(.prior_density_context_density(
        context           = component_context,
        weights           = weights,
        source_transforms = source_transforms
      ))
    }

    .prior_linear_combination_density(
      prior_list        = prior_list,
      weights           = weights,
      n_grid            = context$n_grid,
      tail_prob         = context$tail_prob,
      source_transforms = source_transforms
    )
  })

  dx <- min(vapply(dists, function(dist){
    if(!is.null(dist$density) && length(dist$density$x) > 1){
      return(dist$density$x[2] - dist$density$x[1])
    }
    Inf
  }, numeric(1)))
  if(!is.finite(dx)){
    dx <- NA_real_
  }

  dist <- .prior_linear_density_mix(
    dists   = dists,
    weights = context$model_weights,
    dx      = dx,
    n_grid  = context$n_grid
  )

  .prior_linear_density_transform(dist, output_transformation,
                                  output_transformation_arguments,
                                  n_grid = context$n_grid)
}

.prior_density_build_context <- function(prior_list, column_names, formula_scale = NULL,
                                         n_grid = .prior_linear_density_default_grid(),
                                         tail_prob = .prior_linear_density_tail_prob(),
                                         conditional = NULL,
                                         conditional_rule = "AND"){

  if(length(conditional) > 0){
    return(.prior_density_conditional_context(
      prior_list       = prior_list,
      column_names     = column_names,
      conditional      = conditional,
      conditional_rule = conditional_rule,
      formula_scale    = formula_scale,
      n_grid           = n_grid,
      tail_prob        = tail_prob
    ))
  }

  if(all(vapply(prior_list, is.prior, logical(1)))){
    return(.prior_density_context(prior_list, column_names, formula_scale, n_grid, tail_prob))
  }

  if(!is.null(formula_scale) && length(formula_scale) > 0){
    stop("Formula-scale prior densities for model-list mixtures are not implemented.", call. = FALSE)
  }

  .prior_density_model_mixture_context(prior_list, column_names, n_grid, tail_prob)
}

.prior_density_from_context <- function(context, weights,
                                        source_transforms = NULL,
                                        output_transformation = NULL,
                                        output_transformation_arguments = NULL){

  if(inherits(context, "prior_density_context")){
    return(.prior_density_context_density(
      context                         = context,
      weights                         = weights,
      source_transforms               = source_transforms,
      output_transformation           = output_transformation,
      output_transformation_arguments = output_transformation_arguments
    ))
  }

  if(inherits(context, "prior_density_model_mixture_context")){
    if(!is.null(source_transforms)){
      stop("Source transformations are not supported for model-list prior mixtures.", call. = FALSE)
    }
    return(.prior_density_model_mixture_density(
      context                         = context,
      weights                         = weights,
      output_transformation           = output_transformation,
      output_transformation_arguments = output_transformation_arguments
    ))
  }

  if(inherits(context, "prior_density_conditional_context")){
    return(.prior_density_conditional_context_density(
      context                         = context,
      weights                         = weights,
      source_transforms               = source_transforms,
      output_transformation           = output_transformation,
      output_transformation_arguments = output_transformation_arguments
    ))
  }

  stop("Unknown prior density context.", call. = FALSE)
}

.prior_density_from_context_rows <- function(context, weights,
                                             source_transforms = NULL,
                                             output_transformation = NULL,
                                             output_transformation_arguments = NULL){

  if(is.null(dim(weights))){
    return(.prior_density_from_context(
      context                         = context,
      weights                         = weights,
      source_transforms               = source_transforms,
      output_transformation           = output_transformation,
      output_transformation_arguments = output_transformation_arguments
    ))
  }

  weights <- as.matrix(weights)
  if(nrow(weights) == 0){
    return(.prior_linear_density_point(0))
  }
  if(nrow(weights) == 1){
    return(.prior_density_from_context(
      context                         = context,
      weights                         = weights[1, ],
      source_transforms               = source_transforms,
      output_transformation           = output_transformation,
      output_transformation_arguments = output_transformation_arguments
    ))
  }

  row_keys <- apply(signif(weights, 15), 1, paste0, collapse = "\r")
  unique_keys <- unique(row_keys)
  row_counts <- tabulate(match(row_keys, unique_keys), nbins = length(unique_keys))
  row_indices <- match(unique_keys, row_keys)

  dists <- lapply(row_indices, function(row_i){
    .prior_density_from_context(
      context                         = context,
      weights                         = weights[row_i, ],
      source_transforms               = source_transforms,
      output_transformation           = output_transformation,
      output_transformation_arguments = output_transformation_arguments
    )
  })

  dx <- min(vapply(dists, function(dist){
    if(!is.null(dist$density) && length(dist$density$x) > 1){
      return(dist$density$x[2] - dist$density$x[1])
    }
    Inf
  }, numeric(1)))
  if(!is.finite(dx)){
    dx <- NA_real_
  }

  .prior_linear_density_mix(
    dists   = dists,
    weights = row_counts,
    dx      = dx,
    n_grid  = if(!is.null(context$n_grid)) context$n_grid else NULL
  )
}

.prior_density_coefficient_weights <- function(column_names, parameter){

  weights <- rep(0, length(column_names))
  names(weights) <- column_names
  if(parameter %in% names(weights)){
    weights[[parameter]] <- 1
  }
  weights
}

.generate_transformed_prior_densities <- function(prior_list, column_names, formula_scale = NULL,
                                                  conditional = NULL, conditional_rule = "AND",
                                                  n_grid = .prior_linear_density_default_grid(),
                                                  tail_prob = .prior_linear_density_tail_prob()){

  context <- .prior_density_build_context(
    prior_list       = prior_list,
    column_names     = column_names,
    formula_scale    = formula_scale,
    n_grid           = n_grid,
    tail_prob        = tail_prob,
    conditional      = conditional,
    conditional_rule = conditional_rule
  )

  prior_columns <- unlist(lapply(names(prior_list), function(parameter){
    .prior_linear_prior_columns(parameter, prior_list[[parameter]])
  }), use.names = FALSE)
  prior_columns <- intersect(prior_columns, column_names)

  out <- list()
  for(parameter in prior_columns){
    weights <- .prior_density_coefficient_weights(column_names, parameter)
    source_transforms <- NULL
    output_transformation <- NULL

    for(transform in context$transforms){
      if(transform$log_intercept && identical(parameter, transform$intercept)){
        weights <- rep(0, length(column_names))
        names(weights) <- column_names
        weights[transform$columns] <- transform$matrix[parameter, ]

        source_transforms <- rep(NA_character_, length(column_names))
        names(source_transforms) <- column_names
        source_transforms[[transform$intercept]] <- "log"
        output_transformation <- "exp"
        break
      }
    }

    if(!is.null(output_transformation)){
      out[[parameter]] <- .prior_linear_combination_density(
        prior_list             = prior_list,
        weights                = weights,
        n_grid                 = n_grid,
        tail_prob              = tail_prob,
        source_transforms      = source_transforms,
        output_transformation  = output_transformation
      )
    }else{
      out[[parameter]] <- .prior_density_from_context(
        context               = context,
        weights               = weights,
        source_transforms     = source_transforms,
        output_transformation = output_transformation
      )
    }
  }

  attr(out, "context") <- context
  class(out) <- c("prior_density_list", "list")
  return(out)
}

#' @title Plot Transformed Prior Densities
#'
#' @description Plots a prior density after applying the same deterministic
#' coefficient transformation used for formula-scale back-transforms. The helper
#' supports continuous densities and point-mass priors in the same plot.
#'
#' @param prior_list named list of prior distributions.
#' @param column_names character vector of coefficient column names defining the
#' fitted parameter space.
#' @param formula_scale optional nested formula-scale metadata, in the same shape
#' as \code{attr(fit, "formula_scale")}.
#' @param parameter coefficient name to plot.
#' @param n_points number of plotting points.
#' @param x_range optional plotting range on the untransformed scale.
#' @param transformation optional output transformation passed to the prior
#' plotting machinery.
#' @param transformation_arguments optional list of transformation arguments.
#' @param transformation_settings whether \code{x_range} is supplied on the
#' transformed scale.
#' @param plot_type either \code{"base"} or \code{"ggplot"}.
#' @param par_name optional plotting label.
#' @param ... additional plotting arguments.
#'
#' @return For \code{plot_type = "base"}, invisibly returns base-plot metadata.
#' For \code{plot_type = "ggplot"}, returns a \code{ggplot} object. Returns
#' \code{NULL} when the requested parameter is an identity transform and no
#' output transformation was requested.
#'
#' @seealso [plot_prior_list()] [transform_scale_samples()]
#' @export
plot_transformed_prior <- function(prior_list, column_names, formula_scale = NULL,
                                   parameter, n_points = 1000, x_range = NULL,
                                   transformation = NULL,
                                   transformation_arguments = NULL,
                                   transformation_settings = FALSE,
                                   plot_type = c("base", "ggplot"),
                                   par_name = NULL, ...){

  check_list(prior_list, "prior_list")
  check_char(column_names, "column_names", check_length = FALSE)
  check_list(formula_scale, "formula_scale", allow_NULL = TRUE)
  check_char(parameter, "parameter")
  check_int(n_points, "n_points", lower = 2)
  check_real(x_range, "x_range", check_length = 2, allow_NULL = TRUE)
  plot_type <- match.arg(plot_type)
  .check_transformation_input(transformation, transformation_arguments, transformation_settings)

  if(!parameter %in% column_names){
    stop("Parameter '", parameter, "' was not found in 'column_names'.", call. = FALSE)
  }

  if(is.null(transformation) &&
     .plot_transformed_prior_is_identity(prior_list, column_names, formula_scale, parameter)){
    return(NULL)
  }

  prior_densities <- .generate_transformed_prior_densities(
    prior_list    = prior_list,
    column_names  = column_names,
    formula_scale = formula_scale
  )

  if(!parameter %in% names(prior_densities)){
    return(NULL)
  }

  plot_data <- .prior_linear_density_to_plot_data(
    prior_densities[[parameter]],
    n_points                 = n_points,
    x_range                  = x_range,
    transformation           = transformation,
    transformation_arguments = transformation_arguments,
    transformation_settings  = transformation_settings
  )

  .plot_prior_list.both(
    plot_data = plot_data,
    plot_type = plot_type,
    par_name  = if(is.null(par_name)) parameter else par_name,
    ...
  )
}

.plot_transformed_prior_is_identity <- function(prior_list, column_names, formula_scale, parameter){

  if(is.null(formula_scale) || length(formula_scale) == 0L){
    return(TRUE)
  }

  context <- .prior_density_build_context(
    prior_list    = prior_list,
    column_names  = column_names,
    formula_scale = formula_scale
  )

  if(!inherits(context, "prior_density_context")){
    return(FALSE)
  }

  weights <- .prior_density_coefficient_weights(column_names, parameter)

  for(transform in context$transforms){
    if(transform$log_intercept && identical(parameter, transform$intercept)){
      return(FALSE)
    }
  }

  transformed_weights <- .prior_density_context_standardized_weights(context, weights)
  weights <- weights[abs(weights) > .prior_linear_density_zero_tol()]

  all_names <- union(names(weights), names(transformed_weights))
  raw <- rep(0, length(all_names))
  names(raw) <- all_names
  transformed <- raw
  raw[names(weights)] <- weights
  transformed[names(transformed_weights)] <- transformed_weights

  isTRUE(all.equal(raw, transformed, tolerance = .prior_linear_density_zero_tol()))
}

.prior_factor_level_weight_matrix <- function(sample_metadata, parameter, samples = NULL){

  metadata <- sample_metadata
  if(!is.null(samples)){
    metadata <- .add_factor_metadata_from_named_objects(metadata, parameter, samples)
  }

  if(isTRUE(attr(metadata, "orthonormal")) ||
     isTRUE(attr(metadata, "meandif")) ||
     inherits(metadata, "mixed_posteriors.treatment_transformed")){
    design_info <- .factor_term_design_from_metadata(metadata)
    weights <- design_info$design
    rownames(weights) <- .factor_contrast_parameter_names(
      parameter   = parameter,
      level_names = design_info$level_names,
      cell_names  = design_info$cell_names
    )
    prior <- attr(metadata, "prior_list")
    if(!is.null(prior) && !is.prior(prior)){
      prior_candidates <- prior[vapply(prior, is.prior, logical(1))]
      prior <- if(length(prior_candidates) > 0) prior_candidates[[1]] else NULL
    }
    prior_columns <- if(!is.null(prior) && is.prior(prior)){
      .prior_linear_prior_columns(parameter, prior)
    }else{
      NULL
    }
    colnames(weights) <- if(length(prior_columns) == ncol(weights)){
      prior_columns
    }else{
      colnames(as.matrix(sample_metadata))
    }
    return(weights)
  }

  raw_columns <- colnames(as.matrix(sample_metadata))
  weights <- diag(length(raw_columns))
  rownames(weights) <- raw_columns
  prior <- attr(metadata, "prior_list")
  if(!is.null(prior) && !is.prior(prior)){
    prior_candidates <- prior[vapply(prior, is.prior, logical(1))]
    prior <- if(length(prior_candidates) > 0) prior_candidates[[1]] else NULL
  }
  prior_columns <- if(!is.null(prior) && is.prior(prior)){
    .prior_linear_prior_columns(parameter, prior)
  }else{
    NULL
  }
  colnames(weights) <- if(length(prior_columns) == ncol(weights)){
    prior_columns
  }else{
    raw_columns
  }
  return(weights)
}

---
title: "Comparison to other R packages"
author: "František Bartoš"
date: "`r Sys.Date()`"
output:
  rmarkdown::html_vignette:
    self_contained: yes
bibliography: ../inst/REFERENCES.bib
csl: ../inst/apa.csl
vignette: >
  %\VignetteIndexEntry{Comparison to other R packages}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---


```{r setup, include = FALSE}
is_check <- ("CheckExEnv" %in% search()) ||
             any(c("_R_CHECK_TIMINGS_", "_R_CHECK_LICENSE_") %in% names(Sys.getenv()))
knitr::opts_chunk$set(
  collapse = TRUE,
  comment  = "#>",
  eval     = !is_check,
  dev      = "png"
)
if(.Platform$OS.type == "windows"){
  knitr::opts_chunk$set(dev.args = list(type = "cairo"))
}
```

BayesTools R package allows package developers and users to conveniently specify a wide variety of models. In this vignette, I show how to reproduce some commonly used models and compare the results to output from other, more specialized, R packages.

## Bayesian Two-Sample T-Test

Bayesian two-sample t-test is one of the most commonly used tests. It compares means of two independent groups and the underlying model is usually defined as<br>
$x_i \sim \text{Normal}(\mu - \frac{\alpha}{2}, \sigma^2), i = 1,\dots,N_x$,<br>
$y_i \sim \text{Normal}(\mu + \frac{\alpha}{2}, \sigma^2), i = 1,\dots,N_y$.<br>
where $\mu$ corresponds to the grand mean, $\sigma$ to the grand standard deviation, and $\alpha$ corresponds to the total effect. The standardized effect size, the usual quantity of interest, $\delta$ can be then obtained as $\delta = \frac{\alpha}{\sigma}$ [@rouder2009bayesian].

To perform the test, we specify two competing hypotheses:

 - the null hypothesis $\mathcal{H}_0: \delta = 0$ assuming that the standardized effect size is zero,
 - and the alternative hypothesis $\mathcal{H}_1: \delta \sim g()$ assuming that the standardized effect size is non-zero and prior distribution $g()$ characterizes our hypothesis about the possible values of the standardized effect size.


### Kitchen Rolls Data Set

To illustrate the two-sample t-test specification with BayesTools, I use a data set from a replication study conducted by [@wagenmakers2015turning]. @wagenmakers2015turning replicated the 2nd experiment from a study by @topolinski2012turning who claimed that clockwise movements induce psychological states of progression in time and increases orientation toward the future and novelty. In their second experiment, @topolinski2012turning let 60 participants to rotate kitchen rolls either clock or counter-clock wise, inducing the progression in time, and increasing their mean score in NEO PI-R items measuring openness to experience *t*(58) = 2.21, *p* = 0.031, Cohen's *d* = 0.58. In the replication study, @wagenmakers2015turning collected data from 102 participants and found a $\text{BF}_{0+}$ = 10.76 in the favor of null hypothesis of no effect against a positive effect in the direction of the original study.

First, we load the package, data set, and split the data set into the two groups.

```{r}
library(BayesTools)
data("kitchen_rolls")
x <- kitchen_rolls$mean_NEO[kitchen_rolls$rotation == "counter"]
y <- kitchen_rolls$mean_NEO[kitchen_rolls$rotation == "clock"]
```

```{r fig_dist, fig.width = 4, fig.height = 4, dpi = 300, out.width = "60%", fig.align = "center"}
h1 <- hist(x, breaks = 15, plot = FALSE)
h2 <- hist(y, breaks = 15, plot = FALSE)
par(mar = c(4, 4, 0, 1))
plot(h1, col = rgb(0,0,1,1/4), xlim = c(-1, 2), ylim = c(0, 16), las = 1, main = "", xlab = "mean NEO PI-R")
plot(h2, col = rgb(1,0,0,1/4), add = TRUE)
legend("topright", legend = c("Counter", "Clock"), fill = c(rgb(0,0,1,1/4), rgb(1,0,0,1/4)), bty = "n")
```


### Implementation with BayesTools

I first specify the model likelihood using the JAGS modeling syntax. Note that JAGS uses mean and precision to define a normal distribution, therefore, I need to specify $\sigma^-2$ as the second parameter. Otherwise, the syntax closely resembles the mathematical notation used to outline the two-sample t-test (notice the substitution of $\delta$ in the place of $\alpha$). See [JAGS user manual](https://people.stat.sc.edu/hansont/stat740/jags_user_manual.pdf) for more details about the syntax and supported distributions.

```{r}
ttest_model <- 
"model{
  for(i in 1:Nx){
    x[i] ~ dnorm(mu - delta*sigma/2, pow(sigma, -2))
  }
  for(i in 1:Ny){
    y[i] ~ dnorm(mu + delta*sigma/2, pow(sigma, -2))
  }
}
"
```

Researchers familiar with specifying models in JAGS might notice that I did not specify the prior distribution for any of our parameters ($\mu, \sigma, \delta$). In fact, all operations related to prior distributions are handled automatically by the BayesTools package. Here, I specify two lists of prior distributions, the first one for the null hypothesis $\mathcal{H}_0$ of no effect and second for the alternative hypothesis of a positive effect $\mathcal{H}_+$. As in @wagenmakers2015turning, I use a point null hypothesis for the $\delta$ for $\mathcal{H}_0$ and a half Cauchy distribution with scale $\sqrt{2}/2$ for the alternative hypothesis. Unfortunately, JAGS does not allow us to specify Jeffreys' priors for the common mean $\mu$ and standard deviation $\sigma$, therefore I use a relative wide Cauchy prior distribution for the mean and an exponential distribution for the standard deviation, both of which have approximately the same shape as the Jeffreys' priors in areas where the likelihood is the highest.$^1$

```{r} 
ttest_priors_H0 <- list(
  mu    = prior("cauchy",      parameters = list(location = 0, scale = 10)),
  sigma = prior("exponential", parameters = list(rate = 2)),
  delta = prior("spike",       parameters = list(location = 0))
)
ttest_priors_Hp <- list(
  mu    = prior("cauchy",      parameters = list(location = 0, scale = 10)),
  sigma = prior("exponential", parameters = list(rate = 2)),
  delta = prior("cauchy",      parameters = list(location = 0, scale = 1),
                truncation = list(lower = 0, upper = Inf))
)
```

I can visualize and check the prior distributions with the `plot()` function.

```{r fig_priors, fig.width = 8, fig.height = 8, dpi = 300, out.width = "100%", fig.align = "center"}
par(mfrow = c(2, 2), mar = c(4, 4, 2, 1))
plot(ttest_priors_H0$mu,    par_name = bquote(mu), xlim = c(-50, 50))
plot(ttest_priors_H0$sigma, par_name = bquote(sigma), xlim = c(0, 50))
plot(ttest_priors_H0$delta, par_name = bquote(delta), 
     xlim = c(-1, 1), main = bquote(H[0]))
plot(ttest_priors_Hp$delta, par_name = bquote(delta), 
     xlim = c(0, 5), main = bquote(H[1]))
```

The last step needed before actually fitting the models is specifying a list with the data set. Here, I need to pass both the observations for $x$ and $y$, but also the number of observations in each group, $N_x$ and $N_y$, since I used it the model syntax I specified earlier.

```{r} 
ttest_data <- list(
  x  = x,
  y  = y,
  Nx = length(x),
  Ny = length(y)
)
```

Then, I pass the arguments into the `JAGS_fit()` function that wraps around the runjags R package [@runjags], specifying the prior distributions, preparing the starting values, etc...

```{r} 
ttest_model_H0 <- JAGS_fit(
  model_syntax = ttest_model,
  data         = ttest_data,
  prior_list   = ttest_priors_H0,
  seed         = 0
) 

ttest_model_Hp <- JAGS_fit(
  model_syntax = ttest_model,
  data         = ttest_data,
  prior_list   = ttest_priors_Hp,
  seed         = 1
) 
```

I can print the estimated model summary with `runjags_estimates_table()` function.

```{r}
runjags_estimates_table(ttest_model_Hp)
```

To compute the Bayes factor in favor of $\mathcal{H}_+$ over $\mathcal{H}_0$, I need to first obtain the marginal likelihoods. The only additional step is specifying an R function that returns the log likelihood of the model. Here, instead of writing out the loops, I take advantage of the fact that summation on the log scale equals to multiplication on the original scale.

```{r}
log_posterior <- function(parameters, data){
  loglik_x <- sum(dnorm(
    x    = data[["x"]],
    mean = parameters[["mu"]] - parameters[["delta"]] * parameters[["sigma"]] / 2,
    sd   = parameters[["sigma"]],
    log  = TRUE
  ))
  loglik_y <- sum(dnorm(
    x    = data[["y"]],
    mean = parameters[["mu"]] + parameters[["delta"]] * parameters[["sigma"]] / 2,
    sd   = parameters[["sigma"]], 
    log  = TRUE
  ))
  return(loglik_x + loglik_y)
}
```

The function requires two arguments, `parameters` and `data`, both of which are forwarded to it via the `JAGS_bridgesampling` function that wraps around the bridge sampling R package [@bridgesampling], specifying the marginal likelihood for prior distributions and taking care of all required transformations etc...

```{r}
marglik_model_H0 <- JAGS_bridgesampling(
  fit           = ttest_model_H0,
  log_posterior = log_posterior,
  data          = ttest_data,
  prior_list    = ttest_priors_H0
)

marglik_model_Hp <- JAGS_bridgesampling(
  fit           = ttest_model_Hp,
  log_posterior = log_posterior,
  data          = ttest_data,
  prior_list    = ttest_priors_Hp
)
```

To obtain the resulting Bayes factor in favor of the null hypothesis, I can use the `bf()` function from the bridgesampling R package

```{r}
bridgesampling::bf(marglik_model_H0, marglik_model_Hp)
```

or specify a BayesTools model ensemble object that can be further inspected by a variety of functions.

```{r}
models_list <- models_inference(list(
  list(model = ttest_model_H0, marglik = marglik_model_H0, prior_weights = 1/2),
  list(model = ttest_model_Hp, marglik = marglik_model_Hp, prior_weights = 1/2)
))
ensemble_info <- ensemble_inference(models_list, parameters = "delta", is_null_list = list("delta" = c(TRUE, FALSE)))

ensemble_inference_table(ensemble_info, "delta", BF01 = TRUE)
```

The corresponding results can be also obtained numerically with the `ttestBF()` function from the BayesFactor R package [@BayesFactor].

```{r}
BayesFactor_ttest <- BayesFactor::ttestBF(x = x, y = y, rscale = "wide", nullInterval = c(0, Inf))
BayesFactor_ttest
1/exp(BayesFactor_ttest@bayesFactor$bf[2])
```

This is of course a trivial example, however, it showcases flexibility of the BayesTools package. The prior distribution on the standardized effect size parameter $\delta$ can be changed to any of the multitude of supported distributions without effecting rest of the code, allowing users and package developers to specify more customized tests.



## Footnotes
$^1$ Although the Jeffreys' prior distributions integrate out from the numerical solution for the default Bayesian t-test/ANOVA, the prior distribution for sigma can influence the resulting Bayes factor in favor of the effect since it scales the standardized effect size coefficient. 


### References

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/interpret.R
\name{interpret}
\alias{interpret}
\alias{interpret2}
\title{Interpret ensemble inference and estimates}
\usage{
interpret(inference, samples, specification, method)

interpret2(specification, method = NULL)
}
\arguments{
\item{inference}{model inference created by \link{ensemble_inference}}

\item{samples}{posterior samples created by \link{mix_posteriors}}

\item{specification}{list of lists specifying the generated text.
Each inner list carries: (1) \code{inference} specifying the name of
in the \code{inference} entry and optionally \code{inference_name}
as a name to use in the text and \code{inference_BF_name} as a
symbol to be used instead of the default \code{"BF"}. Finite-sample
BF bounds can be supplied either as a \code{bound_operator} attribute
on the numeric BF or as \code{inference_BF_bound_operator} /
\code{BF_bound_operator} with value \code{"<"} or \code{">"}. (2) \code{samples}
specifying the name of in the \code{samples} entry and optionally
\code{samples_name} as a name to use in the text, \code{samples_units} as
a unit text to be appended after the estimate, and \code{samples_conditional}
specifying whether the estimate is conditional or model-averaged.}

\item{method}{character specifying name of the method to be
appended at the beginning of each sentence.}
}
\value{
\code{interpret} returns character.
}
\description{
Provides textual summary for posterior
distributions created by \link{mix_posteriors} and ensemble inference
created by \link{ensemble_inference}.
}
\seealso{
\link{ensemble_inference} \link{mix_posteriors} \link{BayesTools_model_tables} \link{BayesTools_ensemble_tables}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/distributions-weightfunctions.R
\name{weightfunctions}
\alias{weightfunctions}
\alias{mdone.sided}
\alias{mdtwo.sided}
\alias{mdone.sided_fixed}
\alias{mdtwo.sided_fixed}
\alias{rone.sided}
\alias{rtwo.sided}
\alias{rone.sided_fixed}
\alias{rtwo.sided_fixed}
\alias{mpone.sided}
\alias{mptwo.sided}
\alias{mpone.sided_fixed}
\alias{mptwo.sided_fixed}
\alias{mqone.sided}
\alias{mqtwo.sided}
\alias{mqone.sided_fixed}
\alias{mqtwo.sided_fixed}
\title{Weight functions}
\usage{
mdone.sided(x, alpha = NULL, alpha1 = NULL, alpha2 = NULL, log = FALSE)

mdtwo.sided(x, alpha, log = FALSE)

mdone.sided_fixed(x, omega, log = FALSE)

mdtwo.sided_fixed(x, omega, log = FALSE)

rone.sided(n, alpha = NULL, alpha1 = NULL, alpha2 = NULL)

rtwo.sided(n, alpha)

rone.sided_fixed(n, omega)

rtwo.sided_fixed(n, omega)

mpone.sided(
  q,
  alpha = NULL,
  alpha1 = NULL,
  alpha2 = NULL,
  lower.tail = TRUE,
  log.p = FALSE
)

mptwo.sided(q, alpha, lower.tail = TRUE, log.p = FALSE)

mpone.sided_fixed(q, omega, lower.tail = TRUE, log.p = FALSE)

mptwo.sided_fixed(q, omega, lower.tail = TRUE, log.p = FALSE)

mqone.sided(
  p,
  alpha = NULL,
  alpha1 = NULL,
  alpha2 = NULL,
  lower.tail = TRUE,
  log.p = FALSE
)

mqtwo.sided(p, alpha, lower.tail = TRUE, log.p = FALSE)

mqone.sided_fixed(p, omega, lower.tail = TRUE, log.p = FALSE)

mqtwo.sided_fixed(p, omega, lower.tail = TRUE, log.p = FALSE)
}
\arguments{
\item{x, q}{vector or matrix of quantiles.}

\item{alpha}{vector or matrix with concentration parameters for
the Dirichlet distribution for a monotonic one.sided or a two.sided
weight function.}

\item{alpha1}{vector or matrix with concentration parameters for
the Dirichlet distribution for the expected direction of non-monotonic
one.sided of weight function.}

\item{alpha2}{vector or matrix with concentration parameters for
the Dirichlet distribution for the unexpected direction of non-monotonic
one.sided of weight function.}

\item{log, log.p}{logical; if \code{TRUE}, probabilities
\code{p} are given as \code{log(p)}.}

\item{omega}{vector or matrix of fixed non-negative relative weights for a
one.sided or a two.sided weight function. The first/reference weight must be
exactly 1.}

\item{n}{number of observations.}

\item{lower.tail}{logical; if \code{TRUE} (default), probabilities
are \eqn{P[X \le x]}, otherwise, \eqn{P[X > x]}.}

\item{p}{vector of probabilities.}
}
\value{
\code{mdone.sided}, \code{mdtwo.sided}, \code{mdone.sided_fixed},
and \code{mdtwo.sided_fixed} give the marginal density,
\code{mpone.sided}, \code{mptwo.sided}, \code{mpone.sided_fixed},
and \code{mptwo.sided_fixed} give the marginal distribution function,
\code{mqone.sided}, \code{mqtwo.sided}, \code{mqone.sided_fixed},
and \code{mqtwo.sided_fixed} give the marginal quantile function,
and \code{rone.sided}, \code{rtwo.sided}, \code{rone.sided_fixed},
and \code{rtwo.sided_fixed} generate random deviates.
}
\description{
Marginal density, marginal distribution function,
marginal quantile function and random generation for weight functions.
}
\examples{
# draw samples from a two-sided weight function
rtwo.sided(10, alpha = c(1, 1))

# draw samples from a monotone one-sided weight function
rone.sided(10, alpha = c(1, 1, 1))

# draw samples from a non-monotone one-sided weight function
rone.sided(10, alpha1 = c(1, 1), alpha2 = c(1, 1))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model-averaging-plots.R
\name{plot_posterior}
\alias{plot_posterior}
\title{Plot samples from the mixed posterior distributions}
\usage{
plot_posterior(
  samples,
  parameter,
  plot_type = "base",
  prior = FALSE,
  n_points = 1000,
  n_samples = 10000,
  force_samples = FALSE,
  individual = FALSE,
  show_figures = NULL,
  transformation = NULL,
  transformation_arguments = NULL,
  transformation_settings = FALSE,
  rescale_x = FALSE,
  par_name = NULL,
  effect_direction = "positive",
  dots_prior = list(),
  data = NULL,
  show_data = FALSE,
  dots_data = list(),
  legend = TRUE,
  legend_title = NULL,
  legend_labels = NULL,
  legend_position = NULL,
  ...
)
}
\arguments{
\item{samples}{samples from a posterior distribution for a
parameter generated by \link{mix_posteriors} or \link{as_mixed_posteriors}.}

\item{parameter}{parameter name to be plotted. Use \code{"PETPEESE"}
for PET-PEESE plot with parameters \code{"PET"} and \code{"PEESE"},
and \code{"weightfunction"} for plotting a weightfunction with
parameters \code{"omega"}.}

\item{plot_type}{whether to use a base plot \code{"base"}
or ggplot2 \code{"ggplot"} for plotting.}

\item{prior}{whether prior distribution should be added to the figure.
When samples were prepared with \code{as_mixed_posteriors(..., transform_scaled = TRUE)},
the transformed prior samples are automatically used.}

\item{n_points}{number of equally spaced points
in the \code{x_range} if \code{x_seq} is unspecified}

\item{n_samples}{number of samples from the prior
distribution if the density cannot be obtained
analytically (or if samples are forced with
\code{force_samples = TRUE})}

\item{force_samples}{should prior be sampled instead
of obtaining analytic solution whenever possible}

\item{individual}{should individual densities be returned
(e.g., in case of weightfunction)}

\item{show_figures}{which figures should be returned in case of
multiple plots are generated. Useful when priors for the omega
parameter are plotted and \code{individual = TRUE}.}

\item{transformation}{transformation to be applied
to the prior distribution. Either a character
specifying one of the prepared transformations:
\describe{
\item{lin}{linear transformation in form of \code{a + b*x}}
\item{tanh}{hyperbolic tangent transformation}
\item{exp}{exponential transformation}
}, or a list containing the transformation function \code{fun},
inverse transformation function \code{inv}, and derivative of the
transformation \code{jac}, evaluated on the original support. See examples
for details.}

\item{transformation_arguments}{a list with named arguments for
the \code{transformation}}

\item{transformation_settings}{boolean indicating whether the
settings the \code{x_seq} or \code{x_range} was specified on
the transformed support}

\item{rescale_x}{allows to rescale x-axis in case a
weightfunction is plotted.}

\item{par_name}{a type of parameter for which the prior is
specified. Only relevant if the prior corresponds to a mu
parameter that needs to be transformed.}

\item{effect_direction}{direction of the effect for PET-PEESE
regression. Use \code{"positive"} (default) for
\code{mu + PET*se + PEESE*se^2} or \code{"negative"} for
\code{mu - PET*se - PEESE*se^2}.}

\item{dots_prior}{additional arguments for the prior distribution plot}

\item{data}{optional numeric vector of observed p-values in \code{[0, 1]}, or
a data frame with a \code{p} column. Used only for weightfunction plots.}

\item{show_data}{whether observed p-values should be shown as rug marks on
the weightfunction x-axis.}

\item{dots_data}{additional styling arguments for observed p-value rug marks.
Supports \code{col}/\code{color}, \code{alpha}, \code{lwd}/\code{linewidth},
\code{side}/\code{rug_side}, and \code{height}/\code{rug_height}.}

\item{legend}{whether factor legends should be drawn.}

\item{legend_title}{optional title for factor legends.}

\item{legend_labels}{optional labels for factor legend levels.}

\item{legend_position}{optional legend position for factor legends.}

\item{...}{additional arguments}
}
\value{
\code{plot_posterior} returns either \code{NULL} or
an object of class 'ggplot' if plot_type is \code{plot_type = "ggplot"}.
}
\description{
Plot samples from the mixed posterior distributions
}
\details{
When using scaled predictors (via \code{formula_scale_list} in \link{JAGS_fit}),
you can plot posteriors on the original (unscaled) scale by preparing samples with
\code{as_mixed_posteriors(..., transform_scaled = TRUE)}. The function automatically
detects this and uses the pre-computed transformed prior samples when \code{prior = TRUE}.
}
\seealso{
\code{\link[=prior]{prior()}} \code{\link[=lines_prior_list]{lines_prior_list()}}  \code{\link[=geom_prior_list]{geom_prior_list()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/priors.R
\name{prior_weightfunction}
\alias{prior_weightfunction}
\alias{wf_cumulative}
\alias{wf_fixed}
\alias{wf_independent}
\title{Creates a prior distribution for a weight function}
\usage{
prior_weightfunction(
  side = "one-sided",
  steps = c(0.025, 0.05),
  weights = wf_cumulative(),
  reference = "most_significant",
  prior_weights = 1
)

wf_cumulative(alpha = NULL)

wf_fixed(omega)

wf_independent(prior, scale = "omega")
}
\arguments{
\item{side}{side geometry. Either \code{"one-sided"} or \code{"two-sided"}.}

\item{steps}{increasing p-value cut points between 0 and 1.}

\item{weights}{a weight-prior object created by \code{wf_cumulative()},
\code{wf_fixed()}, or \code{wf_independent()}.}

\item{reference}{reference bin. Currently only \code{"most_significant"} is
supported and fixes the most significant bin to \code{omega = 1}.}

\item{prior_weights}{prior odds associated with a given distribution.}

\item{alpha}{positive cumulative-Dirichlet concentration parameters. If
omitted, a flat Dirichlet prior is used with one concentration parameter per
bin.}

\item{omega}{fixed non-negative relative publication weights, one per bin.
The reference-bin weight must be exactly 1.}

\item{prior}{prior distribution for each non-reference weight.}

\item{scale}{latent scale for independent weights. \code{"omega"} places the
prior directly on the non-negative publication weight. \code{"log_omega"}
places the prior on \code{log(omega)} and transforms with
\code{omega = exp(log_omega)}, allowing weights above one whenever the log
prior assigns mass above zero.}
}
\value{
\code{prior_weightfunction} returns an object of class 'prior'.
}
\description{
\code{prior_weightfunction} creates a prior distribution for
fitting a RoBMA selection model. The \code{side} and \code{steps} arguments
define the p-value bins, and the \code{weights} argument defines the prior on
the publication weights in those bins.
}
\examples{
p1 <- prior_weightfunction(
  side = "one-sided",
  steps = c(.05, .10),
  weights = wf_cumulative(alpha = c(1, 1, 1))
)

p2 <- prior_weightfunction(
  side = "one-sided",
  steps = c(.05),
  weights = wf_independent(prior("beta", list(1, 1)))
)

}
\seealso{
\code{\link[=plot.prior]{plot.prior()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/BayesTools-package.R
\docType{package}
\name{BayesTools}
\alias{BayesTools}
\alias{BayesTools-package}
\title{BayesTools}
\description{
BayesTools: Provides tools for conducting Bayesian analyses. The package contains
functions for creating a wide range of prior distribution objects, mixing posterior
samples from JAGS and Stan models, plotting posterior distributions, and etc...
The tools for working with prior distribution span from visualization, generating JAGS
and bridgesampling syntax to basic functions such as rng, quantile, and distribution functions.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://fbartos.github.io/BayesTools/}
  \item Report bugs at \url{https://github.com/FBartos/BayesTools/issues}
}

}
\author{
František Bartoš \email{f.bartos96@gmail.com}
}
\keyword{package}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/priors-informed.R
\name{prior_informed}
\alias{prior_informed}
\title{Creates an informed prior distribution based on research}
\usage{
prior_informed(name, parameter = NULL, type = "smd")
}
\arguments{
\item{name}{name of the prior distribution. There are many options based on prior psychological
or medical research.
For psychology, the possible options are
\describe{
\item{\code{"van Erp"}}{for an informed prior distribution for the heterogeneity parameter tau
of meta-analytic effect size estimates based on standardized mean differences
\insertCite{erp2017estimates}{BayesTools},}
\item{\code{"Oosterwijk"}}{for an informed prior distribution for the effect sizes expected in
social psychology based on prior elicitation with dr. Oosterwijk
\insertCite{gronau2017bayesian}{BayesTools}.}
}
For medicine, the possible options are based on \insertCite{bartos2021bayesian;textual}{BayesTools}
and  \insertCite{bartos2023empirical;textual}{BayesTools}
who developed empirical prior distributions for the effect size and heterogeneity parameters of the
continuous outcomes (standardized mean differences), dichotomous outcomes (logOR, logRR, and risk differences),
and time to event outcomes (logHR) based on the Cochrane database of systematic reviews.
Use \code{"Cochrane"} for a prior distribution based on the whole database or call
\code{print(prior_informed_medicine_names)} to inspect the names of
available subfields and set the appropriate \code{parameter} and \code{type};
availability of subfield-specific priors depends on the selected \code{type}.}

\item{parameter}{parameter name describing what prior distribution is supposed to be produced in cases
where the \code{name} corresponds to multiple prior distributions. Relevant only for the empirical medical
prior distributions.}

\item{type}{prior type describing what prior distribution is supposed to be produced in cases
where the \code{name} and \code{parameter} correspond to multiple prior distributions. Relevant only for
the empirical medical prior distributions with the following options
\describe{
\item{\code{"smd"}}{for standardized mean differences}
\item{\code{"logOR"}}{for log odds ratios}
\item{\code{"logRR"}}{for log risk ratios}
\item{\code{"RD"}}{for risk differences}
\item{\code{"logHR"}}{for log hazard ratios}
}}
}
\value{
\code{prior_informed} returns an object of class 'prior'.
}
\description{
\code{prior_informed} creates an informed prior distribution based on past
research. The prior can be visualized by the \code{plot} function.
}
\examples{
# prior distribution representing expected effect sizes in social psychology
# based on prior elicitation with dr. Oosterwijk
p1 <- prior_informed("Oosterwijk")

# the prior distribution can be visualized using the plot function
# (see ?plot.prior for all options)
plot(p1)


# empirical prior distribution for the standardized mean differences from the oral health
# medical subfield based on meta-analytic effect size estimates from the
# Cochrane database of systematic reviews
p2 <- prior_informed("Oral Health", parameter ="effect", type ="smd")
print(p2)

}
\references{
\insertAllCited{}
}
\seealso{
\code{\link[=prior]{prior()}}, \link{prior_informed_medicine_names}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/JAGS-fit.R
\name{JAGS_add_priors}
\alias{JAGS_add_priors}
\title{Add 'JAGS' prior}
\usage{
JAGS_add_priors(syntax, prior_list)
}
\arguments{
\item{syntax}{JAGS model syntax}

\item{prior_list}{named list of prior distribution
(names correspond to the parameter names)}
}
\value{
\code{JAGS_add_priors} returns a JAGS syntax.
}
\description{
Adds priors to a 'JAGS' syntax.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/summary-tables.R
\name{remove_column}
\alias{remove_column}
\title{Removes column to BayesTools table}
\usage{
remove_column(table, column_position = NULL)
}
\arguments{
\item{table}{BayesTools table}

\item{column_position}{position of the to be removed column (defaults to \code{NULL} which
removes the last column)}
}
\value{
returns an object of 'BayesTools_table' class.
}
\description{
Removes column to a BayesTools table while not
breaking formatting, attributes, etc...
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/JAGS-formula.R
\name{transform_scale_samples}
\alias{transform_scale_samples}
\title{Transform standardized posterior samples back to original scale}
\usage{
transform_scale_samples(fit, formula_scale = NULL)
}
\arguments{
\item{fit}{a fitted model object with \code{formula_scale} attribute, or
a matrix of posterior samples}

\item{formula_scale}{nested list containing standardization information keyed by
parameter name. Each parameter entry contains scaling info (mean and sd) for
each standardized predictor, e.g., \code{list(mu = list(mu_x1 = list(mean = 0, sd = 1)))}.
If \code{fit} is provided and has a \code{formula_scale} attribute, this will be used automatically.}
}
\value{
\code{transform_scale_samples} returns posterior samples transformed
back to the original predictor scale.
}
\description{
Transforms posterior samples from standardized continuous
predictors back to the original scale. This function is used when predictors
were standardized during model fitting via the \code{formula_scale} parameter.
}
\details{
The function transforms regression coefficients and intercepts
to account for predictor standardization using a combinatorial approach that
correctly handles interactions of any order.

For a k-way interaction between standardized predictors, the expansion of
\eqn{\prod_{i} (x_i - \mu_i)/\sigma_i} contributes to all lower-order terms.
The contribution to a target term T from a source term S (where T is a subset
of S's scaled components) is:
\deqn{(-1)^{|extra|} \cdot \prod_{i \in extra} \mu_i / \prod_{i \in S_{scaled}} \sigma_i}
where \eqn{extra = S_{scaled} \setminus T_{scaled}}.
}
\seealso{
\code{\link[=JAGS_formula]{JAGS_formula()}} \code{\link[=JAGS_fit]{JAGS_fit()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/summary-tables.R
\name{print.BayesTools_table}
\alias{print.BayesTools_table}
\title{Print a BayesTools table}
\usage{
\method{print}{BayesTools_table}(x, ...)
}
\arguments{
\item{x}{a BayesTools_values_tables}

\item{...}{additional arguments.}
}
\value{
\code{print.BayesTools_table} returns \code{NULL}.
}
\description{
Print a BayesTools table
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/interpret.R
\name{interpret_records}
\alias{interpret_records}
\alias{interpret_tables}
\title{Normalize structured interpretation sources}
\usage{
interpret_records(
  sources,
  plan,
  output = c("records", "text"),
  missing = c("error", "skip", "warn"),
  method = NULL,
  digits = 3
)

interpret_tables(sources, spec, ...)
}
\arguments{
\item{sources}{named list of interpretation sources. Each source can be a
\code{data.frame}/\code{BayesTools_table}, or a list with entries
\code{data}, optional \code{type}, and optional per-source \code{schema}.
Supported source types are \code{"table"}, \code{"record"}, and
\code{"records"}. Table schemas can define columns such as \code{row},
\code{BF}, \code{central}, \code{lower}, \code{upper}, and metadata such as
\code{BF_orientation}, \code{BF_scale}, \code{BF_name},
\code{BF_bound_operator}, \code{lower_prob}, \code{upper_prob},
\code{interval_level}, \code{units}, and \code{conditioning}.}

\item{plan}{ordered list specifying which records to create. Plan items can
have \code{kind = "evidence"}, \code{"estimate"}, \code{"pair"},
\code{"test_estimate"}, \code{"for_each"}, \code{"header"}, \code{"note"},
\code{"prior"}, or \code{"footnote"}. Pair items contain \code{evidence}
and/or \code{estimate} references. Dynamic \code{"for_each"} items generate
records from source row order.}

\item{output}{whether to return normalized \code{"records"} or a simple
generic \code{"text"} rendering. Downstream packages should usually request
\code{"records"} and render domain-specific text themselves.}

\item{missing}{how missing optional sources or rows should be handled.}

\item{method}{optional method name used only by the generic text renderer.}

\item{digits}{number of digits used only by the generic text renderer.}

\item{spec}{alias for \code{plan} used by \code{interpret_tables}.}

\item{...}{additional arguments passed from \code{interpret_tables} to
\code{interpret_records}.}
}
\value{
\code{interpret_records} returns a data frame with class
\code{"BayesTools_interpret_records"} when \code{output = "records"}, or a
character vector when \code{output = "text"}. \code{interpret_tables} is a
convenience wrapper around \code{interpret_records}.
}
\description{
\code{interpret_records} normalizes interpretation inputs from
one or more structured sources into traceable evidence, estimate, header,
note, prior, and footnote records. It is intended for downstream packages
that already own the domain-specific ordering and wording of an
interpretation, but want BayesTools to consistently extract table rows,
preserve Bayes-factor metadata, and carry traceability information.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model-averaging.R
\name{as_mixed_posteriors}
\alias{as_mixed_posteriors}
\title{Export BayesTools JAGS model posterior distribution as model-average posterior distributions via \code{mix_posteriors}}
\usage{
as_mixed_posteriors(
  model,
  parameters,
  conditional = NULL,
  conditional_rule = "AND",
  force_plots = FALSE,
  transform_scaled = FALSE,
  n_prior_samples = 10000
)
}
\arguments{
\item{model}{model fit via the \link{JAGS_fit} function}

\item{parameters}{vector of parameters names for which inference should
be drawn}

\item{conditional}{a character vector of parameters to be conditioned on}

\item{conditional_rule}{a character string specifying the rule for conditioning.
Either "AND" or "OR". Defaults to "AND".}

\item{force_plots}{temporal argument allowing to generate conditional posterior samples
suitable for prior and posterior plots. Only available when conditioning on a
single parameter.}

\item{transform_scaled}{whether to transform samples from standardized (scaled) to
original (unscaled) scale. When \code{TRUE}, posterior samples are
transformed, and the result can be directly passed to \link{plot_posterior} which will
automatically detect the transformation and use transformed deterministic prior densities.
Requires a model fitted with \code{formula_scale_list}. Defaults to \code{FALSE}.}

\item{n_prior_samples}{controls the numerical grid used for transformed
prior densities when \code{transform_scaled = TRUE}. Defaults to 10000.}
}
\value{
\code{as_mix_posteriors} returns a named list of mixed posterior
distributions (either a vector of matrix).
}
\description{
Creates a model-averages posterior distributions on a single
model that allows mimicking the \link{mix_posteriors} functionality. This function
is useful when the model-averaged ensemble is based on \link{prior_spike_and_slab}
or \link{prior_mixture} priors - the model-averaging is done within the model.
}
\seealso{
\link{mix_posteriors}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/JAGS-marglik.R
\name{bridgesampling_object}
\alias{bridgesampling_object}
\title{Create a 'bridgesampling' object}
\usage{
bridgesampling_object(logml = -Inf)
}
\arguments{
\item{logml}{log marginal likelihood. Defaults to \code{-Inf}.}
}
\value{
\code{JAGS_bridgesampling} returns an object of class 'bridge'.
}
\description{
prepares a 'bridgesampling' object with a given
log marginal likelihood.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/summary-tables.R
\name{BayesTools_model_tables}
\alias{BayesTools_model_tables}
\alias{model_summary_table}
\alias{runjags_estimates_table}
\alias{runjags_inference_table}
\alias{JAGS_estimates_table}
\alias{JAGS_inference_table}
\alias{JAGS_summary_table}
\alias{model_summary_empty_table}
\alias{runjags_estimates_empty_table}
\alias{runjags_inference_empty_table}
\alias{JAGS_estimates_empty_table}
\alias{JAGS_inference_empty_table}
\alias{stan_estimates_table}
\title{Create BayesTools model tables}
\usage{
model_summary_table(
  model,
  model_description = NULL,
  title = NULL,
  footnotes = NULL,
  warnings = NULL,
  remove_spike_0 = TRUE,
  short_name = FALSE,
  formula_prefix = TRUE,
  remove_parameters = NULL
)

runjags_estimates_table(
  fit,
  transformations = NULL,
  title = NULL,
  footnotes = NULL,
  warnings = NULL,
  conditional = FALSE,
  probs = c(0.025, 0.5, 0.975),
  remove_spike_0 = TRUE,
  transform_factors = FALSE,
  transform_orthonormal = FALSE,
  formula_prefix = TRUE,
  remove_inclusion = FALSE,
  remove_parameters = NULL,
  remove_formulas = NULL,
  keep_parameters = NULL,
  keep_formulas = NULL,
  return_samples = FALSE,
  transform_scaled = FALSE,
  remove_diagnostics = FALSE,
  diagnostic_columns = getOption("BayesTools.JAGS_estimates_diagnostic_columns", if
    (remove_diagnostics) "none" else "all")
)

runjags_inference_table(
  fit,
  title = NULL,
  footnotes = NULL,
  warnings = NULL,
  formula_prefix = TRUE,
  logBF = FALSE,
  BF01 = FALSE,
  BF_diagnostics = FALSE,
  BF_diagnostic_columns = getOption("BayesTools.JAGS_BF_diagnostic_columns", if
    (BF_diagnostics) "all" else "none")
)

JAGS_estimates_table(
  fit,
  transformations = NULL,
  title = NULL,
  footnotes = NULL,
  warnings = NULL,
  conditional = FALSE,
  probs = c(0.025, 0.5, 0.975),
  remove_spike_0 = TRUE,
  transform_factors = FALSE,
  transform_orthonormal = FALSE,
  formula_prefix = TRUE,
  remove_inclusion = FALSE,
  remove_parameters = NULL,
  remove_formulas = NULL,
  keep_parameters = NULL,
  keep_formulas = NULL,
  return_samples = FALSE,
  transform_scaled = FALSE,
  remove_diagnostics = FALSE,
  diagnostic_columns = getOption("BayesTools.JAGS_estimates_diagnostic_columns", if
    (remove_diagnostics) "none" else "all")
)

JAGS_inference_table(
  fit,
  title = NULL,
  footnotes = NULL,
  warnings = NULL,
  formula_prefix = TRUE,
  logBF = FALSE,
  BF01 = FALSE,
  BF_diagnostics = FALSE,
  BF_diagnostic_columns = getOption("BayesTools.JAGS_BF_diagnostic_columns", if
    (BF_diagnostics) "all" else "none")
)

JAGS_summary_table(
  model,
  model_description = NULL,
  title = NULL,
  footnotes = NULL,
  warnings = NULL,
  remove_spike_0 = TRUE,
  short_name = FALSE,
  formula_prefix = TRUE,
  remove_parameters = NULL
)

model_summary_empty_table(
  model_description = NULL,
  title = NULL,
  footnotes = NULL,
  warnings = NULL
)

runjags_estimates_empty_table(
  probs = c(0.025, 0.5, 0.975),
  title = NULL,
  footnotes = NULL,
  warnings = NULL,
  remove_diagnostics = FALSE,
  diagnostic_columns = getOption("BayesTools.JAGS_estimates_diagnostic_columns", if
    (remove_diagnostics) "none" else "all")
)

runjags_inference_empty_table(
  title = NULL,
  footnotes = NULL,
  warnings = NULL,
  logBF = FALSE,
  BF01 = FALSE,
  BF_diagnostics = FALSE,
  BF_diagnostic_columns = getOption("BayesTools.JAGS_BF_diagnostic_columns", if
    (BF_diagnostics) "all" else "none")
)

JAGS_estimates_empty_table(
  probs = c(0.025, 0.5, 0.975),
  title = NULL,
  footnotes = NULL,
  warnings = NULL,
  remove_diagnostics = FALSE,
  diagnostic_columns = getOption("BayesTools.JAGS_estimates_diagnostic_columns", if
    (remove_diagnostics) "none" else "all")
)

JAGS_inference_empty_table(
  title = NULL,
  footnotes = NULL,
  warnings = NULL,
  logBF = FALSE,
  BF01 = FALSE,
  BF_diagnostics = FALSE,
  BF_diagnostic_columns = getOption("BayesTools.JAGS_BF_diagnostic_columns", if
    (BF_diagnostics) "all" else "none")
)

stan_estimates_table(
  fit,
  transformations = NULL,
  title = NULL,
  footnotes = NULL,
  warnings = NULL
)
}
\arguments{
\item{model}{model object containing a list of \code{priors}
and \code{inference} object, The \code{inference} must be a
named list with information about the model: model number
\code{m_number}, marginal likelihood \code{marglik}, prior and
posterior probability \code{prior_prob} and \code{post_prob},
and model inclusion Bayes factor \code{inclusion_BF}}

\item{model_description}{named list with additional description
to be added to the table}

\item{title}{title to be added to the table}

\item{footnotes}{footnotes to be added to the table}

\item{warnings}{warnings to be added to the table}

\item{remove_spike_0}{whether prior distributions equal to spike
at 0 should be removed from the \code{prior_list}}

\item{short_name}{whether the prior distribution names should be
shortened. Defaults to \code{FALSE}.}

\item{formula_prefix}{whether the parameter prefix from formula should
be printed. Defaults to \code{TRUE}.}

\item{remove_parameters}{parameters to be removed from the summary.
Can be \code{NULL} (default, no removal), a character vector of parameter
names to remove, or \code{TRUE} to remove all parameters that are not
part of any formula.}

\item{fit}{runjags model fit}

\item{transformations}{named list of transformations to be applied
to specific parameters}

\item{conditional}{summarizes estimates conditional on being included
in the model for spike and slab priors. Defaults to \code{FALSE}.}

\item{probs}{quantiles for parameter estimates}

\item{transform_factors}{whether factors with orthonormal/meandif
prior distribution should be transformed to differences from the
grand mean}

\item{transform_orthonormal}{(to be depreciated) whether factors
with orthonormal prior distributions should be transformed to
differences from the grand mean}

\item{remove_inclusion}{whether estimates of the inclusion probabilities
should be excluded from the summary table. Defaults to \code{FALSE}.}

\item{remove_formulas}{character vector of formula names whose parameters
should be removed from the summary. Defaults to \code{NULL}.}

\item{keep_parameters}{character vector of parameter names to keep.
All other parameters will be removed unless they belong to formulas
specified in \code{keep_formulas}. Defaults to \code{NULL}.}

\item{keep_formulas}{character vector of formula names whose parameters
should be kept. All other parameters will be removed unless they are
specified in \code{keep_parameters}. Defaults to \code{NULL}.}

\item{return_samples}{whether to return the transoformed and formated samples
instead of the table. Defaults to \code{FALSE}.}

\item{transform_scaled}{whether coefficients from standardized
continuous predictors should be transformed back to the original
scale. Defaults to \code{FALSE}.}

\item{remove_diagnostics}{whether to exclude MCMC diagnostics (MCMC error,
ESS, R-hat) from the output table. Defaults to \code{FALSE}. Setting to
\code{TRUE} will exclude diagnostics columns regardless of the
\code{conditional} setting.}

\item{diagnostic_columns}{MCMC diagnostic columns to display in JAGS
estimates tables. Can be \code{"all"}, \code{"none"}, \code{TRUE},
\code{FALSE}, or a character vector containing any subset of
\code{"MCMC_error"}, \code{"MCMC_SD_error"}, \code{"ESS"}, and
\code{"R_hat"}. Defaults to the
\code{BayesTools.JAGS_estimates_diagnostic_columns} option, or all
diagnostics unless \code{remove_diagnostics = TRUE}.}

\item{logBF}{whether the Bayes factor should be on log scale}

\item{BF01}{whether the Bayes factor should be inverted}

\item{BF_diagnostics}{whether to add MCMC diagnostics for Bayes factors
computed from model indicator frequencies. The Bayes factor error is
reported as a relative Monte Carlo standard error percentage. Defaults to
\code{FALSE}.}

\item{BF_diagnostic_columns}{MCMC diagnostic columns to display in JAGS
inclusion Bayes factor tables. Can be \code{"all"}, \code{"none"},
\code{TRUE}, \code{FALSE}, or a character vector containing any subset of
\code{"ESS"}, \code{"MCMC_error"}, and \code{"BF_error_percent"}.
Defaults to the \code{BayesTools.JAGS_BF_diagnostic_columns} option, or all
diagnostics when \code{BF_diagnostics = TRUE} and none otherwise.}
}
\value{
\code{model_summary_table} returns a table with
overview of the fitted model, \code{runjags_estimates_table} returns
a table with MCMC estimates, and \code{runjags_estimates_empty_table}
returns an empty estimates table. All of the tables are objects of
class 'BayesTools_table'.
}
\description{
Creates model summary based on a model objects or
provides estimates table for a runjags fit.
}
\details{
For product-space JAGS inclusion Bayes factors, posterior inclusion
probabilities of exactly 0 or 1 cannot produce a finite point estimate of
the model odds ratio. In that case, inclusion BFs marked with \code{"<"} or
\code{">"} are finite-sample bounds: posterior inclusion probabilities of
0 or 1 were replaced by \code{1/S} or \code{(S - 1)/S}, where \code{S} is
the number of posterior samples. This is a reporting convention, not an
unbiased finite Bayes-factor estimate. If the prior inclusion probability
is exactly 0 or 1, the inclusion Bayes factor is undefined and reported as
\code{NA}, because the corresponding inclusion/exclusion comparison was not
tested.
}
\seealso{
\link{BayesTools_ensemble_tables}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/priors.R
\name{var}
\alias{var}
\title{Creates generic for var function}
\usage{
var(x, ...)
}
\arguments{
\item{x}{main argument}

\item{...}{additional arguments}
}
\value{
\code{var} returns a variance
of the supplied object (if it is either a numeric vector
or an object of class 'prior').
}
\description{
Creates generic for var function
}
\seealso{
\link[stats]{cor}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/selection-kernels.R
\name{prior_bias}
\alias{prior_bias}
\title{Creates a composed publication-bias prior}
\usage{
prior_bias(selection = NULL, phacking = NULL, prior_weights = 1)
}
\arguments{
\item{selection}{optional \code{prior_weightfunction()} object.}

\item{phacking}{optional \code{prior_phacking()} object.}

\item{prior_weights}{prior odds associated with a given distribution.}
}
\value{
\code{prior_bias()} returns an object of class \code{"prior"}.
}
\description{
\code{prior_bias()} composes a step-selection
\code{prior_weightfunction()} and/or a \code{prior_phacking()} object. It does
not introduce new math; the backend compiler treats the composition as the
product of both kernels.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model-averaging-plots.R
\name{plot_marginal}
\alias{plot_marginal}
\title{Plot samples from the marginal posterior distributions}
\usage{
plot_marginal(
  samples,
  parameter,
  plot_type = "base",
  prior = FALSE,
  n_points = 1000,
  transformation = NULL,
  transformation_arguments = NULL,
  transformation_settings = FALSE,
  rescale_x = FALSE,
  par_name = NULL,
  dots_prior = list(),
  legend = TRUE,
  legend_title = NULL,
  legend_labels = NULL,
  legend_position = NULL,
  ...
)
}
\arguments{
\item{samples}{samples from a posterior distribution for a
parameter generated by \link{marginal_inference}.}

\item{parameter}{parameter name to be plotted.}

\item{plot_type}{whether to use a base plot \code{"base"}
or ggplot2 \code{"ggplot"} for plotting.}

\item{prior}{whether prior distribution should be added to the figure.
When samples were prepared with \code{as_mixed_posteriors(..., transform_scaled = TRUE)},
the transformed prior samples are automatically used.}

\item{n_points}{number of equally spaced points
in the \code{x_range} if \code{x_seq} is unspecified}

\item{transformation}{transformation to be applied
to the prior distribution. Either a character
specifying one of the prepared transformations:
\describe{
\item{lin}{linear transformation in form of \code{a + b*x}}
\item{tanh}{hyperbolic tangent transformation}
\item{exp}{exponential transformation}
}, or a list containing the transformation function \code{fun},
inverse transformation function \code{inv}, and derivative of the
transformation \code{jac}, evaluated on the original support. See examples
for details.}

\item{transformation_arguments}{a list with named arguments for
the \code{transformation}}

\item{transformation_settings}{boolean indicating whether the
settings the \code{x_seq} or \code{x_range} was specified on
the transformed support}

\item{rescale_x}{allows to rescale x-axis in case a
weightfunction is plotted.}

\item{par_name}{a type of parameter for which the prior is
specified. Only relevant if the prior corresponds to a mu
parameter that needs to be transformed.}

\item{dots_prior}{additional arguments for the prior distribution plot}

\item{legend}{whether factor legends should be drawn.}

\item{legend_title}{optional title for factor legends.}

\item{legend_labels}{optional labels for factor legend levels.}

\item{legend_position}{optional legend position for factor legends.}

\item{...}{additional arguments}
}
\value{
\code{plot_marginal} returns either \code{NULL} or
an object of class 'ggplot' if plot_type is \code{plot_type = "ggplot"}.
}
\description{
Plot samples from the marginal posterior distributions
}
\seealso{
\code{\link[=prior]{prior()}} \code{\link[=marginal_inference]{marginal_inference()}}  \code{\link[=plot_posterior]{plot_posterior()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/distributions-mpoint.R
\name{mpoint}
\alias{mpoint}
\alias{dmpoint}
\alias{rmpoint}
\alias{pmpoint}
\alias{qmpoint}
\title{Multivariate point mass distribution}
\usage{
dmpoint(x, location, log = FALSE)

rmpoint(n, location)

pmpoint(q, location, lower.tail = TRUE, log.p = FALSE)

qmpoint(p, location, lower.tail = TRUE, log.p = FALSE)
}
\arguments{
\item{x, q}{vector or matrix of quantiles.}

\item{location}{vector of locations corresponding to the location of individual points.
Alternatively, a matrix with rows corresponding to the location of individual samples
and columns correspond to the location of individual points.}

\item{log, log.p}{logical; if \code{TRUE}, probabilities
\code{p} are given as \code{log(p)}.}

\item{n}{number of observations.}

\item{lower.tail}{logical; if \code{TRUE} (default), probabilities
are \eqn{P[X \le x]}, otherwise, \eqn{P[X > x]}.}

\item{p}{vector of probabilities.}
}
\value{
\code{dpoint} gives the density, \code{ppoint} gives the
distribution function, \code{qpoint} gives the quantile function,
and \code{rpoint} generates random deviates.
}
\description{
Density, distribution function,
quantile function and random generation for multivariate point distribution.
}
\examples{
# draw samples from a multivariate point distribution
rmpoint(10, location = c(0, 1))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/priors-plot.R
\name{plot.prior}
\alias{plot.prior}
\title{Plots a prior object}
\usage{
\method{plot}{prior}(
  x,
  plot_type = "base",
  x_seq = NULL,
  xlim = NULL,
  x_range_quant = NULL,
  n_points = 1000,
  n_samples = 10000,
  force_samples = FALSE,
  transformation = NULL,
  transformation_arguments = NULL,
  transformation_settings = FALSE,
  show_figures = if (individual) -1 else NULL,
  individual = FALSE,
  rescale_x = FALSE,
  par_name = NULL,
  ...
)
}
\arguments{
\item{x}{a prior}

\item{plot_type}{whether to use a base plot \code{"base"}
or ggplot2 \code{"ggplot"} for plotting.}

\item{x_seq}{sequence of x coordinates}

\item{xlim}{x plotting range}

\item{x_range_quant}{quantile used for
automatically obtaining \code{x_range}
if both \code{x_range} and \code{x_seq}
are unspecified. Defaults to \code{0.005}
for all but Cauchy, Student-t, Gamma, and
Inverse-gamme distributions that use
\code{0.010}.}

\item{n_points}{number of equally spaced points
in the \code{x_range} if \code{x_seq} is unspecified}

\item{n_samples}{number of samples from the prior
distribution if the density cannot be obtained
analytically (or if samples are forced with
\code{force_samples = TRUE})}

\item{force_samples}{should prior be sampled instead
of obtaining analytic solution whenever possible}

\item{transformation}{transformation to be applied
to the prior distribution. Either a character
specifying one of the prepared transformations:
\describe{
\item{lin}{linear transformation in form of \code{a + b*x}}
\item{tanh}{hyperbolic tangent transformation}
\item{exp}{exponential transformation}
}, or a list containing the transformation function \code{fun},
inverse transformation function \code{inv}, and derivative of the
transformation \code{jac}, evaluated on the original support. See examples
for details.}

\item{transformation_arguments}{a list with named arguments for
the \code{transformation}}

\item{transformation_settings}{boolean indicating whether the
settings the \code{x_seq} or \code{x_range} was specified on
the transformed support}

\item{show_figures}{which figures should be returned in case of
multiple plots are generated. Useful when priors for the omega
parameter are plotted and \code{individual = TRUE}.}

\item{individual}{should individual densities be returned
(e.g., in case of weightfunction)}

\item{rescale_x}{allows to rescale x-axis in case a
weightfunction is plotted.}

\item{par_name}{a type of parameter for which the prior is
specified. Only relevant if the prior corresponds to a mu
parameter that needs to be transformed.}

\item{...}{additional arguments}
}
\value{
\code{plot.prior} returns either \code{NULL} or
an object of class 'ggplot' if plot_type is \code{plot_type = "ggplot"}.
}
\description{
Plots a prior object
}
\examples{
# create some prior distributions
p0 <- prior(distribution = "point",  parameters = list(location = 0))
p1 <- prior(distribution = "normal", parameters = list(mean = 0, sd = 1))
p2 <- prior(distribution = "normal", parameters = list(mean = 0, sd = 1), truncation = list(0, Inf))

# a default plot
plot(p0)

# manipulate line thickness and color, change the parameter name
plot(p1, lwd = 2, col = "blue", par_name = bquote(mu))

# use ggplot
plot(p2, plot_type = "ggplot")

# utilize the ggplot prior geom
plot(p2, plot_type = "ggplot", xlim = c(-2, 2)) + geom_prior(p1, col = "red", lty = 2)

# apply transformation
plot(p1, transformation = "exp")

}
\seealso{
\code{\link[=prior]{prior()}} \code{\link[=lines.prior]{lines.prior()}}  \code{\link[=geom_prior]{geom_prior()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/JAGS-marglik.R
\name{JAGS_bridgesampling_posterior}
\alias{JAGS_bridgesampling_posterior}
\title{Prepare 'JAGS' posterior for 'bridgesampling'}
\usage{
JAGS_bridgesampling_posterior(
  posterior,
  prior_list,
  add_parameters = NULL,
  add_bounds = NULL
)
}
\arguments{
\item{posterior}{matrix of mcmc samples from the posterior
distribution}

\item{prior_list}{named list of prior distribution
(names correspond to the parameter names) of parameters not specified within the
\code{formula_list}}

\item{add_parameters}{vector of additional parameter names that should be used
in bridgesampling but were not specified in the \code{prior_list}}

\item{add_bounds}{list with two name vectors (\code{"lb"} and \code{"up"})
containing lower and upper bounds of the additional parameters that were not
specified in the \code{prior_list}}
}
\value{
\code{JAGS_bridgesampling_posterior} returns a matrix of
posterior samples with 'lb' and 'ub' attributes carrying the
lower and upper boundaries.
}
\description{
prepares posterior distribution for 'bridgesampling'
by removing unnecessary parameters and attaching lower and upper
bounds of parameters based on a list of prior distributions.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/priors-tools.R
\name{is.prior}
\alias{is.prior}
\alias{is.prior.point}
\alias{is.prior.none}
\alias{is.prior.simple}
\alias{is.prior.discrete}
\alias{is.prior.vector}
\alias{is.prior.PET}
\alias{is.prior.PEESE}
\alias{is.prior.weightfunction}
\alias{is.prior.factor}
\alias{is.prior.orthonormal}
\alias{is.prior.treatment}
\alias{is.prior.independent}
\alias{is.prior.spike_and_slab}
\alias{is.prior.meandif}
\alias{is.prior.mixture}
\title{Reports whether x is a a prior object}
\usage{
is.prior(x)

is.prior.point(x)

is.prior.none(x)

is.prior.simple(x)

is.prior.discrete(x)

is.prior.vector(x)

is.prior.PET(x)

is.prior.PEESE(x)

is.prior.weightfunction(x)

is.prior.factor(x)

is.prior.orthonormal(x)

is.prior.treatment(x)

is.prior.independent(x)

is.prior.spike_and_slab(x)

is.prior.meandif(x)

is.prior.mixture(x)
}
\arguments{
\item{x}{an object of test}
}
\value{
returns a boolean indicating whether the test object
is a prior (of specific type).
}
\description{
Reports whether x is a a prior object. Note that
point priors inherit the prior.simple property
}
\examples{
# create some prior distributions
p0 <- prior(distribution = "point",  parameters = list(location = 0))
p1 <- prior_PET(distribution = "normal", parameters = list(mean = 0, sd = 1))

is.prior(p0)
is.prior.simple(p0)
is.prior.point(p0)
is.prior.PET(p0)

is.prior(p1)
is.prior.simple(p1)
is.prior.point(p1)
is.prior.PET(p1)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/JAGS-formula.R
\name{JAGS_formula}
\alias{JAGS_formula}
\title{Create JAGS formula syntax and data object}
\usage{
JAGS_formula(formula, parameter, data, prior_list, formula_scale = NULL)
}
\arguments{
\item{formula}{formula specifying the right hand side of the assignment (the
left hand side is ignored). If the formula contains \code{-1}, it will be
automatically converted to include an intercept with a spike(0) prior.
The formula can also have a \code{"log(intercept)"} attribute set to \code{TRUE}
to generate syntax of the form \code{log(intercept) + sum(beta_i * x_i)}, which
is useful for parameters that must be positive (e.g., standard deviation).}

\item{parameter}{name of the parameter to be created with the formula}

\item{data}{data.frame containing predictors included in the formula}

\item{prior_list}{named list of prior distribution of parameters specified within
the \code{formula}. When using \code{-1} in the formula, an "intercept" prior
can be explicitly specified; otherwise, \code{prior("spike", list(0))} is
automatically added. The list can also include two special entries:}

\item{formula_scale}{named list specifying whether to standardize continuous predictors.
If \code{NULL} (default), no standardization is applied. If a named list is provided,
continuous predictors with \code{TRUE} values will be standardized (mean-centered and
scaled by standard deviation). The intercept is never standardized.
\describe{
\item{\code{"__default_continuous"}}{A prior to use for any continuous predictors
(including the intercept) that are not explicitly specified in the prior list.}
\item{\code{"__default_factor"}}{A prior to use for any factor predictors
(including interactions involving factors) that are not explicitly specified
in the prior list.}
}
These default priors allow for more concise specification when many predictors
share the same prior distribution.}
}
\value{
\code{JAGS_formula} returns a list containing the formula JAGS syntax,
JAGS data object, modified prior_list, and (if standardization was applied) a
\code{formula_scale} list with standardization information for back-transformation.
}
\description{
Creates a JAGS formula syntax, prepares data input, and
returns modified prior list for further processing in the \code{JAGS_fit}
function.
}
\details{
When a formula with \code{-1} (no intercept) is specified, the
function automatically removes the \code{-1}, adds an intercept back to the
formula, and includes a spike(0) prior for the intercept to ensure equivalent
model behavior while maintaining consistent formula parsing.

When using default priors (\code{"__default_continuous"} or \code{"__default_factor"}),
explicitly specified priors for individual terms take precedence over the defaults.
The defaults are only applied to terms that are not already in the prior list.
}
\examples{
# simulate data
set.seed(1)
df <- data.frame(
  y      = rnorm(60),
  x_cont = rnorm(60),
  x_bin  = rbinom(60, 1, .5),
  x_fac3 = factor(rep(c("A", "B", "C"), 20), levels = c("A", "B", "C")),
  x_fac4 = factor(rep(c("A", "B", "C", "D"), 15), levels = c("A", "B", "C", "D"))
)

# specify priors with intercept
prior_list <- list(
"intercept"     = prior("normal", list(0, 1)),
"x_cont"        = prior("normal", list(0, .5)),
"x_fac3"        = prior_factor("normal",  list(0, 1),  contrast = "treatment"),
"x_fac4"        = prior_factor("mnormal", list(0, 1),  contrast = "orthonormal"),
"x_fac3:x_fac4" = prior_factor("mnormal", list(0, .5), contrast = "orthonormal")
)

# create the formula object
formula_obj <- JAGS_formula(
  formula = ~ x_cont + x_fac3 * x_fac4,
  parameter = "mu", data = df, prior_list = prior_list)

# using -1 notation (automatically adds spike(0) intercept)
prior_list_no_intercept <- list(
  "x_fac3" = prior_factor("normal", list(0, 1), contrast = "treatment")
)
formula_no_intercept <- JAGS_formula(
  formula = ~ x_fac3 - 1,
  parameter = "mu", data = df, prior_list = prior_list_no_intercept)
# Equivalent to specifying intercept = prior("spike", list(0))

# using default priors for simpler specification
prior_list_defaults <- list(
  "__default_continuous" = prior("normal", list(0, 1)),
  "__default_factor"     = prior_factor("normal", list(0, 0.5), contrast = "treatment")
)
formula_defaults <- JAGS_formula(
  formula = ~ x_cont + x_fac3,
  parameter = "mu", data = df, prior_list = prior_list_defaults)
# intercept and x_cont get the default continuous prior
# x_fac3 gets the default factor prior

}
\seealso{
\code{\link[=JAGS_fit]{JAGS_fit()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/marginal-distributions.R
\name{as_marginal_inference}
\alias{as_marginal_inference}
\title{Model-average marginal posterior distributions and
marginal Bayes factors based on BayesTools JAGS model via \code{marginal_inference}}
\usage{
as_marginal_inference(
  model,
  marginal_parameters,
  parameters,
  conditional_list,
  conditional_rule,
  formula,
  null_hypothesis = 0,
  normal_approximation = FALSE,
  n_samples = 10000,
  silent = FALSE,
  force_plots = FALSE
)
}
\arguments{
\item{model}{model fit via the \link{JAGS_fit} function}

\item{marginal_parameters}{parameters for which the the marginal summary
should be created}

\item{parameters}{all parameters included in the model_list that are
relevant for the formula (all of which need to have specification of
\code{is_null_list})}

\item{conditional_list}{list of conditional parameters for each marginal parameter}

\item{conditional_rule}{a character string specifying the rule for conditioning.
Either "AND" or "OR". Defaults to "AND".}

\item{formula}{model formula (needs to be specified if \code{parameter} was part of a formula)}

\item{null_hypothesis}{point null hypothesis to test. Defaults to \code{0}}

\item{normal_approximation}{whether the height of prior and posterior density should be
approximated via a normal distribution (rather than kernel density). Defaults to \code{FALSE}.}

\item{n_samples}{controls the numerical grid used for model-averaged
prior densities}

\item{silent}{whether warnings should be returned silently. Defaults to \code{FALSE}}

\item{force_plots}{temporal argument allowing to generate conditional posterior samples
suitable for prior and posterior plots. Only available when conditioning on a
single parameter.}
}
\value{
\code{as_marginal_inference} returns an object of class 'marginal_inference'.
}
\description{
Creates marginal model-averaged and conditional
posterior distributions based on a BayesTools JAGS model, vector of parameters,
formula, and a list of conditional specifications for each parameter.
Computes inclusion Bayes factors for each marginal estimate via a Savage-Dickey
density approximation.
}
\seealso{
\link{marginal_inference} \link{as_mixed_posteriors}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/priors.R
\name{sd}
\alias{sd}
\title{Creates generic for sd function}
\usage{
sd(x, ...)
}
\arguments{
\item{x}{main argument}

\item{...}{additional arguments}
}
\value{
\code{sd} returns a standard deviation
of the supplied object (if it is either a numeric vector
or an object of class 'prior').
}
\description{
Creates generic for sd function
}
\seealso{
\link[stats]{sd}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/priors-density.R
\name{range.prior}
\alias{range.prior}
\title{Prior range}
\usage{
\method{range}{prior}(x, quantiles = NULL, ..., na.rm = FALSE)
}
\arguments{
\item{x}{a prior}

\item{quantiles}{quantile to be returned in
case of unbounded distribution.}

\item{...}{additional arguments}

\item{na.rm}{unused}
}
\value{
\code{range.prior} returns a numeric vector of
length with a plotting range of a prior distribution.
}
\description{
Computes range of a prior
distribution (if the prior distribution is
unbounded range from \code{quantiles} to
\code{1 -quantiles}) is returned.
}
\seealso{
\code{\link[=prior]{prior()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/summary-tables.R
\name{add_column}
\alias{add_column}
\title{Adds column to BayesTools table}
\usage{
add_column(
  table,
  column_title,
  column_values,
  column_position = NULL,
  column_type = NULL
)
}
\arguments{
\item{table}{BayesTools table}

\item{column_title}{title of the new column}

\item{column_values}{values of the new column}

\item{column_position}{position of the new column (defaults to \code{NULL} which
appends the column to the end)}

\item{column_type}{type of values of the new column table (important for formatting,
defaults to \code{NULL} = the function tries to guess numeric / character based on the
\code{column_values} but many more specific types are available)}
}
\value{
returns an object of 'BayesTools_table' class.
}
\description{
Adds column to a BayesTools table while not
breaking formatting, attributes, etc...
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model-averaging.R
\name{ensemble_inference}
\alias{ensemble_inference}
\alias{compute_inference}
\alias{models_inference}
\title{Compute posterior probabilities and inclusion Bayes factors}
\usage{
compute_inference(prior_weights, margliks, is_null = NULL, conditional = FALSE)

ensemble_inference(model_list, parameters, is_null_list, conditional = FALSE)

models_inference(model_list)
}
\arguments{
\item{prior_weights}{vector of prior model odds}

\item{margliks}{vector of marginal likelihoods}

\item{is_null}{logical vector of indicators specifying whether the model corresponds
to the null or alternative hypothesis (or an integer vector indexing models
corresponding to the null hypothesis)}

\item{conditional}{whether prior and posterior model probabilities should
be returned only for the conditional model. Defaults to \code{FALSE}}

\item{model_list}{list of models, each of which contains marginal
likelihood estimated with bridge sampling \code{marglik} and prior model
odds \code{prior_weights}}

\item{parameters}{vector of parameters names for which inference should
be drawn}

\item{is_null_list}{list with entries for each parameter carrying either
logical vector of indicators specifying whether the model corresponds
to the null or alternative hypothesis (or an integer vector indexing models
corresponding to the null hypothesis)}
}
\value{
\code{compute_inference} returns a named list of prior probabilities,
posterior probabilities, and Bayes factors, \code{ppoint} gives the
distribution function, \code{ensemble_inference} gives a list of named lists of
inferences for each parameter, and \code{models_inference} returns a list of
models, each expanded by the inference list.
}
\description{
Computes prior probabilities, posterior probabilities,
and inclusion Bayes factors based either on (1) a list of models,
vector of parameters, and a list of indicators the models represent the
null or alternative hypothesis for each parameter, (2) on prior
model odds, marginal likelihoods, and indicator whether
the models represent the null or alternative hypothesis, or (3) list of
models for each model.
}
\seealso{
\link{mix_posteriors} \link{BayesTools_ensemble_tables}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/selection-kernels.R
\name{phack_pi_null}
\alias{phack_pi_null}
\alias{phack_alpha_from_pi_null}
\alias{phack_backend_constants}
\title{P-hacking calibration helpers}
\usage{
phack_pi_null(alpha, form, source, destination, target = 0.025)

phack_alpha_from_pi_null(pi_null, form, source, destination, target = 0.025)

phack_backend_constants(form, source, destination, target = 0.025)
}
\arguments{
\item{alpha}{p-hacking severity parameter.}

\item{form}{power depletion form, either \code{"linear"} or
\code{"quadratic"}.}

\item{source}{source p-value cut point.}

\item{destination}{destination p-value cut point.}

\item{target}{target p-value cut point.}

\item{pi_null}{null probability mass depleted from the source interval.}
}
\value{
Numeric vector for the calibration helpers and a named list for
\code{phack_backend_constants()}.
}
\description{
\code{phack_pi_null()} converts the sampled severity
\code{alpha} to the amount of null probability mass depleted from the source
interval. \code{phack_alpha_from_pi_null()} applies the inverse calibration.
\code{phack_backend_constants()} returns the z-scale constants used by the
backend.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/JAGS-marglik.R
\name{JAGS_marglik_priors}
\alias{JAGS_marglik_priors}
\alias{JAGS_marglik_priors_formula}
\title{Compute marginal likelihood for 'JAGS' priors}
\usage{
JAGS_marglik_priors(samples, prior_list)

JAGS_marglik_priors_formula(samples, formula_prior_list)
}
\arguments{
\item{samples}{samples provided by bridgesampling
function}

\item{prior_list}{named list of prior distribution
(names correspond to the parameter names) of parameters not specified within the
\code{formula_list}}

\item{formula_prior_list}{named list of named lists of prior distributions
(names of the lists correspond to the parameter name created by each of the formula and
the names of the prior distribution correspond to the parameter names) of parameters specified
within the \code{formula}}
}
\value{
\code{JAGS_marglik_priors} returns a numeric value
of likelihood evaluated at the current posterior sample.
}
\description{
Computes marginal likelihood for the
prior part of a 'JAGS' model within 'bridgesampling'
function
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/summary-tables.R
\name{update.BayesTools_table}
\alias{update.BayesTools_table}
\title{Updates BayesTools table}
\usage{
\method{update}{BayesTools_table}(
  object,
  title = NULL,
  footnotes = NULL,
  warnings = NULL,
  remove_parameters = NULL,
  logBF = FALSE,
  BF01 = FALSE,
  ...
)
}
\arguments{
\item{object}{a BayesTools table}

\item{title}{title of the table}

\item{footnotes}{add footnotes to the table}

\item{warnings}{add warnings of the table}

\item{remove_parameters}{remove parameters from the table}

\item{logBF}{whether to format Bayes factors as log(BF)}

\item{BF01}{whether to format Bayes factors as 1/BF}

\item{...}{additional arguments.}
}
\value{
returns an object of 'BayesTools_table' class.
}
\description{
Updates BayesTools table while not breaking formatting, attributes, etc...
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/JAGS-fit.R
\name{JAGS_get_inits}
\alias{JAGS_get_inits}
\title{Create initial values for 'JAGS' model}
\usage{
JAGS_get_inits(prior_list, chains, seed)
}
\arguments{
\item{prior_list}{named list of prior distribution
(names correspond to the parameter names)}

\item{chains}{number of chains}

\item{seed}{seed for random number generation}
}
\value{
\code{JAGS_add_priors} returns a list of JAGS
initial values.
}
\description{
Creates initial values for priors in
a 'JAGS' model.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/marginal-distributions.R
\name{Savage_Dickey_BF}
\alias{Savage_Dickey_BF}
\title{Compute Savage-Dickey inclusion Bayes factors}
\usage{
Savage_Dickey_BF(
  posterior,
  null_hypothesis = 0,
  normal_approximation = FALSE,
  silent = FALSE
)
}
\arguments{
\item{posterior}{marginal posterior distribution generated via the
\code{marginal_posterior} function}

\item{null_hypothesis}{point null hypothesis to test. Defaults to \code{0}}

\item{normal_approximation}{whether the height of prior and posterior density should be
approximated via a normal distribution (rather than kernel density). Defaults to \code{FALSE}.}

\item{silent}{whether warnings should be returned silently. Defaults to \code{FALSE}}
}
\value{
\code{Savage_Dickey_BF} returns a Bayes factor.
}
\description{
Computes Savage-Dickey (density ratio) inclusion Bayes factors
based the change of height from prior to posterior distribution at the test value.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/JAGS-fit.R
\name{JAGS_to_monitor}
\alias{JAGS_to_monitor}
\title{Create list of monitored parameters for 'JAGS' model}
\usage{
JAGS_to_monitor(prior_list)
}
\arguments{
\item{prior_list}{named list of prior distribution
(names correspond to the parameter names)}
}
\value{
\code{JAGS_to_monitor} returns a character vector of
parameter names.
}
\description{
Creates a vector of parameter names to be
monitored in a 'JAGS' model.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model-averaging.R
\name{weightfunctions_mapping}
\alias{weightfunctions_mapping}
\title{Create coefficient mapping between multiple weightfunctions}
\usage{
weightfunctions_mapping(prior_list, cuts_only = FALSE, one_sided = FALSE)
}
\arguments{
\item{prior_list}{list of prior distributions}

\item{cuts_only}{whether only p-value cuts should be returned}

\item{one_sided}{force one-sided output}
}
\value{
\code{weightfunctions_mapping} returns a list of indices
mapping the publication weights omega from the individual weightfunctions
into a joint weightfunction.
}
\description{
Creates coefficients mapping between multiple weightfunctions.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/JAGS-marglik.R
\name{JAGS_bridgesampling}
\alias{JAGS_bridgesampling}
\title{Compute marginal likelihood of a 'JAGS' model}
\usage{
JAGS_bridgesampling(
  fit,
  log_posterior,
  data = NULL,
  prior_list = NULL,
  formula_list = NULL,
  formula_data_list = NULL,
  formula_prior_list = NULL,
  formula_scale_list = NULL,
  add_parameters = NULL,
  add_bounds = NULL,
  maxiter = 10000,
  silent = TRUE,
  ...
)
}
\arguments{
\item{fit}{model fitted with either \link[runjags]{runjags} posterior
samples obtained with \link[rjags]{rjags-package}}

\item{log_posterior}{function that takes a named list of samples, the data,
and additional list of parameters passed as \code{...} as input and
returns the log of the unnormalized posterior density of the model part}

\item{data}{list containing data to fit the model (not including data for the formulas)}

\item{prior_list}{named list of prior distribution
(names correspond to the parameter names) of parameters not specified within the
\code{formula_list}}

\item{formula_list}{named list of formulas to be added to the model
(names correspond to the parameter name created by each of the formula)}

\item{formula_data_list}{named list of data frames containing data for each formula
(names of the lists correspond to the parameter name created by each of the formula)}

\item{formula_prior_list}{named list of named lists of prior distributions
(names of the lists correspond to the parameter name created by each of the formula and
the names of the prior distribution correspond to the parameter names) of parameters specified
within the \code{formula}}

\item{formula_scale_list}{named list of named lists for standardizing continuous predictors
(names of the lists correspond to the parameter name created by each of the formula).
Each entry should be a named list where continuous predictors with \code{TRUE} values will
be standardized. Defaults to \code{NULL} (no standardization).}

\item{add_parameters}{vector of additional parameter names that should be used
in bridgesampling but were not specified in the \code{prior_list}}

\item{add_bounds}{list with two name vectors (\code{"lb"} and \code{"up"})
containing lower and upper bounds of the additional parameters that were not
specified in the \code{prior_list}}

\item{maxiter}{maximum number of iterations for the
\link[bridgesampling]{bridge_sampler}}

\item{silent}{whether the progress should be printed, defaults to \code{TRUE}}

\item{...}{additional argument to the \link[bridgesampling]{bridge_sampler}
and \code{log_posterior} function}
}
\value{
\code{JAGS_bridgesampling} returns an object of class 'bridge'.
}
\description{
A wrapper around
\link[bridgesampling]{bridge_sampler} that automatically
computes likelihood part dependent on the prior distribution
and prepares parameter samples. \code{log_posterior} must
specify a function that takes two arguments - a named list
of samples from the prior distributions and the data, and returns
log likelihood of the model part.
}
\examples{
\dontrun{
# simulate data
set.seed(1)
data <- list(
  x = rnorm(10),
  N = 10
)
data$x

# define priors
priors_list <- list(mu = prior("normal", list(0, 1)))

# define likelihood for the data
model_syntax <-
  "model{
    for(i in 1:N){
      x[i] ~ dnorm(mu, 1)
    }
  }"

# fit the models
fit <- JAGS_fit(model_syntax, data, priors_list)

# define log posterior for bridge sampling
log_posterior <- function(parameters, data){
  sum(dnorm(data$x, parameters$mu, 1, log = TRUE))
}

# get marginal likelihoods
marglik <- JAGS_bridgesampling(fit, log_posterior, data, priors_list)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model-averaging.R
\name{inclusion_BF}
\alias{inclusion_BF}
\title{Compute inclusion Bayes factors}
\usage{
inclusion_BF(prior_probs, post_probs, margliks, is_null)
}
\arguments{
\item{prior_probs}{vector of prior model probabilities}

\item{post_probs}{vector of posterior model probabilities}

\item{margliks}{vector of marginal likelihoods.}

\item{is_null}{logical vector of indicators whether the model corresponds
to the null or alternative hypothesis (or an integer vector indexing models
corresponding to the null hypothesis)}
}
\value{
\code{inclusion_BF} returns a Bayes factor.
}
\description{
Computes inclusion Bayes factors based on prior model probabilities,
posterior model probabilities (or marginal likelihoods), and indicator whether
the models represent the null or alternative hypothesis.
}
\details{
Supplying \code{margliks} as the input is preferred since it is better at dealing with
under/overflow (posterior probabilities are very close to either 0 or 1). In case that both the
\code{post_probs} and \code{margliks} are supplied, the results are based on \code{margliks}.
If the prior probability of either the null or alternative hypothesis is
zero, the Bayes factor is undefined and \code{NA} is returned.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/selection-kernels.R
\name{selection_backend_spec}
\alias{selection_backend_spec}
\title{Compile selection priors for backend consumers}
\usage{
selection_backend_spec(
  priors,
  backend = "jags",
  names = list(omega = "omega", alpha = "alpha"),
  global_breaks = NULL
)
}
\arguments{
\item{priors}{a selection prior, p-hacking prior, composed bias prior,
\code{prior_none()}, \code{prior_mixture()}, or a list of those priors.}

\item{backend}{backend target. Currently only \code{"jags"} is supported.}

\item{names}{list of backend parameter names.}

\item{global_breaks}{optional global p-value break grid.}
}
\value{
A list describing the compiled backend specification.
}
\description{
\code{selection_backend_spec()} compiles step-selection and
p-hacking prior objects into active backend parameters. The returned object
contains stable p/z geometry, JAGS prior/transform code, monitor names,
initial values, and data constants.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model-averaging.R
\name{mix_posteriors}
\alias{mix_posteriors}
\title{Model-average posterior distributions}
\usage{
mix_posteriors(
  model_list,
  parameters,
  is_null_list,
  conditional = FALSE,
  seed = NULL,
  n_samples = 10000
)
}
\arguments{
\item{model_list}{list of models, each of which contains marginal
likelihood estimated with bridge sampling \code{marglik} and prior model
odds \code{prior_weights}}

\item{parameters}{vector of parameters names for which inference should
be drawn}

\item{is_null_list}{list with entries for each parameter carrying either
logical vector of indicators specifying whether the model corresponds
to the null or alternative hypothesis (or an integer vector indexing models
corresponding to the null hypothesis)}

\item{conditional}{whether prior and posterior model probabilities should
be returned only for the conditional model. Defaults to \code{FALSE}}

\item{seed}{integer specifying seed for sampling posteriors for
model averaging. Defaults to \code{NULL}.}

\item{n_samples}{number of samples to be drawn for the model-averaged
posterior distribution}
}
\value{
\code{mix_posteriors} returns a named list of mixed posterior
distributions (either a vector of matrix).
}
\description{
Model-averages posterior distributions based on
a list of models, vector of parameters, and a list of
indicators of the null or alternative hypothesis models
for each parameter.
}
\seealso{
\link{ensemble_inference} \link{BayesTools_ensemble_tables} \link{as_mixed_posteriors}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/JAGS-formula.R
\name{parameter_names}
\alias{parameter_names}
\alias{format_parameter_names}
\alias{JAGS_parameter_names}
\title{Clean parameter names from JAGS}
\usage{
format_parameter_names(
  parameters,
  formula_parameters = NULL,
  formula_random = NULL,
  formula_prefix = TRUE,
  formula_scale = NULL
)

JAGS_parameter_names(parameters, formula_parameter = NULL)
}
\arguments{
\item{parameters}{a vector of parameter names}

\item{formula_parameters}{a vector of formula parameter prefix names}

\item{formula_random}{a vector of random effects grouping factors}

\item{formula_prefix}{whether the \code{formula_parameters} names should be
kept. Defaults to \code{TRUE}.}

\item{formula_scale}{optional nested list containing scaling info. When provided,
intercepts from parameters with \code{log_intercept = TRUE} attribute will be
renamed to \code{exp(intercept)}.}

\item{formula_parameter}{a formula parameter prefix name}
}
\value{
A character vector with reformatted parameter names.
}
\description{
Removes additional formatting from parameter names outputted from
JAGS.
}
\examples{
format_parameter_names(c("mu_x_cont", "mu_x_fac3t", "mu_x_fac3t__xXx__x_cont"),
                       formula_parameters = "mu")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/JAGS-formula.R
\name{JAGS_formula_design}
\alias{JAGS_formula_design}
\title{Extract Fitted JAGS Formula Design Metadata}
\usage{
JAGS_formula_design(fit, parameter = NULL)
}
\arguments{
\item{fit}{a fitted object returned by \code{\link[=JAGS_fit]{JAGS_fit()}}.}

\item{parameter}{optional formula parameter name. If \code{NULL}, all stored
formula designs are returned.}
}
\value{
A named list of formula designs, or one formula design when
\code{parameter} is supplied. Returns \code{NULL} when no formula design
metadata is stored on \code{fit}.
}
\description{
Returns the fitted formula design metadata stored by
\code{\link[=JAGS_fit]{JAGS_fit()}}. The design contains the processed formula, fitted model frame,
exact model matrix used for JAGS data construction, JAGS-safe coefficient
names, contrast and factor-level metadata, rank diagnostics, prior metadata,
and formula-scale information.
}
\seealso{
\code{\link[=JAGS_fit]{JAGS_fit()}} \code{\link[=JAGS_formula]{JAGS_formula()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model-averaging-plots.R
\name{plot_models}
\alias{plot_models}
\title{Plot estimates from models}
\usage{
plot_models(
  model_list,
  samples,
  inference,
  parameter,
  plot_type = "base",
  prior = FALSE,
  conditional = FALSE,
  order = NULL,
  transformation = NULL,
  transformation_arguments = NULL,
  transformation_settings = FALSE,
  par_name = NULL,
  formula_prefix = TRUE,
  ...
)
}
\arguments{
\item{model_list}{list of models, each of which contains marginal
likelihood estimated with bridge sampling \code{marglik} and prior model
odds \code{prior_weights}}

\item{samples}{samples from a posterior distribution for a
parameter generated by \link{mix_posteriors} or \link{as_mixed_posteriors}.}

\item{inference}{object created by \link{ensemble_inference} function}

\item{parameter}{parameter name to be plotted. Does not support
PET-PEESE and weightfunction.}

\item{plot_type}{whether to use a base plot \code{"base"}
or ggplot2 \code{"ggplot"} for plotting.}

\item{prior}{whether prior distribution should be added to the figure.
When samples were prepared with \code{as_mixed_posteriors(..., transform_scaled = TRUE)},
the transformed prior samples are automatically used.}

\item{conditional}{whether conditional models should be displayed}

\item{order}{list specifying ordering of the models. The first
element describes whether the ordering should be \code{"increasing"}
or \code{"decreasing"} and the second element describes whether
the ordering should be based \code{"model"} order, \code{"estimate"}
size, posterior \code{"probability"}, or the inclusion \code{"BF"}.}

\item{transformation}{transformation to be applied
to the prior distribution. Either a character
specifying one of the prepared transformations:
\describe{
\item{lin}{linear transformation in form of \code{a + b*x}}
\item{tanh}{hyperbolic tangent transformation}
\item{exp}{exponential transformation}
}, or a list containing the transformation function \code{fun},
inverse transformation function \code{inv}, and derivative of the
transformation \code{jac}, evaluated on the original support. See examples
for details.}

\item{transformation_arguments}{a list with named arguments for
the \code{transformation}}

\item{transformation_settings}{boolean indicating whether the
settings the \code{x_seq} or \code{x_range} was specified on
the transformed support}

\item{par_name}{a type of parameter for which the prior is
specified. Only relevant if the prior corresponds to a mu
parameter that needs to be transformed.}

\item{formula_prefix}{whether the \code{formula_parameters} names should be
kept. Defaults to \code{TRUE}.}

\item{...}{additional arguments. E.g.:
\describe{
\item{\code{"show_updating"}}{whether Bayes factors and change from
prior to posterior odds should be shown on the secondary y-axis}
\item{\code{"show_estimates"}}{whether posterior estimates and 95\% CI
should be shown on the secondary y-axis}
\item{\code{"y_axis2"}}{whether the secondary y-axis should be shown}
}}
}
\value{
\code{plot_models} returns either \code{NULL} or
an object of class 'ggplot' if plot_type is \code{plot_type = "ggplot"}.
}
\description{
Plot estimates from models
}
\details{
Plots prior and posterior estimates of the same parameter
across multiple models (prior distributions with orthonormal/meandif contrast
are always plotted as differences from the grand mean).
}
\seealso{
\code{\link[=prior]{prior()}} \code{\link[=lines_prior_list]{lines_prior_list()}}  \code{\link[=geom_prior_list]{geom_prior_list()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/priors.R
\name{prior_functions}
\alias{prior_functions}
\alias{rng.prior}
\alias{cdf.prior}
\alias{ccdf.prior}
\alias{lpdf.prior}
\alias{pdf.prior}
\alias{quant.prior}
\alias{mcdf.prior}
\alias{mccdf.prior}
\alias{mlpdf.prior}
\alias{mpdf.prior}
\alias{mquant.prior}
\title{Elementary prior related functions}
\usage{
\method{rng}{prior}(x, n, ...)

\method{cdf}{prior}(x, q, ...)

\method{ccdf}{prior}(x, q, ...)

\method{lpdf}{prior}(x, y, ...)

\method{pdf}{prior}(x, y, ...)

\method{quant}{prior}(x, p, ...)

\method{mcdf}{prior}(x, q, ...)

\method{mccdf}{prior}(x, q, ...)

\method{mlpdf}{prior}(x, y, ...)

\method{mpdf}{prior}(x, y, ...)

\method{mquant}{prior}(x, p, ...)
}
\arguments{
\item{x}{prior distribution}

\item{n}{number of observations}

\item{...}{unused arguments}

\item{q}{vector or matrix of quantiles}

\item{y}{vector of observations}

\item{p}{vector of probabilities}
}
\value{
\code{pdf} (\code{mpdf}) and \code{lpdf} (\code{mlpdf}) give
the (marginal) density and the log of (marginal) density,
\code{cdf} (\code{mcdf}) and \code{ccdf} (\code{mccdf}) give the
(marginal) distribution and the complement of (marginal) distribution function,
\code{quant} (\code{mquant}) give the (marginal) quantile function,
and \code{rng} generates random deviates for an object of class 'prior'.
}
\description{
Density (pdf / lpdf), distribution
function (cdf / ccdf), quantile function (quant),
random generation (rng), mean, standard deviation (sd),
and marginal variants of the functions (mpdf, mlpf, mcdf,
mccdf, mquant) for prior distributions.
}
\examples{
# create a standard normal prior distribution
p1 <- prior(distribution = "normal", parameters = list(mean = 1, sd = 1))

# generate a random sample from the prior
rng(p1, 10)

# compute cumulative density function
cdf(p1, 0)

# obtain quantile
quant(p1, .5)

# compute probability density
pdf(p1, c(0, 1, 2))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/priors.R
\name{prior_mixture}
\alias{prior_mixture}
\title{Creates a mixture of prior distributions}
\usage{
prior_mixture(
  prior_list,
  is_null = rep(FALSE, length(prior_list)),
  components = NULL
)
}
\arguments{
\item{prior_list}{a list of prior distributions to be mixed.}

\item{is_null}{a logical vector indicating which of the prior distributions
should be considered as a null distribution. Defaults to \code{rep(FALSE, length(prior_list))}.}

\item{components}{a character vector indicating which of the prior distributions
belong to the same mixture component (this is an alternative specification to the \code{is_null} argument).
Defaults to \code{NULL} (i.e., \code{is_null} is used.}
}
\description{
\code{prior_mixture} creates a mixture of prior distributions.
This is a more generic version of the \code{prior_spike_and_slab} function.
}
\seealso{
\code{\link[=prior]{prior()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model-averaging-plots.R
\name{lines_prior_list}
\alias{lines_prior_list}
\title{Add list of prior objects to a plot}
\usage{
lines_prior_list(
  prior_list,
  xlim = NULL,
  x_seq = NULL,
  x_range_quant = NULL,
  n_points = 500,
  n_samples = 10000,
  force_samples = FALSE,
  individual = FALSE,
  show_figures = if (individual) 1 else NULL,
  transformation = NULL,
  transformation_arguments = NULL,
  transformation_settings = FALSE,
  rescale_x = FALSE,
  scale_y2 = NULL,
  prior_list_mu = NULL,
  effect_direction = "positive",
  ...
)
}
\arguments{
\item{prior_list}{list of prior distributions}

\item{xlim}{x plotting range}

\item{x_seq}{sequence of x coordinates}

\item{x_range_quant}{quantile used for
automatically obtaining \code{x_range}
if both \code{x_range} and \code{x_seq}
are unspecified. Defaults to \code{0.005}
for all but Cauchy, Student-t, Gamma, and
Inverse-gamme distributions that use
\code{0.010}.}

\item{n_points}{number of equally spaced points
in the \code{x_range} if \code{x_seq} is unspecified}

\item{n_samples}{number of samples from the prior
distribution if the density cannot be obtained
analytically (or if samples are forced with
\code{force_samples = TRUE})}

\item{force_samples}{should prior be sampled instead
of obtaining analytic solution whenever possible}

\item{individual}{should individual densities be returned
(e.g., in case of weightfunction)}

\item{show_figures}{which figures should be returned in case of
multiple plots are generated. Useful when priors for the omega
parameter are plotted and \code{individual = TRUE}.}

\item{transformation}{transformation to be applied
to the prior distribution. Either a character
specifying one of the prepared transformations:
\describe{
\item{lin}{linear transformation in form of \code{a + b*x}}
\item{tanh}{hyperbolic tangent transformation}
\item{exp}{exponential transformation}
}, or a list containing the transformation function \code{fun},
inverse transformation function \code{inv}, and derivative of the
transformation \code{jac}, evaluated on the original support. See examples
for details.}

\item{transformation_arguments}{a list with named arguments for
the \code{transformation}}

\item{transformation_settings}{boolean indicating whether the
settings the \code{x_seq} or \code{x_range} was specified on
the transformed support}

\item{rescale_x}{allows to rescale x-axis in case a
weightfunction is plotted.}

\item{scale_y2}{scaling factor for a secondary axis}

\item{prior_list_mu}{list of priors for the mu parameter
required when plotting PET-PEESE}

\item{effect_direction}{direction of the effect for PET-PEESE
regression. Use \code{"positive"} (default) for
\code{mu + PET*se + PEESE*se^2} or \code{"negative"} for
\code{mu - PET*se - PEESE*se^2}.}

\item{...}{additional arguments}
}
\value{
\code{lines_prior_list} returns \code{NULL}.
}
\description{
Add list of prior objects to a plot
}
\seealso{
\code{\link[=plot_prior_list]{plot_prior_list()}} \code{\link[=geom_prior_list]{geom_prior_list()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/JAGS-fit.R
\name{JAGS_check_and_list}
\alias{JAGS_check_and_list}
\alias{JAGS_check_and_list_fit_settings}
\alias{JAGS_check_and_list_autofit_settings}
\title{Check and list 'JAGS' fitting settings}
\usage{
JAGS_check_and_list_fit_settings(
  chains,
  adapt,
  burnin,
  sample,
  thin,
  autofit,
  parallel,
  cores,
  silent,
  seed,
  check_mins = list(chains = 1, adapt = 50, burnin = 50, sample = 100, thin = 1),
  call = ""
)

JAGS_check_and_list_autofit_settings(
  autofit_control,
  skip_sample_extend = FALSE,
  call = ""
)
}
\arguments{
\item{chains}{number of chains to be run, defaults to \code{4}}

\item{adapt}{number of samples used for adapting the MCMC chains, defaults to \code{500}}

\item{burnin}{number of burnin iterations of the MCMC chains, defaults to \code{1000}}

\item{sample}{number of sampling iterations of the MCMC chains, defaults to \code{4000}}

\item{thin}{thinning interval for the MCMC samples, defaults to \code{1}}

\item{autofit}{whether the models should be refitted until convergence criteria
specified in \code{autofit_control}. Defaults to \code{FALSE}.}

\item{parallel}{whether the chains should be run in parallel \code{FALSE}}

\item{cores}{number of cores used for multithreading if \code{parallel = TRUE},
defaults to \code{chains}}

\item{silent}{whether the function should proceed silently, defaults to \code{TRUE}}

\item{seed}{seed for random number generation}

\item{check_mins}{named list of minimal values for which
should some input be checked. Defaults to:
\describe{
\item{chains}{\code{1}}
\item{adapt}{\code{50}}
\item{burnin}{\code{50}}
\item{sample}{\code{100}}
\item{thin}{\code{1}}
}}

\item{call}{string to be placed as a prefix to the error call.}

\item{autofit_control}{a list of arguments controlling the autofit function.
Possible options are:
\describe{
\item{max_Rhat}{maximum R-hat error for the autofit function.
Defaults to \code{1.05}.}
\item{min_ESS}{minimum effective sample size. Defaults to \code{500}.}
\item{max_error}{maximum MCMC error. Defaults to \code{0.01}.}
\item{max_SD_error}{maximum MCMC error as the proportion of standard
deviation of the parameters. Defaults to \code{0.05}.}
\item{max_time}{list specifying the time \code{time} and \code{units}
after which the automatic fitting function is stopped. The units arguments
need to correspond to \code{units} passed to \link[base]{difftime} function.}
\item{max_extend}{number of times after which the automatic fitting function is stopped.}
\item{sample_extend}{number of samples between each convergence check. Defaults to
\code{1000}.}
\item{restarts}{number of times new initial values should be generated in case the model
fails to initialize. Defaults to \code{10}.}
\item{check_indicators}{whether model indicator variables should be included
in convergence checks. Defaults to \code{FALSE}.}
}}

\item{skip_sample_extend}{whether \code{sample_extend}
is allowed to be NULL and skipped in the check}
}
\value{
\code{JAGS_check_and_list_fit_settings} invisibly returns a
list of checked fit settings. \code{JAGS_check_and_list_autofit_settings}
invisibly returns a list of checked autofit settings.
parameter names.
}
\description{
Checks and lists settings for the
\link{JAGS_fit} function.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/JAGS-diagnostics.R
\name{JAGS_diagnostics}
\alias{JAGS_diagnostics}
\alias{JAGS_diagnostics_density}
\alias{JAGS_diagnostics_autocorrelation}
\alias{JAGS_diagnostics_trace}
\title{Plot diagnostics of a 'JAGS' model}
\usage{
JAGS_diagnostics(
  fit,
  parameter,
  type,
  plot_type = "base",
  xlim = NULL,
  ylim = NULL,
  lags = 30,
  n_points = 1000,
  transformations = NULL,
  transform_factors = FALSE,
  transform_orthonormal = FALSE,
  short_name = FALSE,
  parameter_names = FALSE,
  formula_prefix = TRUE,
  ...
)

JAGS_diagnostics_density(
  fit,
  parameter,
  plot_type = "base",
  xlim = NULL,
  n_points = 1000,
  transformations = NULL,
  transform_factors = FALSE,
  transform_orthonormal = FALSE,
  short_name = FALSE,
  parameter_names = FALSE,
  formula_prefix = TRUE,
  ...
)

JAGS_diagnostics_trace(
  fit,
  parameter,
  plot_type = "base",
  ylim = NULL,
  transformations = NULL,
  transform_factors = FALSE,
  transform_orthonormal = FALSE,
  short_name = FALSE,
  parameter_names = FALSE,
  formula_prefix = TRUE,
  ...
)

JAGS_diagnostics_autocorrelation(
  fit,
  parameter,
  plot_type = "base",
  lags = 30,
  transformations = NULL,
  transform_factors = FALSE,
  transform_orthonormal = FALSE,
  short_name = FALSE,
  parameter_names = FALSE,
  formula_prefix = TRUE,
  ...
)
}
\arguments{
\item{fit}{a JAGS model fitted via \code{\link[=JAGS_fit]{JAGS_fit()}}}

\item{parameter}{parameter to be plotted}

\item{type}{what type of model diagnostic should be plotted. The available
options are \code{"density"}, \code{"trace"}, and \code{"autocorrelation"}}

\item{plot_type}{whether to use a base plot \code{"base"}
or ggplot2 \code{"ggplot"} for plotting.}

\item{xlim}{x plotting range}

\item{ylim}{y plotting range}

\item{lags}{number of lags to be shown for the autocorrelation plot.
Defaults to \code{30}.}

\item{n_points}{number of equally spaced points
in the \code{x_range} if \code{x_seq} is unspecified}

\item{transformations}{named list of transformations to be applied
to specific parameters}

\item{transform_factors}{whether factors with orthonormal/meandif
prior distribution should be transformed to differences from the
grand mean}

\item{transform_orthonormal}{(to be depreciated) whether factors
with orthonormal prior distributions should be transformed to
differences from the grand mean}

\item{short_name}{whether prior distribution names should be
shorted}

\item{parameter_names}{whether parameter names should be
printed}

\item{formula_prefix}{whether the parameter prefix from formula should
be printed. Defaults to \code{TRUE}.}

\item{...}{additional arguments}
}
\value{
\code{diagnostics} returns either \code{NULL} if \code{plot_type = "base"}
or an object/list of objects (depending on the number of parameters to be plotted)
of class 'ggplot2' if \code{plot_type = "ggplot2"}.
}
\description{
Creates density plots, trace plots, and autocorrelation plots
for a given parameter of a JAGS model.
}
\seealso{
\code{\link[=JAGS_fit]{JAGS_fit()}} \code{\link[=JAGS_check_convergence]{JAGS_check_convergence()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/priors.R
\name{prior_factor}
\alias{prior_factor}
\title{Creates a prior distribution for factors}
\usage{
prior_factor(
  distribution,
  parameters,
  truncation = list(lower = -Inf, upper = Inf),
  prior_weights = 1,
  contrast = "meandif"
)
}
\arguments{
\item{distribution}{name of the prior distribution. The
possible options are
\describe{
\item{\code{"point"}}{for a point density characterized by a
\code{location} parameter.}
\item{\code{"normal"}}{for a normal distribution characterized
by a \code{mean} and \code{sd} parameters.}
\item{\code{"lognormal"}}{for a lognormal distribution characterized
by a \code{meanlog} and \code{sdlog} parameters.}
\item{\code{"cauchy"}}{for a Cauchy distribution characterized
by a \code{location} and \code{scale} parameters. Internally
converted into a generalized t-distribution with \code{df = 1}.}
\item{\code{"t"}}{for a generalized t-distribution characterized
by a \code{location}, \code{scale}, and \code{df} parameters.}
\item{\code{"gamma"}}{for a gamma distribution characterized
by either \code{shape} and \code{rate}, or \code{shape} and
\code{scale} parameters. The later is internally converted to
the \code{shape} and \code{rate} parametrization}
\item{\code{"invgamma"}}{for an inverse-gamma distribution
characterized by a \code{shape} and \code{scale} parameters. The
JAGS part uses a 1/gamma distribution with a shape and rate
parameter.}
\item{\code{"beta"}}{for a beta distribution
characterized by an \code{alpha} and \code{beta} parameters.}
\item{\code{"exp"}}{for an exponential distribution
characterized by either \code{rate} or \code{scale}
parameter. The later is internally converted to
\code{rate}.}
\item{\code{"uniform"}}{for a uniform distribution defined on a
range from \code{a} to \code{b}}
}}

\item{parameters}{list of appropriate parameters for a given
\code{distribution}.}

\item{truncation}{list with two elements, \code{lower} and
\code{upper}, that define the lower and upper truncation of the
distribution. Defaults to \code{list(lower = -Inf, upper = Inf)}.
The truncation is automatically set to the bounds of the support.}

\item{prior_weights}{prior odds associated with a given distribution.
The value is passed into the model fitting function, which creates models
corresponding to all combinations of prior distributions for each of
the model parameters and sets the model priors odds to the product
of its prior distributions.}

\item{contrast}{type of contrast for the prior distribution. The possible options are
\describe{
\item{\code{"meandif"}}{for contrast centered around the grand mean
with equal marginal distributions, making the prior distribution exchangeable
across factor levels. In contrast to \code{"orthonormal"}, the marginal distributions
are identical regardless of the number of factor levels and the specified prior
distribution corresponds to the difference from grand mean for each factor level.
Only supports \code{distribution = "mnormal"} and \code{distribution = "mt"}
which generates the corresponding multivariate normal/t distributions.}
\item{\code{"orthonormal"}}{for contrast centered around the grand mean
with equal marginal distributions, making the prior distribution exchangeable
across factor levels. Only supports \code{distribution = "mnormal"} and
\code{distribution = "mt"} which generates the corresponding multivariate normal/t
distributions.}
\item{\code{"treatment"}}{for contrasts using the first level as a comparison
group and setting equal prior distribution on differences between the individual
factor levels and the comparison level.}
\item{\code{"independent"}}{for contrasts specifying dependent prior distribution
for each factor level (note that this leads to an overparameterized model if the
intercept is included).}
}}
}
\value{
return an object of class 'prior'.
}
\description{
\code{prior_factor} creates a prior distribution for fitting
models with factor predictors. (Note that results across different operating
systems might vary due to differences in JAGS numerical precision.)
}
\examples{
# create an orthonormal prior distribution
p1 <- prior_factor(distribution = "mnormal", contrast = "orthonormal",
                   parameters = list(mean = 0, sd = 1))

}
\seealso{
\code{\link[=prior]{prior()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/marginal-distributions.R
\name{marginal_inference}
\alias{marginal_inference}
\title{Model-average marginal posterior distributions and
marginal Bayes factors}
\usage{
marginal_inference(
  model_list,
  marginal_parameters,
  parameters,
  is_null_list,
  formula,
  null_hypothesis = 0,
  normal_approximation = FALSE,
  n_samples = 10000,
  seed = NULL,
  silent = FALSE
)
}
\arguments{
\item{model_list}{list of models, each of which contains marginal
likelihood estimated with bridge sampling \code{marglik} and prior model
odds \code{prior_weights}}

\item{marginal_parameters}{parameters for which the the marginal summary
should be created}

\item{parameters}{all parameters included in the model_list that are
relevant for the formula (all of which need to have specification of
\code{is_null_list})}

\item{is_null_list}{list with entries for each parameter carrying either
logical vector of indicators specifying whether the model corresponds
to the null or alternative hypothesis (or an integer vector indexing models
corresponding to the null hypothesis)}

\item{formula}{model formula (needs to be specified if \code{parameter} was part of a formula)}

\item{null_hypothesis}{point null hypothesis to test. Defaults to \code{0}}

\item{normal_approximation}{whether the height of prior and posterior density should be
approximated via a normal distribution (rather than kernel density). Defaults to \code{FALSE}.}

\item{n_samples}{controls the numerical grid used for model-averaged
prior densities}

\item{seed}{seed for random number generation}

\item{silent}{whether warnings should be returned silently. Defaults to \code{FALSE}}
}
\value{
\code{marginal_inference} returns an object of class 'marginal_inference'.
}
\description{
Creates marginal model-averaged and conditional
posterior distributions based on a list of models, vector of parameters,
formula, and a list of indicators of the null or alternative hypothesis models
for each parameter. Computes inclusion Bayes factors for each
marginal estimate via a Savage-Dickey density approximation.
}
\seealso{
\link{ensemble_inference} \link{mix_posteriors} \link{BayesTools_ensemble_tables}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/JAGS-formula.R
\name{formula_add_intercept}
\alias{formula_add_intercept}
\title{Add an Intercept to a Formula}
\usage{
formula_add_intercept(formula)
}
\arguments{
\item{formula}{a formula object.}
}
\value{
A formula object with an intercept.
}
\description{
Converts a no-intercept formula to the corresponding formula
with an intercept while preserving the formula environment. Top-level
no-intercept encodings such as \code{- 1}, \code{+ 0}, and \code{0 +} are
removed without editing transformed calls such as \code{I(x - 1)} or
\code{offset(x - 1)}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/JAGS-fit.R
\name{JAGS_check_convergence}
\alias{JAGS_check_convergence}
\title{Assess convergence of a runjags model}
\usage{
JAGS_check_convergence(
  fit,
  prior_list,
  max_Rhat = 1.05,
  min_ESS = 500,
  max_error = 0.01,
  max_SD_error = 0.05,
  add_parameters = NULL,
  fail_fast = FALSE,
  check_indicators = FALSE
)
}
\arguments{
\item{fit}{a runjags model}

\item{prior_list}{named list of prior distribution
(names correspond to the parameter names)}

\item{max_Rhat}{maximum R-hat error for the autofit function.
Defaults to \code{1.05}.}

\item{min_ESS}{minimum effective sample size. Defaults to \code{500}.}

\item{max_error}{maximum MCMC error. Defaults to \code{0.01}.}

\item{max_SD_error}{maximum MCMC error as the proportion of standard
deviation of the parameters. Defaults to \code{0.05}.}

\item{add_parameters}{vector of additional parameter names that should be used
(only allows removing last, fixed, omega element if omega is tracked manually).}

\item{fail_fast}{whether the function should stop after the first failed convergence check.}

\item{check_indicators}{whether model indicator variables should be included
in convergence checks. Defaults to \code{FALSE}.}
}
\value{
\code{JAGS_check_convergence} returns a boolean
indicating whether the model converged or not, with an
attribute 'errors' carrying the failed convergence checks (if any).
}
\description{
Checks whether the supplied \link[runjags]{runjags-package} model
satisfied convergence criteria.
}
\examples{
\dontrun{
# simulate data
set.seed(1)
data <- list(
  x = rnorm(10),
  N = 10
)
data$x

# define priors
priors_list <- list(mu = prior("normal", list(0, 1)))

# define likelihood for the data
model_syntax <-
  "model{
    for(i in 1:N){
      x[i] ~ dnorm(mu, 1)
    }
  }"

# fit the models
fit <- JAGS_fit(model_syntax, data, priors_list)
JAGS_check_convergence(fit, priors_list)
}
}
\seealso{
\code{\link[=JAGS_fit]{JAGS_fit()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/JAGS-formula.R
\name{transform_orthonormal_samples}
\alias{transform_orthonormal_samples}
\title{Transform orthonomal posterior samples into differences from the mean}
\usage{
transform_orthonormal_samples(samples)
}
\arguments{
\item{samples}{(a list) of mixed posterior distributions created with
\code{mix_posteriors} function}
}
\value{
\code{transform_orthonormal_samples} returns a named list of mixed posterior
distributions (either a vector of matrix).
}
\description{
Transforms posterior samples from model-averaged posterior
distributions based on orthonormal prior distributions into differences from
the mean.
}
\seealso{
\link{mix_posteriors} \link{contr.orthonormal}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/JAGS-formula.R
\name{transform_prior_samples}
\alias{transform_prior_samples}
\title{Transform prior samples to original scale}
\usage{
transform_prior_samples(
  fit,
  n_samples = 10000,
  seed = NULL,
  formula_scale = NULL
)
}
\arguments{
\item{fit}{a fitted model object with \code{prior_list} and optionally
\code{formula_scale} attributes}

\item{n_samples}{number of samples to generate (default: 10000)}

\item{seed}{random seed for reproducibility (optional)}

\item{formula_scale}{optional nested list containing standardization information.
If not provided, extracted from \code{fit} attribute.}
}
\value{
A matrix of prior samples on the original (unscaled) scale, with
columns matching the structure of posterior samples.
}
\description{
Generate prior samples and transform them using the same
matrix transformation as posterior samples. This is the correct approach for
visualizing priors on the original (unscaled) scale, especially for the intercept
which depends on contributions from multiple coefficient priors.
}
\details{
When models use auto-scaling (standardizing predictors), the posterior
samples are on the standardized scale. To correctly visualize priors on the
original scale, we cannot simply apply a linear transformation to individual
priors because the intercept on the original scale is a weighted sum of
multiple priors:

\deqn{\beta_0^{orig} = \beta_0^* - \sum_i \frac{\mu_i}{\sigma_i} \beta_i^*}

This function generates samples from ALL priors simultaneously and applies
the same matrix transformation used for posterior samples, which correctly
handles the intercept and all other parameters.
}
\examples{
# With a fitted model that used formula_scale:
# prior_samples <- transform_prior_samples(fit, n_samples = 10000)
# This can then be used with density() or for custom plotting

}
\seealso{
\code{\link[=transform_scale_samples]{transform_scale_samples()}} \code{\link[=plot_posterior]{plot_posterior()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/priors.R
\name{var.prior}
\alias{var.prior}
\title{Prior var}
\usage{
\method{var}{prior}(x, ...)
}
\arguments{
\item{x}{a prior}

\item{...}{unused arguments}
}
\value{
a variance of an object of class 'prior'.
}
\description{
Computes variance
of a prior distribution.
}
\examples{
# create a standard normal prior distribution
p1 <- prior(distribution = "normal", parameters = list(mean = 1, sd = 1))

# compute variance of the prior distribution
var(p1)

}
\seealso{
\code{\link[=prior]{prior()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{kitchen_rolls}
\alias{kitchen_rolls}
\title{Kitchen Rolls data from
\insertCite{wagenmakers2015turning;textual}{BayesTools} replication study.}
\format{
A data.frame with 2 columns and 102 observations.
}
\usage{
kitchen_rolls
}
\value{
a data.frame.
}
\description{
The data set contains mean NEO PI-R scores for two groups of students.
Each of them filled a personality questionnaire while rotating a kitchen roll either
clock or counter-clock wise. See \insertCite{wagenmakers2015turning;textual}{BayesTools}
for more details about the replication study and the https://osf.io/uszvx/ for the
original data.
}
\references{
\insertAllCited{}
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/selection-kernels.R
\name{is_prior_bias}
\alias{is_prior_bias}
\title{Reports whether x is a composed publication-bias prior}
\usage{
is_prior_bias(x)
}
\arguments{
\item{x}{object to test.}
}
\value{
A logical value.
}
\description{
Reports whether x is a composed publication-bias prior
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/marginal-distributions.R
\name{marginal_posterior}
\alias{marginal_posterior}
\title{Model-average marginal posterior distributions}
\usage{
marginal_posterior(
  samples,
  parameter,
  formula = NULL,
  at = NULL,
  prior_samples = FALSE,
  use_formula = TRUE,
  transformation = NULL,
  transformation_arguments = NULL,
  transformation_settings = FALSE,
  n_samples = 10000,
  ...
)
}
\arguments{
\item{samples}{model-averaged posterior samples created by \code{mix_posteriors()}}

\item{parameter}{parameter of interest}

\item{formula}{model formula (needs to be specified if \code{parameter} was part of a formula)}

\item{at}{named list with predictor levels of the formula for which marginalization
should be performed. If a predictor level is missing, \code{0} is used for continuous
predictors, the baseline factor level is used for factors with \code{contrast = "treatment"} prior
distributions, and the parameter is completely omitted for for factors with \code{contrast = "meandif"},}

\item{prior_samples}{whether marginal prior distributions should be generated
\code{contrast = "orthonormal"}, and \code{contrast = "independent"} levels}

\item{use_formula}{whether the parameter should be evaluated as a part of supplied formula}

\item{transformation}{transformation to be applied
to the prior distribution. Either a character
specifying one of the prepared transformations:
\describe{
\item{lin}{linear transformation in form of \code{a + b*x}}
\item{tanh}{hyperbolic tangent transformation}
\item{exp}{exponential transformation}
}, or a list containing the transformation function \code{fun},
inverse transformation function \code{inv}, and derivative of the
transformation \code{jac}, evaluated on the original support. See examples
for details.}

\item{transformation_arguments}{a list with named arguments for
the \code{transformation}}

\item{transformation_settings}{boolean indicating whether the
settings the \code{x_seq} or \code{x_range} was specified on
the transformed support}

\item{n_samples}{controls the numerical grid used for model-averaged
prior densities}

\item{...}{additional arguments}
}
\value{
\code{marginal_posterior} returns a named list of mixed marginal posterior
distributions (either a vector of matrix).
#'
}
\description{
Creates marginal model-averages posterior distributions for a given
parameter based on model-averaged posterior samples and parameter name
(and formula with at specification).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/summary-tables.R
\name{BayesTools_ensemble_tables}
\alias{BayesTools_ensemble_tables}
\alias{ensemble_estimates_table}
\alias{ensemble_inference_table}
\alias{ensemble_summary_table}
\alias{ensemble_diagnostics_table}
\alias{ensemble_estimates_empty_table}
\alias{ensemble_inference_empty_table}
\alias{ensemble_summary_empty_table}
\alias{ensemble_diagnostics_empty_table}
\alias{marginal_estimates_table}
\title{Create BayesTools ensemble summary tables}
\usage{
ensemble_estimates_table(
  samples,
  parameters,
  probs = c(0.025, 0.975),
  title = NULL,
  footnotes = NULL,
  warnings = NULL,
  transform_factors = FALSE,
  transform_orthonormal = FALSE,
  formula_prefix = TRUE,
  transform_scaled = FALSE,
  formula_scale = NULL
)

ensemble_inference_table(
  inference,
  parameters,
  logBF = FALSE,
  BF01 = FALSE,
  title = NULL,
  footnotes = NULL,
  warnings = NULL
)

ensemble_summary_table(
  models,
  parameters,
  logBF = FALSE,
  BF01 = FALSE,
  title = NULL,
  footnotes = NULL,
  warnings = NULL,
  remove_spike_0 = TRUE,
  short_name = FALSE
)

ensemble_diagnostics_table(
  models,
  parameters,
  title = NULL,
  footnotes = NULL,
  warnings = NULL,
  remove_spike_0 = TRUE,
  short_name = FALSE
)

ensemble_estimates_empty_table(
  probs = c(0.025, 0.975),
  title = NULL,
  footnotes = NULL,
  warnings = NULL
)

ensemble_inference_empty_table(title = NULL, footnotes = NULL, warnings = NULL)

ensemble_summary_empty_table(title = NULL, footnotes = NULL, warnings = NULL)

ensemble_diagnostics_empty_table(
  title = NULL,
  footnotes = NULL,
  warnings = NULL
)

marginal_estimates_table(
  samples,
  inference,
  parameters,
  probs = c(0.025, 0.5, 0.975),
  logBF = FALSE,
  BF01 = FALSE,
  title = NULL,
  footnotes = NULL,
  warnings = NULL,
  formula_prefix = TRUE,
  transform_scaled = FALSE,
  formula_scale = NULL
)
}
\arguments{
\item{samples}{posterior samples created by \link{mix_posteriors}}

\item{parameters}{character vector of parameters (or a
named list with of character vectors for summary and
diagnostics tables) specifying the parameters
(and their grouping) for the summary table}

\item{probs}{quantiles for parameter estimates}

\item{title}{title to be added to the table}

\item{footnotes}{footnotes to be added to the table}

\item{warnings}{warnings to be added to the table}

\item{transform_factors}{whether factors with orthonormal/meandif
prior distribution should be transformed to differences from the
grand mean}

\item{transform_orthonormal}{(to be depreciated) whether factors
with orthonormal prior distributions should be transformed to
differences from the grand mean}

\item{formula_prefix}{whether the parameter prefix from formula should
be printed. Defaults to \code{TRUE}.}

\item{transform_scaled}{whether coefficients from standardized
continuous predictors should be transformed back to the original
scale. Defaults to \code{FALSE}.}

\item{formula_scale}{named list containing standardization information
(mean and sd) for each standardized predictor. Required when
\code{transform_scaled = TRUE} for ensemble/marginal tables. For
\code{runjags_estimates_table}, this is automatically extracted from
the fit object's \code{formula_scale} attribute.}

\item{inference}{model inference created by \link{ensemble_inference}}

\item{logBF}{whether the Bayes factor should be on log scale}

\item{BF01}{whether the Bayes factor should be inverted}

\item{models}{list of \link{models_inference} model objects,
each of which containing a list of \code{priors}
and \code{inference} object, The \code{inference} must be a
named list with information about the model: model number
\code{m_number}, marginal likelihood \code{marglik}, prior and
posterior probability \code{prior_prob} and \code{post_prob},
inclusion Bayes factor \code{inclusion_BF}, and fit summary
generated by \link{runjags_estimates_table} for the diagnostics
table}

\item{remove_spike_0}{whether prior distributions equal to spike
at 0 should be removed from the \code{prior_list}}

\item{short_name}{whether the prior distribution names should be
shortened. Defaults to \code{FALSE}.}
}
\value{
\code{ensemble_estimates_table} returns a table with the
model-averaged estimates, \code{ensemble_inference_table} returns
a table with the prior and posterior probabilities and inclusion
Bayes factors, \code{ensemble_summary_table} returns a table with
overview of the models included in the ensemble, and
\code{ensemble_diagnostics_table} returns an overview of the MCMC
diagnostics for the models included in the ensemble. All of the
tables are objects of class 'BayesTools_table'.
}
\description{
Creates estimate summaries based on posterior
distributions created by \link{mix_posteriors}, inference summaries
based on inference created by \link{ensemble_inference}, or ensemble
summary/diagnostics based on a list of \link{models_inference} models
(or \link{marginal_inference} in case of \link{marginal_estimates_table}).
}
\seealso{
\link{ensemble_inference} \link{mix_posteriors} \link{BayesTools_model_tables}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/summary-tables.R
\name{format_BF}
\alias{format_BF}
\title{Format Bayes factor}
\usage{
format_BF(BF, logBF = FALSE, BF01 = FALSE, inclusion = FALSE)
}
\arguments{
\item{BF}{Bayes factor(s)}

\item{logBF}{log(BF)}

\item{BF01}{1/BF}

\item{inclusion}{whether the Bayes factor is an inclusion BF (for naming purposes)}
}
\value{
\code{format_BF} returns a formatted Bayes factor.
}
\description{
Formats Bayes factor
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/JAGS-marglik.R
\name{JAGS_marglik_parameters}
\alias{JAGS_marglik_parameters}
\alias{JAGS_marglik_parameters_formula}
\title{Extract parameters for 'JAGS' priors}
\usage{
JAGS_marglik_parameters(samples, prior_list)

JAGS_marglik_parameters_formula(
  samples,
  formula_list,
  formula_data_list,
  formula_prior_list,
  prior_list_parameters
)
}
\arguments{
\item{samples}{samples provided by bridgesampling
function}

\item{prior_list}{named list of prior distribution
(names correspond to the parameter names) of parameters not specified within the
\code{formula_list}}

\item{formula_list}{named list of formulas to be added to the model
(names correspond to the parameter name created by each of the formula)}

\item{formula_data_list}{named list of data frames containing data for each formula
(names of the lists correspond to the parameter name created by each of the formula)}

\item{formula_prior_list}{named list of named lists of prior distributions
(names of the lists correspond to the parameter name created by each of the formula and
the names of the prior distribution correspond to the parameter names) of parameters specified
within the \code{formula}}

\item{prior_list_parameters}{named list of prior distributions on model parameters
(not specified within the formula but that might scale the formula parameters)}
}
\value{
\code{JAGS_marglik_parameters} returns a named list
of (transformed) posterior samples.
}
\description{
Extracts transformed parameters from the
prior part of a 'JAGS' model inside of a 'bridgesampling'
function (returns them as a named list)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/priors-informed.R
\docType{data}
\name{prior_informed_medicine_names}
\alias{prior_informed_medicine_names}
\title{Names of medical subfields from the Cochrane database of systematic reviews}
\format{
An object of class \code{character} of length 57.
}
\usage{
prior_informed_medicine_names
}
\value{
returns a character vector with names of medical subfields from Cochrane database of
systematic reviews.
}
\description{
Contains the union of names identifying the individual subfields from
the Cochrane database of systematic reviews. Type-specific availability depends on
the \code{type} argument passed to \code{\link[=prior_informed]{prior_informed()}}.
}
\examples{
print(prior_informed_medicine_names)

}
\seealso{
\code{\link[=prior_informed]{prior_informed()}}
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/JAGS-fit.R
\name{JAGS_fit}
\alias{JAGS_fit}
\alias{JAGS_extend}
\title{Fits a 'JAGS' model}
\usage{
JAGS_fit(
  model_syntax,
  data = NULL,
  prior_list = NULL,
  formula_list = NULL,
  formula_data_list = NULL,
  formula_prior_list = NULL,
  formula_scale_list = NULL,
  chains = 4,
  adapt = 500,
  burnin = 1000,
  sample = 4000,
  thin = 1,
  autofit = FALSE,
  autofit_control = list(max_Rhat = 1.05, min_ESS = 500, max_error = 0.01, max_SD_error =
    0.05, max_time = list(time = 60, unit = "mins"), sample_extend = 1000, restarts = 10,
    max_extend = 10, check_indicators = FALSE),
  parallel = FALSE,
  cores = chains,
  silent = TRUE,
  seed = NULL,
  add_parameters = NULL,
  required_packages = NULL,
  ...
)

JAGS_extend(
  fit,
  autofit_control = list(max_Rhat = 1.05, min_ESS = 500, max_error = 0.01, max_SD_error =
    0.05, max_time = list(time = 60, unit = "mins"), sample_extend = 1000, restarts = 10,
    max_extend = 10, check_indicators = FALSE),
  parallel = FALSE,
  cores = NULL,
  silent = TRUE,
  seed = NULL
)
}
\arguments{
\item{model_syntax}{jags syntax for the model part}

\item{data}{list containing data to fit the model (not including data for the formulas)}

\item{prior_list}{named list of prior distribution
(names correspond to the parameter names) of parameters not specified within the
\code{formula_list}}

\item{formula_list}{named list of formulas to be added to the model
(names correspond to the parameter name created by each of the formula)}

\item{formula_data_list}{named list of data frames containing data for each formula
(names of the lists correspond to the parameter name created by each of the formula)}

\item{formula_prior_list}{named list of named lists of prior distributions
(names of the lists correspond to the parameter name created by each of the formula and
the names of the prior distribution correspond to the parameter names) of parameters specified
within the \code{formula}}

\item{formula_scale_list}{named list of named lists for standardizing continuous predictors
(names of the lists correspond to the parameter name created by each of the formula).
Each entry should be a named list where continuous predictors with \code{TRUE} values will
be standardized. Defaults to \code{NULL} (no standardization).}

\item{chains}{number of chains to be run, defaults to \code{4}}

\item{adapt}{number of samples used for adapting the MCMC chains, defaults to \code{500}}

\item{burnin}{number of burnin iterations of the MCMC chains, defaults to \code{1000}}

\item{sample}{number of sampling iterations of the MCMC chains, defaults to \code{4000}}

\item{thin}{thinning interval for the MCMC samples, defaults to \code{1}}

\item{autofit}{whether the models should be refitted until convergence criteria
specified in \code{autofit_control}. Defaults to \code{FALSE}.}

\item{autofit_control}{a list of arguments controlling the autofit function.
Possible options are:
\describe{
\item{max_Rhat}{maximum R-hat error for the autofit function.
Defaults to \code{1.05}.}
\item{min_ESS}{minimum effective sample size. Defaults to \code{500}.}
\item{max_error}{maximum MCMC error. Defaults to \code{0.01}.}
\item{max_SD_error}{maximum MCMC error as the proportion of standard
deviation of the parameters. Defaults to \code{0.05}.}
\item{max_time}{list specifying the time \code{time} and \code{units}
after which the automatic fitting function is stopped. The units arguments
need to correspond to \code{units} passed to \link[base]{difftime} function.}
\item{max_extend}{number of times after which the automatic fitting function is stopped.}
\item{sample_extend}{number of samples between each convergence check. Defaults to
\code{1000}.}
\item{restarts}{number of times new initial values should be generated in case the model
fails to initialize. Defaults to \code{10}.}
\item{check_indicators}{whether model indicator variables should be included
in convergence checks. Defaults to \code{FALSE}.}
}}

\item{parallel}{whether the chains should be run in parallel \code{FALSE}}

\item{cores}{number of cores used for multithreading if \code{parallel = TRUE},
defaults to \code{chains}}

\item{silent}{whether the function should proceed silently, defaults to \code{TRUE}}

\item{seed}{seed for random number generation}

\item{add_parameters}{vector of additional parameter names that should be used
monitored but were not specified in the \code{prior_list}}

\item{required_packages}{character vector specifying list of packages containing
JAGS models required for sampling (in case that the function is run in parallel or in
detached R session). Defaults to \code{NULL}.}

\item{...}{additional hidden arguments}

\item{fit}{a 'BayesTools_fit' object (created by \code{JAGS_fit()} function) to be
extended}
}
\value{
\code{JAGS_fit} returns an object of class 'runjags' and 'BayesTools_fit'.
}
\description{
A wrapper around
\link[runjags]{run.jags}  that simplifies fitting 'JAGS' models
with usage with pre-specified model part of the 'JAGS' syntax, data and list
of prior distributions.
}
\examples{
\dontrun{
# simulate data
set.seed(1)
data <- list(
  x = rnorm(10),
  N = 10
)
data$x

# define priors
priors_list <- list(mu = prior("normal", list(0, 1)))

# define likelihood for the data
model_syntax <-
  "model{
    for(i in 1:N){
      x[i] ~ dnorm(mu, 1)
    }
  }"

# fit the models
fit <- JAGS_fit(model_syntax, data, priors_list)
}

}
\seealso{
\code{\link[=JAGS_check_convergence]{JAGS_check_convergence()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/priors.R
\name{prior_functions_methods}
\alias{prior_functions_methods}
\alias{rng}
\alias{cdf}
\alias{ccdf}
\alias{quant}
\alias{lpdf}
\alias{pdf}
\alias{mcdf}
\alias{mccdf}
\alias{mquant}
\alias{mlpdf}
\alias{mpdf}
\title{Creates generics for common statistical functions}
\usage{
rng(x, ...)

cdf(x, ...)

ccdf(x, ...)

quant(x, ...)

lpdf(x, ...)

pdf(x, ...)

mcdf(x, ...)

mccdf(x, ...)

mquant(x, ...)

mlpdf(x, ...)

mpdf(x, ...)
}
\arguments{
\item{x}{main argument}

\item{...}{unused arguments}
}
\value{
\code{pdf} (\code{mpdf}) and \code{lpdf} (\code{mlpdf}) give
the (marginal) density and the log of (marginal) density,
\code{cdf} (\code{mcdf}) and \code{ccdf} (\code{mccdf}) give the
(marginal) distribution and the complement of (marginal) distribution function,
\code{quant} (\code{mquant}) give the (marginal) quantile function,
and \code{rng} generates random deviates for an object of class 'prior'.

The \code{pdf} function proceeds to PDF graphics device if \code{x} is a character.
}
\description{
Density (pdf / lpdf), distribution
function (cdf / ccdf), quantile function (quant),
random generation (rng), mean, standard deviation (sd),
and marginal variants of the functions (mpdf, mlpf, mcdf,
mccdf, mquant).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/posterior-extraction.R
\name{posterior_extraction_helpers}
\alias{posterior_extraction_helpers}
\alias{.extract_posterior_samples}
\alias{.remove_auxiliary_parameters}
\alias{.remove_parameter_columns}
\alias{.filter_parameters}
\alias{.process_spike_and_slab}
\alias{.apply_parameter_transformations}
\alias{.transform_factor_contrasts}
\alias{.rename_factor_levels}
\title{Helper functions for extracting and formatting posterior distributions}
\usage{
.extract_posterior_samples(fit, as_list = FALSE)

.remove_auxiliary_parameters(
  model_samples,
  prior_list,
  remove_parameters = NULL
)

.remove_parameter_columns(model_samples, prior, par_name)

.filter_parameters(
  prior_list,
  remove_parameters = NULL,
  remove_formulas = NULL,
  keep_parameters = NULL,
  keep_formulas = NULL,
  remove_spike_0 = TRUE
)

.process_spike_and_slab(
  model_samples,
  prior_list,
  par,
  conditional = FALSE,
  remove_inclusion = FALSE,
  warnings = NULL
)

.apply_parameter_transformations(
  model_samples,
  transformations,
  prior_list,
  transform_factors = FALSE
)

.transform_factor_contrasts(
  model_samples,
  prior_list,
  transform_factors = FALSE,
  transformations = NULL
)

.rename_factor_levels(model_samples, prior_list)
}
\arguments{
\item{fit}{a JAGS model fit object}

\item{as_list}{whether to return samples as mcmc.list (TRUE) or merged matrix (FALSE)}

\item{model_samples}{matrix of posterior samples}

\item{prior_list}{list of prior objects}

\item{remove_parameters}{character vector of parameter names to remove, or TRUE to remove all non-formula parameters.
If "bias" is specified and the bias prior contains PET, PEESE, or weightfunction priors,
the corresponding parameters (PET, PEESE, omega) are also added to the removal list.}

\item{prior}{prior object for the parameter}

\item{par_name}{name of the parameter}

\item{remove_formulas}{character vector of formula names whose parameters should be removed}

\item{keep_parameters}{character vector of parameter names to keep (all others removed unless in keep_formulas).
If "bias" is specified and the bias prior contains PET, PEESE, or weightfunction priors,
the corresponding parameters (PET, PEESE, omega) are also added to the keep list.}

\item{keep_formulas}{character vector of formula names whose parameters should be kept (all others removed unless in keep_parameters)}

\item{remove_spike_0}{whether to remove spike at 0 priors}

\item{par}{parameter name}

\item{conditional}{whether to compute conditional summary}

\item{remove_inclusion}{whether to remove inclusion indicators}

\item{warnings}{character vector for collecting warnings}

\item{transformations}{list of transformations to apply}

\item{transform_factors}{whether to transform orthonormal/meandif to differences}
}
\value{
matrix or mcmc.list of posterior samples

list with cleaned model_samples and updated prior_list

updated model_samples matrix

list with filtered model_samples and prior_list

list with updated model_samples, prior_list, and warnings

updated model_samples matrix

updated model_samples matrix

updated model_samples matrix with renamed columns
}
\description{
Internal helper functions to extract posterior samples from JAGS
fits and reformat them for further processing (summary tables, diagnostics, plots).
These functions consolidate common logic that was duplicated across
\code{runjags_estimates_table}, \code{.diagnostics_plot_data}, and plotting functions.

Helper to remove all columns associated with a parameter
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/priors.R
\name{prior_PP}
\alias{prior_PP}
\alias{prior_PET}
\alias{prior_PEESE}
\title{Creates a prior distribution for PET or PEESE models}
\usage{
prior_PET(
  distribution,
  parameters,
  truncation = list(lower = 0, upper = Inf),
  prior_weights = 1
)

prior_PEESE(
  distribution,
  parameters,
  truncation = list(lower = 0, upper = Inf),
  prior_weights = 1
)
}
\arguments{
\item{distribution}{name of the prior distribution. The
possible options are
\describe{
\item{\code{"point"}}{for a point density characterized by a
\code{location} parameter.}
\item{\code{"normal"}}{for a normal distribution characterized
by a \code{mean} and \code{sd} parameters.}
\item{\code{"lognormal"}}{for a lognormal distribution characterized
by a \code{meanlog} and \code{sdlog} parameters.}
\item{\code{"cauchy"}}{for a Cauchy distribution characterized
by a \code{location} and \code{scale} parameters. Internally
converted into a generalized t-distribution with \code{df = 1}.}
\item{\code{"t"}}{for a generalized t-distribution characterized
by a \code{location}, \code{scale}, and \code{df} parameters.}
\item{\code{"gamma"}}{for a gamma distribution characterized
by either \code{shape} and \code{rate}, or \code{shape} and
\code{scale} parameters. The later is internally converted to
the \code{shape} and \code{rate} parametrization}
\item{\code{"invgamma"}}{for an inverse-gamma distribution
characterized by a \code{shape} and \code{scale} parameters. The
JAGS part uses a 1/gamma distribution with a shape and rate
parameter.}
\item{\code{"beta"}}{for a beta distribution
characterized by an \code{alpha} and \code{beta} parameters.}
\item{\code{"exp"}}{for an exponential distribution
characterized by either \code{rate} or \code{scale}
parameter. The later is internally converted to
\code{rate}.}
\item{\code{"uniform"}}{for a uniform distribution defined on a
range from \code{a} to \code{b}}
}}

\item{parameters}{list of appropriate parameters for a given
\code{distribution}.}

\item{truncation}{list with two elements, \code{lower} and
\code{upper}, that define the lower and upper truncation of the
distribution. Defaults to \code{list(lower = -Inf, upper = Inf)}.
The truncation is automatically set to the bounds of the support.}

\item{prior_weights}{prior odds associated with a given distribution.
The value is passed into the model fitting function, which creates models
corresponding to all combinations of prior distributions for each of
the model parameters and sets the model priors odds to the product
of its prior distributions.}
}
\value{
\code{prior_PET} and \code{prior_PEESE} return an object of class 'prior'.
}
\description{
\code{prior} creates a prior distribution for fitting a PET or
PEESE style models in RoBMA. The prior distribution can be visualized
by the \code{plot} function.
}
\examples{
# create a half-Cauchy prior distribution
# (PET and PEESE specific functions automatically set lower truncation at 0)
p1 <- prior_PET(distribution = "Cauchy", parameters = list(location = 0, scale = 1))

plot(p1)

}
\seealso{
\code{\link[=plot.prior]{plot.prior()}}, \code{\link[=prior]{prior()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/selection-kernels.R
\name{prior_phacking}
\alias{prior_phacking}
\title{Creates a prior distribution for a p-hacking selection kernel}
\usage{
prior_phacking(
  side = "one-sided",
  target = 0.025,
  source = 0.25,
  destination = 0.005,
  form = c("linear", "quadratic"),
  alpha = prior("beta", list(1, 1)),
  report_scale = "pi_null",
  prior_weights = 1
)
}
\arguments{
\item{side}{side geometry. Currently only \code{"one-sided"} is supported.}

\item{target}{target p-value cut point.}

\item{source}{source p-value cut point. Must be larger than \code{target}.}

\item{destination}{destination p-value cut point. Must be smaller than
\code{target}.}

\item{form}{power depletion form, either \code{"linear"} or
\code{"quadratic"}.}

\item{alpha}{prior distribution for the p-hacking severity parameter.}

\item{report_scale}{reporting scale for deterministic summaries.}

\item{prior_weights}{prior odds associated with a given distribution.}
}
\value{
\code{prior_phacking()} returns an object of class \code{"prior"}.
}
\description{
\code{prior_phacking()} creates a prior distribution for a
mass-preserving p-hacking redistribution kernel. The initial backend supports
linear and quadratic power depletion between \code{source} and
\code{target}, with depleted probability mass redistributed between
\code{target} and \code{destination}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/JAGS-formula.R
\name{JAGS_evaluate_formula}
\alias{JAGS_evaluate_formula}
\title{Evaluate JAGS formula using posterior samples}
\usage{
JAGS_evaluate_formula(fit, formula, parameter, data, prior_list)
}
\arguments{
\item{fit}{model fitted with either \link[runjags]{runjags} posterior
samples obtained with \link[rjags]{rjags-package}}

\item{formula}{formula specifying the right hand side of the assignment (the
left hand side is ignored). If the formula has a \code{"log(intercept)"}
attribute set to \code{TRUE}, the intercept values will be log-transformed
before computing the linear predictor.}

\item{parameter}{name of the parameter created with the formula}

\item{data}{data.frame containing predictors included in the formula}

\item{prior_list}{named list of prior distribution of parameters specified within
the \code{formula}}
}
\value{
\code{JAGS_evaluate_formula} returns a matrix of the evaluated posterior samples on
the supplied data.
}
\description{
Evaluates a JAGS formula on a posterior distribution
obtained from a fitted model.
}
\seealso{
\code{\link[=JAGS_fit]{JAGS_fit()}} \code{\link[=JAGS_formula]{JAGS_formula()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/priors-linear-density.R
\name{plot_transformed_prior}
\alias{plot_transformed_prior}
\title{Plot Transformed Prior Densities}
\usage{
plot_transformed_prior(
  prior_list,
  column_names,
  formula_scale = NULL,
  parameter,
  n_points = 1000,
  x_range = NULL,
  transformation = NULL,
  transformation_arguments = NULL,
  transformation_settings = FALSE,
  plot_type = c("base", "ggplot"),
  par_name = NULL,
  ...
)
}
\arguments{
\item{prior_list}{named list of prior distributions.}

\item{column_names}{character vector of coefficient column names defining the
fitted parameter space.}

\item{formula_scale}{optional nested formula-scale metadata, in the same shape
as \code{attr(fit, "formula_scale")}.}

\item{parameter}{coefficient name to plot.}

\item{n_points}{number of plotting points.}

\item{x_range}{optional plotting range on the untransformed scale.}

\item{transformation}{optional output transformation passed to the prior
plotting machinery.}

\item{transformation_arguments}{optional list of transformation arguments.}

\item{transformation_settings}{whether \code{x_range} is supplied on the
transformed scale.}

\item{plot_type}{either \code{"base"} or \code{"ggplot"}.}

\item{par_name}{optional plotting label.}

\item{...}{additional plotting arguments.}
}
\value{
For \code{plot_type = "base"}, invisibly returns base-plot metadata.
For \code{plot_type = "ggplot"}, returns a \code{ggplot} object. Returns
\code{NULL} when the requested parameter is an identity transform and no
output transformation was requested.
}
\description{
Plots a prior density after applying the same deterministic
coefficient transformation used for formula-scale back-transforms. The helper
supports continuous densities and point-mass priors in the same plot.
}
\seealso{
\code{\link[=plot_prior_list]{plot_prior_list()}} \code{\link[=transform_scale_samples]{transform_scale_samples()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tools.R
\name{check_input}
\alias{check_input}
\alias{check_bool}
\alias{check_char}
\alias{check_real}
\alias{check_int}
\alias{check_list}
\title{Check input}
\usage{
check_bool(
  x,
  name = deparse(substitute(x)),
  check_length = 1,
  allow_NULL = FALSE,
  allow_NA = TRUE,
  call = ""
)

check_char(
  x,
  name = deparse(substitute(x)),
  check_length = 1,
  allow_values = NULL,
  allow_NULL = FALSE,
  allow_NA = TRUE,
  call = ""
)

check_real(
  x,
  name = deparse(substitute(x)),
  lower = -Inf,
  upper = Inf,
  allow_bound = TRUE,
  check_length = 1,
  allow_NULL = FALSE,
  allow_NA = TRUE,
  call = ""
)

check_int(
  x,
  name = deparse(substitute(x)),
  lower = -Inf,
  upper = Inf,
  allow_bound = TRUE,
  check_length = 1,
  allow_NULL = FALSE,
  allow_NA = TRUE,
  call = ""
)

check_list(
  x,
  name = deparse(substitute(x)),
  check_length = 0,
  check_names = NULL,
  all_objects = FALSE,
  allow_other = FALSE,
  allow_NULL = FALSE,
  call = ""
)
}
\arguments{
\item{x}{object to be checked}

\item{name}{name of the object that will be print in the error
message.}

\item{check_length}{length of the object to be checked. Defaults to
\code{1}. Set to \code{0} in order to not check object length.}

\item{allow_NULL}{whether the object can be \code{NULL}.
If so, no checks are executed.}

\item{allow_NA}{whether the object can contain \code{NA} or \code{NaN}
values.}

\item{call}{string to be placed as a prefix to the error call.}

\item{allow_values}{names of values allowed in a character vector.
Defaults to \code{NULL} (do not check).}

\item{lower}{lower bound of allowed values.
Defaults to \code{-Inf} (do not check).}

\item{upper}{upper bound of allowed values.
Defaults to \code{Inf} (do not check).}

\item{allow_bound}{whether the values at the boundary are allowed.
Defaults to \code{TRUE}.}

\item{check_names}{names of entries allowed in a list. Defaults to
\code{NULL} (do not check).}

\item{all_objects}{whether all entries in \code{check_names} must be
present. Defaults to \code{FALSE}.}

\item{allow_other}{whether additional entries then the specified in
\code{check_names} might be present}
}
\value{
returns \code{NULL}, called for the input check.
}
\description{
A set of convenience functions for checking
objects/arguments to a function passed by a user.
}
\examples{
# check whether the object is logical
check_bool(TRUE, name = "input")

# will throw an error on any other type
\dontrun{
  check_bool("TRUE", name = "input")
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/priors-print.R
\name{print.prior}
\alias{print.prior}
\title{Prints a prior object}
\usage{
\method{print}{prior}(
  x,
  short_name = FALSE,
  parameter_names = FALSE,
  plot = FALSE,
  digits_estimates = 2,
  silent = FALSE,
  ...
)
}
\arguments{
\item{x}{a prior}

\item{short_name}{whether prior distribution names should be
shorted}

\item{parameter_names}{whether parameter names should be
printed}

\item{plot}{to return \link[base]{bquote} formatted
prior name for plotting.}

\item{digits_estimates}{number of decimals to be displayed
for printed parameters.}

\item{silent}{to silently return the print message.}

\item{...}{additional arguments}
}
\value{
\code{print.prior} invisibly returns the print statement.
}
\description{
Prints a prior object
}
\examples{
# create some prior distributions
p0 <- prior(distribution = "point",  parameters = list(location = 0))
p1 <- prior(distribution = "normal", parameters = list(mean = 0, sd = 1))

# print them
p0
p1

# use short names
print(p1, short_name = TRUE)

# print parameter names
print(p1, parameter_names = TRUE)

# generate bquote plotting syntax
plot(0, main = print(p1, plot = TRUE))

}
\seealso{
\code{\link[=prior]{prior()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/priors.R
\name{prior_spike_and_slab}
\alias{prior_spike_and_slab}
\title{Creates a spike and slab prior distribution}
\usage{
prior_spike_and_slab(
  prior_parameter,
  prior_inclusion = prior(distribution = "spike", parameters = list(location = 0.5)),
  prior_weights = 1
)
}
\arguments{
\item{prior_parameter}{a prior distribution for the parameter}

\item{prior_inclusion}{a prior distribution for the inclusion probability. The
inclusion probability must be bounded within 0 and 1 range. Defaults to
\code{prior("spike", parameters = list(location = 0.5))} which corresponds to 1/2
prior probability of including the slab prior distribution (but other prior
distributions, like beta etc can be also specified).}

\item{prior_weights}{prior odds associated with a given distribution.
The value is passed into the model fitting function, which creates models
corresponding to all combinations of prior distributions for each of
the model parameters and sets the model priors odds to the product
of its prior distributions.}
}
\value{
return an object of class 'prior'.
}
\description{
\code{prior_spike_and_slab} creates a spike and slab prior
distribution corresponding to the specification in
\insertCite{kuo1998variable;textual}{BayesTools} (see
\insertCite{ohara2009review;textual}{BayesTools} for further details). I.e.,
a prior distribution is multiplied by an independent indicator with values
either zero or one.
}
\examples{
# create a spike and slab prior distribution
p1 <- prior_spike_and_slab(
   prior(distribution = "normal", parameters = list(mean = 0, sd = 1)),
   prior_inclusion = prior(distribution = "beta", parameters = list(alpha = 1, beta = 1))
)

}
\seealso{
\code{\link[=prior]{prior()}}
}
